/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.config;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Properties;

import org.ourgrid.common.util.JavaFileUtil;

/**
 * Description: This class provides access to the ourgrid's configuration model
 * Each module that needs specific configuration properties should extend this
 * class and sets its own properties.
 */
public abstract class Configuration {

	public static final String VERSION = "3.3.2";

	protected static Configuration singleInstance = null;

	public static final String MYGRID = MyGridConfiguration.class.getName();

	public static final String PEER = PeerConfiguration.class.getName();

	public static final String COREPEER = CorePeerConfiguration.class.getName();

	public static final String USERAGENT = UserAgentConfiguration.class.getName();

	// Properties
	private Properties currentProperties = System.getProperties();

	private boolean usingDefaultProperties = false;

	public static final String PROP_LOGFILE = "logfile";

	public static final String PROP_LOG_PROPERTIES_FILE = "logProperties";

	public static final String PROP_NAME = "name";

	public static final String PROP_EXTERNAL_NAME = "externalname";

	public static final String PROP_PORT = "port";

	public static final String PROP_STARTED_TIME = "started_time";

	// The mygrid/ourgrid root dir.
	public static final String MGROOT = "MGROOT";

	// Security options
	public static final String PROP_SECURITY = "security";

	public static final String PROP_SECURE_PORT = "secureport";

	public static final String PROP_KEYSTORE_FILENAME = "keystorefilename";

	public static final String PROP_KEYSTORE_PASSWORD = "keystorepassword";

	public static final String PROP_TRUSTSTORE_FILENAME = "truststorefilename";

	public static final String PROP_TRUSTSTORE_PASSWORD = "truststorepassword";

	public static final int QUERY_FREQ = 5 * 60 * 1000; // 5 minutes


	public synchronized static final Configuration getInstance() {

		if ( Configuration.singleInstance == null ) {

			throw new Error( "You need to create a configuration!" );

		}
		return Configuration.singleInstance;
	}


	/**
	 * Creates, if not already created, the unique configuration instance.
	 * 
	 * @param configurationType The configuration type to be created.
	 * @return The unique configuration instance.
	 */
	public synchronized static final Configuration getInstance( String configurationType ) {

		if ( Configuration.singleInstance == null ) {
			Configuration.singleInstance = Configuration.getConfigurationInstance( configurationType );
		}

		return Configuration.singleInstance;

	}


	/**
	 * Sets the singleton instance to null. Used only by test classes.
	 */
	public synchronized static final void reset() {

		Configuration.singleInstance = null;

	}


	/**
	 * Gets the value of a given property.
	 * 
	 * @param key The property name without the module specific prefix.
	 * @return The property value or null if the property was no setted.
	 */
	public final String getProperty( String key ) {

		if ( key.indexOf( getPrefix() ) != -1 ) {
			throw new RuntimeException( "You should use Configuration constants as property names!" );
		}

		String prop = (String) currentProperties.get( getPrefix() + key.trim() );

		if ( prop == null ) {
			prop = (String) currentProperties.get( key.trim() );
		}

		if ( prop != null ) {
			prop = prop.trim();
		}

		return prop;

	}


	public boolean isSecurityEnabled() throws ConfigException {

		String propSecurePort = (String) currentProperties.get( getPrefix() + PROP_SECURE_PORT );
		String propKeyPwd = (String) currentProperties.get( getPrefix() + PROP_KEYSTORE_PASSWORD );
		String propTrustFile = (String) currentProperties.get( getPrefix() + PROP_TRUSTSTORE_FILENAME );
		String propTrustPwd = (String) currentProperties.get( getPrefix() + PROP_TRUSTSTORE_PASSWORD );

		if ( isEnabled( PROP_SECURITY ) ) {
			if ( propSecurePort != null && propKeyPwd != null && propTrustFile != null && propTrustPwd != null ) {
				return true;
			}
			throw new ConfigException( "To enable security all the requeried options must be set:\n" + getPrefix()
					+ PROP_SECURITY + "\n" + getPrefix() + PROP_SECURE_PORT + "\n" + getPrefix()
					+ PROP_KEYSTORE_PASSWORD + "\n" + getPrefix() + PROP_TRUSTSTORE_FILENAME + "\n" + getPrefix()
					+ PROP_TRUSTSTORE_PASSWORD );
		}
		return false;

	}


	/**
	 * Sets the value of a given property.
	 * 
	 * @param key The property name without the module specific prefix.
	 * @param value The property value.
	 */
	public final void setProperty( String key, String value ) {

		if ( key.indexOf( getPrefix() ) != -1 ) {
			throw new RuntimeException( "You should use Configuration constants as property names!" );
		}

		currentProperties.put( getPrefix() + key.trim(), value.trim() );
	}


	/**
	 * Verify if the property passed is enabled
	 * 
	 * @param propertyKey the key name of the property.
	 * @return true if property is yes, false otherwise.
	 */
	public final boolean isEnabled( String propertyKey ) {

		String property = getProperty( propertyKey );
		return property != null && property.equalsIgnoreCase( "yes" );

	}


	/**
	 * This method reads properties from a file.
	 * 
	 * @param file The property file to be loaded.
	 */
	protected void loadPropertiesFromFile( String file ) {

		/** Get an abstraction for the properties file */
		File propertiesFile = new File( file );

		/* load the properties file, if it exists */
		try {
			currentProperties.load( new FileInputStream( propertiesFile ) );
		} catch ( IOException e ) {
			usingDefaultProperties = true;
		}

	}


	public boolean isUsingDefaultProperties() {

		return usingDefaultProperties;
	}


	/**
	 * Gets an instance of a <code>Configuration</code> subclass by reflection
	 * and throws an <code>ConfigurationInstantiationRunTimeException</code>
	 * if any error occurs.
	 * 
	 * @param configurationClassName The fully qualified
	 *        <code>Configuration</code> subclass name.
	 * @return An <code>Configuration</code> subclass instance.
	 */
	private static Configuration getConfigurationInstance( String configurationClassName ) {

		try {
			Class configurationClassDefinition = Class.forName( configurationClassName );
			Constructor constructor = configurationClassDefinition.getConstructor( new Class[ 0 ] );
			return (Configuration) constructor.newInstance( new Object[ 0 ] );
		} catch ( Exception e ) {
			throw new ConfigurationInstantiationRunTimeException();
		}
	}


	/**
	 * @return local host name.
	 */
	public final String getHostname() {

		String hostname = "localhost";

		try {
			hostname = InetAddress.getLocalHost().getCanonicalHostName();
		} catch ( UnknownHostException e ) {
			// Using default peer name.
		}

		return hostname;

	}


	/**
	 * This method provides the absolute path to the log file.
	 * 
	 * @return The absolute path where the log file is located
	 */

	public String getLogPath() {

		String logfile = getProperty( PROP_LOGFILE );
		if ( JavaFileUtil.isAbsolutePath( logfile ) ) {
			return (new File( logfile )).getAbsolutePath();
		}
		return getRootDir() + File.separator + getProperty( PROP_LOGFILE );
	}


	public String getLogPropertiesPath() {

		String logfile = getProperty( PROP_LOG_PROPERTIES_FILE );
		if ( JavaFileUtil.isAbsolutePath( logfile ) ) {
			return (new File( logfile )).getAbsolutePath();
		}
		return getRootDir() + File.separator + getProperty( PROP_LOG_PROPERTIES_FILE );
	}


	/**
	 * This method provides the absolute path to the log file.
	 * 
	 * @return The absolute path where the log file is located
	 */

	public final String getXMLLogPath() {

		return getLogPath() + ".xml";
	}


	/**
	 * Returns the module specific prefix used to indentify it's own properties.
	 * For example, the MyGrid properties prefix is "mg." , the OurGrid's one is
	 * "peer." .
	 * 
	 * @return The module specific property prefix.
	 */
	protected abstract String getPrefix();


	/**
	 * Gets the module configuration rootdir.
	 * 
	 * @return The module configuration rootdir path.
	 */
	public abstract String getRootDir();


	/**
	 * Creates the unique configuration instance.
	 * 
	 * @param configurationType The configuration type to be created.
	 * @return The unique configuration instance.
	 */
	public static Configuration createInstance( String configurationType ) {

		reset();
		return getInstance( configurationType );
	}
}
