/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.config;

import java.io.FileInputStream;
import java.io.InputStream;
import java.util.Properties;

import org.apache.log4j.PropertyConfigurator;

/**
 * This class configures Log4j to store log messages.
 */
public class LogConfiguration {

	/** This key-property is in log.properties */
	private static final String LOG4J_FILE_KEY = "log4j.appender.ourgrid.file";

	/** This key-property is in log.properties */
	private static final String LOG4J_FILE_XML_KEY = "log4j.appender.ourgridxml.file";


	/**
	 * It loads log4j properties file.
	 * 
	 * @param propFilePath The path for log4j properties file to be loaded.
	 * @param defaultLogPath If the filePath property in the log4j properties
	 *        file is empty, this parameter defines a default file path to write
	 *        log files.
	 */
	public static void configure( String propFilePath, String defaultLogPath ) {

		Properties prop = new Properties();
		InputStream input;
		try {

			input = new FileInputStream( propFilePath );
			prop.load( input );

			verifyLogFilePath( prop, LOG4J_FILE_KEY, defaultLogPath );

			PropertyConfigurator.configure( prop );

		} catch ( Exception e ) {
			// Nothing to do
		}

	}


	/**
	 * It loads log4j properties file. Note that you must use this method only
	 * if the Xml Appender is defined in the respective properties file.
	 * 
	 * @param propFilePath The path for log4j properties file to be loaded.
	 * @param defaultLogPath If the filePath property in the log4j properties
	 *        file is empty, this parameter defines a default file path to write
	 *        log files.
	 * @param defaultXMLLogPath If the XmlFilePath property in the log4j
	 *        properties file is empty, this parameter defines a default XML
	 *        file path to write log files.
	 */
	public static void configure( String propFilePath, String defaultLogPath, String defaultXMLLogPath ) {

		Properties prop = new Properties();
		InputStream input;
		try {

			input = new FileInputStream( propFilePath );
			prop.load( input );

			verifyLogFilePath( prop, LOG4J_FILE_XML_KEY, defaultXMLLogPath );
			verifyLogFilePath( prop, LOG4J_FILE_KEY, defaultLogPath );

			PropertyConfigurator.configure( prop );

		} catch ( Exception e ) {
			// Nothing to do
		}

	}


	/**
	 * Verify if the property that represents the file to log, in respective
	 * log.properties, is empty. If it is, this property is replaced by the
	 * defaultLogPath parameter.
	 * 
	 * @param prop The <code>Properties</code> where log.properties is loaded.
	 * @param propKey The name of the property that must be analized if it is
	 *        empty.
	 * @param defaultLogPath If the respective filepath property is empty in the
	 *        respective log.properties, it is replaced by this default log path
	 *        (parameter).
	 */
	private static void verifyLogFilePath( Properties prop, String propKey, String defaultLogPath ) {

		String pathValue = prop.getProperty( propKey );
		if ( (pathValue == null) || pathValue.trim().equals( "" ) ) {
			prop.setProperty( propKey, defaultLogPath );
		}

	}

}
