/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.event;

import java.util.LinkedList;

/**
 * This class is desired to queue events in a synchronized way.
 * 
 * @param <T>
 */
public class EventQueue<T extends Event> {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( EventQueue.class );

	/**
	 * The queued events.
	 */
	private LinkedList<T> eventQueue;


	/**
	 * The constructor.
	 */
	public EventQueue() {

		super();
		eventQueue = new LinkedList<T>();

	}


	/**
	 * Adds an event to the end of the queue.
	 * 
	 * @param event The event to be queued
	 */
	public synchronized void put( T event ) {

		eventQueue.addLast( event );
		notify();
	}


	/**
	 * Removes the first event from the queue.
	 * 
	 * @return The first event if exists, otherwise returns <code>null</code>
	 */
	public synchronized T unblockingRemove() {

		if ( eventQueue.size() > 0 ) {
			return eventQueue.removeFirst();
		}
		return null;
	}


	/**
	 * Removes the first event of the queue. If there is no event in the queue,
	 * this method blocks until an event is queued.
	 * 
	 * @return The first event
	 */
	public synchronized T blockingRemove() {

		try {
			while ( eventQueue.size() == 0 ) {
				wait();
			}
		} catch ( InterruptedException e ) {

			LOG.error( "", e );
		}
		return eventQueue.removeFirst();
	}


	/**
	 * The queue size.
	 * 
	 * @return The size
	 */
	public synchronized int size() {

		return eventQueue.size();
	}

}
