/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.executor;

import java.io.File;
import java.io.Serializable;
import java.util.Map;

/**
 * This interface defines the contract between platform dependent layer and the
 * upper layers. The idea is to provide a clean interface that is a minimum set
 * of features to perform native operations.
 */
public interface Executor extends Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	static final long serialVersionUID = 33L;


	/**
	 * Execute a local command in a specific dirName, with the environment
	 * defined by envVars
	 * 
	 * @param dirName directory to execute the script
	 * @param command command to be executed
	 * @param envVars environment variables used by this command
	 * @return a handle that identifies this execution
	 * @throws ExecutorException when there is a problem in the execution.
	 */
	public ExecutorHandle execute( String dirName, String command, Map envVars ) throws ExecutorException;


	/**
	 * Execute a local command in a specific dirName. This method doens't set
	 * any enviornment variables.
	 * 
	 * @param dirName directory to execute the script
	 * @param command command to be executed
	 * @return a handle that identifies this execution
	 * @throws ExecutorException when there is a problem in the execution.
	 */
	public ExecutorHandle execute( String dirName, String command ) throws ExecutorException;


	/**
	 * Changes the permissions for the indicated file. Permissions are described
	 * using the "rwx" Linux convention. The permissions are changed in a way
	 * that further access from this JVM will see the new permissions.
	 * 
	 * @param file the File object
	 * @param perm the new permission in "rwx" format
	 * @exception ExecutorException error changing the permissions
	 */
	public void chmod( File file, String perm ) throws ExecutorException;


	/**
	 * Kills command that was issued via an execute method
	 * 
	 * @param handle the command handle
	 * @throws ExecutorException when there is a problem while changing .the
	 *         permissions of a file.
	 */
	public void kill( ExecutorHandle handle ) throws ExecutorException;


	/**
	 * Blocks until the command finishes and returns its result
	 * 
	 * @param handle the command handle
	 * @return ExecutorResult StdOut, StdErr and exitValue from kill command.
	 * @throws ExecutorException when there is a problem while obtaining the
	 *         result of an execution.
	 */
	public ExecutorResult getResult( ExecutorHandle handle ) throws ExecutorException;
}
