/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.executor;

import java.io.Serializable;

/**
 * This class represents the result of an execution.
 */
public class ExecutorResult implements Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/**
	 * The exit value of an execution.
	 */
	private int exitValue;

	/**
	 * The standard output of an exection.
	 */
	private String stdout;

	/**
	 * The standard error output.
	 */
	private String stderr;


	/**
	 * Returns a textual representation of an <code>ExecutorResult</code>.
	 */
	public String toString() {

		StringBuffer str = new StringBuffer();
		str.append( "ExitValue: " );
		str.append( exitValue );
		str.append( " StdOut: " );
		str.append( stdout );
		str.append( " StdErr: " );
		str.append( stderr );
		return str.toString();
	}


	/**
	 * Returns the exit value of the process
	 * 
	 * @return The exit value
	 */
	public int getExitValue() {

		return exitValue;
	}


	/**
	 * Returns the error of the process
	 * 
	 * @return The error
	 */
	public String getStderr() {

		return stderr;
	}


	/**
	 * Returns the standard output of the process
	 * 
	 * @return The standard output
	 */
	public String getStdout() {

		return stdout;
	}


	/**
	 * Sets the exit value of the process.
	 * 
	 * @param exitValue The exit value of the process
	 */
	public void setExitValue( int exitValue ) {

		this.exitValue = exitValue;
	}


	/**
	 * Set the error of the process.
	 * 
	 * @param stderr The standard err of the process
	 */
	public void setStderr( String stderr ) {

		this.stderr = stderr;
	}


	/**
	 * Set the standard output of the process.
	 * 
	 * @param stdout The standard output of the process
	 */
	public void setStdout( String stdout ) {

		this.stdout = stdout;
	}


	/**
	 * Two executor results are equal if their stderr, stdout and exitvalue are
	 * equal.
	 * 
	 * @return true If the two executor results are equals. false, otherwise.
	 */
	public boolean equals( Object obj ) {

		if ( !(obj instanceof ExecutorResult) ) {
			return false;
		}
		ExecutorResult toCompare = (ExecutorResult) obj;
		boolean isEqual = true;

		if ( (this.stderr == null) || (toCompare.stderr == null) ) {
			isEqual = (isEqual && toCompare.stderr == this.stderr);
		} else {
			isEqual = (isEqual && toCompare.stderr.equals( this.stderr ));
		}

		if ( (this.stdout == null) || (toCompare.stdout == null) ) {
			isEqual = (isEqual && toCompare.stdout == this.stdout);
		} else {
			isEqual = (isEqual && toCompare.stdout.equals( this.stdout ));
		}
		isEqual = (isEqual && toCompare.exitValue == this.exitValue);

		return isEqual;
	}

}
