/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.executor;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.test.ReplicaExecutorFastTest;
import org.ourgrid.test.TestUtils;

/**
 * This class is the test suite for the LinuxExecutor class.
 */
public class LinuxExecutorTest extends TestCase implements ReplicaExecutorFastTest {

	/** An instance of LinuxExecutor to be tested */
	private LinuxExecutor executor;

	/** A file used to test the LinuxExecutor class. */
	private File file;

	/** The system temporary directory. */
	private File tmpDir;

	private TestUtils testUtils;

	private File tmpFile;


	/**
	 * This method constructs a new instance of LinuxExecutor class.
	 * 
	 * @param name The name of class file that corresponds to this test suite.
	 */
	public LinuxExecutorTest( String name ) {

		super( name );
		testUtils = new TestUtils();
		testUtils.load();
		tmpDir = new File( testUtils.getProperty( TestUtils.SYSTEM_TEMP_DIR ) );
		tmpFile = new File( testUtils.getProperty( TestUtils.FAKE_FILE ) );
	}


	/**  */
	protected void setUp() throws Exception {

		super.setUp();
		Configuration.getInstance( Configuration.MYGRID );
		createTempFile();
		executor = new LinuxExecutor();

		File mypids = new File( ".mypids" );
		if ( mypids.exists() ) {
			mypids.delete();
		}
	}


	/**
	 *  
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		Configuration.reset();
		removeTempFile();
	}


	/**
	 * This method test if an expected correct execution will be performed well
	 * indeed.
	 */
	public void testGoodExecute() throws Exception {

		/* A very simple ls command. There is no mistery here! */
		ExecutorHandle handle = executor.execute( tmpDir.getAbsolutePath(), "ls -la" );

		/* Get the result */
		ExecutorResult result = executor.getResult( handle );

		assertEquals( "the command must be successful", 0, result.getExitValue() );
		assertTrue( "if the command was successful, there is something in the standard output", !result.getStdout()
				.equals( "" ) );
	}


	/** This method tests the verification of a bad execution of native command. */
	public void testBadExecute() throws Exception {

		/* Execute an invalid command. */
		ExecutorHandle handle = executor.execute( tmpDir.getAbsolutePath(), tmpFile.getName() );

		/* Get the result of execution */
		ExecutorResult result = executor.getResult( handle );

		assertEquals( "The command must not be successful (Command not found <127>)", result.getExitValue(), 127 );
		assertTrue( "If the command was successful, there is some chars in the error output", !result.getStderr()
				.equals( "" ) );
	}


	/**
	 * This method tests the execution of native commands that are dependent of
	 * environment variables.
	 */
	public void testExecuteWithVariables() throws Exception {

		/* The handle for a specific execution */
		ExecutorHandle handle = null;

		/*
		 * The result of a specific execution that is get by using the
		 * associated handle
		 */
		ExecutorResult result = null;

		/* The enviroment variables used by the command */
		Map<String,String> envVars = new HashMap<String,String>();

		/* Populate the enviroment variables Map. */
		envVars.put( "TEST", "test" );
		envVars.put( "TEST2", "test 2" );

		/* Execute the command and get a handle */
		handle = executor.execute( tmpDir.getAbsolutePath(), "echo $TEST", envVars );

		/* Get the result */
		result = executor.getResult( handle );

		System.out.println( result.toString() );

		assertTrue( "the command must be successful", result.getExitValue() == 0 );
		assertEquals( "the command must return TEST value", result.getStdout().trim(), "test" );

		/* Execute other command and get a new handle */
		handle = executor.execute( tmpDir.getAbsolutePath(), "echo $TEST2", envVars );

		/* Get the result */
		result = executor.getResult( handle );

		assertTrue( "the command must be successful", result.getExitValue() == 0 );
		assertEquals( "the command must return TEST2 value", result.getStdout().trim(), "test 2" );
	}


	/**
	 * This method tests the execution of native commands that are dependent of
	 * environment variables.
	 */
	public void testTwoExecuteInvocations() throws Exception {

		/* The handle for a specific execution */
		ExecutorHandle handle1 = null;
		ExecutorHandle handle2 = null;

		/*
		 * The result of a specific execution that is get by using the
		 * associated handle
		 */
		ExecutorResult result1 = null;
		ExecutorResult result2 = null;

		/* The enviroment variables used by the command */
		Map<String,String> envVars = new HashMap<String,String>();

		/* Populate the enviroment variables Map. */
		envVars.put( "TEST", "test" );
		envVars.put( "TEST2", "test 2" );

		/* Execute the 1st command and get a handle */
		handle1 = executor.execute( tmpDir.getAbsolutePath(),
				"I=1; while [ $I -le 5 ]; do sleep 1; echo $TEST $((I++)); done;", envVars );

		/* Execute the 2nd command and get a handle */
		handle2 = executor.execute( tmpDir.getAbsolutePath(), tmpFile.getName() + " $TEST2", envVars );

		/* Get the second result */
		result2 = executor.getResult( handle2 );

		assertEquals( "The command must not be found for execution", result2.getExitValue(), 127 );
		assertTrue( "It must have something in the std error output", !result2.getStderr().trim().equals( "" ) );

		/* Get the first result */
		result1 = executor.getResult( handle1 );

		assertTrue( "A lot of chars must be printed out in the std output", !result1.getStdout().trim().equals( "" ) );
		assertEquals( "The command must perform well", result1.getExitValue(), 0 );

	}


	/** Test an execution that uses environment variables but these are not set */
	public void testExecuteWithoutVariables() throws Exception {

		/* The handle */
		ExecutorHandle handle = null;

		/* The result */
		ExecutorResult result = null;

		/* Execute and got a handle */
		handle = executor.execute( tmpDir.getAbsolutePath(), "echo $TASK" );

		/* Get the execution result */
		result = executor.getResult( handle );

		System.out.println( result.toString() );

		assertTrue( "the command must be successful", result.getExitValue() == 0 );
	}


	/** Test if a permission denied is detected during the execution */
	public void testNotPermitedLocalExecute() throws Exception {

		ExecutorHandle handle = null;
		ExecutorResult result = null;
		handle = executor.execute( testUtils.getProperty( TestUtils.PROTECTED_DIR ), "ls" );

		result = executor.getResult( handle );

		System.out.println( result );

		assertTrue( "The execution should be denied!", result.getStderr().indexOf( "Permis" ) != -1 );

	}


	/** Test the handle management */
	public void testHandleManagement() throws Exception {

		/* 1st handle and executor result */
		ExecutorHandle handle1 = null;

		/* 2nd handle and executor result */
		ExecutorHandle handle2 = null;

		handle1 = executor.execute( ".", "ls" );
		handle2 = executor.execute( ".", "ls" );

		executor.getResult( handle1 );
		executor.getResult( handle2 );

		assertTrue( "The handles should be different!", handle1 != handle2 );
	}


	/**
	 * This method is a utility method just to be used by this test.
	 * 
	 * @return <b>true </b> if the file is created successfully. <b>false </b>
	 *         otherwise.
	 * @throws Exception if some problem occurs during the IO operations.
	 */
	private boolean createTempFile() throws Exception {

		/* The abstract representaion for the temp file */
		file = new File( tmpDir.getAbsolutePath() + "/executorTestFile" );

		/* create the file and return the result flag */
		return file.createNewFile();
	}


	/**
	 * This method just remove the temp file created by the createTempFile()
	 * method.
	 * 
	 * @return <b>true </b> if the file is removed successfully. <b>false </b>
	 *         otherwise.
	 * @throws Exception if some problem occurs during the IO operations
	 */
	private boolean removeTempFile() throws Exception {

		return file.delete();
	}


	/**
	 * This method just get the file permissions.
	 * 
	 * @param file The abstract representation for the file will have the
	 *        permissions determined
	 * @return The string representation "rwx" of file permissions
	 * @throws Exception if some problem occurs during the IO or execution
	 *         operation.
	 */
	private String getFilePermissions( File file ) throws Exception {

		/* Execute and get the handle */
		ExecutorHandle handle = executor.execute( ".", "ls -l " + file.getAbsolutePath() );

		/* get the result of execution */
		ExecutorResult result = executor.getResult( handle );

		/* parse the result */
		return result.getStdout().substring( 1, 10 );
	}


	/**
	 * This method test the modification of permissions feature.
	 * 
	 * @throws Exception
	 */
	public void testChmod1() throws Exception {

		/** Try to change the permissions of a file */
		executor.chmod( file, "0" );

		String permissions = getFilePermissions( file );

		assertTrue( "Permission's results 0", permissions.indexOf( "---------" ) != -1 );
	}


	public void testChmod2() throws Exception {

		executor.chmod( file, "1" );
		String permissions = getFilePermissions( file );
		assertTrue( "Permission's results 1", permissions.indexOf( "--------x" ) != -1 );
	}


	public void testChmod3() throws Exception {

		executor.chmod( file, "123" );
		String permissions = getFilePermissions( file );
		assertTrue( "Permission's results 123", permissions.indexOf( "--x-w--wx" ) != -1 );

	}


	public void testChmod4() throws Exception {

		executor.chmod( file, "405" );
		String permissions = getFilePermissions( file );
		assertEquals( "Permission's results 405", "r-----r-x", permissions );

	}


	public void testChmod5() throws Exception {

		executor.chmod( file, "471" );
		String permissions = getFilePermissions( file );
		assertEquals( "Permission's results 471", "r--rwx--x", permissions );

	}


	public void testChmod6() throws Exception {

		executor.chmod( file, "724" );
		String permissions = getFilePermissions( file );
		assertEquals( "Permission's results 724", "rwx-w-r--", permissions );

	}


	public void testChmod7() throws Exception {

		executor.chmod( file, "352" );
		String permissions = getFilePermissions( file );
		assertEquals( "Permission's results 352", "-wxr-x-w-", permissions );

	}


	public void testChmod8() throws Exception {

		executor.chmod( file, "41" );
		String permissions = getFilePermissions( file );
		assertEquals( "Permission's results 41", "---r----x", permissions );

	}


	public void testChmod9() throws Exception {

		executor.chmod( file, "707" );
		String permissions = getFilePermissions( file );
		assertEquals( "Permission's results 707", "rwx---rwx", permissions );

	}


	public void testChmod10() throws Exception {

		executor.chmod( file, "404" );
		String permissions = getFilePermissions( file );
		assertEquals( "Permission's results 404", "r-----r--", permissions );

	}


	public void testChmod11() throws Exception {

		executor.chmod( file, "525" );
		String permissions = getFilePermissions( file );
		assertEquals( "Permission's results 525", "r-x-w-r-x", permissions );

	}


	public void testChmod12() throws Exception {

		executor.chmod( file, "321" );
		String permissions = getFilePermissions( file );
		assertEquals( "Permission's results 321", "-wx-w---x", permissions );

	}


	public void testChmod13() throws Exception {

		executor.chmod( file, "70" );
		String permissions = getFilePermissions( file );
		assertEquals( "Permission's results 70", "---rwx---", permissions );
	}


	/**
	 * Tests if the chmod feature performs well when receiving a bad permission.
	 * 
	 * @throws Exception
	 */
	public void testChmod14() throws Exception {

		/* Get permissions before trying to change */
		String permissionsBefore = getFilePermissions( file );

		try {
			executor.chmod( file, "T" );
		} catch ( ExecutorException e ) {
			// Expected exception
		}

		String permissionsAfter = getFilePermissions( file );

		assertEquals( "Permission should be the same [" + permissionsBefore + " != " + permissionsAfter + "]",
				permissionsBefore, permissionsAfter );

		permissionsBefore = getFilePermissions( file );

		try {
			executor.chmod( file, "999" );
		} catch ( ExecutorException e ) {
			System.out.println( "EXPECTED EXCEPTION: " + e.getMessage() );
		}

		permissionsAfter = getFilePermissions( file );

		assertEquals( "Permissions should be the same [" + permissionsBefore + " != " + permissionsAfter + "]",
				permissionsBefore, permissionsAfter );
	}


	/** Test the kill operation */
	public void testKill() throws Exception {

		ExecutorResult result;
		ExecutorHandle handle = executor.execute( ".", "sleep 300" );

		executor.kill( handle );

		result = executor.getResult( handle );

		assertTrue( "The exit value must be nor '0' neither '1'.", (result.getExitValue() != 0)
				&& (result.getExitValue() != 1) );

	}


	/** Test the kill operation */
	public void testKillEqualsCommands() throws Exception {

		ExecutorResult result;
		ExecutorHandle handle = executor.execute( ".", "sleep 20" );
		ExecutorHandle otherHandle = executor.execute( ".", "sleep 20" );

		/** Kill the second process */
		executor.kill( otherHandle );

		/** get the result of second process */
		result = executor.getResult( otherHandle );

		assertTrue( "The exit value must be nor '0' neither '1'.", (result.getExitValue() != 0)
				&& (result.getExitValue() != 1) );

		/** wait first command finish */
		result = executor.getResult( handle );

		assertEquals( "The exit value must be '0'.", 0, result.getExitValue() );
	}

}
