/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.executor;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.util.TempFileManager;
import org.ourgrid.test.ReplicaExecutorFastTest;

/**
 * Test for SwanExecutor.
 */
public class SWANExecutorTest extends TestCase implements ReplicaExecutorFastTest {

	private LinuxExecutor linuxExec;

	private SWANExecutor swanExec;

	private File stdOutput;

	private File errorOutput;

	private File exitValue;


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();
		linuxExec = EasyMock.createMock( LinuxExecutor.class );

		stdOutput = TempFileManager.createTempFile( "stdOutput_", ".txt" );
		errorOutput = TempFileManager.createTempFile( "stdError_", ".txt" );
		exitValue = TempFileManager.createTempFile( "exitValue_", ".txt" );
		swanExec = new SWANExecutor( linuxExec, stdOutput, errorOutput, exitValue );
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
	}


	/*
	 * Test method for 'org.ourgrid.common.executor.SWANExecutor.execute(String,
	 * String, Map)'
	 */
	public void testExecuteStringStringMap() throws Exception {

		Map<String,String> envVars = new HashMap<String,String>();
		envVars.put( "USER", "test" );

		ExecutorHandle handle = EasyMock.createMock( ExecutorHandle.class );
		linuxExec.execute( EasyMock.eq( "dir" ), (String) EasyMock.anyObject() );
		EasyMock.expectLastCall().andReturn( handle ).once();

		EasyMock.replay( linuxExec );
		swanExec.execute( "dir", "echo hello", envVars );
		EasyMock.verify( linuxExec );
	}


	/*
	 * Test method for 'org.ourgrid.common.executor.SWANExecutor.chmod(File,
	 * String)'
	 */
	public void testChmod() throws Exception {

		File file = new File( "test" );
		String mod = "0";

		linuxExec.chmod( file, mod );
		EasyMock.expectLastCall().once();

		EasyMock.replay( linuxExec );
		swanExec.chmod( file, mod );
		EasyMock.verify( linuxExec );
	}


	public void testGetResult() throws Exception {

		ExecutorHandle handle = EasyMock.createMock( ExecutorHandle.class );
		assertNull( swanExec.getResult( handle ) );

		Map<String,String> envVars = new HashMap<String,String>();
		envVars.put( "USER", "test" );
		ExecutorResult result = new ExecutorResult();
		result.setExitValue( 0 );
		result.setStdout( "" );
		result.setStderr( "" );

		linuxExec.execute( EasyMock.eq( "dir" ), (String) EasyMock.anyObject() );
		EasyMock.expectLastCall().andReturn( handle ).once();

		linuxExec.getResult( handle );
		EasyMock.expectLastCall().andReturn( result ).once();

		EasyMock.replay( linuxExec );
		swanExec.execute( "dir", "echo hello", envVars );
		assertEquals( result, swanExec.getResult( handle ) );
		EasyMock.verify( linuxExec );
	}


	public void testKill() throws Exception {

		ExecutorHandle handle = EasyMock.createMock( ExecutorHandle.class );
		swanExec.kill( handle );

		Map<String,String> envVars = new HashMap<String,String>();
		envVars.put( "USER", "test" );
		ExecutorResult result = new ExecutorResult();
		result.setExitValue( 0 );
		result.setStdout( "" );
		result.setStderr( "" );

		linuxExec.execute( EasyMock.eq( "dir" ), (String) EasyMock.anyObject() );
		EasyMock.expectLastCall().andReturn( handle ).once();

		linuxExec.kill( handle );
		EasyMock.expectLastCall().once();

		linuxExec.execute( EasyMock.eq( "dir" ), (String) EasyMock.anyObject() );
		EasyMock.expectLastCall().andReturn( EasyMock.createMock( ExecutorHandle.class ) ).once();

		linuxExec.getResult( (ExecutorHandle) EasyMock.anyObject() );
		EasyMock.expectLastCall().andReturn( result ).once();

		EasyMock.replay( linuxExec );
		ExecutorHandle handleResult = swanExec.execute( "dir", "echo hello", envVars );
		swanExec.kill( handleResult );
		assertNull( swanExec.getResult( handle ) );
		EasyMock.verify( linuxExec );
	}


	/*
	 * Test method for
	 * 'org.ourgrid.common.executor.SWANExecutor.createScript(String, String,
	 * Map)'
	 */
	public void testCreateScriptStringStringMap() throws Exception {

		Map<String,String> envVars = new HashMap<String,String>();
		envVars.put( "USER", "test" );
		File script = swanExec.createScript( "echo hello", "/tmp", envVars );
		script.deleteOnExit();
		BufferedReader fileReader = new BufferedReader( new FileReader( script ) );
		String[ ] testScript = { "USER='test'", "export  USER", "PATH=$PATH:$STORAGE:.", "export  USER PATH",
								"echo hello" };
		for ( int i = 0; i < testScript.length; i++ ) {
			assertEquals( testScript[i], fileReader.readLine() );
		}

	}

}
