/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.fd;

import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.security.SecureRandom;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Timer;
import java.util.TreeMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.ourgrid.common.event.Shutdownable;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;

/**
 * This object is responsible for providing an updated information about
 * <code>Monitorable</code>s fails. <br>
 * <br>
 * The rationale behind the <code>abstract</code> is that the
 * <code>Monitorable</code> interface methods will be implemented by the
 * failure detector. We plan three main implementors to this interface:
 * <ul>
 * <li>UserAgentFailureDetector</li>
 * <li>PeerFailureDetector</li>
 * <li>MyGridFailureDetector</li>
 * </ul>
 * <br>
 * <br>
 * P.S.: Remember that the <code>UserAgent</code> has three states. :-)
 * 
 * @see org.ourgrid.common.event.Shutdownable
 * @see org.ourgrid.common.fd.Monitor
 * @see org.ourgrid.common.fd.Monitorable
 */
public abstract class AbstractFailureDetector extends OurgridUnicastRemoteObject implements Shutdownable, Monitor,
		Monitorable, Remote {

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( AbstractFailureDetector.class );

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	public static final long serialVersionUID = 33L;

	/**
	 * The name to be used to bound the remote service.
	 */
	public static final String BOUND_NAME = "FAILURE_DETECTOR";

	public static final String DEFAULT_PROTOCOL = "rmi";

	/**
	 * This constant indicates the query frequency of active grid machines state
	 * verification (in milliseconds).
	 */
	private int queryFrequency;

	private static final int DEFAULT_QUERY_FREQUENCY = 10 * 60 * 1000; // 10 minutes

	// minutes

	private Map<String,MonitoringUnit> monitoringUnits;

	Timer pingerTimerTask;

	private double encarnationNumber;

	private MonitoringUnitFactory monitoringUnitFactory;

	private ExecutorService threadpool;


	/**
	 * Initializes an <code>AbstractFailureDetector</code> with
	 * <code>AbstractFailureDetector#DEFAULT_QUERY_FREQUENCY</code>
	 * 
	 * @throws RemoteException If something wrong occurs at the initialization
	 *         of the <code>OurgridUnicastRemoteObject</code>
	 */
	public AbstractFailureDetector() throws RemoteException {

		this( DEFAULT_QUERY_FREQUENCY );
	}


	public AbstractFailureDetector( MonitoringUnitFactory factory ) throws RemoteException {

		this( DEFAULT_QUERY_FREQUENCY, factory );
	}


	/**
	 * Initializes an <code>AbstractFailureDetector</code>
	 * 
	 * @param queryFrequency The frequency when the state of
	 *        <code>Monitorable</code> objects will me verified.
	 * @throws RemoteException If something wrong occurs at the initialization
	 *         of the <code>OurgridUnicastRemoteObject</code>
	 */
	public AbstractFailureDetector( int queryFrequency ) throws RemoteException {

		this( queryFrequency, new RMIMonitoringUnitFactory() );
	}


	public AbstractFailureDetector( int queryFrequency, MonitoringUnitFactory factory ) throws RemoteException {

		super();
		this.queryFrequency = queryFrequency;
		this.monitoringUnits = new TreeMap<String,MonitoringUnit>();

		this.threadpool = new ThreadPoolExecutor( 0, 100, 600L, TimeUnit.SECONDS, new SynchronousQueue<Runnable>() );

		generateEncarnationNumber();
		this.monitoringUnitFactory = factory;
	}


	public void register( Notifiable notifiable, String monitorableAddress, boolean imediatePing )
		throws RemoteException, MalformedURLException {

		register( notifiable, monitorableAddress );
		if ( imediatePing ) {
			pingNow( monitorableAddress );
		}

	}


	public void pingNow( String monitorableAddress ) throws RemoteException {

		MonitoringUnit monitoringUnit = monitoringUnits.get( monitorableAddress );
		if ( monitoringUnit == null ) {
			throw new RemoteException( "Monitorable " + monitorableAddress + " does not exist" );
		}

		MachinesPingerRunnable pinger = new MachinesPingerRunnable( monitoringUnit );
		threadpool.execute( pinger );
	}


	/**
	 * @throws MalformedURLException If has something wrong with monitorable
	 *         URL.
	 * @see org.ourgrid.common.fd.Monitor#register(org.ourgrid.common.fd.Notifiable,
	 *      String)
	 */
	public void register( Notifiable notifiable, String monitorableAddress ) throws RemoteException,
		MalformedURLException {

		// Local calls
		String notifiableName = notifiable.getName();
		MonitoringUnit monitoringUnit = monitoringUnits.get( monitorableAddress );

		if ( monitoringUnit == null ) {
			// First time
			// See that the address given to the monitoring unit is the failure
			// detector address.
			monitoringUnit = this.monitoringUnitFactory.createMonitoringUnit( new GenericState( State.DOWN ),
					getFailureDetectorAddress( monitorableAddress ), monitorableAddress );
			synchronized ( monitoringUnits ) {
				monitoringUnits.put( monitorableAddress, monitoringUnit );
			}
		}
		monitoringUnit.addInterested( notifiableName, notifiable );
	}


	/**
	 * @see org.ourgrid.common.fd.Monitor#unregister(org.ourgrid.common.fd.Notifiable)
	 */
	public void unregister( Notifiable notifiable ) throws RemoteException {

		String notifiableName = notifiable.getName();
		Collection<MonitoringUnit> monitoringUnit = monitoringUnits.values();

		for ( Iterator<MonitoringUnit> iter = monitoringUnit.iterator(); iter.hasNext(); ) {

			MonitoringUnit unit = iter.next();
			unit.removeInterested( notifiableName );

			if ( !unit.hasInteresteds() ) {
				iter.remove();
			}
		}
	}


	/**
	 * @see org.ourgrid.common.fd.Monitor#unregister(org.ourgrid.common.fd.Notifiable,
	 *      String)
	 */
	public void unregister( Notifiable notifiable, String monitorableAddress ) throws RemoteException {

		// Local calls
		String notifiableName = notifiable.getName();
		MonitoringUnit monitoringUnit = monitoringUnits.get( monitorableAddress );
		if ( monitoringUnit != null ) {
			monitoringUnit.removeInterested( notifiableName );
			if ( !monitoringUnit.hasInteresteds() ) {
				// If has nobody interested in this monitorable
				synchronized ( monitoringUnits ) {
					monitoringUnits.remove( monitorableAddress );
				}
			}
		}

	}


	/**
	 * The <code>pingerTimerTask</code> is initialized here because in certain
	 * situations (e.g. Sensors) the monitoring will not occurs.
	 * 
	 * @see org.ourgrid.common.fd.Monitor#startMonitoring()
	 */
	public void startMonitoring() throws RemoteException {

		pingerTimerTask = new Timer( true ); // true because is a daemon
		// thread.
		pingerTimerTask.schedule( new MachinesPinger( this, threadpool ), 0, this.queryFrequency );
	}


	/**
	 * @see org.ourgrid.common.fd.Monitor#getMonitorables()
	 */
	public synchronized Map<String,MonitoringUnit> getMonitorables() throws RemoteException {

		return this.monitoringUnits;
	}


	/**
	 * @see org.ourgrid.common.fd.Monitorable#getEncarnationNumber()
	 */
	public double getEncarnationNumber() throws RemoteException {

		return this.encarnationNumber;
	}


	private void generateEncarnationNumber() {

		SecureRandom random = new SecureRandom();
		this.encarnationNumber = random.nextDouble();
	}


	/**
	 * Allows that an external entity updates the state of a monitorable
	 * 
	 * @param monitorableAddress
	 * @param newState
	 */
	public void setMonitorableState( String monitorableAddress, GenericState newState ) {

		LOG.debug( "Receiving advice about state of address: " + monitorableAddress + ", new state: " + newState );

		MonitoringUnit monitoringUnit = monitoringUnits.get( monitorableAddress );

		if ( monitoringUnit != null ) {
			monitoringUnit.changeState( newState );
		}
		
		this.encarnationNumber = 0;
	}


	/**
	 * This method gets the address of the <code>FailureDetector</code>
	 * service bounded at the same machine of the monitorable service given by
	 * parameter.
	 * 
	 * @param monitorableAddress The service the user would like to monitor
	 * @return the address of the <code>FailureDetector</code> service bounded
	 *         at the same machine of the monitorable service given by
	 *         parameter.
	 * @throws MalformedURLException If has something wrong with monitorable
	 *         URL.
	 */
	protected String getFailureDetectorAddress( String monitorableAddress ) throws MalformedURLException {

		URI uri;
		try {
			uri = new URI( monitorableAddress );
		} catch ( URISyntaxException uriSE ) {
			throw new MalformedURLException( uriSE.getMessage() );
		}
		validateURL( uri );

		return uri.getScheme() + "://" + uri.getHost() + ":" + uri.getPort() + "/" + BOUND_NAME;
	}


	private void validateURL( URI uri ) throws MalformedURLException {

		if ( uri.getPath() == null ) {
			throw new MalformedURLException( "Invalid remote object " );
		}
		if ( !uri.getScheme().equals( DEFAULT_PROTOCOL ) ) {
			throw new MalformedURLException( "Invalid protocol" );
		}
		if ( uri.getHost() == null ) {
			throw new MalformedURLException( "Invalid hostname" );
		}
		if ( uri.getPort() == -1 ) {
			throw new MalformedURLException( "Invalid port number" );
		}
	}


	/**
	 * @see org.ourgrid.common.event.Shutdownable#shutdown()
	 */
	public void shutdown() {

		if ( pingerTimerTask != null ) {
			pingerTimerTask.cancel();
			pingerTimerTask.purge();
		}
	}


	/*
	 * Abstract section
	 */
	public abstract GenericState getCurrentState() throws RemoteException;

}
