/*
 * Copyright (c) 2002-2004 Universidade Federal de Campina Grande
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.ourgrid.common.fd;

import java.net.MalformedURLException;
import java.util.Map;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.test.FailureDetectorTest;
import org.ourgrid.test.TestUtils;

/**
 * Test case for <code>AbstractFailureDetector</code>.
 * 
 * @see org.ourgrid.common.fd.AbstractFailureDetector
 */
public class AbstractFailureDetectorTest extends TestCase implements FailureDetectorTest {

	private final String NOTIFIABLE_IDENTIFIER = "notifiable";

	private final double DEFAULT_REENCARNATION_NUMBER = 1d;

	private String DEFAULT_MONITORABLE_ADDRESS = "rmi://monitorable:80/monitorable";

	private String DEFAULT_MONITORABLE_ADDRESS_1 = "rmi://monitorable:80/monitorable1";

	SimpleTestFailureDetector fd;

	Monitorable monitorable;

	Notifiable notifiable;

	FakeMonitoringUnitFactory factory;

	private GenericState downState;

	private GenericState upState;

	private TestUtils testUtils;


	protected void setUp() throws Exception {

		Configuration.getInstance( Configuration.MYGRID );

		testUtils = new TestUtils();
		testUtils.load();

		Configuration.getInstance().setProperty( Configuration.PROP_PORT, Integer.toString( testUtils.getRMIPort() ) );

		monitorable = createMonitorableMock();
		notifiable = createNotifiableMock();
		factory = new FakeMonitoringUnitFactory();
		fd = new SimpleTestFailureDetector( factory );
		downState = new GenericState( State.DOWN );
		upState = new GenericState( State.UP );
	}


	protected void tearDown() throws Exception {

		fd.shutdown();

		Configuration.reset();
		testUtils.reset();
	}


	/**
	 * Verfiying if a monitorable is invoked. <br>
	 * <br>
	 * <br>
	 * 
	 * @throws Exception If anything wrong occurs
	 */
	public void testRegistter() throws Exception {

		monitorable.getCurrentState();
		EasyMock.expectLastCall().andReturn( downState ).once();
		monitorable.getEncarnationNumber();
		EasyMock.expectLastCall().andReturn( DEFAULT_REENCARNATION_NUMBER ).once();

		factory.setMonitorable( DEFAULT_MONITORABLE_ADDRESS, monitorable );

		notifiable.getName();
		EasyMock.expectLastCall().andReturn( NOTIFIABLE_IDENTIFIER ).atLeastOnce();

		EasyMock.replay( monitorable );
		EasyMock.replay( notifiable );

		fd.register( notifiable, DEFAULT_MONITORABLE_ADDRESS );
		assertFalse( fd.getMonitorables().isEmpty() );

		fd.startMonitoring( 1 );

		EasyMock.verify( monitorable );
		EasyMock.verify( notifiable );
	}


	/**
	 * The same test above, but with two <code>Monitorables</code> and one
	 * <code>Notifiable</code>. So, when the notifiable unregister the
	 * interest in one, the other must continue to be monitored.
	 * 
	 * @throws Exception If anything wrong occurs
	 */
	public void testUnregister1() throws Exception {

		monitorable.getCurrentState();
		EasyMock.expectLastCall().andReturn( downState ).once();
		monitorable.getEncarnationNumber();
		EasyMock.expectLastCall().andReturn( DEFAULT_REENCARNATION_NUMBER ).once();

		Monitorable monitorable1 = createMonitorableMock();
		monitorable1.getCurrentState();
		EasyMock.expectLastCall().andReturn( downState ).once();
		monitorable1.getEncarnationNumber();
		EasyMock.expectLastCall().andReturn( DEFAULT_REENCARNATION_NUMBER ).once();

		notifiable.getName();
		EasyMock.expectLastCall().andReturn( NOTIFIABLE_IDENTIFIER ).times( 3 );

		factory.setMonitorable( DEFAULT_MONITORABLE_ADDRESS, monitorable );
		factory.setMonitorable( DEFAULT_MONITORABLE_ADDRESS_1, monitorable1 );

		EasyMock.replay( monitorable );
		EasyMock.replay( monitorable1 );
		EasyMock.replay( notifiable );

		fd.register( notifiable, DEFAULT_MONITORABLE_ADDRESS );
		fd.register( notifiable, DEFAULT_MONITORABLE_ADDRESS_1 );
		fd.startMonitoring( 1 );

		// There is not any more interested in this notifiable
		fd.unregister( notifiable, DEFAULT_MONITORABLE_ADDRESS );

		Map<String,MonitoringUnit> monitoringUnits = fd.getMonitorables();

		assertFalse( monitoringUnits.isEmpty() );
		assertFalse( monitoringUnits.containsKey( DEFAULT_MONITORABLE_ADDRESS ) );
		assertTrue( monitoringUnits.containsKey( DEFAULT_MONITORABLE_ADDRESS_1 ) );

		EasyMock.verify( monitorable );
		EasyMock.verify( monitorable1 );
		EasyMock.verify( notifiable );
	}


	/**
	 * Tests if the last interested in a monitorable unregister, then this
	 * monitorable must not be monitored any more.
	 * 
	 * @throws Exception If anything wrong occurs
	 */
	public void testUnregister2() throws Exception {

		monitorable.getCurrentState();
		EasyMock.expectLastCall().andReturn( downState ).anyTimes(); // Because
		// could
		// not
		// be
		// enough
		// time to call this method
		monitorable.getEncarnationNumber();
		EasyMock.expectLastCall().andReturn( DEFAULT_REENCARNATION_NUMBER ).atLeastOnce();

		notifiable.getName();
		EasyMock.expectLastCall().andReturn( NOTIFIABLE_IDENTIFIER ).atLeastOnce();

		factory.setMonitorable( DEFAULT_MONITORABLE_ADDRESS, monitorable );

		EasyMock.replay( monitorable );
		EasyMock.replay( notifiable );

		fd.register( notifiable, DEFAULT_MONITORABLE_ADDRESS );
		fd.startMonitoring( 1 );

		// There is not any more interested in this notifiable
		fd.unregister( notifiable, DEFAULT_MONITORABLE_ADDRESS );

		assertTrue( fd.getMonitorables().isEmpty() );
		EasyMock.verify( monitorable );
		EasyMock.verify( notifiable );
	}


	/**
	 * Now, two <code>Notifiable</code> are interested in the same
	 * <code>Monitorable</code>, so, if one <code>Notifiable</code>
	 * unregister, the <code>Monitorable</code> must continue to be monitored.
	 * 
	 * @throws Exception If anything wrong occurs
	 */
	public void testUnregister3() throws Exception {

		monitorable.getCurrentState();
		EasyMock.expectLastCall().andReturn( downState ).once(); // Because
		// could not
		// be enough
		// time to call this method
		monitorable.getEncarnationNumber();
		EasyMock.expectLastCall().andReturn( DEFAULT_REENCARNATION_NUMBER ).once();

		notifiable.getName();
		EasyMock.expectLastCall().andReturn( NOTIFIABLE_IDENTIFIER ).times( 2 );

		Notifiable notifiable1 = createNotifiableMock();
		notifiable1.getName();
		EasyMock.expectLastCall().andReturn( NOTIFIABLE_IDENTIFIER + "1" ).once();

		factory.setMonitorable( DEFAULT_MONITORABLE_ADDRESS, monitorable );

		EasyMock.replay( monitorable );
		EasyMock.replay( notifiable );
		EasyMock.replay( notifiable1 );

		fd.register( notifiable, DEFAULT_MONITORABLE_ADDRESS );
		fd.register( notifiable1, DEFAULT_MONITORABLE_ADDRESS );
		fd.startMonitoring( 1 );

		// There is not any more interested in this notifiable
		fd.unregister( notifiable, DEFAULT_MONITORABLE_ADDRESS );
		Map<String,MonitoringUnit> monitoringUnits = fd.getMonitorables();

		assertFalse( monitoringUnits.isEmpty() );
		assertTrue( monitoringUnits.containsKey( DEFAULT_MONITORABLE_ADDRESS ) );

		EasyMock.verify( monitorable );
		EasyMock.verify( notifiable );
		EasyMock.verify( notifiable1 );
	}


	/**
	 * Test if the <b>FailureDetector</b> is notifying when a
	 * <code>Monitorade</code> changes it state.
	 * 
	 * @throws Exception If anything wrong occurs
	 */
	public void testMonitorableStateChanged() throws Exception {

		monitorable.getCurrentState();
		EasyMock.expectLastCall().andReturn( upState ).once();
		monitorable.getEncarnationNumber();
		EasyMock.expectLastCall().andReturn( DEFAULT_REENCARNATION_NUMBER ).once();

		notifiable.getName();
		EasyMock.expectLastCall().andReturn( NOTIFIABLE_IDENTIFIER ).once();

		notifiable.notifyMonitorableStateChange( DEFAULT_MONITORABLE_ADDRESS, upState );
		EasyMock.expectLastCall().once();

		factory.setMonitorable( DEFAULT_MONITORABLE_ADDRESS, monitorable );

		EasyMock.replay( monitorable );
		EasyMock.replay( notifiable );

		fd.register( notifiable, DEFAULT_MONITORABLE_ADDRESS );
		fd.startMonitoring( 1 );

		EasyMock.verify( notifiable );
		EasyMock.verify( monitorable );

	}


	/**
	 * Test if the <b>FailureDetector</b> is notifying when a
	 * <code>Monitorade</code> changes it state. Now the state will change
	 * twice
	 * 
	 * @throws Exception If anything wrong occurs
	 */
	public void testMonitorableStateChanged2() throws Exception {

		monitorable.getCurrentState();
		EasyMock.expectLastCall().andReturn( upState ).once();
		monitorable.getCurrentState();
		EasyMock.expectLastCall().andReturn( downState ).once();
		monitorable.getEncarnationNumber();
		EasyMock.expectLastCall().andReturn( DEFAULT_REENCARNATION_NUMBER ).times( 2 );

		notifiable.getName();
		EasyMock.expectLastCall().andReturn( NOTIFIABLE_IDENTIFIER ).once();

		notifiable.notifyMonitorableStateChange( DEFAULT_MONITORABLE_ADDRESS, upState );
		EasyMock.expectLastCall().once();
		notifiable.notifyMonitorableStateChange( DEFAULT_MONITORABLE_ADDRESS, downState );
		EasyMock.expectLastCall().atLeastOnce();

		factory.setMonitorable( DEFAULT_MONITORABLE_ADDRESS, monitorable );

		EasyMock.replay( monitorable );
		EasyMock.replay( notifiable );

		fd.register( notifiable, DEFAULT_MONITORABLE_ADDRESS );
		fd.startMonitoring( 2 );

		EasyMock.verify( notifiable );
		EasyMock.verify( monitorable );

	}


	/**
	 * In this test the state will chance once but the name (that will be a
	 * encarnationNumber)
	 * 
	 * @throws Exception If anything wrong occurs
	 */
	public void testWhenReencarnationNumberChanges() throws Exception {

		final double newEncarnationNumber = 2d;

		EasyMock.expect( monitorable.getCurrentState() ).andReturn( upState ).times( 3 );

		EasyMock.expect( monitorable.getEncarnationNumber() ).andReturn( DEFAULT_REENCARNATION_NUMBER ).once();

		EasyMock.expect( monitorable.getEncarnationNumber() ).andReturn( newEncarnationNumber ).times( 2 );

		EasyMock.expect( notifiable.getName() ).andReturn( NOTIFIABLE_IDENTIFIER ).once();

		notifiable.notifyMonitorableStateChange( DEFAULT_MONITORABLE_ADDRESS, upState );
		EasyMock.expectLastCall().times( 2 );

		notifiable.notifyMonitorableStateChange( DEFAULT_MONITORABLE_ADDRESS, downState );
		EasyMock.expectLastCall().once();

		factory.setMonitorable( DEFAULT_MONITORABLE_ADDRESS, monitorable );

		EasyMock.replay( monitorable );
		EasyMock.replay( notifiable );

		fd.register( notifiable, DEFAULT_MONITORABLE_ADDRESS );
		fd.startMonitoring( 3 );

		Map<String,MonitoringUnit> monitoringUnits = fd.getMonitorables();

		assertFalse( monitoringUnits.isEmpty() );
		MonitoringUnit unit = monitoringUnits.values().iterator().next();
		assertEquals( newEncarnationNumber, unit.getLastEncarnationNumber() );
		assertEquals( DEFAULT_MONITORABLE_ADDRESS, unit.getMonitorableIdentifier() );

		EasyMock.verify( notifiable );
		EasyMock.verify( monitorable );

	}


	public void testGetFailureDetectorAddress() throws Exception {

		String monitorableAddress = "rmi://fireman:80/UserAgenServer";
		String expectedResponse = "rmi://fireman:80/" + AbstractFailureDetector.BOUND_NAME;
		assertEquals( expectedResponse, fd.getFailureDetectorAddress( monitorableAddress ) );

		monitorableAddress = "rmi://150.111.111.111:4543/UserAgenServer";
		expectedResponse = "rmi://150.111.111.111:4543/" + AbstractFailureDetector.BOUND_NAME;
		assertEquals( expectedResponse, fd.getFailureDetectorAddress( monitorableAddress ) );

		try {
			monitorableAddress = "rmi://fireman/UserAgenServer";
			fd.getFailureDetectorAddress( monitorableAddress );
			fail();
		} catch ( MalformedURLException mue ) {
			// Expected Exception
		}

		try {
			monitorableAddress = "rmi://fireman/";
			fd.getFailureDetectorAddress( monitorableAddress );
			fail();
		} catch ( MalformedURLException mue ) {
			// Expected Exception
		}

		try {
			monitorableAddress = "rmi://150.111.111.111/";
			fd.getFailureDetectorAddress( monitorableAddress );
			fail();
		} catch ( MalformedURLException mue ) {
			// Expected Exception
		}

		try {
			monitorableAddress = "rmi://150.111.111.111";
			fd.getFailureDetectorAddress( monitorableAddress );
			fail();
		} catch ( MalformedURLException mue ) {
			// Expected Exception
		}

	}


	private Monitorable createMonitorableMock() {

		return org.easymock.EasyMock.createMock( Monitorable.class );
	}


	private Notifiable createNotifiableMock() {

		return org.easymock.EasyMock.createMock( Notifiable.class );
	}

}
