/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.fd;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Map;
import java.util.TimerTask;
import java.util.concurrent.ExecutorService;

/**
 * Objects of this class are used to execute real pings on Grid Machines through
 * a <code>MachinesPingerThread</code>. It is created a new thread to ping
 * each machine avoiding the necessity to wait for a ping to return. When the
 * ping returns, the Grid Machine state is updated in the <code>Monitor</code>
 * passed in the constructor.
 */
public class MachinesPinger extends TimerTask {

	private Monitor fd;

	private final ExecutorService threadpool;


	/**
	 * The constructor.
	 * 
	 * @param fd Object responsible for maintaining the updated states of Grid
	 *        Machines and providing machines to be pinged.
	 * @param threadpool Pool to execute pingers in.
	 */
	public MachinesPinger( Monitor fd, ExecutorService threadpool ) {

		this.fd = fd;
		this.threadpool = threadpool;
	}


	/**
	 * This method retrieves the Grid Machines to be pinged from the
	 * <code>FailureDetector</code> and fires a
	 * <code>MachinesPingerThread</code> for each machine. This thread makes
	 * real pings on them and updates their state on the
	 * <code>FailureDetector</code>.
	 */
	@Override
	public void run() {

		Map<String,MonitoringUnit> monitoringUnitsMap = null;
		try {
			monitoringUnitsMap = this.fd.getMonitorables();
		} catch ( RemoteException re ) {
			// If the failureDetector dies
			this.cancel();
		}

		// Every body must sycronize in this map to avoid
		// ConcurrentModificationException
		synchronized ( monitoringUnitsMap ) {
			Collection<MonitoringUnit> monitoringUnits = monitoringUnitsMap.values();
			for ( MonitoringUnit unit : monitoringUnits ) {
				threadpool.execute( new MachinesPingerRunnable( unit ) );
			}
		}
	}


	@Override
	public boolean cancel() {

		return super.cancel();
	}
}
