/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.fd;

import java.net.MalformedURLException;
import java.rmi.RemoteException;
import java.util.Map;

/**
 * Objects that implement this interface are responsible for monitoring Grid
 * Machines and providing an updated information about their states.
 */
public interface Monitor {

	/**
	 * Adds an object interested in receiving Grid Machine state changes
	 * notifications of the specified Grid Machine.
	 * 
	 * @param notifiable object interested in receiving the notifications
	 * @param monitorableAddress the address of the service (<code>FailureDetector</code> ..
	 *        <code>Monitorable</code>) to monitor
	 * @throws RemoteException if anything wrong occurs in communication.
	 * @throws MalformedURLException
	 */

	public void register( Notifiable notifiable, String monitorableAddress ) throws RemoteException,
		MalformedURLException;


	/**
	 * Removes the object interested in receiving Grid Machine state changes
	 * notifications.
	 * 
	 * @param notifiable the object to be removed
	 * @param monitorableAddress the <code>Monitorable</code> address that the
	 *        <code>Notifiable</code> will not be interested.
	 * @throws RemoteException if anything wrong occurs in communication.
	 */
	public void unregister( Notifiable notifiable, String monitorableAddress ) throws RemoteException;


	/**
	 * Removes the object interested in receiving Grid Machine state changes
	 * notifications of all monitorables interested list.
	 * 
	 * @param notifiable the object to be removed
	 * @throws RemoteException if anything wrong occurs in communication.
	 */
	public void unregister( Notifiable notifiable ) throws RemoteException;


	/**
	 * Starts monitoring Grid Machines.
	 * 
	 * @throws RemoteException if anything wrong occurs in communication.
	 */
	public void startMonitoring() throws RemoteException;


	/**
	 * Retrieves a collection of monitorables
	 * 
	 * @return The entities to be monitored
	 * @throws RemoteException If anything wrong occurs in communication.
	 */
	public Map<String,MonitoringUnit> getMonitorables() throws RemoteException;

}
