/*
 * Copyright (c) 2002-2004 Universidade Federal de Campina Grande
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ourgrid.common.fd;

import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.Map;
import java.util.TreeMap;

/**
 * Create to organize the monitoring action this class knows all information
 * necessary to monitor a <code>Monitorable</code>. <br>
 * <br>
 * We have the remote entity and an local state to recover at any time.
 */
public class MonitoringUnit {

	/**
	 * This is the name used when the <code>MonitoringUnit</code> does not
	 * know (or can not get yet) the <code>Monitorable</code> name
	 */
	public static final String UNDEFINED_MONITORABLE_NAME = "UNDEFINED";

	/** The interested */
	Map<String,Notifiable> interesteds;

	/** The local state. Can be recovery at any time. */
	GenericState lastState;

	String monitorableServiceIndex;

	double lastEncarnationNumber;

	private String address;


	/**
	 * Constructor
	 * 
	 * @param initialState The <code>MonitoringUnit</code> initial state
	 * @param remoteFailureDetectorAddress The address of the
	 *        <code>MonitoringUnit</code> to be created
	 * @param monitorableServiceIndex An index name given by the
	 *        <code>Notifier</code> when register it interests to the
	 *        <code>Monitorable</code>. It will be used to notify that a
	 *        state has changed.
	 */
	public MonitoringUnit( GenericState initialState, String remoteFailureDetectorAddress,
							String monitorableServiceIndex ) {

		this.interesteds = Collections.synchronizedMap( new TreeMap<String,Notifiable>() );
		this.lastState = initialState;
		this.address = remoteFailureDetectorAddress;
		this.lastEncarnationNumber = 0;
		this.monitorableServiceIndex = monitorableServiceIndex;
	}


	/**
	 * @return Returns the interesteds.
	 */
	public Map<String,Notifiable> getInteresteds() {

		return interesteds;
	}


	/**
	 * @return Returns the state.
	 */
	public GenericState getLastState() {

		return lastState;
	}


	/**
	 * Adds a <code>Notifiable</code> like interested in this
	 * <code>Monitorable</code>.
	 * 
	 * @param notifiableName The name of the <code>Notifiable</code> to be
	 *        used as key.
	 * @param notifiable The <code>Notifiable</code> implementation.
	 */
	public void addInterested( String notifiableName, Notifiable notifiable ) {

		interesteds.put( notifiableName, notifiable );
	}


	/**
	 * Remove a <code>Notifiable</code> like interested in this
	 * <code>Monitorable</code>.
	 * 
	 * @param notifiableName The <code>Notifiable</code> name
	 */
	public void removeInterested( String notifiableName ) {

		interesteds.remove( notifiableName );
	}


	/**
	 * @return Returns true if the <code>Monitorable</code> at least one
	 *         <code>Notifiable</code> interested.
	 */
	public boolean hasInteresteds() {

		return !interesteds.isEmpty();
	}


	/**
	 * Executes a call (note that call can be remote or local), asking to the
	 * <code>Monitorable</code> the new state. If anything wrong occurs sets
	 * the monitorable state to <code>State.DOWN</code> Here, it is necessary
	 * to <code>getSynchronizedRemoteReference</code> when the object goes
	 * down. This synchronization is necessary because the remote references are
	 * modified at startup moment and old (deprecated) references could point to
	 * wrong remote references.
	 * 
	 * @see MonitoringUnit#getSynchronizedRemoteReference()
	 */
	public void updateState() {

		try {

			// Remote call block
			Monitorable monitorable = getSynchronizedRemoteReference();
			GenericState newState = monitorable.getCurrentState();
			double newEncarnationNumber = monitorable.getEncarnationNumber();
			// End of remote call block

			if ( encarnationNumberNotChange( newEncarnationNumber ) ) {
				changeState( newState );
			} else {
				this.lastEncarnationNumber = newEncarnationNumber;
				changeState( new GenericState( State.DOWN ) );
			}

		} catch ( RemoteException e ) {
			changeState( new GenericState( State.DOWN ) );
		} catch ( MalformedURLException e ) {
			// Case can not contact the service
			changeState( new GenericState( State.DOWN ) );
		} catch ( NotBoundException e ) {
			// Case can not contact the service
			changeState( new GenericState( State.DOWN ) );
		}
	}


	private boolean encarnationNumberNotChange( double newEncarnationNumber ) {

		// At the first iteration
		if ( getLastEncarnationNumber() == 0 ) {
			this.lastEncarnationNumber = newEncarnationNumber;
			return true;
		}
		return (getLastEncarnationNumber() == newEncarnationNumber);
	}


	/**
	 * This method is used to synchronize the reference the the
	 * <code>MonitoringUnit</code> has with the remote one.
	 * 
	 * @return A synchronized object remote reference.
	 * @throws RemoteException If could not contact the remote server
	 * @throws NotBoundException If the object in exported remotely.
	 * @throws MalformedURLException If the address is a not valid
	 *         <code>RMI</code> url.
	 */
	protected Monitorable getSynchronizedRemoteReference() throws RemoteException, MalformedURLException,
		NotBoundException {

		return (Monitorable) Naming.lookup( getAddress() );
	}


	/**
	 * Returns the object remote address.
	 * 
	 * @return The object remote address.
	 */
	public String getAddress() {

		return this.address;
	}


	/**
	 * @return Returns the last Reencarnation number associated with this
	 *         <code>MonitoringUnit</code>
	 */
	public double getLastEncarnationNumber() {

		return this.lastEncarnationNumber;
	}


	/**
	 * The state should be changed by this method only
	 * 
	 * @param newState
	 */
	public synchronized void changeState( GenericState newState ) {

		if ( !lastState.equals( newState ) ) {
			this.lastState = newState;
			notifyInteresteds( newState );
		}
	}


	/**
	 * Will notify the interested about a state change of a Monitorable. We use
	 * a cloned map due to the Notifiables actions can cause a deadlock if it
	 * try to execute anything at the failure detector.
	 * 
	 * @param newState The new state to be notified.
	 */
	private void notifyInteresteds( GenericState newState ) {

		Collection<Notifiable> interestedsClone = new LinkedList<Notifiable>( interesteds.values() );

		for ( Notifiable notifiable : interestedsClone ) {
			notifiable.notifyMonitorableStateChange( getMonitorableIdentifier(), newState );
		}
	}


	/**
	 * An index name given by the <code>Notifier</code> when register it
	 * interests to the <code>Monitorable</code>.
	 * 
	 * @return An index name given by the <code>Notifier</code> when register
	 *         it interests to the <code>Monitorable</code>.
	 */
	public String getMonitorableIdentifier() {

		return this.monitorableServiceIndex;
	}

}
