/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ourgrid.common.gum;

import java.io.File;
import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.Map;

import org.ourgrid.common.exception.UnableToCreatePlaypenException;
import org.ourgrid.common.exception.UnableToCreateStorageException;
import org.ourgrid.common.exception.UnableToDigestFileException;
import org.ourgrid.common.executor.ExecutorException;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.ExecutorResult;

/**
 * Represents a grid machine, which is an abstraction that allows a user to use
 * a remote machine on the grid without knowing about its particularities (for
 * example: the installed OS) . This interface defines the minimal set of
 * operations needed to execute a task on machines in a grid, abstracting their
 * differences. Namely, the operations that comprise this minimal set are:
 * <ul>
 * <li>Running a command on the machine, that is: set up an user environment
 * (environment variables) and execute a command in a specified directory in the
 * grid machine.
 * <li>Transfer a file from the home machine to the grid machine, keeping the
 * file name and permissions
 * <li>Transfer a file from the grid machine back to the machine where is the
 * object which initiates the transfer. This operation will keep the file
 * permissions
 * </ul>
 * <p>
 * All implementations of <code>Gum</code> <i>must be</i> thread-safe. That
 * is, they must support calling of their methods by multiple threads
 * consistently. Furthermore, implementations that also support parallel threads
 * running each method are very much encouraged (that is: multiple file
 * transfers or multiple commands running).
 * <p>
 * Each instance of <code>Gum</code> also has an unique name that identifies
 * it, and a list of attributes. Such name is the network address (for example:
 * baleia.dsc.ufpb.edu.br) plus a port number (for example: 3089). The tuple
 * (address,port) identifies a single machine in the grid and is used to
 * reference the grid machine.
 * <p>
 * The attributes of a <code>Gum</code> are a set of name/value pairs that
 * represent the machine characteristics. The semantics of the attributes is
 * given by the user. They are used to verify if a grid machine satisfies the
 * requirements of a request. This verification is done by the provider.
 * <p>
 * This interface is part of the core interfaces of the OurGrid project, along
 * with <code>Gump</code> and <code>GumpClient</code>. They define a
 * general way of getting access to and use machines on a grid, abstracting
 * differences in doing so.
 * 
 * @see org.ourgrid.common.gum.GumClient
 */
public interface Gum extends Remote {

	/**
	 * Executes a remote task command in a specific directory on the grid
	 * machine.
	 * 
	 * @param remoteDir directory where the command will be executed. For
	 *        example, in a Unix environment, a <code>cd</code> to this
	 *        directory will be done before running the specified command.
	 * @param command the command that will be executed in the specified
	 *        directory.
	 * @param environmentVars a <code>Map</code> containing a set of
	 *        environment variable names as the keys and the corresponding
	 *        values as the values set. All these variables will be set before
	 *        the execution of the command. The keys and values of Map should be
	 *        String or, at least, override the <code>Object#toString</code>
	 *        method.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation.
	 * @throws ExecutorException if some error occurs during the remote
	 *         execution of the command
	 * @return A handle that identifies executions dispatched to instances of a
	 *         native <code>Executor</code> class.
	 */
	public ExecutorHandle remoteExecute( String remoteDir, String command, Map<String,String> environmentVars )
		throws RemoteException, ExecutorException;


	/**
	 * Put an array of bytes INTO the file specified by <code>removeFile</code>
	 * at the Grid Machine.
	 * 
	 * @param remoteFile the name of file that where the array of bytes will be
	 *        appended
	 * @param data an array of bytes that must be appended to the file.
	 * @param firstSlice Indicates if the array of bytes is the first slice of
	 *        the file or not. If <b>true</b> and <i>remoteFile</i> exists,
	 *        the file will be overwritten. Otherwise, the array of bytes will
	 *        just be appended to the file.
	 * @throws RemoteException if some RMI error occurs during the putFile()
	 *         method invocation.
	 * @throws IOException if some error occurs during the writing process.
	 */
	public void putFile( String remoteFile, byte[ ] data, boolean firstSlice ) throws RemoteException, IOException;


	/**
	 * Get an array of bytes FROM a file that is located at the Grid Machine.
	 * 
	 * @param filePath the path where the file is located.
	 * @param filePos the position into the file from the array will be read.
	 * @param size the size of array of bytes that must be read.
	 * @return the array of bytes read from the file.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation
	 * @throws IOException if some error occurs during the reading process.
	 */
	public byte[ ] getFile( String filePath, long filePos, int size ) throws RemoteException, IOException;


	/**
	 * Gives access to the result of execution.
	 * 
	 * @param handle a handle that identifies the execution of interest
	 * @return an object that contains information about the execution (exit
	 *         value, stderr, stdout).
	 */
	public ExecutorResult getResult( ExecutorHandle handle ) throws RemoteException, ExecutorException;


	/**
	 * Check if the file denoted by file abstraction exists.
	 * 
	 * @param file the file that will be searched
	 * @return <b>true</b> if the file exists, <b>false</b> otherwise.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation
	 */
	public boolean fileExists( File file ) throws RemoteException;


	/**
	 * Change the permissions of a file located at the Grid Machine. For
	 * example, change the permissions of a script file, in order to execute it.
	 * 
	 * @param file the file which permissions will be changed.
	 * @param permissions a string specifying the permissions. The string must
	 *        be composed of three numbers, like in the UNIX-like systems. For
	 *        instance, "rwxrwx---" must be represented by "770".
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation
	 * @throws IOException if some error occurs while trying to read or write
	 *         current permission of file.
	 */
	public void changeFilePermissions( File file, int permissions ) throws RemoteException, IOException;


	/**
	 * Determines the remote file size in bytes.
	 * 
	 * @param remoteFile The file located at the Grid Machine which the size
	 *        must be determined.
	 * @return The size of the file in bytes.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation
	 * @throws IOException If some problem occurs during the access to the file.
	 */
	public long getRemoteFileLength( File remoteFile ) throws RemoteException, IOException;


	/**
	 * Gets the time when the last modification in the specified file has
	 * occurred.
	 * 
	 * @param remoteFile The file located at the Grid Machine which the
	 *        modification time will be determined.
	 * @param normalized Indicates if the last modification time must be
	 *        normalized to the TIMEZONE-0.
	 * @return The number of milliseconds from 1/Jan/1970 to the moment of file
	 *         last modification.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation
	 * @throws IOException if during the access to the file some problem occurs.
	 */
	public long getRemoteFileLastModification( File remoteFile, boolean normalized ) throws RemoteException,
		IOException;


	/**
	 * Used to return the digest representation of a file. Files with the same
	 * content must have the same digest. Different content files must have
	 * different digests.
	 * 
	 * @param remoteFile if the file used to generate the digest
	 * @return The digest representation of the file or 0 if its a digest from a
	 *         directory.
	 * @throws UnableToDigestFileException if there is any problem on the digest
	 *         generation, like the file is not found, I/O errors or the digest
	 *         algorithm is not valid.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation
	 */
	public String getRemoteFileDigest( File remoteFile ) throws RemoteException, UnableToDigestFileException;


	/**
	 * Creates a playpen in the Grid Machine. The playpen is supposed to be an
	 * area reserved to store temporary files related to the execution of a
	 * Task. It is important that a gum consumer always call stopReplica() after
	 * using the gum to assure the playpen will be destroyed.
	 * 
	 * @param size The number of bytes that must be reserved to the playpen.
	 * @return The path to the directory at the Grid Machine corresponding to
	 *         the playpen.
	 * @throws RemoteException If some RMI error occurs during this method
	 *         invocation.
	 * @throws UnableToCreatePlaypenException if the playpen could not be
	 *         created. This will be usually due to I/O issues, like user space
	 *         quota exceeded
	 */
	public String createPlaypen( int size ) throws RemoteException, UnableToCreatePlaypenException;


	/**
	 * Creates the storage directory. If it already exists, nothing is done.
	 * 
	 * @return the absolute path of the storage directory.
	 * @throws RemoteException if some RMI error occurs during this method
	 *         invocation.
	 * @throws UnableToCreateStorageException if the storage directory could not
	 *         be created. This will be usually due to I/O issues, like user
	 *         space quota exceeded
	 */
	public String createStorageDirectory() throws RemoteException, UnableToCreateStorageException;


	/**
	 * Returns the remote storage directory. If it does not exist, it should be
	 * created.
	 * 
	 * @return the path of the storage directory.
	 */
	public String getStorageDirectory() throws RemoteException;


	/**
	 * Gets the status of the grid machine. It may assume one of five states:
	 * UP, READY, RUNNING, OWNER and DOWN, described above.
	 * 
	 * @return the status of this GridMachine.
	 * @throws RemoteException if some RMI error occurs during this method
	 *         invocation.
	 * @see org.ourgrid.common.gum.GumStatus
	 */
	public GumStatus getStatus() throws RemoteException;


	/**
	 * Determines a replica start. When a replica is started the gum must listen
	 * and hold information about created directories, process and anything that
	 * is created during the replica execution. It's impossible to execute
	 * anything or create a playpen at the gum before this method is called.
	 * 
	 * @throws RemoteException if some RMI error occurs during this method
	 *         invocation.
	 */
	public void startReplica() throws RemoteException;


	/**
	 * Tells the user agent to kill any executing tasks and also clean the
	 * environment to a next execution. This is done by cleaning any playpen
	 * created. Any invalidation or shutdown must invoke this method to assure
	 * that the gum is clean.
	 * 
	 * @throws RemoteException If some RMI error occurs during this method
	 *         invocation.
	 */
	public void stopReplica() throws RemoteException;
}
