/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ourgrid.common.gum;

import java.io.File;
import java.io.Serializable;
import java.util.Map;

import org.ourgrid.common.exception.UnableToCreateStorageException;
import org.ourgrid.common.exception.UnavailableGumException;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.ExecutorResult;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.gridmachine.exception.FailCauseException;
import org.ourgrid.gridmachine.exception.FileTransferException;
import org.ourgrid.gridmachine.exception.UnableToExecuteException;

/**
 * <code>GumClient</code> is an abstraction that represents all entities that
 * will consume <code>Gum</code>s being the one who will effectively use the
 * <code>Gum</code>. Doing that, it will have a reference to the last and
 * will execute remote tasks on it after making special treatments useful to the
 * client and providing callback methods.
 */
public interface GumClient extends Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	static final long serialVersionUID = 33L;


	/**
	 * @throws UnavailableGumException Throws if the contact with the remote
	 *         machine is lost.
	 * @throws UnableToExecuteException Throws if cannot execute the command on
	 *         the machine. For example, if the command does not exist or the
	 *         machine already is executing a remote task command.
	 * @see Gum#remoteExecute(String, String, Map)
	 */
	public ExecutorHandle remoteExecute( String remoteDir, String command, Map<String,String> environmentVars )
		throws FailCauseException;


	/**
	 * @see Gum#putFile(String, byte[], boolean)
	 * @throws FileTransferException Throws if can not transfer the file.
	 */
	public void putFile( String filePath, String remoteFile ) throws FailCauseException;


	/**
	 * @see Gum#getFile(String, long, int)
	 * @throws FileTransferException if the file cannot be transfered.
	 */
	public void getFile( String remoteFilePath, String localFile ) throws FailCauseException;


	/**
	 * @see Gum#getResult(ExecutorHandle)
	 */
	public ExecutorResult getResult( ExecutorHandle handle ) throws FailCauseException;


	/**
	 * This method stores a local file into the remote storage location. In
	 * general that location is the ".mgstorage" directory in a shared file
	 * system at the remote site.
	 * 
	 * @param localFile A string representing the local file path.
	 * @param remoteFile A string representing the remote file path.
	 * @throws UnavailableGumException if the requested GM is not available.
	 * @throws FileTransferException if some transfer problems happen.
	 */
	public void storeFile( String localFile, String remoteFile ) throws FailCauseException;


	/**
	 * @see Gum#fileExists(File) *
	 * @throws UnavailableGumException Occurs if the requested processor is not
	 *         available.
	 * @throws UnableToExecuteException Occurs if some operation could not be
	 *         performed in order to verify if the file exists.
	 * @throws FileTransferException if any file that must be transfered for the
	 *         execution (in case of transfer a script) fails.
	 */
	public boolean fileExists( File file ) throws FailCauseException;


	/**
	 * @see Gum#getRemoteFileLength(File)
	 * @throws UnavailableGumException if the requested GM is not available.
	 * @throws UnableToExecuteException Occurs if some operation could not be
	 *         performed in order to get the remote file length.
	 * @throws FileTransferException if some transfer problems happen.
	 */
	public long getRemoteFileLength( File file ) throws FailCauseException;


	/**
	 * @see Gum#getRemoteFileLastModification(File, boolean)
	 * @throws UnavailableGumException if the requested GM is not available.
	 * @throws UnableToExecuteException Occurs if some operation could not be
	 *         performed in order to get the remote file length.
	 */
	public long getRemoteFileLastModification( File remoteFile, boolean normalized ) throws FailCauseException;


	/**
	 * @throws UnavailableGumException if the requested GM is not available.
	 * @throws UnableToExecuteException Occurs if some operation could not be
	 *         performed in order to create playpen directory.
	 * @throws FileTransferException if some transfer problems happen.
	 * @see Gum#createPlaypen(int)
	 */
	public String createPlaypen( int size ) throws FailCauseException;


	/**
	 * @see Gum#createStorageDirectory()
	 * @throws UnavailableGumException if the requested GM is not available.
	 * @throws UnableToExecuteException Occurs if some operation could not be
	 *         performed in order to create the storage directory.
	 * @throws UnableToCreateStorageException if the creation of the directory
	 *         was not possible due to permission denied, disk space is full,
	 *         etc.
	 * @throws FileTransferException if some transfer problems happen.
	 */
	public String createStorageDirectory() throws FailCauseException;


	/**
	 * @see Gum#getStorageDirectory()
	 * @throws UnavailableGumException if the requested GM is not available.
	 * @throws UnableToExecuteException Occurs if some operation could not be
	 *         performed in order to get the remote storage directory path.
	 * @throws FileTransferException if some transfer problems happen.
	 */
	public String getStorageDirectory() throws FailCauseException;


	/**
	 * Retrieves the gum specification cached on the client side
	 */
	public GumSpec getGumSpec();


	/**
	 * @see Gum#getStatus()
	 */
	public GumStatus getStatus();


	/**
	 * @see Gum#startReplica()
	 * @throws UnavailableGumException If the contact with grid machine is lost.
	 */
	public void startReplica() throws FailCauseException;


	/**
	 * @see Gum#stopReplica()
	 * @throws UnavailableGumException If the contact with grid machine is lost.
	 */
	public void stopReplica() throws FailCauseException;

}
