/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.gump;

import java.rmi.Naming;
import java.rmi.RemoteException;
import java.util.concurrent.ConcurrentHashMap;

import org.ourgrid.common.exception.RequestAlreadyExistsException;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.security.MGSecureClientSocketFactory;
import org.ourgrid.common.security.MGSecureServerSocketFactory;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.common.url.PeerURLProvider;
import org.ourgrid.peer.manager.EBPeerManagerFacade;

public class BrokerAccessImpl extends OurgridUnicastRemoteObject implements Gump {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/**
	 * Logger to store log messages.
	 */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( BrokerAccessImpl.class );

	private ObjectID objectID;

	private ConcurrentHashMap<Long,RequestSpec> requestSpecsMap = new ConcurrentHashMap<Long,RequestSpec>();

	private EBPeerManagerFacade ebPeerManagerFacade;


	/**
	 * Default constructor.
	 * 
	 * @param brokerAccessID ID used to identify this broker access.
	 * @param peerFacade Facade used to communicate with peer module.
	 */
	public BrokerAccessImpl( ObjectID brokerAccessID, EBPeerManagerFacade peerFacade ) throws RemoteException {

		super();
		construct( brokerAccessID, peerFacade );
	}


	/**
	 * Constructor with security options.
	 * 
	 * @param brokerAccessID ID used to identify this broker access.
	 * @param ebPeerManagerFacade Facade used to communicate with peer module.
	 * @param secureClientSocketFactory The SecureClientSocketFactory.
	 * @param secureServerSocketFactory The SecureServerSocketFactory.
	 * @param securePort The port where the secure services will be bound.
	 */
	public BrokerAccessImpl( ObjectID brokerAccessID, EBPeerManagerFacade ebPeerManagerFacade,
								MGSecureClientSocketFactory secureClientSocketFactory,
								MGSecureServerSocketFactory secureServerSocketFactory, int securePort )
		throws RemoteException {

		super( securePort, secureClientSocketFactory, secureServerSocketFactory );
		construct( brokerAccessID, ebPeerManagerFacade );
	}


	/**
	 * Used by constructor to set parameters.
	 * 
	 * @param brokerAccessID ID used to identify this broker access.
	 * @param peerFacade Facade used to communicate with peer module.
	 */
	private void construct( ObjectID brokerAccessID, EBPeerManagerFacade peerFacade ) {

		this.ebPeerManagerFacade = peerFacade;
		this.objectID = brokerAccessID;
	}


	/**
	 * @see org.ourgrid.common.gump.Gump#requestGums(RequestSpec)
	 */
	public void requestGums( RequestSpec requestSpec ) throws RemoteException, RequestAlreadyExistsException {

		if ( requestSpec.getRequiredGums() == 0 ) {
			throw new RemoteException( "Invalid number of required gums." );
		}
		requestSpec.setAsFromLocalBroker();
		requestSpecsMap.put( requestSpec.getID(), requestSpec );
		this.ebPeerManagerFacade.requestGums( requestSpec );
	}


	/**
	 * @see org.ourgrid.common.gump.Gump#finishRequestGums(long)
	 */
	public void finishRequestGums( long requestID ) {

		requestSpecsMap.remove( requestID );
		this.ebPeerManagerFacade.finishRequestGums( requestID );
	}


	/**
	 * @see org.ourgrid.common.gump.Gump#disposeGum(long, GumID)
	 */
	public void disposeGum( long requestID, GumID gumID ) {

		this.ebPeerManagerFacade.disposeGum( requestID, gumID );
	}


	/**
	 * @see org.ourgrid.common.gump.Gump#pauseRequestGums(long)
	 */
	public void pauseRequestGums( long requestID ) {

		this.ebPeerManagerFacade.pauseRequestGums( requestID );
	}


	/**
	 * @see org.ourgrid.common.gump.Gump#resumeRequestGums(long)
	 */
	public void resumeRequestGums( long requestID ) {

		this.ebPeerManagerFacade.resumeRequestGums( requestID );
	}


	/**
	 * @see org.ourgrid.common.gump.Gump#getObjectID()
	 */
	public ObjectID getObjectID() {

		/*
		 * This entity does not need to provide an implementation for this
		 * method. In the future, when using JIC, every interface that will be
		 * remotly accessed must implement this method, which will return a
		 * valid Jabber Identification as ObjectID. By this time, the only
		 * entity that is implementing this method is PeerImpl, which needs to
		 * provide an Identification for the peer it represents.
		 */
		return objectID;
	}


	public void shutdown() {

		try {
			Naming.unbind( PeerURLProvider.localAccess() );
		} catch ( Exception e ) {
			LOG.error( "Problems during shutdown" );
		}
	}


	/**
	 * @see org.ourgrid.common.gump.Gump#unwantedGum(long,
	 *      org.ourgrid.common.id.GumID)
	 */
	public void unwantedGum( long requestID, GumID gumID ) throws RemoteException {

		this.ebPeerManagerFacade.unwantedGum( requestID, gumID );
	}
}
