/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.gump;

import java.rmi.Remote;
import java.rmi.RemoteException;

import org.ourgrid.common.exception.RequestAlreadyExistsException;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.RequestSpec;

/**
 * This interface represents a grid machine provider. It provides grid machines
 * to consumers. The consumers that will receive the grid machines in the form
 * of <code>Gum</code> instances, must implement the <code>GumpClient</code>
 * interface. The consumers will receive Gums asynchronously through invocations
 * of the <code>GumpClient.hereIsGum()</code> method.
 * <p>
 * There are no guarantee that the provider will return any grid machine at all.
 * They will be provided accordingly to the availability and prioritizing
 * policies of the <code>Gump</code> implementation. Moreover, the
 * <code>Gum</code>s provided may become unavailable at any time.
 * <p>
 * The request for <code>Gum</code>s should work in the following way: When a
 * consumer need grid machines it will invoke the method
 * <code>Gump#requestGums</code>. The request will be registered at this
 * provider and should receive <code>Gum</code>s, if available. If a request
 * is satisfied (i.e. it's got all the machines it's asked), the consumer should
 * invoke the <code>Gump#pauseRequest</code> method, to tell the consumer it
 * need no more <code>Gum</code>s. If a machine currently allocated to a
 * satisfied request is lost, the consumer will invoke the
 * <code>Gump#resumeRequest</code> method. This will tell the provider that
 * the request need machines again. When a request should be finished (i.e. the
 * job is finished) the consumer should invoke the
 * <code>Gump#finishRequestGums</code> at the provider. The provider will then
 * take all the machines in use from the consumer, and clean the information
 * about the request.
 * <p>
 * The methods Gump#requestGums, Gump#pauseRequest, Gump#resumeRequest and
 * Gump#finishRequestGums must result in a broadcast to the entire Ourgrid
 * community (i.e. all know peers).
 * <p>
 * This interface is part of the core interfaces of the OurGrid project, along
 * with <code>Gum</code> and <code>GumpClient</code>. They define a general
 * way of getting access to and use machines on a grid, abstracting differences
 * on doing so.
 */
public interface Gump extends Remote {

	/**
	 * Requests grid machines according to the specification contained in a
	 * <code>RequestSpec</code>. This event must be broadcasted to the entire
	 * Ourgrid community (that is, all know peers).
	 * 
	 * @param requestSpec specification of the request
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation.
	 * @throws RequestAlreadyExistsException another request with the same ID
	 *         already exists
	 */
	public void requestGums( RequestSpec requestSpec ) throws RemoteException, RequestAlreadyExistsException;


	/**
	 * Informs the provider that it must stop providing grid machines for the
	 * consumer who made a given request. This event must be broadcasted to the
	 * entire Ourgrid community (that is, all know peers).
	 * 
	 * @param requestID the unique ID that represents the request that needs no
	 *        more grid machines.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation.
	 */
	public void finishRequestGums( long requestID ) throws RemoteException;


	/**
	 * Suspends a request for gums. This event must be broadcasted to the entire
	 * Ourgrid community (that is, all know peers).
	 * 
	 * @param requestID the identification of the request that will be
	 *        suspended.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation.
	 */
	public void pauseRequestGums( long requestID ) throws RemoteException;


	/**
	 * Resumes a request for gums. This event must be broadcasted to the entire
	 * Ourgrid community (that is, all know peers).
	 * 
	 * @param requestID the identification of the request that will be resumed.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation.
	 */
	public void resumeRequestGums( long requestID ) throws RemoteException;


	/**
	 * Informs the provider that a given grid machine is not needed anymore.
	 * Thus, the grid machine can be rescheduled to other requests. The same
	 * machine may scheduled to the same consumer in the future, for the same
	 * request in case it needs machines again or for a completely new request.
	 * If it is required that the given request no longer receives the grid
	 * machine being disposed, the
	 * <code>unwantedGum(long, org.ourgrid.common.id.GumID)</code> call should
	 * be used. <br>
	 * </br> This event must be forwarded to the gum owner, only.
	 * 
	 * @param requestID the unique ID that represents the request that wants to
	 *        dispose the grid machine.
	 * @param gumID identification of the grid machine
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation.
	 * @see org.ourgrid.common.gump.Gump#unwantedGum(long, GumID)
	 */
	public void disposeGum( long requestID, GumID gumID ) throws RemoteException;


	/**
	 * Informs the provider that a given grid machine is not needed anymore.
	 * Thus, the grid machine can be rescheduled to other requests. Different
	 * from the <code>disposeGum(long, org.ourgrid.common.id.GumID)</code>,
	 * when a Peer receives this call it will no longer deliver this machines to
	 * the same request. <br>
	 * </br> This event must be forwarded to the gum owner, only.
	 * 
	 * @param requestID the unique ID that represents the request which no
	 *        longer wants this grid machine.
	 * @param gumID identification of the grid machine
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation.
	 * @see org.ourgrid.common.gump.Gump#disposeGum(long, GumID)
	 */
	public void unwantedGum( long requestID, GumID gumID ) throws RemoteException;


	/**
	 * Returns the identification object related to this
	 * <code>GridMachineProvider</code>.
	 * 
	 * @return the provider identification.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation.
	 */
	public ObjectID getObjectID() throws RemoteException;
}
