/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.gump;

import java.rmi.Remote;
import java.rmi.RemoteException;

import org.ourgrid.common.id.ObjectID;
import org.ourgrid.mygrid.scheduler.gridmanager.GumResponseEntry;

/**
 * <p>
 * Represents a grid consumer that wants to use grid machines. This consumer
 * gains access to grid machines through a grid machines provider. These two
 * entities, grid machines and providers, are represented, respectively, by
 * <code>Gum</code> and <code>Gump</code> instances.
 * <p>
 * To request grid machines, an instance of <code>GumClient</code> uses the
 * <code>Gump.requestGums()</code> method. The provider then calls back the
 * consumer to deliver <code>Gum</code> instances through a set of invocations
 * of <code>hereIsGum()</code>.
 * <p>
 * This interface is part of the core interfaces of the OurGrid project, along
 * with <code>Gum</code> and <code>Gump</code>. They define a general way
 * of getting access to and use machines on a grid, abstracting differences on
 * doing so.
 */
public interface GumpClient extends Remote {

	/**
	 * <p>
	 * Delivers a grid machine (<code>Gum</code>) to this consumer. The grid
	 * machine satisfies the requirements of a specific request made by this
	 * consumer. The respective request is informed through a requestID which
	 * travels inside the parameter of type <code>GumResponseEntry</code>,
	 * which also contains a reference to the <code>Gum</code>, a gum
	 * specification (<code>GumSpec</code>) and a ID of the grid machine
	 * provider.
	 * <p>
	 * The grid machine availability has no guarantees. It may become
	 * unavailable due to failures or to the provider's will. For example, the
	 * provider may decide to schedule the grid machine to another consumer.
	 * <p>
	 * Also, it is expected that the consumer releases the <code>Gum</code>
	 * when it does not need it anymore. To do this the consumer must invoke the
	 * method <code>Gump.finishRequestGums()
	 * </code>.
	 * 
	 * @param gumResponseEntry an object containing the information about the
	 *        delivery. It contains a reference to the <code>Gum</code>, a
	 *        gum specification (<code>GumSpec</code>) and a ID of the grid
	 *        machine provider.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation.
	 */
	public void hereIsGum( GumResponseEntry gumResponseEntry ) throws RemoteException;


	/**
	 * Returns the Identification object related to this <code>GumpClient</code>.
	 * 
	 * @return the consumer identification.
	 * @throws RemoteException if some RMI error occurs during the method
	 *         invocation.
	 */
	public ObjectID getObjectID() throws RemoteException;
}
