/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.gump;

import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.ourgrid.common.event.Shutdownable;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.mygrid.scheduler.gridmanager.GumResponseEntry;
import org.ourgrid.peer.manager.EBPeerManagerFacade;

/**
 * Represents a ThreadPool responsible for managing and executing Threads that
 * will invoke RMI methods on the external consumers using the Peer, for
 * example: MyGrid.
 */
public class GumpClientThreadManager implements Shutdownable {

	/**
	 * The Thread pool.
	 */
	private ThreadPoolExecutor threadPool;

	/**
	 * Logger to store log messages.
	 */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( GumpClientThreadManager.class );

	/**
	 * The ebPeerManagerFacade who provides Grid Machines.
	 */
	private EBPeerManagerFacade ebPeerManagerFacade;


	/**
	 * Default constructor
	 */
	public GumpClientThreadManager( EBPeerManagerFacade facade ) {

		this.threadPool = new ThreadPoolExecutor( 1, Integer.MAX_VALUE, 60, TimeUnit.SECONDS,
				new SynchronousQueue<Runnable>() );
		this.ebPeerManagerFacade = facade;
	}


	/**
	 * Creates a <code>DelivererThread</code> and arrange to be executed by a
	 * thread in the pool. The Thread created will try to deliver a
	 * <code>Gum</code> to a <code>GumpClient.</code> If the gum cannot be
	 * successfully delivered, the deliverer thread will dispose the gum.
	 * 
	 * @param gumClientID ID of the consumer that will receive the
	 *        <code>GridMachine</code>
	 * @param gumResponseEntry an object which encapsulates information about
	 *        the gum to be delivered, including a reference to a
	 *        <code>Gum</code> type
	 */
	public void createAndExecuteDelivererThread( ObjectID gumClientID, GumResponseEntry gumResponseEntry ) {

		Runnable threadDeliverer = new DelivererThread( gumClientID, gumResponseEntry, ebPeerManagerFacade );
		this.threadPool.execute( threadDeliverer );
	}


	/**
	 * @see org.ourgrid.common.event.Shutdownable#shutdown()
	 */
	public void shutdown() {

		LOG.info( "The ConsumerThreadManager is going down" );

		threadPool.shutdownNow();
	}

}
