/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.gump;

import java.rmi.Naming;
import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.UnicastRemoteObject;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.mygrid.scheduler.gridmanager.GumResponseEntry;
import org.ourgrid.peer.manager.EBPeerManagerFacade;
import org.ourgrid.test.PeerPreCommitTest;
import org.ourgrid.test.TestUtils;
import org.ourgrid.test.threadServices.ThreadPoolServices;
import org.ourgrid.test.threadServices.ThreadServices;

/**
 * The GridMachineProviderClientThreadManager tests.
 */
public class GumpClientThreadManagerTest extends TestCase implements PeerPreCommitTest {

	private GumpClientThreadManager consumerThreadManager;

	private GumpClient mockGumpClient;

	private Gump mockGump;

	private EBPeerManagerFacade facade;

	private GumSpec gumSpec;

	private ObjectID gumpClientID;

	private ObjectID gumpID;

	private GumResponseEntry gumResponseEntry;

	private TestUtils testUtils;

	private int registryPort;

	private BindableMockGumpClientWrapper bindableMockGumpClientWrapper;

	private BindableMockGumpWrapper bindableMockGumpWrapper;


	/*
	 * @see TestCase#setUp()
	 */
	@Override
	protected void setUp() throws Exception {

		super.setUp();
		this.testUtils = new TestUtils();
		this.testUtils.load();
		this.testUtils.createOGProperties();

		this.registryPort = Integer.parseInt( testUtils.getProperty( TestUtils.RMI_PORT ) );

		try {
			LocateRegistry.createRegistry( registryPort );
		} catch ( Exception e ) {
			LocateRegistry.getRegistry( registryPort );
		}

		Configuration.getInstance( Configuration.PEER );

		this.facade = EasyMock.createMock( EBPeerManagerFacade.class );
		this.consumerThreadManager = new GumpClientThreadManager( facade );

		this.mockGump = EasyMock.createMock( Gump.class );
		this.mockGumpClient = EasyMock.createMock( GumpClient.class );

		this.gumSpec = new GumSpec();
		this.gumSpec.putAttribute( GumSpec.ATT_NAME, "MockGum" );
		this.gumSpec.putAttribute( GumSpec.ATT_PORT, "1111" );
		this.gumSpec.putAttribute( GumSpec.ATT_PEERNAME, "MyPeer" );
		this.gumSpec.putAttribute( GumSpec.ATT_PEERPORT, "2222" );

		this.gumpClientID = new ObjectID( "MockGumClient", (long) 1, "rmi://localhost:" + registryPort
				+ "/MockGumpClient" );
		this.gumpID = new ObjectID( "MockGump", (long) 1, "rmi://localhost:" + registryPort + "/MockGump" );
		this.bindableMockGumpWrapper = new BindableMockGumpWrapper( mockGump );
		this.bindableMockGumpClientWrapper = new BindableMockGumpClientWrapper( mockGumpClient );

		Naming.rebind( gumpClientID.getObjectURL(), bindableMockGumpClientWrapper );
		Naming.rebind( gumpID.getObjectURL(), bindableMockGumpWrapper );

		this.mockGump.getObjectID();
		EasyMock.expectLastCall().andReturn( gumpID ).anyTimes();

		this.mockGumpClient.getObjectID();
		EasyMock.expectLastCall().andReturn( gumpClientID ).anyTimes();

		long id = 7171;
		this.gumResponseEntry = new GumResponseEntry( null, gumSpec, id, gumpID );
	}


	/*
	 * @see TestCase#tearDown()
	 */
	@Override
	protected void tearDown() throws Exception {

		super.tearDown();

		Naming.unbind( gumpClientID.getObjectURL() );
		Naming.unbind( gumpID.getObjectURL() );

		try {
			UnicastRemoteObject.unexportObject( bindableMockGumpClientWrapper, true );
			UnicastRemoteObject.unexportObject( bindableMockGumpWrapper, true );
		} catch ( NoSuchObjectException e ) {
		}

		testUtils.reset();
		Configuration.reset();
	}


	/**
	 * Tests the CreateAndDeliverThread method, delivering a gum for a consumer.
	 * Everything should works ok.
	 * 
	 * @throws RemoteException
	 */
	public void testCreateAndExecuteDelivererThread1() throws RemoteException {

		this.mockGumpClient.hereIsGum( EasyMock.eq( gumResponseEntry ) );
		EasyMock.expectLastCall().once();
		EasyMock.replay( this.mockGumpClient );

		this.consumerThreadManager.createAndExecuteDelivererThread( gumpClientID, gumResponseEntry );
		ThreadServices.waitUntilWorkIsDone();
		ThreadPoolServices.waitUntilWorkIsDone();

		EasyMock.verify( mockGumpClient );
	}


	/**
	 * Test if the CreateAndDeliverThread method is working as unexpected if the
	 * consumer fail. In case of the consumer fail, the ConsumerThreadManager
	 * should notify the EBPeerManagerFacade informing that that can is not
	 * being used.
	 * 
	 * @throws RemoteException
	 */
	public void testCreateAndExecuteDeliverThread2() throws Exception {

		this.mockGumpClient.hereIsGum( EasyMock.eq( gumResponseEntry ) );
		EasyMock.expectLastCall().andThrow( new RemoteException() );

		this.facade.finishRequestGums( EasyMock.anyLong() );
		EasyMock.expectLastCall().once();

		GumID gumID = gumSpec.getGumID();
		long id = gumResponseEntry.getRequestID();

		this.mockGump.disposeGum( EasyMock.eq( id ), EasyMock.eq( gumID ) );
		EasyMock.expectLastCall().once();

		EasyMock.replay( mockGumpClient );
		EasyMock.replay( mockGump );
		EasyMock.replay( facade );

		this.consumerThreadManager.createAndExecuteDelivererThread( gumpClientID, gumResponseEntry );

		ThreadServices.waitUntilWorkIsDone();
		ThreadPoolServices.waitUntilWorkIsDone();

		EasyMock.verify( mockGumpClient );
		EasyMock.verify( mockGump );
		EasyMock.verify( facade );
	}
}
