/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.security;

import org.apache.log4j.Level;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.executor.Executor;
import org.ourgrid.common.executor.ExecutorException;
import org.ourgrid.common.executor.ExecutorFactory;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.ExecutorResult;
import org.ourgrid.common.logger.MGLogWriter4j;
import org.ourgrid.common.xml.XMLLayout;
import org.ourgrid.common.xml.XMLLogWriter4j;

/**
 * This class provides an abstraction to use the KeyTool features. The idea is
 * to generate self-signed certificates, as well import and export signatures to
 * the user KeyStore file. The keystore file is located at the
 * $MGROOT/var/$USER.keys by deafult.
 */
public class KeyToolWrapper {

	/** Determine the maximum keysize supported by the KeyTool */
	protected static final int MAXSIZE = 1024;

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( KeyToolWrapper.class );

	/** The keytool command line */
	protected static final String KEYTOOL = "keytool";

	/** Generation a key entry option */
	public static final String GENKEY = "-genkey";

	/** Import certificates to the KeyStore option */
	public static final String IMPORT = "-import";

	/** Export certificates from a KeyStore */
	public static final String EXPORT = "-export";

	/** Keystore file specification option */
	protected static final String KEYSTORE = "-keystore";

	/** Exported certificate file specification option */
	protected static final String FILE = "-file";

	/** the keysize option that determines the size in bits of the generated key */
	protected static final String KEYSIZE = "-keysize";

	/** This option determines the validity of the certificate */
	protected static final String VALIDITY = "-validity";

	/** Option to specify personal information of the certificate */
	protected static final String DNAME = "-dname";

	/** This option is used to specify a password associated to the KeyStore */
	protected static final String STOREPASS = "-storepass";

	/** This option is used to specify a password associated to the generated Key */
	protected static final String KEYPASS = "-keypass";

	/** This option is used to specify an alias to identfy the certificate */
	protected static final String ALIAS = "-alias";

	/** This option is used to list KEYSTORE content */
	protected static final String LIST = "-list";

	/** To execute the KeyTool application */
	protected Executor executor;

	/** Store the result of the execution of the KeyTool */
	protected ExecutorResult result;


	/**
	 * This method constructs a new instance of the
	 */
	public KeyToolWrapper() {

		/** Get an instance of the executor class */
		executor = ExecutorFactory.getExecutorInstance();

		/** **** Configuring Log4j ******* */
		org.apache.log4j.Logger root = org.apache.log4j.Logger.getRootLogger(); // root
																				// logger

		// creating appender
		MGLogWriter4j txtAppender = MGLogWriter4j.getInstance( new LogMessagesLayout(), Configuration.getInstance()
				.getLogPath(), false );
		XMLLogWriter4j xmlAppender = XMLLogWriter4j.getInstance( new XMLLayout(), Configuration.getInstance()
				.getXMLLogPath(), false );

		// adding appender
		root.addAppender( txtAppender );
		root.addAppender( xmlAppender );

		root.setLevel( Level.DEBUG );
		/** ************************************* */
	}


	/**
	 * This method provides an interface to the key entry generation into the
	 * Keystore specified.
	 * 
	 * @param alias an alias to identify the certificate
	 * @param commonName the CN field that is used to identify the key entry.
	 * @param keystoreName the filename where the key entries are stored.
	 * @param keysize the size of generated key in bytes
	 * @param validity how many days the key will be a valid key
	 * @return <i>true</i> if the execution has been sucessfull. <i>false</i>
	 *         otherwise.
	 */
	public boolean generateKey( String alias, String commonName, String keystoreName, int keysize, int validity ) {

		/** Check if the validity is positive and grater than zero */
		if ( validity <= 0 ) {
			LOG.info( "Certificate VALIDITY was lower than 0 (zero), this value was fixed to 365 (a year)" );

			validity = 365;
		}

		/** Check if the keysize is grater than the supported size */
		if ( (keysize <= 0) || ((keysize % 64) != 0) || (keysize > MAXSIZE) ) {
			keysize = 1024;
		}

		/** Check if the alias already exists */
		if ( this.aliasExists( alias, keystoreName ) ) {
			LOG.info( "The alias <" + alias + "> already exists in the keystore <" + keystoreName + ">" );
			return true;
		}

		String commandLine = KEYTOOL + " " + GENKEY + " " + ALIAS + " " + alias + " " + DNAME + " \"CN=" + commonName
				+ "\" " + KEYSIZE + " " + keysize + " " + VALIDITY + " " + validity + " " + STOREPASS + " " + alias
				+ " " + KEYSTORE + " " + keystoreName + " " + KEYPASS + " " + alias;

		return this.execute( commandLine );
	}


	/**
	 * This method perform the key importation and exportation. The operation is
	 * defined by the parameter and the fields <i>KeyToolWrapper.IMPORT</i> and
	 * <i>KeyToolWrapper.EXPORT</i> must be used.
	 * 
	 * @param alias the identificator of the key that must be transported
	 * @param keystoreName the name of the file that store the certificates
	 * @param certificateFilename the name of the file to/from the certificate
	 *        goes/comes
	 * @param operation the operation that must be performed.
	 *        <i>KeyToolWrapper.IMPORT</i> and <i>KeyToolWrapper.EXPORT</i>
	 * @return <i>true</i> if the operation was successfully executed <i>false</i>
	 *         otherwise.
	 */
	public boolean transportCertificate( String alias, String keystoreName, String certificateFilename, String operation ) {

		if ( !(operation.equals( IMPORT ) || operation.equals( EXPORT )) ) {
			return false;
		}

		String commandLine = KEYTOOL + " " + operation + " " + ALIAS + " " + alias + " " + STOREPASS + " " + alias
				+ " " + KEYSTORE + " " + keystoreName + " " + KEYPASS + " " + alias + " " + FILE + " "
				+ certificateFilename;

		return this.execute( commandLine );
	}


	/**
	 * This method verifies the existence of the alias into the keystore.
	 * 
	 * @param alias The alias name
	 * @param keystoreName The name of the file that stores the certificates
	 * @return <i>true</i> if alias already exists. <i>false</i> otherwise.
	 */
	protected boolean aliasExists( String alias, String keystoreName ) {

		String commandLine = KEYTOOL + " " + LIST + " " + ALIAS + " " + alias + " " + STOREPASS + " " + alias + " "
				+ KEYSTORE + " " + keystoreName;

		/**
		 * It is needed to put a not operator. Since the execute() method return
		 * true in success.
		 */
		return (this.execute( commandLine ));
	}


	/**
	 * This method executes the key tool command that is determined by the
	 * public methods.
	 * 
	 * @param commandLine the command line that represents a key generation,
	 *        certificate importation or exportation.
	 * @return <i>true</i> if the execution has been successfull. <i>false</i>
	 *         otherwise.
	 */
	protected boolean execute( String commandLine ) {

		if ( commandLine != null ) {

			try {
				ExecutorHandle handle = this.executor.execute( ".", commandLine );
				this.result = this.executor.getResult( handle );

			} catch ( ExecutorException e ) {

				LOG.info( "Problems when execution KeyTool operation." );
				LOG.debug( e.getMessage() );

				return false;
			}

		} else {
			return false;
		}

		/** Print the return of the KeyTool execution */
		LOG.debug( this.result.getStdout() );

		return (this.result.getExitValue() == 0);
	}
}
