/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.security;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.net.ServerSocket;
import java.rmi.server.RMIServerSocketFactory;
import java.security.KeyStore;

import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLServerSocketFactory;

import org.ourgrid.common.config.Configuration;

/**
 * The SSL Server SocketFactory. Based on the Sun's example.
 */
public class MGSecureServerSocketFactory implements RMIServerSocketFactory, Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/**
	 * The password to the keystore.
	 */
	private String keyStorePassword;

	/**
	 * The keystore filename.
	 */
	private String keyStoreFilename;

	private static MGSecureServerSocketFactory instance;


	/**
	 * Creates a MGSecureServerSocketFactory.
	 * 
	 * @param keyStoreFilename The keystore's filename.
	 * @param keyStorePassword The keystore's password.
	 */
	protected MGSecureServerSocketFactory( String keyStoreFilename, String keyStorePassword ) {

		this.keyStorePassword = keyStorePassword;
		this.keyStoreFilename = keyStoreFilename;
	}


	/**
	 * Creates a server socket that will only accept connections if it is
	 * certified.
	 * 
	 * @param port The port.
	 */
	public ServerSocket createServerSocket( int port ) throws IOException {

		SSLServerSocketFactory ssf = null;
		try {
			// set up key manager to do server authentication
			SSLContext ctx;
			KeyManagerFactory kmf;
			KeyStore ks;
			char[ ] password = keyStorePassword.toCharArray();

			ctx = SSLContext.getInstance( "TLS" );
			kmf = KeyManagerFactory.getInstance( "SunX509" );
			ks = KeyStore.getInstance( "JKS" );

			ks.load( new FileInputStream( keyStoreFilename ), password );
			kmf.init( ks, password );
			ctx.init( kmf.getKeyManagers(), null, null );

			ssf = ctx.getServerSocketFactory();
		} catch ( Exception e ) {
			throw new IOException( "SecureServerSocket cannot be created! " + e.getMessage() );
		}
		return ssf.createServerSocket( port );
	}


	public synchronized static MGSecureServerSocketFactory getInstance() {

		if ( instance == null ) {
			String keystoreFileName = Configuration.getInstance().getProperty( Configuration.PROP_KEYSTORE_FILENAME );
			String keystorePassword = Configuration.getInstance().getProperty( Configuration.PROP_KEYSTORE_PASSWORD );
			instance = new MGSecureServerSocketFactory( keystoreFileName, keystorePassword );
		}
		return instance;
	}

}
