/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.security;

import java.io.File;
import java.rmi.Naming;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.ExportException;
import java.rmi.server.RMIClientSocketFactory;
import java.rmi.server.RMIServerSocketFactory;
import java.rmi.server.UnicastRemoteObject;

import junit.framework.TestCase;

import org.ourgrid.common.config.ErrorCode;
import org.ourgrid.test.SecurityTest;
import org.ourgrid.test.TestUtils;

/**
 * Tests the MGSecureClientSocketFactory and the MGSecureServerSocketFactory.
 */
public class MGSecureSocketFactoryTest extends TestCase implements SecurityTest {

	private TestUtils testUtils;

	private int port;


	/**
	 * Sets up the environment to the test.
	 */
	public void setUp() throws Exception {

		super.setUp();
		testUtils = new TestUtils();
		testUtils.load();
		port = Integer.parseInt( testUtils.getProperty( TestUtils.RMI_PORT ) );

		try {
			LocateRegistry.createRegistry( port );
		} catch ( ExportException e ) {
			LocateRegistry.getRegistry( port );
		}

	}


	/**
	 * Tests the SecureSocketFactories.
	 */
	public void testSocketFactory() throws Exception {

		System.setProperty( "javax.net.ssl.keyStore", testUtils.getProperty( TestUtils.KEYSTORE_FILENAME ) );
		System.setProperty( "javax.net.ssl.keyStorePassword", testUtils.getProperty( TestUtils.KEYSTORE_PASSWORD ) );

		String keystoreFileName = testUtils.getProperty( TestUtils.KEYSTORE_FILENAME );
		String password = testUtils.getProperty( TestUtils.KEYSTORE_PASSWORD );
		Hello secureHello = new HelloImpl( 0, new MGSecureClientSocketFactory(), new MGSecureServerSocketFactory(
				keystoreFileName, password ) );

		String url = "rmi://" + testUtils.getProperty( TestUtils.MG_HOMEMACHINE ) + ":" + port + "/SecureHello";
		System.out.println( "Binding 'Hello' in " + url );
		Naming.rebind( url, secureHello );
		System.out.println( "Hello bound." );

		String urlCheck = url;
		String[ ] helloCheckerCommand = {
											"java",
											"-Djavax.net.ssl.trustStore="
													+ testUtils.getProperty( TestUtils.TRUSTSTORE_FILENAME ),
											"-Djavax.net.ssl.trustStorePassword="
													+ testUtils.getProperty( TestUtils.TRUSTSTORE_PASSWORD ), "-cp",
											"lib" + File.separator + "ourgrid.jar",
											"org.ourgrid.common.security.HelloChecker", urlCheck,
											Integer.toString( port ) };
		TestUtils.runCommand( helloCheckerCommand, 0 );

		try {
			String[ ] anotherHelloCheckerCommand = {
													"java",
													"-Djavax.net.ssl.trustStore="
															+ testUtils
																	.getProperty( TestUtils.INVALID_TRUSTSTORE_FILENAME ),
													"-Djavax.net.ssl.trustStorePassword="
															+ testUtils
																	.getProperty( TestUtils.INVALID_TRUSTSTORE_PASSWORD ),
													"-cp", "lib" + File.separator + "ourgrid.jar",
													"org.ourgrid.common.security.HelloChecker", urlCheck,
													Integer.toString( port ) };
			TestUtils.runCommand( anotherHelloCheckerCommand, 0 );
			fail( "An exception should be thrown." );
		} catch ( Exception e ) {
			System.out.println( "================================================" );
			System.out.println( "I'm sure there was lots of error messages above," );
			System.out.println( "but this is expected! =)" );
			System.out.println( "================================================" );
			System.out.println( "EXPECTED EXCEPTION: " + e.getMessage() );
		}
	}

}


/**
 * Interface only to be used in tests.
 * 
 * @author Guilherme Mauro Germoglio - guiga@dsc.ufcg.edu.br
 */
interface Hello extends Remote {

	/**
	 * Returns a well known message.
	 * 
	 * @return A well known message.
	 */
	public String sayHello() throws RemoteException;

}


class HelloImpl extends UnicastRemoteObject implements Hello {

	private static final long serialVersionUID = 33L;

	private String message;


	/**
	 * Creates a HelloImpl.
	 * 
	 * @param port The port where it will be exported.
	 * @param csf The ClientSocketFactory it will use.
	 * @param ssf The ServerSocketFactory it will use.
	 * @throws RemoteException If there any problem with communication.
	 */
	protected HelloImpl( int port, RMIClientSocketFactory csf, RMIServerSocketFactory ssf ) throws RemoteException {

		super( port, csf, ssf );
		this.message = "Hello World!";
	}


	/**
	 * @see org.ourgrid.common.security.Hello#sayHello()
	 */
	public String sayHello() throws RemoteException {

		return this.message;
	}

}


class HelloChecker {

	public static void main( String[ ] args ) throws RemoteException {

		int port = Integer.parseInt( args[1] );
		try {
			LocateRegistry.createRegistry( port );
		} catch ( ExportException e ) {
			LocateRegistry.getRegistry( port );
		}

		Hello remoteHello;
		try {
			String url = args[0];
			remoteHello = (Hello) Naming.lookup( url );
			if ( "Hello World!".equals( remoteHello.sayHello() ) ) {
				System.out.println( "Hello ok!" );
				System.exit( ErrorCode.OK );
			} else {
				System.out.println( "Hello not ok!" );
				System.exit( ErrorCode.FAIL );
			}
		} catch ( Exception e ) {
			System.out.println( "Hello not ok!" );
			e.printStackTrace();
			System.exit( ErrorCode.FAIL );
		}
	}
}
