/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.spec;

import java.io.Serializable;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

import org.ourgrid.common.id.GumID;

/**
 * Entity that encapsulates all the information given by the user about a Gum To
 * provide the information, the user uses the Description Files that can be
 * compiled by CommonCompiler.
 * 
 * @see org.ourgrid.common.spec.main.CommonCompiler
 */
public class GumSpec implements Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/**
	 * The attributes of a gum describe its features, like memory size, OS, etc.
	 */
	private Map<String,String> attributes;

	/** A unique identification for the gum */
	private GumID gumID = null;

	/** A series of standard attribute names */

	public static final String ATT_NAME = "name";

	public static final String ATT_PEERNAME = "peername";

	public static final String ATT_PEERPORT = "peerport";

	public static final String ATT_TYPE = "type";

	public static final String ATT_REM_EXEC = "remexec";

	public static final String ATT_PORT = "port";

	public static final String ATT_COPY_FROM = "copyfrom";

	public static final String ATT_COPY_TO = "copyto";

	public static final String ATT_PLAYPEN_ROOT = "playpenroot";

	public static final String ATT_PLAYPEN_SIZE = "playpensize";

	public static final String ATT_STORAGE_DIR = "storagedir";

	public static final String ATT_STORAGE_SIZE = "storagesize";

	public static final String ATT_OS = "os";

	public static final String ATT_PROCESSOR_FAMILY = "processorfamily";

	public static final String ATT_MEM = "mem";

	public static final String ATT_BOGO_MIPS = "bogomips";

	public static final String ATT_DEBUG = "debug";

	public static final String ATT_SECURITY = "security";

	public static final String ATT_SITE = "site";

	public static final String ATT_STORAGE_SHARED = "storageshared"; // Added
																		// by
																		// Hunter
																		// Hagewood
																		// 02/03/2005

	public static final String ATT_ENVIRONMENT = "environment"; // Added by
																// Gustavo and
																// Alisson on
																// 12/12/2005

	public static final String TYPE_UA_LINUX = "ualinux";

	public static final String TYPE_UA_WINDOWS = "uawindows";

	public static final String TYPE_UA_SWAN = "uaswan";

	public static final String TYPE_FAKE_GUM = "fakegum";

	public static final String OS_LINUX = "linux";

	public static final String OS_WINDOWS = "windows";

	public static final String OS_HPUX = "hpux";

	public static final String OS_SOLARIS = "solaris";

	public static final String OS_LINUX_XEN = "/proc/xen";


	/**
	 * Default constructor.
	 */
	public GumSpec() {

		this.attributes = new TreeMap<String,String>( new IgnoreCaseComparator() );
	}


	/**
	 * Builds a machine specification with a given map of attributes.
	 * 
	 * @param attributes A map with the pairs of name + value of the machine's
	 *        attributes
	 */
	public GumSpec( Map<String,String> attributes ) {

		this.attributes = attributes;
	}


	/**
	 * Retrieves the map of attributes.
	 * 
	 * @return a map with all the attributes of this grid machine indexed by the
	 *         name of the attribute.
	 */
	public Map<String,String> getAttributes() {

		return this.attributes;
	}


	/**
	 * Inserts a new attribute at this gum specification.
	 * 
	 * @param name the name of the new attribute. If the name already exists the
	 *        new value will substitute the old one.
	 * @param value the value for this attribute.
	 */
	public void putAttribute( String name, String value ) {

		this.attributes.put( name, value );
	}


	/**
	 * Inserts a set of attributes at this gum specification.
	 * 
	 * @param mapAttributes the attributes to be added.
	 */
	public void putAttributes( Map<String,String> mapAttributes ) {

		this.attributes.putAll( mapAttributes );
	}


	/**
	 * Removes all the given attributes.
	 * 
	 * @param keys The keys to be removed.
	 */
	public void removeAttributes( Collection<String> keys ) {

		Iterator<String> it = keys.iterator();
		while ( it.hasNext() ) {
			String currentKey = it.next();
			this.attributes.remove( currentKey );
		}
	}


	/**
	 * Checks if the given attribute name is contained in the machine attribute
	 * set.
	 * 
	 * @param name the attribute name.
	 * @return true if the attribute name already exists and false otherwise.
	 */
	public boolean hasAttribute( String name ) {

		return this.attributes.containsKey( name );
	}


	/**
	 * Returns the value of the attribute named as given.
	 * 
	 * @param attName the attribute name.
	 * @return the attribute value.
	 */
	public String getAttribute( String attName ) {

		Object value = this.attributes.get( attName );
		if ( value != null ) {
			return (String) value;
		}
		return null;
	}


	/**
	 * Checks if an object is equals to this grid machine specification.
	 * 
	 * @param o the object that has to be compared.
	 * @return true if the object is equals and false otherwise
	 */
	@Override
	public boolean equals( Object o ) {

		if ( o != null && o instanceof GumSpec ) {
			GumSpec gum = (GumSpec) o;
			return this.attributes.equals( gum.attributes );
		}
		return false;
	}


	/**
	 * Returns the gum identification.
	 * 
	 * @return the gum identification
	 */
	public GumID getGumID() {

		if ( gumID == null )
			gumID = new GumID( this );
		return gumID;
	}

	/**
	 * A special comparator to use with the attribute map. It will make the
	 * comparisons of attributes case insensitive
	 */
	class IgnoreCaseComparator implements Comparator<String>, Serializable {

		private static final long serialVersionUID = 33L;


		public int compare( String o1, String o2 ) {

			return o1.compareToIgnoreCase( o2 );
		}
	}
}
