/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.spec;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

/**
 * This entity handles the input and output entries for a task.
 * 
 * @see IOEntry Created on Jul 1, 2004
 */
public class IOBlock implements Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	HashMap<String,Vector<IOEntry>> entries = new HashMap<String,Vector<IOEntry>>();


	/**
	 * An empty constructor
	 */
	public IOBlock() {

	}


	/**
	 * Inserts a new input/output entry at this block of I/O commands.
	 * 
	 * @param condition The condition that tells if the I/O command will be
	 *        executed. It happens only when the command was written inside a
	 *        if/else block.
	 * @param entry A IOEntry object that defines the command and the paths of
	 *        origin and destiny of a file.
	 */
	public void putEntry( String condition, IOEntry entry ) {

		Vector<IOEntry> set;
		if ( entries.containsKey( condition ) ) {
			set = entries.get( condition );
		} else {
			set = new Vector<IOEntry>();
		}
		set.add( entry );
		entries.put( condition, set );
	}


	/**
	 * Inserts a new input/output entry at this block of I/O commands, using as
	 * condition the empty string, that means this entry will always be used.
	 * 
	 * @param entry The input/output entry.
	 */
	public void putEntry( IOEntry entry ) {

		this.putEntry( "", entry );
	}


	/**
	 * Tells how many entries this I/O block has. Notice that the lenght means
	 * how many conditions blocks where inserted and not how many commands has
	 * been inserted.
	 * 
	 * @return How many entries this block has.
	 */
	public int length() {

		return entries.size();
	}


	/**
	 * Returns a collection with the entries related with a condition. To obtain
	 * all the conditions at this block use this.getConditions.
	 * 
	 * @param condition The condition that indexes all the entries that will be
	 *        used if it is true.
	 * @return The collection of entries related with a condition - null if the
	 *         condition does not exist.
	 */
	public Vector<IOEntry> getEntry( String condition ) {

		if ( entries.containsKey( condition ) ) {
			return entries.get( condition );
		}
		return null;

	}


	/**
	 * Returns all the conditions that indexes the entry blocks.
	 * 
	 * @return A iterator with all the valid conditions that indexes the entries
	 *         at this input/output block.
	 */
	public Iterator<String> getConditions() {

		return entries.keySet().iterator();
	}


	/**
	 * Returns a string representation of an IOBlock.
	 */
	@Override
	public String toString() {

		StringBuffer message = new StringBuffer();
		Iterator<String> it = entries.keySet().iterator();
		while ( it.hasNext() ) {
			String key = it.next();
			message.append( "    Condition: [" + key + "]\n" );
			Iterator<IOEntry> it1 = (entries.get( key )).iterator();
			while ( it1.hasNext() ) {
				message.append( "      IOEntry: [" + it1.next().toString() + "]\n" );
			}
		}
		return message.toString();
	}

}
