/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.spec;

import java.io.Serializable;
import java.util.List;

import org.ourgrid.common.spec.exception.JobSpecificationException;

/**
 * Entity that encapsulates all the information given by the user about each
 * job. To inform, the user uses the Description Files that can be compiled by
 * the CommonCompiler.
 * 
 * @see org.ourgrid.common.spec.main.CommonCompiler
 * @version 1.0
 */
public class JobSpec implements Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	private List<TaskSpec> taskSpecs;

	private String requirements, label;


	/**
	 * Constructor.
	 * 
	 * @param label The label for the job.
	 * @param requirements The logical expression that defines the job. It will
	 *        be used to choose machines that are able to run its tasks. To
	 *        define it well, check the OurGrid manual.
	 * @param taskSpecs A list with all the task specifications of this job.
	 */
	public JobSpec( String label, String requirements, List<TaskSpec> taskSpecs ) throws JobSpecificationException {

		this.label = label;
		this.requirements = requirements;
		this.taskSpecs = taskSpecs;
		validate();
	}


	/**
	 * This method validates the attributes of this Job Spec
	 */
	private void validate() throws JobSpecificationException {

		if ( this.taskSpecs == null ) {
			throw new JobSpecificationException( "A Job Spec could not be initialized with a null list of Task Specs." );
		}

		if ( this.taskSpecs.size() == 0 ) {
			throw new JobSpecificationException(
					"A Job Spec could not be initialized with an empty list of Task Specs." );
		}

		if ( this.taskSpecs.contains( null ) ) {
			throw new JobSpecificationException(
					"A Job Spec could not contain a null element into the list of Task Specs." );
		}

	}


	/**
	 * The constructor
	 * 
	 * @param label The label for the job.
	 */
	public JobSpec( String label ) {

		this.label = label;
		this.requirements = "";
	}


	/**
	 * @return A list with the tasks specification in this job.
	 */
	public List<TaskSpec> getTaskSpecs() {

		return taskSpecs;
	}


	/**
	 * Inserts a list of task specifications.
	 * 
	 * @param taskSpecs The list of tasks that will be contained by this job.
	 */
	public void setTaskSpecs( List<TaskSpec> taskSpecs ) throws JobSpecificationException {

		this.taskSpecs = taskSpecs;
		validate();
	}


	/**
	 * @return The logical expression that will be used to choose machines to
	 *         run the tasks in this job.
	 */
	public String getRequirements() {

		return this.requirements;
	}


	/**
	 * Sets the logical expression for the job.
	 * 
	 * @param expression The logical expression that defines the job. It will be
	 *        used to choose machines that are able to run its tasks.
	 */
	public void setRequirements( String expression ) {

		this.requirements = expression;
	}


	/**
	 * @return The label of the job.
	 */
	public String getLabel() {

		return this.label;
	}
}
