/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.spec;

import java.io.File;
import java.io.Serializable;

import org.ourgrid.common.url.PeerURLProvider;

/**
 * This GuMPSpec identifies the container of informations about the remote GUMPS
 * that the local instance of MyGrid will be able to use to require aditional
 * processing power. Created on Jul 9, 2004
 */
public class PeerSpec implements Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/**
	 * The remote gump name. It is the Internet name of the machine, it means
	 * that the machine can be contacted into the internet by this name (or IP)
	 */
	private String name;

	/**
	 * The port the gump is listening.
	 */
	private int port;

	/**
	 * The GuMP label.
	 */
	private String label;


	/**
	 * Constructor.
	 * 
	 * @param name The gump machine internet name.
	 * @param port The gump machine port.
	 */
	public PeerSpec( String name, int port ) {

		super();
		this.name = name;
		this.port = port;
	}


	/**
	 * Returns the gump name.
	 * 
	 * @return The internet name of the machine
	 */
	public String getName() {

		if ( name == null ) {
			return "";
		}

		return name;
	}


	/**
	 * Returns the gump port.
	 * 
	 * @return The port.
	 */
	public int getPort() {

		return port;
	}


	public String getLabel() {

		if ( label == null ) {
			return "";
		}
		return label;
	}


	/**
	 * Returns the url that is like :
	 * rmi://MachineName:MachinePort/REMOTE_GUMP_OBJECT_NAME.
	 * 
	 * @return The URL of the gump.
	 */
	public String getURL() {

		return "rmi://" + name + ":" + port + File.separator + PeerURLProvider.LOCAL_ACCESS;
	}


	public void setLabel( String label ) {

		this.label = label;
	}


	@Override
	public boolean equals( Object o ) {

		if ( !(o instanceof PeerSpec) ) {
			return false;
		}
		PeerSpec otherSpec = (PeerSpec) o;
		return this.getURL().equals( otherSpec.getURL() ) && this.getLabel().equals( otherSpec.getLabel() );
	}


	@Override
	public int hashCode() {

		return this.getURL().concat( this.getLabel() ).hashCode();
	}
}
