/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.spec;

import java.io.Serializable;
import java.rmi.RemoteException;

import org.ourgrid.common.gump.GumpClient;
import org.ourgrid.common.id.ObjectID;

/**
 * The class represents a request made by a consumer. It stores information
 * about a request.
 */
public class RequestSpec implements Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/** An ID that uniquely identifies this request. */
	private long requestID;

	/** Represents the expression that must be satisfied by processor. */
	private String requirements;

	/**
	 * This is useful for limiting the consumption of resources by the consumer,
	 * if it is needed. Otherwise, if it's not needed, the consumer can request
	 * an unlimited number of processors. Integer.MAX_VALUE means an unlimited
	 * number of processors.
	 */
	private int requiredGums;

	/** The entity which is requesting gums. This entity creates the request */
	private GumpClient gumpClient;

	/**
	 * Defines if the request has been received from a local broker or from a
	 * remote peer.
	 */
	private RequestSource source;

	/**
	 * <code>ObjectID</code> representing <code>GumpClient</code>
	 * responsible f or this request.
	 */
	private ObjectID objectID;


	/**
	 * Default constructor.
	 * 
	 * @param requestID request's identifier.
	 * @param requirements represents the expression that must be satisfied by
	 *        processor.
	 * @param maxNumberOfGums number of processors requested. Integer.MAX_VALUE
	 *        is unlimited number of processors.
	 * @param gumpClient a reference to the consumer
	 */
	public RequestSpec( long requestID, String requirements, int maxNumberOfGums, GumpClient gumpClient ) {

		this.requestID = requestID;
		this.requirements = requirements;
		this.requiredGums = maxNumberOfGums;
		// assumes by default that a request is from local brokers
		this.source = RequestSource.FROM_LOCAL_BROKER;

		setGumpClient( gumpClient );
	}


	/**
	 * Sets a <code>gumpClient</code> and caches its object id.
	 * 
	 * @param gumpClient
	 */
	private void setGumpClient( GumpClient gumpClient ) {

		this.gumpClient = gumpClient;
		try {
			this.objectID = gumpClient.getObjectID();
		} catch ( RemoteException e ) {
			// Unexpected, remote object is local.
		}
	}


	/**
	 * Gets the <code>ObjectID</code> representing the <code>GumpClient</code>
	 * responsible for this request.
	 * 
	 * @return the cached object identification of the consumer
	 */
	public ObjectID getObjectID() {

		return this.objectID;
	}


	public long getID() {

		return this.requestID;
	}


	public String getRequirements() {

		return requirements;
	}


	public int getRequiredGums() {

		return this.requiredGums;
	}


	public GumpClient getGumpClient() {

		return this.gumpClient;
	}


	public RequestSource getSource() {

		return this.source;
	}


	public void setAsFromLocalBroker() {

		this.source = RequestSource.FROM_LOCAL_BROKER;
	}


	public void setAsFromRemotePeer() {

		this.source = RequestSource.FROM_REMOTE_PEER;
	}


	public boolean isFromALocalBroker() {

		return this.source.isLocal();
	}


	public boolean isFromARemotePeer() {

		return this.source.isRemote();
	}


	public String toString() {

		return "ID: " + String.valueOf( this.getID() ) + " Is Local: " + isFromALocalBroker() + " Source ID: "
				+ objectID;
	}

	/**
	 * Indicates the request source (from local brokers or from remote peers).
	 */
	public enum RequestSource {

		/** Types of source allowed */
		FROM_LOCAL_BROKER, FROM_REMOTE_PEER;

		public boolean isLocal() {

			return (this.equals( FROM_LOCAL_BROKER ));
		}


		public boolean isRemote() {

			return (this.equals( FROM_REMOTE_PEER ));
		}
	}

}
