/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.spec;

import java.io.Serializable;

import org.ourgrid.common.spec.exception.TaskSpecificationException;

/**
 * Entity that encapsulates all the infomations given by the user about each
 * task. To inform, the user uses the Description Files that can be compiled by
 * CommonCompiler.
 * 
 * @version 1.0
 * @see org.ourgrid.common.spec.main.CommonCompiler
 */
public class TaskSpec implements Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/**
	 * The remote part of a task specification.
	 */
	private String remoteExec;

	/**
	 * The init part of a task specification.
	 */
	private IOBlock initBlock;

	/**
	 * The final part of a task specification.
	 */
	private IOBlock finalBlock;


	/**
	 * The constructor.
	 * 
	 * @param initBlock The information about the files that will be transfered
	 *        to the remote machine that will be used at the execution of the
	 *        task.
	 * @param remoteExec The command that will be run at the remote machine to
	 *        execute the task.
	 * @param finalBlock The information about the files that will be transfered
	 *        from the remote machine as results of the execution.
	 * @throws TaskSpecificationException If the informations about the task are
	 *         not valid.
	 */
	public TaskSpec( IOBlock initBlock, String remoteExec, IOBlock finalBlock ) throws TaskSpecificationException {

		this.initBlock = initBlock;
		this.remoteExec = remoteExec;
		this.finalBlock = finalBlock;
		validate();
	}


	/**
	 * Validates the informations passed as arguments to this task specification
	 * 
	 * @throws TaskSpecificationException If the task specs are not valid
	 */
	private void validate() throws TaskSpecificationException {

		if ( remoteExec == null || remoteExec == "" )
			throw new TaskSpecificationException( CompilerMessages.BAD_TASK_SPEC_REMEXEC_MISSING );
	}


	/**
	 * Gets the init part of a task.
	 * 
	 * @return Returns the initBlock.
	 */
	public IOBlock getInitBlock() {

		return initBlock;
	}


	/**
	 * Gets the final part of a task.
	 * 
	 * @return Returns the finalBlock.
	 */
	public IOBlock getFinalBlock() {

		return finalBlock;
	}


	/**
	 * Gets the remote part of a task.
	 * 
	 * @return Returns the remoteExec.
	 */
	public String getRemoteExec() {

		return remoteExec;
	}


	@Override
	public String toString() {

		StringBuffer message = new StringBuffer();

		message.append( "   INIT PHASE:\n" );
		if ( initBlock != null ) {
			message.append( initBlock.toString() );
		} else {
			message.append( "      empty!" );
		}

		message.append( "   REMOTE PHASE:\n" );
		message.append( "      " + remoteExec + "\n" );

		message.append( "   FINAL PHASE:\n" );
		if ( finalBlock != null ) {
			message.append( finalBlock.toString() );
		} else {
			message.append( "      empty!" );
		}

		return message.toString();

	}


	@Override
	public boolean equals( Object o ) {

		if ( o instanceof TaskSpec ) {
			TaskSpec taskSpec = (TaskSpec) o;
			return toString().equals( taskSpec.toString() );
		}

		return false;
	}

}
