/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.spec.semantic;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

import org.ourgrid.common.config.UserAgentConfiguration;
import org.ourgrid.common.spec.CompilerMessages;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.semantic.exception.SemanticException;
import org.ourgrid.common.spec.syntactical.CommonSyntacticalAnalyzer;
import org.ourgrid.common.spec.token.Token;
import org.ourgrid.gridmachine.services.GumSpecServices;

/**
 * This entity is the set of actions that the GRID grammar uses to build a
 * answer to the compilation of sources wrote in this language. Created on Jul
 * 8, 2004
 */
public class SDFSemanticActions implements SemanticActions {

	private Stack<String> stack;

	private Token actualToken;

	private int mode = CommonSyntacticalAnalyzer.MODE_NORMAL;

	private HashMap<String,String> defaultAttrib = null;

	private boolean defaults;

	private LinkedList<GumSpec> gums = new LinkedList<GumSpec>();

	private GumSpec actualGUM = null;

	private List<GumSpec> result;


	/**
	 * The Constructor
	 */
	public SDFSemanticActions() {

		this.stack = new Stack<String>();
	}


	/**
	 * @see org.ourgrid.common.spec.semantic.SemanticActions#performAction(java.lang.String,
	 *      org.ourgrid.common.spec.token.Token)
	 */
	public void performAction( String action, Token token ) throws SemanticException {

		this.actualToken = token;
		try {
			Class semantic = Class.forName( this.getClass().getName() );
			Method method = semantic.getMethod( action );
			method.invoke( this );
		} catch ( NoSuchMethodException nsmex ) {
			throw new SemanticException( CompilerMessages.SEMANTIC_ACTION_NOT_FOUND, nsmex );
		} catch ( ClassNotFoundException cnfex ) {
			throw new SemanticException( CompilerMessages.SEMANTIC_CLASS_NOT_FOUND, cnfex );
		} catch ( InvocationTargetException itex ) {
			if ( itex.getCause() instanceof SemanticException ) {
				throw (SemanticException) itex.getCause();
			}
			throw new SemanticException( CompilerMessages.SEMANTIC_FATAL_ERROR(), itex.getCause() );
		} catch ( IllegalAccessException iaex ) {
			throw new SemanticException( CompilerMessages.SEMANTIC_FATAL_ILLEGAL_ACCESS, iaex );
		}
	}


	/**
	 * @see org.ourgrid.common.spec.semantic.SemanticActions#getOperationalMode()
	 */
	public int getOperationalMode() {

		return this.mode;
	}


	/**
	 * @see org.ourgrid.common.spec.semantic.SemanticActions#getResult()
	 */
	public List<GumSpec> getResult() {

		return this.result;
	}


	/**
	 * This action: Sets the read line mode
	 */
	public void action1() {

		this.mode = CommonSyntacticalAnalyzer.MODE_READLINE;
	}


	/**
	 * This action: Puts the value string for a attribute at the top of the
	 * stack.
	 * 
	 * @throws SemanticException
	 */
	public void action2() throws SemanticException {

		String tempAttValue = actualToken.getSymbol();
		if ( tempAttValue.equals( "" ) ) {
			throw new SemanticException( CompilerMessages.SEMANTIC_EMPTY_ATTRIBUTE_VALUE( stack.pop(), actualToken
					.getLine() ) );
		}
		this.stack.push( tempAttValue );
	}


	/**
	 * This action: Puts the name string for a attribute at the top of the
	 * stack.
	 */
	public void action3() {

		this.stack.push( actualToken.getSymbol() );
	}


	/**
	 * This action: Inserts a new attribute at the actual GUM specification or
	 * at the default map depending of the type of the attibute (GUM or
	 * default).
	 */
	public void action4() {

		String attValue = stack.pop();
		String attName = stack.pop();
		if ( defaults == false ) {
			this.actualGUM.putAttribute( attName, attValue );
		} else {
			this.defaultAttrib.put( attName, attValue );
		}
	}


	/**
	 * This action: Initializes a new GUM as the actual GUM.
	 */
	public void action6() {

		this.actualGUM = new GumSpec();
	}


	/**
	 * This action: Closes the actual GUM specification and inserts it at MyGUM.
	 * The close process includes the insertion of the default atributtes at the
	 * actual GUMP specification where each default attibute is only inserted if
	 * the entry for a value does not exists.
	 */
	public void action8() throws SemanticException {

		// insert the default attributes at the actualGUM
		Iterator<String> defaultAttribNames = defaultAttrib.keySet().iterator();
		while ( defaultAttribNames.hasNext() ) {
			String defAttName = defaultAttribNames.next();
			if ( !actualGUM.hasAttribute( defAttName ) ) {
				actualGUM.putAttribute( defAttName, defaultAttrib.get( defAttName ) );
			}
		}
		boolean isValid = GumSpecServices.isSpecValid( actualGUM );

		if ( isValid ) {
			this.gums.add( actualGUM );
			this.actualGUM = null;
		} else {
			throw new SemanticException( CompilerMessages.BAD_GUM_DEFINITION( actualGUM
					.getAttribute( UserAgentConfiguration.PROP_NAME ) ) );
		}
	}


	/**
	 * This action:Prepare the environment to read and indicates that the
	 * attributes are the defaults.
	 */
	public void action9() {

		this.defaultAttrib = new HashMap<String,String>();
		this.defaults = true;
	}


	/**
	 * This action: Finishes the default attibutes reading process.
	 */
	public void action10() {

		this.defaults = false;
	}


	/**
	 * This action: Sets the final result LIST object.
	 */
	public void action12() {

		this.result = this.gums;
	}


	/**
	 * This action: Sets the mode to NORMAL
	 */
	public void action14() {

		this.mode = CommonSyntacticalAnalyzer.MODE_NORMAL;
	}

}
