/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.spec.semantic.test;

import java.io.File;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import junit.framework.TestCase;

import org.ourgrid.common.spec.IOBlock;
import org.ourgrid.common.spec.IOEntry;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.common.spec.main.CommonCompiler;
import org.ourgrid.common.spec.main.CompilerException;
import org.ourgrid.common.spec.semantic.JDFSemanticActions;
import org.ourgrid.common.spec.semantic.exception.SemanticException;
import org.ourgrid.common.spec.token.Token;
import org.ourgrid.test.CommonTest;

public class JDFSemanticActionsTest extends TestCase implements CommonTest {

	CommonCompiler compiler;


	/**
	 * Constructor for JobSemanticActionsTest.
	 * 
	 * @param arg0
	 */
	public JDFSemanticActionsTest( String arg0 ) {

		super( arg0 );
	}


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();
		compiler = new CommonCompiler();
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		compiler = null;
	}


	public void testInitRemoteFinalEntries() throws Exception {

		compiler.compile( "test/specs/semantic/JobTest1.jdf", CommonCompiler.JDF_TYPE );
		List result = compiler.getResult();

		JobSpec jobInfo = (JobSpec) result.remove( 0 );
		assertEquals( "testing:label", jobInfo.getLabel() );
		assertEquals( "OS == Linux && PLAYPEN_SIZE >= 200", jobInfo.getRequirements() );

		LinkedList tasksList = (LinkedList) jobInfo.getTaskSpecs();

		TaskSpec taskInfo = (TaskSpec) tasksList.removeFirst();
		// Scripts
		assertEquals( "$STORAGE/execute2 input1", taskInfo.getRemoteExec() );
		// InputBlock
		IOBlock input = taskInfo.getInitBlock();
		assertTrue( input.length() == 3 );
		// InputEntry - without condition
		Vector entries = input.getEntry( "" ); // Means that this entry is the
		// set of I/O commands that have
		// no conditions to be used.
		assertTrue( entries.size() == 1 );
		IOEntry entry = (IOEntry) entries.get( 0 );
		assertEquals( "PUT", entry.getCommand() );
		assertEquals( CommonCompiler.getSourceParentDir() + File.separator + "sample/execute", entry.getSourceFile() );
		assertEquals( ".", entry.getDestiny() );
		// InputEntry - with condition
		entries = input.getEntry( "( PLAYPEN_SIZE < 300 )" );
		assertTrue( entries.size() == 2 );
		entry = (IOEntry) entries.get( 1 );
		assertEquals( "STORE", entry.getCommand() );
		assertEquals( CommonCompiler.getSourceParentDir() + File.separator + "sample/input2", entry.getSourceFile() );
		assertEquals( "input", entry.getDestiny() );
		// InputEntry - with ELSE condition
		entries = input.getEntry( "! ( ( PLAYPEN_SIZE < 300 ) )" );
		assertTrue( entries.size() == 1 );
		entry = (IOEntry) entries.get( 0 );
		assertEquals( "PUT", entry.getCommand() );
		assertEquals( CommonCompiler.getSourceParentDir() + File.separator + "sample/input2", entry.getSourceFile() );
		assertEquals( "$STORAGE/input", entry.getDestiny() );

		taskInfo = (TaskSpec) tasksList.removeFirst();
		assertEquals( "$PLAYPEN/execute input", taskInfo.getRemoteExec() );
		input = taskInfo.getInitBlock();
		assertTrue( input.length() == 0 );
		IOBlock output = taskInfo.getFinalBlock();
		assertTrue( output.length() == 2 );
		entries = output.getEntry( "" ); // Means that this entry is the set
											// of
		// I/O commands that have no conditions
		// to be used.
		assertTrue( entries.size() == 2 );
		entry = (IOEntry) entries.get( 0 );
		assertEquals( "GET", entry.getCommand() );
		assertEquals( "t1", entry.getSourceFile() );
		assertEquals( CommonCompiler.getSourceParentDir() + File.separator + "teste/result:1", entry.getDestiny() );

		entries = output.getEntry( "( ( XPTO >= 2 && IN_PATH T ) || BLA != linux )" );
		assertTrue( entries.size() == 1 );

		taskInfo = (TaskSpec) tasksList.removeFirst();
		assertEquals( "xpto_script:1", taskInfo.getRemoteExec() );
		output = taskInfo.getFinalBlock();
		assertTrue( output.length() == 0 );

	}


	public void testNonValideTask() throws Exception {

		try {
			compiler.compile( "test/specs/semantic/JobTest2.jdf", CommonCompiler.JDF_TYPE );
			fail( "Should fail because exists a non valide task at the file." );
		} catch ( CompilerException e ) {

		}
	}


	public void testCaseInsensitiveFile() throws Exception {

		compiler.compile( "test/specs/semantic/JobTest3.jdf", CommonCompiler.JDF_TYPE );
		List result = compiler.getResult();

		JobSpec jobInfo = (JobSpec) result.remove( 0 );
		assertEquals( "job3_test", jobInfo.getLabel() );
		assertEquals( "( os == linux && ram >= 400 ) OR ( os != solaris && ram <= 300 )", jobInfo.getRequirements() );

		LinkedList tasks = (LinkedList) jobInfo.getTaskSpecs();
		assertEquals( 7, tasks.size() );

		TaskSpec taskToTest = (TaskSpec) tasks.removeFirst();
		assertEquals( "run_app.sh", taskToTest.getRemoteExec() );

		IOBlock output = taskToTest.getFinalBlock();
		Vector entries = output.getEntry( "( os == linux )" );
		IOEntry entry = (IOEntry) entries.get( 0 );
		assertEquals( "get", entry.getCommand() );
		assertEquals( CommonCompiler.getSourceParentDir() + File.separator + "output/outputL.$TASK", entry.getDestiny() );
		assertEquals( "$PLAYPEN/output_l", entry.getSourceFile() );
		entries = output.getEntry( "! ( ( os == linux ) )" );
		entry = (IOEntry) entries.get( 0 );
		assertEquals( "get", entry.getCommand() );
		assertEquals( "/temp/mygrid/output/outputS.$TASK", entry.getDestiny() );
		assertEquals( "$PLAYPEN/output_l", entry.getSourceFile() );

		IOBlock input = taskToTest.getInitBlock();
		entries = input.getEntry( "" );
		entry = (IOEntry) entries.get( 0 );
		assertEquals( "put", entry.getCommand() );
		assertEquals( "$PLAYPEN", entry.getDestiny() );
		assertEquals( CommonCompiler.getSourceParentDir() + File.separator + "entry/input1.txt", entry.getSourceFile() );
		entry = (IOEntry) entries.get( 1 );
		assertEquals( "store", entry.getCommand() );
		assertEquals( "$PLAYPEN/run_app.sh", entry.getDestiny() );
		assertEquals( "/temp/mygrid/app/run.sh", entry.getSourceFile() );

		taskToTest = (TaskSpec) tasks.removeFirst();
		input = taskToTest.getInitBlock();
		entries = input.getEntry( "" );
		entry = (IOEntry) entries.get( 0 );
		assertEquals( "put", entry.getCommand() );
		assertEquals( "$PLAYPEN", entry.getDestiny() );
		assertEquals( CommonCompiler.getSourceParentDir() + File.separator + "entry/input2.txt", entry.getSourceFile() );
		entry = (IOEntry) entries.get( 1 );
		assertEquals( "store", entry.getCommand() );
		assertEquals( "$PLAYPEN/run_app.sh", entry.getDestiny() );
		assertEquals( "/temp/mygrid/app/run.sh", entry.getSourceFile() );

	}


	public void testNewOperators() throws Exception {

		compiler.compile( "test/specs/semantic/JobTest4.jdf", CommonCompiler.JDF_TYPE );
		List result = compiler.getResult();

		JobSpec jobInfo = (JobSpec) result.remove( 0 );
		assertEquals( "testing", jobInfo.getLabel() );
		assertEquals( "NOT ( ( OS = Linux AND PLAYPEN_SIZE >= 200 ) or ( 0s = Windows ) )", jobInfo.getRequirements() );

		LinkedList tasks = (LinkedList) jobInfo.getTaskSpecs();
		assertEquals( 5, tasks.size() );

		TaskSpec taskToTest = (TaskSpec) tasks.removeFirst();
		assertEquals( "echo \"bla ble bli blo blu\"", taskToTest.getRemoteExec() );

	}


	public void testErrorCallsAtPerformAction() {

		JDFSemanticActions actions = new JDFSemanticActions();
		try {
			actions.performAction( "nonexistent_action", new Token() );
			fail( "Sould fail because the action does not exists to be executed." );
		} catch ( SemanticException sex ) {
			assertTrue( sex.getCause() instanceof NoSuchMethodException );
		}

		try {
			actions.performAction( "action4", null );
			fail( "Sould fail because the action uses the token object passed as \"null\"." );
		} catch ( SemanticException sex ) {
			assertTrue( sex.getCause() instanceof NullPointerException );
		}

	}


	public void testBlankSpaces() throws CompilerException {

		compiler.compile( "test/specs/semantic/JobTest5.jdf", CommonCompiler.JDF_TYPE );
		List result = compiler.getResult();

		JobSpec jobInfo = (JobSpec) result.remove( 0 );
		assertEquals( "testing:label", jobInfo.getLabel() );
		assertEquals( "OS == Linux && PLAYPEN_SIZE >= 200", jobInfo.getRequirements() );

		LinkedList tasks = (LinkedList) jobInfo.getTaskSpecs();
		assertEquals( 1, tasks.size() );

		TaskSpec taskToTest = (TaskSpec) tasks.removeFirst();
		assertEquals( "$PLAYPEN/execute input", taskToTest.getRemoteExec() );

		IOBlock input = taskToTest.getInitBlock();
		Vector entries = input.getEntry( "( PLAYPEN_SIZE < 300 )" );
		IOEntry entry = (IOEntry) entries.get( 0 );

		assertEquals( "PUT", entry.getCommand() );
		assertEquals( CommonCompiler.getSourceParentDir() + File.separator + "sample/input 1", entry.getSourceFile() );
		assertEquals( "input1", entry.getDestiny() );

		entry = (IOEntry) entries.get( 1 );
		assertEquals( "STORE", entry.getCommand() );
		assertEquals( CommonCompiler.getSourceParentDir() + File.separator + "sample/input2", entry.getSourceFile() );
		assertEquals( "input 2", entry.getDestiny() );
	}

}
