/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.spec.syntactical;

import java.util.Stack;

import org.ourgrid.common.spec.grammar.Rule;
import org.ourgrid.common.spec.grammar.Symbol;

/**
 * This is the entity that will control the derivation of a word from a language
 * described by a grammar. Here a word is the source that has to be compiled,
 * then; a derivation is the process of, beginning from the initial symbol of a
 * grammar, obtain the read source.
 */
public class SyntacticalDerivationTree {

	private Stack<Symbol> sdt;


	/**
	 * Initializes the stack that will control the derivation.
	 * 
	 * @param endOfSourceSymbol The symbol that defines that the source has
	 *        ended.
	 * @param initialSymbol The initial symbol of the grammar language.
	 */
	public SyntacticalDerivationTree( Symbol endOfSourceSymbol, Symbol initialSymbol ) {

		sdt = new Stack<Symbol>();
		initializeSDT( endOfSourceSymbol, initialSymbol );
	}


	/*
	 * Pushes the two symbol received as paramethers at the constructor... It is
	 * necessary to have at the end of the stack the final symbol from source
	 * and at the top (in the begging moment) the initial symbol from grammar.
	 */
	private void initializeSDT( Symbol endOfSourceSymbol, Symbol initialSymbol ) {

		sdt.push( endOfSourceSymbol );
		sdt.push( initialSymbol );
	}


	/**
	 * Returns the symbol at the top of the derivation stack.
	 * 
	 * @return The symbol at the top of the derivation stack.
	 */
	public Symbol top() {

		return sdt.peek();
	}


	/**
	 * Pops (removes) the symbol at the top of the derivation stack.
	 * 
	 * @return The symbol that was removed from the top of the derivation stack.
	 */
	public Symbol pop() {

		return sdt.pop();
	}


	/**
	 * Pushes (insert) a given rule at the top of the derivation stack. A rule
	 * is a set of symbols (terminal or not). To push a rule at the derivation
	 * stack, is necessary to insert from the end to the beggining of the rule's
	 * symbols.
	 * 
	 * @param rule The rule that have to be inserted at the derivation stack
	 *        top.
	 */
	public void pushRule( Rule rule ) {

		for ( int count = (rule.getBody().length) - 1; count >= 0; count-- ) {
			Symbol symbol = ((rule.getBody())[count]);
			sdt.push( symbol );
		}
	}

}
