/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.url;

import org.ourgrid.common.fd.AbstractFailureDetector;

/**
 * This class provides URLs for the main components of User Agent that needed to
 * be remotely referenced.
 */
public class UserAgentURLProvider extends URLProvider {

	public static final String USERAGENT_UIMANAGER = "UseragentUIManager";

	public static final String USERAGENT = "USERAGENT";


	/**
	 * This method builds an user agent URL based on the name and port number
	 * used on the ua.properties configuration file.
	 * 
	 * @return A well-formed URL for an UserAgentServer object.
	 */
	public static String userAgent() {

		return "rmi://" + getName() + ":" + getPort() + "/" + USERAGENT;
	}


	/**
	 * This method builds a UserAgent URL based on a hostname and a port number.
	 * 
	 * @param hostname Hostname used to build this URL.
	 * @param port TCP port number used to build this URL.
	 * @return A well formed URL for a UserAgent server object.
	 */
	public static String userAgent( String hostname, int port ) {

		return "rmi://" + hostname + ":" + port + "/" + USERAGENT;
	}


	/**
	 * This method is used like a common point to get the address of the
	 * ShutdownManager.
	 * 
	 * @return the RMI URL that points to the ShutdownManager
	 */
	public static String shutdownManager() {

		return "rmi://" + getName() + ":" + getPort() + "/ShutdownManager";
	}


	/**
	 * This method is used like a common point to get the address of the
	 * ShutdownManager from a defined UserAgent.
	 * 
	 * @param name Host name of the User Agent.
	 * @param port Port of this UserAgent.
	 * @return the RMI URL that points to the ShutdownManager
	 */
	public static String shutdownManager( String name, String port ) {

		return "rmi://" + name + ":" + port + "/ShutdownManager";
	}


	/**
	 * This method is used like a common point to get the address of the all
	 * UserAgent published objects.
	 * 
	 * @return the RMI URL that points to all UserAgent published objects.
	 */
	public static String publishedObjects() {

		return "rmi://" + getName() + ":" + getPort();
	}


	/**
	 * This method is used like a common point to get the address of the
	 * UserAgent Failure Detector.
	 * 
	 * @return the RMI URL that points to the UserAgent Failure Detector.
	 */
	public static String userAgentFailureDetector() {

		return "rmi://" + getName() + ":" + getPort() + "/" + AbstractFailureDetector.BOUND_NAME;
	}


	/**
	 * This method is used like a common point to get the address of the all
	 * UserAgent published objects.
	 * 
	 * @return the RMI URL that points to all UserAgent published objects.
	 */
	public static String userAgentUIManager() {

		return "rmi://" + getName() + ":" + getPort() + "/" + USERAGENT_UIMANAGER;
	}

}
