/*
 * Copyright (c) 2002-2004 Universidade Federal de Campina Grande
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ourgrid.common.util;

/**
 * Class to provide a Bidimensional Array
 */
public class BidimensionalArray<E> {

	private int initialLength;

	private Object[ ][ ] arrays;

	private final static String LINE_SEPARATOR = System.getProperty( "line.separator" );


	public BidimensionalArray( int initialArrayLength ) {

		arrays = new Object[ initialArrayLength ][ initialArrayLength ];
		this.initialLength = initialArrayLength;
	}


	/**
	 * Inserts a new element at the determinated coordenate. If the coordenate
	 * already exists in this <code>BidimensionalArray</code>, the old
	 * element is replaced by the specified element.
	 * 
	 * @param x_index Represents the "x" position, as a cartesian plane.
	 * @param y_index Represents the "y" position, as a cartesian plane.
	 * @param element The element to be added.
	 */
	public boolean insert( int x_index, int y_index, E element ) {

		needResize( x_index, y_index );

		if ( arrays[x_index][y_index] != null )
			return false;

		arrays[x_index][y_index] = element;
		return true;

	}


	private void needResize( int x_index, int y_index ) {

		if ( x_index >= initialLength - 1 || y_index >= initialLength - 1 )
			resizeArray();
	}


	private void resizeArray() {

		int oldLength = initialLength;
		initialLength = 2 * initialLength;

		Object[ ][ ] tmp = new Object[ initialLength ][ initialLength ];
		for ( int x = 0; x < oldLength; x++ ) {
			for ( int y = 0; y < oldLength; y++ ) {
				tmp[x][y] = arrays[x][y];
			}
		}

		arrays = tmp;

	}


	public void insertAtRow( int row, E element ) {

		for ( int i = 0; i < initialLength; i++ ) {
			needResize( row, i );
			if ( arrays[row][i] == null ) {
				arrays[row][i] = element;
				break;
			}
		}

	}


	public void insertAtColumn( int column, E element ) {

		for ( int i = 0; i < initialLength; i++ ) {
			needResize( i, column );
			if ( arrays[i][column] == null ) {
				arrays[i][column] = element;
				break;
			}

		}

	}


	/**
	 * Gets the element at the determinated position. Returns null if the
	 * position is not assigned.
	 * 
	 * @param x_index Represents the "x" position, as a cartesian plane.
	 * @param y_index Represents the "y" position, as a cartesian plane.
	 */
	public E get( int x_index, int y_index ) {

		Object element = arrays[x_index][y_index];
		E generic = null;

		if ( element != null )
			generic = (E) element;
		return generic;

	}


	public E remove( int x_index, int y_index ) {

		E generic = get( x_index, y_index );
		arrays[x_index][y_index] = null;
		return generic;

	}


	public void replace( int x_index, int y_index, E element ) {

		arrays[x_index][y_index] = element;

	}


	public String toString() {

		String str = "| ";
		for ( int x = 0; x < initialLength; x++ ) {
			for ( int y = 0; y < initialLength; y++ ) {
				str += arrays[x][y] + " | ";
			}
			str += x != initialLength ? LINE_SEPARATOR : "";
		}
		str += " |";

		return str;
	}
}
