/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

/**
 * Class used to clone objects through stream writing and reading.
 */
public class StreamCloner {

	/**
	 * Clones the given object through stream writing and reading.
	 * 
	 * @param objectToBeCloned The object to be cloned
	 * @return The cloned object
	 * @throws CloneNotSupportedException If some object involved in the cloning
	 *         operation does not implement the
	 *         <code>java.io.Serializable</code> interface.
	 */
	public static <T extends Serializable> T clone( T objectToBeCloned ) throws CloneNotSupportedException {

		ByteArrayOutputStream objectBAOS = new ByteArrayOutputStream();
		ObjectOutputStream objectOOS = null;
		ObjectInputStream objectOIS = null;
		T clonedObject = null;
		try {
			objectOOS = new ObjectOutputStream( objectBAOS );
			objectOOS.writeObject( objectToBeCloned );

			objectOIS = new ObjectInputStream( new ByteArrayInputStream( objectBAOS.toByteArray() ) );
			clonedObject = (T) objectOIS.readObject();

		} catch ( Exception e ) {
			throw new CloneNotSupportedException(
					"The given object could not be cloned. Verify if all objects involved in the cloning operation implement the java.io.Serializable interface: "
							+ e.getMessage() );
		} finally {

			if ( objectOOS != null ) {
				try {
					objectOOS.close();
				} catch ( IOException e1 ) {

				}
			}

			if ( objectOIS != null ) {
				try {
					objectOIS.close();
				} catch ( IOException e1 ) {

				}
			}

			objectBAOS.reset();
			objectBAOS = null;

		}

		System.gc();

		return clonedObject;

	}
}
