/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.xml;

import java.io.IOException;

import org.ourgrid.common.exception.XMLParserException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Objects of this class knows how to parse a MyGrid Doctor test description XML
 * file.
 */
public class XMLParser {

	/**
	 * Object that represents the XML document.
	 */
	private Document doc = null;


	/**
	 * The constructor.
	 * 
	 * @param filePath XML configuration file path
	 * @throws IOException
	 * @throws XMLParserException
	 */
	public XMLParser( String filePath ) throws XMLParserException, IOException {

		doc = XMLUtil.parseXMLFile( filePath );
	}


	/**
	 * This method returns the root element of the XML configuration file.
	 * 
	 * @return The root element of the XML configuration file
	 */
	public Element getDocumentElement() {

		return doc.getDocumentElement();
	}


	/**
	 * This method returns a diagnostic by a given id.
	 * 
	 * @param id Diagnostic id
	 * @return The diagnostic that matches the given id
	 */
	public String getDiagnosticById( String id ) {

		Node node = getNodeById( "GMDDiagnostics", id );
		if ( node == null )
			return "";

		String diagnostic = node.getChildNodes().item( 0 ).getNodeValue().trim();
		return diagnostic;
	}


	/**
	 * This method returns a cause by a given id
	 * 
	 * @param id Cause id
	 * @return The cause that matches the given id
	 */
	public String getCauseById( String id ) {

		Node node = getNodeById( "GMDCauses", id );
		if ( node == null )
			return "";

		String cause = node.getChildNodes().item( 0 ).getNodeValue().trim();
		return cause;
	}


	/**
	 * This method returns a prescription by a given id.
	 * 
	 * @param id Prescription id
	 * @return A String containing the prescription that matches the specified
	 *         id
	 */
	public String getPrescriptionById( String id ) {

		Node node = getNodeById( "GMDPrescriptions", id );
		if ( node == null )
			return "";

		String prescription = node.getChildNodes().item( 0 ).getNodeValue().trim();
		return prescription;
	}


	/**
	 * This method returns a node by a given id.
	 * 
	 * @param node Node name
	 * @param id Node id
	 * @return The node that matches the given id
	 */
	private Node getNodeById( String node, String id ) {

		NodeList nodeList = doc.getDocumentElement().getElementsByTagName( node ).item( 0 ).getChildNodes();
		Node returnNode = null;
		for ( int x = 0; x < nodeList.getLength(); x++ )
			if ( nodeList.item( x ).getNodeType() == Node.ELEMENT_NODE ) {
				returnNode = nodeList.item( nodeList.getLength() - 2 );
				if ( nodeList.item( x ).getAttributes().getNamedItem( "id" ).getNodeValue().equals( id ) ) {
					returnNode = nodeList.item( x );
					break;
				}
			}

		return returnNode;
	}


	/**
	 * This method returns the Exception list of a given test.
	 * 
	 * @param testName Test name
	 * @return A NodeList with exceptions that can ocurrs in the test
	 */
	public NodeList getExceptionList( String testName ) {

		NodeList testsList = doc.getDocumentElement().getElementsByTagName( "GMDTests" ).item( 0 ).getChildNodes();

		// tests loop
		for ( int a = 0; a < testsList.getLength(); a++ ) {
			if ( (testsList.item( a ).getNodeType() == Node.ELEMENT_NODE)
					&& (testsList.item( a ).getAttributes().getNamedItem( "name" ).getNodeValue().equals( testName )) ) {
				return testsList.item( a ).getChildNodes();
			}
		}
		return null;
	}


	/**
	 * This method returns the dependencies of a test.
	 * 
	 * @param testName Test name
	 * @return An array of string containing the dependencies of the given test
	 */
	public String[ ] getDependencies( String testName ) {

		NodeList testsList = doc.getDocumentElement().getElementsByTagName( "GMDTests" ).item( 0 ).getChildNodes();

		// tests loop
		for ( int a = 0; a < testsList.getLength(); a++ ) {
			if ( (testsList.item( a ).getNodeType() == Node.ELEMENT_NODE)
					&& (testsList.item( a ).getAttributes().getNamedItem( "name" ).getNodeValue().equals( testName )) ) {
				return testsList.item( a ).getAttributes().getNamedItem( "depends" ).getNodeValue().split( "," );
			}
		}
		return null;
	}


	/**
	 * This method returns the detailed exception node according to exception.
	 * 
	 * @param exceptionList Node list of exceptions
	 * @param exception Exception catched
	 * @return The detailed exception node according to exception
	 */
	public Node getExceptionDetails( NodeList exceptionList, Throwable exception ) {

		Throwable cause = getCause( exceptionList, exception );

		if ( exceptionList != null ) {
			for ( int y = 0; y < exceptionList.getLength(); y++ ) {
				if ( exceptionList.item( y ).getNodeType() == Node.ELEMENT_NODE ) {
					String exceptionName = exceptionList.item( y ).getAttributes().getNamedItem( "name" )
							.getNodeValue().trim();
					if ( cause != null )
						if ( cause.getClass().getName().indexOf( exceptionName ) != -1 )
							return exceptionList.item( y );
				}
			}
		}
		return null;
	}


	/**
	 * This method returns the exception cause.
	 * 
	 * @param exceptionList Node list of exceptions
	 * @param exception Exception catched
	 * @return The cause of the given exception
	 */
	public Throwable getCause( NodeList exceptionList, Throwable exception ) {

		Throwable cause = null;
		Throwable nonMGException = getNonMyGridException( exception );
		exception = nonMGException == null ? exception : nonMGException;

		if ( exceptionList != null ) {
			for ( int y = 0; y < exceptionList.getLength(); y++ ) {
				if ( exceptionList.item( y ).getNodeType() == Node.ELEMENT_NODE ) {
					String exceptionName = exceptionList.item( y ).getAttributes().getNamedItem( "name" )
							.getNodeValue().trim();
					cause = exception;

					while ( cause.getClass().getName().indexOf( exceptionName ) == -1 ) {
						cause = cause.getCause();
						if ( cause == null )
							break;
					} // end while
					if ( cause != null )
						return cause;
				} // end if
			} // end for
		}
		return exception;
	}


	/**
	 * This method return the first non MyGrid Exception found in exception.
	 * 
	 * @param exception Exception to extract a non MyGrid Exception
	 * @return The first non MyGrid exception found
	 */
	private Throwable getNonMyGridException( Throwable exception ) {

		Throwable myThrowable = exception;
		while ( (myThrowable != null) && (myThrowable.getClass().getName().indexOf( "mygrid" ) >= 0) )
			myThrowable = myThrowable.getCause();
		return myThrowable;

	}


	/**
	 * This method returns the exception name of a given exception node.
	 * 
	 * @param node Exception node
	 * @return Exception name
	 */
	public String getExceptionName( Node node ) {

		if ( node != null )
			return node.getAttributes().getNamedItem( "name" ).getNodeValue();
		return "Unexpected Exception";
	}


	/**
	 * This method returns the diagnostic of a given diagnostic node.
	 * 
	 * @param node Diagnostic node
	 * @return The diagnostic obtained from the given node
	 */
	public String getDiagnostic( Node node ) {

		NodeList details = node.getChildNodes();
		for ( int x = 0; x < details.getLength(); x++ )
			if ( details.item( x ).getNodeName().equals( "diagnostic" ) )
				return getDiagnosticById( details.item( x ).getAttributes().getNamedItem( "id" ).getNodeValue() );

		return null;
	}


	/**
	 * This method returns a node list of causes of a given causes node.
	 * 
	 * @param node Causes node
	 * @return A NodeList containing the causes of the given node
	 */
	public NodeList getCauses( Node node ) {

		return getNodeListByName( node, "causes" );
	}


	/**
	 * This method returns a node list of prescriptions of a given prescriptions
	 * node.
	 * 
	 * @param node Prescriptions node
	 * @return A NodeList containing the prescriptions of the given node
	 */
	public NodeList getPrescriptions( Node node ) {

		return getNodeListByName( node, "prescriptions" );
	}


	/**
	 * This method returns a node list from a child node of the given node.
	 * 
	 * @param node Node to be searched
	 * @param name Child node name
	 * @return A NodeList of the given child node
	 */
	private NodeList getNodeListByName( Node node, String name ) {

		NodeList details = node.getChildNodes();
		for ( int x = 0; x < details.getLength(); x++ )
			if ( details.item( x ).getNodeName().equals( name ) )
				return details.item( x ).getChildNodes();

		return null;
	}
}
