/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.common.xml;

/**
 * Description: This class represents a utility to XML operations.
 * 
 * @version 1.0 Created on Sep 12, 2003 Last update: Sep 12, 2003
 */

import java.io.ByteArrayInputStream;
import java.io.IOException;

import org.ourgrid.common.exception.XMLParserException;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.sun.org.apache.xerces.internal.parsers.DOMParser;

public class XMLUtil {

	/** Validation feature id (http://xml.org/sax/features/validation). */
	protected static final String VALIDATION_FEATURE_ID = "http://xml.org/sax/features/validation";

	/**
	 * Schema validation feature id
	 * (http://apache.org/xml/features/validation/schema).
	 */
	protected static final String SCHEMA_VALIDATION_FEATURE_ID = "http://apache.org/xml/features/validation/schema";

	// The possible origins of the XML( string with the content or a file )
	private static final int XML_FILE_TYPE = 0;

	private static final int XML_CONTENT_TYPE = 1;


	/**
	 * Parse the xmlContent string, verifying if the content is well-formed and
	 * validating with the schema, if it exists.
	 * 
	 * @param xmlContent The xml content
	 * @return The document parsed
	 * @throws XMLParserException If there is a problem of validation or
	 *         well-formed.
	 */
	public static Document parseXMLContent( String xmlContent ) throws XMLParserException {

		// As xmlContent will not do a IO operation it will never throw an
		// IOException
		try {
			return parse( xmlContent, XML_CONTENT_TYPE );
		} catch ( IOException e ) {
		}
		return null;
	}


	/**
	 * Parse the fileName in xml, verifying if the content is well-formed and
	 * validating with the schema, if it exists.
	 * 
	 * @param fileName The xml file to be parsed
	 * @return The document parsed
	 * @throws XMLParserException If there is a problem of validation or
	 *         well-formed.
	 */
	public static Document parseXMLFile( String fileName ) throws XMLParserException, IOException {

		return parse( fileName, XML_FILE_TYPE );
	}


	// really parse the file or the string in xml format
	private static Document parse( String xml, int type ) throws IOException, XMLParserException {

		DOMParser parser = new DOMParser();
		try {
			parser.setFeature( VALIDATION_FEATURE_ID, true );
			parser.setFeature( SCHEMA_VALIDATION_FEATURE_ID, true );

			// Validating the xml instance fileName
			if ( type == XML_FILE_TYPE ) {
				parser.parse( xml );
			} else {
				parser.parse( new InputSource( new ByteArrayInputStream( xml.getBytes() ) ) );
			}

			return parser.getDocument();
		} catch ( SAXException e ) {
			throw new XMLParserException( "Error parsing the XML: [ " + xml + " ]", e );
		}
	}


	/**
	 * Gets the string value of the given attribute name( nameLowerCase or
	 * nameUpperCase )
	 * 
	 * @param nameLowerCase The name of the attribute in lowerCase
	 * @param nameUpperCase The name of the attribute in upperCase
	 * @param attributes The attributes where is the attribute finded
	 */
	public static String getAttributeValue( String nameLowerCase, String nameUpperCase, NamedNodeMap attributes ) {

		Node node = attributes.getNamedItem( nameLowerCase );
		if ( node == null ) {
			node = attributes.getNamedItem( nameUpperCase );
		}
		return node.getNodeValue();
	}

}
