/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.corepeer;

import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.CorePeerConfiguration;
import org.ourgrid.common.fd.AbstractFailureDetector;
import org.ourgrid.common.fd.Notifiable;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.url.CorePeerURLProvider;
import org.ourgrid.common.url.URLProvider;
import org.ourgrid.corepeer.fd.CorePeerFailureDetector;
import org.ourgrid.corepeer.fd.PeerStateNotifiable;

/**
 * Description: Implementation of CorePeer interface.
 * 
 * @version 1.0 Date24/09/2004
 */
public class CorePeerImpl extends OurgridUnicastRemoteObject implements CorePeer, CorePeerStatusProvider {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	public static final long serialVersionUID = 33L;

	/**
	 * A timer to execute a pinger task.
	 */
	private transient Timer t = new Timer();

	/**
	 * A map with all know peers on the community.
	 */
	private transient Map<Long,ObjectID> knowPeers;

	/**
	 * A map with all know peers on the community, associated by URL.
	 */
	private transient Map<String,ObjectID> knowPeersbyURL;

	/**
	 * A Map to story Peer recently arrived. These Peers will stay here until
	 * the Peer Failure Detector notify the new Peer is UP.
	 */
	private transient Map<String,ObjectID> temporaryArrivedPeers;

	/**
	 * Logger to store log messages
	 */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( CorePeerImpl.class );

	/**
	 * A thread pool.
	 */
	private transient ThreadPoolExecutor threadPool;

	/**
	 * Notifiable to receive notification from FD about Peer Status
	 */
	private Notifiable peerStateNotifiable;

	/** Representation of Peer Failure Detector */
	private static CorePeerFailureDetector failureDetector;

	/** Pinger Factory * */
	private transient PingerFactory pingerFactory;


	/**
	 * Default constructor.
	 * 
	 * @throws RemoteException If could not create the object.
	 */
	public CorePeerImpl() throws RemoteException {

		super();

		pingerFactory = new PingerFactory();
		knowPeers = new HashMap<Long,ObjectID>();
		knowPeersbyURL = new HashMap<String,ObjectID>();
		temporaryArrivedPeers = new HashMap<String,ObjectID>();

		threadPool = new ThreadPoolExecutor( 1, Integer.MAX_VALUE, 60, TimeUnit.SECONDS,
				new SynchronousQueue<Runnable>() );

		peerStateNotifiable = new PeerStateNotifiable( this );

		long updateDelay = Long.parseLong( Configuration.getInstance().getProperty(
				CorePeerConfiguration.PROP_UPDATE_DELAY ) );
		t.schedule( new CorePeerTimerTask(), updateDelay, updateDelay );

		try {
			failureDetector = new CorePeerFailureDetector( Configuration.QUERY_FREQ );
			startFailureDetector();
		} catch ( RemoteException e ) {
			LOG.error( "Error trying to create Failure Detector", e );
		}

	}


	/**
	 * Starts Failure Detector
	 */
	private void startFailureDetector() {

		try {

			failureDetector.startMonitoring();
			Naming.rebind( CorePeerURLProvider.corePeerFailureDetector(), failureDetector );

		} catch ( RemoteException e ) {
			LOG.error( "Cannot bind Peer Failure Detector Object.", e );
		} catch ( MalformedURLException e ) {
			LOG.error( "Cannot bind Peer Failure Detector Object.", e );
		}
	}


	/**
	 * @see org.ourgrid.corepeer.CorePeer#hereIAm(ObjectID)
	 */
	public List<ObjectID> hereIAm( ObjectID peerID ) throws RemoteException {

		try {

			String peerURL = URLProvider.getScheme( peerID.getObjectURL() ) + "://"
					+ URLProvider.getHost( peerID.getObjectURL() ) + ":" + URLProvider.getPort( peerID.getObjectURL() );

			String peerFdURL = peerURL + "/" + AbstractFailureDetector.BOUND_NAME;

			// adding new peer to temporary cache
			temporaryArrivedPeers.put( peerURL, peerID );

			failureDetector.register( peerStateNotifiable, peerFdURL, true );

		} catch ( Exception e ) {
			e.printStackTrace();
			LOG.error( "Cannot get CorePeerFailureDetector RMI Object", e );
		}

		return new Vector<ObjectID>( knowPeers.values() );

	}


	public void newPeerArrived( ObjectID peerID ) {

		synchronized ( knowPeers ) {

			try {
				String peerURL = URLProvider.getScheme( peerID.getObjectURL() ) + "://"
						+ URLProvider.getHost( peerID.getObjectURL() ) + ":"
						+ URLProvider.getPort( peerID.getObjectURL() );

				ObjectID objID = temporaryArrivedPeers.remove( peerURL );
				verifyAndAdd( objID );

			} catch ( Exception e ) {
				LOG.error( "Error trying to remove " + peerID + " from temporary arrived peers.", e );
			}

			t.schedule( new CorePeerTimerTask(), 5000 );

		}

	}


	/**
	 * @see org.ourgrid.corepeer.CorePeer#removePeer(org.ourgrid.common.id.ObjectID)
	 */
	public void removePeer( ObjectID peerID ) {

		synchronized ( knowPeers ) {

			Collection<ObjectID> peers = knowPeers.values();

			for ( Iterator<ObjectID> iter = peers.iterator(); iter.hasNext(); ) {

				ObjectID objID = iter.next();

				try {
					if ( comparePeers( objID, peerID ) ) {
						iter.remove();
					}
				} catch ( URISyntaxException e ) {
					LOG.error( "Cannot compare URL's.", e );
				}

			}

			t.schedule( new CorePeerTimerTask(), 5000 );

		}

	}


	private boolean comparePeers( ObjectID obj1, ObjectID obj2 ) throws URISyntaxException {

		String objURL1, objURL2;

		objURL1 = URLProvider.getScheme( obj1.getObjectURL() ) + "://" + URLProvider.getHost( obj1.getObjectURL() )
				+ ":" + URLProvider.getPort( obj1.getObjectURL() );
		objURL2 = URLProvider.getScheme( obj2.getObjectURL() ) + "://" + URLProvider.getHost( obj2.getObjectURL() )
				+ ":" + URLProvider.getPort( obj2.getObjectURL() );
		return objURL1.equals( objURL2 );

	}


	/**
	 * Verifies if a peer has the same URL of a previous known peerid, but a
	 * different peerID.
	 * 
	 * @param peerID
	 */
	private void verifyAndAdd( ObjectID peerID ) {

		if ( peerID == null ) {
			return;
		}

		ObjectID p = knowPeersbyURL.get( peerID.getObjectURL() );

		if ( p != null ) {

			if ( !p.getIdValue().equals( peerID.getIdValue() ) ) {

				LOG.info( "Peer " + peerID.getIdValue() + " at " + peerID.getObjectURL()
						+ " was previously identified as " + p.getIdValue() );

				knowPeers.remove( p.getIdValue() );
				knowPeersbyURL.remove( p.getObjectURL() );

			}
		}

		LOG.info( "Peer " + peerID.getName() + " at " + peerID.getObjectURL() + " has joined the OurGrid community." );
		knowPeersbyURL.put( peerID.getObjectURL(), peerID );
		knowPeers.put( peerID.getIdValue(), peerID );

	}


	public void shutdown() {

		failureDetector.shutdown();
		t.cancel();
	}


	public void setFailureDetector( CorePeerFailureDetector fd ) {

		CorePeerImpl.failureDetector = fd;
	}

	private class CorePeerTimerTask extends TimerTask {

		/**
		 * Sends the peer list to all online peers. If a peer is offline, it's
		 * is removed from the know peers map.
		 */
		public void run() {

			Vector<ObjectID> peerList = null;

			synchronized ( knowPeers ) {
				peerList = new Vector<ObjectID>( knowPeers.values() );
			}

			Iterator<ObjectID> peersIt = peerList.iterator();
			while ( peersIt.hasNext() ) {

				ObjectID peerID = peersIt.next();

				threadPool.execute( pingerFactory.makePinger( peerID, knowPeers, peerList, knowPeersbyURL ) );

			}

		}

	}


	public List<ObjectID> getKnownPeers( ObjectID peerId ) throws RemoteException {

		synchronized ( knowPeers ) {
			return new Vector<ObjectID>( knowPeers.values() );
		}
	}


	public Collection<ObjectID> getPeerList() {

		synchronized ( knowPeers ) {
			return new Vector<ObjectID>( knowPeers.values() );
		}
	}


	public void setPingerFactory( PingerFactory pingerFactory ) {

		this.pingerFactory = pingerFactory;
	}

}
