/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.corepeer.ui;

import java.io.File;
import java.net.MalformedURLException;
import java.rmi.AccessException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.Collection;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.CorePeerConfiguration;
import org.ourgrid.common.config.LogConfiguration;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.url.CorePeerURLProvider;
import org.ourgrid.corepeer.CorePeer;
import org.ourgrid.corepeer.CorePeerImpl;
import org.ourgrid.corepeer.CorePeerStatusProvider;
import org.ourgrid.corepeer.RestrictedCorePeerImpl;
import org.ourgrid.corepeer.ShutdownManager;
import org.ourgrid.corepeer.ShutdownManagerImpl;

public class CorePeerUIManager extends OurgridUnicastRemoteObject implements UIManager {

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( CorePeerUIManager.class );

	private static final long serialVersionUID = 33L;

	private static UIManager uniqueInstance;

	private static Registry registry = null;


	private CorePeerUIManager() throws RemoteException {

		super();

	}


	public static UIManager getInstance() throws RemoteException {

		if ( uniqueInstance == null ) {

			Configuration.getInstance( Configuration.COREPEER );

			String logfile = Configuration.getInstance().getLogPath();
			String xmlLogfile = Configuration.getInstance().getXMLLogPath();

			String parents = new File( logfile ).getParent();

			if ( parents != null ) {
				(new File( parents )).mkdirs();
			}

			// Configuring Log4j
			String logPropertiesPath = Configuration.getInstance().getLogPropertiesPath();
			LogConfiguration.configure( logPropertiesPath, logfile, xmlLogfile );

			int port = Integer.parseInt( Configuration.getInstance( Configuration.COREPEER ).getProperty(
					CorePeerConfiguration.PROP_PORT ) );

			try {
				// Starting registry
				registry = LocateRegistry.createRegistry( port );
			} catch ( RemoteException e ) {
				try {
					registry = LocateRegistry.getRegistry( port );
				} catch ( RemoteException e1 ) {
					LOG.error( "Could not get registry.", e1 );
					System.exit( ErrorCode.COULD_NOT_GET_REGISTRY );
				}
			}

			try {
				// tries to get a remote instance already bound by another
				// invocation of this method.
				uniqueInstance = (UIManager) Naming.lookup( CorePeerURLProvider.CorePeerUIManager() );
			} catch ( MalformedURLException e ) {
				LOG.debug( "Error getting reference to corepeer UIManager (will create a new one). Invalid URL: "
						+ CorePeerURLProvider.CorePeerUIManager() );
			} catch ( NotBoundException e ) {
				LOG.debug( "Error getting reference to corepeer UIManager (will create a new one). Name not bound: "
						+ CorePeerURLProvider.CorePeerUIManager() );
			} catch ( RemoteException e ) {
				LOG
						.debug( "Error getting reference to corepeer UIManager (will create a new one). Cannot contact URL: "
								+ CorePeerURLProvider.CorePeerUIManager() );
			}

			if ( uniqueInstance == null ) {
				uniqueInstance = new CorePeerUIManager();

				try {
					Naming.rebind( CorePeerURLProvider.CorePeerUIManager(), uniqueInstance );
				} catch ( MalformedURLException e ) {
					LOG.debug( "Error trying to rebind at URL " + CorePeerURLProvider.CorePeerUIManager() );
				}
			}
		}
		return uniqueInstance;
	}


	public void startCorePeerService() throws AccessException, RemoteException {

		Configuration.getInstance( Configuration.COREPEER );

		File f = new File( RestrictedCorePeerImpl.ALLOWED_PEERS_FILE );

		if ( f.exists() )
			registry.rebind( CorePeerURLProvider.COREPEER_OBJECT_NAME, new RestrictedCorePeerImpl() );
		else
			registry.rebind( CorePeerURLProvider.COREPEER_OBJECT_NAME, new CorePeerImpl() );

		registry.rebind( CorePeerURLProvider.SHUTDOWN_MANAGER, new ShutdownManagerImpl() );

		LOG.info( "OurGrid Core Peer was successfully started." );

	}


	public void stopCorePeerService() throws CorePeerUIException, AccessException, RemoteException, NotBoundException {

		if ( isCorePeerRunning() ) {

			ShutdownManager shutdownManager = null;
			CorePeer corepeer = (CorePeer) registry.lookup( CorePeerURLProvider.COREPEER_OBJECT_NAME );
			shutdownManager = (ShutdownManager) registry.lookup( CorePeerURLProvider.SHUTDOWN_MANAGER );
			corepeer.shutdown();
			registry.unbind( CorePeerURLProvider.COREPEER_OBJECT_NAME );

			LOG.info( "OurGrid CorePeer was successfully stopped." );
			shutdownManager.shutdown();

		} else {
			throw new CorePeerUIException( "CorePeer is not running!" );
		}
	}


	public boolean isCorePeerRunning() throws RemoteException {

		try {
			findExportedObjects();
			return true;
		} catch ( Exception e ) {
			return false;
		}
	}


	private void findExportedObjects() throws MalformedURLException, RemoteException, NotBoundException {

		Naming.lookup( CorePeerURLProvider.peerExternalURL() );
	}


	public Collection<ObjectID> statusCorePeerService() throws AccessException, RemoteException, NotBoundException {

		CorePeerStatusProvider corePeer = (CorePeerStatusProvider) registry
				.lookup( CorePeerURLProvider.COREPEER_OBJECT_NAME );
		return corePeer.getPeerList();
	}

}
