/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.deployer;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.LogConfiguration;
import org.ourgrid.common.config.UserAgentConfiguration;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.main.CommonCompiler;
import org.ourgrid.common.spec.main.CompilerException;

public class Deployer {

	/** directory where UA.properties files will be stored */
	public static final String UASCONF_DIR = "uasconf";

	private List<GumSpec> gumsList;


	/**
	 * Creates a Deployer object that uses the gums described in the SDF file
	 * 
	 * @param siteDescriptionFile The SDF path
	 * @throws CompilerException If any error occurs during SDF compilation
	 */
	public Deployer( String siteDescriptionFile ) throws CompilerException {

		/** get GUMPSpec from SDF */
		CommonCompiler compiler = new CommonCompiler();

		/** tries to compile GDF */
		compiler.compile( siteDescriptionFile, CommonCompiler.SDF_TYPE );

		gumsList = compiler.getResult();
	}


	/**
	 * Main.
	 * 
	 * @param args First argument is the command to execute and the second is
	 *        GDF path
	 */
	public static void main( String[ ] args ) {

		/** **** Configuring Log4j ******* */
		/**
		 * setting to write 'uaadmin log' in a diferent file of 'peer log'. For
		 * all others log properties configuration, uaadmin log is equivalent to
		 * peer
		 */

		/** log file path for uaadmin log */
		final String UAADMIN_LOG_FILE_PATH = Configuration.getInstance( Configuration.PEER ).getRootDir()
				+ "/log/uaadmin.log";

		/** xml log file path for uaadmin xml log */
		final String UAADMIN_XML_LOG_FILE_PATH = UAADMIN_LOG_FILE_PATH + ".xml";

		String logfile = UAADMIN_LOG_FILE_PATH;
		String xmlLogFile = UAADMIN_XML_LOG_FILE_PATH;
		String logPropertiesfile = Configuration.getInstance().getLogPropertiesPath();

		String parents = new File( logfile ).getParent();

		if ( parents != null )
			(new File( parents )).mkdirs();

		LogConfiguration.configure( logPropertiesfile, logfile, xmlLogFile );

		Configuration.reset();

		/** ****** */

		Configuration.getInstance( Configuration.MYGRID );

		/** holds the command to execute */
		String command = args[0];

		/** holds the gdf path */
		String sdf = args[1];

		checkEnv( sdf );

		Deployer deployer;
		try {
			/** creates a new deployer */
			deployer = new Deployer( sdf );

			/** run deployer commmand */
			if ( command.equalsIgnoreCase( "start" ) || command.equalsIgnoreCase( "restart" ) ) {
				deployer.config();
			}
			deployer.startThreads( command );
		} catch ( CompilerException e ) {
			String message = "Error compiling the " + sdf + " file.";
			System.out.println( message );
			System.exit( ErrorCode.ERROR_COMPILING_THE_SDF_FILE );
		}
	}


	/**
	 * Checks if the environment is ok. If the gdf file not exists the program
	 * will exit.
	 * 
	 * @param file A file that must exist.
	 */
	private static void checkEnv( String file ) {

		int exitCode = ErrorCode.OK;

		if ( !fileExists( file ) ) {
			System.out.println( "The file \'" + file + "\' does not exist." );
			exitCode = ErrorCode.THE_FILE_NOT_EXIST;
		}

		if ( exitCode != ErrorCode.OK ) {
			System.exit( exitCode );
		}

	}


	/**
	 * Check if the gdf file exists.
	 * 
	 * @param fileName The file name.
	 * @return If the file exists or not.
	 */
	private static boolean fileExists( String fileName ) {

		if ( fileName == null ) {
			return false;
		}
		File file = new File( fileName );
		return file.exists();
	}


	/**
	 * Generates configuration files for user agents in localGuMP;
	 */
	private void config() {

		/** iterating on gum in localGuMP */
		for ( GumSpec currentGumSpec : gumsList ) {
			if ( gumIsUserAgent( currentGumSpec ) ) {
				generateUAProperties( currentGumSpec );
			}
		}
	}


	/**
	 * Verifies whether the specified GuM is a User Agent public static boolean
	 * gumIsUserAgent(GuMSpec currentGuMSpec) {
	 * 
	 * @param currentGuMSpec The remote Gum specification
	 * @return <b>true </b> if the GuM is a User Agent, <b>false </b> otherwise.
	 */
	public static boolean gumIsUserAgent( GumSpec currentGuMSpec ) {

		return currentGuMSpec.getAttribute( GumSpec.ATT_TYPE ).equalsIgnoreCase( GumSpec.TYPE_UA_LINUX )
				|| currentGuMSpec.getAttribute( GumSpec.ATT_TYPE ).equalsIgnoreCase( GumSpec.TYPE_UA_WINDOWS )
				|| currentGuMSpec.getAttribute( GumSpec.ATT_TYPE ).equalsIgnoreCase( GumSpec.TYPE_UA_SWAN );
	}


	/**
	 * Generates UA.properties file to a specific GuM
	 * 
	 * @param currentGuMSpec The GuM description
	 */
	private void generateUAProperties( GumSpec currentGuMSpec ) {

		/** create "uaproperties" directory */
		File fileUAPropertiesDir = new File( UASCONF_DIR );

		if ( !fileUAPropertiesDir.exists() ) {
			fileUAPropertiesDir.mkdir();
		}

		/** create a UA.properties file with GuM name */
		File fileUAPropertiesFile = defineUAPropertiesFile( currentGuMSpec );

		/** write UA.properties contents */
		writePropertiesToUAPropertiesFile( currentGuMSpec, fileUAPropertiesFile );
	}


	/**
	 * Reads user agent properties' defaults and, then, replace them with
	 * properties in the GuM description object (GuMSpec)
	 * 
	 * @param currentGuMSpec The GuM description
	 * @return A Map object that contains up to date user agent properties
	 *         values
	 */
	private Map readAndReplaceDefaultProperties( GumSpec currentGuMSpec ) {

		/** loading default values * */
		Map<String,String> uaPropertiesContents = UserAgentConfiguration.getDefaults();

		/** replacing default values with user-defined values * */

		Configuration.getInstance();

		// this if-clause wouldn't be necessary here, since we have decided that
		// UA_ATT_PORT is mandatory for User Agents (03/08/2004)
		// defensive programming is applied here
		if ( currentGuMSpec.hasAttribute( GumSpec.ATT_PORT ) ) {
			uaPropertiesContents.put( UserAgentConfiguration.PREFIX + Configuration.PROP_PORT, currentGuMSpec
					.getAttribute( GumSpec.ATT_PORT ) );
		}
		if ( currentGuMSpec.hasAttribute( GumSpec.ATT_NAME ) ) {
			uaPropertiesContents.put( UserAgentConfiguration.PREFIX + Configuration.PROP_NAME, currentGuMSpec
					.getAttribute( GumSpec.ATT_NAME ) );
		}
		if ( currentGuMSpec.hasAttribute( GumSpec.ATT_PLAYPEN_ROOT ) ) {
			uaPropertiesContents.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.PROP_PLAYPEN_ROOT,
					currentGuMSpec.getAttribute( GumSpec.ATT_PLAYPEN_ROOT ) );
		}
		if ( currentGuMSpec.hasAttribute( GumSpec.ATT_STORAGE_DIR ) ) {
			uaPropertiesContents.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.PROP_STORAGE_DIR,
					currentGuMSpec.getAttribute( GumSpec.ATT_STORAGE_DIR ) );
		}
		if ( currentGuMSpec.hasAttribute( GumSpec.ATT_PLAYPEN_SIZE ) ) {
			uaPropertiesContents.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.PROP_PLAYPEN_SIZE,
					currentGuMSpec.getAttribute( GumSpec.ATT_PLAYPEN_SIZE ) );
		}
		if ( currentGuMSpec.hasAttribute( GumSpec.ATT_STORAGE_SIZE ) ) {
			uaPropertiesContents.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.PROP_STORAGE_SIZE,
					currentGuMSpec.getAttribute( GumSpec.ATT_STORAGE_SIZE ) );
		}
		if ( currentGuMSpec.hasAttribute( GumSpec.ATT_SECURITY ) ) {
			uaPropertiesContents.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.PROP_SECURITY,
					currentGuMSpec.getAttribute( GumSpec.ATT_SECURITY ) );
		}

		if ( currentGuMSpec.hasAttribute( UserAgentConfiguration.PROP_IDLENESS_DETECTOR ) ) {
			uaPropertiesContents.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.PROP_IDLENESS_DETECTOR,
					currentGuMSpec.getAttribute( UserAgentConfiguration.PROP_IDLENESS_DETECTOR ) );
		}
		if ( currentGuMSpec.hasAttribute( UserAgentConfiguration.PROP_IDLENESS_TIME ) ) {
			uaPropertiesContents.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.PROP_IDLENESS_TIME,
					currentGuMSpec.getAttribute( UserAgentConfiguration.PROP_IDLENESS_TIME ) );
		}

		return uaPropertiesContents;
	}


	/**
	 * Writes user agent properties to UA.properties file. User agent properties
	 * come from a GuMSpec object
	 * 
	 * @param currentGuMSpec The GuM description
	 * @param uaPropertiesFile A reference object to the UA.properties file
	 */
	private void writePropertiesToUAPropertiesFile( GumSpec currentGuMSpec, File uaPropertiesFile ) {

		/** BufferedWriter object to write on UA.properties file */
		BufferedWriter out = null;
		try {
			out = new BufferedWriter( new FileWriter( uaPropertiesFile ) );

			/** a map that contains UA.properties file contents */
			Map uaPropertiesContents = readAndReplaceDefaultProperties( currentGuMSpec );

			/** writing UA.properties contents to the file * */
			Set entrySet = uaPropertiesContents.entrySet();
			Iterator itEntrySet = entrySet.iterator();
			while ( itEntrySet.hasNext() ) {
				Map.Entry mapEntry = (Map.Entry) itEntrySet.next();
				out.write( mapEntry.getKey() + " = " + mapEntry.getValue() );
				out.write( System.getProperty( "line.separator" ) );

			}
		} catch ( IOException e ) {
			String message = "Error when trying to write ua.properties file for gum "
					+ currentGuMSpec.getAttribute( GumSpec.ATT_NAME ) + ": " + e.getMessage();
			System.out.println( message );
		} finally {
			try {
				if ( out != null ) {
					out.close();
				}
			} catch ( IOException e ) {
				String message = "Error when trying to write ua.properties file for gum "
						+ currentGuMSpec.getAttribute( GumSpec.ATT_NAME ) + ": " + e.getMessage();
				System.out.println( message );
			}
		}
	}


	/**
	 * Defines the name of the UA.properties file, composing it with the GuM
	 * name and asserting it autoinstallable or not. It is autoinstallable if it
	 * is possible to use this Deployer to install and start it.
	 * 
	 * @param currentGuMSpec The GuM descritpion
	 * @return A reference to the UA.properties file
	 */
	private File defineUAPropertiesFile( GumSpec currentGuMSpec ) {

		/** get GuM name from GuMSpec */
		String gumName = currentGuMSpec.getAttribute( GumSpec.ATT_NAME ) + "_"
				+ currentGuMSpec.getAttribute( GumSpec.ATT_PORT );

		if ( uaIsAutoInstallable( currentGuMSpec ) ) {
			return new File( UASCONF_DIR + "/ua.properties-" + gumName + "-auto" );
		}
		return new File( UASCONF_DIR + "/ua.properties-" + gumName );

	}


	/**
	 * Verifies whether it is possible to install User Agent through deployer.
	 * 
	 * @param currentGuMSpec The remote Gum specification
	 * @return <b>true </b> if it is possible to automatically install, <b>false
	 *         </b> otherwise.
	 */
	public static boolean uaIsAutoInstallable( GumSpec currentGuMSpec ) {

		/**
		 * some conditions must be satisfied for the user agent to be
		 * autoinstallable
		 */
		return currentGuMSpec.getAttribute( GumSpec.ATT_TYPE ).equalsIgnoreCase( GumSpec.TYPE_UA_LINUX );
	}


	/**
	 * Verifies whether the gum spec has the attributes that are mandatory to
	 * use uaadmin.
	 * 
	 * @param currentGuMSpec The remote Gum specification
	 * @return <b>true </b> if the gum has the mandatory attributes, <b>false
	 *         </b> otherwise.
	 */
	public static boolean uaHasMandatoryAttributes( GumSpec currentGuMSpec ) {

		/**
		 * some conditions must be satisfied for the user agent to be
		 * autoinstallable
		 */
		return currentGuMSpec.hasAttribute( GumSpec.ATT_REM_EXEC )
				&& currentGuMSpec.hasAttribute( GumSpec.ATT_COPY_FROM )
				&& currentGuMSpec.hasAttribute( GumSpec.ATT_COPY_TO );
	}


	/**
	 * Starts all Threads to execute the deployer command.
	 * 
	 * @param command The deployer command.
	 */
	private void startThreads( String command ) {

		System.out.println( "-- Initializing " + command + " execution --" );
		for ( GumSpec currentGuMSpec : gumsList ) {
			String gumName = currentGuMSpec.getAttribute( GumSpec.ATT_NAME );
			if ( uaIsAutoInstallable( currentGuMSpec ) ) {
				if ( uaHasMandatoryAttributes( currentGuMSpec ) ) {
					try {
						Thread t = new Thread( new DeployerRunnable( command, currentGuMSpec ), "DeployerRunnable "
								+ currentGuMSpec.getAttribute( GumSpec.ATT_NAME ) );
						t.start();

					} catch ( Exception e ) {
						System.out.println( "--> " + gumName + " [ Problems when trying to " + command + " ("
								+ e.getMessage() + ") ] " );
					}
				} else {
					System.out.println( "--> " + gumName + " [ Problems (missing mandatory attributes" + "'"
							+ GumSpec.ATT_REM_EXEC.toString() + "' '" + GumSpec.ATT_COPY_TO.toString() + "' '"
							+ GumSpec.ATT_COPY_FROM.toString() + " ) ] " );
				}
			} else {
				System.out.println( "--> " + gumName + " [ Unable to " + command + " (not ualinux) ] " );
			}
		}
	}
}
