/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.deployer;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.executor.Executor;
import org.ourgrid.common.executor.ExecutorException;
import org.ourgrid.common.executor.ExecutorFactory;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.ExecutorResult;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.exception.GumSpecificationException;
import org.ourgrid.common.url.UserAgentURLProvider;
import org.ourgrid.common.util.SimpleZip;
import org.ourgrid.common.util.StringUtil;
import org.ourgrid.gridmachine.services.GumSpecServices;
import org.ourgrid.gridmachine.useragent.ShutdownManager;
import org.ourgrid.gridmachine.useragent.UserAgentClient;

/**
 * The deployer Thread. It is responsible for initializing, installing,
 * testing... User Agents.
 */
public class DeployerRunnable implements Runnable {

	public static final String flagMachine = "$machine";

	public static final String flagLocalFile = "$localfile";

	public static final String flagRemoteFile = "$remotefile";

	public static final String flagCommand = "$command";

	/** mygrid home directory */
	private String mgroot;

	/** directory where UA.properties files will be stored */
	private String uaPropertiesDir;

	/** The deployer command */
	private String command;

	/** The GuMSpec */
	private GumSpec currentGuMSpec;

	/** The tar package extension */
	private static final String ZIP_EXTENSION = ".zip";

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( DeployerRunnable.class );

	// grid machine rem_exec field
	private String gumRemExec;

	// grid machine copy_to field
	private String gumCopyTo;

	// grid machine name field
	private String dirName, gumName;

	Executor executor;

	ExecutorHandle eh;

	ExecutorResult execResult;


	/**
	 * Constructor.
	 * 
	 * @param uaadminCommand The command the Thread will execute.
	 * @param currentGuMSpec The GuMSpec.
	 */
	public DeployerRunnable( String uaadminCommand, GumSpec currentGuMSpec ) {

		this.command = uaadminCommand;
		this.currentGuMSpec = currentGuMSpec;

		gumRemExec = currentGuMSpec.getAttribute( GumSpec.ATT_REM_EXEC );
		gumCopyTo = currentGuMSpec.getAttribute( GumSpec.ATT_COPY_TO );
		gumName = currentGuMSpec.getAttribute( GumSpec.ATT_NAME );
		dirName = currentGuMSpec.getAttribute( GumSpec.ATT_NAME ) + "_"
				+ currentGuMSpec.getAttribute( GumSpec.ATT_PORT );

		/** mygrid home directory */
		mgroot = Configuration.getInstance().getProperty( Configuration.MGROOT );

		/** directory where UA.properties files will be stored */
		uaPropertiesDir = Deployer.UASCONF_DIR;

		executor = ExecutorFactory.getExecutorInstance();
	}


	private String[ ] getFilesToZip( String uaPropertiesFilePath ) {

		if ( new File( mgroot + "/lib/aspectjrt.jar" ).exists() ) {
			return new String[ ] { mgroot + "/bin/useragent", mgroot + "/lib/log4j.jar", mgroot + "/ua.log.properties",
									mgroot + "/lib/aspectjrt.jar", mgroot + "/LICENSE",
									uaPropertiesFilePath + File.separator + "ua.properties" };
		}
		return new String[ ] { mgroot + "/bin/useragent", mgroot + "/lib/log4j.jar", mgroot + "/ua.log.properties",
								uaPropertiesFilePath + File.separator + "ua.properties" };

	}


	/**
	 * The Thread behavior - it will execute the deployer command.
	 */
	public void run() {

		try {
			if ( command.equalsIgnoreCase( "start" ) ) {
				start( currentGuMSpec );
			} else if ( command.equalsIgnoreCase( "status" ) ) {
				status( currentGuMSpec );
			} else if ( command.equalsIgnoreCase( "stop" ) ) {
				stop( currentGuMSpec );
			} else if ( command.equalsIgnoreCase( "restart" ) ) {
				restart( currentGuMSpec );
			} else if ( command.equalsIgnoreCase( "clean" ) ) {
				clean();
			}
			delUasconf();

		} catch ( DeployerException de ) {
			System.err.println( de.getMessage() );
		}
	}


	/**
	 * @throws DeployerException
	 * @throws
	 */
	private void start( GumSpec currentGuMSpec ) throws DeployerException {

		GumClient uaGuM;
		String startCommand = "useragent start -";
		String message;

		try {

			try {
				uaGuM = new UserAgentClient( GumSpecServices.createGum( currentGuMSpec ), currentGuMSpec );
				if ( uaGuM.getStatus().isUp() )
					printGumStatus( currentGuMSpec.getAttribute( GumSpec.ATT_NAME ), currentGuMSpec
							.getAttribute( GumSpec.ATT_PORT ), UAAdminStatus.ALREADY_STARTED );

			} catch ( RemoteException e ) {

				install();

				// command to be executed on the remote gum for starting the
				// useragent
				String execCommand = StringUtil.replace( gumRemExec, flagMachine, gumName );
				String remoteCommand = "'sh " + DeployerRunnable.getUADirName( dirName ) + File.separator
						+ startCommand + "'";
				execCommand = StringUtil.replace( execCommand, flagCommand, remoteCommand );

				// using Executor class for remote execution
				try {
					eh = executor.execute( ".", execCommand );
					execResult = ExecutorFactory.getExecutorInstance().getResult( eh );

					if ( execResult.getExitValue() == 0 ) {
						printGumStatus( currentGuMSpec.getAttribute( GumSpec.ATT_NAME ), currentGuMSpec
								.getAttribute( GumSpec.ATT_PORT ), UAAdminStatus.STARTED );

					} else {
						message = "Could not start useragent at " + currentGuMSpec.getAttribute( GumSpec.ATT_NAME );
						System.out.println( message );
						LOG.info( message );
					}
				} catch ( ExecutorException e1 ) {
					message = "There was a problem during starting on "
							+ currentGuMSpec.getAttribute( GumSpec.ATT_NAME ) + ". User Agent was not started there.";
					LOG.error( message );
				}

			}
		} catch ( GumSpecificationException e ) {
			throw new DeployerException( mgroot, e );

		} catch ( Exception e ) {
			throw new DeployerException( mgroot, e );
		}
	}


	/**
	 * Tries to stop user agents in localGuMP
	 * 
	 * @throws DeployerException
	 */
	private void stop( GumSpec currentGuMSpec ) throws DeployerException {

		try {

			String url = UserAgentURLProvider.shutdownManager( currentGuMSpec.getAttribute( GumSpec.ATT_NAME ),
					currentGuMSpec.getAttribute( GumSpec.ATT_PORT ) );

			ShutdownManager shutdownManager = (ShutdownManager) Naming.lookup( url );

			shutdownManager.shutdown();

			LOG.error( "Cannot stop " + currentGuMSpec.getAttribute( GumSpec.ATT_NAME ) );
			throw new DeployerException( mgroot, new Exception( "Cannot stop "
					+ currentGuMSpec.getAttribute( GumSpec.ATT_NAME ) ) );

		} catch ( UnmarshalException e ) {
			// This exception is expected

			printGumStatus( currentGuMSpec.getAttribute( GumSpec.ATT_NAME ), currentGuMSpec
					.getAttribute( GumSpec.ATT_PORT ), UAAdminStatus.STOPPED );

		} catch ( RemoteException e ) {
			printGumStatus( currentGuMSpec.getAttribute( GumSpec.ATT_NAME ), currentGuMSpec
					.getAttribute( GumSpec.ATT_PORT ), UAAdminStatus.ALREADY_STOPPED );

		} catch ( Exception e ) {
			throw new DeployerException( mgroot, e );
		}

	}


	/**
	 * Checks status of each user agent GuMs from MyGump
	 * 
	 * @throws DeployerException if any problem occurs during the test of the
	 *         user agent.
	 */
	private void status( GumSpec currentGuMSpec ) {

		if ( Deployer.gumIsUserAgent( currentGuMSpec ) ) {

			GumClient ua;

			try {
				ua = new UserAgentClient( GumSpecServices.createGum( currentGuMSpec ), currentGuMSpec );
				printGumStatus( currentGuMSpec.getAttribute( GumSpec.ATT_NAME ), currentGuMSpec
						.getAttribute( GumSpec.ATT_PORT ), ua.getStatus() );
			} catch ( Exception e ) {
				printGumStatus( currentGuMSpec.getAttribute( GumSpec.ATT_NAME ), currentGuMSpec
						.getAttribute( GumSpec.ATT_PORT ), UAAdminStatus.DOWN );
			}

		} else {
			System.out.println( "--> " + currentGuMSpec.getAttribute( GumSpec.ATT_NAME ) + ":"
					+ currentGuMSpec.getAttribute( GumSpec.ATT_PORT ) + " [ Can't verify. Not ualinux type. ] " );
		}

	}


	/**
	 * Tries to restart user agents from localGuMP
	 * 
	 * @throws DeployerException
	 * @throws DeployerException if any error occurs during this initialization.
	 */

	private void restart( GumSpec currentGuMSpec ) throws DeployerException {

		GumClient uaGuM;
		try {

			try {

				uaGuM = new UserAgentClient( GumSpecServices.createGum( currentGuMSpec ), currentGuMSpec );

				if ( uaGuM.getStatus().isUp() ) {
					stop( currentGuMSpec );

				}

			} catch ( RemoteException e ) {
				// Nothing to be done, the gum is already down, just start it

			} finally {

				start( currentGuMSpec );
			}

		} catch ( GumSpecificationException e ) {
			throw new DeployerException( mgroot, e );
		} catch ( Exception e ) {
			throw new DeployerException( mgroot, e );
		}
	}


	/**
	 * Tries to install user agent software at a specific remote GuM
	 * 
	 * @param currentGuMSpec The GuMSpec object that describes the GuM in which
	 *        User Agent software will be installed
	 * @throws DeployerException if could not install the User Agent software.
	 */
	private void install() throws DeployerException {

		String message;
		String finalDirName = getUADirName( dirName );

		String uaPropertiesFilePath = uaPropertiesDir + File.separator + dirName;
		String zipFileName = dirName + ZIP_EXTENSION;
		String zipLocationName = uaPropertiesDir + File.separator + zipFileName;

		try {
			/** getting correct UA.properties file for this GuM */
			File currentUAPropertiesFile = locateUAPropertiesFiles( dirName );

			// command to be executed on the remote gum for creating the
			// .ourgrid/gumName directory at the gum
			String execCommand = StringUtil.replace( gumRemExec, flagMachine, gumName );
			String remoteCommand = "'mkdir -p " + finalDirName + "'";
			execCommand = StringUtil.replace( execCommand, flagCommand, remoteCommand );

			// using Executor class for remote execution
			eh = executor.execute( "/tmp", execCommand );
			execResult = ExecutorFactory.getExecutorInstance().getResult( eh );

			if ( execResult.getExitValue() == 0 ) {
				// dirName succesfully created on gumName

				// renaming local UA.properties file to remove gum name and
				// "auto" flag
				File gumDirectory = new File( uaPropertiesFilePath );
				gumDirectory.mkdir();
				currentUAPropertiesFile.renameTo( new File( uaPropertiesFilePath + File.separator + "ua.properties" ) );
				String currentFilesToZip[] = getFilesToZip( uaPropertiesFilePath );
				File zipFile = createZip( zipLocationName, currentFilesToZip );

				// command for transferring zipName file to the gum
				String destFile = finalDirName;
				String copyCommand = StringUtil.replace( gumCopyTo, flagMachine, gumName );
				copyCommand = StringUtil.replace( copyCommand, flagLocalFile, zipLocationName );
				copyCommand = StringUtil.replace( copyCommand, flagRemoteFile, destFile );

				// Transferring zipFile to the remote gum
				eh = executor.execute( ".", copyCommand );
				execResult = ExecutorFactory.getExecutorInstance().getResult( eh );

				if ( execResult.getExitValue() == 0 ) {
					// zipName succesfully transferred to the gum 'gumName'

					// command for transferring jarFile file to the gum
					destFile = finalDirName + File.separator + "useragent.jar";
					copyCommand = StringUtil.replace( gumCopyTo, flagMachine, gumName );
					copyCommand = StringUtil.replace( copyCommand, flagLocalFile, mgroot + "/lib/useragent.jar" );
					copyCommand = StringUtil.replace( copyCommand, flagRemoteFile, destFile );

					// Transferring jarFile to the remote gum
					eh = executor.execute( "/tmp", copyCommand );
					execResult = ExecutorFactory.getExecutorInstance().getResult( eh );

					if ( execResult.getExitValue() == 0 ) {
						// jarFile succesfully transferred to the gum
						// 'gumName'

						// unzip the zipFile on the remote gum
						execCommand = StringUtil.replace( gumRemExec, flagMachine, gumName );
						String remoteCmd = "'cd " + finalDirName
								+ "; java -cp useragent.jar org.ourgrid.common.util.SimpleZip -x " + zipFileName
								+ " . ; rm -f " + zipFileName + "'";
						execCommand = StringUtil.replace( execCommand, flagCommand, remoteCmd );

						// Unzipping zipName at the 'gumName'
						eh = executor.execute( "/tmp", execCommand );
						execResult = ExecutorFactory.getExecutorInstance().getResult( eh );

						if ( execResult.getExitValue() == 0 ) {
							// unzip was successfully done at the 'gumName'
							// deleting renamed local UA.properties file
							// from "uasconf" directory
							String localCommand = "'/bin/rm -Rf " + uaPropertiesFilePath + "'";
							eh = executor.execute( ".", localCommand );
							execResult = ExecutorFactory.getExecutorInstance().getResult( eh );

							zipFile.delete();
							gumDirectory.delete();

						} else {
							message = "Could not remove local temp file ' " + uaPropertiesFilePath;
							LOG.error( message );
						}
					} else {
						message = "Could not transfer 'useragent.jar' to the gum " + dirName;
						LOG.error( message );
					}
				} else {
					message = "Could not transfer '" + zipLocationName + "' to the gum " + dirName;
					LOG.error( message );
				}
			} else {
				message = "Could not create " + finalDirName + " at " + dirName;
				LOG.error( message );
			}

			// Deleting the UAProperties file.
			currentUAPropertiesFile.delete();

		} catch ( Exception e ) {
			message = "There was a problem during installation on " + dirName + ". User Agent was not installed there.";
			LOG.error( message, e );
		}
	}


	/**
	 * Locates the UA.properties file associated with the specified GuM
	 * 
	 * @param gumName The GuM name
	 * @return A File object that references the UA.properties file associated
	 *         with the specified GuM
	 * @throws FileNotFoundException When it was not possible to find the
	 *         UA.properties file that matches the specified GuM
	 */
	private File locateUAPropertiesFiles( String gumName ) throws FileNotFoundException {

		/** get UA.properties files list */
		File uaPropertiesDirFile = new File( uaPropertiesDir );
		File[ ] uaPropertiesFiles = uaPropertiesDirFile.listFiles();

		/** iterating on UA.properties files */
		if ( uaPropertiesFiles != null ) {

			for ( int i = 0; i < uaPropertiesFiles.length; i++ ) {
				/** getting the correct UA.properties file */
				if ( uaPropertiesFiles[i].getName().equals( "ua.properties-" + gumName + "-auto" ) ) {
					return uaPropertiesFiles[i];
				}
			}
		}

		/** couldn't find the correct UA.properties file */
		String message = "Couldn't find ua.properties file for GuM " + gumName + ".";
		throw new FileNotFoundException( message );
	}


	/**
	 * Creates a zip package containing files to transfer to grid machine.
	 * 
	 * @param zipName The package filename (including the extension)
	 * @param filesToTransfer An array containing the files names that will be
	 *        into the package.
	 * @return A File that is a logical representation of the package created.
	 * @throws IOException In case unable to zip.
	 */
	private File createZip( String zipName, String[ ] filesToTransfer ) throws IOException {

		SimpleZip.zipIt( filesToTransfer, zipName );

		return new File( zipName );
	}


	/**
	 * Compose user agent installation directory with the GuM name
	 * 
	 * @param gumName The remote GuM name
	 * @return A String with a USER AGENT INSTALLATION directory complete path.
	 */
	public static String getUADirName( String gumName ) {

		return ".ourgrid/" + gumName;
	}


	/**
	 * Cleans the environment, deleting the user's <b>.ougrid</b> dir, and the
	 * <b>uasconf</b> dir.
	 */
	private void clean() {

		delOurgrid();

		delUasconf();
	}


	private void delUasconf() {

		try {
			executor.execute( ".", "/bin/rm -rf uasconf/" + dirName + ZIP_EXTENSION );
			executor.execute( ".", "/bin/rm -rf uasconf/" + dirName + "/" );
		} catch ( ExecutorException e ) {
			LOG.error( "Cannot delete uasconf" );
		}

	}


	private void delOurgrid() {

		try {
			executor.execute( "/tmp", "cd ~; /bin/rm -Rf .ourgrid" );
		} catch ( ExecutorException e ) {
			LOG.error( "Cannot delete .ourgrid" );
		}

	}


	private void printGumStatus( String gumName, String port, Enum gumStatus ) {

		String message = "--> " + gumName + ":" + port + " [ " + gumStatus + " ]";
		System.out.println( message );
		LOG.info( message );
	}

	private enum UAAdminStatus {

		UP, DOWN, RESTARTED, STARTED, STOPPED, ALREADY_STARTED, ALREADY_STOPPED;
	}

}
