/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.gridmachine.services;

import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.exception.GumSpecificationException;
import org.ourgrid.gridmachine.services.creators.FakeGumCreator;
import org.ourgrid.gridmachine.services.creators.GumCreator;
import org.ourgrid.gridmachine.services.creators.UserAgentClientCreator;
import org.ourgrid.gridmachine.services.validators.DefaultGumSpecValidator;
import org.ourgrid.gridmachine.services.validators.GumSpecValidator;
import org.ourgrid.gridmachine.services.validators.UserAgentClientValidator;

/**
 * This class defines a kind of GuMServicesBuilder able of creating and
 * validating GuMSpecs of the following types: user_agent, grid_script,
 * ua_gateway, fake_gum. <br>
 */
public class DefaultGumServicesBuilder implements GumServicesBuilder {

	/**
	 * The specific GuMCreator used to create a gum according to a GuMSpec
	 * passed as a parameter to thisb
	 */
	private GumCreator specificGumCreator;

	/**
	 * The specific validator for gums according to the GuMSpec passed as
	 * argument to the constructor.
	 */
	private GumSpecValidator specificGumSpecValidator;


	/**
	 * Constructor.
	 * 
	 * @param spec a given GuMSpec.
	 * @throws GumSpecificationException if it is not possible to create a
	 *         GuMCreator or a GuMSpecValidator for a GuMSpec.
	 */
	public DefaultGumServicesBuilder( GumSpec spec ) throws GumSpecificationException {

		this.loadCreatorAndValidatorOfGuM( spec );
	}


	/**
	 * Loads the specific GuMCreator and GuMSpecValidator.
	 * 
	 * @throws GumSpecificationException when the type attribute is not valid.
	 */
	private void loadCreatorAndValidatorOfGuM( GumSpec spec ) throws GumSpecificationException {

		Object typeObj = spec.getAttribute( GumSpec.ATT_TYPE );
		if ( typeObj == null ) {
			throw new GumSpecificationException( "There is no type attribute for this spec" );
		}
		String typeStr = ((String) typeObj).toLowerCase();

		if ( typeStr.equalsIgnoreCase( GumSpec.TYPE_FAKE_GUM ) ) {
			this.specificGumCreator = new FakeGumCreator();
			this.specificGumSpecValidator = new DefaultGumSpecValidator();

			// if GuMSpec type is UA_LINUX
		} else if ( typeStr.equalsIgnoreCase( GumSpec.TYPE_UA_LINUX ) ) {
			this.specificGumCreator = new UserAgentClientCreator();
			this.specificGumSpecValidator = new UserAgentClientValidator();

			// TODO implement UA_WINDOWS
		} else if ( typeStr.equalsIgnoreCase( GumSpec.TYPE_UA_WINDOWS ) ) {
			this.specificGumCreator = new UserAgentClientCreator();
			this.specificGumSpecValidator = new UserAgentClientValidator();

			// TODO implement UA_SWAN
		} else if ( typeStr.equalsIgnoreCase( GumSpec.TYPE_UA_SWAN ) ) {
			this.specificGumCreator = new UserAgentClientCreator();
			this.specificGumSpecValidator = new UserAgentClientValidator();

		} else {
			throw new GumSpecificationException( "The type \"" + typeStr + "\" of this GuM is not valid." );
		}
	}


	/**
	 * Gets an specific GuMCreator according to the GuMSpec passed as parameter
	 * to the constructor.
	 * 
	 * @return an object that can create gums.
	 */
	public GumCreator getGumCreator() {

		return this.specificGumCreator;
	}


	/**
	 * Gets the GuMSpecValidator according to the GuMSpec passed as parameter to
	 * the constructor.
	 * 
	 * @return an object that validates gums.
	 */
	public GumSpecValidator getGumSpecValidator() {

		return this.specificGumSpecValidator;
	}

}
