/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.gridmachine.services;

import java.net.MalformedURLException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.ExportException;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.exception.GumSpecificationException;
import org.ourgrid.common.url.UserAgentURLProvider;
import org.ourgrid.gridmachine.useragent.UserAgentServerImpl;
import org.ourgrid.gridmachine.useragent.test.FakeUserAgentServer;
import org.ourgrid.test.SchedulerFastTest;
import org.ourgrid.test.TestUtils;

/**
 * This class is used to test the service of creating and validating gums. <br>
 */
public class GumSpecServicesTest extends TestCase implements SchedulerFastTest {

	private TestUtils testUtils;


	public void setUp() throws Exception {

		testUtils = new TestUtils();
		testUtils.load();

		Configuration.getInstance( Configuration.USERAGENT );

	}


	public void tearDown() throws Exception {

		Configuration.reset();
	}


	/**
	 * Tests the creating of a FakeGuM from a GuMSpec.
	 * 
	 * @throws GumSpecificationException
	 * @throws NotBoundException
	 * @throws RemoteException
	 * @throws MalformedURLException
	 */
	public void testFakeSpecServices() throws GumSpecificationException, MalformedURLException, RemoteException,
		NotBoundException {

		GumSpec spec = new GumSpec();
		spec.putAttribute( GumSpec.ATT_OS, "linux" );
		spec.putAttribute( GumSpec.ATT_TYPE, GumSpec.TYPE_FAKE_GUM );

		assertTrue( GumSpecServices.isSpecValid( spec ) );
		Gum gum;
		gum = GumSpecServices.createGum( spec );
		assertTrue( gum instanceof FakeUserAgentServer );

	}


	/**
	 * Tests the creation of a UserAgent gum from a GuMSpec.
	 * 
	 * @throws GumSpecificationException When the GuMSpec is invalid.
	 */
	public void testUserAgentSpecServices() throws Exception {

		GumSpec spec = testUtils.getLocalGuMSpec( TestUtils.TYPE_UALINUX );
		spec.putAttribute( GumSpec.ATT_PORT, testUtils.getProperty( TestUtils.RMI_PORT ) );
		
		Configuration.getInstance().setProperty( Configuration.PROP_PORT, testUtils.getProperty( TestUtils.RMI_PORT ) );

		int port = Integer.parseInt( spec.getAttribute( GumSpec.ATT_PORT ) );
		Gum uaServer = new UserAgentServerImpl();

		Registry reg = null;

		try {

			reg = LocateRegistry.createRegistry( port );

		} catch ( ExportException e ) {
			System.out.println( "MESSAGE: " + e.getMessage() );
			System.out.println( "WARNING: There is already a REGISTRY created on this port (" + port
					+ "). Trying to get the current Registry." );
			reg = LocateRegistry.getRegistry( port );
			System.out.println( "MGGuMSpecTest....." );
		}

		reg.rebind( UserAgentURLProvider.USERAGENT, uaServer );

		assertTrue( GumSpecServices.isSpecValid( spec ) );
		Gum gum = GumSpecServices.createGum( spec );
		assertNotNull( gum );

		reg.unbind( UserAgentURLProvider.USERAGENT );
	}


	/**
	 * Tests the creation of a UserAgent gum from a GuMSpec, when it is not
	 * valid.
	 * 
	 * @throws NotBoundException
	 * @throws RemoteException
	 * @throws MalformedURLException
	 */
	public void testBadUserAgentSpecWithoutPort() throws MalformedURLException, RemoteException, NotBoundException {

		GumSpec spec = new GumSpec();
		spec.putAttribute( GumSpec.ATT_OS, "linux" );
		spec.putAttribute( GumSpec.ATT_TYPE, GumSpec.TYPE_UA_LINUX );
		spec.putAttribute( GumSpec.ATT_NAME, "localhost" );

		// As can be seen, we have omitted the UserAgent port attribute.

		try {
			assertFalse( GumSpecServices.isSpecValid( spec ) );
			GumSpecServices.createGum( spec );
			fail( "A GuMSpecificationException should be thrown" );
		} catch ( GumSpecificationException e ) {
			assertTrue( "A GuMSpecificationException must be thrown in this condition", true );
		}
	}


	/**
	 * Tests if a exception is throwaded with a GuMSpec, when it is not have
	 * type.
	 * 
	 * @throws NotBoundException
	 * @throws RemoteException
	 * @throws MalformedURLException
	 */
	public void testGuMCreationWithoutType() throws MalformedURLException, RemoteException, NotBoundException {

		GumSpec spec = new GumSpec();
		spec.putAttribute( GumSpec.ATT_OS, "linux" );
		spec.putAttribute( GumSpec.ATT_REM_EXEC, "testUserAgent" );
		spec.putAttribute( GumSpec.ATT_PORT, "7891" );

		// As can be seen, we have omitted the UserAgent Type attribute.

		try {
			assertFalse( GumSpecServices.isSpecValid( spec ) );
			GumSpecServices.createGum( spec );
			fail( "A GuMSpecificationException should be thrown" );
		} catch ( GumSpecificationException e ) {
			assertTrue( "A GuMSpecificationException must be thrown in this condition", true );
		}
	}


	/**
	 * Tests if a exception is throwaded with a GuMSpec, when it is not have
	 * name.
	 * 
	 * @throws NotBoundException
	 * @throws RemoteException
	 * @throws MalformedURLException
	 */
	public void testUserAgentGuMCreationWithoutName() throws MalformedURLException, RemoteException, NotBoundException {

		GumSpec spec = new GumSpec();
		spec.putAttribute( GumSpec.ATT_OS, "linux" );
		spec.putAttribute( GumSpec.ATT_REM_EXEC, "testUserAgent" );
		spec.putAttribute( GumSpec.ATT_PORT, "7891" );
		spec.putAttribute( GumSpec.ATT_TYPE, GumSpec.TYPE_UA_LINUX );

		// As can be seen, we have omitted the UserAgent Name attribute.

		try {
			assertFalse( GumSpecServices.isSpecValid( spec ) );
			GumSpecServices.createGum( spec );
			fail( "A GuMSpecificationException should be thrown" );
		} catch ( GumSpecificationException e ) {
			assertTrue( "A GuMSpecificationException must be thrown in this condition", true );
		}
	}

}
