/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.gridmachine.useragent;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.StringTokenizer;

import org.ourgrid.gridmachine.useragent.exception.UserAgentPauseException;
import org.ourgrid.gridmachine.useragent.ui.ErrorCode;

/**
 * This idleness detector is not multiplataform (I'm sorry) as JAVA is. It was
 * made to detect if the user is using hers workstation at Linux. At the
 * development it will be increasely done by adding more paramether to be more
 * and more sure about its assertion. Created on 10/20/2004
 * 
 * @version 0.1 - At this first version it will only check if the keyboard and
 *          mouse are used. To implement this, we check the "file"
 *          /proc/interrupts.
 */
public class IdlenessCheckerDaemon extends Thread {

	/* Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( IdlenessCheckerDaemon.class );

	/** Store the last readed interrupt file status */
	private LinkedList<String[ ]> actualStatus = new LinkedList<String[ ]>();

	private LinkedList<String[ ]> oldStatus = new LinkedList<String[ ]>();

	/**
	 * The time that the machine have to be IDLE to initialize the UserAgent in
	 * seconds
	 */
	private int cicleTime;

	/** The UA managed by this daemon */
	private UserAgentServerImpl ua;

	/**
	 * Variable that indicates if the deamon should stop.
	 */
	private boolean shouldShutdown = false;


	/**
	 * Initializes the daemon that will check (until be killed) the machine
	 * idleness. If the machine, at any time, stay idle for more than the
	 * specified time then the detector will start the managed OurGrid UserAgent
	 * that will wait for tasks. When the UA is started, the daemon will begin
	 * to wait for the user return to pause the agent.
	 * 
	 * @param ua the UserAgent managed by this daemon
	 * @param timeForCicle the time in minutes that the computer have to be idle
	 *        before the UA starts.
	 */
	public IdlenessCheckerDaemon( UserAgentServerImpl ua, int timeForCicle ) {

		this.ua = ua;
		cicleTime = timeForCicle;
		start();
	}


	/**
	 * Thread that do what this class have to do :)
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run() {

		try {
			this.readInterruptsFile(); // Makes a initial status before call
			// the wasUsed method.
			while ( !shouldShutdown ) {
				this.tryToStartUA();
				this.tryToPauseUA();
			}
		} catch ( InterruptedException iex ) {
			LOG
					.error( "Could not count to start/pause the execution! Probably this is a temporary error and will not cause damages." );

			this.run(); // TODO: This is a good way to continue the work of this
			// method????
		} catch ( IOException ioex ) {
			LOG
					.error( "Idleness detection will only work on the Linux plataform.\n"
							+ "The problem is that the file \"/proc/interrupts\" could not be used properly. PLEASE REPORT THIS!!!" );
			System.exit( ErrorCode.FAIL );
		}
	}


	/**
	 * Sends a signal for the deamon to shutdown
	 */
	protected void shutdown() {

		this.shouldShutdown = true;

	}


	/**
	 * Just how the name says, waits until the machine is idle and trys to start
	 * the OurGrid UserAgent.
	 * 
	 * @throws InterruptedException This method uses Thread.sleep to coult the
	 *         times so this exception is thrown if any error occurs at this
	 *         call.
	 */
	private void tryToStartUA() throws InterruptedException {

		int counter = 0;
		while ( counter < (cicleTime * 1000) ) { // Translating seconds do
			// miliseconds
			boolean idle = this.isIdle();

			if ( idle )
				counter = counter + (5 * 1000); // Adds each seconds that
			// the thread will sleep
			else
				counter = 0;

			Thread.sleep( 5 * 1000 );
		}
		LOG.info( "Starting UA!" );
		this.ua.restart();
	}


	/**
	 * Just how the name says, waits until the machine not-idle-anymore time and
	 * trys to pause the OurGrid UserAgent.
	 * 
	 * @throws InterruptedException This method uses Thread.sleep to coult the
	 *         times so this exception is thrown if any error occurs at this
	 *         call.
	 */
	private void tryToPauseUA() throws InterruptedException {

		boolean idle = true;

		while ( idle ) {
			idle = this.isIdle();

			if ( idle )
				Thread.sleep( 1 * 1000 );
		}
		LOG.info( "Pausing UA!" );
		try {
			this.ua.pause();
		} catch ( UserAgentPauseException e ) {
			LOG.error( e.getMessage() );
		}
	}


	/**
	 * Compare the last readed status with the actual readed status to tell if
	 * the machine is idle. It prepares the old and actual status before to
	 * compare.
	 * 
	 * @return <code>true</code> if the machine is idle and <code>false</code>
	 *         otherwise.
	 */
	private boolean isIdle() {

		try {
			this.oldStatus = actualStatus;
			this.actualStatus = new LinkedList<String[ ]>();
			this.readInterruptsFile();
		} catch ( IOException ioe ) {
			ioe.printStackTrace();
		}

		return this.checkIdleness();
	}


	/**
	 * Complets the work of this.isIdle() method. It makes the comparation
	 * itself.
	 * 
	 * @return <code>true</code> if the computer is idle and
	 *         <code>false</code> otherwise.
	 */
	private boolean checkIdleness() {

		boolean idle = true;
		Iterator<String[ ]> it = actualStatus.iterator();
		while ( it.hasNext() && idle == true ) {

			String[ ] device = it.next();

			if ( device.length == 4 ) {
				String name = device[3];
				String deviceId = device[0];
				if ( name.equals( "keyboard" ) || name.equals( "PS/2Mouse" ) || deviceId.equals( "1:" )
						|| deviceId.equals( "12:" ) ) {
					String interrupt = device[1];
					int index = actualStatus.indexOf( device );
					String[ ] oldStatusObj = oldStatus.get( index );
					String oldInterrupt = oldStatusObj[1];
					if ( !interrupt.equals( oldInterrupt ) )
						idle = false;
				}
			}
		}
		return idle;
	}


	/**
	 * This method will get the last status of interruptions provided by the
	 * linux system. It reads the file "/proc/interrupts" and stores the actual
	 * state at a data structure.
	 * 
	 * @throws IOException signalize the non existence of the "/proc/interrupts"
	 *         file or if it could not be correctly used. Note that the file
	 *         should exist if you are using Linux system. If not PLEASE
	 *         REPORT!!!
	 */
	private void readInterruptsFile() throws IOException {

		BufferedReader interruptState = null;
		try {
			interruptState = new BufferedReader( new FileReader( "/proc/interrupts" ) );

			interruptState.readLine(); // Ignoring the first line that is not a
			// device.
			String deviceStatus = interruptState.readLine();
			while ( !(deviceStatus == null) ) {
				this.storeStatus( deviceStatus );
				deviceStatus = interruptState.readLine();
			}
			interruptState.close();
		} catch ( IOException e ) {
			throw e;
		} finally {
			if ( interruptState != null ) {
				try {
					interruptState.close();
				} catch ( IOException e1 ) {
					LOG.debug( "Failed to close stream on Exception." );
				}
			}
		}
	}


	/**
	 * Stores informations about a device at a array and puts it at the List
	 * mantained by this class. The number of informations about the devices can
	 * vary and because of this the algorithm is a bit more complicated.
	 * 
	 * @param interruptFileLine A line readed from the interruptions file.
	 */
	private void storeStatus( String interruptFileLine ) {

		String[ ] info;
		StringTokenizer dsTokenizer = new StringTokenizer( interruptFileLine, " " );
		int tokensNumber = dsTokenizer.countTokens();
		if ( tokensNumber > 4 ) {
			info = new String[ 4 ];
		} else {
			info = new String[ tokensNumber ];
		}
		for ( int counter = 0; counter < info.length; counter++ ) {
			info[counter] = dsTokenizer.nextToken().trim();
		}
		if ( tokensNumber == 5 ) {
			info[3] = info[3] + dsTokenizer.nextToken().trim();
		}
		// Store the readed informations always at the actualStatus list.
		this.actualStatus.addLast( info );
	}

}
