/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ourgrid.gridmachine.useragent;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.ourgrid.common.exception.UnableToCreatePlaypenException;
import org.ourgrid.common.exception.UnableToCreateStorageException;
import org.ourgrid.common.exception.UnableToDigestFileException;
import org.ourgrid.common.exception.UnavailableGumException;
import org.ourgrid.common.executor.ExecutorException;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.ExecutorResult;
import org.ourgrid.common.executor.Win32Executor;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.util.JavaFileUtil;
import org.ourgrid.gridmachine.exception.FilePermissionException;
import org.ourgrid.gridmachine.exception.FileTransferException;
import org.ourgrid.gridmachine.exception.UnableToExecuteException;
import org.ourgrid.mygrid.scheduler.DataDiscovery;

/**
 * Concrete implementation of the <code>GumClient</code> interface that
 * provides methods for remote execution, file transfer and information
 * retrieval like remote file information.
 * <p>
 * It is the representation of a <code>Gum</code> in the broker side. It will
 * use the methods of <code>Gum</code> to provide services to the broker users
 */
public class UserAgentClient implements GumClient {

	private static final int FILE_PERMISSION = 700;

	/** Logger (log4j) object to store log events. */
	private static final transient org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( UserAgentClient.class );

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/**
	 * This constant is the flag used in the REM_EXEC spec field to indicate the
	 * command must be executed at the Gum.
	 */
	public static final String FLAG_COMMAND = "$command";

	/**
	 * This constant is the flag used in the REM_EXEC spec field to indicate
	 * which Gum must execute the command.
	 */
	public static final String FLAG_MACHINE = "$machine";

	/** An spec object that contains the specification of the Grid Machine. */
	private GumSpec spec;

	/**
	 * The FileInputStream used to read a local file that will be transfered to
	 * the grid machine.
	 */
	public transient FileInputStream localFileStream;

	/** size in bytes of the read file blocks. */
	private static int sizeEachRead = 4096;

	/** Lock-related objects. */
	protected transient FileChannel channel;

	/** A reference to the grid machine. */
	protected Gum ua;

	/**
	 * Map containing all the executor handles of the remote executions on this
	 * UserAgent. The map is updated on each getResult() invocation.
	 */
	protected Map<ExecutorHandle,ExecutorHandle> executorHandles;

	/**
	 * This abstract representation for a Random Access file is used to obtain a
	 * exclusive lock during file transfer operations.
	 */
	public transient RandomAccessFile randomFile;

	/**
	 * This object represents the lock for a given file. This lock is intended
	 * to be exclusive and should avoid concurrent writing operations.
	 */
	public transient FileLock localFileLock;

	/** The string representation for the playpen used in a given execution. */
	protected String playpenPath;


	/**
	 * Default constructor.
	 * 
	 * @param gum a remote reference the the grid machine
	 * @param gumSpec information about the gum features
	 */
	public UserAgentClient( Gum gum, GumSpec gumSpec ) {

		this.spec = gumSpec;
		this.ua = gum;
	}


	public void storeFile( String localFilePath, String remoteFilePath ) throws UnableToExecuteException,
		UnavailableGumException, FileTransferException {

		/* Get access to the DataDiscovery reference */
		DataDiscovery dd = DataDiscovery.getInstance();

		/* The local and remote file digests to be compared */
		String localFileDigest;
		String remoteFileDigest;

		String storageDirectory;

		storageDirectory = getStorageDirectory();

		if ( storageDirectory != null )
			remoteFilePath = getRemoteFileAbsolutePath( remoteFilePath, storageDirectory );

		/* That directory describes the shared remote repository */
		File remoteFile;

		/* Get the file to be transferred */
		File fileToStore = new File( localFilePath );

		if ( !fileToStore.exists() ) {
			throw new FileTransferException( fileToStore.getAbsolutePath(), new FileNotFoundException( "File "
					+ localFilePath + " does not exists in the home machine!" ), true );
		}

		LOG.debug( "The lock for file: " + localFilePath + " is being requested by " + this.spec.getGumID() );

		FileLockManager.getInstance().requestLock( fileToStore.getAbsolutePath() ).blockingRemove();

		LOG.debug( "The lock for file:" + localFilePath + " was granted to " + this.spec.getGumID() );

		remoteFile = new File( remoteFilePath );

		try {
			// Moved by Hunter Hagewood 28/02/2005
			// Trying to digest the local file.
			/*
			 * This is done separately from the remote file to, eventually,
			 * throw an exactly exception message!!!
			 */
			try {
				localFileDigest = JavaFileUtil.getDigestRepresentation( fileToStore );
			} catch ( UnableToDigestFileException exception ) {
				throw new FileTransferException( fileToStore.getAbsolutePath(), exception, true );
			}

			/* Check about file existence */
			if ( !this.fileExists( remoteFile ) ) {

				/* transfer file indeed */
				this.putFile( fileToStore.getAbsolutePath(), remoteFilePath );

				/*
				 * Added by Hunter Hagewood 28/02/2005 Record the transfer in
				 * the DataDiscovery table
				 */
				dd.putDDEntry( localFileDigest, remoteFilePath, fileToStore.length(), spec );

			} else if ( fileToStore.exists() ) {

				/*
				 * Check about file contents. If local and remote files contents
				 * are the same, it is not necessary to transfer
				 */
				try {

					/** Trying to digest the remote file */
					remoteFileDigest = this.getUAServer().getRemoteFileDigest( remoteFile );

					/** ** Comparing the digest values *** */
					if ( !localFileDigest.equals( remoteFileDigest ) ) {
						/* transfer file indeed */
						this.putFile( fileToStore.getAbsolutePath(), remoteFilePath );

						/*
						 * Added by Hunter Hagewood 28/02/2005 Record the
						 * transfer in the Data Discovery table
						 */
						dd.putDDEntry( localFileDigest, remoteFilePath, fileToStore.length(), spec );

						/* Remove any invalid locations */
						dd.modDDEntry( remoteFileDigest, remoteFilePath, spec );

					} else {
						this.changeFilePermissions( remoteFile, FILE_PERMISSION );
					}

				} catch ( UnableToDigestFileException exception ) {
					throw new FileTransferException( fileToStore.getAbsolutePath(), exception, true );

				} catch ( RemoteException exception ) {
					this.ua = null;
					throw new UnavailableGumException( this.spec.getGumID().toString(), exception );
				}

			}

		} finally {

			FileLockManager.getInstance().releaseLock( fileToStore.getAbsolutePath() );
			LOG.debug( "The lock for file:" + localFilePath + " was released by " + this.spec.getGumID() );

		}

	}


	/**
	 * Converts a relative remote file path to absolute path.
	 * 
	 * @param remoteFile remoteFile to check and convert to absolute file path
	 * @param relativeDirectory base directory for the file
	 * @return absolute remote file path
	 */
	private String getRemoteFileAbsolutePath( String remoteFile, String relativeDirectory ) {

		if ( this.spec.getAttribute( GumSpec.ATT_TYPE ).equalsIgnoreCase( GumSpec.TYPE_UA_WINDOWS ) ) {
			remoteFile = Win32Executor.convert2WinStyle( remoteFile );
			if ( !JavaFileUtil.isAbsolutePath( remoteFile ) ) {
				remoteFile = relativeDirectory + '\\' + remoteFile;
			}
		} else {
			if ( !JavaFileUtil.isAbsolutePath( remoteFile ) ) {
				remoteFile = relativeDirectory + '/' + remoteFile;
			}
		}
		return remoteFile;
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#fileExists(java.io.File)
	 */
	public boolean fileExists( File file ) throws UnavailableGumException {

		try {
			return getUAServer().fileExists( file );
		} catch ( RemoteException e ) {
			this.ua = null;
			throw new UnavailableGumException( getGumID().toString(), e );
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getRemoteFileLength(java.io.File)
	 */
	public long getRemoteFileLength( File file ) throws UnavailableGumException, UnableToExecuteException {

		try {
			return getUAServer().getRemoteFileLength( file );
		} catch ( RemoteException e ) {
			this.ua = null;
			throw new UnavailableGumException( getGumID().toString(), e );
		} catch ( IOException e ) {
			throw new UnableToExecuteException( getGumID().toString(), e );
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getRemoteFileLastModification
	 *      (java.io.File, boolean)
	 */
	public long getRemoteFileLastModification( File remoteFile, boolean normalized ) throws UnavailableGumException,
		UnableToExecuteException {

		try {
			return this.getUAServer().getRemoteFileLastModification( remoteFile, normalized );
		} catch ( RemoteException e ) {
			this.ua = null;
			throw new UnavailableGumException( getGumID().toString(), e );
		} catch ( IOException e ) {
			throw new UnableToExecuteException( getGumID().toString(), e );
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#createPlaypen(int)
	 */
	public String createPlaypen( int size ) throws UnavailableGumException, UnableToCreatePlaypenException {

		try {
			playpenPath = this.getUAServer().createPlaypen( size );
			return playpenPath;
		} catch ( RemoteException e ) {
			this.ua = null;
			throw new UnavailableGumException( getGumID().toString(), e );
		}
	}


	/**
	 * @return the gum reference
	 */
	public Gum getUAServer() throws RemoteException {

		if ( ua == null ) {
			throw new RemoteException( "UserAgent is Unavailable" );
		}
		return ua;
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#createStorageDirectory()
	 */
	public String createStorageDirectory() throws UnavailableGumException, UnableToCreateStorageException {

		try {
			return this.getUAServer().createStorageDirectory();
		} catch ( RemoteException exception ) {
			this.ua = null;
			throw new UnavailableGumException( getGumID().toString(), exception );
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getStorageDirectory()
	 */
	public String getStorageDirectory() throws UnavailableGumException {

		try {
			return this.getUAServer().getStorageDirectory();
		} catch ( RemoteException exception ) {
			this.ua = null;
			throw new UnavailableGumException( getGumID().toString(), exception );
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#remoteExecute(java.lang.String,
	 *      java.lang.String, java.util.Map)
	 */
	public ExecutorHandle remoteExecute( String remoteDir, String command, Map<String,String> environmentVars )
		throws UnavailableGumException, UnableToExecuteException {

		try {

			ExecutorHandle handle = this.getUAServer().remoteExecute( remoteDir, command, environmentVars );

			/*
			 * The executor handles map already stores the real handles, so
			 * putting the handle during the remote execution sounds useless,
			 * but the executor handles map is only updated when the getResult
			 * method is invoked. It means that if someone does a remote
			 * execution and right after tries to kill the execution, the UA
			 * would fail because there would be no reference to the new
			 * execution in the map.
			 */
			if ( executorHandles == null ) {
				executorHandles = new HashMap<ExecutorHandle,ExecutorHandle>();
			}
			executorHandles.put( handle, handle );

			return handle;

		} catch ( RemoteException e ) {
			this.ua = null;
			throw new UnavailableGumException( getGumID().toString(), e );
		} catch ( ExecutorException e ) {
			throw new UnableToExecuteException( getGumID().toString(), e );
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#putFile(String, byte[], boolean)
	 */
	public void putFile( String localFile, String remoteFile ) throws UnavailableGumException, FileTransferException {

		if ( playpenPath != null ) {
			remoteFile = getRemoteFileAbsolutePath( remoteFile, playpenPath );
		}

		/* verify if filePath exists */
		File file = null;
		if ( localFile != null ) {
			file = new File( localFile );
		} else {
			LOG.debug( "Local file '" + localFile + "' does not exist." );
			throw new FileTransferException( localFile, new NullPointerException(), true );
		}

		/* Check if the file exists */
		if ( !file.exists() ) {
			LOG.debug( "File exists: " + file.getAbsolutePath() );
			throw new FileTransferException( localFile, new FileNotFoundException( localFile ), true );
		}

		/*
		 * Determines if this is the first slice of the file that must be
		 * transfered
		 */
		boolean firstSlice = true;

		try {

			int pos = 0;
			LOG.info( "Sending the file " + file.getName() + " to " + remoteFile + " at " + this.spec.getGumID() );

			try {

				/* Create an Inputstream to get the bytes from files */
				localFileStream = new FileInputStream( file );

				do {

					/* size is the amount of bytes that will be sent */
					int size = sizeEachRead;

					/* set the correct value of size up */
					if ( (file.length() - pos) < sizeEachRead ) {
						size = (int) (file.length() - pos);
					}

					/* creating a block to be writted */
					byte[ ] data = new byte[ size ];

					localFileStream.read( data );

					try {

						/* Send the byte stream to the remote machine */
						getUAServer().putFile( remoteFile, data, firstSlice );

					} catch ( RemoteException re ) {
						this.ua = null;
						/* If the Grid Machine could not be contacted */
						throw new UnavailableGumException( this.spec.getGumID().toString(), re );

					} catch ( IOException ioe ) {

						LOG
								.error( "Problems when transferring the local file: \"" + localFile
										+ "\". to the Grid Machine (" + this.getGumID() + ").\n" + "Cause: "
										+ ioe.getMessage() );

						throw new FileTransferException( localFile, ioe, true );
					}

					/* Update the first slice mark */
					firstSlice = false;

					/* increase the position in the file */
					pos += sizeEachRead;

				} while ( pos < file.length() );

			} finally {
				/* This line must prevent "too many open files" message appears */
				if ( localFileStream != null ) {
					localFileStream.close();
				}
			}

		} catch ( IOException ioe ) {

			LOG.error( "Could not close the local file stream. "
					+ "This can lead to a \"too many files open\" message. Cause Error: " + ioe.getMessage(), ioe );

		}

		try {
			/* set the file permissions of the remote file */
			this.changeFilePermissions( new File( remoteFile ), FILE_PERMISSION );

		} catch ( UnableToExecuteException e ) {
			LOG.error( "Could not change the remote file permissions: " + e.getMessage() );
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getFile(String, long, int)
	 */
	public void getFile( String remoteFile, String localFile ) throws UnavailableGumException, FileTransferException {

		/* this is an abstraction for represents the */
		File file = new File( localFile );

		/* An output stream to write the output file */
		FileOutputStream fosFile = null;

		/* Control variable that determine the current offset in the file */
		long filePos = 0;

		/* Data array must be written in the file */
		byte[ ] data;

		/* Check if it is allowed to write to file, in case file exists. */
		if ( file.exists() && !file.canWrite() ) {
			throw new FileTransferException( remoteFile, new FilePermissionException( remoteFile ), false );
		}

		try {

			File dirOfFile = new File( file.getParent() );
			if ( !dirOfFile.exists() ) {
				dirOfFile.mkdirs();
			}
			if ( !file.exists() ) {
				file.createNewFile();
			}
			/* Init the Output stream */
			fosFile = new FileOutputStream( file );

			LOG.info( "Sending file " + remoteFile );

			do {

				/* Get bytes from remote location */
				data = getUAServer().getFile( remoteFile, filePos, sizeEachRead );

				/* Inc the local file offset */
				filePos += data.length;

				/* Write data */
				fosFile.write( data );

			} while ( data.length == sizeEachRead );

		} catch ( RemoteException re ) {
			this.ua = null;
			throw new UnavailableGumException( this.spec.getGumID().toString(), re );
		} catch ( IOException ioe ) {
			throw new FileTransferException( remoteFile, ioe, false );
		} finally {
			/*
			 * Close file. This is mandatory to avoid "Too many open files
			 * bug!". Ok, man?!
			 */
			if ( fosFile != null ) {
				try {
					fosFile.close();
				} catch ( IOException e ) {
					// Nothing to be done

				}
			}
		}
	}


	private void changeFilePermissions( File file, int permissions ) throws UnavailableGumException,
		UnableToExecuteException {

		try {
			this.getUAServer().changeFilePermissions( file, permissions );
		} catch ( RemoteException e ) {
			this.ua = null;
			throw new UnavailableGumException( getGumID().toString(), e );
		} catch ( IOException e ) {
			throw new UnableToExecuteException( getGumID().toString(), e );
		}
	}


	/**
	 * Helper method used by remoteExecute to retrieve a value's key on a Map,
	 * if this value exists. It is used to check if there's already a handle
	 * referring to this execution handle in the handle map.
	 * 
	 * @param map
	 * @param value
	 * @return A reference for the key, or null if if the value does not exist.
	 *         It will ignore matches if the map's key equals the map's object.
	 */
	private Object getKeyFrom( Map map, Object value ) {

		if ( !map.containsValue( value ) ) {
			return null;
		}

		Collection keys = map.keySet();
		Iterator it = keys.iterator();

		while ( it.hasNext() ) {
			Object key = it.next();
			Object val = map.get( key );

			if ( value == null && val == null && key != null ) {
				return key;
			}

			if ( (val != null && val.equals( value )) && (key == null || !key.equals( value )) ) {
				return key;
			}
		}

		return null;
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getResult(org.ourgrid.common.executor.ExecutorHandle)
	 */
	public ExecutorResult getResult( ExecutorHandle handle ) throws UnavailableGumException, UnableToExecuteException {

		ExecutorResult result = null;

		// Assigning the executor handle with the first execution
		// handle.
		ExecutorHandle facadeHandle = (ExecutorHandle) getKeyFrom( executorHandles, handle );

		if ( facadeHandle == null ) {
			facadeHandle = handle;
			executorHandles.put( handle, handle );
		}

		// Remotely invoking getResult on UAServer
		try {
			// See description of currentHandle for the reason
			// for not using the parameter value.
			result = this.getUAServer().getResult( handle );
		} catch ( RemoteException e ) {
			this.ua = null;
			throw new UnavailableGumException( "Problems when contacting ", e );
		} catch ( ExecutorException e ) {
			throw new UnableToExecuteException( getGumID().toString(), e );
		}
		return result;
	}


	public GumStatus getStatus() {

		GumStatus uaStatus;

		try {
			uaStatus = getUAServer().getStatus();
			return uaStatus;
		} catch ( RemoteException e ) {
			ua = null;
			return GumStatus.CONTACTING;
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#startReplica()
	 */
	public void startReplica() throws UnavailableGumException {

		try {
			getUAServer().startReplica();
		} catch ( RemoteException e ) {
			this.ua = null;
			throw new UnavailableGumException( "Problems when contacting ", e );
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#stopReplica()
	 */
	public void stopReplica() throws UnavailableGumException {

		try {
			getUAServer().stopReplica();
		} catch ( RemoteException e ) {
			this.ua = null;
			throw new UnavailableGumException( "Problems when contacting ", e );
		}
	}


	public boolean equals( Object o ) {

		if ( o instanceof UserAgentClient ) {
			UserAgentClient newClient = (UserAgentClient) o;
			return this.getGumID().equals( newClient.getGumID() );
		}

		return false;

	}


	public GumSpec getGumSpec() {

		return spec;
	}


	private GumID getGumID() {

		return getGumSpec().getGumID();
	}
}
