/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.gridmachine.useragent;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.UserAgentConfiguration;
import org.ourgrid.common.exception.UnableToCreatePlaypenException;
import org.ourgrid.common.exception.UnableToDigestFileException;
import org.ourgrid.common.executor.Executor;
import org.ourgrid.common.executor.ExecutorException;
import org.ourgrid.common.executor.ExecutorFactory;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.ExecutorResult;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.security.MGSecureClientSocketFactory;
import org.ourgrid.common.security.MGSecureServerSocketFactory;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.util.JavaFileUtil;
import org.ourgrid.gridmachine.useragent.exception.UserAgentPauseException;

public class UserAgentServerImpl extends OurgridUnicastRemoteObject implements Gum {

	/** Serial identification of the class. */
	public static final long serialVersionUID = 33L;

	/** This is the time that the machine have to be idle to start the UA. */
	public static final int TIME_TO_WAIT = 20;

	/** The remote storage directory path. */
	private String storagePath = "";

	/** String representing the XML log file absolute path. */
	protected static String XMLLogFile = "";

	/** The executor object that provides platform-dependent features. */
	protected Executor executor = ExecutorFactory.getExecutorInstance();

	/**
	 * Store all the "active" handles returned by the Executor when
	 * this.remoteExecute is called
	 */
	List<ExecutorHandle> ownHandles = Collections.synchronizedList( new ArrayList<ExecutorHandle>() );

	/** Show the actual status of this UA. * */
	private GumStatus gumState;

	/** The GumSpec of this Gum. */
	private GumSpec spec;

	/** Set to hold information about created playpens. */
	private Set<String> createdPlaypens = Collections.synchronizedSet( new HashSet<String>() );

	/** Boolean that signalizes a job start. */
	private boolean replicaStarted = false;

	/** Logger (log4j) object to store log events */
	static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( UserAgentServerImpl.class );


	/**
	 * Construct a default UserAgent.
	 * 
	 * @throws RemoteException If some problem during the creation of the
	 *         RemoteObject occurs.
	 */
	public UserAgentServerImpl() throws RemoteException {

		super();
		Runtime.getRuntime().addShutdownHook( new UAShutdownThread() );
		this.setState( GumStatus.UP );
		startsIdlenessDetectorDaemon();
	}


	/**
	 * Construct a Secure User Agent remote object. The communication between
	 * the client and this server will be done through SSL sockets.
	 * 
	 * @param port The number of the port where the remote object must be
	 *        listening to. The port number must be greater than 1024.
	 * @param clientFactory A factory for client sockets that provide secure
	 *        client sockets over SSL.
	 * @param serverFactory A factory for server sockets that provide secure
	 *        server sockets over SSL.
	 * @throws RemoteException If some problem during the creation of the
	 *         RemoteObject occurs.
	 */
	public UserAgentServerImpl( int port, MGSecureClientSocketFactory clientFactory,
								MGSecureServerSocketFactory serverFactory ) throws RemoteException {

		super( port, clientFactory, serverFactory );
		Runtime.getRuntime().addShutdownHook( new UAShutdownThread() );
		this.setState( GumStatus.UP );
		startsIdlenessDetectorDaemon();
	}


	/**
	 * Starts the IdlenessDetectorDaemon. It sets the gumstate to owner and wait
	 * the setted idleness time to change the gumstate to idle. In a time
	 * interval it will verify if the gum is in use by the owner so it will
	 * change the current GumStatus to OWNER or to IDLE when it is possible.
	 */
	private void startsIdlenessDetectorDaemon() {

		String startsDaemon = Configuration.getInstance().getProperty( UserAgentConfiguration.PROP_IDLENESS_DETECTOR );
		if ( startsDaemon.equals( "yes" ) ) {
			this.setState( GumStatus.OWNER );
			if ( !System.getProperty( "os.name" ).startsWith( "Windows" ) ) {
				int timeToWait = Integer.parseInt( Configuration.getInstance().getProperty(
						UserAgentConfiguration.PROP_IDLENESS_TIME ) );
				new IdlenessCheckerDaemon( this, timeToWait );
			}
		}
	}


	/**
	 * Verify if this UserAgent is in use by his owner.
	 * 
	 * @throws RemoteException An exception is throw when the owner is using its
	 *         machine.
	 */
	private void checkIsUsedByOwner() throws RemoteException {

		if ( this.getStatus() == GumStatus.OWNER )
			throw new RemoteException( "The UserAgent has been paused because the machine is not idle." );
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getRemoteFileLength(java.io.File)
	 */
	public long getRemoteFileLength( File remoteFile ) throws RemoteException, IOException {

		checkIsUsedByOwner();

		LOG.debug( "Verifying file (" + remoteFile.getAbsolutePath() + ") size matching... REMOTE("
				+ remoteFile.length() + ")" );

		return remoteFile.length();
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getRemoteFileLastModification(java.io.File,
	 *      boolean)
	 */
	public long getRemoteFileLastModification( File remoteFile, boolean normalized ) throws RemoteException,
		IOException {

		checkIsUsedByOwner();

		Calendar currentTime = Calendar.getInstance();
		long lastModified = -1;

		LOG.debug( "Verifying file last modified time matching... REMOTE(" + remoteFile.lastModified() + ")" );

		if ( normalized ) {
			lastModified = remoteFile.lastModified() + (currentTime.get( Calendar.ZONE_OFFSET ) * (-1));
		} else {
			lastModified = remoteFile.lastModified();
		}

		return lastModified;
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getRemoteFileDigest(java.io.File)
	 */
	public String getRemoteFileDigest( File remoteFile ) throws RemoteException, UnableToDigestFileException {

		checkIsUsedByOwner();

		if ( remoteFile.isDirectory() )
			return "0"; // Digest from a directory.
		return JavaFileUtil.getDigestRepresentation( remoteFile );
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#putFile(java.lang.String, byte[],
	 *      boolean)
	 */
	public void putFile( String remoteFile, byte[ ] data, boolean firstSlice ) throws RemoteException, IOException {

		checkIsUsedByOwner();

		File file; // Abstraction used to fill the file with bytes passed.
		FileOutputStream fileOutput; // The output stream used to store bytes
		// in the file.

		if ( data == null ) {
			LOG.error( "The data could not be stored because the array of bytes was NULL" );
			throw new IOException( "The data array is NULL" );
		}

		file = new File( remoteFile );

		/*
		 * Verify if it is needed to delete the file to store a new set of
		 * bytes.
		 */
		if ( firstSlice ) {
			LOG.info( "Receiving file " + remoteFile );
			if ( file.exists() ) {
				if ( file.isDirectory() ) {
					deleteFilesInDir( file );
				}
				file.delete();
			}
			if ( file.getParent() != null ) {
				File temp = new File( file.getParent() );
				temp.mkdirs();
			}
			file.createNewFile();
		}

		/*
		 * Initialize the output stream using the file abstraction. The array of
		 * bytes must be appended, it is indicated by the boolean parameter
		 * "true"
		 */
		fileOutput = new FileOutputStream( file.getPath(), true );

		try {
			fileOutput.write( data ); // Write the array of bytes on the file.
		} catch ( IOException e ) {
			LOG.error( "Problems when writing the file \"" + file.getPath() + "\" : " + e.getMessage() );
			throw new IOException( e.getMessage() + " (" + file.getPath() + ")." );
		} finally {
			fileOutput.close(); // In the case of exception or not we need to
			// close the stream.
		}

	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getFile(java.lang.String, long, int)
	 */
	public byte[ ] getFile( String filePath, long filePos, int size ) throws RemoteException, IOException {

		checkIsUsedByOwner();

		File file = new File( filePath );

		if ( !file.exists() )
			throw new FileNotFoundException( filePath );

		if ( !file.canRead() )
			throw new IOException( filePath );

		FileInputStream fis = null;
		try {
			if ( filePos > file.length() ) {
				size = 0;
			} else if ( (file.length() - filePos) < size ) {
				size = (int) (file.length() - filePos);
			}
			byte[ ] data = new byte[ size ];
			fis = new FileInputStream( file );
			fis.skip( filePos );
			fis.read( data );

			return data;
		} catch ( IOException e ) {
			LOG.error( "Problems on file From -> To transfer." );
			throw new IOException( e.getMessage() + " (" + file.getPath() + ")." );
		} finally {
			if ( fis != null ) {
				fis.close();
			}
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#createPlaypen(int)
	 */
	public String createPlaypen( int size ) throws RemoteException, UnableToCreatePlaypenException {

		checkIsUsedByOwner();

		long FolderPathNumber = (long) (Math.random() * Long.MAX_VALUE);

		File newPath = new File( Configuration.getInstance().getProperty( UserAgentConfiguration.PROP_PLAYPEN_ROOT )
				+ File.separator + "mygrid-" + FolderPathNumber );

		String result;

		if ( newPath.mkdirs() ) {
			result = newPath.getPath();
		} else {
			throw new UnableToCreatePlaypenException();
		}

		createdPlaypens.add( result );

		LOG.debug( "Playpen " + result + " has been created" );

		return result;
	}


	private synchronized boolean isReplicaStarted() {

		return replicaStarted;
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#createStorageDirectory()
	 */
	public String createStorageDirectory() throws RemoteException {

		checkIsUsedByOwner();

		String storageDir = Configuration.getInstance().getProperty( UserAgentConfiguration.PROP_STORAGE_DIR );

		if ( !JavaFileUtil.isAbsolutePath( storageDir ) )
			storageDir = System.getProperty( "user.home" ) + File.separator + storageDir;

		File storage = new File( storageDir );

		boolean createdStorageSuccessfully = false;

		if ( storage.exists() )
			createdStorageSuccessfully = true;
		else
			createdStorageSuccessfully = storage.mkdirs();

		if ( !createdStorageSuccessfully )
			throw new RemoteException( "Could not create the Storage Directory: " + storageDir );

		this.storagePath = storageDir;

		return this.storagePath;
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getStorageDirectory()
	 */
	public String getStorageDirectory() throws RemoteException {

		checkIsUsedByOwner();

		if ( this.storagePath.trim().equals( "" ) )
			return createStorageDirectory();

		return this.storagePath;
	}


	protected void destroyPlaypen( String playpenPath ) throws IOException {

		File file = new File( playpenPath );
		if ( file.exists() && file.isDirectory() ) {
			deleteFilesInDir( file );
			file.delete();
			LOG.debug( "Playpen " + playpenPath + " has been destroyed" );
		} else {
			LOG.error( playpenPath + " is not a directory or does not exists." );
			throw new IOException( "'" + playpenPath + "' is not a directory or does not exists." );
		}

		createdPlaypens.remove( playpenPath );

	}


	/**
	 * Check if the file denoted by file abstraction exists
	 * 
	 * @param file The file to have the existence checked
	 * @return <code>true</code> if the file exists and <code>false</code>
	 *         otherwise
	 * @throws RemoteException if any communication problem occurs
	 */
	public boolean fileExists( File file ) throws RemoteException {

		checkIsUsedByOwner();

		String fileToCheck;
		fileToCheck = file.getAbsolutePath();
		return this.fileExists( fileToCheck );
	}


	/**
	 * Check if the file named by parameter exists
	 * 
	 * @param filename The name of the file to have the existence checked
	 * @return <code>true</code> if the file exists and <code>false</code>
	 *         otherwise
	 * @throws RemoteException if any communication problem occurs
	 */
	private boolean fileExists( String filename ) throws RemoteException {

		checkIsUsedByOwner();

		File file = new File( filename );
		LOG.debug( "Does " + filename + " exist ? " + file.exists() );
		return (file.exists());
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#changeFilePermissions(File, int)
	 */
	public void changeFilePermissions( File file, int permissions ) throws RemoteException, IOException {

		checkIsUsedByOwner();

		/** Get an instance of a native Executor class */
		Executor exec = ExecutorFactory.getExecutorInstance();

		try {
			exec.chmod( file, Integer.toString( permissions ) );
		} catch ( ExecutorException e ) {
			throw new RemoteException( e.getMessage(), e );
		}

	}


	protected void kill( ExecutorHandle handle ) throws ExecutorException {

		LOG.debug( "Killing handle " + handle );

		Executor exec = ExecutorFactory.getExecutorInstance();

		exec.kill( handle );
		this.ownHandles.remove( handle );
		LOG.debug( "Handle " + handle + " killed" );
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#remoteExecute(String, String, Map)
	 */
	public ExecutorHandle remoteExecute( String remoteDir, String command, Map<String,String> envVars )
		throws RemoteException, ExecutorException {

		checkIsUsedByOwner();

		Executor exec = ExecutorFactory.getExecutorInstance();
		ExecutorHandle handle = exec.execute( remoteDir, command, envVars );

		LOG.info( "Command '" + command + "' has dispatched to execution. Handle: " + handle );

		this.ownHandles.add( handle );

		return handle;

	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getResult(ExecutorHandle)
	 */
	public ExecutorResult getResult( ExecutorHandle handle ) throws RemoteException, ExecutorException {

		checkIsUsedByOwner();

		try {

			Executor exec = ExecutorFactory.getExecutorInstance();
			ExecutorResult result = exec.getResult( handle );

			if ( !this.ownHandles.remove( handle ) ) {
				throw new ExecutorException( "Could not get execution result for handle " + handle
						+ ". Maybe the process has been killed." );
			}
			return result;

		} catch ( ExecutorException e ) {
			checkIsUsedByOwner();
			throw new ExecutorException( "Could not get execution result, maybe the process has been killed." );
		}

	}


	public GumID getID() {

		return spec.getGumID();
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getStatus()
	 */
	public synchronized GumStatus getStatus() {

		return gumState;
	}


	protected synchronized void setState( GumStatus state ) {

		gumState = state;
	}


	/**
	 * Called when the machine is detected do be NOT idle. This turn ON the FLAG
	 * that tells that any remote method might return a RemoteException.
	 * 
	 * @throws UserAgentPauseException If any problem happens at the kill
	 *         handles time.
	 */
	protected void pause() throws UserAgentPauseException {

		// set this UA's status to "notUsable"
		this.setState( GumStatus.OWNER );

		// kill the processes and destroy playpens
		try {
			this.stopReplica();
		} catch ( RemoteException e ) {
			throw new UserAgentPauseException( e );
		}
		LOG.info( "UA was paused!" );
	}


	/**
	 * Called when the machine is detected do be idle. This turn OFF the FLAG
	 * that tells that any remote method might return a RemoteException. So,
	 * they (the remote methods) will execute correctly.
	 */
	protected void restart() {

		this.setState( GumStatus.UP );

		LOG.info( "UA was re-started!" );
	}

	protected class UAShutdownThread extends Thread {

		@Override
		public void run() {

			try {
				stopReplica();
			} catch ( RemoteException e ) {
			}
		}
	}


	/**
	 * Removes all the files in a specified directory.
	 * 
	 * @param dir The directory from which all the files will be removed.
	 * @throws IOException if any I/O problem happens at the deletion process.
	 */
	private void deleteFilesInDir( File dir ) throws IOException {

		File[ ] files = dir.listFiles();
		if ( files != null ) {
			for ( int i = 0; i < files.length; i++ ) {
				File file = files[i];
				if ( file.isDirectory() ) {
					deleteFilesInDir( file );
				}
				file.delete();
			}
		}
	}


	/**
	 * Prepare a job to be executed. This is done by setting the boolean that
	 * control if a job is started to true.
	 * 
	 * @see org.ourgrid.common.gum.Gum#startReplica()
	 */
	public void startReplica() throws RemoteException {

		LOG.debug( "Starting replica" );
		setReplicaStarted();
	}


	private synchronized void setReplicaStarted() {

		replicaStarted = true;
	}


	/**
	 * Stops a job execution. It clones the collections with playpens and own
	 * handles to iterate to destroy and kill any execution that is active by
	 * the user. Its important to see that the original collections will be
	 * updated at destroyPlaypen and kill methods.
	 * 
	 * @see org.ourgrid.common.gum.Gum#stopReplica()
	 */
	public void stopReplica() throws RemoteException {

		LOG.debug( "Stop replica request received, will kill any pending executions" );

		setReplicaStopped();

		Set<String> createdPlaypensClone;
		synchronized ( createdPlaypens ) {
			createdPlaypensClone = new HashSet<String>( createdPlaypens );
		}

		List<ExecutorHandle> ownHandlesClone;
		synchronized ( ownHandles ) {
			ownHandlesClone = new ArrayList<ExecutorHandle>( ownHandles );
		}

		for ( ExecutorHandle executorHandle : ownHandlesClone ) {
			try {
				kill( executorHandle );
			} catch ( Exception e ) {
				LOG.error( "Could not kill process with Handle " + executorHandle, e );
			}
		}

		for ( String directory : createdPlaypensClone ) {
			try {
				destroyPlaypen( directory );
			} catch ( Exception e ) {
				LOG.error( "Could not destroy this playpen: " + directory, e );
			}
		}
	}


	private synchronized void setReplicaStopped() {

		replicaStarted = false;
	}
}
