/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.gridmachine.useragent;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.rmi.AlreadyBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.ExportException;
import java.util.HashMap;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.executor.ExecutorException;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.ExecutorResult;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.util.JavaFileUtil;
import org.ourgrid.common.util.TempFileManager;
import org.ourgrid.test.GumPreCommitTest;
import org.ourgrid.test.TestUtils;

public class UserAgentServerImplTest extends TestCase implements GumPreCommitTest {

	/** User Agent Object */
	UserAgentServerImpl ua;

	/** User Agent Object will be created with a port number */
	UserAgentServerImpl uaPort;

	/** User Agent Object will be created with an ObjID and a port number */
	UserAgentServerImpl uaIdAndPort;

	/** The data array used to test the put file method */
	byte[ ] arrayOfData;

	/** The temporary file used to test the putFile() method */
	File file;

	Registry registry;

	private TestUtils testUtils;


	/**
	 * This method constructs a new instance of this TestCase.
	 * 
	 * @param name The name of the TestCase class.
	 */
	public UserAgentServerImplTest( String name ) {

		super( name );
	}


	/*
	 * Initialization of the test
	 * 
	 * @see junit.framework.TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		testUtils = new TestUtils();
		testUtils.load();

		Configuration.getInstance( Configuration.USERAGENT );

		int port = Integer.parseInt( testUtils.getProperty( TestUtils.RMI_PORT ) );

		Configuration.getInstance().setProperty( Configuration.PROP_PORT, Integer.toString( port ) );

		File mypids = new File( ".mypids" );
		if ( mypids.exists() ) {
			mypids.delete();
		}

		try {
			registry = LocateRegistry.createRegistry( port );
		} catch ( ExportException e ) {
			registry = LocateRegistry.getRegistry( port );
		}

		try {

			ua = new UserAgentServerImpl();

		} catch ( RemoteException re ) {

			re.printStackTrace();

			fail();
		}

		/** Set an array of bytes with null values */
		arrayOfData = new byte[ 13 ];

		/** Init the array of bytes */
		for ( int i = 0; i < 13; i++ ) {
			arrayOfData[i] = 1;
		}

		/** The temporary file used to test the putFile() method */
		file = new File( "/tmp/UserServerImplTest.testPutFile" );
		file.deleteOnExit();

		/** Abstract representaion for the storage */
		File storage = new File( System.getProperty( "user.home" ) + File.separator + ".mgstorage" );
		storage.delete();
	}


	/*
	 * Finishes the test
	 * 
	 * @see junit.framework.TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		Configuration.reset();
		testUtils.reset();
	}


	/**
	 * This test verifies if the remote file length determination works on
	 * several situations: - File has size Zero. - File is so "big"
	 */
	public void testGetRemoteFileLength() {

		File tempFile = new File( "/tmp/zeroUATest" );

		/** Try to create a temporary file */
		try {
			tempFile.createNewFile();

			/** Check the size returned by a getRemoteFileLength() */
			assertEquals( ua.getRemoteFileLength( tempFile ), 0 );

			tempFile.delete();
		} catch ( IOException e ) {
			e.printStackTrace();
			fail();
		}
	}


	/** Test the object creation and its attatchment on a port. */
	public void testUserAgentAttatchment() throws Exception {

		UserAgentServerImpl attatchedUA = null;
		Gum referenceToUA = null;

		try {
			attatchedUA = new UserAgentServerImpl();
		} catch ( ExportException e ) {
			e.printStackTrace();
			fail( "Object could not be attatched!" );
		} catch ( RemoteException e ) {
			e.printStackTrace();
			fail();
		}

		/* Register the UA object */
		registry.bind( "UserAgent", attatchedUA );

		try {
			referenceToUA = (Gum) registry.lookup( "UserAgent" );
		} catch ( RemoteException e ) {
			e.printStackTrace();
			fail( "Could not bind to the GridMachine Object!" );
		}

		try {
			referenceToUA.getStatus();
		} catch ( RemoteException e1 ) {
			e1.printStackTrace();
			fail( "Could not contact the object!" );
		}

		registry.unbind( "UserAgent" );
	}


	/**
	 * This method tests the Put File Method by trying the extreme situations!
	 */
	public void testPutNullFile() throws Exception {

		File nullFile = new File( "/tmp/null" );

		try {
			this.ua.putFile( this.ua.getStorageDirectory() + File.separator + nullFile.getName(), null, true );
			fail( "This should throw an IOException!" );
		} catch ( IOException e ) {
			// Expected Exception
		}

		// assertFalse("The file [" + nullFile.getName() + "] should not exist
		// at '.' .", nullFile.exists());
	}


	/**
	 * The the putFile() method when is intended to create a file on
	 * unauthorized area
	 */
	public void testPutFileOnUnauthorizedArea() {

		try {
			this.ua.putFile( "/root/" + file.getName(), arrayOfData, true );
		} catch ( RemoteException e ) {
			e.printStackTrace();
			fail();
		} catch ( IOException e ) {
			System.out.println( e.getMessage() + ". IT WAS EXPECTED!" );
		}

		/** The temporary file used to test the putFile() method */
		File unauthorizedFile = new File( "/root/UserServerImplTest.testPutFile" );

		assertTrue( !unauthorizedFile.exists() );
	}


	/**
	 * Test the putFile() method in a common situation. Store a non-null array
	 * of bytes in a file located at an authorized place. The idea is verify if
	 * FIRST-SLICE flag is working well. Store a 13-bytes-length array twice
	 * (firstly FIRST SLICE flag is TRUE, secondly the FIRST SLICE flag is
	 * FALSE) and check the file length. The expected result is the 26-bytes
	 * file length.
	 */
	public void testPutFileCommon() throws Exception {

		try {
			this.ua.putFile( file.getAbsolutePath(), arrayOfData, true );
			this.ua.putFile( file.getAbsolutePath(), arrayOfData, false );
		} catch ( RemoteException e ) {
			e.printStackTrace();
			fail();
		} catch ( IOException e ) {
			e.printStackTrace();
			fail();
		}

		assertTrue( file.length() == 26 );
	}


	/**
	 * Test the put file method using an UserAgent created by default
	 * constructor!
	 */
	public void testPutFileDefaultConstructor() throws Exception {

		/** Index */
		int sliceIndex;

		/** Contructs a new UserAgent */
		Gum defaultUA = new UserAgentServerImpl();

		/** First action try to ping the new UA! */
		defaultUA.getStatus();

		/** put a new file there! */
		defaultUA.putFile( "/tmp/" + file.getName(), this.arrayOfData, true );

		for ( sliceIndex = 0; sliceIndex < 10; sliceIndex++ ) {
			defaultUA.putFile( "/tmp/" + file.getName(), this.arrayOfData, false );
		}

		File remoteFile = new File( "/tmp" + File.separator + file.getName() );

		/** Check the number of bytes transfered to the remote file */
		assertEquals( "The file sizes must be equals", 13 * 11, remoteFile.length() );

		/** Delete file */
		remoteFile.delete();
	}


	/**
	 * Test the get file method using an UserAgent created by default
	 * constructor!
	 */
	public void testGetFile() throws Exception {

		/** Received bytes */
		byte[ ] receivedBytes;

		/** Index */
		int sliceIndex;

		/** The received files */
		FileOutputStream outputFile = null;

		/** File created from bytes */
		File receivedFile;

		/** First action try to ping the new UA! */
		this.ua.getStatus();

		try {

			File remoteFile = new File( "/tmp" + File.separator + file.getName() );

			/** Just creates!!! */
			this.createNewFile( remoteFile );

			receivedFile = new File( "/tmp/ReceivedFile" );
			outputFile = new FileOutputStream( receivedFile );

			sliceIndex = 0;

			do {

				/** Get the transferred file */
				receivedBytes = this.ua.getFile( "/tmp" + File.separator + file.getName(), sliceIndex * 13, 13 );

				sliceIndex++;

				/** Write to the file */
				outputFile.write( receivedBytes );

			} while ( receivedBytes.length == 13 );

			assertEquals( "Received and remote file sizes must be equals", remoteFile.length(), receivedFile.length() );

			/** Close the output stream */
			outputFile.close();

			/** Delete the files */
			receivedFile.delete();

		} catch ( IOException e ) {
			throw e;
		} finally {
			try {
				outputFile.close();
			} catch ( IOException e1 ) {
				// Nothing to be done
			}
		}
	}


	/**
	 * Test the get file method using an UserAgent created by default
	 * constructor!
	 */
	public void testGetFileDoesNotExist() throws Exception {

		/** Index */
		int sliceIndex = 0;

		try {
			/** Get the transferred file */
			this.ua.getFile( "/tmp/FOO/BAR", sliceIndex * 13, 13 );
		} catch ( IOException e ) {
			System.out.println( "EXPECTED EXCEPTION: " + e.getClass() + ":: File '/tmp/FOO/BAR' does not exist." );
			assertFalse( "File '/tmp/FOO/BAR' should not exists.", (new File( "/tmp/FOO/BAR" )).exists() );
		}

	}


	/**
	 * Try to create two User Agents using the same port number and perform some
	 * operations using one of the UAs created.
	 */
	public void testCreateStorageTwoUAOnePort() throws Exception {

		Gum userAgent1 = null;
		Gum userAgent2 = null;

		int port = Integer.parseInt( Configuration.getInstance( Configuration.MYGRID ).getProperty(
				Configuration.PROP_PORT ) );

		try {
			userAgent1 = new UserAgentServerImpl();
			userAgent2 = new UserAgentServerImpl();

			registry.bind( "UserAgent", userAgent1 );
			registry.bind( "UserAgent", userAgent2 );

		} catch ( AlreadyBoundException e ) {
			System.out.println( "Expected Exception: " + e.getMessage() );
		}

		/* Just ping */
		userAgent1.getStatus();

		/** Try to get the stub for a published and exported UA */
		Gum uaByStub = (Gum) registry.lookup( "UserAgent" );

		uaByStub.getStatus(); // This is just to throws an exception if the
		// object is not up
		System.out.println( "Ping UA stub! Ok! " );

		/** Abstract representaion for the storage */
		File mgStorage = new File( System.getProperty( "user.home" ) + File.separator + ".mgstorage" );

		/** Stoarage must not exists */
		if ( !mgStorage.exists() ) {

			/** Creates the storage */
			mgStorage = new File( uaByStub.createStorageDirectory() );

			assertTrue( "The storage should be created", mgStorage.exists() );
		}

		mgStorage.delete();

		/* Unregister the object */
		registry.unbind( "UserAgent" );
	}


	/**
	 * Test the createPlaypen() method by a UA constructed using the default
	 * constructor
	 * 
	 * @throws Exception If some error occurs during the test.
	 */
	public void testCreatePlaypen() throws Exception {

		ua.startReplica();

		/** Just ping to know if the object has been exported */
		this.ua.getStatus();

		/** Create a storage directory */
		File playpen = new File( this.ua.createPlaypen( 0 ) );

		assertTrue( "Playpen should exist!", playpen.exists() );

		ua.stopReplica();

	}


	public void testChangeFilePermissions() throws Exception {

		/** A temp file used during the test */
		File regularFile = TempFileManager.createTempFile();
		regularFile.deleteOnExit();

		/** Just to test the object */
		this.ua.getStatus();

		/* parse the result */
		String permissions = JavaFileUtil.getFilePermissions( regularFile );

		System.out.println( "[" + getClass().getName() + ".testChangeFilePermissions()] Original file permissions: "
				+ permissions );

		/** Change the file permissions to rwxrwxrwx */
		this.ua.changeFilePermissions( regularFile, 777 );

		permissions = JavaFileUtil.getFilePermissions( regularFile );

		assertTrue( "The permissions should be 777", permissions.equals( "rwxrwxrwx" ) );

		try {
			/** Try to change the file permissions to an invalid number */
			this.ua.changeFilePermissions( regularFile, -1 );
		} catch ( RemoteException e ) {
			System.out.println( "EXPECTED EXCEPTION: Invalid Permissions <-1> :: " + e.getClass() );
			assertTrue( "The permissions should be 777", permissions.equals( "rwxrwxrwx" ) );
		}

		/** Change the file permissions to 000 */
		this.ua.changeFilePermissions( regularFile, 0 );

		permissions = JavaFileUtil.getFilePermissions( regularFile );

		assertTrue( "The permissions should be 000", permissions.equals( "---------" ) );

		try {
			/** Change the file permissions to another invalid number */
			this.ua.changeFilePermissions( regularFile, 999 );
		} catch ( RemoteException e ) {
			System.out.println( "EXPECTED EXCEPTION: Invalid Permissions <999> :: " + e.getClass() );
			assertTrue( "The permissions should be 000", permissions.equals( "---------" ) );
		}
	}


	/**
	 * Try to destroy a playpen in several situations!
	 * 
	 * @throws Exception if some unexpected error occurs.
	 */
	public void testDestroyPlaypen() throws Exception {

		this.ua.getStatus();

		this.ua.startReplica();

		File playpen = new File( this.ua.createPlaypen( 0 ) );

		assertTrue( "Playpen [" + playpen.getAbsolutePath() + "] should exists.", playpen.exists() );

		this.ua.destroyPlaypen( playpen.getAbsolutePath() );

		assertFalse( "Playpen [" + playpen.getAbsolutePath() + "] should not exist anymore.", playpen.exists() );

		File otherPlaypen = new File( "/xyz/foo/bar" );

		try {
			this.ua.destroyPlaypen( otherPlaypen.getAbsolutePath() );
		} catch ( IOException e ) {
			System.out.println( "EXPECTED EXCEPTION: " + e.getClass() + " :: Dir does not exist." );
			assertFalse( "Playpen [" + otherPlaypen.getAbsolutePath() + "] should not exist.", otherPlaypen.exists() );
		}

		this.ua.stopReplica();

	}


	/**
	 * This method tests the behavior of testing the existence of a remote file
	 * using an UA.
	 * 
	 * @throws Exception if an unexpected error occurs
	 */
	public void testFileExists() throws Exception {

		this.ua.getStatus();

		this.ua.startReplica();

		File playpen = new File( this.ua.createPlaypen( 0 ) );

		assertTrue( "The file should exist.", this.ua.fileExists( new File( playpen.getAbsolutePath() ) ) );
		assertTrue( "The file should exist.", this.ua.fileExists( playpen ) );
		assertFalse( "The file should not exist.", this.ua.fileExists( new File( "/FOO/BAR" ) ) );
		assertFalse( "The file should not exist.", this.ua.fileExists( new File( "/FOO/BAR" ) ) );

		this.ua.stopReplica();

	}


	/**
	 * This method tests the checksum generation is correct
	 * 
	 * @throws Exception
	 */
	public void testGetRemoteFileDigest() throws Exception {

		this.ua.getStatus();

		/** Create a temporary file */
		File regularFile = TempFileManager.createTempFile();

		String localDigest = JavaFileUtil.getDigestRepresentation( regularFile );
		String remoteDigest = this.ua.getRemoteFileDigest( regularFile );

		assertTrue( "The checksum should be equal.", localDigest.equals( remoteDigest ) );

		regularFile.delete();

		/** Different files */
		regularFile = TempFileManager.createTempFile();

		localDigest = JavaFileUtil.getDigestRepresentation( regularFile );
		remoteDigest = this.ua.getRemoteFileDigest( new File( "/etc/passwd" ) );

		assertFalse( "The checksum should not be equal", localDigest.equals( remoteDigest ) );
	}


	/**
	 * This method test the execution of a correct command. The expected result
	 * is a correct execution of command.
	 * 
	 * @throws Exception
	 */
	public void testGoodExecute() throws Exception {

		this.ua.getStatus();

		this.ua.startReplica();

		ExecutorHandle handle = this.ua.remoteExecute( ".", "ls -la /tmp", new HashMap<String,String>() );
		ExecutorResult result = this.ua.getResult( handle );

		assertEquals( result.getExitValue(), 0 );

		this.ua.stopReplica();
	}


	/**
	 * This method test the execution of a wrong command. The expected result is
	 * an erroneous execution of command because a permission denied problem.
	 * 
	 * @throws Exception
	 */
	public void testBadExecute() throws Exception {

		ExecutorHandle handle = null;

		this.ua.getStatus();
		this.ua.startReplica();

		try {
			/** Permission denied */
			handle = this.ua.remoteExecute( "/root", "ls -la", new HashMap<String,String>() );
			this.ua.getResult( handle );
		} catch ( ExecutorException e ) {
			System.out.println( "EXPECTED EXCEPTION: " + e.getClass() + " :: " + e.getMessage() );
		}
	}


	/**
	 * Use a inner thread object to test the kill method.
	 * 
	 * @throws Exception
	 */
	public void testUaKill() throws Exception {

		this.ua.startReplica();

		ExecutorResult killResult = null;
		ExecutorHandle killHandle = null;

		killHandle = this.ua.remoteExecute( ".", "ls -Ral /usr", new HashMap<String,String>() );

		this.ua.kill( killHandle );

		try {
			killResult = this.ua.getResult( killHandle );
			fail( "An exception should be thrown." );
		} catch ( Exception e ) {
			System.out.println( "EXPECTED EXCEPTION: " + e );
		}

		assertNull( "The kill result must be null, as there is no result because the execution was killed.", killResult );

		this.ua.stopReplica();

	}


	/**
	 * Testing the start and stop replica at a gum. Those method should prepare
	 * a gum to execution and also clean the gum after any execution. The test
	 * will create two playpen and also do two executions. The sleep execution
	 * that will be hold until the stop job occurs must be killed and the call
	 * to recover the execution must fail.
	 * 
	 * @throws Exception If any exception happen.
	 */
	public void testStartAndStopReplica() throws Exception {

		this.ua.startReplica();

		String playpenDir = this.ua.createPlaypen( 0 );
		String anotherPlaypenDir = this.ua.createPlaypen( 0 );

		ExecutorHandle simpleExecutionExecutorHandle = this.ua.remoteExecute( ".", "echo UATest", null );
		ExecutorHandle sleepExecutionExecutorHandle = this.ua.remoteExecute( ".", "sleep 10000", null );

		assertTrue( ua.fileExists( new File( playpenDir ) ) );
		assertTrue( ua.fileExists( new File( anotherPlaypenDir ) ) );

		assertEquals( 0, ua.getResult( simpleExecutionExecutorHandle ).getExitValue() );

		this.ua.stopReplica();

		try {
			ua.getResult( sleepExecutionExecutorHandle ).getExitValue();
			fail( "This execution was killed." );
		} catch ( RemoteException e1 ) {
			fail( "Unexpected exception: " + e1 );
		} catch ( ExecutorException e1 ) {

		}

		assertFalse( ua.fileExists( new File( playpenDir ) ) );
		assertFalse( ua.fileExists( new File( anotherPlaypenDir ) ) );

	}


	/**
	 * An auxiliar method just to create a new file filling with some bytes.
	 * 
	 * @param newFile the filename will be created at /tmp.
	 */
	private void createNewFile( File newFile ) throws Exception {

		FileOutputStream output = null;
		try {

			byte[ ] data = new byte[ 13 ];

			output = new FileOutputStream( newFile );

			newFile.deleteOnExit();

			for ( int i = 0; i < 10; i++ ) {

				output.write( data );

			}

			output.close();

		} catch ( IOException e ) {

			throw e;

		} finally {

			try {
				if ( output != null ) {
					output.close();
				}
			} catch ( IOException e1 ) {
				// Nothing left to do.
			}

		}
	}

}
