/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.gridmachine.useragent.test;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.ExportException;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.exception.UnavailableGumException;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.util.TempFileManager;
import org.ourgrid.gridmachine.exception.FileTransferException;
import org.ourgrid.gridmachine.exception.UnableToExecuteException;
import org.ourgrid.gridmachine.useragent.UserAgentClient;
import org.ourgrid.gridmachine.useragent.UserAgentServerImpl;
import org.ourgrid.mygrid.scheduler.DataDiscovery;
import org.ourgrid.mygrid.scheduler.exception.DDInitException;
import org.ourgrid.test.GumPreCommitTest;
import org.ourgrid.test.TestUtils;
import org.ourgrid.test.UniqueTest;

/**
 * Tests just a little part of the UserAgentClient class, exactly at methods
 * where exists some information process. This little test range is because of
 * the tested class functions just like a proxy to GridMachines.
 */
public class UserAgentClientTest extends TestCase implements GumPreCommitTest, UniqueTest {

	public static final String TEST_ALL_TEST_TYPE = "UNIT";

	private static final int SIZE_OF_READ_BLOCKS = 4096;

	private UserAgentClient uaClient;

	private File fileToTransfer;

	private GumSpec gumSpec;

	private Gum uaServer;

	private int port;

	Registry registry = null;

	private TestUtils testUtils;

	private GumSpec fakeSpec;


	/**
	 * Creates a default Processor object that will be used at all the test
	 * without tests.
	 */
	public UserAgentClientTest( String name ) {

		super( name );
		Configuration.reset();
	}


	/**
	 * Set up the environment to run one test!
	 */
	protected void setUp() throws Exception {

		super.setUp();

		testUtils = new TestUtils();

		testUtils.load();
		testUtils.createMGProperties();
		testUtils.createOGProperties();

		Configuration.reset();
		Configuration.getInstance( Configuration.USERAGENT );

		try {
			DataDiscovery.initDataDiscovery( false );
		} catch ( DDInitException e ) {
			DataDiscovery.getInstance();
		}

		gumSpec = testUtils.getLocalGuMSpec( TestUtils.TYPE_UALINUX );
		gumSpec.putAttribute( GumSpec.ATT_PORT, testUtils.getProperty( TestUtils.RMI_PORT ) );

		port = Integer.parseInt( gumSpec.getAttribute( GumSpec.ATT_PORT ) );

		// Create the file to use in tests!!!
		fileToTransfer = TempFileManager.createTempFile();

		/* get port and init the UserAgent server */
		Configuration.getInstance().setProperty( Configuration.PROP_PORT, testUtils.getProperty( TestUtils.RMI_PORT ) );
		uaServer = new UserAgentServerImpl();

		try {
			registry = LocateRegistry.createRegistry( port );
		} catch ( ExportException e ) {
			System.out.println( "WARNING: There is already a REGISTRY created on this port (" + port
					+ "). Trying to get the current Registry." );
			registry = LocateRegistry.getRegistry( port );
		}

		registry.rebind( "UserAgent", uaServer );

		fakeSpec = new GumSpec();
		fakeSpec.putAttribute( GumSpec.ATT_NAME, "fakeGuM" );
		fakeSpec.putAttribute( GumSpec.ATT_PORT, "1000" );
		fakeSpec.putAttribute( GumSpec.ATT_PEERNAME, "fakePeer" );
		fakeSpec.putAttribute( GumSpec.ATT_PEERPORT, "1000" );
		fakeSpec.putAttribute( GumSpec.ATT_TYPE, GumSpec.TYPE_FAKE_GUM );

		this.uaClient = new UserAgentClient( this.uaServer, gumSpec );

	}


	/**
	 * Clean what it have to clean!
	 */
	protected void tearDown() throws Exception {

		super.tearDown();

		uaClient = null;

		fileToTransfer = null;

		/** Unbind the UA */
		registry.unbind( "UserAgent" );

		uaServer = null;

		Configuration.reset();

		testUtils.reset();
	}


	public void testUserAgerClient() throws Exception {

		this.uaClient = new UserAgentClient( uaServer, gumSpec );
		assertEquals( "Verifying constructor of UAClient", uaServer, this.uaClient.getUAServer() );

	}


	/**
	 * Tests if the putFile algorithm is fragmenting correctly the file! In this
	 * case test for an empty file.
	 * 
	 * @throws UnavailableGumException
	 * @throws FileTransferException
	 * @throws RemoteException
	 */
	public void testPutFile_Size0() throws UnavailableGumException, FileTransferException, RemoteException {

		String remoteFile = testUtils.getProperty( TestUtils.REMOTE_TEMP_DIR ) + System.currentTimeMillis();

		// Preparing the Server mock!!!
		FakeUserAgentServer fakeServer = new FakeUserAgentServer( fakeSpec );

		uaClient = new UserAgentClient( fakeServer, fakeSpec );
		uaClient.putFile( fileToTransfer.getAbsolutePath(), remoteFile );

		// Checking
		assertTrue( fakeServer.fileExists( new File( remoteFile ) ) );
	}


	/**
	 * Tests if the putFile algorithm is fragmenting correctly the file! In this
	 * case test for a file that has exactly the lenght of the block and 2 times
	 * it.
	 * 
	 * @throws IOException
	 */
	public void testPutFile_MultipleOfReadBlock_Size() throws IOException {

		String remoteFile = "tmp/Remote/file.test";

		// Preparing the Server mock!!!
		FakeUserAgentServer fakeServer = new FakeUserAgentServer( fakeSpec );

		// Overriding the existent client
		uaClient = new UserAgentClient( fakeServer, fakeSpec );

		// Preparing the file with 1 block exactly
		writeBlock();

		// Call methods
		putFile( remoteFile );

		// Checking
		assertTrue( fakeServer.getPutFileCalls() == 1 );

		// Putting 2 times the block size
		writeBlock();

		// Call methods
		putFile( remoteFile );

		// Checking
		assertTrue( fakeServer.getPutFileCalls() == 3 );

	}


	/**
	 * Tests if the stream is closed. The stream must be closed when an
	 * exception occurs. The idea is to avoid the "too many open files" message
	 * bothers the users!
	 * 
	 * @throws IOException
	 */
	public void testPutFileStreamCloseWithException() throws IOException {

		String remoteFile = "/tmp/Remote/file.test";

		// Preparing the Server mock!!!
		FakeUserAgentServer fakeServer = new FakeUserAgentServer( fakeSpec );
		fakeServer.setPutFileException( true );

		// Preparing the file with 1 block exactly
		writeBlock();

		// Call methods
		try {
			uaClient = new UserAgentClient( fakeServer, fakeSpec );
			uaClient.putFile( fileToTransfer.getAbsolutePath(), remoteFile );
		} catch ( UnavailableGumException e ) {
			e.printStackTrace();
			fail( "This exception was not expected!" );
		} catch ( FileTransferException e1 ) {
			// OK... this is an expected exception!
		}

		// Checking
		assertTrue( fakeServer.getPutFileCalls() == 1 );

		// Check if stream is closed
		try {
			assertTrue( !this.uaClient.localFileStream.getFD().valid() );
		} catch ( IOException e ) {
			e.printStackTrace();
			fail( "Could not possible to get the file descriptor." );
		}
	}


	/**
	 * Tests if the stream is closed. The stream must be closed even when no
	 * exception occurs. The idea is to avoid the "too many open files" message
	 * bothers the users!
	 * 
	 * @throws IOException
	 */
	public void testPutFileStreamCloseWithoutException() throws IOException {

		String remoteDir = "/tmp/Remote/file.test";

		// Preparing the Server mock!!!
		FakeUserAgentServer fakeServer = new FakeUserAgentServer( fakeSpec );

		uaClient = new UserAgentClient( fakeServer, fakeSpec );

		// Preparing the file with 1 block exactly
		writeBlock();

		// Call methods
		putFile( remoteDir );

		// Checking
		assertEquals( 1, fakeServer.getPutFileCalls() );

		// Check if stream is closed
		try {
			assertTrue( !this.uaClient.localFileStream.getFD().valid() );
		} catch ( IOException e ) {
			e.printStackTrace();
			fail( "Could not possible to get the file descriptor." );
		}
	}


	/**
	 * Tests the GetFile method... It is testing just about the thrown exeptions
	 * at some times like: the local dir does not exist, or has not permition to
	 * write in it...
	 * 
	 * @throws RemoteException
	 */
	public void testGetFile() throws RemoteException {

		// This is made in a way that this directory cannot already exists.
		String localDir = testUtils.getProperty( TestUtils.SYSTEM_TEMP_DIR ) + File.separator;

		// These string are not use to create any file.
		String localStorageDir = testUtils.getProperty( TestUtils.SYSTEM_TEMP_DIR ) + "/Test_UAClient_Storage/";
		String fileName = "file.test";

		/*
		 * Testing the throwing action of the exception when the directory do
		 * not exists
		 */
		try {
			uaClient = new UserAgentClient( this.uaServer, fakeSpec );
			uaClient.getFile( localStorageDir + fileName, localDir + fileName );

			fail( "The exception FileTransferException had to occur!!!" );
		} catch ( UnavailableGumException e ) {
			fail( "Exception not expected: UnavailableGMException!" );
		} catch ( FileTransferException e ) {
			assertTrue( true );
		}

		/*
		 * Testing the throwing action of the exception when the directory do
		 * not exists
		 */
		File dir = new File( localDir );
		dir.mkdirs();
		dir.setReadOnly();

		try {
			uaClient.getFile( localStorageDir + fileName, localDir + fileName );

			fail( "The exception FileTransferException had to ocurre!!!" );
		} catch ( UnavailableGumException e ) {
			fail( "Exception not expected: UnavailableGMException!" );
		} catch ( FileTransferException e ) {
			assertTrue( true );
		}

		// Clean the garbage created at test!
		dir.delete();

	}


	/**
	 * Tests if the RandomAccessFile is closed and the FileLock is not released.
	 * The file must be closed even if the exception does not occur. The idea is
	 * to avoid the "too many open files" and "Bad Descriptor File" errors hang
	 * the entire system. The similar behavior will occurs if the FileLock is
	 * not released. Therefore, CLOSE and RELEASE!!! TESTS: BID #136
	 * 
	 * @throws IOException
	 * @throws UnavailableGumException
	 * @throws FileNotFoundException
	 * @throws FileTransferException
	 * @throws UnableToExecuteException
	 */
	public void testStoreCloseWithoutException() throws IOException, UnableToExecuteException, UnavailableGumException,
		FileTransferException, FileNotFoundException {

		/* Preparing the User Aget Server mock object */
		FakeUserAgentServer fakeServer = new FakeUserAgentServer( fakeSpec );
		fakeServer.setStorageDirectory( "/tmp/Remote/" );

		// Preparing the file with 1 block exactly
		writeBlock();

		// Starting test...
		uaClient = new UserAgentClient( fakeServer, fakeSpec );

		uaClient.storeFile( fileToTransfer.getAbsolutePath(), fileToTransfer.getName() );

		/* Check if the Mock Object behavior is ok! */
		assertEquals( 1, fakeServer.getFileExistsCalls() );
		assertEquals( 1, fakeServer.getGetStorageDirCalls() );
		assertEquals( 1, fakeServer.getPutFileCalls() );

	}


	/**
	 * Tests if a put and store file occurs with success in a windows GuM.
	 * 
	 * @throws FileNotFoundException
	 * @throws UnavailableGumException
	 * @throws FileTransferException
	 * @throws FileNotFoundException
	 * @throws UnableToExecuteException
	 * @throws RemoteException
	 */
	public void testPutAndStoreFilesWithGuMTypeWindows() throws UnableToExecuteException, UnavailableGumException,
		FileTransferException, FileNotFoundException, RemoteException {

		fakeSpec = new GumSpec();

		fakeSpec.putAttribute( GumSpec.ATT_NAME, "GuM 1" );
		fakeSpec.putAttribute( GumSpec.ATT_PORT, "1000" );
		fakeSpec.putAttribute( GumSpec.ATT_PEERNAME, "fakePeer" );
		fakeSpec.putAttribute( GumSpec.ATT_PEERPORT, "1000" );
		fakeSpec.putAttribute( GumSpec.ATT_TYPE, GumSpec.TYPE_UA_WINDOWS );

		FakeUserAgentServer fakeServer = new FakeUserAgentServer( fakeSpec );
		fakeServer.setStorageDirectory( "C:\\tmp1" );

		// Starting test...
		uaClient = new UserAgentClient( fakeServer, fakeSpec );

		uaClient.storeFile( fileToTransfer.getAbsolutePath(), "testFile" );
		uaClient.storeFile( fileToTransfer.getAbsolutePath(), "test1\\testFile" );
		uaClient.storeFile( fileToTransfer.getAbsolutePath(), "test2/testFile" );
		uaClient.storeFile( fileToTransfer.getAbsolutePath(), "/test1/testFile" );
		uaClient.storeFile( fileToTransfer.getAbsolutePath(), "C:\\test2\\testFile" );

		assertTrue( fakeServer.fileNameExists( "C:\\tmp1\\testFile" ) );
		assertTrue( fakeServer.fileNameExists( "C:\\tmp1\\test1\\testFile" ) );
		assertTrue( fakeServer.fileNameExists( "C:\\tmp1\\test2\\testFile" ) );
		assertTrue( fakeServer.fileNameExists( "\\test1\\testFile" ) );
		assertTrue( fakeServer.fileNameExists( "C:\\test2\\testFile" ) );

		assertTrue( uaClient.fileExists( new File( "C:\\tmp1\\testFile" ) ) );
		assertTrue( uaClient.fileExists( new File( "C:\\tmp1\\test1\\testFile" ) ) );
		assertTrue( uaClient.fileExists( new File( "C:\\tmp1\\test2\\testFile" ) ) );
		assertTrue( uaClient.fileExists( new File( "\\test1\\testFile" ) ) );
		assertTrue( uaClient.fileExists( new File( "C:\\test2\\testFile" ) ) );

		assertTrue( fakeServer.fileExists( new File( "C:\\tmp1\\testFile" ) ) );
		assertTrue( fakeServer.fileExists( new File( "C:\\tmp1\\test1\\testFile" ) ) );
		assertTrue( fakeServer.fileExists( new File( "C:\\tmp1\\test2\\testFile" ) ) );
		assertTrue( fakeServer.fileExists( new File( "\\test1\\testFile" ) ) );
		assertTrue( fakeServer.fileExists( new File( "C:\\test2\\testFile" ) ) );

	}


	/**
	 * Tests if a put and store file occurs with success in a linux GuM.
	 * 
	 * @throws FileNotFoundException
	 * @throws UnavailableGumException
	 * @throws FileTransferException
	 * @throws FileNotFoundException
	 * @throws UnableToExecuteException
	 * @throws RemoteException
	 */
	public void testPutAndStoreFilesWithGuMTypeLinux() throws UnableToExecuteException, UnavailableGumException,
		FileTransferException, FileNotFoundException, RemoteException {

		FakeUserAgentServer fakeServer = new FakeUserAgentServer( fakeSpec );
		fakeServer.setStorageDirectory( "/tmp2" );

		// Starting test...

		uaClient = new UserAgentClient( fakeServer, fakeSpec );

		uaClient.storeFile( fileToTransfer.getAbsolutePath(), "testFile" );
		uaClient.storeFile( fileToTransfer.getAbsolutePath(), "test1/testFile" );
		uaClient.storeFile( fileToTransfer.getAbsolutePath(), "/test1/testFile" );

		assertTrue( fakeServer.fileNameExists( "/tmp2/testFile" ) );
		assertTrue( fakeServer.fileNameExists( "/tmp2/test1/testFile" ) );
		assertTrue( fakeServer.fileNameExists( "/test1/testFile" ) );

		assertTrue( fakeServer.fileExists( new File( "/tmp2/testFile" ) ) );
		assertTrue( fakeServer.fileExists( new File( "/tmp2/test1/testFile" ) ) );
		assertTrue( fakeServer.fileExists( new File( "/test1/testFile" ) ) );

	}


	/**
	 * Tests if the RandomAccessFile is not open when File does not exist.
	 * TESTS: BID #136
	 * 
	 * @throws RemoteException
	 */
	public void testStoreCloseWithException() throws RemoteException {

		File fakeFile = new File( testUtils.getProperty( TestUtils.SYSTEM_TEMP_DIR ) + "/fakefiletest.tmp" );
		fakeFile.deleteOnExit();

		try {
			uaClient = new UserAgentClient( this.uaServer, fakeSpec );
			uaClient.startReplica();
			uaClient.storeFile( fakeFile.getAbsolutePath(), fakeFile.getName() );
			fail( "This should throw an exception!" );
		} catch ( FileTransferException e ) {
			// Expected Exception
		} catch ( UnableToExecuteException e ) {
			fail( "Could not create a GridMachineClient." );
			e.printStackTrace();
		} catch ( UnavailableGumException e ) {
			fail( "Could not create a GridMachineClient." );
			e.printStackTrace();
		}

		/* Check if the FILE is NULL */
		assertTrue( "The Random File must be NULL", this.uaClient.randomFile == null );

		/* Check if the FileLock is NULL . */
		assertTrue( "The File Lock must be NULL", this.uaClient.localFileLock == null );

	}


	/* TESTS WITH REAL GridMachine and UserAgentConnector */

	/**
	 * Tests if the putFile algorithm is fragmenting correctly the file! In this
	 * case test for an empty file.
	 * 
	 * @throws UnavailableGumException
	 * @throws FileTransferException
	 */
	public void testRealPutFile_Size0() throws RemoteException, IOException, UnavailableGumException,
		FileTransferException {

		String remoteDir = testUtils.getProperty( TestUtils.REMOTE_TEMP_DIR ) + File.separator;

		uaClient = new UserAgentClient( this.uaServer, fakeSpec );
		uaClient.putFile( fileToTransfer.getAbsolutePath(), remoteDir + File.separator + fileToTransfer.getName() );

		// Checking
		assertTrue( uaServer.fileExists( fileToTransfer ) );
		OurgridUnicastRemoteObject.unexportObject( uaServer, true );
	}


	/**
	 * Tests if the putFile algorithm is fragmenting correctly the file! In this
	 * case test for a file that has exactly the lenght of the block and 2 times
	 * it.
	 * 
	 * @throws IOException
	 */
	public void testRealPutFile_MultipleOfReadBlock_Size() throws IOException {

		String remoteFile = testUtils.getProperty( TestUtils.REMOTE_TEMP_DIR ) + File.separator
				+ this.fileToTransfer.getName();

		int port = Integer.parseInt( gumSpec.getAttribute( GumSpec.ATT_PORT ) );
		Gum uaServer = new UserAgentServerImpl();

		// Preparing the file with 1 block exactly
		writeBlock();

		// Call methods
		uaClient = new UserAgentClient( this.uaServer, fakeSpec );
		putFile( remoteFile );

		// Putting 2 times the block size
		writeBlock();

		// Call methods
		putFile( remoteFile );

		// Checking
		OurgridUnicastRemoteObject.unexportObject( uaServer, true );

		// TODO Check if putFile method was called 3 times with real
		// GridMachine.

	}


	/**
	 * Tests if the stream is closed. The stream must be closed when an
	 * exception occurs. The idea is to avoid the "too many open files" message
	 * bothers the users!
	 * 
	 * @throws IOException
	 */
	public void testRealPutFileStreamCloseWithException() throws IOException {

		String remoteDir = testUtils.getProperty( TestUtils.REMOTE_TEMP_DIR ) + File.separator;

		// Preparing the file with 1 block exactly
		writeBlock();

		// Call methods
		try {
			uaClient = new UserAgentClient( this.uaServer, fakeSpec );
			uaClient.putFile( fileToTransfer.getAbsolutePath(), remoteDir + File.separator + fileToTransfer.getName() );
		} catch ( UnavailableGumException e ) {
			e.printStackTrace();
			fail( "This exception was not expected!" );
		} catch ( FileTransferException e1 ) {
			// OK... this is an expected exception!
		}

		// Check if stream is closed
		try {
			assertTrue( !this.uaClient.localFileStream.getFD().valid() );
		} catch ( IOException e ) {
			e.printStackTrace();
			fail( "Could not possible to get the file descriptor." );
		}
	}


	/**
	 * Tests if the stream is closed. The stream must be closed even when no
	 * exception occurs. The idea is to avoid the "too many open files" message
	 * bothers the users!
	 * 
	 * @throws IOException
	 */
	public void testRealPutFileStreamCloseWithoutException() throws IOException {

		String remoteFile = testUtils.getProperty( TestUtils.REMOTE_TEMP_DIR ) + File.separator
				+ this.fileToTransfer.getName();

		// Preparing the file with 1 block exactly
		writeBlock();

		// Call methods
		uaClient = new UserAgentClient( this.uaServer, fakeSpec );
		putFile( remoteFile );

		// Check if stream is closed
		try {
			assertTrue( !this.uaClient.localFileStream.getFD().valid() );
		} catch ( IOException e ) {
			e.printStackTrace();
			fail( "Could not possible to get the file descriptor." );
		}
	}


	/**
	 * Tests the GetFile method... It is testing just about the thrown exeptions
	 * at some times like: the local dir does not exist, or has not permition to
	 * write in it...
	 * 
	 * @throws RemoteException
	 */
	public void testRealGetFile() throws RemoteException {

		// This is made in a way that this directory cannot already exists.
		String localDir = testUtils.getProperty( TestUtils.FAKE_FILE ) + File.separator;

		// These string are not use to create any file.
		String localStorageDir = testUtils.getProperty( TestUtils.SYSTEM_TEMP_DIR ) + "/Test_UAClient_Storage/";
		String fileName = "file.test";

		int port = Integer.parseInt( gumSpec.getAttribute( GumSpec.ATT_PORT ) );
		Gum uaServer = new UserAgentServerImpl();

		/*
		 * Testing the throwing action of the exception when the directory do
		 * not exists
		 */
		try {
			uaClient = new UserAgentClient( this.uaServer, fakeSpec );
			uaClient.getFile( localStorageDir + fileName, localDir + fileName );

			fail( "The exception FileTransferException had to ocurre!!!" );
		} catch ( UnavailableGumException e ) {
			fail( "Exception not expected: UnavailableGMException!" );
		} catch ( FileTransferException e ) {
			assertTrue( true );
		}

		(new File( localDir + File.separator + fileName )).delete();

		/*
		 * Testing the throwing action of the exception when the directory do
		 * not exists
		 */
		File dir = new File( localDir );
		dir.mkdirs();
		dir.setReadOnly();

		try {
			uaClient.getFile( localStorageDir + fileName, localDir + File.separator + fileName );

			fail( "The exception FileTransferException had to ocurre!!!" );
		} catch ( UnavailableGumException e ) {
			fail( "Exception not expected: UnavailableGMException!" );
		} catch ( FileTransferException e ) {
			assertTrue( true );
		}

		// Clean the garbage created at test!
		(new File( localDir + File.separator + fileName )).delete();

		dir.delete();
		OurgridUnicastRemoteObject.unexportObject( uaServer, true );

	}


	/**
	 * Tests if the RandomAccessFile is closed and the FileLock is not released.
	 * The file must be closed even if the exception does not occur. The idea is
	 * to avoid the "too many open files" and "Bad Descriptor File" errors hang
	 * the entire system. The similar behavior will occurs if the FileLock is
	 * not released. Therefore, CLOSE and RELEASE!!! TESTS: BID #136
	 * 
	 * @throws IOException
	 * @throws UnavailableGumException
	 * @throws FileNotFoundException
	 * @throws FileTransferException
	 * @throws UnableToExecuteException
	 */
	public void testRealStoreCloseWithoutException() throws IOException, UnableToExecuteException,
		UnavailableGumException, FileTransferException, FileNotFoundException {

		int port = Integer.parseInt( gumSpec.getAttribute( GumSpec.ATT_PORT ) );
		Gum uaServer = new UserAgentServerImpl();

		// Preparing the file with 1 block exactly
		writeBlock();

		uaClient = new UserAgentClient( this.uaServer, fakeSpec );
		uaClient.startReplica();
		uaClient.storeFile( fileToTransfer.getAbsolutePath(), fileToTransfer.getName() );

		OurgridUnicastRemoteObject.unexportObject( uaServer, true );
	}


	/**
	 * Tests if the RandomAccessFile is not open when File does not exist.
	 * TESTS: BID #136
	 * 
	 * @throws RemoteException
	 * @throws UnavailableGumException
	 * @throws FileNotFoundException
	 * @throws FileTransferException
	 * @throws UnableToExecuteException
	 */
	public void testRealStoreCloseWithException() throws Exception {

		File fakeFile = new File( testUtils.getProperty( TestUtils.SYSTEM_TEMP_DIR ) + "fake_file.tmp" );

		int port = Integer.parseInt( gumSpec.getAttribute( GumSpec.ATT_PORT ) );
		Gum uaServer = new UserAgentServerImpl();

		try {
			uaClient = new UserAgentClient( this.uaServer, fakeSpec );
			uaClient.startReplica();
			uaClient.storeFile( fakeFile.getAbsolutePath(), fakeFile.getName() );
			fail( "This should throw an exception!" );
		} catch ( Exception e ) {
			// This exception is expected
		}

		/* Check if the FILE is NULL */
		assertTrue( "The Random File must be NULL", this.uaClient.randomFile == null );

		/* Check if the FileLock is NULL . */
		assertTrue( "The File Lock must be NULL", this.uaClient.localFileLock == null );

		OurgridUnicastRemoteObject.unexportObject( uaServer, true );
	}


	/* PRIVATE METHODS */

	/**
	 * This method writes SIZE_OF_READ_BLOCKS bytes in the file fileToTransfer.
	 * REMEMBER: The SIZE_OF_READ_BLOCKS has to be equals with
	 * UserAgentClient.sizeEachRead.
	 * 
	 * @throws IOException
	 */
	private void writeBlock() throws IOException {

		BufferedWriter writer = null;
		try {
			writer = new BufferedWriter( new FileWriter( this.fileToTransfer.getAbsoluteFile(), true ) );
			char[ ] dataArray = new char[ SIZE_OF_READ_BLOCKS ];
			for ( int count = 1; count < SIZE_OF_READ_BLOCKS; count++ ) {
				dataArray[count - 1] = 'x';
			}
			writer.write( dataArray );
			writer.close();
		} catch ( IOException e ) {
			throw e;
		} finally {
			try {
				if ( writer != null ) {
					writer.close();
				}
			} catch ( IOException e1 ) {
				// Nothing to be done
			}
		}

	}


	/**
	 * Calls the (this) UserAgentClient method "putFile()"! It is here because
	 * of the treatement of the exceptions.
	 * 
	 * @param remoteFile The remote file
	 */
	private void putFile( String remoteFile ) {

		try {

			uaClient.putFile( fileToTransfer.getAbsolutePath(), remoteFile );
		} catch ( UnavailableGumException e ) {
			e.printStackTrace();
			fail( "Exception not expected: UnavailableGMException!" );
		} catch ( FileTransferException e ) {
			e.printStackTrace();
			fail( "Exception not expected: FileTransferException!" );
		}
	}

}
