/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.gridmachine.useragent.ui;

import static org.ourgrid.gridmachine.useragent.ui.UIErrors.USERAGENT_STARTING_FAIL;
import static org.ourgrid.gridmachine.useragent.ui.UIMessages.STARTING_BIND;
import static org.ourgrid.gridmachine.useragent.ui.UIMessages.STARTING_SECURITY;
import static org.ourgrid.gridmachine.useragent.ui.UIMessages.START_SUCCESSFUL;
import static org.ourgrid.gridmachine.useragent.ui.UIMessages.USERAGENT_STARTING;

import java.io.File;
import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.LogConfiguration;
import org.ourgrid.common.config.UserAgentConfiguration;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.security.MGSecureClientSocketFactory;
import org.ourgrid.common.security.MGSecureServerSocketFactory;
import org.ourgrid.common.security.NullSecurityManager;
import org.ourgrid.common.url.UserAgentURLProvider;
import org.ourgrid.gridmachine.useragent.ShutdownManager;
import org.ourgrid.gridmachine.useragent.ShutdownManagerImpl;
import org.ourgrid.gridmachine.useragent.UserAgentServerImpl;
import org.ourgrid.gridmachine.useragent.fd.UserAgentFailureDetector;

/**
 * Main class to UserAgent initialization. Any UI call is made at this class so
 * the interface logic is done only at this point.
 */
public class UserAgentUIManager extends OurgridUnicastRemoteObject implements UIManager {

	private static final long serialVersionUID = 33L;

	/** Start argument. */
	public static final String START = "start";

	/** Stop argument. */
	public static final String STOP = "stop";

	/** Status argument. */
	public static final String STATUS = "status";

	/** Array with valid arguments. */
	public static final String[ ] VALID_ARGS = { START, STOP, STATUS };

	private static UserAgentFailureDetector failureDetector;

	private static UIManager uniqueInstance;

	/** Logger object to store log events */
	static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( UserAgentUIManager.class );

	private Configuration config;

	// private int port;

	private Gum userAgent;

	private ShutdownManager shutdownManager;


	/**
	 * Private constructor to the Singleton.
	 * 
	 * @throws RemoteException If its not possible create this remote object.
	 */
	private UserAgentUIManager() throws RemoteException {

		super();
		/*
		 * This line creates and installs a Security Manager. A Security Manager
		 * is resposible for allowing or preventing access to all system
		 * resources. It is the primary interface between the CORE Java API and
		 * the operating System. In order to allow the UserAgent access several
		 * key-resources (e.g. Network) it is needed to install a security
		 * manager that grant this access <code>NullSecurotyManager</code>.
		 */
		System.setSecurityManager( new NullSecurityManager() );
		this.config = Configuration.getInstance( Configuration.USERAGENT );
	}


	public static UIManager getInstance() throws RemoteException {

		if ( uniqueInstance == null ) {
			Configuration config = Configuration.getInstance( Configuration.USERAGENT );

			prepareLog();

			try {
				// tries to get a remote instance already bound by another
				// invocation of this method.
				uniqueInstance = (UIManager) Naming.lookup( UserAgentURLProvider.userAgentUIManager() );
			} catch ( MalformedURLException e ) {
			} catch ( NotBoundException e ) {
			} catch ( RemoteException e ) {
			}

			if ( uniqueInstance == null ) {
				uniqueInstance = new UserAgentUIManager();

				try {
					Naming.rebind( UserAgentURLProvider.userAgentUIManager(), uniqueInstance );
				} catch ( MalformedURLException e ) {
					LOG.debug( "Error trying to rebind at URL " + UserAgentURLProvider.userAgentUIManager() );
				}
			}
		}
		return uniqueInstance;
	}


	/**
	 * Prepares a log to be used. When configuring a log, it will create the
	 * necessary directories and also check for a log properties file to be used
	 * to make this configuration.
	 */
	private static void prepareLog() {

		String logFile = Configuration.getInstance().getLogPath();

		String parents = new File( logFile ).getParent();

		if ( parents != null )
			(new File( parents )).mkdirs();

		LogConfiguration.configure( Configuration.getInstance().getLogPropertiesPath(), logFile );
	}


	private void startFailureDetector( Gum userAgent ) {

		try {
			failureDetector = new UserAgentFailureDetector( UserAgentConfiguration.QUERY_FREQ, userAgent );
			Naming.rebind( UserAgentURLProvider.userAgentFailureDetector(), failureDetector );
		} catch ( RemoteException e ) {
			LOG.error( "Cannot bind Peer Failure Detector Object.", e );
		} catch ( MalformedURLException e ) {
			LOG.error( "Cannot bind Peer Failure Detector Object.", e );
		}
	}


	/**
	 * Initialize the <code>ShutDownManagerImpl</code> to be used for shutdown
	 * the User Agent.
	 * 
	 * @param port Port to publish this object.
	 * @return The ShutDownManagerImpl object.
	 */
	private ShutdownManager initShutDownManagerImpl() {

		ShutdownManager shutdownManager = null;
		try {
			shutdownManager = new ShutdownManagerImpl( userAgent );
		} catch ( RemoteException e ) {

			System.out.println( USERAGENT_STARTING_FAIL.getMessage() );
			LOG.error( USERAGENT_STARTING_FAIL.getMessage(), e );
			System.exit( USERAGENT_STARTING_FAIL.getExitCode() );
		}
		return shutdownManager;
	}


	/**
	 * Bind remote UserAgent server objects. This includes the UA itself and a
	 * Shutdown Manager.
	 * 
	 * @param userAgent UserAgent to be binded.
	 * @param shutDownManager ShutDownManager to be binded.
	 * @throws MalformedURLException If the URL to bind this object is
	 *         malformed.
	 * @throws RemoteException If is not possible bind those objects.
	 */
	private static void bindRemoteObjects( Gum userAgent, ShutdownManager shutdownManager ) throws RemoteException,
		MalformedURLException {

		String userAgentURL = UserAgentURLProvider.userAgent();

		Naming.rebind( UserAgentURLProvider.shutdownManager(), shutdownManager );
		Naming.rebind( userAgentURL, userAgent );
	}


	/**
	 * Initializate the <code>UserAgentImpl</code> this also includes start a
	 * registry and prepare the configuration to be used.
	 * 
	 * @param config Configuration to be used.
	 * @param port Desired UserAgent port.
	 * @return The <code>UserAgentServerImpl</code> object created.
	 * @throws RemoteException If is not possible to bind UserAgent, a remote
	 *         exception is trhown.
	 * @throws ConfigException
	 */
	private UserAgentServerImpl initUserAgentImpl() throws RemoteException, ConfigException {

		UserAgentServerImpl userAgent = null;

		if ( config.isSecurityEnabled() ) {

			LOG.info( STARTING_SECURITY.getMessage() );

			int securePort = Integer
					.parseInt( Configuration.getInstance().getProperty( Configuration.PROP_SECURE_PORT ) );

			String keyStoreFilename = config.getProperty( Configuration.PROP_KEYSTORE_FILENAME );
			String keyStorePassword = config.getProperty( Configuration.PROP_KEYSTORE_PASSWORD );

			System.setProperty( "javax.net.ssl.keyStore", keyStoreFilename );
			System.setProperty( "javax.net.ssl.keyStorePassword", keyStorePassword );

			String trustStoreFilename = config.getProperty( Configuration.PROP_TRUSTSTORE_FILENAME );
			String trustStorePassword = config.getProperty( Configuration.PROP_TRUSTSTORE_PASSWORD );

			System.setProperty( "javax.net.ssl.trustStore", trustStoreFilename );
			System.setProperty( "javax.net.ssl.trustStorePassword", trustStorePassword );

			/**
			 * Create a Secure UA based on the SSL Comunnication through the MG
			 * implementation of socket factories
			 */
			userAgent = new UserAgentServerImpl( securePort, MGSecureClientSocketFactory.getInstance(),
					MGSecureServerSocketFactory.getInstance() );

		} else {
			LOG.info( STARTING_BIND.getMessage() );
			userAgent = new UserAgentServerImpl();
		}
		return userAgent;
	}


	/**
	 * Verify if UserAgent is running.
	 */
	public boolean isUserAgentUp() {

		return (userAgent != null && shutdownManager != null);
	}


	/**
	 * @throws ConfigException
	 * @see UIManager#startUserAgentService()
	 */
	public void startUserAgentService() throws RemoteException, MalformedURLException, InterruptedException,
		UserAgentUIException, ConfigException {

		LOG.info( USERAGENT_STARTING.getMessage() );

		if ( isUserAgentUp() ) {
			throw new UserAgentUIException( "User agent already started." );
		}

		userAgent = initUserAgentImpl();
		shutdownManager = initShutDownManagerImpl();
		bindRemoteObjects( userAgent, shutdownManager );

		startFailureDetector( userAgent );

		LOG.info( START_SUCCESSFUL.getMessage() );

		/**
		 * The current Thread, in this case Thread.main, must be maintained
		 * alive because the UserAgent object must be listening to the port
		 * continuously.
		 */
		while ( true ) {
			Thread.sleep( Long.MAX_VALUE );
		}

	}


	/**
	 * @see UIManager#stopUserAgentService()
	 */
	public void stopUserAgentService() throws RemoteException, UserAgentUIException, MalformedURLException,
		NotBoundException {

		ShutdownManager shutdownManager;

		LOG.info( "Stopping user agent." );

		if ( isUserAgentUp() ) {
			shutdownManager = (ShutdownManager) Naming.lookup( UserAgentURLProvider.shutdownManager() );
			failureDetector.shutdown();
			shutdownManager.shutdown();
		} else {
			throw new UserAgentUIException( "UserAgent already down." );
		}

	}


	/**
	 * @see UIManager#statusUserAgentService()
	 */
	public GumStatus statusUserAgentService() throws RemoteException {

		if ( userAgent == null )
			return GumStatus.CONTACTING;
		return userAgent.getStatus();
	}

}
