/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.replicaexecutor;

import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.event.ShutdownRequestEvent;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.mygrid.replicaexecutor.event.CancelReplicasOfJobRequestEvent;
import org.ourgrid.mygrid.replicaexecutor.event.ExecuteReplicaRequestEvent;
import org.ourgrid.mygrid.replicaexecutor.event.PermissionRequestEvent;
import org.ourgrid.mygrid.replicaexecutor.event.ReplicaAbortedNotiftyEvent;
import org.ourgrid.mygrid.replicaexecutor.event.ReplicaCanceledNotifyEvent;
import org.ourgrid.mygrid.replicaexecutor.event.ReplicaFailedNotifyEvent;
import org.ourgrid.mygrid.replicaexecutor.event.ReplicaFinishedNotifyEvent;
import org.ourgrid.mygrid.scheduler.EBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;

/**
 * Description: This class is responsible for queueing events in the
 * <code>ReplicaExecutorEventEngine</code>. All communication with objects
 * inside the <code>ReplicaExecutorEventEngine</code> is made through this
 * facade.
 */
public class EBReplicaExecutorFacade {

	/**
	 * The <code>ReplicaExecutorEventEngine</code>
	 */
	private ReplicaExecutorEventEngine replicaExecutorEventProcessor;


	/**
	 * Creates a <code>EBReplicaExecutorFacade</code>
	 */
	public EBReplicaExecutorFacade() {

		replicaExecutorEventProcessor = null;

	}


	/**
	 * Executes a given <code>Replica</code> with a specified
	 * <code>GridMachine</code>
	 * 
	 * @param replica The replica to be executed
	 * @param gumClient The grid machine where the replica should be executed
	 */
	public void executeReplica( ReplicaEntry replica, GumClient gumClient ) {

		ExecuteReplicaRequestEvent executeReplicaRequestEvent = new ExecuteReplicaRequestEvent(
				replicaExecutorEventProcessor.getEBReplicaManager(), replica, gumClient );
		replicaExecutorEventProcessor.putEvent( executeReplicaRequestEvent );

	}


	/**
	 * Cancel all replicas of a given <code>Job</code>
	 * 
	 * @param jobId The id of the job whose replicas should be canceled.
	 */
	public void cancelReplicasOfJob( int jobId ) {

		CancelReplicasOfJobRequestEvent cancelReplicasOfJobRequestEvent = new CancelReplicasOfJobRequestEvent(
				replicaExecutorEventProcessor.getEBReplicaManager(), jobId );
		replicaExecutorEventProcessor.putEvent( cancelReplicasOfJobRequestEvent );

	}


	/**
	 * Notifies the replica executor module that a given replica has failed
	 * 
	 * @param replica The replica who has failed.
	 */
	protected void replicaFailed( ReplicaEntry replica ) {

		ReplicaFailedNotifyEvent replicaFailedRequestEvent = new ReplicaFailedNotifyEvent(
				replicaExecutorEventProcessor.getEBReplicaManager(), replica );
		replicaExecutorEventProcessor.putEvent( replicaFailedRequestEvent );
	}


	/**
	 * Notifies the replica executor module that a given replica has finished.
	 * 
	 * @param replica The replica who has finished.
	 */
	protected void replicaFinished( ReplicaEntry replica ) {

		ReplicaFinishedNotifyEvent replicaFinishedRequestEvent = new ReplicaFinishedNotifyEvent(
				replicaExecutorEventProcessor.getEBReplicaManager(), replica );
		replicaExecutorEventProcessor.putEvent( replicaFinishedRequestEvent );
	}


	/**
	 * Notifies the replica executor module that a given replica has been
	 * aborted.
	 * 
	 * @param replica The replica who has been been aborted.
	 */
	public void replicaAborted( ReplicaEntry replica ) {

		ReplicaAbortedNotiftyEvent replicaAbortedRequestEvent = new ReplicaAbortedNotiftyEvent(
				replicaExecutorEventProcessor.getEBReplicaManager(), replica );
		replicaExecutorEventProcessor.putEvent( replicaAbortedRequestEvent );
	}


	/**
	 * Notifies the replica executor module that a given replica has been
	 * cancelled.
	 * 
	 * @param replica The replica who has been cancelled.
	 */
	public void replicaCanceled( ReplicaEntry replica ) {

		ReplicaCanceledNotifyEvent replicaCanceledRequestEvent = new ReplicaCanceledNotifyEvent(
				replicaExecutorEventProcessor.getEBReplicaManager(), replica );
		replicaExecutorEventProcessor.putEvent( replicaCanceledRequestEvent );
	}


	/**
	 * Request permission to access the task's critical region.
	 * 
	 * @param replica The replica who wants to access it's task's critical
	 *        region.
	 * @return An EventQueue that will contain the result of the permission
	 *         request.
	 */
	protected EventQueue<ResponseEvent<Boolean>> requestPermission( ReplicaEntry replica ) {

		EventQueue<ResponseEvent<Boolean>> responseQueue = new EventQueue<ResponseEvent<Boolean>>();

		PermissionRequestEvent permissionRequestEvent = new PermissionRequestEvent( replicaExecutorEventProcessor
				.getEBReplicaManager(), replica, responseQueue );
		replicaExecutorEventProcessor.putEvent( permissionRequestEvent );

		return responseQueue;

	}


	/**
	 * Configures the EBReplicaExecutorFacade by setting the
	 * <code>EBSchedulerFacade</code>.
	 * 
	 * @param ebSchedulerFacade The EBSchedulerFacade
	 */
	public void config( EBSchedulerFacade ebSchedulerFacade ) {

		replicaExecutorEventProcessor = new ReplicaExecutorEventEngine( this, ebSchedulerFacade );

	}


	/**
	 * Configures the EBReplicaExecutorFacade by setting the
	 * <code>ReplicaExecutorEventEngine</code>.
	 * 
	 * @param replicaExecutorEventProcessor The
	 *        <code>ReplicaExecutorEventEngine</code>
	 */
	public void config( ReplicaExecutorEventEngine replicaExecutorEventProcessor ) {

		this.replicaExecutorEventProcessor = replicaExecutorEventProcessor;

	}


	/**
	 * @return The <code>ReplicaExecutroEventProcessor</code>
	 */
	public ReplicaExecutorEventEngine getEventProcessor() {

		return replicaExecutorEventProcessor;

	}


	/**
	 * Starts the <code>ReplicaExecutorEventEngine</code>
	 */
	public void startEventProcessor() {

		replicaExecutorEventProcessor.startProcessing();

	}


	/**
	 * Shuts down the facade.
	 */
	public EventQueue<ShutdownResponseEvent> shutdown() {

		EventQueue<ShutdownResponseEvent> responseQueue = new EventQueue<ShutdownResponseEvent>();
		ShutdownRequestEvent shutdownRequestEvent = new ShutdownRequestEvent( responseQueue,
				replicaExecutorEventProcessor );
		replicaExecutorEventProcessor.putEvent( shutdownRequestEvent );

		return responseQueue;
	}


	/**
	 * @return Returns the number of replicas cancelled.
	 */
	protected long getNumberOfCancelledReplicas() {

		return replicaExecutorEventProcessor.getEBReplicaManager().getNumberOfCancelledReplicas();
	}


	/**
	 * @return Returns the the number of replicas failed.
	 */
	protected long getNumberOfFailedReplicas() {

		return replicaExecutorEventProcessor.getEBReplicaManager().getNumberOfFailedReplicas();
	}


	/**
	 * @return Returns the number of replicas Finished.
	 */
	protected long getNumberOfFinishedReplicas() {

		return replicaExecutorEventProcessor.getEBReplicaManager().getNumberOfFinishedReplicas();
	}


	/**
	 * @return Returns the number of replicas Aborted.
	 */
	protected long getNumberOfAbortedReplicas() {

		return replicaExecutorEventProcessor.getEBReplicaManager().getNumberOfAbortedReplicas();
	}


	/**
	 * @return Returns the number of replicas running.
	 */
	protected long getNumberOfRunningReplicas() {

		return replicaExecutorEventProcessor.getEBReplicaManager().getNumberOfRunningReplicas();
	}

}
