/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.replicaexecutor;

import java.util.Iterator;

import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.event.Shutdownable;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.mygrid.replicaexecutor.exception.ReplicaExecutorThreadException;
import org.ourgrid.mygrid.scheduler.EBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;

/**
 * Description: This is the main manager of the replica executor module. It
 * coordinates the <code>ThreadManager</code> and the
 * <code>PermissionManager</code> to do the main job of the module: provide a
 * way to execute a replica.
 * 
 * @version 1.0 Created on 08/07/2004
 */
public class EBReplicaManager implements Shutdownable {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( EBReplicaManager.class );

	/**
	 * The <code>EBSchedulerFacade</code>
	 */
	private EBSchedulerFacade ebSchedulerFacade;

	/**
	 * The <code>ThreadManager</code>
	 */
	private ReplicaExecutorThreadManager replicaExecutorThreadManager;

	/**
	 * The <code>PermissionManager</code>
	 */
	private PermissionManager permissionManager;

	/**
	 * the number of replicas failed.
	 */
	private long failedReplicas = 0;

	/**
	 * the number of replicas aborted.
	 */
	private long abortedReplicas = 0;

	/**
	 * the number of replicas cancelled.
	 */
	private long cancelledReplicas = 0;

	/**
	 * the number of replicas finished.
	 */
	private long finishedReplicas = 0;

	/**
	 * the number of replicas running.
	 */
	private long runningReplicas = 0;

	/**
	 * @version 1.0 Created on 23/07/2004
	 */
	public class KillerThread implements Runnable {

		private ReplicaExecutor aThread;


		public KillerThread( ReplicaExecutor aThread ) {

			super();
			this.aThread = aThread;
			Thread engine = new Thread( this, "Killer of " + aThread.getReplica() );
			engine.setPriority( Thread.MAX_PRIORITY );
			engine.start();
		}


		public void run() {

			aThread.abortReplica();

		}
	}

	/**
	 * A service thread used to cancel a replica.
	 * 
	 * @version 1.0 Created on 23/07/2004
	 */
	private class CancellerThread implements Runnable {

		private ReplicaExecutor aThread;


		public CancellerThread( ReplicaExecutor aThread ) {

			super();
			this.aThread = aThread;
			Thread engine = new Thread( this, "Canceller of " + aThread.getReplica() );
			engine.setPriority( Thread.MAX_PRIORITY );
			engine.start();
		}


		public void run() {

			aThread.cancelReplica();
		}
	}


	/**
	 * Creates a EBReplicaManager with a <code>EBSchedulerFacade</code>, a
	 * <code>ThreadManager</code> and a <code>PermissionManager</code>.
	 * 
	 * @param ebSchedulerFacade The <code>EBSchedulerFacade</code>
	 * @param replicaExecutorThreadManager The <code>ThreadManager</code>
	 * @param permissionManager The <code>PermissionManager</code>
	 */
	public EBReplicaManager( EBSchedulerFacade ebSchedulerFacade,
								ReplicaExecutorThreadManager replicaExecutorThreadManager,
								PermissionManager permissionManager ) {

		this.ebSchedulerFacade = ebSchedulerFacade;

		this.replicaExecutorThreadManager = replicaExecutorThreadManager;
		this.permissionManager = permissionManager;

	}


	/**
	 * Assign the specified <code>Replica</code> to be executed int the given
	 * <code>GridMachine</code>
	 * 
	 * @param replica The <code>Replica</code> to be executed
	 * @param gumClient The <code>GridMachine</code> where the replica will be
	 *        executed.
	 */
	public void executeReplica( ReplicaEntry replica, GumClient gumClient ) {

		try {

			LOG.info( "About to start " + replica.toString() );
			replicaExecutorThreadManager.obtainReplicaExecutorThread( replica, gumClient );
			runningReplicas++;

		} catch ( ReplicaExecutorThreadException e ) {

			LOG.error( "", e );

		}

	}


	/**
	 * Notifies the <code>PermissionManager</code> and the
	 * <code>ThreadManager</code> that the given replica has been cancelled.
	 */
	public void replicaCanceled( ReplicaEntry replica ) {

		ebSchedulerFacade.replicaCanceled( replica.getResult() );
		permissionManager.replicaCanceled( replica );

		try {

			LOG.debug( "Removing thread of canceled replica:" + replica.toString() );
			replicaExecutorThreadManager.releaseReplicaExecutorThread( replica );
			runningReplicas--;
			cancelledReplicas++;

		} catch ( ReplicaExecutorThreadException e ) {

			LOG.error( "", e );

		}

	}


	/**
	 * Notifies the <code>EBSchedulerFacade</code>, the
	 * <code>PermissioManager</code> and the <code>ThreadManager</code> that
	 * the given replica has failed.
	 */
	public void replicaFailed( ReplicaEntry replica ) {

		ebSchedulerFacade.replicaFailed( replica.getResult() );
		permissionManager.replicaFailed( replica );

		try {

			LOG.debug( "Removing thread of failed replica:" + replica.toString() );
			replicaExecutorThreadManager.releaseReplicaExecutorThread( replica );
			failedReplicas++;
			runningReplicas--;

		} catch ( ReplicaExecutorThreadException e ) {

			LOG.error( "", e );

		}

	}


	/**
	 * Notifies the <code>EBSchedulerFacade</code>, the
	 * <code>PermissioManager</code> and the <code>ThreadManager</code> that
	 * the given replica has finished. When a replica is Finished all the other
	 * replicas of the same task should be aborted.
	 */
	public void replicaFinished( ReplicaEntry replica ) {

		ebSchedulerFacade.replicaFinished( replica.getResult() );
		replicaExecutorThreadManager.replicaFinished( replica );

		try {

			LOG.debug( "Removing thread of finished replica:" + replica.toString() );
			replicaExecutorThreadManager.releaseReplicaExecutorThread( replica );
			runningReplicas--;
			finishedReplicas++;

			LOG.debug( "Attepting to abort the other replicas of task " + replica.getJobId() + "."
					+ replica.getTaskId() );
			abortReplicasOfTask( replica.getJobId(), replica.getTaskId() );

			permissionManager.replicaFinished( replica );

		} catch ( ReplicaExecutorThreadException e ) {

			LOG.error( "", e );

		}

	}


	/**
	 * Notifies the <code>PermissioManager</code> and the
	 * <code>ThreadManager</code> that the given replica has been aborted.
	 */
	public void replicaAborted( ReplicaEntry replica ) {

		ebSchedulerFacade.replicaAborted( replica.getResult() );
		permissionManager.replicaAborted( replica );

		try {

			LOG.debug( "Removing thread of aborted replica: " + replica.toString() );
			replicaExecutorThreadManager.releaseReplicaExecutorThread( replica );
			runningReplicas--;
			abortedReplicas++;

		} catch ( ReplicaExecutorThreadException e ) {

			LOG.error( "", e );

		}

	}


	/**
	 * Cancel all replicas of a given <code>Job</code>
	 * 
	 * @param jobId The id of the job whose replicas should be canceled.
	 */
	public void cancelReplicasOfJob( int jobId ) {

		Iterator<ReplicaExecutor> threadIterator = replicaExecutorThreadManager.getReplicaExecutorThreads().iterator();

		while ( threadIterator.hasNext() ) {

			ReplicaExecutor aThread = threadIterator.next();
			ReplicaEntry replica = aThread.getReplica();

			if ( replica.getJobId() == jobId ) {

				LOG.debug( "Canceling replica :" + replica.toString() );
				new CancellerThread( aThread );

			}

		}

	}


	/**
	 * Abort all replicas of a given <code>Task</code>
	 * 
	 * @param jobId The id of the job whose replicas should be aborted.
	 * @param taskId The id of the task whose replicas should be aborted.
	 */
	protected void abortReplicasOfTask( int jobId, int taskId ) {

		Iterator<ReplicaExecutor> threadIterator = replicaExecutorThreadManager.getReplicaExecutorThreads().iterator();

		while ( threadIterator.hasNext() ) {

			ReplicaExecutor aThread = threadIterator.next();
			ReplicaEntry replica = aThread.getReplica();

			if ( replica.getJobId() == jobId && replica.getTaskId() == taskId ) {

				LOG.debug( "Killing replica :" + replica.toString() );
				new KillerThread( aThread );

			}

		}

	}


	/**
	 * Request permission to access the task's critical region.
	 * 
	 * @param replica The replica who wants to access it's task's critical
	 *        region.
	 * @param responseQueue The replica's response queue.
	 */
	public void requestPermission( ReplicaEntry replica, EventQueue<ResponseEvent<Boolean>> responseQueue ) {

		permissionManager.requestPermission( replica, responseQueue );

	}


	/**
	 * Shuts down the replica manager
	 */
	public void shutdown() {

		LOG.debug( "The EBReplicaManager is going down" );

		replicaExecutorThreadManager.shutdown();
		permissionManager.shutdown();

	}


	/**
	 * Returns the number of replicas running.
	 * 
	 * @return the number of replicas running.
	 */
	public long getNumberOfRunningReplicas() {

		return runningReplicas;

	}


	/**
	 * Returns the number of replicas aborted.
	 * 
	 * @return the number of replicas aborted.
	 */
	public long getNumberOfAbortedReplicas() {

		return abortedReplicas;

	}


	/**
	 * Returns the number of replicas finished.
	 * 
	 * @return the number of replicas Finished.
	 */
	public long getNumberOfFinishedReplicas() {

		return finishedReplicas;
	}


	/**
	 * Returns the number of replicas failed.
	 * 
	 * @return the number of replicas failed.
	 */
	public long getNumberOfFailedReplicas() {

		return failedReplicas;

	}


	/**
	 * Returns the number of replicas cancelled.
	 * 
	 * @return the number of replicas cancelled.
	 */
	public long getNumberOfCancelledReplicas() {

		return cancelledReplicas;
	}

}
