/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.replicaexecutor;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.event.Shutdownable;
import org.ourgrid.mygrid.scheduler.event.SimpleResponseEvent;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;

/**
 * Description: This class is responsible for managing the access of a replica
 * to it's task's critical region. <br>
 * The critical region of a task is composed by it's two final tasks. Only one
 * replica of each task should be capable to execute it's critical region. Once
 * a replica has finished, no other replica of the same task should be allowed
 * to execute it's critical region.
 * 
 * @version 1.0 Created on 09/07/2004
 */
public class PermissionManager implements Shutdownable {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( PermissionManager.class );

	/**
	 * @author Alexandre N?brega Duarte - alex@dsc.ufcg.edu.br Description: A
	 *         class that associates a repica to it's response queue
	 * @version 1.0 Created on 13/07/2004
	 */
	private class Pair {

		/**
		 * The replica's response queue
		 */
		private EventQueue<ResponseEvent<Boolean>> eventQueue;

		/**
		 * The replica.
		 */
		private ReplicaEntry replica;


		/**
		 * Creates a Pair with a replica and a response queue
		 * 
		 * @param replica The replica
		 * @param eventQueue The replica's response queue.
		 */
		public Pair( ReplicaEntry replica, EventQueue<ResponseEvent<Boolean>> eventQueue ) {

			this.replica = replica;
			this.eventQueue = eventQueue;

		}


		/**
		 * Replica's getter method.
		 * 
		 * @return The replica
		 */
		public ReplicaEntry getReplica() {

			return replica;
		}


		/**
		 * Response queue getter method
		 * 
		 * @return The replica's response queue
		 */
		public EventQueue<ResponseEvent<Boolean>> getEventQueue() {

			return eventQueue;
		}

	}

	/**
	 * A map of the replicas that are wainting to enter it's task's critical
	 * region.
	 */
	private Map<String,Collection<Pair>> mapOfBlockedReplicas;

	/**
	 * A map of the replicas that are currently inside it's task's critical
	 * region.
	 */
	private Map<String,Pair> mapOfRunningReplicas;

	private Set<String> setOfFinishedTasks;


	/**
	 * Creates a PermissionManager
	 */
	public PermissionManager() {

		mapOfBlockedReplicas = new HashMap<String,Collection<Pair>>();
		mapOfRunningReplicas = new HashMap<String,Pair>();
		setOfFinishedTasks = new HashSet<String>();
	}


	/**
	 * Sends a event to a replica granting permission to access it's taks's
	 * critical region.
	 * 
	 * @param responseQueue The response queue of the replica
	 */
	private void grantPermission( EventQueue<ResponseEvent<Boolean>> responseQueue ) {

		responseQueue.put( new SimpleResponseEvent<Boolean>( true ) );

	}


	/**
	 * Sends a event to a replica denying permission to access it's taks's
	 * critical region.
	 * 
	 * @param responseQueue The response queue of the replica
	 */
	private void denyPermission( EventQueue<ResponseEvent<Boolean>> responseQueue ) {

		responseQueue.put( new SimpleResponseEvent<Boolean>( false ) );

	}


	/**
	 * Denies permission for executing output and final phases to a all replicas
	 * of a Task.
	 * 
	 * @param replica The reference replica to deny permissions.
	 */
	private void denyPermissionToReplicasOfTask( ReplicaEntry replica ) {

		Collection<Pair> c = getBlockedReplicas( replica );

		if ( c != null ) {
			Iterator<Pair> it = c.iterator();

			while ( it.hasNext() ) {
				Pair p = it.next();
				denyPermission( p.getEventQueue() );
				LOG.debug( "Permission denied to replica " + p.toString() );

			}

		}

	}


	/**
	 * @return A collection with all replica of a task that are blocked, waiting
	 *         to access it's task's critical region.
	 */
	protected Collection<Pair> getBlockedReplicas( ReplicaEntry replica ) {

		return mapOfBlockedReplicas.get( replica.getJobId() + "." + replica.getTaskId() );

	}


	/**
	 * @return A collection with all replica that are blocked, waiting to access
	 *         it's task's critical region.
	 */
	public Collection<Pair> getBlockedReplicas() {

		Vector<Pair> v = new Vector<Pair>();

		for ( Collection<Pair> collection : mapOfBlockedReplicas.values() ) {
			for ( Pair p : collection ) {
				v.add( p );
			}

		}

		return v;

	}


	/**
	 * @return A collection with all replicas that are currently inside it's
	 *         taks's critical region.
	 */
	public Collection<ReplicaEntry> getRunningReplicas() {

		Vector<ReplicaEntry> v = new Vector<ReplicaEntry>();
		for ( Pair p : mapOfRunningReplicas.values() ) {
			v.add( p.getReplica() );
		}

		return v;

	}


	/**
	 * Ask the <code>PermissionManager</code> to verify if a given replica can
	 * access it's task's critical region. If no replica of the same task is
	 * already inside it's critical region, the permission is granted,
	 * otherwise, the replica will be blocked until the task's critical region
	 * is free.
	 * 
	 * @param replica The replica that asks for permission
	 * @param responseQueue The replica's response queue
	 */
	public void requestPermission( ReplicaEntry replica, EventQueue<ResponseEvent<Boolean>> responseQueue ) {

		LOG.debug( "Permission request from replica " + replica.toString() );

		if ( setOfFinishedTasks.contains( replica.getJobId() + "." + replica.getTaskId() ) ) {
			denyPermission( responseQueue );
		}

		else if ( mapOfRunningReplicas.get( replica.getJobId() + "." + replica.getTaskId() ) == null ) {

			mapOfRunningReplicas
					.put( replica.getJobId() + "." + replica.getTaskId(), new Pair( replica, responseQueue ) );
			grantPermission( responseQueue );
			LOG.debug( "Permission granted to replica " + replica.toString() );

		} else {

			Collection<Pair> l = mapOfBlockedReplicas.get( replica.getJobId() + "." + replica.getTaskId() );

			if ( l == null ) {

				l = new LinkedList<Pair>();
				mapOfBlockedReplicas.put( replica.getJobId() + "." + replica.getTaskId(), l );

			}

			l.add( new Pair( replica, responseQueue ) );

		}

	}


	/**
	 * If a replica fails inside it's critical region and there are other
	 * replicas waiting to access it's critical region the
	 * <code>PermissionManager</code> should allow the first replica of the
	 * queue to access it's critical region. Elsewhere, if a replica fails
	 * outside it's critical region, the <code>PermissionManager</code> should
	 * remove it from the waiting queue.
	 * 
	 * @param replica The replica who has failed.
	 */
	public void replicaFailed( ReplicaEntry replica ) {

		String key = replica.getJobId() + "." + replica.getTaskId();

		Pair pair;

		pair = mapOfRunningReplicas.get( key );

		if ( pair != null && pair.getReplica().equals( replica ) ) {

			mapOfRunningReplicas.remove( key );

			List<Pair> l = (List<Pair>) mapOfBlockedReplicas.get( key );

			if ( l != null && l.size() != 0 ) {

				Pair other = l.get( 0 );
				l.remove( 0 );
				mapOfRunningReplicas.put( key, other );
				grantPermission( other.getEventQueue() );
				LOG.debug( "Replica " + replica.toString() + " has failed. Permission granted to replica "
						+ other.getReplica().toString() );

				if ( l.size() == 0 ) {

					mapOfBlockedReplicas.remove( key );

				}

			}

		} else {

			List<Pair> l = (List<Pair>) mapOfBlockedReplicas.get( key );

			if ( l != null ) {

				for ( int i = 0; i < l.size(); i++ ) {

					if ( l.get( i ).getReplica().equals( replica ) ) {

						l.remove( i );

						if ( l.size() == 0 ) {

							mapOfBlockedReplicas.remove( key );

						}

						break;

					}

				}

			}

		}

	}


	/**
	 * Removes all references to replica of a specified task from the
	 * <code>PermissionManager</code>'s internal data structures.
	 * 
	 * @param replica One replica from the task.
	 */
	private void cleanReplicaMaps( ReplicaEntry replica ) {

		String key = replica.getJobId() + "." + replica.getTaskId();

		denyPermissionToReplicasOfTask( replica );

		mapOfRunningReplicas.remove( key );
		mapOfBlockedReplicas.remove( key );

	}


	/**
	 * If a replica has been aborted, all replicas of the same task will be
	 * aborted too soon. So, the <code>PermissionManager</code> should remove
	 * all references for replicas of the same task from it's internal data
	 * structures.
	 * 
	 * @param replica
	 */
	public void replicaAborted( ReplicaEntry replica ) {

		cleanReplicaMaps( replica );

	}


	/**
	 * If a replica has finished, all replicas of the same task will be aborted
	 * soon. So, the <code>PermissionManager</code> should remove all
	 * references for replicas of the same task from it's internal data
	 * structures.
	 * 
	 * @param replica
	 */
	public void replicaFinished( ReplicaEntry replica ) {

		setOfFinishedTasks.add( replica.getJobId() + "." + replica.getTaskId() );
		cleanReplicaMaps( replica );

	}


	/**
	 * If a replica has been cancelled, all replicas of the same task will be
	 * cancelled too soon. So, the <code>PermissionManager</code> should
	 * remove all references for replicas of the same task from it's internal
	 * data structures.
	 * 
	 * @param replica
	 */
	public void replicaCanceled( ReplicaEntry replica ) {

		setOfFinishedTasks.add( replica.getJobId() + "." + replica.getTaskId() );
		cleanReplicaMaps( replica );

	}


	/**
	 * Shutdown the <code>PermissionManager</code>
	 */
	public void shutdown() {

		LOG.debug( "The PermissionManager is going down" );

	}

}
