/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.replicaexecutor;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.event.ActionEvent;
import org.ourgrid.common.event.Event;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ShutdownEventEngine;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.mygrid.scheduler.EBSchedulerFacade;

/**
 * Description: This class is responsible for receiving events that come from
 * the <code>EBReplicaExecutorFacade</code> and process then.
 * 
 * @version 1.0 Created on 13/07/2004
 */
public class ReplicaExecutorEventEngine implements ShutdownEventEngine {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( ReplicaExecutorEventEngine.class );

	/**
	 * The <code>EventQueue</code>
	 */
	private EventQueue<Event> eventQueue;

	/**
	 * A control flag
	 */
	private boolean mustShutdown;

	/**
	 * The <code>EBReplicaManager</code>
	 */
	private EBReplicaManager ebReplicaManager;

	/**
	 * The ReplicaExecutorEventEngine own <code>Thread</code>.
	 */
	private Thread myThread;

	/**
	 * Indicates if this JICEventHandler is alive.
	 */
	private boolean isAlive;


	/**
	 * Creates an empty ReplicaExecutorEventEngine This constructor should not
	 * be used.
	 */
	protected ReplicaExecutorEventEngine() {

	}


	/**
	 * Creates a ReplicaExecutorEventEngine with a
	 * <code>EBReplicaExecutorFacade</code> and a
	 * <code>EBSchedulerFacade</code>
	 * 
	 * @param ebReplicaExecutorFacade The EBReplicaExecutorFacade who sends
	 *        events to this event processor .
	 * @param ebSchedulerFacade The EBSchedulerFacade who sends requests to the
	 *        EBReplicaExecutorFacade
	 */
	public ReplicaExecutorEventEngine( EBReplicaExecutorFacade ebReplicaExecutorFacade,
										EBSchedulerFacade ebSchedulerFacade ) {

		this.eventQueue = new EventQueue<Event>();

		this.ebReplicaManager = new EBReplicaManager( ebSchedulerFacade, new ReplicaExecutorThreadManager(
				ebReplicaExecutorFacade ), new PermissionManager() );

		this.mustShutdown = false;
		this.myThread = new Thread( this, "ReplicaExecutorEventProcessorThread" );
	}


	/**
	 * Starts the ReplicaExecutorEventEngine <code>Thread</code>.
	 */
	public void startProcessing() {

		this.isAlive = true;
		this.myThread.start();
	}


	/**
	 * Puts an <code>Event</code> in the EventQueue.
	 * 
	 * @param event The <code>Event</code>.
	 */
	public void putEvent( ActionEvent event ) {

		eventQueue.put( event );

	}


	/**
	 * The <code>Thread</code> behavior.
	 */
	public void run() {

		while ( !mustShutdown ) {
			ActionEvent actionEvent = (ActionEvent) eventQueue.blockingRemove();

			if ( actionEvent != null ) {

				long start = System.currentTimeMillis();

				try {

					actionEvent.process();

				} catch ( Exception e ) {

					LOG.error( "Error in event process! Cause: " + e.getMessage() );

				}

				if ( Configuration.getInstance().getProperty( "development" ) != null ) {

					LOG.debug( " QUEUE SIZE " + eventQueue.size() + " EVENT " + actionEvent.getClass().getName()
							+ " EXECUTION TIME: " + (System.currentTimeMillis() - start) + " ms" );

				}

			}
		}

	}


	/**
	 * Marks this <code>Runnable</code> to be shutdown.
	 */
	public void shutdown( EventQueue<ShutdownResponseEvent> eq ) {

		this.mustShutdown = true;
		this.ebReplicaManager.shutdown();
		eq.put( new ShutdownResponseEvent() );
		this.isAlive = false;
	}


	/**
	 * Gets the EBReplicaManager object
	 * 
	 * @return The <code>EBReplicaManager</code>
	 */
	public EBReplicaManager getEBReplicaManager() {

		return this.ebReplicaManager;
	}


	/**
	 * Indicates if the thread is alive
	 * 
	 * @return <b>true</b> if the thread is alive, <b>false</b> otherwise
	 */
	public boolean isAlive() {

		return this.isAlive;
	}

}
