/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.replicaexecutor;

import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.ExecutorResult;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.matcher.Matcher;
import org.ourgrid.common.spec.IOBlock;
import org.ourgrid.common.spec.IOEntry;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.common.util.JavaFileUtil;
import org.ourgrid.common.util.StringUtil;
import org.ourgrid.common.util.TimeDataGenerator;
import org.ourgrid.gridmachine.exception.FailCauseException;
import org.ourgrid.gridmachine.exception.GenericFailCauseException;
import org.ourgrid.mygrid.replicaexecutor.exception.ReplicaFailedException;
import org.ourgrid.mygrid.replicaexecutor.exception.UnableToFinishReplicaException;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.ReplicaExecutorResult;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;

/**
 * Description: This class is in charge of executing a replica.
 * 
 * @version 1.0 Created on 13/07/2004
 */
public class ReplicaExecutorThread implements ReplicaExecutor {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( ReplicaExecutorThread.class );

	/**
	 * The grid machine where the replica will be executed
	 */
	private GumClient gumClient;

	/**
	 * The <code>EBReplicaExecutorFacade</code> object
	 */
	private EBReplicaExecutorFacade ebReplicaExecutorFacade;

	/**
	 * The replica to be executed
	 */
	private ReplicaEntry replica;

	/**
	 * This thread's name
	 */
	private String threadName;

	/**
	 * A <code>TimeDataGenerator</code> for init phase
	 */
	private TimeDataGenerator initData;

	/**
	 * A <code>TimeDataGenerator</code> for remote phase
	 */
	private TimeDataGenerator remoteData;

	/**
	 * A <code>TimeDataGenerator</code> for final phase
	 */
	private TimeDataGenerator finalData;

	/**
	 * Map with environment variables.
	 */
	private Map<String,String> envVars;

	/**
	 * A <code>ReplicaExecutorResult</code> object that represents the result
	 * of the replica execution
	 */
	private ReplicaExecutorResult replicaExecutorResult;

	/**
	 * The specification of the task that is associated with the replica
	 */
	private TaskSpec spec;

	/**
	 * The handle that identifies that command executing at the grid machine
	 */
	private ExecutorHandle taskHandle;


	/**
	 * Main constructor
	 * 
	 * @param threadName This thread's name
	 * @param replica The replica to be executed
	 * @param gumClient The grid machine where the replica will be executed
	 * @param ebReplicaExecutorFacade The <code>EBReplicaExecutorFacade</code>
	 *        object
	 */
	public ReplicaExecutorThread( String threadName, ReplicaEntry replica, GumClient gumClient,
									EBReplicaExecutorFacade ebReplicaExecutorFacade ) {

		this.threadName = threadName;
		this.replica = replica;
		this.replica.setStatus( ExecutionStatus.RUNNING );
		this.spec = replica.getSpec();
		this.gumClient = gumClient;
		this.ebReplicaExecutorFacade = ebReplicaExecutorFacade;

		replicaExecutorResult = new ReplicaExecutorResult();
		replicaExecutorResult.setReplica( replica );
		replica.setResult( replicaExecutorResult );

		replicaExecutorResult.setGumClient( gumClient );

	}


	/**
	 * Verifies whether the replica being executed may enter output/final
	 * phases.
	 * 
	 * @return <b>true </b> if the replica may enter output phase, <b>false </b>
	 *         otherwise
	 */
	private boolean requestPermission() {

		boolean perm = false;

		try {
			perm = ebReplicaExecutorFacade.requestPermission( replica ).blockingRemove().getResponse();
		} catch ( Exception e ) {
			LOG.error( "", e );
		}

		return perm;

	}


	/**
	 * Tries to cancel the replica associated with this thread
	 */
	public void cancelReplica() {

		replica.setStatus( ExecutionStatus.CANCELLED );
		ebReplicaExecutorFacade.replicaCanceled( replica );
	}


	/**
	 * Tries to abort the replica associated with this thread.
	 */
	public void abortReplica() {

		replica.setStatus( ExecutionStatus.ABORTED );
		try {
			gumClient.stopReplica();
		} catch ( FailCauseException e ) {
			LOG.debug( "Could not stop replica at gum: " + gumClient.getGumSpec().getGumID().getUrl() );
		}
		ebReplicaExecutorFacade.replicaAborted( replica );
	}


	/**
	 * Finalize execution of the replica associated with this thread.
	 */
	private void finishReplica() {

		replica.setStatus( ExecutionStatus.FINISHED );
		replica.setResult( replicaExecutorResult );
		ebReplicaExecutorFacade.replicaFinished( replica );
	}


	/**
	 * Change the replica state to FAILED and notifies the replica failure to
	 * the ReplicaExecutor.
	 */
	private void failReplica( UnableToFinishReplicaException e ) {

		replica.setStatus( ExecutionStatus.FAILED );
		replica.setResult( replicaExecutorResult );

		if ( e.getCause() != null ) {
			replicaExecutorResult.setErrorCause( e.getFailCause() );
		} else {
			replicaExecutorResult.setErrorCause( new GenericFailCauseException( "Cannot determine error cause" ) );
			LOG.error( "Exception caught: " + e.getMessage(), e );
		}

		ebReplicaExecutorFacade.replicaFailed( replica );
		LOG.error( "FAILED REPLICA CAUSE: " + e.getCause() );

	}


	/**
	 * Status Gets the replica associated with this thread
	 * 
	 * @return The replica associated with this thread
	 */
	public ReplicaEntry getReplica() {

		return replica;

	}


	/**
	 * Gets the name of this thread
	 * 
	 * @return The name of this thread
	 */
	public String getName() {

		return threadName;

	}


	/**
	 * Gets the state of the replica associated with this thread
	 * 
	 * @return The state of the replica associated with this thread
	 */
	public ExecutionStatus getReplicaState() {

		return replica.getStatus();
	}


	/**
	 * Gets this thread's execution state
	 * 
	 * @return This thread's execution state
	 */
	public synchronized ExecutionStatus getStatus() {

		return replica.getStatus();

	}


	/**
	 * Gets the replica's execution result
	 * 
	 * @return A <code>ReplicaExecutorResult</code> object that represents the
	 *         result of the replica execution
	 */
	public ReplicaExecutorResult getReplicaExecutorResult() {

		return replicaExecutorResult;

	}


	/**
	 * Tries to execute the replica
	 */
	public void run() {

		ExecutionStatus finalReplicaStatus = null;
		UnableToFinishReplicaException failCause = null;

		try {
			if ( this.replica.getStatus().equals( ExecutionStatus.RUNNING ) )
				createEnvVars();
			if ( this.replica.getStatus().equals( ExecutionStatus.RUNNING ) )
				initPhase();
			if ( this.replica.getStatus().equals( ExecutionStatus.RUNNING ) )
				remotePhase();
			if ( this.replica.getStatus().equals( ExecutionStatus.RUNNING ) ) {
				if ( requestPermission() ) {
					finalPhase();
					finalReplicaStatus = ExecutionStatus.FINISHED;
				} else {
					abortReplica();
					return;
				}
			}
		} catch ( UnableToFinishReplicaException e ) {
			if ( this.replica.getStatus().equals( ExecutionStatus.RUNNING ) ) {
				failCause = e;
				finalReplicaStatus = ExecutionStatus.FAILED;
			}
		} finally {
			report();
		}

		if ( ExecutionStatus.FINISHED.equals( finalReplicaStatus ) ) {
			try {
				gumClient.stopReplica();
			} catch ( FailCauseException e ) {
				LOG.debug( "Could not stop replica at gum: " + gumClient.getGumSpec().getGumID().getUrl(), e );
			}
			finishReplica();
		} else if ( ExecutionStatus.FAILED.equals( finalReplicaStatus ) ) {
			try {
				gumClient.stopReplica();
			} catch ( FailCauseException e ) {
				LOG.debug( "Could not stop replica at gum: " + gumClient.getGumSpec().getGumID().getUrl(), e );
			}
			failReplica( failCause );
		}
	}


	/**
	 * Creates environment variables necessary to proceed with execution
	 * 
	 * @throws UnableToFinishReplicaException
	 */
	private void createEnvVars() throws UnableToFinishReplicaException {

		String playpenName = null, storageDir = null;

		initData = new TimeDataGenerator( "Replica " + replica.toString() + " Init phase duration: " );
		remoteData = new TimeDataGenerator( "Replica " + replica.toString() + " assigned to "
				+ gumClient.getGumSpec().getGumID() + " Remote phase duration: " );
		finalData = new TimeDataGenerator( "Replica " + replica.toString() + " Final phase duration: " );

		this.replicaExecutorResult.setExecutionTimes( initData, remoteData, finalData );

		envVars = new HashMap<String,String>( gumClient.getGumSpec().getAttributes() );

		try {

			gumClient.startReplica();

			playpenName = gumClient.createPlaypen( 1 );

			storageDir = gumClient.createStorageDirectory();

		} catch ( FailCauseException e ) {

			throw new UnableToFinishReplicaException( "Unable to create envVars map.", e );

		}

		envVars.put( "PROC", gumClient.getGumSpec().getAttributes().get( Configuration.PROP_NAME ) );
		envVars.put( "PLAYPEN", playpenName );
		envVars.put( "TASK", Integer.toString( replica.getTaskId() ) );
		envVars.put( "JOB", Integer.toString( replica.getJobId() ) );
		envVars.put( "STORAGE", storageDir );
	}


	/**
	 * In this method, the Task Executor will process the INPUT field entries,
	 * verifying if it is necessary transfer files to MG.STORAGE or to another
	 * remote place.
	 * 
	 * @throws UnableToFinishReplicaException if any problem occurs during init
	 *         phase.
	 */
	private void initPhase() throws UnableToFinishReplicaException {

		LOG.info( "Replica " + replica.toString() + " init phase execution started." );

		initData.setStartTime();

		IOBlock input = spec.getInitBlock();

		if ( input != null ) {

			Iterator<String> conditionIterator = input.getConditions();

			while ( conditionIterator.hasNext() ) {

				String condition = conditionIterator.next();

				Vector<IOEntry> ioEntries = input.getEntry( condition );

				if ( ioEntries != null ) {

					Iterator<IOEntry> ioEntriesIterator = ioEntries.iterator();

					while ( ioEntriesIterator.hasNext() ) {

						IOEntry ioEntry = ioEntriesIterator.next();

						String localFile = ioEntry.getSourceFile();
						String remoteFile = ioEntry.getDestiny();

						try {

							if ( localFile == null || remoteFile == null ) {
								throw new GenericFailCauseException( "Localfile and remotefile must be specified!" );
							}

							localFile = StringUtil.replaceVariables( localFile, envVars );
							remoteFile = StringUtil.replaceVariables( remoteFile, envVars );

							if ( Matcher.match( condition, gumClient.getGumSpec().getAttributes() ) ) {

								String operation = ioEntry.getCommand();

								if ( operation.trim().equalsIgnoreCase( "PUT" ) ) {

									LOG.debug( "Replica " + replica.toString() + " about put file '" + localFile
											+ "' to '" + remoteFile + "'" );

									if ( !JavaFileUtil.isAbsolutePath( remoteFile ) ) {
										remoteFile = envVars.get( "PLAYPEN" ) + File.separator + remoteFile;
									}

									gumClient.putFile( localFile, JavaFileUtil.getTranslatedFilePath( remoteFile,
											gumClient.getGumSpec().getAttributes() ) );
									LOG.info( "Replica " + replica.toString() + " put file " + localFile + " to "
											+ remoteFile + " sucessfully." );

								} else if ( operation.trim().equalsIgnoreCase( "STORE" ) ) {

									LOG.debug( "Replica " + replica.toString() + " about to store file " + localFile );

									if ( !JavaFileUtil.isAbsolutePath( remoteFile ) ) {
										remoteFile = envVars.get( "STORAGE" ) + File.separator + remoteFile;
									}
									gumClient.storeFile( localFile, JavaFileUtil.getTranslatedFilePath( remoteFile,
											gumClient.getGumSpec().getAttributes() ) );
									LOG.info( "Replica " + replica.toString() + " store file " + localFile
											+ " sucessfully." );

								}

							}

						} catch ( FailCauseException e ) {

							throw new UnableToFinishReplicaException( "Error during init phase: ", e );

						}

					}

				}

			}

		}

		ExecutorResult execResult = new ExecutorResult();
		execResult.setExitValue( 0 );
		execResult.setStdout( "Init phase OK." );

		replicaExecutorResult.setInitResult( execResult );
		initData.setEndTime();

		LOG.info( "Replica " + replica.toString() + " init phase execution completed: " + execResult );

	}


	/**
	 * This method executes the remote-part of the task.
	 * 
	 * @throws UnableToFinishReplicaException if any problem occurs during
	 *         remote phase.
	 */
	private void remotePhase() throws UnableToFinishReplicaException {

		LOG.info( "Replica " + replica.toString() + " remote phase execution started. " );

		remoteData.setStartTime();
		ExecutorResult execResult = null;

		try {

			taskHandle = gumClient.remoteExecute( envVars.get( "PLAYPEN" ), spec.getRemoteExec(), envVars );

			execResult = gumClient.getResult( taskHandle );

			if ( execResult == null ) {
				execResult = new ExecutorResult();
				execResult.setExitValue( 1 );
				execResult.setStderr( "Unknown Error" );
				execResult.setStdout( "Unknown Error" );
			}

			if ( execResult.getExitValue() != 0 ) {
				throw new ReplicaFailedException( replica, execResult );
			}

		} catch ( FailCauseException e ) {

			throw new UnableToFinishReplicaException( "Error during remote phase: ", e );

		}

		replicaExecutorResult.setRemoteResult( execResult );
		remoteData.setEndTime();

		LOG.info( "Replica " + replica.toString() + " remote phase execution completed:  " + execResult );

	}


	/**
	 * In this method, the Task Executor will process the OUTPUT field entries,
	 * verifying if it is necessary transfer files from grid machine to home
	 * machine.
	 * 
	 * @throws UnableToFinishReplicaException if any problem occurs during final
	 *         phase.
	 */
	private void finalPhase() throws UnableToFinishReplicaException {

		LOG.info( "Replica " + replica.toString() + " final phase execution started." );

		finalData.setStartTime();

		IOBlock output = spec.getFinalBlock();

		if ( output != null ) {

			Iterator<String> conditionIterator = output.getConditions();

			while ( conditionIterator.hasNext() ) {

				String condition = conditionIterator.next();

				Vector<IOEntry> ioEntries = output.getEntry( condition );

				if ( ioEntries != null ) {

					Iterator<IOEntry> ioEntriesIterator = ioEntries.iterator();

					while ( ioEntriesIterator.hasNext() ) {

						IOEntry ioEntry = ioEntriesIterator.next();

						try {

							if ( Matcher.match( condition, gumClient.getGumSpec().getAttributes() ) ) {

								String operation = ioEntry.getCommand();

								if ( operation.trim().equalsIgnoreCase( "GET" ) ) {
									String remoteFile = StringUtil.replaceVariables( ioEntry.getSourceFile(), envVars );
									String localFile = StringUtil.replaceVariables( ioEntry.getDestiny(), envVars );

									if ( !JavaFileUtil.isAbsolutePath( remoteFile ) ) {
										remoteFile = envVars.get( "PLAYPEN" ) + File.separator + remoteFile;
									}

									gumClient.getFile( JavaFileUtil.getTranslatedFilePath( remoteFile, gumClient
											.getGumSpec().getAttributes() ), localFile );
								}

							}

						} catch ( FailCauseException e ) {
							throw new UnableToFinishReplicaException( "Error during final phase: ", e );

						}

					}

				}

			}

		}

		ExecutorResult execResult = new ExecutorResult();
		execResult.setExitValue( 0 );
		execResult.setStdout( "final phase OK." );

		replicaExecutorResult.setFinalResult( execResult );
		finalData.setEndTime();

		LOG.info( "Replica " + replica.toString() + " final phase execution completed: " + execResult );

	}


	/**
	 * Logs start and end times of each phase
	 */
	private void report() {

		if ( initData != null ) {
			initData.report();
		}

		if ( remoteData != null ) {
			remoteData.report();
		}

		if ( finalData != null ) {
			finalData.report();
		}

	}

}
