/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.replicaexecutor;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.ourgrid.common.event.Shutdownable;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.mygrid.replicaexecutor.exception.ReplicaExecutorThreadException;
import org.ourgrid.mygrid.replicaexecutor.test.FakeReplicaExecutorThread;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;

/**
 * This class takes care of a thread pool. Threads inside the pool are in charge
 * of executing replicas. Description: A thread pool that contains all
 * ReplicaExecutorThread objects.
 * 
 * @version 1.0 Created on 12/07/2004
 */
public class ReplicaExecutorThreadManager extends Thread implements Shutdownable {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( ReplicaExecutorThreadManager.class );

	/**
	 * Real execution identifier
	 */
	public static final int MG = 0;

	/**
	 * Fake execution identifier
	 */
	public static final int FAKE = 1;

	/**
	 * Amount of threads in the pool TODO This number should be calculated in a
	 * dynamic way based on the number of available GuMs
	 */
	public static int THREAD_POOL_SIZE = 200;

	/**
	 * Defines whether this class must use either
	 * <code>FakeReplicaExecutorThread</code> (usually for tests) or
	 * <code>MGReplicaExecutorThread</code> (for real execution)
	 */
	private int type;

	/**
	 * Maps a replica identifier to its executor thread
	 */
	private Map<String,ReplicaExecutor> replicaExecutorThreads;

	/**
	 * The <code>EBReplicaExecutorFacade</code>
	 */
	private EBReplicaExecutorFacade ebReplicaExecutorFacade;

	/**
	 * The thread pool itself
	 */
	private ThreadPoolExecutor threadPool;

	/**
	 * A Map with all task that had a finished replica.
	 */
	private Map<String,Object> finishedTasks;


	/**
	 * Simple constructor.
	 * 
	 * @param ebReplicaExecutorFacade The <code>EBReplicaExecutorFacade</code>
	 */
	public ReplicaExecutorThreadManager( EBReplicaExecutorFacade ebReplicaExecutorFacade ) {

		this( ebReplicaExecutorFacade, MG, THREAD_POOL_SIZE );

	}


	/**
	 * Main constructor
	 * 
	 * @param ebReplicaExecutorFacade The <code>EBReplicaExecutorFacade</code>
	 * @param type Defines whether this class must use either
	 *        <code>FakeReplicaExecutorThread</code> (usually for tests) or
	 *        <code>MGReplicaExecutorThread</code> (for real execution)
	 * @param threadPoolSize The number of threads in the threadPool
	 */
	public ReplicaExecutorThreadManager( EBReplicaExecutorFacade ebReplicaExecutorFacade, int type, int threadPoolSize ) {

		super( "ThreadManager" );

		this.ebReplicaExecutorFacade = ebReplicaExecutorFacade;
		this.type = type;

		threadPool = new ThreadPoolExecutor( threadPoolSize, Integer.MAX_VALUE, 60, TimeUnit.SECONDS,
				new LinkedBlockingQueue<Runnable>() );

		replicaExecutorThreads = new HashMap<String,ReplicaExecutor>();
		finishedTasks = new HashMap<String,Object>();

	}


	/**
	 * Verifies whether a task already has a finished replica
	 * 
	 * @param r A reference replica
	 * @return <b>true</b> if there's a finished replica for the task, <b>false</b>
	 *         otherwise
	 */
	private boolean taskHasAlreadyAFinishedReplica( ReplicaEntry r ) {

		return finishedTasks.get( r.getJobId() + "." + r.getTaskId() ) != null;

	}


	/**
	 * Provides a replicaExecutorThread
	 * 
	 * @param replica Replica to be executed by the
	 *        <code>ReplicaExecutorThread</code>
	 * @param gumClient Grid machine where the replica will be executed.
	 * @return A <code>ReplicaExecutorThread</code> object to execute the
	 *         replica.
	 * @throws ReplicaExecutorThreadException
	 */
	public ReplicaExecutor obtainReplicaExecutorThread( ReplicaEntry replica, GumClient gumClient )
		throws ReplicaExecutorThreadException {

		ReplicaExecutor aThread = replicaExecutorThreads.get( replica.toString() );

		if ( aThread != null ) {

			throw new ReplicaExecutorThreadException( "The replica " + replica.toString()
					+ " already have a ReplicaExecutorThread" );

		}

		if ( type == MG ) {

			aThread = new ReplicaExecutorThread( "ReplicaExecutorThread:" + replica.toString() + "-"
					+ gumClient.getGumSpec().getGumID(), replica, gumClient, ebReplicaExecutorFacade );

		} else {

			aThread = new FakeReplicaExecutorThread( replica, ebReplicaExecutorFacade );

		}
		synchronized ( this ) {

			replicaExecutorThreads.put( replica.toString(), aThread );

		}

		if ( taskHasAlreadyAFinishedReplica( replica ) ) {

			replica.setStatus( ExecutionStatus.ABORTED );
			ebReplicaExecutorFacade.replicaAborted( replica );
		} else {

			threadPool.execute( aThread );
		}

		return aThread;

	}


	/**
	 * Releases a thread to the pool
	 * 
	 * @param replica Replica associated with the
	 *        <code>ReplicaExecutorThread</code> to be released.
	 * @throws ReplicaExecutorThreadException
	 */
	public void releaseReplicaExecutorThread( ReplicaEntry replica ) throws ReplicaExecutorThreadException {

		ReplicaExecutor aThread = null;

		synchronized ( this ) {

			aThread = replicaExecutorThreads.remove( replica.toString() );

		}

		if ( aThread == null ) {

			throw new ReplicaExecutorThreadException( "There's no ReplicaExecutorThread for the replica " + replica );

		}
	}


	/**
	 * Provides a <code>Collection</code> that contains all
	 * <code>ReplicaExecutorThread</code> objects.
	 * 
	 * @return A <code>Collection</code> that contains all
	 *         <code>ReplicaExecutorThread</code> objects.
	 */
	public synchronized Collection<ReplicaExecutor> getReplicaExecutorThreads() {

		Vector<ReplicaExecutor> v = new Vector<ReplicaExecutor>();
		for ( ReplicaExecutor re : replicaExecutorThreads.values() ) {
			v.add( re );
		}
		return v;

	}


	/**
	 * Shuts down all threads in the pool, including the pool itself
	 */
	public void shutdown() {

		LOG.debug( "The ThreadManager is going down" );
		threadPool.shutdownNow();

	}


	/**
	 * Registers that a replica has finished
	 * 
	 * @param replica The replica that has finished.
	 */
	public void replicaFinished( ReplicaEntry replica ) {

		finishedTasks.put( replica.getJobId() + "." + replica.getTaskId(), new Object() );

	}

}
