/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.replicaexecutor.test;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.gum.FakeGumClient;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.mygrid.replicaexecutor.ReplicaExecutor;
import org.ourgrid.mygrid.replicaexecutor.ReplicaExecutorThreadManager;
import org.ourgrid.mygrid.replicaexecutor.exception.ReplicaExecutorThreadException;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.test.ReplicaExecutorFastTest;
import org.ourgrid.test.TestUtils;
import org.ourgrid.test.threadServices.ThreadPoolServices;

/**
 * <code>ThreadManager</code> unit's test.
 */
public class ReplicaExecutorThreadManagerTest extends TestCase implements ReplicaExecutorFastTest {

	private static final int NUMBER_OF_JOBS = 1;

	private static final int TASKS_PER_JOB = 2;

	private static final int REPLICAS_PER_TASK = 3;

	private ReplicaExecutorThreadManager manager;

	private GumClient gumClient;

	private ReplicaEntry replicas[][][];

	private TestUtils testUtils;


	@Override
	protected void setUp() throws Exception {

		super.setUp();

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createMGProperties();

		Configuration.getInstance( Configuration.MYGRID );

		manager = new ReplicaExecutorThreadManager( new FakeEBReplicaExecutorFacade(),
				ReplicaExecutorThreadManager.FAKE, 200 );

		Map<String,String> attribs = new HashMap<String,String>();
		attribs.put( "os", "Linux" );
		attribs.put( "site", "LSD" );

		GumID gumID = new GumID( TestUtils.FAKE_MACHINE_NAME, "0", "peer", "0" );
		GumSpec gumSpec = new GumSpec( attribs );
		gumClient = new FakeGumClient( gumID, gumSpec );

		replicas = new ReplicaEntry[ NUMBER_OF_JOBS ][ TASKS_PER_JOB ][ REPLICAS_PER_TASK ];
		for ( int jobId = 0; jobId < NUMBER_OF_JOBS; jobId++ ) {
			for ( int taskId = 0; taskId < TASKS_PER_JOB; taskId++ ) {
				for ( int replicaId = 0; replicaId < REPLICAS_PER_TASK; replicaId++ ) {
					replicas[jobId][taskId][replicaId] = new ReplicaEntry( replicaId, taskId, jobId, FakeTaskSpec
							.getFakeTaskSpec() );
				}
			}
		}

	}


	@Override
	public void tearDown() throws Exception {

		for ( int jobId = 0; jobId < NUMBER_OF_JOBS; jobId++ ) {
			for ( int taskId = 0; taskId < TASKS_PER_JOB; taskId++ ) {
				for ( int replicaId = 0; replicaId < REPLICAS_PER_TASK; replicaId++ ) {
					replicas[jobId][taskId][replicaId] = null;
				}
			}
		}

		replicas = null;

		manager.shutdown();

		Configuration.reset();
		testUtils.reset();

	}


	public void testGetReplicaExecutorThreads() {

		Collection<ReplicaExecutor> c = manager.getReplicaExecutorThreads();

		assertNotNull( "This collection should not be null ", c );
		assertEquals( "This collection should have no elements ", c.size(), 0 );
	}


	public void testObtainReplicaExecutorThread() {

		assertEquals( "The number of threads should be 0", manager.getReplicaExecutorThreads().size(), 0 );

		for ( int jobId = 0; jobId < NUMBER_OF_JOBS; jobId++ ) {
			for ( int taskId = 0; taskId < TASKS_PER_JOB; taskId++ ) {
				for ( int replicaId = 0; replicaId < REPLICAS_PER_TASK; replicaId++ ) {
					try {

						manager.obtainReplicaExecutorThread( replicas[jobId][taskId][replicaId], gumClient );

					} catch ( ReplicaExecutorThreadException e ) {

						fail( "This exception should not be thrown!" );

					}
				}
			}
		}

		for ( int jobId = 0; jobId < NUMBER_OF_JOBS; jobId++ ) {
			for ( int taskId = 0; taskId < TASKS_PER_JOB; taskId++ ) {
				for ( int replicaId = 0; replicaId < REPLICAS_PER_TASK; replicaId++ ) {
					try {

						manager.obtainReplicaExecutorThread( replicas[jobId][taskId][replicaId], gumClient );
						fail( "The line above should throw an exception" );

					} catch ( ReplicaExecutorThreadException e ) {

					}

				}
			}
		}

		ThreadPoolServices.waitUntilWorkIsDone();

		assertEquals( "The number of threads should be NUMBER_OF_JOBS * TASKS_PER_JOB * REPLICAS_PER_TASK", manager
				.getReplicaExecutorThreads().size(), NUMBER_OF_JOBS * TASKS_PER_JOB * REPLICAS_PER_TASK );
	}


	public void testReleaseReplicaExecutorThread() {

		assertEquals( "The number of threads should be 0", manager.getReplicaExecutorThreads().size(), 0 );

		for ( int jobId = 0; jobId < NUMBER_OF_JOBS; jobId++ ) {
			for ( int taskId = 0; taskId < TASKS_PER_JOB; taskId++ ) {
				for ( int replicaId = 0; replicaId < REPLICAS_PER_TASK; replicaId++ ) {
					try {

						manager.obtainReplicaExecutorThread( replicas[jobId][taskId][replicaId], gumClient );

					} catch ( ReplicaExecutorThreadException e ) {

						fail( "This exception should not be thrown!" );

					}
				}
			}
		}

		assertEquals( "The number of threads should be NUMBER_OF_JOBS * TASKS_PER_JOB * REPLICAS_PER_TASK", manager
				.getReplicaExecutorThreads().size(), NUMBER_OF_JOBS * TASKS_PER_JOB * REPLICAS_PER_TASK );

		for ( int jobId = 0; jobId < NUMBER_OF_JOBS; jobId++ ) {
			for ( int taskId = 0; taskId < TASKS_PER_JOB; taskId++ ) {
				for ( int replicaId = 0; replicaId < REPLICAS_PER_TASK; replicaId++ ) {
					try {

						manager.releaseReplicaExecutorThread( replicas[jobId][taskId][replicaId] );

					} catch ( ReplicaExecutorThreadException e ) {

						fail( "This exception should not be thrown!" );

					}
				}
			}
		}

		assertEquals( "The number of threads should be 0", manager.getReplicaExecutorThreads().size(), 0 );

		for ( int jobId = 0; jobId < NUMBER_OF_JOBS; jobId++ ) {
			for ( int taskId = 0; taskId < TASKS_PER_JOB; taskId++ ) {
				for ( int replicaId = 0; replicaId < REPLICAS_PER_TASK; replicaId++ ) {
					try {

						manager.releaseReplicaExecutorThread( replicas[jobId][taskId][replicaId] );
						fail( "The line above should throw an exception" );

					} catch ( ReplicaExecutorThreadException e ) {

					}
				}
			}
		}

		ThreadPoolServices.waitUntilWorkIsDone();

		assertEquals( "The number of threads should be 0", manager.getReplicaExecutorThreads().size(), 0 );
	}

}
