/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.replicaexecutor.test;

import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.gum.FakeGumClient;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.mygrid.replicaexecutor.EBReplicaExecutorFacade;
import org.ourgrid.mygrid.replicaexecutor.ReplicaExecutorThread;
import org.ourgrid.mygrid.replicaexecutor.ReplicaExecutorThreadManager;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.FakeEBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.test.ReplicaExecutorFastTest;
import org.ourgrid.test.TestUtils;
import org.ourgrid.test.threadServices.ThreadServices;

/**
 * Description: This class tests <code>MGReplicaExecutor</code> class. It is a
 * FastTest.
 * 
 * @version 1.0 Created on 14/07/2004
 */
public class ReplicaExecutorThreadTest extends TestCase implements ReplicaExecutorFastTest {

	private EBReplicaExecutorFacade ebReplicaExecutorFacade;

	private TestUtils testUtils;

	ReplicaExecutorThread aThread;

	ReplicaEntry replicaEntry;


	@Override
	public void setUp() throws Exception {

		super.setUp();

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createMGProperties();

		Configuration.getInstance( Configuration.MYGRID );

		ebReplicaExecutorFacade = new EBReplicaExecutorFacade();

		FakeEBSchedulerFacade fakeEBSchedulerFacade = new FakeEBSchedulerFacade();

		fakeEBSchedulerFacade.ignoreCalls();

		ebReplicaExecutorFacade.config( new FakeReplicaExecutorEventEngine( fakeEBSchedulerFacade,
				new ReplicaExecutorThreadManager( ebReplicaExecutorFacade ) ) );

		replicaEntry = new ReplicaEntry( 0, 0, 0, FakeTaskSpec.getFakeTaskSpec() );
		Map<String,String> attributes = new HashMap<String,String>();
		attributes.put( "foo", "bar" );
		GumID gumID = new GumID( TestUtils.FAKE_MACHINE_NAME, "0", "peer", "0" );
		GumSpec gumSpec = new GumSpec( attributes );

		aThread = new ReplicaExecutorThread( "a replica executor thread", replicaEntry, new FakeGumClient( gumID,
				gumSpec ), ebReplicaExecutorFacade );

	}


	@Override
	public void tearDown() throws Exception {

		super.tearDown();

		ebReplicaExecutorFacade.shutdown();

		/*
		 * This usage of aspects is needed because the Threads invokes the
		 * Configuration before finishing. Therefore, the Configuration just can
		 * be reseted after thread goes down.
		 */
		ThreadServices.waitUntilWorkIsDone();

		Configuration.reset();
		testUtils.reset();

	}


	public void testRun() {

		assertEquals( "The thread execution status should be RUNNING", ExecutionStatus.RUNNING, aThread.getStatus() );
		assertEquals( "The replica status should be RUNNING", ExecutionStatus.RUNNING, replicaEntry.getStatus() );

		aThread.run();

		assertEquals( "The thread execution status should be FINISHED", ExecutionStatus.FINISHED, aThread.getStatus() );
		assertEquals( "The replica status should be FINISHED", ExecutionStatus.FINISHED, replicaEntry.getStatus() );

	}


	public void testCancelReplica() {

		assertEquals( "The thread execution status should be RUNNING", ExecutionStatus.RUNNING, aThread.getStatus() );
		assertEquals( "The replica status should be RUNNING", ExecutionStatus.RUNNING, replicaEntry.getStatus() );

		aThread.cancelReplica();

		assertEquals( "The thread execution status should be CANCELLED", ExecutionStatus.CANCELLED, aThread.getStatus() );
		assertEquals( "The replica status should be CANCELED_BY_USER", ExecutionStatus.CANCELLED, replicaEntry
				.getStatus() );

		aThread.run();

		assertEquals( "The thread execution status should be CANCELLED", ExecutionStatus.CANCELLED, aThread.getStatus() );
		assertEquals( "The replica status should be CANCELED_BY_USER", ExecutionStatus.CANCELLED, replicaEntry
				.getStatus() );
	}


	public void testAbortReplica() {

		assertEquals( "The thread execution status should be RUNNING", ExecutionStatus.RUNNING, aThread.getStatus() );
		assertEquals( "The replica status should be RUNNING", ExecutionStatus.RUNNING, replicaEntry.getStatus() );

		aThread.abortReplica();

		assertEquals( "The thread execution status should be ABORTED", ExecutionStatus.ABORTED, aThread.getStatus() );
		assertEquals( "The replica status should be ABORTED", ExecutionStatus.ABORTED, replicaEntry.getStatus() );

		aThread.run();

		assertEquals( "The thread execution status should be ABORTED", ExecutionStatus.ABORTED, aThread.getStatus() );
		assertEquals( "The replica status should be ABORTED", ExecutionStatus.ABORTED, replicaEntry.getStatus() );
	}


	public void testGetReplica() {

		ReplicaEntry r1 = new ReplicaEntry( 0, 0, 1, FakeTaskSpec.getFakeTaskSpec() );

		assertEquals( "The replicas should be the same.", replicaEntry, aThread.getReplica() );
		assertFalse( "The replicas should not be the same.", r1.equals( aThread.getReplica() ) );

	}


	public void testGetName() {

		assertEquals( "The names should be the same.", "a replica executor thread", aThread.getName() );
		assertFalse( "The names should not be the same.", "a replica executor Thread".equals( aThread.getName() ) );

	}
}
