/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Created at 26/04/2006 */

package org.ourgrid.mygrid.scheduler;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.id.GumID;

/**
 * This entity is responsible to maintain the <b>Black List</b>, the list of
 * <code>Gum</code>s that will not eb assigned to a <code>Request</code>
 * after some failures. The <code>Black List Manager</code> will follow some
 * rules as described bellow :<br>
 * <br>
 * Dentro de um mesmo job, uma mquina poder falhar n vezes, onde n = min(T,
 * |J|) e T = definido pelo usurio.
 * <ul>
 * <li>We consider as failures any reason by which a certain replica has failed
 * execept UnavailableGum</li>
 * <li>A given machine can only fail executing one replica of a given task</li> *
 * <li>A machine can only fail n times executing tasks of a given job, where n =
 * min( |J|, T) ), | J | is the number of tasks in the job and T is the user
 * configurable number of fails that can happen during the use of a machine by a
 * job</li>
 * </ul>
 */
public class BlackListManager {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( BlackListManager.class );

	/**
	 * The name says everything.
	 */
	private final int maxFailsPerJobToEnterInBL;

	/**
	 * Map that maintains the blacklist.
	 */
	private Map<Integer,Set<BlackListEntry>> blackListMap;

	private Map<Integer,Integer> jobSizes;


	/**
	 * Cretes a new <code>EBBlackListManager</code>
	 */
	public BlackListManager() {

		this( Integer.parseInt( Configuration.getInstance().getProperty( MyGridConfiguration.PROP_MAX_BL_FAILS ) ) );
	}


	/**
	 * Cretes a new <code>EBBlackListManager</code>
	 * 
	 * @param maxFailsPerJobToEnterInBL Default maximum number of failues befor
	 *        the gum is blacklisted for a job.
	 */
	public BlackListManager( int maxFailsPerJobToEnterInBL ) {

		this.maxFailsPerJobToEnterInBL = maxFailsPerJobToEnterInBL;
		this.blackListMap = new HashMap<Integer,Set<BlackListEntry>>();
		this.jobSizes = new HashMap<Integer,Integer>();
		LOG.debug( "Black List Manager created" );
	}


	/**
	 * Adds an entry to BlackList registry.
	 * 
	 * @param jobID The request id thats this black list is associated to
	 * @param gumID The "Black Listed" GumID
	 * @return true if the GumID is successfully associated with this request id
	 *         or false if the Gum was already associted.
	 */
	public int executionFailed( int jobID, GumID gumID, int taskID ) {

		Set<BlackListEntry> blackList = this.blackListMap.get( jobID );
		BlackListEntry entry = getEntry( blackList, gumID );
		if ( entry == null ) {
			entry = new BlackListEntry( gumID, this.jobSizes.get( jobID ) );
			blackList.add( entry );
		}

		entry.taskFail( taskID );
		LOG.info( "Gum <" + gumID + "> has been entered into black list of job " + jobID
				+ " and will no longer run any replica of the task <" + taskID + ">" );
		return entry.getNumberOfRemainingFailedExecutions();
	}


	/**
	 * Used in tests.
	 */
	protected Set<BlackListEntry> getBlackListedMachines( int jobID ) {

		return this.blackListMap.get( jobID );
	}


	public int getMaxBLFails() {

		return this.maxFailsPerJobToEnterInBL;
	}


	/**
	 * The number of remaining failed task executions to this Gum enter in the
	 * job black list.
	 * 
	 * @param jobID The identification of the Job
	 * @param gumID The identification of the Gum
	 * @return The number of remaining failed task executions (of this job) to
	 *         this Gum enter in the job black list.
	 */
	public int getRemainingBLFails( int jobID, GumID gumID ) {

		Set<BlackListEntry> blackList = this.blackListMap.get( jobID );
		if ( blackList != null ) {
			BlackListEntry entry = getEntry( blackList, gumID );
			if ( entry != null ) {
				return entry.getNumberOfRemainingFailedExecutions();
			}
		}
		return -1;
	}


	private BlackListEntry getEntry( Set<BlackListEntry> blackList, GumID gumID ) {

		BlackListEntry returnedEntry = null;
		if ( blackList != null ) {
			for ( BlackListEntry entry : blackList ) {

				if ( entry.getGumID().equals( gumID ) ) {
					returnedEntry = entry;
					break;
				}
			}
		}

		return returnedEntry;
	}


	/**
	 * Returns true if no replicas of this task has been failed, false
	 * otherwise.
	 * 
	 * @param jobID The identification of the Job
	 * @param gumID The identification of the Gum
	 * @param taskID The identification of the Task
	 * @return True if no replicas of this task has been failed, false
	 *         otherwise.
	 */
	public boolean canThisGumRunThisTask( int jobID, GumID gumID, int taskID ) {

		Set<BlackListEntry> blackList = this.blackListMap.get( jobID );
		if ( blackList != null ) {
			BlackListEntry entry = getEntry( blackList, gumID );
			if ( entry != null ) {
				return (!entry.getTaskIDs().contains( taskID )) && (entry.getNumberOfRemainingFailedExecutions() > 0);
			}
		}
		return true;
	}


	/**
	 * Init data structures to manage the black list of this job. A call in
	 * <code>executionFailed</code> method can result in a
	 * <code>NullPointerException</code> if the the data structures was not
	 * already initiated.
	 * 
	 * @param jobId The identification of the job
	 * @param jobSize The number of Tasks of the job
	 */
	public void initiateBlackListForJob( int jobId, int jobSize ) {

		this.blackListMap.put( jobId, new HashSet<BlackListEntry>() );
		this.jobSizes.put( jobId, Math.min( this.maxFailsPerJobToEnterInBL, jobSize ) );
		LOG.debug( "Black List has been initiated for the job <" + jobId + "> with " + jobSize + " tasks" );
	}


	/**
	 * Return true if the Gum has had failed min( jobSize, maxBLFails ) tasks of
	 * the job, false otherwise.
	 * 
	 * @param jobId The identification of the Job
	 * @param gumID The identification of the Gum
	 * @return True if the Gum has had failed min( jobSize, maxBLFails ) tasks
	 *         of the job, false otherwise.
	 */
	public boolean isGumBlacklistedForEntireJob( int jobId, GumID gumID ) {

		Set<BlackListEntry> blackList = this.blackListMap.get( jobId );

		BlackListEntry entry = getEntry( blackList, gumID );

		if ( entry != null ) {
			return (entry.getNumberOfRemainingFailedExecutions() < 1);
		}
		return false;
	}
}
