/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * This is the Data Discovery Entry class. It represents a single entry in the
 * Data Discovery lookup table..
 */

public class DataDiscoveryEntry implements Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/**
	 * This represents the digest of a transfered file. Is used as the primary
	 * key
	 */
	private String digest;

	/** This represents the size in bytes of a file within an entry record */
	private long size;

	/**
	 * This is a sub-table used to map the relationship between file
	 * destinations and the date they were last contacted
	 */
	private HashMap<String,Long> location = new HashMap<String,Long>();


	/*
	 * Empty Constructor
	 */
	public DataDiscoveryEntry() {

	}


	/**
	 * Creates a new Data Discovery entry
	 * 
	 * @param fileDigest Digest of a transfered file. Used as primary key.
	 * @param fileSize The size in bytes of the file transfered.
	 * @param fileLocation The full path (file://site/remote_path) of the file's
	 *        destination
	 * @param expireDate The date at which the location information is no longer
	 *        valid. Not yet used.
	 */
	public DataDiscoveryEntry( String fileDigest, long fileSize, String fileLocation, long expireDate ) {

		this.digest = fileDigest;
		this.size = fileSize;
		this.location.put( fileLocation, new Long( expireDate ) );
	}


	/**
	 * Assigns the digest value for this entry
	 * 
	 * @param fileDigest Digest of a transfered file. Used as primary key.
	 */
	public void setDigest( String fileDigest ) {

		this.digest = fileDigest;
	}


	/**
	 * Assigns the file size value for this entry
	 * 
	 * @param fileSize The size in bytes of the file transfered.
	 */
	public void setSize( long fileSize ) {

		this.size = fileSize;
	}


	/**
	 * Adds a valid location and time-stamp to the locations sub-table
	 * 
	 * @param fileLocation The full path (file://site/remote_path) of the file's
	 *        destination
	 * @param expireDate The date at which the location information is no longer
	 *        valid. Not yet used.
	 */
	public void setLocation( String fileLocation, long expireDate ) {

		location.put( fileLocation, new Long( expireDate ) );
	}


	/** Returns an entry's file digest */
	public String getDigest() {

		return this.digest;
	}


	/** Returns an enttry's file size */
	public long getSize() {

		return this.size;
	}


	/** Returns an entry's locations sub-table */
	public Map<String,Long> getLocations() {

		return this.location;
	}


	/** Formats an entry as a string */
	public String toString() {

		StringBuffer message = new StringBuffer();
		Iterator<String> it = this.location.keySet().iterator();

		message.append( "File " + this.digest + " of size " + this.size + " is at:\n" );
		while ( it.hasNext() ) {
			String key = it.next();
			message.append( "     [" + key + " - " + this.location.get( key ) + "]\n" );
		}

		return message.toString();
	}
}
