/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.exception.UnavailableGumException;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.util.StreamCloner;
import org.ourgrid.mygrid.replicaexecutor.EBReplicaExecutorFacade;
import org.ourgrid.mygrid.scheduler.exception.JobCouldNotBeCancelledException;
import org.ourgrid.mygrid.scheduler.exception.JobNotFoundException;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManager;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.TaskEntry;

/**
 * This class should be extended by all scheduling heuristics used by the
 * Scheduler module. It has the basic methods that should be implemented by
 * these heuristics.
 */
abstract class EBAbstractSchedulingHeuristic implements EBSchedulingHeuristic {

	protected JobManager jobManager;

	protected GridManager gridManager;

	private EBReplicaExecutorFacade ebReplicaExecutorFacade;


	public EBAbstractSchedulingHeuristic( JobManager jobManager, GridManager ebGridManager,
											EBReplicaExecutorFacade ebReplicaExecutorFacade ) {

		this.jobManager = jobManager;
		this.gridManager = ebGridManager;
		this.ebReplicaExecutorFacade = ebReplicaExecutorFacade;
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.EBSchedulingHeuristic#getEBGridManager()
	 */
	public final GridManager getEBGridManager() {

		return gridManager;
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.EBSchedulingHeuristic#getEBJobManager()
	 */
	public final JobManager getEBJobManager() {

		return jobManager;
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.EBSchedulingHeuristic#addJob(int,org.ourgrid.common.spec.JobSpec)
	 */
	public void addJob( int jobId, JobSpec jobSpec ) {

		String requirements = jobSpec.getRequirements();
		gridManager.initiateBlackListForJob( jobId, jobSpec.getTaskSpecs().size() );
		gridManager.needGums( jobId, requirements, getNumberOfNeededGums( jobSpec ) );
		jobManager.addJob( jobId, jobSpec );
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.EBSchedulingHeuristic#newPeersHaveBeenSet()
	 */
	public void newPeersHaveBeenSet() {

		Collection<JobEntry> jobs = jobManager.jobList();
		Iterator<JobEntry> jobsIt = jobs.iterator();
		while ( jobsIt.hasNext() ) {
			JobEntry job = jobsIt.next();
			int avaliableTasksSize = job.getAvailableTasks().size();
			if ( avaliableTasksSize > 0 ) {
				gridManager.needGums( job.getId(), job.getRequirements(), job.getAvailableTasks().size() );
			}
		}
	}


	/**
	 * Returns the number of needed Gums according the Job specification.
	 * 
	 * @param jobSpec The Job specification.
	 * @return The number of Gums needed.
	 */
	public int getNumberOfNeededGums( JobSpec jobSpec ) {

		int maxReplicas = Integer.parseInt( Configuration.getInstance().getProperty(
				MyGridConfiguration.PROP_MAX_REPLICAS ) );
		return jobSpec.getTaskSpecs().size() * maxReplicas;
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.EBSchedulingHeuristic#schedule()
	 */
	public abstract boolean schedule();


	/**
	 * It clones the replica and sends it to be executed by the
	 * <code>ReplicaExecutor</code>. It also sets the replica state to
	 * 'Running' and sets the <code>GridMachine</code> as allocated.
	 * 
	 * @param replica The replica to be executed.
	 * @param gumClient The <code>GridMachine</code> to execute the replica
	 */
	public void executeReplica( ReplicaEntry replica, GumClient gumClient ) {

		jobManager.replicaRunning( replica, gumClient.getGumSpec().getGumID() );

		try {
			ReplicaEntry clonedReplica = StreamCloner.clone( replica );

			ebReplicaExecutorFacade.executeReplica( clonedReplica, gumClient );
			gridManager.allocate( gumClient.getGumSpec().getGumID(), replica );
			if ( isJobSatisfied( replica.getJobId() ) ) {
				gridManager.pauseRequest( replica.getJobId() );
			}
		} catch ( CloneNotSupportedException e ) {
			// This should never occur while executing. If it occurs,
			// it is a programming error. All objects within the replica
			// should implement the java.io.Serializable interface
			e.printStackTrace();
		}
	}


	private boolean isJobSatisfied( int jobId ) {

		final List<TaskEntry> availableTasks = jobManager.getJobById( jobId ).getAvailableTasks();
		return availableTasks.isEmpty();
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.EBSchedulingHeuristic#replicaFinished(ReplicaExecutorResult)
	 */
	public void replicaFinished( ReplicaExecutorResult result ) {

		jobManager.replicaFinished( result );

		final ReplicaEntry replica = result.getReplica();
		verifyIfGumIsNeeded( replica );
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.EBSchedulingHeuristic#replicaAborted(ReplicaExecutorResult)
	 */
	public void replicaAborted( ReplicaExecutorResult result ) {

		jobManager.replicaAborted( result );

		final ReplicaEntry replica = result.getReplica();
		verifyIfGumIsNeeded( replica );
	}


	private void verifyIfGumIsNeeded( ReplicaEntry replica ) {

		if ( hasJobEnded( replica.getJobId() ) ) {
			gridManager.cancelRequests( replica.getJobId() );
		} else {
			if ( isGumStillNeeded( replica.getGumID(), replica.getJobId() ) ) {
				gridManager.ready( replica.getGumID() );
			} else {
				gridManager.dispose( replica.getGumID() );
			}
		}
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.EBSchedulingHeuristic#replicaCanceled(ReplicaExecutorResult)
	 */
	public void replicaCanceled( ReplicaExecutorResult result ) {

		jobManager.replicaCanceled( result );
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.EBSchedulingHeuristic#replicaFailed(ReplicaExecutorResult)
	 *      By analyzing the cause of a task failure, it is possible to know
	 *      whether the failure has been provoked by an error in the application
	 *      or by a machine failure. In the first case, the grid machine has its
	 *      status set to READY, so it may be used again. In the latter case,
	 *      the gum entry is removed from the broker's Grid Manager.
	 */
	public void replicaFailed( ReplicaExecutorResult result ) {

		jobManager.replicaFailed( result );

		ReplicaEntry replica = result.getReplica();

		if ( hasJobEnded( result.getReplica().getJobId() ) ) {
			gridManager.cancelRequests( result.getReplica().getJobId() );
		} else {
			Throwable errorCause = result.getErrorCause();

			if ( errorCause instanceof UnavailableGumException ) {
				gridManager.deleteGum( replica.getGumID() );
			} else {
				if ( !gridManager.updateBlackList( replica.getJobId(), replica.getTaskId(), replica.getGumID() ) ) {
					if ( !isGumStillNeeded( replica.getGumID(), replica.getJobId() ) ) {
						gridManager.unwantedGum( replica.getGumID(), replica.getJobId() );
					} else {
						gridManager.ready( replica.getGumID() );
					}
				}
			}

			if ( !isJobSatisfied( replica.getJobId() ) ) {
				gridManager.resumeRequest( replica.getJobId() );
			}
		}
	}


	/**
	 * Since the Gum was not unwanted at the peer, we will verify is this Gum
	 * can still be used. If not it will be unwanted.
	 * 
	 * @param gumID ID of the Gum.
	 * @param jobID ID of the job it has been allocated to.
	 */
	private boolean isGumStillNeeded( GumID gumID, int jobID ) {

		for ( TaskEntry taskEntry : jobManager.getJobById( jobID ).getAvailableTasks() ) {
			if ( gridManager.canGumBeAllocatedToTask( gumID, jobID, taskEntry.getId() ) ) {
				return true;
			}
		}
		return false;
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.EBSchedulingHeuristic#cancelJob(int)
	 */
	public void cancelJob( int jobId ) throws JobNotFoundException, JobCouldNotBeCancelledException {

		jobManager.setCanceled( jobId );
		ebReplicaExecutorFacade.cancelReplicasOfJob( jobId );

		/*
		 * the following call should be made here so that the peer receives a
		 * Gump.finishRequestGums even if there are no running replicas of the
		 * job
		 */
		gridManager.cancelRequests( jobId );
	}


	/**
	 * Cancels the requests for grid machines if the Job is in a final state (
	 * Failed, Aborted, Cancelled or Finished ).
	 * 
	 * @param result The <code>ReplicaExecutorResult</code> that contains all
	 *        information about the replica execution.
	 */
	private boolean hasJobEnded( int jobID ) {

		ExecutionStatus jobState = jobManager.getJobState( jobID );
		return jobState.equals( ExecutionStatus.FAILED ) || jobState.equals( ExecutionStatus.CANCELLED )
				|| jobState.equals( ExecutionStatus.FINISHED );

	}


	/**
	 * Gets the list of tasks to schedule.
	 * 
	 * @return The list of tasks to schedule
	 */
	protected List<TaskEntry> getTasksToScheduleQueue() {

		List<TaskEntry> tasksToScheduleQueue = new LinkedList<TaskEntry>();
		Collection<JobEntry> jobs = jobManager.jobList();

		for ( JobEntry job : jobs ) {

			Collection<TaskEntry> tasks = job.getAvailableTasks();
			tasksToScheduleQueue.addAll( tasks );
		}
		Collections.sort( tasksToScheduleQueue );

		return tasksToScheduleQueue;
	}


	/**
	 * Returns a String representing the given job requirements.
	 * 
	 * @param jobId The identification of the job to get the requirements
	 * @return The job requirements.
	 */
	protected String getRequirementsOfJob( int jobId ) {

		Collection<JobEntry> jobs = jobManager.jobList();
		Iterator<JobEntry> jobsIt = jobs.iterator();
		while ( jobsIt.hasNext() ) {
			JobEntry job = jobsIt.next();
			if ( job.getId() == jobId ) {
				return job.getRequirements();
			}
		}
		return null;
	}

}
