/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.rmi.RemoteException;
import java.util.Collection;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.event.Event;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.NotifyEvent;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.event.ShutdownRequestEvent;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.gump.GumpClient;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.security.MGSecureClientSocketFactory;
import org.ourgrid.common.security.MGSecureServerSocketFactory;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.PeerSpec;
import org.ourgrid.mygrid.replicaexecutor.EBReplicaExecutorFacade;
import org.ourgrid.mygrid.scheduler.event.AddJobRequestEvent;
import org.ourgrid.mygrid.scheduler.event.CancelJobRequestEvent;
import org.ourgrid.mygrid.scheduler.event.CleanFinishedJobsRequestEvent;
import org.ourgrid.mygrid.scheduler.event.GetGumEntriesRequestEvent;
import org.ourgrid.mygrid.scheduler.event.GetGumpSpecsRequestEvent;
import org.ourgrid.mygrid.scheduler.event.GumIsDeadNotifyEvent;
import org.ourgrid.mygrid.scheduler.event.HereIsGumNotifyEvent;
import org.ourgrid.mygrid.scheduler.event.JobListRequestEvent;
import org.ourgrid.mygrid.scheduler.event.PeerAliveNotifyEvent;
import org.ourgrid.mygrid.scheduler.event.PeerIsDeadNotifyEvent;
import org.ourgrid.mygrid.scheduler.event.RemoveJobRequestEvent;
import org.ourgrid.mygrid.scheduler.event.ReplicaAbortedNotifyEvent;
import org.ourgrid.mygrid.scheduler.event.ReplicaCanceledNotifyEvent;
import org.ourgrid.mygrid.scheduler.event.ReplicaFailedNotifyEvent;
import org.ourgrid.mygrid.scheduler.event.ReplicaFinishedNotifyEvent;
import org.ourgrid.mygrid.scheduler.event.SetPeersRequestEvent;
import org.ourgrid.mygrid.scheduler.event.WaitForJobRequestEvent;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerEntry;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerListenerRemote;
import org.ourgrid.mygrid.scheduler.gump.GumpClientImpl;
import org.ourgrid.mygrid.scheduler.gump.GumpManager;
import org.ourgrid.mygrid.scheduler.gump.GumpManagerImpl;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManagerListenerRemote;

/**
 * This class is responsible for queueing events in the
 * <code>SchedulerEventProcessor</code>. All communication with objects
 * inside the <code>SchedulerEventProcessor</code> is made through this
 * facade.
 */
public class EBSchedulerFacade {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( EBSchedulerFacade.class );

	/** The EBReplicaExecutorFacade. */
	private EBReplicaExecutorFacade ebReplicaExecutorFacade;

	/** The <code>SchedulerEventProcessor</code>. */
	protected SchedulerEventEngine schedulerEventProcessor;

	/** The next job identification. */
	private int nextJobId;


	/**
	 * The constructor.
	 */
	public EBSchedulerFacade() {

		this.ebReplicaExecutorFacade = null;
		LOG.debug( "The EBSchedulerFacade has been successfully created." );

		if ( Configuration.getInstance().isEnabled( MyGridConfiguration.PROP_PERSISTJOBID ) ) {
			nextJobId = getLastJobId();
		} else {
			nextJobId = 0;
		}

	}


	private int getLastJobId() {

		BufferedReader reader = null;
		try {
			reader = new BufferedReader( new FileReader( MyGridConfiguration.PROP_JOBCOUNTERFILE ) );
			return Integer.parseInt( reader.readLine() );

		} catch ( Exception e ) {
			return 0;
		} finally {
			if ( reader != null ) {
				try {
					reader.close();
				} catch ( IOException e ) {
					LOG.debug( "stream not closed" );
				}
			}
		}
	}


	/**
	 * Starts the <code>SchedulerEventProcessor</code>
	 */
	public void startEventProcessor() {

		schedulerEventProcessor.startProcessing();
	}


	/**
	 * Configures the <code>EBSchedulerFacade</code>. Creates the RMI objects
	 * that will be accessed by other modules and creates the scheduler event
	 * engine.
	 */
	public void config( EBReplicaExecutorFacade ebReplicaExecutorFacade ) throws ConfigException, RemoteException {

		this.ebReplicaExecutorFacade = ebReplicaExecutorFacade;
		Scheduler scheduler = null;
		GumpManager gumpManager = null;
		GumpClient gumpClient = null;

		if ( Configuration.getInstance().isSecurityEnabled() ) {

			int securePort = Integer
					.parseInt( Configuration.getInstance().getProperty( Configuration.PROP_SECURE_PORT ) );
			MGSecureServerSocketFactory secureServerSocketFactory = MGSecureServerSocketFactory.getInstance();
			MGSecureClientSocketFactory secureClientSocketFactory = MGSecureClientSocketFactory.getInstance();

			scheduler = new SchedulerImpl( this, secureClientSocketFactory, secureServerSocketFactory, securePort );
			gumpManager = new GumpManagerImpl( this, secureClientSocketFactory, secureServerSocketFactory, securePort );
			gumpClient = new GumpClientImpl( this, secureClientSocketFactory, secureServerSocketFactory, securePort );
		} else {
			scheduler = new SchedulerImpl( this );
			gumpManager = new GumpManagerImpl( this );
			gumpClient = new GumpClientImpl( this );
		}

		this.schedulerEventProcessor = createSchedulerEventEngine( scheduler, gumpManager, gumpClient,
				this.ebReplicaExecutorFacade );
	}


	protected SchedulerEventEngine createSchedulerEventEngine( Scheduler scheduler, GumpManager gumpManager,
																GumpClient gumpClient,
																EBReplicaExecutorFacade ebReplicaExecutorFacade )
		throws ConfigException {

		return new SchedulerEventEngine( scheduler, gumpManager, gumpClient, ebReplicaExecutorFacade, this );
	}


	/**
	 * Provides the list of <code>Job</code>s.
	 * 
	 * @param newListener TODO
	 * @return An <code>EventQueue</code> were the list of jobs will be put.
	 */
	public EventQueue<ResponseEvent<Collection<JobEntry>>> jobList( JobManagerListenerRemote newListener ) {

		EventQueue<ResponseEvent<Collection<JobEntry>>> responseQueue = new EventQueue<ResponseEvent<Collection<JobEntry>>>();

		JobListRequestEvent jobListRequestEvent = new JobListRequestEvent( schedulerEventProcessor.getJobManager(),
				newListener, responseQueue );
		schedulerEventProcessor.putEvent( jobListRequestEvent );

		return responseQueue;
	}


	/**
	 * Provides an <code>EventQueue</code> that will only have an element when
	 * the specified <code>Job</code> reaches a terminal state.
	 * 
	 * @param jobId The <code>Job</code> identification.
	 * @return A response <code>EventQueue</code> that will receive an event
	 *         when the specified <code>Job</code> reaches a terminal state.
	 */
	public EventQueue<Event> waitForJob( int jobId ) {

		EventQueue<Event> responseQueue = new EventQueue<Event>();
		JobMonitor jobMonitor = new JobMonitor( responseQueue, jobId );
		WaitForJobRequestEvent waitForJobRequestEvent = new WaitForJobRequestEvent( schedulerEventProcessor
				.getJobManager(), jobMonitor );
		schedulerEventProcessor.putEvent( waitForJobRequestEvent );

		return responseQueue;
	}


	/**
	 * Cancels a specified <code>Job</code>.
	 * 
	 * @param jobId The <code>Job</code> identification.
	 */
	public EventQueue<NotifyEvent> cancelJob( int jobId ) {

		EventQueue<NotifyEvent> responseQueue = new EventQueue<NotifyEvent>();
		CancelJobRequestEvent cancelJobRequestEvent = new CancelJobRequestEvent(
				schedulerEventProcessor.getHeuristic(), jobId, responseQueue );
		schedulerEventProcessor.putEvent( cancelJobRequestEvent );

		return responseQueue;

	}


	/**
	 * Clean the list of terminated jobs
	 * 
	 * @return the response queue
	 */
	public EventQueue<NotifyEvent> cleanFinishedJobs() {

		EventQueue<NotifyEvent> responseQueue = new EventQueue<NotifyEvent>();
		CleanFinishedJobsRequestEvent cleanFinishedJobsRequestEvent = new CleanFinishedJobsRequestEvent(
				schedulerEventProcessor.getJobManager(), responseQueue );
		schedulerEventProcessor.putEvent( cleanFinishedJobsRequestEvent );

		return responseQueue;
	}


	/**
	 * Remove the information about the specified job
	 * 
	 * @param jobId
	 * @return the response queue
	 */
	public EventQueue<ResponseEvent> removeJob( int jobId ) {

		EventQueue<ResponseEvent> responseQueue = new EventQueue<ResponseEvent>();
		RemoveJobRequestEvent removeJobRequestEvent = new RemoveJobRequestEvent( jobId, schedulerEventProcessor
				.getJobManager(), responseQueue );
		schedulerEventProcessor.putEvent( removeJobRequestEvent );

		return responseQueue;

	}


	/**
	 * Adds a new <code>Job</code>.
	 * 
	 * @param newJobSpec The <code>Job</code> specification.
	 * @return An EventQueue that will contain the <code>Job</code>
	 *         identification.
	 */
	public int addJob( JobSpec newJobSpec ) {

		int jobId = this.getNextJobId();
		AddJobRequestEvent addJobRequestEvent = new AddJobRequestEvent( schedulerEventProcessor.getHeuristic(), jobId,
				newJobSpec );
		schedulerEventProcessor.putEvent( addJobRequestEvent );

		return jobId;
	}


	/**
	 * Delivers an available grid machine to the scheduler
	 * 
	 * @param gumClient a reference to the grid machine.
	 * @param requestid identification of the request that will receive the gum
	 * @param peerBrokerAccessID an id to the grid machine provider (peer) to
	 *        which the machine will be given back after execution
	 */
	public void hereIsGum( GumClient gumClient, long requestid, ObjectID peerBrokerAccessID ) {

		HereIsGumNotifyEvent hereIsGuMNotifyEvent = new HereIsGumNotifyEvent( schedulerEventProcessor.getGridManager(),
				gumClient, requestid, peerBrokerAccessID );
		schedulerEventProcessor.putEvent( hereIsGuMNotifyEvent );
	}


	/**
	 * Notifies that a task replica has finished.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	public void replicaFinished( ReplicaExecutorResult result ) {

		ReplicaFinishedNotifyEvent replicaFinishedNotifyEvent = new ReplicaFinishedNotifyEvent( schedulerEventProcessor
				.getHeuristic(), result );
		schedulerEventProcessor.putEvent( replicaFinishedNotifyEvent );
	}


	/**
	 * Notifies that a task replica has failed.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	public void replicaFailed( ReplicaExecutorResult result ) {

		ReplicaFailedNotifyEvent replicaFailedNotifyEvent = new ReplicaFailedNotifyEvent( schedulerEventProcessor
				.getHeuristic(), result );
		schedulerEventProcessor.putEvent( replicaFailedNotifyEvent );
	}


	/**
	 * Notifies that a task replica has been aborted.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	public void replicaAborted( ReplicaExecutorResult result ) {

		ReplicaAbortedNotifyEvent replicaAbortedNotifyEvent = new ReplicaAbortedNotifyEvent( schedulerEventProcessor
				.getHeuristic(), result );
		schedulerEventProcessor.putEvent( replicaAbortedNotifyEvent );
	}


	/**
	 * Notifies that a task replica has canceled.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	public void replicaCanceled( ReplicaExecutorResult result ) {

		ReplicaCanceledNotifyEvent replicaCanceledNotifyEvent = new ReplicaCanceledNotifyEvent( schedulerEventProcessor
				.getHeuristic(), result );
		schedulerEventProcessor.putEvent( replicaCanceledNotifyEvent );
	}


	/**
	 * Notifies that a GridMachine is dead.
	 * 
	 * @param requestId The request identification
	 * @param gumID The GridMachine that is dead
	 */
	public void gumIsDead( long requestId, GumID gumID ) {

		if ( schedulerEventProcessor != null ) {
			GumIsDeadNotifyEvent gumIsDeadNotifyEvent = new GumIsDeadNotifyEvent( schedulerEventProcessor
					.getGridManager(), requestId, gumID );
			schedulerEventProcessor.putEvent( gumIsDeadNotifyEvent );
		}
	}


	/**
	 * Gets all <code>GridMachine</code> s from the EBGridManager.
	 * 
	 * @param newListener TODO
	 * @return An <code>EventQueue</code> containing a
	 *         <code>ResponseEvent</code> whose response value is a list of
	 *         <code>GridMachineExtended</code> objects.
	 */
	public EventQueue<ResponseEvent<Collection<GridManagerEntry>>> getGumEntries( GridManagerListenerRemote newListener ) {

		EventQueue<ResponseEvent<Collection<GridManagerEntry>>> eq = new EventQueue<ResponseEvent<Collection<GridManagerEntry>>>();

		schedulerEventProcessor.putEvent( new GetGumEntriesRequestEvent( schedulerEventProcessor.getGridManager(),
				newListener, eq ) );

		return eq;
	}


	public EventQueue<ResponseEvent<Collection<GumSpec>>> getGumSpecs() {

		EventQueue<ResponseEvent<Collection<GumSpec>>> eq = new EventQueue<ResponseEvent<Collection<GumSpec>>>();

		schedulerEventProcessor.putEvent( new GetGumSpecsRequestEvent( schedulerEventProcessor.getGridManager(), eq ) );

		return eq;
	}


	/**
	 * Gets all GridMachineProvider's specifications from the EBGridManager.
	 * 
	 * @return An <code>EventQueue</code> containing a
	 *         <code>ResponseEvent</code> whose response value is a list of
	 *         <code>GuMPSpec</code> objects.
	 */
	public EventQueue<ResponseEvent<Collection<PeerEntry>>> getPeerEntries( GridManagerListenerRemote newListener ) {

		EventQueue<ResponseEvent<Collection<PeerEntry>>> eq = new EventQueue<ResponseEvent<Collection<PeerEntry>>>();

		schedulerEventProcessor.putEvent( new GetGumpSpecsRequestEvent( schedulerEventProcessor.getGridManager(),
				newListener, eq ) );

		return eq;
	}


	public void removeJobManagerListener( JobManagerListenerRemote listener ) {

		schedulerEventProcessor.putEvent( new RemoveJobManagerListenerRequestEvent( schedulerEventProcessor
				.getJobManager(), listener ) );
	}


	public void removeGridManagerListener( GridManagerListenerRemote listener ) {

		schedulerEventProcessor.putEvent( new RemoveGridManagerListenerRequestEvent( schedulerEventProcessor
				.getGridManager(), listener ) );
	}


	/**
	 * Sets all GridMachineProviders to be used by the Scheduler.
	 * 
	 * @param peerSpecs The new list of GridMachineProviders specifications
	 */
	public void setPeers( Collection<PeerSpec> peerSpecs ) {

		schedulerEventProcessor.putEvent( new SetPeersRequestEvent( schedulerEventProcessor.getHeuristic(),
				schedulerEventProcessor.getGridManager(), peerSpecs ) );
	}


	/**
	 * Notifies EBGridManager that the peer <code>PeerSpec</code> is alive.
	 * 
	 * @param peerSpec
	 */
	public void peerAlive( PeerSpec peerSpec ) {

		schedulerEventProcessor.putEvent( new PeerAliveNotifyEvent( schedulerEventProcessor.getHeuristic(),
				schedulerEventProcessor.getGridManager(), peerSpec ) );
	}


	/**
	 * Notifies EBGridManager that the peer <code>PeerSpec</code> is dead.
	 * 
	 * @param peerSpec
	 */
	public void peerIsDead( PeerSpec peerSpec ) {

		schedulerEventProcessor
				.putEvent( new PeerIsDeadNotifyEvent( schedulerEventProcessor.getGridManager(), peerSpec ) );
	}


	/**
	 * Shuts down the facade.
	 */
	public EventQueue<ShutdownResponseEvent> shutdown() {

		try {
			File file = new File( MyGridConfiguration.PROP_JOBCOUNTERFILE );

			if ( file.exists() ) {
				file.delete();
			}

			PrintWriter write = new PrintWriter( new FileOutputStream( file ) );

			write.write( Integer.toString( nextJobId ) );
			write.flush();
			write.close();
		} catch ( FileNotFoundException e ) {
			LOG.error( "Job ID file not found." );
		}

		EventQueue<ShutdownResponseEvent> responseQueue = new EventQueue<ShutdownResponseEvent>();
		ShutdownRequestEvent shutdownRequestEvent = new ShutdownRequestEvent( responseQueue, schedulerEventProcessor );
		if ( schedulerEventProcessor != null && schedulerEventProcessor.isAlive() ) {
			schedulerEventProcessor.putEvent( shutdownRequestEvent );
		} else {
			responseQueue.put( new ShutdownResponseEvent() );
		}

		return responseQueue;
	}


	/**
	 * Returns the next job identification.
	 * 
	 * @return The job identification
	 */
	private synchronized int getNextJobId() {

		this.nextJobId++;
		return this.nextJobId;
	}


	public SchedulerEventEngine getEventProcessor() {

		return schedulerEventProcessor;
	}
}
