/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler;

import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.mygrid.scheduler.exception.JobCouldNotBeCancelledException;
import org.ourgrid.mygrid.scheduler.exception.JobNotFoundException;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManager;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManager;

/**
 * Interface to be implemented by the scheduling heuristics. It defines the
 * basic methods that these heuristics should implement.
 */
public interface EBSchedulingHeuristic {

	/**
	 * Adds a new <code>Job</code>.
	 * 
	 * @param jobId The job identification.
	 * @param jobSpec The new Job specification.
	 */
	public void addJob( int jobId, JobSpec jobSpec );


	/**
	 * Notifies the heuristic that new <code>GridMachineProvider</code>s have
	 * been set. <code>GridMachine</code>s should be requested to the
	 * <code>GridMachineProvider</code>s.
	 */
	public void newPeersHaveBeenSet();


	/**
	 * Returns the <code>EBGridManager</code>.
	 * 
	 * @return The <code>EBGridManager</code>.
	 */
	public GridManager getEBGridManager();


	/**
	 * Returns the <code>EBJobManager</code>.
	 * 
	 * @return The <code>EBJobManager</code>.
	 */
	public JobManager getEBJobManager();


	/**
	 * Notifies that a task replica has finished.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code> that contains all
	 *        information about the replica execution.
	 */
	public void replicaFinished( ReplicaExecutorResult result );


	/**
	 * Notifies that a task replica has failed.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code> that contains all
	 *        information about the replica execution.
	 */
	public void replicaFailed( ReplicaExecutorResult result );


	/**
	 * Notifies that a task replica has been aborted.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code> that contains all
	 *        information about the replica execution.
	 */
	public void replicaAborted( ReplicaExecutorResult result );


	/**
	 * Notifies that a task replica has canceled.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code> that contains all
	 *        information about the replica execution.
	 */
	public void replicaCanceled( ReplicaExecutorResult result );


	/**
	 * Schedules a task replica to a machine. Important: only one task replica
	 * should be scheduled in call to this method..
	 */
	public boolean schedule();


	/**
	 * Cancels the given job and put an Event indicating that in the
	 * responseQueue.
	 * 
	 * @param jobId The job identification
	 * @throws JobNotFoundException if asked to cancel an invalid job.
	 * @throws JobCouldNoBeCancelledException if asked to cancel a job that
	 *         cannot be cancelled anymore.
	 */
	public void cancelJob( int jobId ) throws JobNotFoundException, JobCouldNotBeCancelledException;
}
