/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.LinkedList;

import org.ourgrid.common.gump.Gump;

/**
 * This class was created to keep a pair of Job and Request used to cancel
 * requests and simplify as well as downsize the code of
 * <code>EBGridManager</code>
 */
public class JobRequestEntry {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( JobRequestEntry.class );

	/** The job identification to which the request was made. */
	private int jobId;

	private long requestID;

	private Collection<Gump> peers;

	private boolean isPaused;


	/**
	 * Constructs a new Pair of Job and Request.
	 * 
	 * @param jobId The Job to which the request was made.
	 */
	public JobRequestEntry( int jobId, long requestID ) {

		this.jobId = jobId;
		this.peers = new LinkedList<Gump>();
		this.requestID = requestID;
		this.isPaused = false;
	}


	/**
	 * Adds a new pair of Request (by id) and respective GuMP.
	 * 
	 * @param gump The Grid Machine Provider to which the request was made.
	 */
	public void addPeer( Gump gump ) {

		this.peers.add( gump );
	}


	/**
	 * Cancels (calls noMoreGMs) the requests made for the Job.
	 */
	public void cancelRequests() {

		for ( Gump entry : peers ) {
			try {
				entry.finishRequestGums( this.requestID );
			} catch ( RemoteException e ) {
				LOG.error( "Could not contact peer to finishRequest " + requestID );
			}
		}
	}


	/**
	 * Returns a String with Job Id following toString of the reqs Map.
	 * 
	 * @return Job Id + Map of req and gmps.
	 */
	public String toString() {

		return "Job ID:" + this.jobId + " requestID: " + requestID;
	}


	public long getRequestID() {

		return requestID;
	}


	public Collection<Gump> getPeers() {

		return peers;
	}


	public void pauseRequest() {

		if ( !this.isPaused ) {
			for ( Gump entry : peers ) {
				try {
					entry.pauseRequestGums( this.requestID );
				} catch ( RemoteException e ) {
					LOG.error( "Could not contact peer to pause request " + requestID );
				}
			}
			this.isPaused = true;
		}
	}


	public void resumeRequest() {

		if ( this.isPaused ) {
			for ( Gump entry : peers ) {
				try {
					entry.resumeRequestGums( this.requestID );
				} catch ( RemoteException e ) {
					LOG.error( "Could not contact peer to resume request " + requestID );
				}
			}
			this.isPaused = false;
		}
	}


	public boolean isPaused() {

		return isPaused;
	}
}
