/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler;

import java.io.Serializable;

import org.ourgrid.common.executor.ExecutorResult;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.util.TimeDataGenerator;
import org.ourgrid.gridmachine.exception.FailCauseException;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;

/**
 * Description: This class represents the result of a replica execution.
 * 
 * @version 1.0 Created on 08/07/2004
 */
public class ReplicaExecutorResult implements Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/** Constant to indicate the array position for the 'INIT' result. */
	private static int INIT = 0;

	/** Constant to indicate the array position for the 'REMOTE' result. */
	private static int REMOTE = 1;

	/** Constant to indicate the array position for the 'FINAL' result. */
	private static int FINAL = 2;

	/** The array of results. */
	private ExecutorResult results[];

	/** The throwable thrown when this replica fails. */
	private FailCauseException errorCause;

	/** The replica executed. */
	private ReplicaEntry replica;

	/**
	 * Init phase execution time.
	 */
	private TimeDataGenerator initData;

	/**
	 * Remote phase execution time.
	 */
	private TimeDataGenerator remoteData;

	/**
	 * Final phase execution time.
	 */
	private TimeDataGenerator finalData;

	private int guMID;


	/**
	 * The constructor.
	 */
	public ReplicaExecutorResult() {

		results = new ExecutorResult[ 3 ];
	}


	/**
	 * The constructor.
	 * 
	 * @param replica The replica executed
	 */
	public ReplicaExecutorResult( ReplicaEntry replica ) {

		this.replica = replica;
		results = new ExecutorResult[ 3 ];
	}


	/**
	 * Sets the executed replica.
	 * 
	 * @param replica The executed replica
	 */
	public void setReplica( ReplicaEntry replica ) {

		this.replica = replica;
	}


	/**
	 * Return the executed replica.
	 * 
	 * @return The executed replica
	 */
	public ReplicaEntry getReplica() {

		return replica;
	}


	/**
	 * Returns an array containing the results of all replica phases: Init,
	 * Input, Remote, Output and Final.
	 * 
	 * @return The array of results
	 */
	public ExecutorResult[ ] getResults() {

		return results;
	}


	/**
	 * Returns the 'INIT' result.
	 * 
	 * @return The 'INIT' result
	 */
	public ExecutorResult getInitResult() {

		return results[INIT];
	}


	/**
	 * Sets the 'INIT' result.
	 * 
	 * @param initResult The 'INIT' result
	 */
	public void setInitResult( ExecutorResult initResult ) {

		results[INIT] = initResult;
	}


	/**
	 * Returns the 'REMOTE' result.
	 * 
	 * @return The 'REMOTE' result
	 */
	public ExecutorResult getRemoteResult() {

		return results[REMOTE];
	}


	/**
	 * Sets the 'REMOTE' result.
	 * 
	 * @param remoteResult The 'REMOTE' result
	 */
	public void setRemoteResult( ExecutorResult remoteResult ) {

		results[REMOTE] = remoteResult;
	}


	/**
	 * Returns the 'FINAL' result.
	 * 
	 * @return The 'FINAL' result
	 */
	public ExecutorResult getFinalResult() {

		return results[FINAL];
	}


	/**
	 * Sets the 'FINAL' result.
	 * 
	 * @param finalResult The 'FINAL' result
	 */
	public void setFinalResult( ExecutorResult finalResult ) {

		results[FINAL] = finalResult;
	}


	/**
	 * Returns the error that caused this replica to fail.
	 * 
	 * @return the error cause.
	 */
	public FailCauseException getErrorCause() {

		return errorCause;
	}


	/**
	 * Sets the error that caused this replica to fail.
	 * 
	 * @param errorCause The error cause to set.
	 */
	public void setErrorCause( FailCauseException errorCause ) {

		this.errorCause = errorCause;
	}


	/**
	 * @param initData
	 * @param remoteData
	 * @param finalData
	 */
	public void setExecutionTimes( TimeDataGenerator initData, TimeDataGenerator remoteData, TimeDataGenerator finalData ) {

		this.initData = initData;
		this.remoteData = remoteData;
		this.finalData = finalData;

	}


	public TimeDataGenerator getInitData() {

		return initData;
	}


	public TimeDataGenerator getRemoteData() {

		return remoteData;
	}


	public TimeDataGenerator getFinalData() {

		return finalData;
	}


	public void setGumClient( GumClient gum ) {

		this.guMID = gum.hashCode();
	}


	public boolean isSameGuMUsedByReplica( GumClient gum ) {

		return gum.hashCode() == this.guMID;
	}


	@Override
	public boolean equals( Object o ) {

		if ( o != null && o instanceof ReplicaExecutorResult ) {
			ReplicaExecutorResult result = (ReplicaExecutorResult) o;
			return replica.equals( result.getReplica() );
		}

		return false;
	}

}
