/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.Collection;

import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.mygrid.scheduler.exception.JobCannotBeRemovedException;
import org.ourgrid.mygrid.scheduler.exception.JobCouldNotBeCancelledException;
import org.ourgrid.mygrid.scheduler.exception.JobNotFoundException;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerEntry;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerListenerRemote;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManagerListenerRemote;

/**
 * The Scheduler interface that should be implemented by a remote object to
 * provide access to the Scheduler from a different Java Virtual Machine.
 */
public interface Scheduler extends Remote {

	/**
	 * Gives access to the list of jobs on the scheduler.
	 * 
	 * @param newListener TODO
	 * @return The list within all the jobs on the scheduler
	 */
	public Collection<JobEntry> jobList( JobManagerListenerRemote newListener ) throws RemoteException;


	/**
	 * Puts the calling thread to wait until a certain job is finished.
	 * 
	 * @param jobId The identifier of the job the thread must wait for. This
	 *        identifier is obtained when the job is added to the scheduler.
	 * @throws JobNotFoundException if there is no job with this id
	 * @throws RemoteException if any exception occurs while contacting with
	 *         this remote object.
	 */
	public void waitForJob( int jobId ) throws JobNotFoundException, RemoteException;


	/**
	 * Cancels a job specified by jobId parameter. Canceling a job implies on
	 * interrupt the running tasks and consider all the ready and running tasks
	 * as canceled tasks. Once a job is canceled, it must not be scheduled
	 * again.
	 * 
	 * @param jobId The identification of the job to be canceled.
	 * @throws JobNotFoundException If the job specified by jobId does not
	 *         exist.
	 * @throws RemoteException If there is a communication exception during the
	 *         execution of a remote method call.
	 * @throws JobCouldNotBeCancelledException If asked to cancel a job that
	 *         canot be cancelled anymore.
	 */
	public void cancelJob( int jobId ) throws JobNotFoundException, RemoteException, JobCouldNotBeCancelledException;


	/**
	 * Adds a new job to the scheduler.
	 * 
	 * @param newJob The JobInfo object that describes the job to be added.
	 * @return the identifier of this job for the scheduler. This identifier can
	 *         be used in the waitForJob method.
	 * @throws RemoteException If there is a communication exception during the
	 *         execution of a remote method call.
	 */
	public int addJob( JobSpec newJob ) throws RemoteException;


	/**
	 * Clean all jobs that are not Running and will not run anymore.
	 * 
	 * @throws RemoteException If there is a communication exception during the
	 *         execution of a remote method call.
	 */
	public void cleanFinishedJobs() throws RemoteException;


	/**
	 * Remove the information about the specified job
	 * 
	 * @throws RemoteException if there is a communication exception during the
	 *         execution of a remote method call.
	 * @throws JobCannotBeRemovedException
	 */
	public void removeJob( int jobId ) throws RemoteException, JobCannotBeRemovedException;


	/**
	 * Provides information about the gums currently in use by the scheduler
	 * 
	 * @return a collection with the EBGridManager entries
	 * @throws RemoteException if there is a communication exception during the
	 *         execution of a remote method call.
	 */
	public Collection<GridManagerEntry> getGumEntries( GridManagerListenerRemote newListener ) throws RemoteException;


	/**
	 * Provides information about the gums currently in use by the scheduler
	 * 
	 * @return a collection with the gums specifications
	 * @throws RemoteException if there is a communication exception during the
	 *         execution of a remote method call.
	 */
	public Collection<GumSpec> getGumSpecs() throws RemoteException;


	public void removeJobManagerListener( JobManagerListenerRemote listener ) throws RemoteException;


	public void removeGridManagerListener( GridManagerListenerRemote listener ) throws RemoteException;
}
