/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler;

import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NoSuchObjectException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.event.ActionEvent;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ShutdownEventEngine;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.common.gump.GumpClient;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.url.MyGridURLProvider;
import org.ourgrid.mygrid.replicaexecutor.EBReplicaExecutorFacade;
import org.ourgrid.mygrid.scheduler.gridmanager.EBGridManager;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManager;
import org.ourgrid.mygrid.scheduler.gump.GumpManager;
import org.ourgrid.mygrid.scheduler.jobmanager.EBJobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManager;

/**
 * This class is responsible for receiving events that come from the
 * <code>EBSchedulerFacade</code> and process them.
 */
public class SchedulerEventEngine implements ShutdownEventEngine {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( SchedulerEventEngine.class );

	/** GridMachineConsumer accessible through RMI. * */
	private GumpClient gumpClient;

	/** GuMPManager accessible through RMI. * */
	private GumpManager gumpManager;

	/** Scheduler accessible through RMI. * */
	private Scheduler scheduler;

	/** The <code>EventQueue</code> */
	private EventQueue<ActionEvent> eventQueue;

	/** The <code>EBJobManager</code> */
	private JobManager jobManager;

	/** The <code>EBGridManager</code> */
	private EBGridManager ebGridManager;

	/** The <code>EBSchedulingHeuristic</code>. */
	private EBSchedulingHeuristic ebSchedulingHeuristic;

	/** The SchedulerEventEngine own <code>Thread</code>. */
	private Thread myThread;

	/** Indicates if this EventEngine is alive. */
	private boolean isAlive;

	/** A control flag */
	private boolean mustShutdown;

	/** The queue for a shutdown request. */
	private EventQueue<ShutdownResponseEvent> shutdownEventQueue;

	private final EBReplicaExecutorFacade ebReplicaExecutorFacade;


	/**
	 * The constructor.
	 * 
	 * @param scheduler GuMManager accessible through RMI.
	 * @param gumpManager GuMPManager accessible through RMI.
	 * @param gridMachineConsumer GridMachineConsumer accessible through RMI.
	 * @param ebReplicaExecutorFacade The EBReplicaExecutorFacade
	 * @param ebSchedulerFacade The EBSchedulerFacade
	 * @throws ConfigException When the scheduler could not be created
	 */
	public SchedulerEventEngine( Scheduler scheduler, GumpManager gumpManager, GumpClient gridMachineConsumer,
									EBReplicaExecutorFacade ebReplicaExecutorFacade, EBSchedulerFacade ebSchedulerFacade )
		throws ConfigException {

		this.scheduler = scheduler;
		this.gumpManager = gumpManager;
		this.gumpClient = gridMachineConsumer;
		this.ebReplicaExecutorFacade = ebReplicaExecutorFacade;

		this.eventQueue = new EventQueue<ActionEvent>();
		this.jobManager = new EBJobManager();
		this.ebGridManager = new EBGridManager( ebSchedulerFacade, gumpClient );

		String maxReplicas = Configuration.getInstance().getProperty( MyGridConfiguration.PROP_MAX_REPLICAS );
		String maxFails = Configuration.getInstance().getProperty( MyGridConfiguration.PROP_MAX_FAILS );
		jobManager.setMaxReplicas( maxReplicas );
		jobManager.setMaxFails( maxFails );
		this.ebSchedulingHeuristic = createSchedulingHeuristic();
		this.mustShutdown = false;
		this.myThread = new Thread( this, "SchedulerEventProcessorThread" );
	}


	protected EBSchedulingHeuristic createSchedulingHeuristic() throws ConfigException {

		EBSchedulingHeuristicFactory shf = new EBSchedulingHeuristicFactory();
		String heuristicName = Configuration.getInstance().getProperty( MyGridConfiguration.PROP_HEURISTIC );
		return shf.getHeuristic( heuristicName, jobManager, ebGridManager, ebReplicaExecutorFacade );
	}


	/**
	 * Starts the SchedulerEventEngine <code>Thread</code>.
	 */
	public void startProcessing() {

		this.isAlive = true;
		// Binding the remote objects
		try {
			Naming.rebind( MyGridURLProvider.scheduler(), this.scheduler );
			Naming.rebind( MyGridURLProvider.gumpManager(), this.gumpManager );
			Naming.rebind( MyGridURLProvider.gumpClient(), this.gumpClient );
		} catch ( RemoteException e ) {
			LOG.error( "Error trying to bind remote objects", e );
		} catch ( MalformedURLException e ) {
			LOG.error( "Error trying to bind remote objects. Malformed URL address(es)", e );
		}

		myThread.start();
		LOG.debug( "The SchedulerEventEngine was successfully started." );
	}


	/**
	 * Puts an <code>Event</code> in the EventQueue.
	 * 
	 * @param event The <code>Event</code>
	 */
	public void putEvent( ActionEvent event ) {

		eventQueue.put( event );
	}


	/**
	 * Returns the <code>EBJobManager</code>.
	 * 
	 * @return The <code>EBJobManager</code>
	 */
	public JobManager getJobManager() {

		return this.jobManager;
	}


	/**
	 * Returns the <code>EBGridManager</code>.
	 * 
	 * @return The <code>EBGridManager</code>
	 */
	public GridManager getGridManager() {

		return this.ebGridManager;
	}


	/**
	 * Returns the <code>EBSchedulingHeuristic</code>.
	 * 
	 * @return The <code>EBSchedulingHeuristic</code>
	 */
	public EBSchedulingHeuristic getHeuristic() {

		return this.ebSchedulingHeuristic;
	}


	protected EBReplicaExecutorFacade getReplicaExecutorFacade() {

		return ebReplicaExecutorFacade;
	}


	protected boolean schedule( EBSchedulingHeuristic schedulingHeuristic ) {

		return schedulingHeuristic.schedule();
	}


	/**
	 * The <code>Thread</code> behavior.
	 */
	public void run() {

		boolean scheduled;
		while ( !mustShutdown ) {

			scheduled = schedule( ebSchedulingHeuristic );

			ActionEvent actionEvent;

			/*
			 * If the schedule occurred successfully there could be more
			 * replicas to be scheduled, so the schedule method must be called
			 * even if there are no events in the event queue.
			 */

			if ( scheduled ) {
				actionEvent = eventQueue.unblockingRemove();
			} else {
				actionEvent = eventQueue.blockingRemove();
			}

			if ( actionEvent != null ) {

				try {

					actionEvent.process();

				} catch ( Exception e ) {

					LOG.error( "Error in event process! Cause: " + e.getMessage(), e );
				}
			}
		}
		this.isAlive = false;
		shutdownEventQueue.put( new ShutdownResponseEvent() );
	}


	/**
	 * Marks this <code>Runnable</code> to be shutdown and tries to unbind all
	 * remote objects.
	 */
	public void shutdown( EventQueue<ShutdownResponseEvent> eq ) {

		this.ebGridManager.shutdown();
		this.ebSchedulingHeuristic.getEBJobManager().shutdown();
		this.shutdownEventQueue = eq;
		this.mustShutdown = true;

		this.unbindAndUnexportRMIObjects();

	}


	/**
	 * Unbind and unexports all RMI objects created by the Scheduler
	 */
	private void unbindAndUnexportRMIObjects() {

		try {

			Naming.unbind( MyGridURLProvider.scheduler() );

		} catch ( RemoteException e ) {
			LOG.error( "Could not contact registry:" + e.getMessage() );
		} catch ( MalformedURLException e ) {
			LOG.error( "MalformedURLException (this should never happen):" + e.getMessage() );
		} catch ( NotBoundException e ) {
			LOG.debug( MyGridURLProvider.scheduler() + " is already unbound:" + e.getMessage() );
		}

		try {
			OurgridUnicastRemoteObject.unexportObject( this.scheduler, true );
		} catch ( NoSuchObjectException e1 ) {
			LOG.debug( "Object " + this.scheduler + " is already unexported: " + e1.getMessage() );
		}

		try {
			Naming.unbind( MyGridURLProvider.gumpManager() );
		} catch ( RemoteException e ) {
			LOG.error( "Could not contact registry:" + e.getMessage() );
		} catch ( MalformedURLException e ) {
			LOG.error( "MalformedURLException (this should never happen):" + e.getMessage() );
		} catch ( NotBoundException e ) {
			LOG.debug( MyGridURLProvider.gumpManager() + " is already unbound:" + e.getMessage() );
		}

		try {
			OurgridUnicastRemoteObject.unexportObject( this.gumpManager, true );
		} catch ( NoSuchObjectException e2 ) {
			LOG.debug( "Object " + this.gumpManager + " is already unexported: " + e2.getMessage() );
		}

		try {
			Naming.unbind( MyGridURLProvider.gridManager() );
		} catch ( RemoteException e ) {
			LOG.error( "Could not contact registry:" + e.getMessage() );
		} catch ( MalformedURLException e ) {
			LOG.error( "MalformedURLException (this should never happen):" + e.getMessage() );
		} catch ( NotBoundException e ) {
			LOG.debug( MyGridURLProvider.gridManager() + " is already unbound:" + e.getMessage() );
		}

		try {
			Naming.unbind( MyGridURLProvider.gumpClient() );
		} catch ( RemoteException e ) {
			LOG.error( "Could not contact registry:" + e.getMessage() );
		} catch ( MalformedURLException e ) {
			LOG.error( "MalformedURLException (this should never happen):" + e.getMessage() );
		} catch ( NotBoundException e ) {
			LOG.debug( MyGridURLProvider.gumpClient() + " is already unbound:" + e.getMessage() );
		}

		try {
			OurgridUnicastRemoteObject.unexportObject( this.gumpClient, true );
		} catch ( NoSuchObjectException e2 ) {
			LOG.debug( "Object " + this.gumpClient + " is already unexported: " + e2.getMessage() );
		}

	}


	/**
	 * @see org.ourgrid.common.event.EBSyncShutdownable#isAlive()
	 */
	public boolean isAlive() {

		return this.isAlive;
	}
}
