/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler;

import java.rmi.RemoteException;
import java.util.Collection;

import org.ourgrid.common.event.Event;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.NotifyEvent;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.event.ThrowableEvent;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.security.MGSecureClientSocketFactory;
import org.ourgrid.common.security.MGSecureServerSocketFactory;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.mygrid.scheduler.event.JobCouldNotBeCancelledNotifyEvent;
import org.ourgrid.mygrid.scheduler.event.JobNotFoundNotifyEvent;
import org.ourgrid.mygrid.scheduler.exception.JobCannotBeRemovedException;
import org.ourgrid.mygrid.scheduler.exception.JobCouldNotBeCancelledException;
import org.ourgrid.mygrid.scheduler.exception.JobNotFoundException;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerEntry;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerListenerRemote;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManagerListenerRemote;

/**
 * The <code>Scheduler</code> remote object responsible for receiving RMI
 * calls and translating them to calls on the <code>EBSchedulerFacade</code>.
 */
public class SchedulerImpl extends OurgridUnicastRemoteObject implements Scheduler {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/** The <code>EBSchedulerFacade</code> */
	private EBSchedulerFacade ebSchedulerFacade;


	/**
	 * The constructor.
	 * 
	 * @param ebSchedulerFacade The <code>EBSchedulerFacade</code>
	 */
	protected SchedulerImpl( EBSchedulerFacade ebSchedulerFacade ) throws RemoteException {

		super();
		this.ebSchedulerFacade = ebSchedulerFacade;
	}


	/**
	 * The constructor with SecureSocketFactories.
	 * 
	 * @param ebSchedulerFacade The <code>EBSchedulerFacade</code>
	 * @param secureClientSocketFactory The SecureClientSocketFactory.
	 * @param secureServerSocketFactory The SecureServerSocketFactory.
	 * @param securePort The port where the secure services will be bound.
	 */
	public SchedulerImpl( EBSchedulerFacade ebSchedulerFacade, MGSecureClientSocketFactory secureClientSocketFactory,
							MGSecureServerSocketFactory secureServerSocketFactory, int securePort )
		throws RemoteException {

		super( securePort, secureClientSocketFactory, secureServerSocketFactory );
		this.ebSchedulerFacade = ebSchedulerFacade;
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.Scheduler#jobList(JobManagerListenerRemote)
	 */
	public Collection<JobEntry> jobList( JobManagerListenerRemote newListener ) throws RemoteException {

		EventQueue<ResponseEvent<Collection<JobEntry>>> responseQueue = ebSchedulerFacade.jobList( newListener );
		ResponseEvent<Collection<JobEntry>> responseEvent = responseQueue.blockingRemove();
		return responseEvent.getResponse();
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.Scheduler#waitForJob(int)
	 */
	public void waitForJob( int jobId ) throws JobNotFoundException, RemoteException {

		EventQueue responseQueue = ebSchedulerFacade.waitForJob( jobId );
		Event event = responseQueue.blockingRemove();
		if ( event instanceof JobNotFoundNotifyEvent ) {
			throw new JobNotFoundException( jobId );
		}

	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.Scheduler#cancelJob(int)
	 */
	public void cancelJob( int jobId ) throws JobNotFoundException, JobCouldNotBeCancelledException, RemoteException {

		EventQueue<NotifyEvent> responseQueue = ebSchedulerFacade.cancelJob( jobId );
		Event event = responseQueue.blockingRemove();
		if ( event instanceof JobNotFoundNotifyEvent ) {
			throw new JobNotFoundException( jobId );
		} else if ( event instanceof JobCouldNotBeCancelledNotifyEvent ) {
			throw new JobCouldNotBeCancelledException( event.toString() );
		}

	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.Scheduler#addJob(JobSpec)
	 */
	public int addJob( JobSpec jobSpec ) throws RemoteException {

		return ebSchedulerFacade.addJob( jobSpec );
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.Scheduler#cleanFinishedJobs()
	 */
	public void cleanFinishedJobs() throws RemoteException {

		ebSchedulerFacade.cleanFinishedJobs();
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.Scheduler#removeJob(int)
	 */
	public void removeJob( int jobId ) throws JobCannotBeRemovedException, RemoteException {

		EventQueue responseQueue = ebSchedulerFacade.removeJob( jobId );
		Event event = responseQueue.blockingRemove();
		if ( event instanceof ThrowableEvent ) {
			throw new JobCannotBeRemovedException();
		}
	}


	public Collection<GridManagerEntry> getGumEntries( GridManagerListenerRemote newListener ) throws RemoteException {

		EventQueue<ResponseEvent<Collection<GridManagerEntry>>> responseQueue = ebSchedulerFacade
				.getGumEntries( newListener );
		ResponseEvent<Collection<GridManagerEntry>> responseEvent = responseQueue.blockingRemove();
		return responseEvent.getResponse();
	}


	public Collection<GumSpec> getGumSpecs() throws RemoteException {

		EventQueue<ResponseEvent<Collection<GumSpec>>> responseQueue = ebSchedulerFacade.getGumSpecs();
		ResponseEvent<Collection<GumSpec>> responseEvent = responseQueue.blockingRemove();
		return responseEvent.getResponse();
	}


	public void removeJobManagerListener( JobManagerListenerRemote listener ) throws RemoteException {

		ebSchedulerFacade.removeJobManagerListener( listener );
	}


	public void removeGridManagerListener( GridManagerListenerRemote listener ) throws RemoteException {

		ebSchedulerFacade.removeGridManagerListener( listener );
	}

}
