/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler;

import java.io.File;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import org.ourgrid.common.exception.UnavailableGumException;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.spec.IOBlock;
import org.ourgrid.common.spec.IOEntry;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.common.util.JavaFileUtil;
import org.ourgrid.gridmachine.exception.FailCauseException;
import org.ourgrid.gridmachine.exception.FileTransferException;
import org.ourgrid.gridmachine.exception.UnableToExecuteException;
import org.ourgrid.mygrid.replicaexecutor.EBReplicaExecutorFacade;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManager;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.TaskEntry;

/**
 * The StorageAffinity is a scheduling heuristic that uses the affinity between
 * a <code>GridMachine</code> and a <code>Task</code> to schedule. And the
 * affinity, that is in the GuM, is the amount that is calculated using the
 * sizes required by the task. For example: GuM1 (has file1 [200 KB], file2 [760
 * KB]) GuM2 (has file2 [760 KB]) GuM3 (has file1 [200 KB], file3 [600 KB])
 * Task1: (requires file2) Task2: (requires file1, file3) On the
 * <code>schedule()</code> method call, the algorithm will search the biggest
 * affinity between the tasks and gums. The first assignment is GuM3 <-> Task2
 * with affinity 800. The second assignment is GuM1 <-> Task1 with affinity 760.
 * All the tasks are assigned and now, depending on replication degree, the last
 * gum will be assigned to the appropriated task. The third assignment is GuM2
 * <-> Task1 with affinity 760.
 */
public class StorageAffinity extends EBAbstractSchedulingHeuristic {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( StorageAffinity.class );


	/**
	 * The constructor.
	 * 
	 * @param jobManager <code>The EBJobManager</code>
	 * @param ebGridManager <code>The EBGridManager</code>
	 * @param ebReplicaExecutorFacade <code>The EBReplicaExecutorFacade</code>
	 */
	public StorageAffinity( JobManager jobManager, GridManager ebGridManager,
							EBReplicaExecutorFacade ebReplicaExecutorFacade ) {

		super( jobManager, ebGridManager, ebReplicaExecutorFacade );

	}


	/**
	 * Tries to assign a task to a machine.
	 */
	@Override
	public boolean schedule() {

		Collection<GumClient> unavailableGuMs = new HashSet<GumClient>();

		GumClient chosenGum = null;
		TaskEntry chosenTask = null;
		List<TaskEntry> tasksToScheduleQueue = getTasksToScheduleQueue();

		int minimumReplicationDegree = getSmallestNumberOfReplicas( tasksToScheduleQueue );

		Collection<JobEntry> jobs = jobManager.jobList();

		Iterator<JobEntry> jobsIt = jobs.iterator();
		while ( jobsIt.hasNext() ) {
			long maxAffinity = -1;
			long tempAffinity = -1;
			JobEntry job = jobsIt.next();
			Collection<TaskEntry> tasks = job.getAvailableTasks();
			tasksToScheduleQueue.addAll( tasks );

			Iterator<TaskEntry> tasksIt = tasksToScheduleQueue.iterator();

			Collections.sort( tasksToScheduleQueue );

			while ( tasksIt.hasNext() ) {

				TaskEntry task = tasksIt.next();

				if ( task.canReplicate() ) {
					for ( GumClient gum : gridManager.getAvailableGuMs( task.getJobId(), task.getId() ) ) {
						if ( unavailableGuMs.contains( gum ) )
							continue;
						try {
							if ( this.getReplicationDegreeOfTask( task ) == minimumReplicationDegree ) {
								tempAffinity = this.getAffinity( task, gum );
								int tempReplicationDegree = this.getReplicationDegreeOfTask( task );
								if ( ((tempReplicationDegree == 0) || ((tempReplicationDegree > 0) && (tempAffinity > 0)))
										&& (tempAffinity > maxAffinity) ) {
									maxAffinity = tempAffinity;
									chosenGum = gum;
									chosenTask = task;
								}
							}
						} catch ( FailCauseException e ) {
							if ( e instanceof UnavailableGumException ) {
								unavailableGuMs.add( gum );
							}
						}
					}

				}
			}
		}

		if ( chosenGum != null && chosenTask != null ) {
			ReplicaEntry replica = jobManager.createNewReplica( chosenTask );
			executeReplica( replica, chosenGum );
			LOG.info( chosenGum.getGumSpec().getGumID() + " was assigned to execute replica " + replica.toString() );
			return true;
		}

		return false;

	}


	/**
	 * Calculates the affinity between a task and a gum.
	 * 
	 * @param task The task.
	 * @param gumex The grid machine extended.
	 * @return A long value that represents the affinity.
	 * @throws FileTransferException
	 * @throws UnableToExecuteException
	 * @throws UnavailableGumException
	 * @throws UnavailableGumException
	 * @throws FileTransferException
	 * @throws UnableToExecuteException
	 */
	public long getAffinity( TaskEntry task, GumClient gumex ) throws FailCauseException {

		long amount = 0L;

		TaskSpec spec = task.getSpec();
		IOBlock ioBlock = spec.getInitBlock();
		DataDiscovery dd = DataDiscovery.getInstance();

		if ( ioBlock == null )
			return 0;

		Collection<IOEntry> allEntries = new HashSet<IOEntry>();

		Iterator<String> itConditions = ioBlock.getConditions();
		while ( itConditions.hasNext() ) {
			String condition = itConditions.next();
			allEntries.addAll( ioBlock.getEntry( condition ) );
		}

		Iterator<IOEntry> itIOEntries = allEntries.iterator();
		while ( itIOEntries.hasNext() ) {
			IOEntry ioEntry = itIOEntries.next();
			if ( ioEntry.getCommand().equalsIgnoreCase( "STORE" ) ) {
				String remotePath = translatePath( gumex, ioEntry.getDestiny() );
				String localPath = ioEntry.getSourceFile();
				amount += dd.amountAtDestination( localPath, remotePath, gumex );
			}
		}

		return amount;
	}


	/**
	 * Replace a possible value of $STORAGE in the destiny name with its real
	 * value.
	 * 
	 * @param gumex The GuMex that contains the real storage directory.
	 * @param destiny The destiny name to be converted.
	 * @return The destiny name in a valid form.
	 * @throws FileTransferException
	 * @throws UnableToExecuteException
	 * @throws UnavailableGumException
	 */
	private String translatePath( GumClient gumex, String destiny ) throws FailCauseException {

		if ( JavaFileUtil.isAbsolutePath( destiny ) )
			return destiny;

		String result = destiny;
		if ( destiny.startsWith( "$" ) ) {
			result = destiny.substring( 9 );
		}

		result = gumex.getStorageDirectory() + File.separator + result;
		return result;

	}


	/**
	 * Get the smallest number of replicas for the list of Tasks to schedule
	 * 
	 * @param tasksToScheduleQueue - list of tasks to schedule
	 * @return the smallest number of replicas
	 */
	protected int getSmallestNumberOfReplicas( List<TaskEntry> tasksToScheduleQueue ) {

		int smallestNumber = Integer.MAX_VALUE;

		Iterator<TaskEntry> tasksIt = tasksToScheduleQueue.iterator();
		while ( tasksIt.hasNext() ) {
			TaskEntry task = tasksIt.next();
			if ( (getReplicationDegreeOfTask( task )) < smallestNumber ) {
				smallestNumber = getReplicationDegreeOfTask( task );
			}
		}

		return smallestNumber;

	}


	/**
	 * Get the replication degree of a task
	 * 
	 * @param task The task
	 * @return the replication degree of a given task
	 */
	protected int getReplicationDegreeOfTask( TaskEntry task ) {

		return (task.getNumberOfRunningReplicas() + task.getNumberOfUnstartedReplicas());
	}

}
