/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler;

import java.util.Iterator;
import java.util.List;

import org.ourgrid.common.gum.GumClient;
import org.ourgrid.mygrid.replicaexecutor.EBReplicaExecutorFacade;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManager;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.TaskEntry;

// This class should have Friendly visibility
/**
 * The <b>Workqueue with Replication</b> scheduler.<br>
 * <br>
 * It picks tasks from the <i>bag-of-tasks</i> and assigns them to machines in
 * a random way. When the bag is empty and there are free machines, replicas of
 * running tasks are created and assigned to these machines. The idea is that
 * replicas assigned later can finish before the ones previously assigned,
 * improving the performance. When a replica finishes, the other replicas
 * related to the same task are aborted to free the machines.<br>
 * <br>
 * Replicas are made until a pre-defined number. It can be definied in the
 * MG.properties file in a property called <b>MG.maxreplicas</b>. Replicas of
 * tasks are created based on a pre-defined policy.<br>
 * <br>
 * Tasks with lower number of running replicas have priority to have their
 * replicas scheduled. If the number of running replicas is the same, the task
 * with the lower job id has priority. If also the job ids are equals, the task
 * with the lower task id has priority.
 */
public class Workqueue extends EBAbstractSchedulingHeuristic {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( Workqueue.class );


	/**
	 * The constructor.
	 * 
	 * @param jobManager <code>The EBJobManager</code>
	 * @param ebGridManager <code>The EBGridManager</code>
	 * @param ebReplicaExecutorFacade <code>The EBReplicaExecutorFacade</code>
	 */
	public Workqueue( JobManager jobManager, GridManager ebGridManager, EBReplicaExecutorFacade ebReplicaExecutorFacade ) {

		super( jobManager, ebGridManager, ebReplicaExecutorFacade );

	}


	/**
	 * Tries to assign a task to a machine.
	 */
	public boolean schedule() {

		List<TaskEntry> tasksToScheduleQueue = getTasksToScheduleQueue();

		Iterator<TaskEntry> tasksIt = tasksToScheduleQueue.iterator();
		while ( tasksIt.hasNext() ) {
			TaskEntry task = tasksIt.next();
			if ( task.canReplicate() ) {
				GumClient chosenGum = chooseGuM( task );
				if ( chosenGum != null ) {
					ReplicaEntry replica = jobManager.createNewReplica( task );
					executeReplica( replica, chosenGum );

					LOG.info( chosenGum.getGumSpec().getGumID() + " was assigned to execute replica "
							+ replica.toString() );

					tasksIt.remove();
					return true;
				}
			} else {
				tasksIt.remove();
			}
		}

		return false;
	}


	/**
	 * Tries to find a GridMachine that matches the task's requirements.
	 * 
	 * @param task The task to be scheduled
	 * @return A GridMachine that matches the tasks's requirements
	 */
	private GumClient chooseGuM( TaskEntry task ) {

		List<GumClient> gums = gridManager.getAvailableGuMs( task.getJobId(), task.getId() );
		if ( gums.isEmpty() )
			return null;
		return gums.get( 0 );
	}

}
