/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.gridmanager;

import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ourgrid.common.exception.RequestAlreadyExistsException;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.gump.Gump;
import org.ourgrid.common.gump.GumpClient;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.PeerSpec;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.mygrid.scheduler.BlackListManager;
import org.ourgrid.mygrid.scheduler.EBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.JobRequestEntry;
import org.ourgrid.mygrid.scheduler.PeerEntry;
import org.ourgrid.mygrid.scheduler.PeerEntry.PeerState;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;

/**
 * An object of this class is responsible for obtaining <code>GridMachine</code>
 * s to be used by the scheduling heuristic from
 * <code>GridMachineProvider</code>s.
 */
public class EBGridManager extends AbstractGridManager {

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( EBGridManager.class );

	/** Map of JobRequest entries. */
	public Map<Integer,JobRequestEntry> jobRequestsEntryMap;

	private BlackListManager blackListManager;

	private EBSchedulerFacade ebSchedulerFacade;

	/** The grid machines received */
	private Map<GumID,GridManagerEntry> gumEntries;

	/**
	 * The <code>GumpClient</code> to be passed along with a request.
	 * Providers will use this object to deliver gums.
	 */
	private GumpClient gumpClient;

	/** The peers known by MyGrid */
	private Map<PeerSpec,PeerEntry> peerEntries;


	public EBGridManager( EBSchedulerFacade ebSchedulerFacade, GumpClient gumpClient ) {

		this.gumEntries = new HashMap<GumID,GridManagerEntry>();
		this.jobRequestsEntryMap = new HashMap<Integer,JobRequestEntry>();
		this.ebSchedulerFacade = ebSchedulerFacade;
		this.gumpClient = gumpClient;
		this.peerEntries = new HashMap<PeerSpec,PeerEntry>();
		this.blackListManager = new BlackListManager();
	}


	public void dispose( GumID gumID ) {

		GridManagerEntry gumEntry = gumEntries.get( gumID );

		if ( gumEntry != null ) {
			sendDisposeToGump( gumEntry );

			deleteGum( gumID );
			LOG.info( "The GuM '" + gumID + "' was released" );
		} else {
			LOG.error( "Error while trying to free the GuM: " + gumID + ". This GuM is unknown by the EBGridManager" );
		}
	}


	private void sendDisposeToGump( GridManagerEntry gumEntry ) {

		try {
			Gump gump = (Gump) Naming.lookup( gumEntry.getPeerBrokerAccessID().getObjectURL() );
			gump.disposeGum( gumEntry.getRequestId(), gumEntry.getGumID() );
		} catch ( RemoteException e ) {
			LOG.error( "Could not contact Gump at " + gumEntry.getPeerBrokerAccessID().getObjectURL() + e.getMessage() );
		} catch ( MalformedURLException e ) {
			LOG.error( "Could not contact Gump at " + gumEntry.getPeerBrokerAccessID().getObjectURL() + e.getMessage() );
		} catch ( NotBoundException e ) {
			LOG.error( "Could not contact Gump at " + gumEntry.getPeerBrokerAccessID().getObjectURL() + e.getMessage() );
		}
	}


	public HashMap<GumID,GridManagerEntry> getAllGumEntries() {

		return new HashMap<GumID,GridManagerEntry>( gumEntries );
	}


	public List getAvailableAndValidGums() {

		List<GumClient> availableGuMs = new LinkedList<GumClient>();

		Iterator<GridManagerEntry> gumEntriesIt = new HashMap<GumID,GridManagerEntry>( gumEntries ).values().iterator();
		while ( gumEntriesIt.hasNext() ) {
			GridManagerEntry gumEntry = gumEntriesIt.next();
			if ( gumEntry.isReady() ) {
				availableGuMs.add( gumEntry.getGum() );
			}
		}

		return availableGuMs;
	}


	public List<GumClient> getAvailableGuMs( int jobID, int taskID ) {

		List<GumClient> availableGuMs = new LinkedList<GumClient>();

		JobRequestEntry jobRequestEntry = getJobRequestEntry( jobID );

		Iterator<GridManagerEntry> gumEntriesIt = new HashMap<GumID,GridManagerEntry>( gumEntries ).values().iterator();
		while ( gumEntriesIt.hasNext() ) {
			GridManagerEntry gumexEntry = gumEntriesIt.next();

			if ( gumexEntry.isReady() && jobRequestEntry.getRequestID() == gumexEntry.getRequestId()
					&& canGumBeAllocatedToTask( gumexEntry.getGumID(), jobID, taskID ) ) {
				availableGuMs.add( gumexEntry.getGum() );
			}
		}

		return availableGuMs;
	}


	/* Methods related to the black list feature */

	public BlackListManager getEBBlackListManager() {

		return this.blackListManager;
	}


	public boolean canGumBeAllocatedToTask( GumID gumID, int jobID, int taskID ) {

		return this.blackListManager.canThisGumRunThisTask( jobID, gumID, taskID );
	}


	@Override
	public boolean realUnwantedGum( GumID gumID, int jobID ) {

		JobRequestEntry jobRequestEntry = getJobRequestEntry( jobID );

		if ( jobRequestEntry != null ) {

			long requestId = jobRequestEntry.getRequestID();
			GridManagerEntry entry = this.gumEntries.get( gumID );
			if ( entry != null ) {
				sendUnwantedGumsToGump( entry, requestId );
				deleteGum( gumID );

				return true;
			}
		}

		return false;
	}


	private void sendUnwantedGumsToGump( GridManagerEntry entry, long requestId ) {

		try {
			Gump gump = (Gump) Naming.lookup( entry.getPeerBrokerAccessID().getObjectURL() );
			gump.unwantedGum( requestId, entry.getGumID() );
		} catch ( Exception e ) {
			LOG.error( "Could not contact peer to blacklist Gum", e );
		}
	}


	@Override
	public boolean realUpdateBlackList( int jobID, int taskID, GumID gumID ) {

		int remaingFailures = this.blackListManager.executionFailed( jobID, gumID, taskID );
		if ( remaingFailures < 1 ) {
			return unwantedGum( gumID, jobID );
		}

		return false;
	}


	/* */

	public Collection<GumSpec> getGumSpecs() {

		Map<GumID,GridManagerEntry> gumEntriesMap = getAllGumEntries();
		Collection<GumSpec> specList = new LinkedList<GumSpec>();

		for ( GridManagerEntry gum : gumEntriesMap.values() ) {
			specList.add( gum.getGum().getGumSpec() );
		}

		return specList;
	}


	public Collection<GridManagerEntry> getGumEntries() {

		return new LinkedList<GridManagerEntry>( gumEntries.values() );
	}


	public Collection<PeerEntry> getPeerEntries() {

		return new LinkedList<PeerEntry>( this.peerEntries.values() );
	}


	public void initiateBlackListForJob( int jobId, int jobSize ) {

		this.blackListManager.initiateBlackListForJob( jobId, jobSize );
	}


	@Override
	public void realAllocate( GumID gumID, ReplicaEntry replica ) {

		GridManagerEntry gumEntry = gumEntries.get( gumID );
		if ( gumEntry != null ) {
			gumEntry.allocate( replica );
		} else {
			LOG.error( "Error while trying to allocate the Gum: " + gumID
					+ ". This Gum is unknown by the EBGridManager" );
		}
	}


	public void cancelRequests( int jobId ) {

		JobRequestEntry jobRequestsEntry = getJobRequestEntry( jobId );
		if ( jobRequestsEntry != null ) {
			jobRequestsEntry.cancelRequests();
			jobRequestsEntryMap.remove( jobId );

			// Deleting information on gums
			Collection<GridManagerEntry> entriesToDelete = new LinkedList<GridManagerEntry>();
			for ( GridManagerEntry gridManagerEntry : gumEntries.values() ) {
				if ( gridManagerEntry.getRequestId() == jobRequestsEntry.getRequestID() ) {
					entriesToDelete.add( gridManagerEntry );
				}
			}
			for ( GridManagerEntry entryToDelete : entriesToDelete ) {
				deleteGum( entryToDelete.getGumID() );
			}
			LOG.info( "All requests and Gums related to job '" + jobId + "' were removed." );
		}

	}


	@Override
	public void realDeleteGum( GumID gumID ) {

		gumEntries.remove( gumID );
		LOG.debug( "Grid machine " + gumID + " was removed from the EBGridManager" );
	}


	@Override
	public void realGumIsReady( GumID gumID ) {

		GridManagerEntry gumEntry = gumEntries.get( gumID );
		if ( gumEntry != null ) {
			gumEntry.ready();
			LOG.info( "The GuM '" + gumID + "' was released" );
		} else {
			LOG.error( "Error while trying to free the GuM: " + gumID + ". This GuM is unknown by the EBGridManager" );
		}

	}


	@Override
	public int realHereIsGum( GumClient gumClient, long requestId, ObjectID peerBrokerAccessID ) {

		boolean hasRequest = false;
		int jobID = -1; // can be any value
		JobRequestEntry jobEntry = null;

		Set<Map.Entry<Integer,JobRequestEntry>> entries = this.jobRequestsEntryMap.entrySet();
		for ( Map.Entry<Integer,JobRequestEntry> entry : entries ) {
			if ( entry.getValue().getRequestID() == requestId ) {
				hasRequest = true;
				jobID = entry.getKey();
				jobEntry = entry.getValue();
				break;
			}
		}

		if ( !hasRequest ) {
			cleanUnknowRequestAtPeer( peerBrokerAccessID, requestId );
			return 0;
		}

		GumID gumID = gumClient.getGumSpec().getGumID();

		GridManagerEntry gumEntry = gumEntries.get( gumID );

		if ( gumEntry == null ) {
			gumEntry = new GridManagerEntry( requestId, jobID, gumClient, peerBrokerAccessID );

			if ( this.blackListManager.isGumBlacklistedForEntireJob( jobID, gumID ) ) {
				LOG.debug( "The Gum <" + gumID + "> is unwanted to run tasks of the job <" + jobID + ">" );
				sendUnwantedGumsToGump( gumEntry, requestId );
			} else if ( jobEntry != null && jobEntry.isPaused() ) {
				sendDisposeToGump( gumEntry );
				LOG.debug( "Received gum: " + gumID + " for paused request: " + requestId );
			} else {
				gumEntries.put( gumID, gumEntry );
				LOG.debug( "The EBGridManager has received a new Gum (" + gumID + " ) from Peer "
						+ peerBrokerAccessID.getObjectURL() + " for request " + requestId );
				return jobID;
			}
		} else {
			LOG.debug( "The EBGridManager has received a Gum that it already knows ( "
					+ gumClient.getGumSpec().getGumID() + " ) from Peer " + peerBrokerAccessID.getObjectURL()
					+ ". Old Gum will be killed." );

			deleteGum( gumEntry.getGumID() );

			ebSchedulerFacade.hereIsGum( gumClient, requestId, peerBrokerAccessID );
		}

		return 0;
	}


	@Override
	public void realNeedGums( int jobId, String requirements, int numRequiredGums ) {

		if ( numRequiredGums < 1 )
			return;

		RequestSpec requestSpec;
		JobRequestEntry jobRequestEntry;
		long requestID;

		if ( !jobRequestsEntryMap.containsKey( jobId ) ) {
			requestID = this.getAUniqueRequestId();
			jobRequestEntry = new JobRequestEntry( jobId, requestID );
		} else {
			jobRequestEntry = getJobRequestEntry( jobId );
			requestID = jobRequestEntry.getRequestID();
		}

		requestSpec = new RequestSpec( requestID, requirements, numRequiredGums, gumpClient );

		for ( PeerEntry entry : peerEntries.values() ) {

			Gump peer = entry.getPeer();

			if ( peer != null ) {
				LOG.debug( "Asking " + numRequiredGums + " GuMs to " + peer.toString() );
				try {
					peer.requestGums( requestSpec );
					jobRequestEntry.addPeer( peer );
					LOG.debug( "Request " + requestID + " successfully done to peer " + peer );
				} catch ( RemoteException exception ) {
					LOG.error( "Could not contact GridMachineProvider " + peer.toString() + ": "
							+ exception.getMessage(), exception );
				} catch ( RequestAlreadyExistsException exception ) {
					LOG.error( "Trying to resubmit a requestID " + requestID + " for the " + jobId, exception );
				}

			} else {
				LOG.error( "Could not contact GridMachineProvider labeled " + entry.getPeerSpec().getLabel() + " at "
						+ entry.getPeerSpec().getURL() );
			}
		}

		jobRequestsEntryMap.put( new Integer( jobId ), jobRequestEntry );
	}


	@Override
	public void realPauseRequest( int jobId ) {

		JobRequestEntry jobRequestsEntry = getJobRequestEntry( jobId );

		if ( jobRequestsEntry != null ) {
			jobRequestsEntry.pauseRequest();
		}
	}


	@Override
	public void realPeerAlive( PeerSpec peerSpec ) {

		PeerEntry entry = peerEntries.get( peerSpec );

		if ( peerSpec != null ) {
			entry.setPeerStatus( PeerState.UP );
			LOG.info( "Peer " + peerSpec.getName() + " is rejoined at URL " + peerSpec.getURL() );
		}
	}


	@Override
	public void realPeerIsDead( PeerSpec peerSpec ) {

		PeerEntry entry = peerEntries.get( peerSpec );

		if ( peerSpec != null ) {
			entry.setPeerStatus( PeerState.CONTACTING );
			LOG.info( "Peer " + peerSpec.getName() + " at URL " + peerSpec.getURL() + " is dead." );
		}
	}


	@Override
	public void realResumeRequest( int jobId ) {

		JobRequestEntry jobRequestsEntry = getJobRequestEntry( jobId );

		if ( jobRequestsEntry != null ) {
			jobRequestsEntry.resumeRequest();
		}

	}


	@Override
	public void realSetPeers( Collection<PeerSpec> peerSpecs ) {

		this.peerEntries = new HashMap<PeerSpec,PeerEntry>();

		for ( PeerSpec spec : peerSpecs ) {

			PeerEntry peerEntry = new PeerEntry( spec );

			LOG.debug( "Peer Spec [" + spec.getURL() + "," + spec.getLabel() + "] has been set." );

			if ( peerEntry.getPeer() == null ) {

				String errMessage = "The Peer labeled " + spec.getLabel() + " at " + spec.getURL()
						+ " is not UP now, but MyGrid will be trying to contact it ";
				LOG.error( errMessage );

			}

			peerEntries.put( spec, peerEntry );
		}
	}


	@Override
	public void realShutdown() {

		LinkedList<JobRequestEntry> jobRequestsEntriesList = new LinkedList<JobRequestEntry>( jobRequestsEntryMap
				.values() );
		Iterator<JobRequestEntry> jobRequestsEntriesIt = jobRequestsEntriesList.iterator();

		while ( jobRequestsEntriesIt.hasNext() ) {
			JobRequestEntry jobEntry = jobRequestsEntriesIt.next();
			jobEntry.cancelRequests();
		}

		LinkedList<GridManagerEntry> gumEntriesList = new LinkedList<GridManagerEntry>();
		gumEntriesList = new LinkedList<GridManagerEntry>( gumEntries.values() );
		Iterator<GridManagerEntry> gumEntriesIt = gumEntriesList.iterator();

		while ( gumEntriesIt.hasNext() ) {
			GridManagerEntry gumEntry = gumEntriesIt.next();
			sendDisposeToGump( gumEntry );
		}

	}


	/**
	 * Clean any request that is garbage from a previous interaction with the
	 * peer.
	 */
	private void cleanUnknowRequestAtPeer( ObjectID peerBrokerAccessID, long requestId ) {

		LOG.debug( "Cleaning unknown request " + requestId + " at peer " + peerBrokerAccessID );

		try {
			Gump gump = (Gump) Naming.lookup( peerBrokerAccessID.getObjectURL() );
			gump.finishRequestGums( requestId );
		} catch ( Exception e ) {
			LOG.error( "Unable to lookup Peer at " + peerBrokerAccessID.getObjectURL() + " and clean unknown request.",
					e );
		}
	}


	/**
	 * Creates a random ID to be propagated.
	 * 
	 * @return The ID.
	 */
	private synchronized long getAUniqueRequestId() {

		return (long) (Math.random() * Long.MAX_VALUE) + 1;
	}


	private JobRequestEntry getJobRequestEntry( int jobId ) {

		return jobRequestsEntryMap.get( jobId );
	}
}
