/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.gump;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.security.MGSecureClientSocketFactory;
import org.ourgrid.common.security.MGSecureServerSocketFactory;
import org.ourgrid.common.spec.PeerSpec;
import org.ourgrid.mygrid.scheduler.EBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.PeerEntry;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerListenerRemote;

/**
 * The <code>GuMPManager</code> remote object responsible for receiving RMI
 * calls and translating them to calls on the <code>EBSchedulerFacade</code>.
 */
public class GumpManagerImpl extends OurgridUnicastRemoteObject implements GumpManager {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	private Map<String,PeerSpec> peerSpecs;

	/** The <code>EBSchedulerFacade</code> */
	private EBSchedulerFacade ebSchedulerFacade;


	/**
	 * The constructor.
	 * 
	 * @param ebSchedulerFacade The EBSchedulerFacade
	 * @throws RemoteException When a connection problem occurs while trying to
	 *         access this object through RMI
	 */
	public GumpManagerImpl( EBSchedulerFacade ebSchedulerFacade ) throws RemoteException {

		super();
		this.ebSchedulerFacade = ebSchedulerFacade;
		this.peerSpecs = new HashMap<String,PeerSpec>();
	}


	/**
	 * The constructor with SecureSocketFactories.
	 * 
	 * @param ebSchedulerFacade The EBSchedulerFacade.
	 * @param secureClientSocketFactory The SecureClientSocketFactory.
	 * @param secureServerSocketFactory The SecureServerSocketFactory.
	 * @param securePort The port where the secure services will be bound.
	 */
	public GumpManagerImpl( EBSchedulerFacade ebSchedulerFacade, MGSecureClientSocketFactory secureClientSocketFactory,
							MGSecureServerSocketFactory secureServerSocketFactory, int securePort )
		throws RemoteException {

		super( securePort, secureClientSocketFactory, secureServerSocketFactory );
		this.ebSchedulerFacade = ebSchedulerFacade;
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.gump.GumpManager#setPeers(Collection)
	 */
	public void setPeers( Collection<PeerSpec> peerSpecs ) throws RemoteException {

		ebSchedulerFacade.setPeers( peerSpecs );

		this.peerSpecs = new HashMap<String,PeerSpec>();
		for ( PeerSpec spec : peerSpecs ) {
			this.peerSpecs.put( spec.getURL(), spec );
		}
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.gump.GumpManager#getPeerEntries(GridManagerListenerRemote)
	 */
	public Collection<PeerEntry> getPeerEntries( GridManagerListenerRemote newListener ) throws RemoteException {

		EventQueue<ResponseEvent<Collection<PeerEntry>>> responseQueue = ebSchedulerFacade.getPeerEntries( newListener );
		ResponseEvent<Collection<PeerEntry>> responseEvent = responseQueue.blockingRemove();
		return responseEvent.getResponse();
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.gump.GumpManager#peerAlive(java.lang.String)
	 */
	public void peerAlive( String url ) {

		PeerSpec spec = peerSpecs.get( url );

		if ( spec != null ) {
			this.ebSchedulerFacade.peerAlive( spec );
		}
	}


	/**
	 * @see org.ourgrid.mygrid.scheduler.gump.GumpManager#peerIsDead(java.lang.String)
	 */
	public void peerIsDead( String url ) {

		PeerSpec spec = peerSpecs.get( url );

		if ( spec != null ) {
			this.ebSchedulerFacade.peerIsDead( spec );
		}
	}
}
