package org.ourgrid.mygrid.scheduler.jobmanager;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;

import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.mygrid.scheduler.ReplicaExecutorResult;
import org.ourgrid.mygrid.scheduler.exception.JobCannotBeRemovedException;
import org.ourgrid.mygrid.scheduler.exception.JobCouldNotBeCancelledException;
import org.ourgrid.mygrid.scheduler.exception.JobNotFoundException;

public abstract class AbstractJobManager implements JobManager {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( AbstractJobManager.class );

	private Collection<JobManagerListenerRemote> listeners = new LinkedHashSet<JobManagerListenerRemote>();


	public void addJob( int jobId, JobSpec jobSpec ) {

		realAddJob( jobId, jobSpec );

		Iterator<JobManagerListenerRemote> it = listeners.iterator();
		while ( it.hasNext() ) {
			JobManagerListenerRemote listener = it.next();
			try {
				listener.jobCreated( getJobById( jobId ) );
			} catch ( RemoteException e ) {
				it.remove();
				LOG.error( "Unable to contact listener " + listener, e );
			}
		}
	}


	public ReplicaEntry createNewReplica( TaskEntry task ) {

		ReplicaEntry replica = realCreateNewReplica( task );

		Iterator<JobManagerListenerRemote> it = listeners.iterator();
		while ( it.hasNext() ) {
			JobManagerListenerRemote listener = it.next();
			try {
				JobEntry jobEntry = getJobById( replica.getJobId() );
				TaskEntry taskEntry = jobEntry.getTaskById( replica.getTaskId() );
				listener.replicaStateChanged( replica, jobEntry.getStatus(), taskEntry.getState(), taskEntry
						.getActualFails() );
			} catch ( RemoteException e ) {
				it.remove();
				LOG.error( "Unable to contact listener " + listener, e );
			}
		}

		return replica;
	}


	protected abstract ReplicaEntry realCreateNewReplica( TaskEntry task );


	public void replicaFinished( ReplicaExecutorResult result ) {

		realReplicaFinished( result );

		Iterator<JobManagerListenerRemote> it = listeners.iterator();
		while ( it.hasNext() ) {
			JobManagerListenerRemote listener = it.next();
			try {
				ReplicaEntry replica = result.getReplica();
				JobEntry jobEntry = getJobById( replica.getJobId() );
				TaskEntry taskEntry = jobEntry.getTaskById( replica.getTaskId() );
				ReplicaEntry realReplica = taskEntry.getReplicaById( replica.getId() );
				listener.replicaStateChanged( realReplica, jobEntry.getStatus(), taskEntry.getState(), taskEntry
						.getActualFails() );
			} catch ( RemoteException e ) {
				it.remove();
				LOG.error( "Unable to contact listener " + listener, e );
			}
		}
	}


	public void replicaFailed( ReplicaExecutorResult result ) {

		realReplicaFailed( result );

		Iterator<JobManagerListenerRemote> it = listeners.iterator();
		while ( it.hasNext() ) {
			JobManagerListenerRemote listener = it.next();
			try {
				ReplicaEntry replica = result.getReplica();
				JobEntry jobEntry = getJobById( replica.getJobId() );
				TaskEntry taskEntry = jobEntry.getTaskById( replica.getTaskId() );
				ReplicaEntry realReplica = taskEntry.getReplicaById( replica.getId() );
				listener.replicaStateChanged( realReplica, jobEntry.getStatus(), taskEntry.getState(), taskEntry
						.getActualFails() );
			} catch ( RemoteException e ) {
				it.remove();
				LOG.error( "Unable to contact listener " + listener, e );
			}
		}
	}


	public void replicaAborted( ReplicaExecutorResult result ) {

		realReplicaAborted( result );

		Iterator<JobManagerListenerRemote> it = listeners.iterator();
		while ( it.hasNext() ) {
			JobManagerListenerRemote listener = it.next();
			try {
				ReplicaEntry replica = result.getReplica();
				JobEntry jobEntry = getJobById( replica.getJobId() );
				TaskEntry taskEntry = jobEntry.getTaskById( replica.getTaskId() );
				ReplicaEntry realReplica = taskEntry.getReplicaById( replica.getId() );
				listener.replicaStateChanged( realReplica, jobEntry.getStatus(), taskEntry.getState(), taskEntry
						.getActualFails() );
			} catch ( RemoteException e ) {
				it.remove();
				LOG.error( "Unable to contact listener " + listener, e );
			}
		}
	}


	public void replicaCanceled( ReplicaExecutorResult result ) {

		realReplicaCanceled( result );

		Iterator<JobManagerListenerRemote> it = listeners.iterator();
		while ( it.hasNext() ) {
			JobManagerListenerRemote listener = it.next();
			try {
				ReplicaEntry replica = result.getReplica();
				JobEntry jobEntry = getJobById( replica.getJobId() );
				TaskEntry taskEntry = jobEntry.getTaskById( replica.getTaskId() );
				ReplicaEntry realReplica = taskEntry.getReplicaById( replica.getId() );
				listener.replicaStateChanged( realReplica, jobEntry.getStatus(), taskEntry.getState(), taskEntry
						.getActualFails() );
			} catch ( RemoteException e ) {
				it.remove();
				LOG.error( "Unable to contact listener " + listener, e );
			}
		}
	}


	public void setCanceled( int jobId ) throws JobNotFoundException, JobCouldNotBeCancelledException {

		realSetCanceled( jobId );

		Iterator<JobManagerListenerRemote> it = listeners.iterator();
		while ( it.hasNext() ) {
			JobManagerListenerRemote listener = it.next();
			try {
				listener.jobCanceled( jobId );
			} catch ( RemoteException e ) {
				it.remove();
				LOG.error( "Unable to contact listener " + listener, e );
			}
		}
	}


	public void removeJob( int jobId ) throws JobCannotBeRemovedException {

		realRemoveJob( jobId );

		Iterator<JobManagerListenerRemote> it = listeners.iterator();
		while ( it.hasNext() ) {
			JobManagerListenerRemote listener = it.next();
			try {
				listener.removeJob( jobId );
			} catch ( RemoteException e ) {
				it.remove();
				LOG.error( "Unable to contact listener " + listener, e );
			}
		}
	}


	public void replicaRunning( ReplicaEntry replica, GumID gumID ) {

		realReplicaRunning( replica, gumID );

		Iterator<JobManagerListenerRemote> it = listeners.iterator();
		while ( it.hasNext() ) {
			JobManagerListenerRemote listener = it.next();
			try {
				JobEntry jobEntry = getJobById( replica.getJobId() );
				TaskEntry taskEntry = jobEntry.getTaskById( replica.getTaskId() );
				listener.replicaStateChanged( replica, jobEntry.getStatus(), taskEntry.getState(), taskEntry
						.getActualFails() );
			} catch ( RemoteException e ) {
				it.remove();
				LOG.error( "Unable to contact listener " + listener, e );
			}
		}
	}


	protected abstract void realReplicaRunning( ReplicaEntry replica, GumID gumID );


	public void shutdown() {

		realShutdown();

		Iterator<JobManagerListenerRemote> it = listeners.iterator();
		while ( it.hasNext() ) {
			JobManagerListenerRemote listener = it.next();
			try {
				listener.myGridDown();
			} catch ( RemoteException e ) {
				it.remove();
				LOG.error( "Unable to contact listener " + listener, e );
			}
		}
	}


	protected abstract void realShutdown();


	public void addNewListener( JobManagerListenerRemote newListener ) {

		if ( newListener != null ) {
			this.listeners.add( newListener );
		}

	}


	public void removeListener( JobManagerListenerRemote listener ) {

		this.listeners.remove( listener );
	}


	protected abstract void realAddJob( int jobId, JobSpec jobSpec );


	protected abstract void realReplicaFinished( ReplicaExecutorResult result );


	protected abstract void realReplicaFailed( ReplicaExecutorResult result );


	protected abstract void realReplicaAborted( ReplicaExecutorResult result );


	protected abstract void realReplicaCanceled( ReplicaExecutorResult result );


	protected abstract void realSetCanceled( int jobId ) throws JobNotFoundException, JobCouldNotBeCancelledException;


	protected abstract void realRemoveJob( int jobId ) throws JobCannotBeRemovedException;

}
