/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.jobmanager;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.util.StreamCloner;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.JobInterested;
import org.ourgrid.mygrid.scheduler.ReplicaExecutorResult;
import org.ourgrid.mygrid.scheduler.exception.JobCannotBeRemovedException;
import org.ourgrid.mygrid.scheduler.exception.JobCouldNotBeCancelledException;
import org.ourgrid.mygrid.scheduler.exception.JobNotFoundException;

/**
 * Class responsible for managing the jobs. It maintains a list of all jobs
 * submitted by the user and controls all operations made in these jobs.
 */
public class EBJobManager extends AbstractJobManager {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( EBJobManager.class );

	/** The collection of <code>JobInterested</code>s. */
	private List<JobInterested> jobInteresteds;

	/** The number of task replicas allowed to be running at the same time. * */
	private int maxReplicas;

	/** The number of fails that a task replica is allowed to have. * */
	private int maxFails;

	protected Map<Integer,JobEntry> jobs;


	/**
	 * The constructor.
	 */
	public EBJobManager() {

		this.jobs = Collections.synchronizedMap( new LinkedHashMap<Integer,JobEntry>() );
		this.jobInteresteds = new LinkedList<JobInterested>();
		this.maxReplicas = 1;
		this.maxFails = 1;
	}


	/**
	 * Returns a clone of the <code>Job</code> list. All objects within this
	 * list are clones.
	 * 
	 * @return A clone of the <code>Job</code> list.
	 */
	public Collection<JobEntry> clonedJobList() {

		Collection<JobEntry> clonedJobs = null;
		// It clones all objects within the jobs
		try {
			clonedJobs = StreamCloner.clone( new LinkedList<JobEntry>( jobs.values() ) );
		} catch ( CloneNotSupportedException e ) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return clonedJobs;
	}


	/**
	 * Returns the <code>Job</code> list.
	 * 
	 * @return The <code>Job</code> list.
	 */
	public Collection<JobEntry> jobList() {

		return new LinkedList<JobEntry>( jobs.values() );
	}


	/**
	 * Adds a new <code>JobInterested</code>.
	 * 
	 * @param jobInterested The <code>JobInterested</code>.
	 */
	public void addJobInterested( JobInterested jobInterested ) {

		JobEntry job = getJobById( jobInterested.getJobId() );

		if ( job == null ) {

			jobInterested.jobNotFound();

		} else {

			jobInteresteds.add( jobInterested );
			if ( !job.getStatus().isRunning() && !job.getStatus().isUnstarted() ) {
				notifyJobInteresteds( job );
			}
		}
	}


	/**
	 * Sets the <code>Replica</code> state to <i>finished</i>. It also
	 * notifies <code>JobInteresteds</code> case it is necessary.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	@Override
	protected void realReplicaFinished( ReplicaExecutorResult result ) {

		ReplicaEntry replica = result.getReplica();
		JobEntry job = getJobById( replica.getJobId() );
		if ( job != null ) {
			job.replicaFinished( result );
			LOG.info( "Replica " + result.getReplica().toString() + " finished on '" + result.getReplica().getGumID()
					+ "' with status '" + result.getReplica().getStatus() + "'" );
			LOG.debug( "\nInit result: " + result.getInitResult() + "\nRemote result: " + result.getRemoteResult()
					+ "\nFinal result: " + result.getFinalResult() );
			if ( job.getStatus().isFinished() || job.getStatus().isFailed() || job.getStatus().isCancelled() ) {
				notifyJobInteresteds( job );
			}
		}
	}


	/**
	 * Sets the <code>Replica</code> state to <i>failed</i>. It also notifies
	 * <code>JobInteresteds</code> case it is necessary.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	@Override
	protected void realReplicaFailed( ReplicaExecutorResult result ) {

		ReplicaEntry replica = result.getReplica();
		JobEntry job = getJobById( replica.getJobId() );
		if ( job != null ) {
			job.replicaFailed( result );
			LOG.info( "Replica " + result.getReplica().toString() + " failed to execute on '"
					+ result.getReplica().getGumID() + "'" );
			LOG.debug( "\nInit result: " + result.getInitResult() + "\nRemote result: " + result.getRemoteResult()
					+ "\nFinal result: " + result.getFinalResult() );
			if ( job.getStatus().isFinished() || job.getStatus().isFailed() || job.getStatus().isCancelled() ) {
				notifyJobInteresteds( job );
			}
		}
	}


	/**
	 * Sets the <code>Replica</code> state to <i>aborted</i>. It also
	 * notifies <code>JobInteresteds</code> case it is necessary.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	@Override
	protected void realReplicaAborted( ReplicaExecutorResult result ) {

		ReplicaEntry replica = result.getReplica();
		JobEntry job = getJobById( replica.getJobId() );
		if ( job != null ) {
			job.replicaAborted( result );
			LOG.info( "Aborted execution of replica " + result.getReplica().toString() + " on '"
					+ result.getReplica().getGumID() + "'" );
			LOG.debug( "\nInit result: " + result.getInitResult() + "\nRemote result: " + result.getRemoteResult()
					+ "\nFinal result: " + result.getFinalResult() );
			if ( job.getStatus().isFinished() || job.getStatus().isFailed() || job.getStatus().isCancelled() ) {
				notifyJobInteresteds( job );
			}
		}
	}


	/**
	 * Sets the <code>Replica</code> state to <i>canceled</i>. It also
	 * notifies <code>JobInteresteds</code> case it is necessary.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	@Override
	protected void realReplicaCanceled( ReplicaExecutorResult result ) {

		ReplicaEntry replica = result.getReplica();
		JobEntry job = getJobById( replica.getJobId() );
		if ( job != null ) {
			job.replicaCancelled( result );
			LOG.info( "Cancelled execution of replica " + result.getReplica().toString() + " on '"
					+ result.getReplica().getGumID() + "'" );
			LOG.debug( "\nInit result: " + result.getInitResult() + "\nRemote result: " + result.getRemoteResult()
					+ "\nFinal result: " + result.getFinalResult() );
			if ( job.getStatus().isFinished() || job.getStatus().isFailed() || job.getStatus().isCancelled() ) {
				notifyJobInteresteds( job );
			}
		}
	}


	@Override
	protected void realReplicaRunning( ReplicaEntry replica, GumID gumID ) {

		replica.setStatus( ExecutionStatus.RUNNING );
		replica.setGumID( gumID );
	}


	/**
	 * Creates and returns a new <code>Replica</code>.
	 * 
	 * @return A new <code>Replica</code> in case that a replication is
	 *         possible to be made, <code>null</code> otherwise.
	 */
	@Override
	protected ReplicaEntry realCreateNewReplica( TaskEntry task ) {

		JobEntry job = getJobById( task.getJobId() );
		if ( job != null ) {
			ReplicaEntry replica = job.getNewReplica( task );
			LOG.debug( "Replica '" + replica.toString() + "' was created" );
			return replica;
		}
		return null;
	}


	/**
	 * Notifies all interesteds in 'job' and removes them from the list of
	 * interesteds.
	 * 
	 * @param job Interesting job
	 */
	private void notifyJobInteresteds( JobEntry job ) {

		Iterator<JobInterested> jobInterestedsIt = jobInteresteds.iterator();
		while ( jobInterestedsIt.hasNext() ) {
			JobInterested currentJobInterested = jobInterestedsIt.next();
			if ( job.getId() == currentJobInterested.getJobId() ) {
				currentJobInterested.jobFinished( job );
				jobInterestedsIt.remove();
			}
		}
	}


	/**
	 * Returns the state of the given Job.
	 * 
	 * @param jobId The Job identification
	 * @return The Job state
	 */
	public ExecutionStatus getJobState( int jobId ) {

		return getJobById( jobId ).getStatus();
	}


	/**
	 * Returns a <code>Job</code> with the specified identification.
	 * 
	 * @param jobId The job identification.
	 * @return The specified <code>Job</code> case it exists,
	 *         <code>null</code> otherwise.
	 */
	public JobEntry getJobById( int jobId ) {

		return jobs.get( jobId );
	}


	/**
	 * Sets the number of task replicas allowed to be running at the same time.
	 * 
	 * @param maxReplicas The number of max replicas
	 * @throws ConfigException
	 */
	public void setMaxReplicas( String maxReplicas ) throws ConfigException {

		try {
			this.maxReplicas = Integer.parseInt( maxReplicas.trim() );
			LOG.debug( "Max replicas was set to '" + maxReplicas + "'" );
			if ( this.maxReplicas < 1 ) {
				throw new ConfigException( MyGridConfiguration.PROP_MAX_REPLICAS );
			}
		} catch ( NumberFormatException nfe ) {
			throw new ConfigException( MyGridConfiguration.PROP_MAX_REPLICAS );
		}
	}


	/**
	 * Gets the number of task replicas allowed to be running at the same time.
	 * 
	 * @return The number of max replicas
	 */
	public int getMaxReplicas() {

		return this.maxReplicas;
	}


	/**
	 * Sets the number of failed task replicas until consider the Task itself as
	 * failed.
	 * 
	 * @param maxFails The number of max fails
	 */
	public void setMaxFails( String maxFails ) throws ConfigException {

		try {
			this.maxFails = Integer.parseInt( maxFails.trim() );
			LOG.debug( "Max fails was set to '" + maxFails + "'" );
			if ( this.maxFails < 1 ) {
				throw new ConfigException( MyGridConfiguration.PROP_MAX_FAILS );
			}
		} catch ( NumberFormatException nfe ) {
			throw new ConfigException( MyGridConfiguration.PROP_MAX_FAILS );
		}
	}


	/**
	 * Gets the number of task replicas allowed to be running at the same time.
	 * 
	 * @return The number of max fails
	 */
	public int getMaxFails() {

		return this.maxFails;
	}


	@Override
	protected void realSetCanceled( int jobId ) throws JobNotFoundException, JobCouldNotBeCancelledException {

		JobEntry job = getJobById( jobId );
		if ( job == null ) {
			throw new JobNotFoundException( jobId );
		}
		job.setCancelled();
		if ( !job.getStatus().equals( ExecutionStatus.CANCELLED ) ) {
			throw new JobCouldNotBeCancelledException( jobId + " could not be cancelled because it is already "
					+ job.getStatus() );
		}
		notifyJobInteresteds( job );
	}


	public void cleanFinishedJobs() {

		List<Integer> jobIds = new LinkedList<Integer>( this.jobs.keySet() );

		for ( int jobId : jobIds ) {
			ExecutionStatus jobState = getJobState( jobId );

			if ( !jobState.isRunning() && !jobState.isUnstarted() ) {
				try {
					removeJob( jobId );
				} catch ( JobCannotBeRemovedException e ) {
					LOG.error( "Unable to clean job " + jobId, e );
				}
			}
		}

	}


	@Override
	protected void realRemoveJob( int jobId ) throws JobCannotBeRemovedException {

		ExecutionStatus jobState = getJobState( jobId );
		if ( jobState.isRunning() || jobState.isUnstarted() ) {
			throw new JobCannotBeRemovedException();
		}
		jobs.remove( jobId );

	}


	@Override
	protected void realAddJob( int jobId, JobSpec jobSpec ) {

		JobEntry newJob = createJobEntry( jobId, jobSpec, maxReplicas, maxFails );
		addNewJob( jobId, newJob );
		LOG.info( "Job '" + jobId + "' was added." );
	}


	protected JobEntry createJobEntry( int jobId, JobSpec jobSpec, int maxReplicas, int maxFails ) {

		JobEntry newJob = new JobEntry( jobId, jobSpec, maxReplicas, maxFails );
		return newJob;
	}


	protected void addNewJob( int jobId, JobEntry newJob ) {

		jobs.put( new Integer( jobId ), newJob );
	}


	protected void setJobList( Map<Integer,JobEntry> newJobList ) {

		jobs = newJobList;

	}


	@Override
	protected void realShutdown() {

	}
}
