/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.jobmanager;

import java.io.Serializable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.ourgrid.common.exception.IllegalStateException;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.ReplicaExecutorResult;

/**
 * The Job is an abstraction representing a set of tasks that the user requested
 * to be executed.
 */
public class JobEntry implements Serializable {

	private static final long serialVersionUID = 33L;

	/** The Job identification. */
	private int id;

	/** The Maximum number a task can fail. */
	private int maxTaskFails;

	/** The Maximum number of replicas a task can have. */
	private int maxTaskReplicas;

	/** The Job specification. */
	private JobSpec spec;

	/** All job's tasks. */
	protected List<TaskEntry> tasks;

	/** Indicates that a job cancel was requested. */
	private boolean cancelRequested;


	/**
	 * This method constructs a new e constructor.
	 * 
	 * @param jobId The Job identification.
	 * @param jobSpec The job specification.
	 */
	public JobEntry( int jobId, JobSpec jobSpec, int maxReplicas, int maxFails ) {

		this.maxTaskReplicas = maxReplicas;
		this.maxTaskFails = maxFails;
		this.id = jobId;
		this.spec = jobSpec;
		this.cancelRequested = false;
		if ( jobSpec != null ) {
			List<TaskSpec> taskSpecs = jobSpec.getTaskSpecs();
			if ( taskSpecs != null ) {
				this.tasks = createTasks( taskSpecs );
			} else {
				this.tasks = new LinkedList<TaskEntry>();
			}
		} else {
			this.tasks = new LinkedList<TaskEntry>();
		}
	}


	/**
	 * Creates the job's tasks according the specification.
	 * 
	 * @param taskSpecs A list containing <code>TasksSpec</code>s.
	 * @return A list of <code>Task</code>s.
	 */
	protected List<TaskEntry> createTasks( List<TaskSpec> taskSpecs ) {

		List<TaskEntry> tasks = new LinkedList<TaskEntry>();
		if ( taskSpecs != null ) {
			int nextTaskId = 0;
			for ( TaskSpec taskSpec : taskSpecs ) {
				nextTaskId++;
				TaskEntry task = createTaskEntry( taskSpec, nextTaskId, this.id, maxTaskFails, maxTaskReplicas );
				tasks.add( task );
			}
		}
		return tasks;
	}


	protected TaskEntry createTaskEntry( TaskSpec taskSpec, int nextTaskId, int id, int maxTaskFails,
											int maxTaskReplicas ) {

		TaskEntry task = new TaskEntry( taskSpec, nextTaskId, id, maxTaskFails, maxTaskReplicas );
		return task;
	}


	/**
	 * Returns the job identification.
	 * 
	 * @return The job identification.
	 */
	public int getId() {

		return this.id;
	}


	/**
	 * Sets the replica contained in the <code>ReplicaExecutorResult</code> to
	 * the <i>finished </i> state.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	protected void replicaFinished( ReplicaExecutorResult result ) {

		ReplicaEntry replica = result.getReplica();
		TaskEntry task = getTaskById( replica.getTaskId() );
		task.replicaFinished( result );
	}


	/**
	 * Sets the replica contained in the <code>ReplicaExecutorResult</code> to
	 * the <i>failed </i> state.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	protected void replicaFailed( ReplicaExecutorResult result ) {

		ReplicaEntry replica = result.getReplica();
		TaskEntry task = getTaskById( replica.getTaskId() );
		task.replicaFailed( result );
	}


	/**
	 * Sets the replica contained in the <code>ReplicaExecutorResult</code> to
	 * the <i>aborted </i> state.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	protected void replicaAborted( ReplicaExecutorResult result ) {

		ReplicaEntry replica = result.getReplica();
		TaskEntry task = getTaskById( replica.getTaskId() );
		task.replicaAborted( result );
	}


	/**
	 * Sets the replica contained in the <code>ReplicaExecutorResult</code> to
	 * the <i>canceled </i> state.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	protected void replicaCancelled( ReplicaExecutorResult result ) {

		ReplicaEntry replica = result.getReplica();
		TaskEntry task = getTaskById( replica.getTaskId() );
		task.replicaCanceled( result );
	}


	/**
	 * Sets the job as canceled.
	 */
	protected void setCancelled() {

		ExecutionStatus state = getStatus();
		if ( state.equals( ExecutionStatus.UNSTARTED ) || state.equals( ExecutionStatus.RUNNING ) ) {
			this.cancelRequested = true;
			Iterator<TaskEntry> tasksIt = tasks.iterator();
			while ( tasksIt.hasNext() ) {
				TaskEntry task = tasksIt.next();
				task.setCancelled();
			}
		}
	}


	/**
	 * Returns a <code>Task</code> by its identification.
	 * 
	 * @param taskId The <code>Task</code>'s identification.
	 * @return A <code>Task</code> with the specified identification
	 */
	public TaskEntry getTaskById( int taskId ) {

		return tasks.get( taskId - 1 );
	}


	/**
	 * Returns the <code>Job</code> state according its <code>Task</code>s'
	 * states.
	 * 
	 * @return The <code>Job</code> state.
	 */
	public ExecutionStatus getStatus() {

		if ( cancelRequested ) {
			return ExecutionStatus.CANCELLED;
		}

		boolean hasReadyTask = false;
		boolean hasFinishedTask = false;
		boolean hasRunningTask = false;
		boolean hasFailedTask = false;
		boolean hasCanceledTask = false;

		Iterator<TaskEntry> tasksIt = tasks.iterator();
		while ( tasksIt.hasNext() ) {
			TaskEntry task = tasksIt.next();
			ExecutionStatus taskState = task.getState();
			if ( taskState.equals( ExecutionStatus.UNSTARTED ) ) {
				hasReadyTask = true;
			} else if ( taskState.equals( ExecutionStatus.CANCELLED ) ) {
				hasCanceledTask = true;
			} else if ( taskState.equals( ExecutionStatus.FINISHED ) ) {
				hasFinishedTask = true;
			} else if ( taskState.equals( ExecutionStatus.RUNNING ) ) {
				hasRunningTask = true;
			} else if ( taskState.equals( ExecutionStatus.FAILED ) ) {
				hasFailedTask = true;
			}
		}
		boolean isJobReady = hasReadyTask && !hasRunningTask && !hasFinishedTask && !hasCanceledTask && !hasFailedTask;
		if ( isJobReady ) {
			return ExecutionStatus.UNSTARTED;
		}
		boolean isJobRunning = hasRunningTask
				|| (hasReadyTask && (hasFailedTask || hasFinishedTask || hasCanceledTask));
		if ( isJobRunning ) {
			return ExecutionStatus.RUNNING;
		}
		boolean isJobFailed = hasFailedTask && !hasRunningTask && !hasReadyTask && !hasCanceledTask;
		if ( isJobFailed ) {
			return ExecutionStatus.FAILED;
		}
		boolean isJobCanceled = hasCanceledTask && !hasRunningTask && !hasReadyTask;
		if ( isJobCanceled ) {
			return ExecutionStatus.CANCELLED;
		}
		boolean isJobFinished = hasFinishedTask && !hasRunningTask && !hasReadyTask && !hasCanceledTask
				&& !hasFailedTask;
		if ( isJobFinished ) {
			return ExecutionStatus.FINISHED;
		}

		throw new IllegalStateException( "Job is in an illegal state" );
	}


	/**
	 * Returns the <code>Job</code>'s<code>Task</code>s.
	 * 
	 * @return A <code>List</code> containing all the <code>Task</code>s.
	 */
	public List<TaskEntry> getTasks() {

		return new LinkedList<TaskEntry>( tasks );
	}


	/**
	 * Creates and returns a new <code>Replica</code>.
	 * 
	 * @return A new <code>Replica</code> in case that a replication is
	 *         possible to be made, <code>null</code> otherwise.
	 */
	protected ReplicaEntry getNewReplica( TaskEntry task ) {

		TaskEntry myTask = getTaskById( task.getId() );
		if ( myTask != null ) {
			return myTask.getNewReplica();
		}
		return null;
	}


	/**
	 * Returns the <code>Job</code>'s available <code>Task</code>s. If a
	 * task can be replicated, it is included in this list.
	 * 
	 * @return A <code>List</code> containing all available <code>Task</code>
	 *         s.
	 */
	public List<TaskEntry> getAvailableTasks() {

		LinkedList<TaskEntry> availableTasks = new LinkedList<TaskEntry>();
		if ( !this.getStatus().isCancelled() ) {
			Iterator<TaskEntry> tasksIt = tasks.iterator();
			while ( tasksIt.hasNext() ) {
				TaskEntry task = tasksIt.next();
				if ( task.canReplicate() ) {
					availableTasks.add( task );
				}
			}
		}

		return availableTasks;
	}


	public JobSpec getJobSpec() {

		return spec;
	}


	/**
	 * Returns the job requirements expression.
	 * 
	 * @return The job requirements expression.
	 */
	public String getRequirements() {

		return this.spec.getRequirements();
	}


	/**
	 * Returns the job label.
	 * 
	 * @return The label
	 */
	public String getLabel() {

		String label = spec.getLabel();
		if ( label == null || label == "" ) {
			label = Integer.toString( id );
		}

		return label;
	}
}
