package org.ourgrid.mygrid.scheduler.jobmanager;

import java.util.Collection;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.JobInterested;
import org.ourgrid.mygrid.scheduler.ReplicaExecutorResult;
import org.ourgrid.mygrid.scheduler.exception.JobCannotBeRemovedException;
import org.ourgrid.mygrid.scheduler.exception.JobCouldNotBeCancelledException;
import org.ourgrid.mygrid.scheduler.exception.JobNotFoundException;

public interface JobManager {

	/**
	 * Adds a new <code>Job</code>.
	 * 
	 * @param jobSpec The job specification.
	 */
	public void addJob( int jobId, JobSpec jobSpec );


	/**
	 * Sets the <code>Replica</code> state to <i>finished</i>. It also
	 * notifies <code>JobInteresteds</code> case it is necessary.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	public void replicaFinished( ReplicaExecutorResult result );


	/**
	 * Sets the <code>Replica</code> state to <i>failed</i>. It also notifies
	 * <code>JobInteresteds</code> case it is necessary.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	public void replicaFailed( ReplicaExecutorResult result );


	/**
	 * Sets the <code>Replica</code> state to <i>aborted</i>. It also
	 * notifies <code>JobInteresteds</code> case it is necessary.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	public void replicaAborted( ReplicaExecutorResult result );


	/**
	 * Sets the <code>Replica</code> state to <i>canceled</i>. It also
	 * notifies <code>JobInteresteds</code> case it is necessary.
	 * 
	 * @param result The <code>ReplicaExecutorResult</code>.
	 */
	public void replicaCanceled( ReplicaExecutorResult result );


	/**
	 * Sets the given job as canceled.
	 * 
	 * @param jobId The job identification
	 * @throws JobNotFoundException when the job was not found
	 * @throws JobCouldNotBeCancelledException when the job could not be
	 *         canceled
	 */
	public void setCanceled( int jobId ) throws JobNotFoundException, JobCouldNotBeCancelledException;


	/**
	 * Clean the list of terminated jobs
	 */
	public void cleanFinishedJobs();


	/**
	 * Remove the information about the specified job
	 * 
	 * @param jobId
	 * @throws JobCannotBeRemovedException
	 */
	public void removeJob( int jobId ) throws JobCannotBeRemovedException;


	/**
	 * Returns the <code>Job</code> list.
	 * 
	 * @return The <code>Job</code> list.
	 */
	public Collection<JobEntry> jobList();


	public Collection<JobEntry> clonedJobList();


	public JobEntry getJobById( int jobId );


	public ReplicaEntry createNewReplica( TaskEntry task );


	public ExecutionStatus getJobState( int jobID );


	public void addJobInterested( JobInterested jobInterested );


	public void setMaxReplicas( String string ) throws ConfigException;


	public void setMaxFails( String string ) throws ConfigException;


	public void addNewListener( JobManagerListenerRemote newListener );


	public void replicaRunning( ReplicaEntry replica, GumID gumID );


	public void removeListener( JobManagerListenerRemote listener );


	public void shutdown();
}
