/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.jobmanager;

import java.io.Serializable;

import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.ReplicaExecutorResult;

/**
 * A replica is the lower level abstraction that is executed on the
 * <code>GridMachine</code>s. It contains all necessary information to run a
 * a job task. All replicas of a task differ only in their identification, the
 * remaining information is the same to all replicas. In other words, a replica
 * is a representation of a task to be executed on the remote machines.
 */
public class ReplicaEntry implements Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/** The task specification. */
	private TaskSpec spec;

	/** The identification of this replica. */
	private int id;

	/** The identification of the owner task. */
	private int taskId;

	/** The identification of the owner job. */
	private int jobId;

	/** The current state of this replica. */
	private ExecutionStatus status;

	/** The result of the replica execution. */
	private ReplicaExecutorResult replicaExecutorResult;

	/** The <code>GridMachine</code> name assigned to execute this replica. */
	private GumID gumID;


	/**
	 * The constructor.
	 * 
	 * @param id The replica id
	 * @param taskId The task id
	 * @param jobId The job id
	 */
	public ReplicaEntry( int id, int taskId, int jobId, TaskSpec taskSpec ) {

		this.jobId = jobId;
		this.taskId = taskId;
		this.id = id;
		this.status = ExecutionStatus.UNSTARTED;
		this.replicaExecutorResult = null;
		this.spec = taskSpec;
		this.gumID = null;
	}


	/**
	 * Returns the <code>TaskSpec</code> of this replica.
	 * 
	 * @return The <code>TaskSpec</code>
	 */
	public TaskSpec getSpec() {

		return spec;
	}


	/**
	 * Sets the state of the replica.
	 * 
	 * @param state The new state of the replica
	 */
	public void setStatus( ExecutionStatus state ) {

		this.status = state;
	}


	/**
	 * Returns the state of the replica.
	 * 
	 * @return The replica current status
	 */
	public ExecutionStatus getStatus() {

		return this.status;
	}


	/**
	 * Returns the owner job identification.
	 * 
	 * @return The job identification
	 */
	public int getJobId() {

		return this.jobId;
	}


	/**
	 * Returns the owner task identification.
	 * 
	 * @return The task identification
	 */
	public int getTaskId() {

		return this.taskId;
	}


	/**
	 * Returns the identification of this replica.
	 * 
	 * @return The identification
	 */
	public int getId() {

		return id;
	}


	/**
	 * Sets the ReplicaExecutorResult of this replica.
	 * 
	 * @param result The result of the replica execution
	 */
	public void setResult( ReplicaExecutorResult result ) {

		this.replicaExecutorResult = result;
	}


	/**
	 * Gets the ReplicaExecutorResult of this replica.
	 * 
	 * @return The result of the replica execution
	 */
	public ReplicaExecutorResult getResult() {

		return this.replicaExecutorResult;
	}


	/**
	 * Returns a unique string representation for this replica.
	 * 
	 * @return A string representation for this replica.
	 */
	@Override
	public String toString() {

		return getJobId() + "." + getTaskId() + "." + getId();
	}


	@Override
	public boolean equals( Object o ) {

		if ( o != null && o instanceof ReplicaEntry ) {
			ReplicaEntry otherReplica = (ReplicaEntry) o;
			return jobId == otherReplica.jobId && taskId == otherReplica.taskId && id == otherReplica.id;
		}

		return false;
	}


	/**
	 * Returns the ID of the <code>GridMachine</code> assigned to execute this
	 * replica.
	 * 
	 * @return The <code>GridMachine</code> ID.
	 */
	public GumID getGumID() {

		return this.gumID;
	}


	/**
	 * Sets the ID of the <code>GridMachine</code> assigned to execute this
	 * replica.
	 * 
	 * @param gumID The <code>GridMachine</code> ID.
	 */
	public void setGumID( GumID gumID ) {

		this.gumID = gumID;
	}

}
