/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.jobmanager;

import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.ourgrid.common.exception.IllegalStateException;
import org.ourgrid.common.exception.UnableToCreatePlaypenException;
import org.ourgrid.common.exception.UnableToCreateStorageException;
import org.ourgrid.common.exception.UnavailableGumException;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.ReplicaExecutorResult;

/**
 * An abstraction to represent a task of a job sent by user to be executed. A
 * task can be made of <i>n</i> replicas defined by <code>maxReplicas</code>.
 * A replica is the lower level abstraction that is executed on the
 * <code>GridMachine</code>s.
 */
public class TaskEntry implements Comparable<TaskEntry>, Serializable {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/** The <code>Job</code> identification. */
	private int jobId;

	/** The <code>Task</code> identification. */
	private int id;

	/** The maximum number of running replicas the <code>Task</code> can have. */
	private int maxReplicas;

	/** The maximum number of replica fails a <code>Task</code> can have. */
	private int maxFails;

	/** The current number of replica fails. */
	private int actualFails;

	/** This <code>Task</code>'s <code>Replica</code>s. */
	protected List<ReplicaEntry> replicas;

	/** The <code>Task</code> specification. */
	private TaskSpec taskSpec;

	/***************************************************************************
	 * Boolean to identify that a task is canceled. It is necessary when there
	 * is no replica to check if the task is canceled
	 **************************************************************************/
	private boolean cancelRequested;


	/**
	 * The constructor.
	 * 
	 * @param taskSpec The <code>Task</code> specification.
	 * @param id This <code>Task</code> identification
	 * @param jobId The <code>Job</code> identification.
	 * @param maxFails The maximum number of replicas' fails allowed.
	 * @param maxReplicas The maximum number of running <code>Replica</code>s
	 *        permitted.
	 */
	public TaskEntry( TaskSpec taskSpec, int id, int jobId, int maxFails, int maxReplicas ) {

		this.taskSpec = taskSpec;
		this.id = id;
		this.jobId = jobId;
		this.replicas = new LinkedList<ReplicaEntry>();
		this.actualFails = 0;
		this.maxFails = maxFails;
		this.maxReplicas = maxReplicas;
		this.cancelRequested = false;
	}


	/**
	 * Returns the <code>Task</code> identification.
	 * 
	 * @return The <code>Task</code> identification.
	 */
	public int getId() {

		return id;
	}


	/**
	 * Returns the job identification of this <code>Task</code>.
	 * 
	 * @return The job id of this <code>Task</code>.
	 */
	public int getJobId() {

		return jobId;
	}


	/**
	 * Returns the current number of replica fails.
	 * 
	 * @return The current number of replica fails
	 */
	public int getActualFails() {

		return this.actualFails;
	}


	/**
	 * Returns the maximum number of replica fails a <code>Task</code> can
	 * have.
	 * 
	 * @return The maximum number of replica fails
	 */
	public int getMaxFails() {

		return this.maxFails;
	}


	/**
	 * Returns the maximum number of running replicas the <code>Task</code>
	 * can have.
	 * 
	 * @return The maximum number of running replicas
	 */
	public int getMaxReplicas() {

		return this.maxReplicas;
	}


	/**
	 * Returns the <code>Task</code>'s <code>Replica</code>s.
	 * 
	 * @return The <code>Collection</code> containing all the
	 *         <code>Replica</code>s.
	 */
	public Collection<ReplicaEntry> getReplicas() {

		return new LinkedList<ReplicaEntry>( replicas );
	}


	/**
	 * Creates and returns a new <code>Replica</code>.
	 * 
	 * @return A new <code>Replica</code> in case that a replication is
	 *         possible to be made, <code>null</code> otherwise.
	 */
	public ReplicaEntry getNewReplica() {

		if ( canReplicate() ) {
			ReplicaEntry replica = createReplicaEntry( replicas.size() + 1, this.id, this.jobId, taskSpec );
			addNewReplica( replica );

			return replica;
		}
		return null;
	}


	protected ReplicaEntry createReplicaEntry( int id, int taskId, int jobId, TaskSpec taskSpec ) {

		ReplicaEntry replica = new ReplicaEntry( id, taskId, jobId, taskSpec );
		return replica;
	}


	/**
	 * @param replica
	 */
	private void addNewReplica( ReplicaEntry replica ) {

		replicas.add( replica );
	}


	/**
	 * Returns the number of running <code>Replica</code>s.
	 * 
	 * @return The number of running <code>Replica</code>s.
	 */
	public int getNumberOfRunningReplicas() {

		return getNumberOfReplicasByState( ExecutionStatus.RUNNING );
	}


	/**
	 * Returns the number of unstarted <code>Replica</code>s.
	 * 
	 * @return The number of unstarted <code>Replica</code>s.
	 */
	public int getNumberOfUnstartedReplicas() {

		return getNumberOfReplicasByState( ExecutionStatus.UNSTARTED );
	}


	/**
	 * Returns the number of <code>Replica</code>s in the specified state.
	 * 
	 * @param state A possible <code>Replica</code> state.
	 * @return The number of <code>Replica</code>s in that state.
	 */
	private int getNumberOfReplicasByState( ExecutionStatus state ) {

		int counter = 0;
		Iterator<ReplicaEntry> replicasIt = replicas.iterator();
		while ( replicasIt.hasNext() ) {
			ReplicaEntry replica = replicasIt.next();
			if ( replica.getStatus().equals( state ) ) {
				counter++;
			}
		}

		return counter;
	}


	/**
	 * Verifies if a task replica can be done.
	 * 
	 * @return True if it is possible to create a task replica, false otherwise
	 */
	public boolean canReplicate() {

		if ( !hasCanceledReplica() && !hasFinishedReplica() ) {
			if ( ((getNumberOfRunningReplicas() + getNumberOfUnstartedReplicas()) < maxReplicas)
					&& (this.actualFails < this.maxFails) ) {
				return true;
			}
			return false;
		}
		return false;
	}


	/**
	 * Sets a <code>Replica</code> state to <i>Finished</i>, if the replica
	 * is 'Running'.
	 * 
	 * @param result The result of the replica execution
	 */
	public void replicaFinished( ReplicaExecutorResult result ) {

		changeReplicaStateIfItIsRunning( result, ExecutionStatus.FINISHED );
	}


	/**
	 * Sets a <code>Replica</code> state to <i>Failed</i>, if the replica is
	 * 'Running'.
	 * 
	 * @param result The result of the replica execution
	 */
	public void replicaFailed( ReplicaExecutorResult result ) {

		changeReplicaStateIfItIsRunning( result, ExecutionStatus.FAILED );
	}


	/**
	 * Sets a <code>Replica</code> state to <i>Aborted</i>, if the replica is
	 * 'Running'.
	 * 
	 * @param result The result of the replica execution
	 */
	public void replicaAborted( ReplicaExecutorResult result ) {

		changeReplicaStateIfItIsRunning( result, ExecutionStatus.ABORTED );
	}


	/**
	 * Changes the state of a replica to 'state' if it was 'Running'.
	 * 
	 * @param result The ReplicaExecutorResult
	 * @param state The new state of the replica
	 */
	private void changeReplicaStateIfItIsRunning( ReplicaExecutorResult result, ExecutionStatus state ) {

		ReplicaEntry replicaCopy = result.getReplica();
		ReplicaEntry replica = getReplicaById( replicaCopy.getId() );
		if ( replica != null ) {
			replica.setResult( result );
			if ( replica.getStatus().isRunning() ) {
				if ( state.equals( ExecutionStatus.FAILED ) ) {
					Throwable errorCause = result.getErrorCause();
					if ( !(errorCause instanceof UnavailableGumException
							|| errorCause instanceof UnableToCreatePlaypenException || errorCause instanceof UnableToCreateStorageException) ) {
						actualFails++;
					}
				}
				replica.setStatus( state );
			}
		}
	}


	/**
	 * Sets a <code>Replica</code> state to <i>Canceled by the User</i>, if
	 * the replica is 'Running' or 'Unstarted'.
	 * 
	 * @param result The result of the replica execution
	 */
	public void replicaCanceled( ReplicaExecutorResult result ) {

		ReplicaEntry replicaCopy = result.getReplica();
		ReplicaEntry replica = getReplicaById( replicaCopy.getId() );
		if ( replica != null ) {
			replica.setResult( result );
			replica.setStatus( ExecutionStatus.CANCELLED );
		}
	}


	/**
	 * Returns the <code>Task</code> state.
	 * 
	 * @return The <code>Task</code> state.
	 * @throws IllegalStateException - Case the <code>Task</code> is not in
	 *         any of the expected states.
	 */
	public ExecutionStatus getState() {

		if ( cancelRequested ) {
			return ExecutionStatus.CANCELLED;
		}

		boolean hasUnstartedReplica = false;
		boolean hasFinishedReplica = false;
		boolean hasRunningReplica = false;
		boolean hasFailedReplica = false;
		boolean hasCanceledReplica = false;
		boolean hasAbortedReplica = false;

		Iterator<ReplicaEntry> replicasIt = replicas.iterator();
		while ( replicasIt.hasNext() ) {
			ReplicaEntry replica = replicasIt.next();
			if ( replica.getStatus().equals( ExecutionStatus.CANCELLED ) ) {
				hasCanceledReplica = true;
			} else if ( replica.getStatus().equals( ExecutionStatus.ABORTED ) ) {
				hasAbortedReplica = true;
			} else if ( replica.getStatus().equals( ExecutionStatus.FINISHED ) ) {
				hasFinishedReplica = true;
			} else if ( replica.getStatus().equals( ExecutionStatus.RUNNING ) ) {
				hasRunningReplica = true;
			} else if ( replica.getStatus().equals( ExecutionStatus.FAILED ) ) {
				hasFailedReplica = true;
			} else if ( replica.getStatus().equals( ExecutionStatus.UNSTARTED ) ) {
				hasUnstartedReplica = true;
			}
		}

		boolean isInIllegalState = hasAbortedReplica && !hasFinishedReplica;
		if ( isInIllegalState ) {
			throw new IllegalStateException( "Task is in an illegal state" );
		}

		boolean isTaskReady = !hasRunningReplica && !hasFinishedReplica && !hasCanceledReplica && !hasFailedReplica
				&& !hasAbortedReplica;
		if ( isTaskReady ) {
			return ExecutionStatus.UNSTARTED;
		}

		boolean isTaskFailed = this.actualFails >= this.maxFails && !hasRunningReplica && !hasFinishedReplica;
		if ( isTaskFailed ) {
			return ExecutionStatus.FAILED;
		}

		boolean isTaskCanceled = hasCanceledReplica && !hasRunningReplica && !hasUnstartedReplica;
		if ( isTaskCanceled ) {
			return ExecutionStatus.CANCELLED;
		}

		boolean isTaskFinished = hasFinishedReplica && !hasRunningReplica && !hasUnstartedReplica;
		if ( isTaskFinished ) {
			return ExecutionStatus.FINISHED;
		}

		return ExecutionStatus.RUNNING;

	}


	/**
	 * Verifies if there exists a canceled replica.
	 * 
	 * @return True if there exists a canceled replica, false otherwise
	 */
	private boolean hasCanceledReplica() {

		return hasReplicaInState( ExecutionStatus.CANCELLED );
	}


	/**
	 * Verifies if there exists a finished replica.
	 * 
	 * @return True if there exists a finished replica, false otherwise
	 */
	private boolean hasFinishedReplica() {

		return hasReplicaInState( ExecutionStatus.FINISHED );
	}


	/**
	 * Verifies if there exists a replica in the specified state.
	 * 
	 * @param status The state to be verified
	 * @return True if there exists a replica in the specified state, false
	 *         otherwise
	 */
	private boolean hasReplicaInState( ExecutionStatus status ) {

		for ( ReplicaEntry replica : replicas ) {
			if ( replica.getStatus().equals( status ) ) {
				return true;
			}
		}

		return false;
	}


	/**
	 * Returns the <code>Replica</code> with the specified identification.
	 * 
	 * @param id The <code>Replica</code> identification.
	 * @return The <code>Replica</code> case the identification exists,
	 *         <code>null</code> otherwise.
	 */
	public ReplicaEntry getReplicaById( int id ) {

		for ( ReplicaEntry replica : replicas ) {
			if ( replica.getId() == id ) {
				return replica;
			}
		}

		return null;
	}


	/**
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo( TaskEntry objectTask ) {

		TaskEntry task = objectTask;

		if ( this.getNumberOfRunningReplicas() < task.getNumberOfRunningReplicas() ) {
			return -1;
		} else if ( this.getNumberOfRunningReplicas() > task.getNumberOfRunningReplicas() ) {
			return 1;
		} else {
			if ( this.getJobId() < task.getJobId() ) {
				return -1;
			} else if ( this.getJobId() > task.getJobId() ) {
				return 1;
			} else {
				if ( this.getId() < task.getId() ) {
					return -1;
				} else if ( this.getId() > task.getId() ) {
					return 1;
				}
			}
		}
		return 0;
	}


	@Override
	public String toString() {

		return getJobId() + "." + getId();
	}


	/**
	 * Sets the task as canceled.
	 */
	public void setCancelled() {

		ExecutionStatus status = this.getState();
		if ( status.equals( ExecutionStatus.UNSTARTED ) || status.equals( ExecutionStatus.RUNNING ) ) {
			this.cancelRequested = true;
		}
	}


	@Override
	public boolean equals( Object o ) {

		if ( o != null && o instanceof TaskEntry ) {
			TaskEntry taskEntry = (TaskEntry) o;
			return jobId == taskEntry.jobId && id == taskEntry.id && taskEntry.taskSpec.equals( taskSpec );
		}

		return false;
	}


	/**
	 * @return this task's TaskSpec
	 */
	public TaskSpec getSpec() {

		return this.taskSpec;
	}
}
