/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.test;

import java.io.File;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.exception.UnableToDigestFileException;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.util.JavaFileUtil;
import org.ourgrid.mygrid.scheduler.DataDiscovery;
import org.ourgrid.mygrid.scheduler.DataDiscoveryEntry;
import org.ourgrid.mygrid.scheduler.exception.DDEntryNotFoundException;
import org.ourgrid.mygrid.scheduler.exception.DDInitException;
import org.ourgrid.test.SchedulerFastTest;

public class DataDiscoveryTest extends TestCase implements SchedulerFastTest {

	public static final String TEST_ALL_TEST_TYPE = "UNIT";

	GumSpec spec;

	String remotePath = "/this/is/my/file";

	String[ ] localFile = { "File1", "File2", "File3" };

	String[ ] machines = { "GuM1", "GuM2", "GuM3", "GuM4", "GuM5", "GuM6" };

	long myDay = System.currentTimeMillis();


	public DataDiscoveryTest( String s ) {

		super( s );
	}


	protected void setUp() throws Exception {

		super.setUp();
		Configuration.getInstance( Configuration.MYGRID );
		boolean ddCache = false;
		localFile[0] = "test/specs/dd/file.0";
		localFile[1] = "test/specs/dd/file.1";
		localFile[2] = "test/specs/dd/file.2";
		try {
			DataDiscovery.initDataDiscovery( ddCache );
		} catch ( DDInitException e ) {
			DataDiscovery.getInstance();
		}
	}


	protected void tearDown() throws Exception {

		super.tearDown();
	}


	public void testPutDDEntry() {

		DataDiscovery.resetInstance();
		DataDiscovery myDD = DataDiscovery.getInstance();
		int ndx = 0;
		String digested = null;
		File myFile = new File( localFile[0] );
		long size;
		try {
			digested = JavaFileUtil.getDigestRepresentation( myFile );
		} catch ( UnableToDigestFileException exception ) {
			System.err.println( exception );
		}
		size = myFile.length();
		/*
		 * Put file 1 on GuM[1-3] at Site1 Site1 has shared storage file 1
		 * should be available to all GuMs at Site1
		 */
		for ( ndx = 0; ndx < 3; ndx++ ) {
			spec = new GumSpec();
			spec.putAttribute( GumSpec.ATT_STORAGE_SHARED, "yes" );
			spec.putAttribute( GumSpec.ATT_PEERNAME, "Site1" );
			spec.putAttribute( GumSpec.ATT_NAME, machines[ndx] );
			myDD.putDDEntry( digested, remotePath, size, spec );
		}

		File myOFile = new File( localFile[1] );
		try {
			digested = JavaFileUtil.getDigestRepresentation( myOFile );
		} catch ( UnableToDigestFileException exception ) {
			System.err.println( exception );
		}
		size = myOFile.length();
		/*
		 * Put file 2 on GuM[1-5] at Site2 Site2 does not have shared storage
		 * file 2 should only be available to GuM[1-5] at Site2
		 */
		for ( ndx = 0; ndx < 5; ndx++ ) {
			spec = new GumSpec();
			spec.putAttribute( GumSpec.ATT_STORAGE_SHARED, null );
			spec.putAttribute( GumSpec.ATT_PEERNAME, "Site2" );
			spec.putAttribute( GumSpec.ATT_NAME, machines[ndx] );
			myDD.putDDEntry( digested, remotePath, size, spec );
		}
		/*
		 * Put file 2 on GuM5 at Site 3 Site3 has shared storage file 2 should
		 * be available to all GuMs at Site3
		 */
		spec = new GumSpec();
		spec.putAttribute( GumSpec.ATT_STORAGE_SHARED, "yes" );
		spec.putAttribute( GumSpec.ATT_PEERNAME, "Site3" );
		myDD.putDDEntry( digested, remotePath, size, spec );
		assertEquals( "There should be this many entries recorded", 2, myDD.numberOfEntries() );
	}


	public void testGetDDEntry() {

		DataDiscovery myDD = DataDiscovery.getInstance();
		DataDiscoveryEntry entry = null;
		String digest = "somejunk";
		String digest1 = null;
		String digest2 = null;
		File myOFile = new File( localFile[0] );
		try {
			digest1 = JavaFileUtil.getDigestRepresentation( myOFile );
		} catch ( UnableToDigestFileException exception ) {
			System.err.println( exception );
		}
		File myFile = new File( localFile[1] );
		try {
			digest2 = JavaFileUtil.getDigestRepresentation( myFile );
		} catch ( UnableToDigestFileException exception ) {
			System.err.println( exception );
		}

		try {
			entry = myDD.getDDEntry( digest1 );
			assertEquals( "The file should be at this many locations", 1, entry.getLocations().size() );
			// System.out.println(entry.toString());
			entry = myDD.getDDEntry( digest2 );
			assertEquals( "The file should be at this many locations", 6, entry.getLocations().size() );
			// System.out.println(entry.toString());
			entry = myDD.getDDEntry( digest );
		} catch ( DDEntryNotFoundException noDD ) {
			String msg = noDD + "\n";
			assertEquals(
					"Should return the following msg",
					"org.ourgrid.mygrid.scheduler.exception.DDEntryNotFoundException: No DataDiscoveryEntry with such key!\n",
					msg );
		}

	}


	public void testModEntry() {

		DataDiscovery myDD = DataDiscovery.getInstance();
		File myFile = new File( localFile[1] );
		String digest = null;

		try {
			digest = JavaFileUtil.getDigestRepresentation( myFile );
		} catch ( UnableToDigestFileException exception ) {
			System.err.println( exception );
		}

		spec = new GumSpec();
		spec.putAttribute( GumSpec.ATT_STORAGE_SHARED, "no" );
		spec.putAttribute( GumSpec.ATT_PEERNAME, "Site2" );
		spec.putAttribute( GumSpec.ATT_NAME, machines[2] );

		myDD.modDDEntry( digest, remotePath, spec );
		try {
			DataDiscoveryEntry entry = myDD.getDDEntry( digest );
			assertEquals( "The file should be at this many locations", 5, entry.getLocations().size() );
			// System.out.println(entry.toString());
		} catch ( DDEntryNotFoundException noDD ) {
			return;
		}
	}


	// public void testFileAmount() throws GuMSpecificationException {
	//		
	// DataDiscovery myDD = DataDiscovery.getInstance();
	// GridMachine machine;
	// long noFile = 0L;
	// long fileSize, oneSize, twoSize;
	//
	// File fileOne = new File(localFile[0]);
	// oneSize = fileOne.length();
	// File fileTwo = new File(localFile[1]);
	// twoSize = fileTwo.length();
	//
	// /*
	// * Looking for file 1 on GuM3 at Site1. Site1 does have shared storage
	// */
	// spec.putAttribute(GuMSpec.ATT_NAME, machines[2]);
	// spec.putAttribute(GuMSpec.ATT_STORAGE_SHARED, "yes");
	// spec.putAttribute(GuMSpec.ATT_SITE, "Site1");
	// machine = GuMSpecServices.createGuM(spec);
	//	
	// fileSize = myDD.amountAtDestination(localFile[0], remotePath, machine);
	// assertEquals("These file sizes should match", oneSize, fileSize);
	//
	// /*
	// * Looking for file 1 on GuM6 at Site1. Site1 does have shared storage
	// */
	// spec.putAttribute(GuMSpec.ATT_NAME, machines[5]);
	// machine = GuMSpecServices.createGuM(spec);
	// fileSize = myDD.amountAtDestination(localFile[0], remotePath, machine);
	// assertEquals("These file sizes should match", oneSize, fileSize);
	//
	// /*
	// * Looking for file 2 on GuM6 at Site1. Site1 does have shared storage
	// */
	// fileSize = myDD.amountAtDestination(localFile[1], remotePath, machine);
	// assertEquals("These file sizes should match", noFile, fileSize);
	//
	// /*
	// * Looking for file 1 on GuM6 at Site2. Site2 does not have shared
	// * storage
	// */
	// spec.putAttribute(GuMSpec.ATT_SITE, "Site2");
	// spec.putAttribute(GuMSpec.ATT_STORAGE_SHARED, "no");
	// machine = GuMSpecServices.createGuM(spec);
	// fileSize = myDD.amountAtDestination(localFile[0], remotePath, machine);
	// assertEquals("These file sizes should match", noFile, fileSize);
	//
	// /*
	// * Looking for file 2 on GuM6 at Site2. Site2 does not have shared
	// * storage
	// */
	// fileSize = myDD.amountAtDestination(localFile[1], remotePath, machine);
	// assertEquals("These file sizes should match", noFile, fileSize);
	//
	// /*
	// * Looking for file 2 on GuM2 at Site2. Site2 does not have shared
	// * storage
	// */
	// spec.putAttribute(GuMSpec.ATT_NAME, machines[1]);
	// machine = GuMSpecServices.createGuM(spec);
	// fileSize = myDD.amountAtDestination(localFile[1], remotePath, machine);
	// assertEquals("These file sizes should match", twoSize, fileSize);
	//
	// /*
	// * Looking for file 2 on GuM2 at Site3. Site3 does have shared storage
	// */
	// spec.putAttribute(GuMSpec.ATT_SITE, null);
	// spec.putAttribute(GuMSpec.ATT_STORAGE_SHARED, "yes");
	// machine = GuMSpecServices.createGuM(spec);
	// fileSize = myDD.amountAtDestination(localFile[1], remotePath, machine);
	// assertEquals("These file sizes should match", twoSize, fileSize);
	//
	// /*
	// * Looking for file 3 on GuM2 at Site3. Site 3 does have shared storage
	// */
	// fileSize = myDD.amountAtDestination(localFile[2], remotePath, machine);
	// assertEquals("These file sizes should match", noFile, fileSize);
	//		
	// }

	public static void main( String[ ] args ) {

		junit.swingui.TestRunner.run( DataDiscoveryTest.class );
	}
}
