/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.test;

import java.rmi.Naming;
import java.rmi.registry.LocateRegistry;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.gump.BindableMockGumpWrapper;
import org.ourgrid.common.gump.Gump;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.PeerSpec;
import org.ourgrid.mygrid.replicaexecutor.test.FakeEBReplicaExecutorFacade;
import org.ourgrid.mygrid.scheduler.BlackListManager;
import org.ourgrid.mygrid.scheduler.FakeEBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.JobRequestEntry;
import org.ourgrid.mygrid.scheduler.PeerEntry;
import org.ourgrid.mygrid.scheduler.PeerEntry.PeerState;
import org.ourgrid.mygrid.scheduler.gridmanager.EBGridManager;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerEntry;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerListenerRemote;
import org.ourgrid.mygrid.scheduler.gump.MyGridGumpSpecServices;
import org.ourgrid.test.SchedulerFastTest;
import org.ourgrid.test.TestUtils;

public class EBGridManagerTest extends TestCase implements SchedulerFastTest {

	private static final String MOCK_GUM_NAME1 = "MockGum1";

	private static final String MOCK_GUM_NAME2 = "MockGum2";

	private GumClient mockGumClient1;

	private FakeEBSchedulerFacade fakeEBSchedulerFacade;

	private GumClient mockGumClient2;

	private BindableMockGumpWrapper bindableMockGumpWrapper;

	private BindableMockGumpWrapper bindableMockGumpWrapper2;

	private Gump mockGump;

	private Gump mockGump2;

	private boolean verifyMocks;

	private TestUtils testUtils;

	private GumSpec gumSpec;

	private GumSpec gumSpec2;

	private int rmiPort;

	private ObjectID mockGumpObjectID1;

	private ObjectID mockGumpObjectID2;

	private BlackListManager ebBlackListManager;

	private EBGridManager ebGridManager;


	/**
	 * Constructor for EBGridManagerTest.
	 * 
	 * @param name Test name
	 */
	public EBGridManagerTest( String name ) {

		super( name );
	}


	@Override
	protected void setUp() throws Exception {

		super.setUp();

		testUtils = new TestUtils();
		testUtils.load();

		testUtils.setProperty( MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_MAX_BL_FAILS,
				MyGridConfiguration.DEFAULT_MAX_BL_FAILS );

		testUtils.createMGProperties();

		Configuration.reset();
		Configuration.getInstance( Configuration.MYGRID );

		rmiPort = Integer.parseInt( testUtils.getProperty( TestUtils.RMI_PORT ) );
		try {
			LocateRegistry.createRegistry( rmiPort );
		} catch ( Exception e ) {
			LocateRegistry.getRegistry( rmiPort );
		}

		verifyMocks = false;
		fakeEBSchedulerFacade = new FakeEBSchedulerFacade();
		fakeEBSchedulerFacade.config( new FakeEBReplicaExecutorFacade() );
		fakeEBSchedulerFacade.startEventProcessor();
		ebGridManager = (EBGridManager) fakeEBSchedulerFacade.getGridManager();
	}


	@Override
	protected void tearDown() throws Exception {

		super.tearDown();
		// ebGridManager.shutdown();
		EventQueue shutdownResponse = fakeEBSchedulerFacade.shutdown();
		shutdownResponse.blockingRemove();
		if ( verifyMocks ) {
			EasyMock.verify( mockGumClient1 );
			EasyMock.verify( mockGumClient2 );
			EasyMock.verify( mockGump );
			EasyMock.verify( mockGump2 );
		}

		try {
			Naming.unbind( "rmi://localhost:" + rmiPort + "/MockGump" );
			Naming.unbind( "rmi://localhost:" + rmiPort + "/MockGump2" );

			OurgridUnicastRemoteObject.unexportObject( bindableMockGumpWrapper, true );
			OurgridUnicastRemoteObject.unexportObject( bindableMockGumpWrapper2, true );
		} catch ( Exception e ) {
		}

		Configuration.reset();
		testUtils.reset();
	}


	public void testHereIsGuM() throws Exception {

		Collection<GumSpec> gums = ebGridManager.getGumSpecs();
		assertTrue( gums.isEmpty() );

		create2MockGumsAnd2MockGumps();

		mockGump.disposeGum( 1, gumSpec.getGumID() );
		EasyMock.expectLastCall().once();

		mockGump.disposeGum( 2, gumSpec2.getGumID() );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( 1 );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( 2 );
		EasyMock.expectLastCall().once();

		EasyMock.replay( mockGumClient1 );
		EasyMock.replay( mockGumClient2 );
		EasyMock.replay( mockGump );
		EasyMock.replay( mockGump2 );

		JobRequestEntry jobRequestEntry;

		jobRequestEntry = new JobRequestEntry( 1, 1 );
		jobRequestEntry.addPeer( mockGump );

		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		jobRequestEntry = new JobRequestEntry( 2, 2 );
		jobRequestEntry.addPeer( mockGump );

		ebGridManager.jobRequestsEntryMap.put( 2, jobRequestEntry );

		ebGridManager.hereIsGum( mockGumClient1, 1, mockGumpObjectID1 );
		ebGridManager.hereIsGum( mockGumClient2, 2, mockGumpObjectID1 );

		// Getting the gums from the EBGridManager
		gums = ebGridManager.getGumSpecs();

		// Verifying if the mockGum1 is in the EBGridManager
		verifyIfGumIsInList( mockGumClient1.getGumSpec(), gums );

		// Verifying if the mockGum2 is in the EBGridManager
		verifyIfGumIsInList( mockGumClient2.getGumSpec(), gums );

		verifyMocks = true;

	}


	public void testPause() throws Exception {

		Collection<GumSpec> gums = ebGridManager.getGumSpecs();
		assertTrue( gums.isEmpty() );

		create2MockGumsAnd2MockGumps();

		EasyMock.replay( mockGumClient1 );
		EasyMock.replay( mockGumClient2 );

		JobRequestEntry jobRequestEntry;

		jobRequestEntry = new JobRequestEntry( 1, 1 );
		jobRequestEntry.addPeer( mockGump );

		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		jobRequestEntry = new JobRequestEntry( 2, 2 );
		jobRequestEntry.addPeer( mockGump );

		ebGridManager.jobRequestsEntryMap.put( 2, jobRequestEntry );

		mockGump.pauseRequestGums( 1 );
		EasyMock.expectLastCall().once();
		mockGump.disposeGum( 1, mockGumClient1.getGumSpec().getGumID() );
		EasyMock.expectLastCall().once();

		EasyMock.replay( mockGump );

		ebGridManager.pauseRequest( 1 );

		ebGridManager.hereIsGum( mockGumClient1, 1, mockGumpObjectID1 );

		EasyMock.verify( mockGump );

		EasyMock.reset( mockGump );

		mockGump.pauseRequestGums( 2 );
		EasyMock.expectLastCall().once();
		mockGump.resumeRequestGums( 1 );
		EasyMock.expectLastCall().once();
		mockGump.disposeGum( 2, mockGumClient2.getGumSpec().getGumID() );
		EasyMock.expectLastCall().once();

		EasyMock.replay( mockGump );

		ebGridManager.pauseRequest( 2 );

		ebGridManager.resumeRequest( 1 );

		ebGridManager.hereIsGum( mockGumClient2, 2, mockGumpObjectID1 );
		ebGridManager.hereIsGum( mockGumClient2, 1, mockGumpObjectID1 );

		EasyMock.verify( mockGump );

		EasyMock.reset( mockGump );

		mockGump.disposeGum( 1, mockGumClient2.getGumSpec().getGumID() );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( 1 );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( 2 );
		EasyMock.expectLastCall().once();

		EasyMock.replay( mockGump );

		EasyMock.replay( mockGump2 );

		verifyMocks = false;
	}


	public void testGumDiedDuringExecution() throws Exception {

		Collection<GumSpec> gums = ebGridManager.getGumSpecs();
		assertTrue( gums.isEmpty() );

		create2MockGumsAnd2MockGumps();

		mockGump.disposeGum( 1, gumSpec.getGumID() );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( 1 );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( 2 );
		EasyMock.expectLastCall().once();

		EasyMock.replay( mockGumClient1 );
		EasyMock.replay( mockGumClient2 );
		EasyMock.replay( mockGump );
		EasyMock.replay( mockGump2 );

		JobRequestEntry jobRequestEntry;

		jobRequestEntry = new JobRequestEntry( 1, 1 );
		jobRequestEntry.addPeer( mockGump );

		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		jobRequestEntry = new JobRequestEntry( 2, 2 );
		jobRequestEntry.addPeer( mockGump );

		ebGridManager.jobRequestsEntryMap.put( 2, jobRequestEntry );

		ebGridManager.hereIsGum( mockGumClient1, 1, mockGumpObjectID1 );
		ebGridManager.hereIsGum( mockGumClient2, 2, mockGumpObjectID1 );

		// Getting the gums from the EBGridManager
		gums = ebGridManager.getGumSpecs();

		assertTrue( verifyIfGumIsInList( mockGumClient1.getGumSpec(), gums ) );
		assertTrue( verifyIfGumIsInList( mockGumClient2.getGumSpec(), gums ) );

		// Warning the EBGridManager that the gum died during execution
		ebGridManager.deleteGum( gumSpec2.getGumID() );

		// Getting mockGum2. It should not exist anymore
		gums = ebGridManager.getGumSpecs();

		// Verifying if the mockGum1 is in the EBGridManager
		assertTrue( verifyIfGumIsInList( mockGumClient1.getGumSpec(), gums ) );
		assertFalse( verifyIfGumIsInList( mockGumClient2.getGumSpec(), gums ) );

		verifyMocks = true;
	}


	public void testGetAllGumEntries() throws Exception {

		Map gumEntries = ebGridManager.getAllGumEntries();
		assertTrue( gumEntries.isEmpty() );

		create2MockGumsAnd2MockGumps();
		mockGump.disposeGum( 1, gumSpec.getGumID() );
		EasyMock.expectLastCall().once();
		mockGump.disposeGum( 2, gumSpec2.getGumID() );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( 1 );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( 2 );
		EasyMock.expectLastCall().once();

		EasyMock.replay( mockGumClient1 );
		EasyMock.replay( mockGumClient2 );
		EasyMock.replay( mockGump );
		EasyMock.replay( mockGump2 );

		JobRequestEntry jobRequestEntry;

		jobRequestEntry = new JobRequestEntry( 1, 1 );
		jobRequestEntry.addPeer( mockGump );

		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		jobRequestEntry = new JobRequestEntry( 2, 2 );
		jobRequestEntry.addPeer( mockGump );

		ebGridManager.jobRequestsEntryMap.put( 2, jobRequestEntry );

		ebGridManager.hereIsGum( mockGumClient1, 1, mockGumpObjectID1 );
		ebGridManager.hereIsGum( mockGumClient2, 2, mockGumpObjectID1 );

		// Getting gumexEntries from the EBGridManager
		gumEntries = ebGridManager.getAllGumEntries();

		// Verifying GuMeXEntry with mockGum1
		GridManagerEntry gumexEntry = (GridManagerEntry) gumEntries.get( mockGumClient1.getGumSpec().getGumID() );
		assertTrue( "A gumEntry with the ID '" + mockGumClient1.getGumSpec().getGumID() + "' should exist.",
				gumexEntry != null );

		// Verifying GuMeXEntry with mockGum2
		gumexEntry = (GridManagerEntry) gumEntries.get( mockGumClient2.getGumSpec().getGumID() );
		assertTrue( "A gumEntry with the ID '" + mockGumClient2.getGumSpec().getGumID() + "' should exist.",
				gumexEntry != null );
		assertEquals( "The returned gumEntry should be the same as the mockGum2.", gumexEntry.getGum(), mockGumClient2 );

		verifyMocks = true;
	}


	public void testNeedGuMs() {

		PeerSpec specLSD = new PeerSpec( testUtils.getProperty( TestUtils.FAKE_MACHINE_NAME ) + "1", Integer
				.parseInt( testUtils.getProperty( TestUtils.FAKE_PORT ) ) );
		specLSD.setLabel( MyGridGumpSpecServices.FAKE_GUMP );
		PeerSpec specUCSD = new PeerSpec( testUtils.getProperty( TestUtils.FAKE_MACHINE_NAME ) + "2", Integer
				.parseInt( testUtils.getProperty( TestUtils.FAKE_PORT ) ) );
		specUCSD.setLabel( MyGridGumpSpecServices.FAKE_GUMP );
		PeerSpec specLOtR = new PeerSpec( testUtils.getProperty( TestUtils.FAKE_MACHINE_NAME ) + "3", Integer
				.parseInt( testUtils.getProperty( TestUtils.FAKE_PORT ) ) );
		specLOtR.setLabel( MyGridGumpSpecServices.FAKE_GUMP );

		LinkedList<PeerSpec> gumpSpecs = new LinkedList<PeerSpec>();
		gumpSpecs.add( specLSD );
		gumpSpecs.add( specUCSD );
		gumpSpecs.add( specLOtR );
		ebGridManager.setPeers( gumpSpecs );
		int jobId = 1;

		for ( PeerEntry entry : ebGridManager.getPeerEntries() ) {

			entry.setPeerStatus( PeerState.UP );

		}

		ebGridManager.needGums( jobId, "os==linux", 2 );

		JobRequestEntry jobReqEntry = ebGridManager.jobRequestsEntryMap.get( jobId );
		assertEquals( 3, jobReqEntry.getPeers().size() );

	}


	public void testCancelRequests() {

		int jobId = 1;

		testNeedGuMs();

		// This method should result in a noMoreGMs( jobId ) call on the gump.
		ebGridManager.cancelRequests( jobId );

		ebGridManager.jobRequestsEntryMap.get( new Integer( jobId ) );
		assertNull( "No job requests should exists", null );

	}


	public void testSetAndGetGuMPs() {

		PeerSpec specLSD = new PeerSpec( testUtils.getProperty( TestUtils.FAKE_MACHINE_NAME ) + "1", Integer
				.parseInt( testUtils.getProperty( TestUtils.FAKE_PORT ) ) );
		specLSD.setLabel( MyGridGumpSpecServices.FAKE_GUMP );
		PeerSpec specUCSD = new PeerSpec( testUtils.getProperty( TestUtils.FAKE_MACHINE_NAME ) + "2", Integer
				.parseInt( testUtils.getProperty( TestUtils.FAKE_PORT ) ) );
		specUCSD.setLabel( MyGridGumpSpecServices.FAKE_GUMP );
		PeerSpec specLOtR = new PeerSpec( testUtils.getProperty( TestUtils.FAKE_MACHINE_NAME ) + "3", Integer
				.parseInt( testUtils.getProperty( TestUtils.FAKE_PORT ) ) );
		specLOtR.setLabel( MyGridGumpSpecServices.FAKE_GUMP );

		LinkedList<PeerSpec> gumpSpecs = new LinkedList<PeerSpec>();
		gumpSpecs.add( specLSD );
		gumpSpecs.add( specUCSD );
		gumpSpecs.add( specLOtR );
		ebGridManager.setPeers( gumpSpecs );

		List returnedGumpSpecs = new LinkedList();
		returnedGumpSpecs = (List) ebGridManager.getPeerEntries();

		boolean specLSDOK = false;
		boolean specUCSDOK = false;
		boolean specLOtROK = false;
		Iterator gumpSpecsIt = returnedGumpSpecs.iterator();
		while ( gumpSpecsIt.hasNext() ) {
			PeerEntry peerEntry = (PeerEntry) gumpSpecsIt.next();
			if ( peerEntry.getPeerSpec().equals( specLSD ) ) {
				specLSDOK = true;
			}
			if ( peerEntry.getPeerSpec().equals( specUCSD ) ) {
				specUCSDOK = true;
			}
			if ( peerEntry.getPeerSpec().equals( specLOtR ) ) {
				specLOtROK = true;
			}
		}
		if ( !specLSDOK ) {
			fail( "The specLSD ( " + specLSD + " ) should be in the returned list of gumps" );
		}
		if ( !specUCSDOK ) {
			fail( "The specUCSD ( " + specUCSD + " ) should be in the returned list of gumps" );
		}
		if ( !specLOtROK ) {
			fail( "The specLOtR ( " + specLOtR + " ) should be in the returned list of gumps" );
		}

	}


	public void testGetGums() throws Exception {

		Collection<GumSpec> gums = ebGridManager.getGumSpecs();
		assertTrue( gums.isEmpty() );

		create2MockGumsAnd2MockGumps();
		mockGump.disposeGum( 1, gumSpec.getGumID() );
		EasyMock.expectLastCall().once();
		mockGump.disposeGum( 2, gumSpec2.getGumID() );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( 1 );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( 2 );
		EasyMock.expectLastCall().once();

		EasyMock.replay( mockGumClient1 );
		EasyMock.replay( mockGumClient2 );
		EasyMock.replay( mockGump );
		EasyMock.replay( mockGump2 );

		JobRequestEntry jobRequestEntry;

		jobRequestEntry = new JobRequestEntry( 1, 1 );
		jobRequestEntry.addPeer( mockGump );

		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		jobRequestEntry = new JobRequestEntry( 2, 2 );
		jobRequestEntry.addPeer( mockGump );

		ebGridManager.jobRequestsEntryMap.put( 2, jobRequestEntry );

		ebGridManager.hereIsGum( mockGumClient1, 1, mockGumpObjectID1 );
		ebGridManager.hereIsGum( mockGumClient2, 2, mockGumpObjectID1 );

		gums = ebGridManager.getGumSpecs();

		assertTrue( verifyIfGumIsInList( mockGumClient1.getGumSpec(), gums ) );
		assertTrue( verifyIfGumIsInList( mockGumClient2.getGumSpec(), gums ) );

		verifyMocks = true;

	}


	public void testDeleteGridMachine() throws Exception {

		// Sending two gums
		testGetAllGumEntries();

		ebGridManager.deleteGum( mockGumClient1.getGumSpec().getGumID() );

		Map gumexEntries = ebGridManager.getAllGumEntries();

		// Verifying GuMeXEntry with mockGum1
		GridManagerEntry gumexEntry = (GridManagerEntry) gumexEntries.get( mockGumClient1.getGumSpec().getGumID() );
		assertNull( "The GuMeXEntry with the ID '" + mockGumClient1.getGumSpec().getGumID() + "' should not exist.",
				gumexEntry );

		// try to delete an non existent machine
		ebGridManager.deleteGum( new GumID( "Name1", "Port1", "Peer1", "PeerPort1" ) );
		gumexEntries = ebGridManager.getAllGumEntries();
		gumexEntry = (GridManagerEntry) gumexEntries.get( mockGumClient2.getGumSpec().getGumID() );
		assertNotNull( "A GuMeXEntry with the ID '" + mockGumClient2.getGumSpec().getGumID() + "' should exist.",
				gumexEntry );
		assertEquals( "The returned GuMeXEntry should be the same as the mockGum2.", gumexEntry.getGum(),
				mockGumClient2 );

		ebGridManager.deleteGum( mockGumClient2.getGumSpec().getGumID() );
		gumexEntries = ebGridManager.getAllGumEntries();
		gumexEntry = (GridManagerEntry) gumexEntries.get( mockGumClient2.getGumSpec().getGumID() );
		assertNull( "The GuMeXEntry with the ID '" + mockGumClient2.getGumSpec().getGumID() + "' should not exist.",
				gumexEntry );

		verifyMocks = false;

	}


	public void testAllocateAndFree() throws Exception {

		// Sending two gums
		testGetAllGumEntries();

		Map gumEntries = ebGridManager.getAllGumEntries();

		// Verifying GuMeXEntry with mockGum1
		GridManagerEntry gumexEntry = (GridManagerEntry) gumEntries.get( gumSpec.getGumID() );
		assertEquals( "The GuMeXEntry (mockGum1) must be idle.", GumStatus.READY, gumexEntry.getState() );

		// Verifying GuMeXEntry with mockGum2
		gumexEntry = (GridManagerEntry) gumEntries.get( gumSpec2.getGumID() );
		assertEquals( "The GuMeXEntry (mockGum2) must be idle.", GumStatus.READY, gumexEntry.getState() );

		ebGridManager.allocate( mockGumClient1.getGumSpec().getGumID(), null );

		// Verifying GuMeXEntry with mockGum1
		gumexEntry = (GridManagerEntry) gumEntries.get( gumSpec.getGumID() );
		assertEquals( "The GuMeXEntry (mockGum1) must be in use.", GumStatus.RUNNING, gumexEntry.getState() );

		// Verifying GuMeXEntry with mockGum2
		gumexEntry = (GridManagerEntry) gumEntries.get( gumSpec2.getGumID() );
		assertEquals( "The GuMeXEntry (mockGum2) must be idle.", GumStatus.READY, gumexEntry.getState() );

		ebGridManager.ready( mockGumClient1.getGumSpec().getGumID() );
		ebGridManager.allocate( mockGumClient2.getGumSpec().getGumID(), null );

		// Verifying GuMeXEntry with mockGum1
		gumexEntry = (GridManagerEntry) gumEntries.get( gumSpec.getGumID() );
		assertEquals( "The GuMeXEntry (mockGum1) must be idle.", GumStatus.READY, gumexEntry.getState() );

		// Verifying GuMeXEntry with mockGum2
		gumexEntry = (GridManagerEntry) gumEntries.get( gumSpec2.getGumID() );
		assertEquals( "The GuMeXEntry (mockGum2) must be in use.", GumStatus.RUNNING, gumexEntry.getState() );

	}


	public void testGumsAreDisposedAfterJobIsFinished() throws Exception {

		long reqId2 = 22;
		long reqId1 = 11;

		int jobId2 = 2;
		int jobId1 = 1;

		Collection<GumSpec> gums = ebGridManager.getGumSpecs();
		assertTrue( gums.isEmpty() );

		create2MockGumsAnd2MockGumps();

		mockGump.finishRequestGums( reqId1 );
		EasyMock.expectLastCall().once();

		mockGump2.finishRequestGums( reqId2 );
		EasyMock.expectLastCall().once();

		EasyMock.replay( mockGumClient1 );
		EasyMock.replay( mockGumClient2 );
		EasyMock.replay( mockGump );
		EasyMock.replay( mockGump2 );

		JobRequestEntry jobRequestEntry = new JobRequestEntry( jobId1, reqId1 );
		jobRequestEntry.addPeer( mockGump );
		ebGridManager.jobRequestsEntryMap.put( jobId1, jobRequestEntry );

		jobRequestEntry = new JobRequestEntry( jobId2, reqId2 );
		jobRequestEntry.addPeer( mockGump2 );
		ebGridManager.jobRequestsEntryMap.put( jobId2, jobRequestEntry );

		ebGridManager.hereIsGum( mockGumClient1, reqId1, mockGumpObjectID1 );
		ebGridManager.hereIsGum( mockGumClient2, reqId2, mockGumpObjectID2 );

		ebGridManager.cancelRequests( jobId1 );
		ebGridManager.cancelRequests( jobId2 );

		assertFalse( ebGridManager.jobRequestsEntryMap.containsKey( jobId1 ) );
		assertTrue( ebGridManager.getAvailableGuMs( jobId1, 1 ).isEmpty() );

		assertFalse( ebGridManager.jobRequestsEntryMap.containsKey( jobId2 ) );
		assertTrue( ebGridManager.getAvailableGuMs( jobId2, 1 ).isEmpty() );

		verifyMocks = true;
	}


	public void testBlackList() throws Exception {

		create2MockGumsAnd2MockGumps();
		ebBlackListManager = ebGridManager.getEBBlackListManager();

		long reqId1 = 11;
		int jobId1 = 1;
		int jobSize = 7;

		mockGump.unwantedGum( EasyMock.eq( reqId1 ), EasyMock.eq( gumSpec.getGumID() ) );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( reqId1 );
		EasyMock.expectLastCall().anyTimes();

		EasyMock.replay( mockGumClient1 );
		EasyMock.replay( mockGump );

		JobRequestEntry jobRequestEntry = new JobRequestEntry( jobId1, reqId1 );
		jobRequestEntry.addPeer( mockGump );
		ebGridManager.jobRequestsEntryMap.put( jobId1, jobRequestEntry );
		ebGridManager.initiateBlackListForJob( jobId1, jobSize );

		ebGridManager.hereIsGum( mockGumClient1, reqId1, mockGumpObjectID1 );

		for ( int i = 1; i <= jobSize; i++ ) {
			assertTrue( ebBlackListManager.canThisGumRunThisTask( jobId1, gumSpec.getGumID(), i ) );
			ebGridManager.updateBlackList( jobId1, i, gumSpec.getGumID() );
			assertEquals( jobSize - i, ebBlackListManager.getRemainingBLFails( jobId1, gumSpec.getGumID() ) );
			assertFalse( ebBlackListManager.canThisGumRunThisTask( jobId1, gumSpec.getGumID(), i ) );
		}

		EasyMock.verify( mockGumClient1 );
		EasyMock.verify( mockGump );
	}


	public void testBlackList2() throws Exception {

		create2MockGumsAnd2MockGumps();
		ebBlackListManager = ebGridManager.getEBBlackListManager();

		long reqId1 = 11;
		int jobId1 = 1;
		int defMax = Integer.parseInt( MyGridConfiguration.DEFAULT_MAX_BL_FAILS );
		int jobSize = defMax + 1;

		mockGump.unwantedGum( EasyMock.eq( reqId1 ), EasyMock.eq( gumSpec.getGumID() ) );
		EasyMock.expectLastCall().once();

		mockGump.finishRequestGums( reqId1 );
		EasyMock.expectLastCall().anyTimes();

		EasyMock.replay( mockGumClient1 );
		EasyMock.replay( mockGump );

		JobRequestEntry jobRequestEntry = new JobRequestEntry( jobId1, reqId1 );
		jobRequestEntry.addPeer( mockGump );
		ebGridManager.jobRequestsEntryMap.put( jobId1, jobRequestEntry );
		ebGridManager.initiateBlackListForJob( jobId1, jobSize );

		ebGridManager.hereIsGum( mockGumClient1, reqId1, mockGumpObjectID1 );

		for ( int i = 1; i <= defMax; i++ ) {
			assertTrue( ebBlackListManager.canThisGumRunThisTask( jobId1, gumSpec.getGumID(), i ) );
			ebGridManager.updateBlackList( jobId1, i, gumSpec.getGumID() );
			assertEquals( defMax - i, ebBlackListManager.getRemainingBLFails( jobId1, gumSpec.getGumID() ) );
			assertFalse( ebBlackListManager.canThisGumRunThisTask( jobId1, gumSpec.getGumID(), i ) );
		}

		EasyMock.verify( mockGumClient1 );
		EasyMock.verify( mockGump );
	}


	/**
	 * This test intends to check if a hereIsGum method call occurs, and the Gum
	 * is in job black list, this gum can not be assigned to execute any task of
	 * this job.
	 */
	public void testBlackList3() throws Exception {

		create2MockGumsAnd2MockGumps();
		ebBlackListManager = ebGridManager.getEBBlackListManager();

		long reqId1 = 11;
		int jobId1 = 1;
		int defMax = Integer.parseInt( MyGridConfiguration.DEFAULT_MAX_BL_FAILS );
		int jobSize = defMax + 1;

		JobRequestEntry jobRequestEntry = new JobRequestEntry( jobId1, reqId1 );
		jobRequestEntry.addPeer( mockGump );
		ebGridManager.jobRequestsEntryMap.put( jobId1, jobRequestEntry );
		ebGridManager.initiateBlackListForJob( jobId1, jobSize );

		mockGump.unwantedGum( EasyMock.eq( reqId1 ), EasyMock.eq( gumSpec.getGumID() ) );
		EasyMock.expectLastCall().times( 2 );

		mockGump.finishRequestGums( EasyMock.eq( reqId1 ) );
		EasyMock.expectLastCall().anyTimes();

		EasyMock.replay( mockGumClient1 );
		EasyMock.replay( mockGump );

		ebGridManager.hereIsGum( mockGumClient1, reqId1, mockGumpObjectID1 );

		for ( int i = 1; i <= defMax; i++ ) {
			ebGridManager.updateBlackList( jobId1, i, gumSpec.getGumID() );
		}

		assertTrue( ebGridManager.getAllGumEntries().isEmpty() );
		assertTrue( ebGridManager.getAvailableAndValidGums().isEmpty() );

		for ( int i = 1; i <= jobSize; i++ ) {
			assertTrue( ebGridManager.getAvailableGuMs( jobId1, i ).isEmpty() );
		}

		// Now the gum is in the job black list.
		assertTrue( ebBlackListManager.isGumBlacklistedForEntireJob( jobId1, gumSpec.getGumID() ) );

		ebGridManager.hereIsGum( mockGumClient1, reqId1, mockGumpObjectID1 );

		for ( int i = 1; i <= jobSize; i++ ) {
			assertTrue( ebGridManager.getAvailableGuMs( jobId1, i ).isEmpty() );
		}

		EasyMock.verify( mockGumClient1 );
		EasyMock.verify( mockGump );
	}


	/**
	 * If Mygrid receives a blacklisted Gum, the peer that delvered the Gum must
	 * be warned.
	 */
	public void testBlackList4() throws Exception {

		// Redifining properties, defMax = 1
		testUtils.reset();
		testUtils = new TestUtils();
		testUtils.load();
		testUtils.setProperty( MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_MAX_BL_FAILS, 1 + "" );
		testUtils.createMGProperties();

		Configuration.reset();
		Configuration.getInstance( Configuration.MYGRID );
		// end

		fakeEBSchedulerFacade = new FakeEBSchedulerFacade();
		fakeEBSchedulerFacade.config( new FakeEBReplicaExecutorFacade() );
		fakeEBSchedulerFacade.startEventProcessor();
		ebGridManager = (EBGridManager) fakeEBSchedulerFacade.getGridManager();

		create2MockGumsAnd2MockGumps();
		ebBlackListManager = ebGridManager.getEBBlackListManager();

		long reqId1 = 11;
		int jobId1 = 1;
		int defMax = Integer
				.parseInt( Configuration.getInstance().getProperty( MyGridConfiguration.PROP_MAX_BL_FAILS ) );
		int jobSize = 2;

		assertEquals( 1, defMax );

		JobRequestEntry jobRequestEntry = new JobRequestEntry( jobId1, reqId1 );
		jobRequestEntry.addPeer( mockGump );
		ebGridManager.jobRequestsEntryMap.put( jobId1, jobRequestEntry );
		ebGridManager.initiateBlackListForJob( jobId1, jobSize );

		mockGump.unwantedGum( EasyMock.eq( reqId1 ), EasyMock.eq( gumSpec.getGumID() ) );
		EasyMock.expectLastCall().times( 2 );

		mockGump.finishRequestGums( reqId1 );
		EasyMock.expectLastCall().anyTimes();

		EasyMock.replay( mockGumClient1 );
		EasyMock.replay( mockGump );

		ebGridManager.hereIsGum( mockGumClient1, reqId1, mockGumpObjectID1 );

		// Gum is now blacklisted for the entire Job.
		assertFalse( ebBlackListManager.isGumBlacklistedForEntireJob( jobId1, gumSpec.getGumID() ) );
		ebGridManager.updateBlackList( jobId1, 1, gumSpec.getGumID() );
		assertEquals( defMax - 1, ebBlackListManager.getRemainingBLFails( jobId1, gumSpec.getGumID() ) );
		assertTrue( ebBlackListManager.isGumBlacklistedForEntireJob( jobId1, gumSpec.getGumID() ) );

		ebGridManager.hereIsGum( mockGumClient1, reqId1, mockGumpObjectID1 );

		EasyMock.verify( mockGumClient1 );
		EasyMock.verify( mockGump );
	}


	public void testListenerNotify() throws Exception {

		GridManagerListenerRemote mockListener = EasyMock.createMock( GridManagerListenerRemote.class );

		create2MockGumsAnd2MockGumps();

		EasyMock.replay( mockGumClient1 );

		mockListener.hereIsGum( mockGumClient1.getGumSpec(), 1, mockGumpObjectID1 );
		EasyMock.expectLastCall().once();

		EasyMock.replay( mockListener );

		ebGridManager.getGumSpecs();

		mockListener.hereIsGum( mockGumClient1.getGumSpec(), 1, mockGumpObjectID1 );

		EasyMock.verify( mockListener );
	}


	private boolean verifyIfGumIsInList( GumSpec spec, Collection<GumSpec> gums ) {

		for ( GumSpec gum : gums ) {
			if ( spec.getGumID().equals( gum.getGumID() ) ) {
				return true;
			}
		}
		return false;
	}


	private void create2MockGumsAnd2MockGumps() throws Exception {

		// Setting the GridMachine
		mockGumClient1 = EasyMock.createMock( GumClient.class );

		String peerName = "test.peername";
		String peerPort = "0";

		Map<String,String> attrib = generateAttributeMap( MOCK_GUM_NAME1, "1000", peerName, peerPort );
		gumSpec = new GumSpec( attrib );

		EasyMock.expect( mockGumClient1.getGumSpec() ).andReturn( gumSpec ).anyTimes();

		// Setting the GridMachineClient
		mockGumClient2 = EasyMock.createMock( GumClient.class );

		attrib = generateAttributeMap( MOCK_GUM_NAME2, "1000", peerName, peerPort );
		gumSpec2 = new GumSpec( attrib );

		EasyMock.expect( mockGumClient2.getGumSpec() ).andReturn( gumSpec2 ).anyTimes();

		// Creating the mockGump
		mockGump = EasyMock.createMock( Gump.class );

		// Creating the mockGump2
		mockGump2 = EasyMock.createMock( Gump.class );

		bindableMockGumpWrapper = new BindableMockGumpWrapper( mockGump );
		bindableMockGumpWrapper2 = new BindableMockGumpWrapper( mockGump2 );

		mockGumpObjectID1 = new ObjectID( "MockGump", (long) 42, "rmi://localhost:" + rmiPort + "/MockGump" );
		mockGumpObjectID2 = new ObjectID( "MockGump", (long) 42, "rmi://localhost:" + rmiPort + "/MockGump2" );

		Naming.rebind( mockGumpObjectID1.getObjectURL(), bindableMockGumpWrapper );
		Naming.rebind( mockGumpObjectID2.getObjectURL(), bindableMockGumpWrapper2 );
	}


	/**
	 * Generates an attribute map of a Gum.
	 * 
	 * @param os Gum OS.
	 * @param gumName Gum name.
	 * @param gumPort Gum port.
	 * @param peerName Peer name.
	 * @param peerPort Peer port.
	 * @return The attributes map.
	 */
	private Map<String,String> generateAttributeMap( String gumName, String gumPort, String peerName, String peerPort ) {

		Map<String,String> attrib1 = new HashMap<String,String>();
		attrib1.put( GumSpec.ATT_NAME, gumName );
		attrib1.put( GumSpec.ATT_PORT, gumPort );
		attrib1.put( GumSpec.ATT_PEERNAME, peerName );
		attrib1.put( GumSpec.ATT_PEERPORT, peerPort );
		return attrib1;
	}

}
