/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.test;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import junit.framework.TestCase;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.exception.UnableToCreatePlaypenException;
import org.ourgrid.common.exception.UnableToCreateStorageException;
import org.ourgrid.common.exception.UnavailableGumException;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.common.spec.exception.JobSpecificationException;
import org.ourgrid.common.spec.exception.TaskSpecificationException;
import org.ourgrid.gridmachine.exception.FailCauseException;
import org.ourgrid.gridmachine.exception.GenericFailCauseException;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.JobMonitor;
import org.ourgrid.mygrid.scheduler.ReplicaExecutorResult;
import org.ourgrid.mygrid.scheduler.jobmanager.EBJobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.TaskEntry;
import org.ourgrid.test.SchedulerFastTest;
import org.ourgrid.test.TestUtils;

public class EBJobManagerTest extends TestCase implements SchedulerFastTest {

	private JobManager jobManager;

	private TestUtils testUtils;


	/**
	 * Constructor for EBJobManagerTest.
	 * 
	 * @param name Test name
	 */
	public EBJobManagerTest( String name ) {

		super( name );
	}


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createMGProperties();

		Configuration.getInstance( Configuration.MYGRID );
		jobManager = new EBJobManager();
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		jobManager = null;
		Configuration.reset();
		testUtils.reset();
	}


	public void testAddJobAndClonedJobList() throws TaskSpecificationException, JobSpecificationException {

		String expression = "os==linux";
		int numTasks = 5;
		JobSpec jobSpec = createJobSpec( expression, numTasks );
		int jobId = 1;
		jobManager.addJob( jobId, jobSpec );
		Collection jobs = jobManager.clonedJobList();
		assertTrue( "The job added must be in the list returned by the EBJobManager", containsJob( jobs, jobId ) );
	}


	public void testAddJobAndJobList() throws TaskSpecificationException, JobSpecificationException {

		String expression = "os==linux";
		int numTasks = 5;
		JobSpec jobSpec = createJobSpec( expression, numTasks );
		int jobId = 1;
		jobManager.addJob( jobId, jobSpec );
		Collection jobs = jobManager.jobList();
		assertTrue( "The job added must be in the list returned by the EBJobManager", containsJob( jobs, jobId ) );
	}


	public void testAddJobInterestedWithRunningTasks() throws ConfigException, TaskSpecificationException,
		JobSpecificationException {

		int jobId = 1;
		int maxReplicas = 1;
		jobManager.setMaxReplicas( maxReplicas + "" );
		createRunningReplicasInTheJobManager( jobId, 2, maxReplicas );

		EventQueue responseQueue = new EventQueue();
		JobMonitor jobMonitor = new JobMonitor( responseQueue, jobId );
		jobManager.addJobInterested( jobMonitor );

		TaskSpec taskSpec = new TaskSpec( null, "echo 0", null );
		ReplicaEntry clonedReplica = new ReplicaEntry( 1, 1, 1, taskSpec );
		ReplicaExecutorResult result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaFinished( result );

		clonedReplica = new ReplicaEntry( 1, 2, 1, taskSpec );
		result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaFinished( result );

		// This call should not block. The jobMonitor should have been
		// notified when the job finishes.
		responseQueue.blockingRemove();

		ExecutionStatus jobState = jobManager.getJobState( jobId );
		assertEquals( "The job should be 'Finished'", ExecutionStatus.FINISHED, jobState );

	}


	public void testReplicaFinished() throws TaskSpecificationException, JobSpecificationException {

		int jobId = 1;
		int replicasPerTask = 1;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, replicasPerTask + "" );
		createRunningReplicasInTheJobManager( jobId, 2, replicasPerTask );

		TaskSpec taskSpec = new TaskSpec( null, "echo 0", null );
		ReplicaEntry clonedReplica = new ReplicaEntry( 1, 1, 1, taskSpec );
		ReplicaExecutorResult result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaFinished( result );

		clonedReplica = new ReplicaEntry( 1, 2, 1, taskSpec );
		result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaFinished( result );

		ExecutionStatus jobState = jobManager.getJobState( jobId );
		assertEquals( "The job should be 'Finished'", ExecutionStatus.FINISHED, jobState );
	}


	public void testReplicaFailed() throws TaskSpecificationException, JobSpecificationException {

		int jobId = 1;
		int replicasPerTask = 1;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, replicasPerTask + "" );
		createRunningReplicasInTheJobManager( jobId, 2, replicasPerTask );

		TaskSpec taskSpec = new TaskSpec( null, "echo 0", null );

		ReplicaEntry clonedReplica = new ReplicaEntry( 1, 1, 1, taskSpec );
		ReplicaExecutorResult result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaFinished( result );

		clonedReplica = new ReplicaEntry( 1, 2, 1, taskSpec );
		result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaFailed( result );

		ExecutionStatus jobState = jobManager.getJobState( jobId );
		assertEquals( "The job should be 'Failed'", ExecutionStatus.FAILED, jobState );
	}


	public void testReplicaFailedWithExceptions() throws TaskSpecificationException, JobSpecificationException,
		ConfigException {

		int jobId = 1;
		int numTasks = 1;
		int maxReplicas = 1;
		jobManager.setMaxReplicas( maxReplicas + "" );
		jobManager.setMaxFails( "3" );

		// Test with UnavailableGMException, UnableToCreatePlaypenException and
		// UnableToCreateStorageException
		String expression = "os==linux";
		JobSpec jobSpec = createJobSpec( expression, numTasks );
		jobManager.addJob( jobId, jobSpec );

		int replicasLimit = 10;
		int replicasCount = 0;
		ReplicaEntry replica = getReplica();
		while ( replica != null ) {
			replicasCount++;
			sendReplicaFailed( replica, new UnavailableGumException( "" ) );
			if ( replicasCount < replicasLimit ) {
				replica = getReplica();
			} else {
				replica = null;
			}
		}

		replicasCount = 0;
		replica = getReplica();
		while ( replica != null ) {
			replicasCount++;
			sendReplicaFailed( replica, new UnableToCreatePlaypenException() );
			if ( replicasCount < replicasLimit ) {
				replica = getReplica();
			} else {
				replica = null;
			}
		}

		replicasCount = 0;
		replica = getReplica();
		while ( replica != null ) {
			replicasCount++;
			sendReplicaFailed( replica, new UnableToCreateStorageException() );
			if ( replicasCount < replicasLimit ) {
				replica = getReplica();
			} else {
				replica = null;
			}
		}

		int tasksCount = 0;
		Collection jobs = jobManager.jobList();
		Iterator jobsIt = jobs.iterator();
		while ( jobsIt.hasNext() ) {
			JobEntry job = (JobEntry) jobsIt.next();
			Collection tasks = job.getTasks();
			Iterator tasksIt = tasks.iterator();
			while ( tasksIt.hasNext() ) {
				TaskEntry task = (TaskEntry) tasksIt.next();
				tasksCount += task.getReplicas().size();
			}
		}

		ExecutionStatus jobState = jobManager.getJobState( jobId );
		assertEquals( "The number of replicas created should be " + replicasLimit * 3, replicasLimit * 3, tasksCount );
		assertEquals( "The job should be 'Running'", ExecutionStatus.RUNNING, jobState );

		replicasCount = 0;
		replica = getReplica();
		while ( replica != null ) {
			replicasCount++;
			sendReplicaFailed( replica, new GenericFailCauseException() );
			if ( replicasCount < replicasLimit ) {
				replica = getReplica();
			} else {
				replica = null;
			}
		}

		jobState = jobManager.getJobState( jobId );
		assertEquals( "The job should be 'Failed'", ExecutionStatus.FAILED, jobState );

		tasksCount = 0;
		jobs = jobManager.jobList();
		jobsIt = jobs.iterator();
		while ( jobsIt.hasNext() ) {
			JobEntry job = (JobEntry) jobsIt.next();
			Collection tasks = job.getTasks();
			Iterator tasksIt = tasks.iterator();
			while ( tasksIt.hasNext() ) {
				TaskEntry task = (TaskEntry) tasksIt.next();
				tasksCount += task.getReplicas().size();
			}
		}
		assertEquals( "The number of replicas should be " + (replicasLimit * 3 + 3), replicasLimit * 3 + 3, tasksCount );
	}


	public void testReplicaCanceled() throws ConfigException, TaskSpecificationException, JobSpecificationException {

		int jobId = 1;
		int replicasPerTask = 1;
		jobManager.setMaxReplicas( replicasPerTask + "" );

		createRunningReplicasInTheJobManager( jobId, 2, replicasPerTask );

		TaskSpec taskSpec = new TaskSpec( null, "echo 0", null );

		ReplicaEntry clonedReplica = new ReplicaEntry( 1, 1, 1, taskSpec );
		ReplicaExecutorResult result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaFinished( result );

		clonedReplica = new ReplicaEntry( 1, 2, 1, taskSpec );
		result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaCanceled( result );

		ExecutionStatus jobState = jobManager.getJobState( jobId );
		assertEquals( "The job should be 'Canceled'", ExecutionStatus.CANCELLED, jobState );
	}


	public void testReplicaAborted() throws ConfigException, TaskSpecificationException, JobSpecificationException {

		int jobId = 1;
		int replicasPerTask = 2;
		jobManager.setMaxReplicas( replicasPerTask + "" );

		createRunningReplicasInTheJobManager( jobId, 2, replicasPerTask );

		TaskSpec taskSpec = new TaskSpec( null, "echo 0", null );

		ReplicaEntry clonedReplica = new ReplicaEntry( 1, 1, 1, taskSpec );
		ReplicaExecutorResult result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaFinished( result );

		clonedReplica = new ReplicaEntry( 2, 1, 1, taskSpec );
		result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaAborted( result );

		clonedReplica = new ReplicaEntry( 1, 2, 1, taskSpec );
		result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaFinished( result );

		clonedReplica = new ReplicaEntry( 2, 2, 1, taskSpec );
		result = new ReplicaExecutorResult( clonedReplica );
		jobManager.replicaAborted( result );

		ExecutionStatus jobState = jobManager.getJobState( jobId );
		assertEquals( "The job should be 'Finished'", ExecutionStatus.FINISHED, jobState );
	}


	private void sendReplicaFailed( ReplicaEntry replica, FailCauseException cause ) {

		ReplicaExecutorResult rer = new ReplicaExecutorResult( replica );
		rer.setErrorCause( cause );
		jobManager.replicaFailed( rer );
	}


	private ReplicaEntry getReplica() {

		Collection jobs = jobManager.jobList();
		Iterator jobsIt = jobs.iterator();
		JobEntry job = null;
		while ( jobsIt.hasNext() ) {
			job = (JobEntry) jobsIt.next();
			Collection tasks = job.getAvailableTasks();
			Iterator tasksIt = tasks.iterator();
			while ( tasksIt.hasNext() ) {
				TaskEntry task = (TaskEntry) tasksIt.next();
				ReplicaEntry replica = jobManager.createNewReplica( task );
				if ( replica != null ) {
					replica.setStatus( ExecutionStatus.RUNNING );
					return replica;
				}
			}
		}

		return null;
	}


	private void createRunningReplicasInTheJobManager( int jobId, int numTasks, int replicasPerTask )
		throws TaskSpecificationException, JobSpecificationException {

		String expression = "os==linux";
		JobSpec jobSpec = createJobSpec( expression, numTasks );
		jobManager.addJob( jobId, jobSpec );
		Collection jobs = jobManager.jobList();
		Iterator jobsIt = jobs.iterator();
		JobEntry job = null;
		while ( jobsIt.hasNext() ) {
			job = (JobEntry) jobsIt.next();
			Collection tasks = job.getTasks();
			Iterator tasksIt = tasks.iterator();
			while ( tasksIt.hasNext() ) {
				TaskEntry task = (TaskEntry) tasksIt.next();
				for ( int i = 0; i < replicasPerTask; i++ ) {
					if ( task.canReplicate() ) {
						ReplicaEntry replica = jobManager.createNewReplica( task );
						if ( replica != null ) {
							replica.setStatus( ExecutionStatus.RUNNING );
						}
					}
				}
			}
		}
	}


	public void testIfJobsAreBeingCloned() throws ConfigException, TaskSpecificationException,
		JobSpecificationException {

		String expression = "os==linux";
		int numTasks = 1;
		JobSpec jobSpec = createJobSpec( expression, numTasks );
		int jobId = 1;
		jobManager.addJob( jobId, jobSpec );
		jobManager.setMaxReplicas( "1" );
		Collection jobs = jobManager.clonedJobList();
		Iterator jobsIt = jobs.iterator();
		JobEntry job = null;
		ReplicaEntry replica1 = null;
		while ( jobsIt.hasNext() ) {
			job = (JobEntry) jobsIt.next();
			Collection tasks = job.getTasks();
			Iterator tasksIt = tasks.iterator();
			while ( tasksIt.hasNext() ) {
				TaskEntry task = (TaskEntry) tasksIt.next();
				replica1 = jobManager.createNewReplica( task );
			}
		}

		jobs = jobManager.clonedJobList();
		jobsIt = jobs.iterator();
		job = null;
		ReplicaEntry replica2 = null;
		while ( jobsIt.hasNext() ) {
			job = (JobEntry) jobsIt.next();
			Collection tasks = job.getTasks();
			Iterator tasksIt = tasks.iterator();
			while ( tasksIt.hasNext() ) {
				TaskEntry task = (TaskEntry) tasksIt.next();
				Collection replicas = task.getReplicas();
				Iterator replicasIt = replicas.iterator();
				while ( replicasIt.hasNext() ) {
					replica2 = (ReplicaEntry) replicasIt.next();
				}
			}
		}

		assertTrue( "The replicas returned by the EBJobManager should be different. They must be clones.",
				replica1 != replica2 );
		assertTrue( "The cloned replicas must represent the same object.", replica1.equals( replica2 ) );
	}


	/**
	 * Creates a job with <code>numTasks</code> identical tasks. Each task
	 * simply performs "echo 0".
	 * 
	 * @param expression The job attribute expression
	 * @param numTasks The number of tasks in the job
	 * @return The created job
	 * @throws TaskSpecificationException When the <code>TaskSpec</code> is
	 *         not valid
	 * @throws JobSpecificationException
	 */
	private JobSpec createJobSpec( String expression, int numTasks ) throws TaskSpecificationException,
		JobSpecificationException {

		List<TaskSpec> taskSpecList = new LinkedList<TaskSpec>();

		for ( int i = 0; i < numTasks; i++ ) {
			TaskSpec t = new TaskSpec( null, "echo 0", null );
			taskSpecList.add( t );
		}

		return new JobSpec( "TestJob", expression, taskSpecList );
	}


	/**
	 * Verifies if a job with 'jobId' is an element of a given collection of
	 * jobs.
	 * 
	 * @param jobs Job collection
	 * @param jobId The job id
	 * @return True if the job is present in the collection of jobs, false
	 *         otherwise
	 */
	private boolean containsJob( Collection jobs, int jobId ) {

		Iterator jobsIt = jobs.iterator();
		boolean containsJob = false;
		while ( jobsIt.hasNext() ) {
			JobEntry job = (JobEntry) jobsIt.next();
			if ( job.getId() == jobId ) {
				containsJob = true;
			}
		}
		return containsJob;
	}

}
