/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.test;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.common.spec.exception.JobSpecificationException;
import org.ourgrid.common.spec.exception.TaskSpecificationException;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.ReplicaExecutorResult;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.TaskEntry;
import org.ourgrid.test.SchedulerFastTest;

public class JobTest extends TestCase implements SchedulerFastTest {

	/**
	 * Constructor for JobTest.
	 * 
	 * @param name Test name
	 */
	public JobTest( String name ) {

		super( name );
	}


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();
		Configuration.reset();
		Configuration.getInstance( Configuration.MYGRID );
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
	}


	public void testReplicaFinished() {

		// Already tested in the EBJobManagerTest
	}


	public void testReplicaFailed() {

		// Already tested in the EBJobManagerTest
	}


	public void testCancel() {

		// Already tested in the EBJobManagerTest
	}


	public void testSetReplicaState() {

		// Already tested in the EBJobManagerTest
	}


	public void testGetNewReplica() {

		// Already tested in the EBJobManagerTest
	}


	public void testNewJobWithNullJobSpec() {

		JobEntry job = new JobEntry( 1, null, 1, 1 );
		List tasks = job.getTasks();
		assertNotNull( "The tasks list should not be null", tasks );
		assertTrue( "The tasks list should be empty", tasks.isEmpty() );
	}


	/**
	 * This method tests if the JobSpec is verified during its construction. The
	 * validated parameter is the list of Tasks Specs.
	 */
	public void testNewJobWithNullTaskSpecs() {

		JobSpec jobSpec;

		try {
			jobSpec = new JobSpec( "Job1", "os == linux", null );
			fail( getClass().getName() + ".testNewJobWithNullTasksSpecs(): Job Spec should not be created." );
		} catch ( JobSpecificationException e ) {
			System.out.println( getClass().getName() + ".testNewJobWithNullTasksSpecs(): EXPECTED EXCEPTION: "
					+ "A Job Spec could not bw constructed based on a null list of Tasks Specs" );
			jobSpec = new JobSpec( "Job1" );
		}

		JobEntry job = new JobEntry( 1, jobSpec, 1, 1 );
		List tasks = job.getTasks();

		assertNotNull( "The tasks list should not be null", tasks );
		assertTrue( "The tasks list should be empty", tasks.isEmpty() );
	}


	public void testGetAvailableTasks() throws TaskSpecificationException, JobSpecificationException {

		String expression = "so==linux";
		int maxReplicas = 1;
		int maxFails = 1;

		int numTasks = 5;
		JobSpec jobSpec = createJobSpec( expression, numTasks );
		JobEntry job = new JobEntry( 1, jobSpec, maxReplicas, maxFails );
		LinkedList tasks = (LinkedList) job.getTasks();
		LinkedList<ReplicaEntry> replicas = new LinkedList<ReplicaEntry>();
		List availableTasks = new LinkedList();

		Iterator tasksIt = tasks.iterator();
		int tasksCounter = 0;
		while ( tasksIt.hasNext() ) {
			TaskEntry task = (TaskEntry) tasksIt.next();
			while ( task.canReplicate() ) {
				ReplicaEntry replica = task.getNewReplica();
				if ( replica != null ) {
					replicas.add( replica );
				}
			}
			tasksCounter++;
			availableTasks = job.getAvailableTasks();
			assertEquals( numTasks - tasksCounter, availableTasks.size() );
		}
	}


	/**
	 * Tests if, according to the Tasks states, a Job is in the correct state.
	 * It tests a several combination of Tasks states.
	 * 
	 * @throws TaskSpecificationException
	 * @throws JobSpecificationException
	 */
	public void testGetState() throws TaskSpecificationException, JobSpecificationException {

		String expression = "so==linux";
		int maxReplicas = 1;
		int maxFails = 1;

		int numTasks = 5;
		JobSpec jobSpec = createJobSpec( expression, numTasks );
		JobEntry job = new JobEntry( 1, jobSpec, maxReplicas, maxFails );
		LinkedList tasks = (LinkedList) job.getTasks();
		createReplicas( tasks );
		LinkedList<String> states = new LinkedList<String>();
		assertTrue( job.getStatus().equals( ExecutionStatus.UNSTARTED ) );

		// There is at least one running replica
		numTasks = 5;
		jobSpec = createJobSpec( expression, numTasks );
		job = new JobEntry( 1, jobSpec, maxReplicas, maxFails );
		tasks = (LinkedList) job.getTasks();
		createReplicas( tasks );
		states = new LinkedList<String>();
		String[ ] statesArrayRunning = { "1", "2", "3", "4", "5" };
		ExecutionStatus[ ] replicasStatesRunning = { ExecutionStatus.UNSTARTED, ExecutionStatus.RUNNING,
													ExecutionStatus.FAILED, ExecutionStatus.CANCELLED,
													ExecutionStatus.FINISHED };
		assertAll( job, tasks, ExecutionStatus.RUNNING, ExecutionStatus.RUNNING, states, statesArrayRunning,
				replicasStatesRunning, numTasks - 1 );

		// There is at least one unstarted replica (no running replicas)
		numTasks = 4;
		jobSpec = createJobSpec( expression, numTasks );
		job = new JobEntry( 1, jobSpec, maxReplicas, maxFails );
		tasks = (LinkedList) job.getTasks();
		createReplicas( tasks );
		states = new LinkedList<String>();
		String[ ] statesArrayUnstarted = { "1", "2", "3", "4" };
		ExecutionStatus[ ] replicasStatesUnstarted = { ExecutionStatus.UNSTARTED, ExecutionStatus.FAILED,
														ExecutionStatus.CANCELLED, ExecutionStatus.FINISHED };
		assertAll( job, tasks, ExecutionStatus.RUNNING, ExecutionStatus.UNSTARTED, states, statesArrayUnstarted,
				replicasStatesUnstarted, numTasks - 1 );

		// There is at least one canceled replica (no running and unstarted
		// replicas)
		numTasks = 3;
		jobSpec = createJobSpec( expression, numTasks );
		job = new JobEntry( 1, jobSpec, maxReplicas, maxFails );
		tasks = (LinkedList) job.getTasks();
		createReplicas( tasks );
		states = new LinkedList<String>();
		String[ ] statesArrayCanceled = { "1", "2", "3" };
		ExecutionStatus[ ] replicasStatesCanceled = { ExecutionStatus.FAILED, ExecutionStatus.CANCELLED,
														ExecutionStatus.FINISHED };
		assertAll( job, tasks, ExecutionStatus.CANCELLED, ExecutionStatus.CANCELLED, states, statesArrayCanceled,
				replicasStatesCanceled, numTasks - 1 );

		// There is at least one failed replica (no running, unstarted and
		// canceled replicas)
		numTasks = 2;
		jobSpec = createJobSpec( expression, numTasks );
		job = new JobEntry( 1, jobSpec, maxReplicas, maxFails );
		tasks = (LinkedList) job.getTasks();
		createReplicas( tasks );
		states = new LinkedList<String>();
		String[ ] statesArrayFailed = { "1", "2" };
		ExecutionStatus[ ] replicasStatesFailed = { ExecutionStatus.FAILED, ExecutionStatus.FINISHED };
		assertAll( job, tasks, ExecutionStatus.FAILED, ExecutionStatus.FAILED, states, statesArrayFailed,
				replicasStatesFailed, numTasks - 1 );

		// There is at least one failed replica (no running, unstarted and
		// canceled replicas)
		numTasks = 3;
		jobSpec = createJobSpec( expression, numTasks );
		job = new JobEntry( 1, jobSpec, maxReplicas, maxFails );
		tasks = (LinkedList) job.getTasks();
		createReplicas( tasks );
		states = new LinkedList<String>();
		String[ ] statesArrayFinished = { "1", "2" };
		ExecutionStatus[ ] replicasStatesFinished = { ExecutionStatus.FINISHED, ExecutionStatus.FINISHED };
		assertAll( job, tasks, ExecutionStatus.FINISHED, ExecutionStatus.FINISHED, states, statesArrayFinished,
				replicasStatesFinished, numTasks - 1 );
	}


	/**
	 * Asserts if the Job is in the expected state in several combinations of
	 * Tasks states.
	 * 
	 * @param job The Job to be tested.
	 * @param tasks Its Collection of Tasks.
	 * @param expectedJobState The expected state the Job will be in all
	 *        combinations.
	 * @param fixedReplicaState A Replica state that will be fixed to fix one
	 *        Task state - to be possible to know which state Job will be in
	 *        advance.
	 * @param states Where the Replica states will be held.
	 * @param statesArray The possible Task state array - to foresee which state
	 *        the Job will be.
	 * @param replicasStates The possible replica state array - to foresee which
	 *        state the Job will be.
	 * @param depth The depth of the combination.
	 */
	private void assertAll( JobEntry job, LinkedList tasks, ExecutionStatus expectedJobState,
							ExecutionStatus fixedReplicaState, LinkedList<String> states, String[ ] statesArray,
							ExecutionStatus[ ] replicasStates, int depth ) {

		for ( int x = 0; x < statesArray.length; x++ ) {
			states.add( statesArray[x] );
		}

		for ( int a = 1; a < depth; a++ ) {
			states = copyStates( statesArray, states );
		}

		for ( int x = 0; x < states.size(); x++ ) {
			TaskEntry task = (TaskEntry) tasks.get( 0 );
			setTaskState( task, fixedReplicaState );
			String statesString = states.get( x );
			for ( int y = 0; y < statesString.length(); y++ ) {
				task = (TaskEntry) tasks.get( y + 1 );
				int intState = Integer.parseInt( statesString.charAt( y ) + "" ) - 1;
				setTaskState( task, replicasStates[intState] );
			}
			if ( areAllReplicasUnstarted( tasks ) ) {
				assertEquals( ExecutionStatus.UNSTARTED, job.getStatus() );
			} else {
				assertEquals( expectedJobState, job.getStatus() );
			}
		}
	}


	/**
	 * Checks if all Replicas are unstarted.
	 * 
	 * @param tasks The a list of Tasks.
	 * @return <code>true</code> if all Replicas are unstarted,
	 *         <code>false</code> otherwise.
	 */
	private boolean areAllReplicasUnstarted( LinkedList tasks ) {

		Iterator tasksIt = tasks.iterator();
		while ( tasksIt.hasNext() ) {
			TaskEntry task = (TaskEntry) tasksIt.next();
			Collection replicas = task.getReplicas();
			Iterator replicasIt = replicas.iterator();
			while ( replicasIt.hasNext() ) {
				ReplicaEntry replica = (ReplicaEntry) replicasIt.next();
				if ( !replica.getStatus().equals( ExecutionStatus.UNSTARTED ) ) {
					return false;
				}
			}
		}
		return true;
	}


	/**
	 * Sets a Task state.
	 * 
	 * @param task The Task will have its state changed.
	 * @param state The replica state.
	 */
	private void setTaskState( TaskEntry task, ExecutionStatus state ) {

		Collection replicas = task.getReplicas();
		Iterator replicasIt = replicas.iterator();
		while ( replicasIt.hasNext() ) {
			ReplicaEntry replica = (ReplicaEntry) replicasIt.next();
			if ( state.equals( ExecutionStatus.FAILED ) ) {
				replica.setStatus( ExecutionStatus.RUNNING );
				ReplicaExecutorResult rer = new ReplicaExecutorResult( replica );
				task.replicaFailed( rer );
			} else {
				replica.setStatus( state );
			}
		}
	}


	private LinkedList<String> copyStates( String[ ] statesArray, LinkedList states ) {

		LinkedList<String> newStates = new LinkedList<String>();
		for ( int x = 0; x < statesArray.length; x++ ) {
			for ( int y = 0; y < states.size(); y++ ) {
				newStates.add( ((String) states.get( y )) + statesArray[x] );
			}
		}
		return newStates;
	}


	/**
	 * Creates a job with <code>numTasks</code> identical tasks. Each task
	 * simply performs "echo 0".
	 * 
	 * @param expression The job attribute expression
	 * @param numTasks The number of tasks in the job
	 * @return The created job
	 * @throws TaskSpecificationException When the <code>TaskSpec</code> is
	 *         not valid
	 * @throws JobSpecificationException
	 */
	private JobSpec createJobSpec( String expression, int numTasks ) throws TaskSpecificationException,
		JobSpecificationException {

		List<TaskSpec> taskSpecList = new LinkedList<TaskSpec>();

		for ( int i = 0; i < numTasks; i++ ) {
			TaskSpec t = new TaskSpec( null, "echo 0", null );
			taskSpecList.add( t );
		}

		return new JobSpec( "TestJob", expression, taskSpecList );
	}


	private List createReplicas( Collection tasks ) {

		List<ReplicaEntry> replicas = new LinkedList<ReplicaEntry>();
		Iterator tasksIt = tasks.iterator();
		while ( tasksIt.hasNext() ) {
			TaskEntry task = (TaskEntry) tasksIt.next();
			int replicasCounter = 0;
			while ( task.canReplicate() ) {
				ReplicaEntry replica = task.getNewReplica();
				if ( replica != null ) {
					replicas.add( replica );
					replicasCounter++;
				}
			}
		}

		return replicas;
	}
}
