/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.test;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.common.exception.UnavailableGumException;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.gump.Gump;
import org.ourgrid.common.gump.GumpClient;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.common.spec.exception.JobSpecificationException;
import org.ourgrid.common.spec.exception.TaskSpecificationException;
import org.ourgrid.mygrid.replicaexecutor.test.FakeEBReplicaExecutorFacade;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.FakeEBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.Scheduler;
import org.ourgrid.mygrid.scheduler.exception.JobCouldNotBeCancelledException;
import org.ourgrid.mygrid.scheduler.exception.JobNotFoundException;
import org.ourgrid.mygrid.scheduler.gridmanager.GumResponseEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.test.SchedulerFastTest;
import org.ourgrid.test.TestUtils;

public class SchedulerImplTest extends TestCase implements SchedulerFastTest {

	private static final String PEER_PORT = "4242";

	private static final String GUM_NAME = "ArthurDentGum";

	private static final String GUM_PORT = "1000";

	private static final String PEER_NAME = "ArthurDent";

	private static final ObjectID OBJECT_ID = new ObjectID( PEER_NAME, (long) 4242, "rmi://arthurdent:" + PEER_PORT
			+ "/HeartOfGold" );

	private FakeEBSchedulerFacade fakeEBSchedulerFacade;

	private FakeEBReplicaExecutorFacade fakeEBReplicaExecutorFacade;

	private Scheduler schedulerImpl;

	private GumpClient gumc;

	private Gum mockGuM;

	private Gump mockGump;

	private boolean verifyMocks;

	private TestUtils testUtils;

	private GumID mockGumID;


	@Override
	protected void setUp() throws Exception {

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createMGProperties();

		Configuration.getInstance( Configuration.MYGRID );
		verifyMocks = false;
		fakeEBSchedulerFacade = new FakeEBSchedulerFacade();
		fakeEBReplicaExecutorFacade = new FakeEBReplicaExecutorFacade();
		fakeEBReplicaExecutorFacade.config( fakeEBSchedulerFacade );
		fakeEBSchedulerFacade.config( fakeEBReplicaExecutorFacade );
		fakeEBSchedulerFacade.startEventProcessor();
		schedulerImpl = fakeEBSchedulerFacade.getScheduler();
		gumc = fakeEBSchedulerFacade.getGumpClient();
	}


	@Override
	protected void tearDown() throws Exception {

		EventQueue<ShutdownResponseEvent> responseQueue = fakeEBSchedulerFacade.shutdown();
		responseQueue.blockingRemove();

		if ( verifyMocks ) {
			EasyMock.verify( mockGuM );
			EasyMock.verify( mockGump );
		}

		Configuration.reset();
		testUtils.reset();
	}


	public void testJobList() throws RemoteException {

		Collection<JobEntry> jobs = schedulerImpl.jobList( null );
		assertTrue( jobs.isEmpty() );
	}


	public void testWaitForJobWithCancelJob() throws JobNotFoundException, RemoteException,
		JobCouldNotBeCancelledException, TaskSpecificationException, JobSpecificationException, UnavailableGumException {

		createMockGuM();
		createMockGuMP();

		EasyMock.replay( mockGump );
		EasyMock.replay( mockGuM );

		// Sending gum and gumex
		Map<String,String> attrib = new HashMap<String,String>();
		attrib.put( GumSpec.ATT_NAME, GUM_NAME );
		attrib.put( GumSpec.ATT_PORT, GUM_PORT );
		attrib.put( GumSpec.ATT_PEERNAME, PEER_NAME );
		attrib.put( GumSpec.ATT_PEERPORT, PEER_PORT );

		gumc.hereIsGum( new GumResponseEntry( mockGuM, new GumSpec(), 1, OBJECT_ID ) );

		String expression = "os==linux";
		int numTasks = 1;
		JobSpec jobSpec = createJobSpec( expression, numTasks );
		int jobId = schedulerImpl.addJob( jobSpec );
		assertEquals( "The first job id should be '1'", 1, jobId );

		// Cancelling the job
		schedulerImpl.cancelJob( jobId );

		// This method should not block. The job was canceled.
		schedulerImpl.waitForJob( jobId );

		Collection<JobEntry> jobs = schedulerImpl.jobList( null );
		Iterator<JobEntry> jobsIt = jobs.iterator();
		while ( jobsIt.hasNext() ) {
			JobEntry job = jobsIt.next();
			if ( job.getId() == jobId ) {
				assertEquals( "The job should be in 'Canceled' state.", ExecutionStatus.CANCELLED, job.getStatus() );
			}
		}

		verifyMocks = true;

	}


	private void createMockGuM() throws RemoteException {

		// Setting the GridMachine
		mockGuM = EasyMock.createMock( Gum.class );

		mockGuM.getStatus();
		EasyMock.expectLastCall().andReturn( GumStatus.UP ).anyTimes();
	}


	private void createMockGuMP() throws RemoteException {

		// Creating the mockGump
		mockGump = EasyMock.createMock( Gump.class );
		mockGump.disposeGum( 1, mockGumID );
		EasyMock.expectLastCall().anyTimes();

		mockGump.finishRequestGums( EasyMock.anyLong() );
		EasyMock.expectLastCall().anyTimes();
	}


	public void testWaitForJobWithRunningTasks() throws JobNotFoundException, RemoteException,
		TaskSpecificationException, JobSpecificationException {

		createMockGuM();

		// Creating the mockGump
		createMockGuMP();

		EasyMock.replay( mockGump );
		EasyMock.replay( mockGuM );

		Map<String,String> attrib = new HashMap<String,String>();
		attrib.put( GumSpec.ATT_NAME, GUM_NAME );
		attrib.put( GumSpec.ATT_PORT, GUM_PORT );
		attrib.put( GumSpec.ATT_PEERNAME, PEER_NAME );
		attrib.put( GumSpec.ATT_PEERPORT, PEER_PORT );

		gumc.hereIsGum( new GumResponseEntry( mockGuM, new GumSpec(), 1, OBJECT_ID ) );

		String expression = "location==lsd && os==linux";
		int numTasks = 5;
		fakeEBSchedulerFacade.setReplicaStateToBeReturned( ExecutionStatus.FAILED );
		JobSpec jobSpec = createJobSpec( expression, numTasks );
		int jobId1 = schedulerImpl.addJob( jobSpec );
		int jobId2 = schedulerImpl.addJob( jobSpec );
		int jobId3 = schedulerImpl.addJob( jobSpec );
		schedulerImpl.waitForJob( jobId1 );
		schedulerImpl.waitForJob( jobId2 );
		schedulerImpl.waitForJob( jobId3 );

		assertEquals( "The third job id should be '3'", 3, jobId3 );

		verifyMocks = true;
	}


	public void testAddJob() throws RemoteException, TaskSpecificationException, JobSpecificationException {

		String expression = "location=lsd && so=linux";
		int numTasks = 5;
		JobSpec jobSpec = createJobSpec( expression, numTasks );
		int jobId = schedulerImpl.addJob( jobSpec );
		assertEquals( "The first job id should be '1'", 1, jobId );

		// Verifying if the added job is in the list of jobs
		Collection<JobEntry> jobs = schedulerImpl.jobList( null );
		assertTrue( "The job must be in the job list", containsJob( jobs, 1 ) );
	}


	/**
	 * Creates a job with <code>numTasks</code> identical tasks. Each task
	 * simply performs "echo 0".
	 * 
	 * @param expression The job attribute expression
	 * @param numTasks The number of tasks in the job
	 * @return The created job
	 * @throws TaskSpecificationException When the <code>TaskSpec</code> is
	 *         not valid
	 * @throws JobSpecificationException
	 */
	private JobSpec createJobSpec( String expression, int numTasks ) throws TaskSpecificationException,
		JobSpecificationException {

		List<TaskSpec> taskSpecList = new LinkedList<TaskSpec>();

		for ( int i = 0; i < numTasks; i++ ) {
			TaskSpec t = new TaskSpec( null, "echo 0", null );
			taskSpecList.add( t );
		}

		return new JobSpec( "TestJob", expression, taskSpecList );
	}


	/**
	 * Verifies if a job with 'jobId' is an element of a given collection of
	 * jobs.
	 * 
	 * @param jobs Job collection
	 * @param jobId The job id
	 * @return True if the job is present in the collection of jobs, false
	 *         otherwise
	 */
	private boolean containsJob( Collection<JobEntry> jobs, int jobId ) {

		Iterator<JobEntry> jobsIt = jobs.iterator();
		boolean containsJob = false;
		while ( jobsIt.hasNext() ) {
			JobEntry job = jobsIt.next();
			if ( job.getId() == jobId ) {
				containsJob = true;
			}
		}
		return containsJob;
	}

}
