package org.ourgrid.mygrid.scheduler.test;

import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.event.DummyEvent;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.exception.UnableToCreatePlaypenException;
import org.ourgrid.common.exception.UnableToCreateStorageException;
import org.ourgrid.common.executor.ExecutorException;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.IntegerExecutorHandle;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.gump.BindableMockGumpWrapper;
import org.ourgrid.common.gump.Gump;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.common.spec.exception.JobSpecificationException;
import org.ourgrid.common.spec.exception.TaskSpecificationException;
import org.ourgrid.gridmachine.exception.GenericFailCauseException;
import org.ourgrid.gridmachine.useragent.UserAgentClient;
import org.ourgrid.mygrid.replicaexecutor.EBReplicaExecutorFacade;
import org.ourgrid.mygrid.replicaexecutor.test.FakeEBReplicaExecutorFacade;
import org.ourgrid.mygrid.scheduler.EBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.FakeEBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.FakeSchedulingHeuristic;
import org.ourgrid.mygrid.scheduler.JobRequestEntry;
import org.ourgrid.mygrid.scheduler.ReplicaExecutorResult;
import org.ourgrid.mygrid.scheduler.gridmanager.EBGridManager;
import org.ourgrid.mygrid.scheduler.gump.GumpClientImpl;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.test.SchedulerFastTest;
import org.ourgrid.test.TestUtils;
import org.ourgrid.test.UniqueTest;
import org.ourgrid.test.threadServices.ThreadPoolServices;
import org.ourgrid.test.threadServices.ThreadServices;

/**
 * This class will test features that are independent of the scheduling
 * heuristic implementation. Basically, the tests must assure that the abstract
 * methods of <code>EBAbstractSchedulingHeuristic</code> are working.
 */
public class SchedulingHeuristicTest extends TestCase implements SchedulerFastTest {

	private static final String EXIT_COMMAND = "exit 1";

	private static final String ECHO_COMMAND = "echo 0";

	private static final int REQUESTID_1 = 1;

	private static final int JOBID_1 = 1;

	private FakeSchedulingHeuristic schedulingHeuristic;

	private JobManager ebJobManager;

	private GumpClientImpl gridMachineConsumerImpl;

	private EBGridManager ebGridManager;

	private EBReplicaExecutorFacade ebReplicaExecutorFacade;

	private TestUtils testUtils;

	private EBSchedulerFacade ebSchedulerFacade;

	private int rmiPort;

	private ObjectID gumpID;

	private BindableMockGumpWrapper bindableMockGumpWrapper;

	private GumSpec gumSpec1;

	private Gump gump;

	private boolean gumpReplayed;

	private BindableMockGumpWrapper strictBindableMockGumpWrapper;

	private ObjectID strictGumpID;


	protected void setUp() throws Exception {

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createMGProperties();

		rmiPort = Integer.parseInt( testUtils.getProperty( TestUtils.RMI_PORT ) );
		try {
			LocateRegistry.createRegistry( rmiPort );
		} catch ( Exception e ) {
			LocateRegistry.getRegistry( rmiPort );
		}

		Configuration.reset();
		Configuration.getInstance( Configuration.MYGRID );

		gumSpec1 = new GumSpec();
		gumSpec1.putAttribute( GumSpec.ATT_NAME, "mockGum111111" );

		gumpID = new ObjectID( "TestGump", 1L, "rmi://" );

		gump = this.createMockGump();
		this.gumpReplayed = false;
	}


	protected void tearDown() throws Exception {

		if ( !gumpReplayed ) {
			EasyMock.replay( gump );
		}

		EasyMock.verify( gump );
		EasyMock.reset( gump );

		EventQueue shutdownResponse = ebSchedulerFacade.shutdown();
		shutdownResponse.blockingRemove();

		try {
			Naming.unbind( gumpID.getObjectURL() );
			Naming.unbind( strictGumpID.getObjectURL() );

			OurgridUnicastRemoteObject.unexportObject( bindableMockGumpWrapper, true );
			OurgridUnicastRemoteObject.unexportObject( strictBindableMockGumpWrapper, true );
			OurgridUnicastRemoteObject.unexportObject( this.gridMachineConsumerImpl, true );
		} catch ( Exception e ) {
		}

		ebGridManager = null;
		ebSchedulerFacade = null;

		Configuration.reset();
		testUtils.reset();

	}


	private JobSpec createJobSpec( String expression, int numTasks, String command ) throws TaskSpecificationException,
		JobSpecificationException {

		List<TaskSpec> taskSpecList = new LinkedList<TaskSpec>();

		for ( int i = 0; i < numTasks; i++ ) {
			TaskSpec t = new TaskSpec( null, command, null );
			taskSpecList.add( t );
		}

		return new JobSpec( "TestJob", expression, taskSpecList );
	}


	/*
	 * This test intends to check if the scheduling steps involving a failure
	 * Gum are correct.
	 */
	public void testRunningFailTask1() throws Exception {

		configureEnvironment( new FakeEBSchedulerFacade(), new FakeEBReplicaExecutorFacade() );

		Gum mockGum = getMockGumReturningExecutorException( 1 );
		UserAgentClient uaClient1 = new UserAgentClient( mockGum, gumSpec1 );

		/* Adding the job */
		JobSpec jobSpecs = createJobSpec( "", 2, ECHO_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );
		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		/* Adding the Gum */
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		/* Scheduling */
		assertTrue( schedulingHeuristic.schedule() );

		ReplicaEntry entry = new ReplicaEntry( 1, 1, 1, jobSpecs.getTaskSpecs().get( 0 ) );
		entry.setGumID( uaClient1.getGumSpec().getGumID() );

		schedulingHeuristic.replicaFailed( new ReplicaExecutorResult( entry ) );

		ThreadServices.waitUntilWorkIsDone();
		ThreadPoolServices.waitUntilWorkIsDone();

		/* Task has failed, this machine can no longer be used by this task */
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );
		assertFalse( ebGridManager.getAvailableGuMs( JOBID_1, 2 ).isEmpty() );
	}


	/*
	 * Tests the behavior in face of a Gum failure
	 */
	public void testRunningFailTask2() throws Exception {

		configureEnvironment( new EBSchedulerFacade(), new EBReplicaExecutorFacade() );
		this.ebSchedulerFacade.startEventProcessor();
		this.ebReplicaExecutorFacade.startEventProcessor();

		final int TASKS = 2;
		Gum mockGum = getMockGumReturningRemoteException( 1 );
		UserAgentClient uaClient1 = new UserAgentClient( mockGum, gumSpec1 );

		JobSpec jobSpecs = createJobSpec( "", TASKS, EXIT_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );
		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		jobRequestEntry.addPeer( gump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		this.ebSchedulerFacade.getEventProcessor().putEvent( new DummyEvent() );

		ThreadServices.waitUntilWorkIsDone();
		ThreadPoolServices.waitUntilWorkIsDone();

		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 2 ).isEmpty() );

		EasyMock.verify( mockGum );
	}


	public void testBlacklist2() throws Exception {

		configureEnvironment( new EBSchedulerFacade(), new FakeEBReplicaExecutorFacade() );

		UserAgentClient uaClient1 = new UserAgentClient( this.getNiceMockGum(), gumSpec1 );

		EasyMock.reset( gump );

		gump.unwantedGum( 1, uaClient1.getGumSpec().getGumID() );
		EasyMock.expectLastCall().once();

		EasyMock.replay( gump );
		this.gumpReplayed = true;

		JobSpec jobSpecs = createJobSpec( "", 3, EXIT_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );
		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		jobRequestEntry.addPeer( gump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		assertTrue( this.schedulingHeuristic.schedule() );

		ReplicaExecutorResult res1_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		res1_1.setErrorCause( new GenericFailCauseException() );
		this.schedulingHeuristic.replicaFailed( res1_1 );

		assertTrue( this.schedulingHeuristic.schedule() );

		ReplicaExecutorResult res2_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		this.schedulingHeuristic.replicaFinished( res2_1 );

		assertTrue( this.schedulingHeuristic.schedule() );

		ReplicaExecutorResult res3_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		res3_1.setErrorCause( new GenericFailCauseException() );

		this.schedulingHeuristic.replicaFailed( res3_1 );

		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );

	}


	public void testBlacklist1() throws Exception {

		configureEnvironment( new EBSchedulerFacade(), new FakeEBReplicaExecutorFacade() );

		UserAgentClient uaClient1 = new UserAgentClient( this.getNiceMockGum(), gumSpec1 );

		EasyMock.reset( gump );

		gump.unwantedGum( 1, uaClient1.getGumSpec().getGumID() );
		EasyMock.expectLastCall().once();

		EasyMock.replay( gump );
		this.gumpReplayed = true;

		JobSpec jobSpecs = createJobSpec( "", 3, EXIT_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );
		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		jobRequestEntry.addPeer( gump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		assertTrue( this.schedulingHeuristic.schedule() );

		ReplicaExecutorResult res1_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		res1_1.setErrorCause( new GenericFailCauseException() );
		this.schedulingHeuristic.replicaFailed( res1_1 );

		assertTrue( this.schedulingHeuristic.schedule() );

		ReplicaExecutorResult res2_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		this.schedulingHeuristic.replicaFailed( res2_1 );

		assertTrue( this.schedulingHeuristic.schedule() );

		ReplicaExecutorResult res3_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );

		this.schedulingHeuristic.replicaFailed( res3_1 );

		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );

	}


	private Gum getMockGumReturningExecutorException( int executions ) throws Exception {

		Gum mockGum = EasyMock.createMock( Gum.class );

		ExecutorHandle handle = new IntegerExecutorHandle( 0 );

		mockGum.startReplica();
		EasyMock.expectLastCall().times( executions );
		mockGum.createPlaypen( EasyMock.anyInt() );
		EasyMock.expectLastCall().andReturn( "/tmp" ).times( executions );
		mockGum.createStorageDirectory();
		EasyMock.expectLastCall().andReturn( "/tmp" ).times( executions );
		mockGum.remoteExecute( EasyMock.eq( "/tmp" ), EasyMock.eq( ECHO_COMMAND ), (Map<String,String>) EasyMock
				.anyObject() );
		EasyMock.expectLastCall().andReturn( handle ).times( executions );
		mockGum.getResult( handle );
		EasyMock.expectLastCall().andThrow( new ExecutorException( "" ) ).times( executions );

		EasyMock.replay( mockGum );

		return mockGum;
	}


	private Gum getMockGumReturningRemoteException( int executions ) throws Exception {

		Gum mockGum = EasyMock.createMock( Gum.class );

		ExecutorHandle handle = new IntegerExecutorHandle( 0 );

		mockGum.startReplica();
		EasyMock.expectLastCall().times( executions );
		mockGum.createPlaypen( EasyMock.anyInt() );
		EasyMock.expectLastCall().andReturn( "/tmp" ).times( executions );
		mockGum.createStorageDirectory();
		EasyMock.expectLastCall().andReturn( "/tmp" ).times( executions );
		mockGum.remoteExecute( EasyMock.eq( "/tmp" ), EasyMock.eq( EXIT_COMMAND ), (Map<String,String>) EasyMock
				.anyObject() );
		EasyMock.expectLastCall().andReturn( handle ).times( executions );
		mockGum.getResult( handle );
		EasyMock.expectLastCall().andThrow( new RemoteException( "" ) ).times( executions );

		EasyMock.replay( mockGum );

		return mockGum;
	}


	/**
	 * This test verifies if the treatment of a fail occurred at playpen
	 * creation is the same as an execution fail occurred during getResult
	 * phase.
	 */
	public void testGumCouldNotCreatePlaypen() throws Exception {

		configureEnvironment( new FakeEBSchedulerFacade(), new EBReplicaExecutorFacade() );
		this.ebReplicaExecutorFacade.startEventProcessor();

		Gum mockGum = EasyMock.createMock( Gum.class );
		UserAgentClient uaClient1 = new UserAgentClient( mockGum, gumSpec1 );

		mockGum.startReplica();
		EasyMock.expectLastCall().once();

		mockGum.createPlaypen( EasyMock.anyInt() );
		EasyMock.expectLastCall().andThrow( new UnableToCreatePlaypenException() ).once();

		mockGum.stopReplica();

		EasyMock.replay( mockGum );

		gump.unwantedGum( REQUESTID_1, uaClient1.getGumSpec().getGumID() );

		EasyMock.replay( gump );
		this.gumpReplayed = true;

		/* Adding the job */
		JobSpec jobSpecs = createJobSpec( "", 1, ECHO_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );
		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		/* Adding the Gum */
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		/* Scheduling */
		assertTrue( schedulingHeuristic.schedule() );
		this.ebSchedulerFacade.startEventProcessor();

		ThreadServices.waitUntilWorkIsDone();
		ThreadPoolServices.waitUntilWorkIsDone();

		/* Task has failed, this machine can no longer be used by this task */
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );

		/* Verifying the Gum method invocation */
		EasyMock.verify( mockGum );

	}


	/**
	 * This test verifies if the treatement of a fail ocurred at storage
	 * creation is the same as an execution fail ocurred during getResult phase.
	 */
	public void testGumCouldNotCreateStorage() throws Exception {

		configureEnvironment( new FakeEBSchedulerFacade(), new EBReplicaExecutorFacade() );
		this.ebReplicaExecutorFacade.startEventProcessor();

		Gum mockGum = EasyMock.createMock( Gum.class );
		UserAgentClient uaClient1 = new UserAgentClient( mockGum, gumSpec1 );

		mockGum.startReplica();
		EasyMock.expectLastCall().once();

		mockGum.createPlaypen( EasyMock.anyInt() );
		EasyMock.expectLastCall().andReturn( "/tmp" ).once();

		mockGum.createStorageDirectory();
		EasyMock.expectLastCall().andThrow( new UnableToCreateStorageException() ).once();

		mockGum.stopReplica();

		EasyMock.replay( mockGum );

		gump.unwantedGum( REQUESTID_1, uaClient1.getGumSpec().getGumID() );

		EasyMock.replay( gump );
		this.gumpReplayed = true;

		/* Adding the job */
		JobSpec jobSpecs = createJobSpec( "", 1, ECHO_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );
		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		/* Adding the Gum */
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		/* Scheduling */
		assertTrue( schedulingHeuristic.schedule() );
		this.ebSchedulerFacade.startEventProcessor();

		ThreadServices.waitUntilWorkIsDone();
		ThreadPoolServices.waitUntilWorkIsDone();

		/* Task has failed, this machine can no longer be used by this task */
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );

		/* Verifying the Gum method invocation */
		EasyMock.verify( mockGum );
	}


	public void testDispose() throws Exception {

		configureEnvironment( new EBSchedulerFacade(), new FakeEBReplicaExecutorFacade() );

		GumSpec gumSpec2 = new GumSpec();
		gumSpec2.putAttribute( GumSpec.ATT_NAME, "mockGum2" );
		GumSpec gumSpec3 = new GumSpec();
		gumSpec3.putAttribute( GumSpec.ATT_NAME, "mockGum3" );
		GumSpec gumSpec4 = new GumSpec();
		gumSpec4.putAttribute( GumSpec.ATT_NAME, "mockGum4" );

		final Gum mockGum1 = getNiceMockGum();
		UserAgentClient uaClient1 = new UserAgentClient( mockGum1, gumSpec1 );
		final Gum mockGum2 = getNiceMockGum();
		UserAgentClient uaClient2 = new UserAgentClient( mockGum2, gumSpec2 );
		final Gum mockGum3 = getNiceMockGum();
		UserAgentClient uaClient3 = new UserAgentClient( mockGum3, gumSpec3 );
		final Gum mockGum4 = getNiceMockGum();
		UserAgentClient uaClient4 = new UserAgentClient( mockGum4, gumSpec4 );

		gump.pauseRequestGums( REQUESTID_1 );

		gump.disposeGum( REQUESTID_1, gumSpec1.getGumID() );
		gump.disposeGum( REQUESTID_1, gumSpec2.getGumID() );
		gump.disposeGum( REQUESTID_1, gumSpec3.getGumID() );

		gump.finishRequestGums( REQUESTID_1 );

		EasyMock.replay( gump );
		this.gumpReplayed = true;

		/* Adding the job */
		JobSpec jobSpecs = createJobSpec( "", 4, ECHO_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );
		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		jobRequestEntry.addPeer( gump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		/* Delivering Gums to Job */
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		ebGridManager.hereIsGum( uaClient2, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient2.getGumSpec().getGumID() );

		ebGridManager.hereIsGum( uaClient3, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient3.getGumSpec().getGumID() );

		ebGridManager.hereIsGum( uaClient4, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient4.getGumSpec().getGumID() );

		this.schedulingHeuristic.schedule();
		ReplicaExecutorResult res1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		this.schedulingHeuristic.schedule();
		ReplicaExecutorResult res2 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		this.schedulingHeuristic.schedule();
		ReplicaExecutorResult res3 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		this.schedulingHeuristic.schedule();
		ReplicaExecutorResult res4 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );

		this.schedulingHeuristic.replicaFinished( res1 );
		this.schedulingHeuristic.replicaFinished( res2 );
		this.schedulingHeuristic.replicaFinished( res3 );
		this.schedulingHeuristic.replicaFinished( res4 );

		/* Machines should have been disposed */
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 2 ).isEmpty() );
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 3 ).isEmpty() );
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 4 ).isEmpty() );
	}


	public void testDispose2() throws Exception {

		configureEnvironment( new EBSchedulerFacade(), new FakeEBReplicaExecutorFacade() );

		GumSpec gumSpec11 = new GumSpec();
		gumSpec11.putAttribute( GumSpec.ATT_NAME, "mockGum11" );
		GumSpec gumSpec22 = new GumSpec();
		gumSpec22.putAttribute( GumSpec.ATT_NAME, "mockGum22" );
		GumSpec gumSpec33 = new GumSpec();
		gumSpec33.putAttribute( GumSpec.ATT_NAME, "mockGum33" );

		final Gum mockGum11 = getNiceMockGum();
		UserAgentClient uaClient1 = new UserAgentClient( mockGum11, gumSpec11 );
		final Gum mockGum22 = getNiceMockGum();
		UserAgentClient uaClient2 = new UserAgentClient( mockGum22, gumSpec22 );
		final Gum mockGum33 = getNiceMockGum();
		UserAgentClient uaClient3 = new UserAgentClient( mockGum33, gumSpec33 );

		gump.pauseRequestGums( REQUESTID_1 );

		gump.disposeGum( REQUESTID_1, gumSpec11.getGumID() );
		gump.disposeGum( REQUESTID_1, gumSpec22.getGumID() );

		gump.finishRequestGums( REQUESTID_1 );

		EasyMock.replay( gump );
		this.gumpReplayed = true;

		/* Adding the job */
		this.ebJobManager.setMaxReplicas( "3" );
		JobSpec jobSpecs = createJobSpec( "", 1, ECHO_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );

		assertEquals( 1, this.ebJobManager.getJobById( JOBID_1 ).getAvailableTasks().size() );

		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		jobRequestEntry.addPeer( gump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		/* Delivering Gums to Job */
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		ebGridManager.hereIsGum( uaClient2, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient2.getGumSpec().getGumID() );

		ebGridManager.hereIsGum( uaClient3, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient3.getGumSpec().getGumID() );

		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaEntry replicaEntry1 = this.schedulingHeuristic.getLatestScheduledReplica();
		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaEntry replicaEntry2 = this.schedulingHeuristic.getLatestScheduledReplica();
		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaEntry replicaEntry3 = this.schedulingHeuristic.getLatestScheduledReplica();

		ReplicaExecutorResult res1 = new ReplicaExecutorResult( replicaEntry1 );
		res1.setGumClient( uaClient1 );

		ReplicaExecutorResult res2 = new ReplicaExecutorResult( replicaEntry2 );
		res1.setGumClient( uaClient2 );

		ReplicaExecutorResult res3 = new ReplicaExecutorResult( replicaEntry3 );
		res1.setGumClient( uaClient3 );

		this.schedulingHeuristic.replicaFinished( res1 );
		this.schedulingHeuristic.replicaAborted( res2 );
		this.schedulingHeuristic.replicaAborted( res3 );

		/* Machines should have been disposed */
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );
	}


	public void testDispose3() throws Exception {

		configureEnvironment( new EBSchedulerFacade(), new FakeEBReplicaExecutorFacade() );

		final Gum mockGum1 = getNiceMockGum();
		UserAgentClient uaClient1 = new UserAgentClient( mockGum1, gumSpec1 );

		gump.pauseRequestGums( REQUESTID_1 );

		gump.finishRequestGums( REQUESTID_1 );

		EasyMock.replay( gump );
		this.gumpReplayed = true;

		/* Adding the job */
		JobSpec jobSpecs = createJobSpec( "", 4, ECHO_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );
		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		jobRequestEntry.addPeer( gump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		/* Delivering Gums to Job */
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, gumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		/* Scheduling */
		// Task 1
		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaEntry replicaEntry = this.schedulingHeuristic.getLatestScheduledReplica();

		ReplicaExecutorResult res = new ReplicaExecutorResult( replicaEntry );
		res.setGumClient( uaClient1 );
		this.schedulingHeuristic.replicaFinished( res );

		// Task 2
		assertTrue( this.schedulingHeuristic.schedule() );
		replicaEntry = this.schedulingHeuristic.getLatestScheduledReplica();

		res = new ReplicaExecutorResult( replicaEntry );
		res.setGumClient( uaClient1 );
		this.schedulingHeuristic.replicaFinished( res );

		// Task 3
		assertTrue( this.schedulingHeuristic.schedule() );
		replicaEntry = this.schedulingHeuristic.getLatestScheduledReplica();

		res = new ReplicaExecutorResult( replicaEntry );
		res.setGumClient( uaClient1 );
		this.schedulingHeuristic.replicaFinished( res );

		// Task 4
		assertTrue( this.schedulingHeuristic.schedule() );
		replicaEntry = this.schedulingHeuristic.getLatestScheduledReplica();

		res = new ReplicaExecutorResult( replicaEntry );
		res.setGumClient( uaClient1 );
		this.schedulingHeuristic.replicaFinished( res );

		assertTrue( ebJobManager.getJobById( JOBID_1 ).getStatus().equals( ExecutionStatus.FINISHED ) );

		/* Machines should have been disposed */
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 2 ).isEmpty() );
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 3 ).isEmpty() );
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 4 ).isEmpty() );
	}


	public void testPauseResume() throws Exception {

		configureEnvironment( new EBSchedulerFacade(), new FakeEBReplicaExecutorFacade() );

		GumSpec gumSpec11 = new GumSpec();
		gumSpec11.putAttribute( GumSpec.ATT_NAME, "mockGum11" );
		GumSpec gumSpec22 = new GumSpec();
		gumSpec22.putAttribute( GumSpec.ATT_NAME, "mockGum22" );

		final Gum mockGum11 = getNiceMockGum();
		UserAgentClient uaClient1 = new UserAgentClient( mockGum11, gumSpec11 );
		final Gum mockGum22 = getNiceMockGum();
		UserAgentClient uaClient2 = new UserAgentClient( mockGum22, gumSpec22 );

		/* Reconfigure mock gump */
		Gump strictGump = this.createStrictMockGump();

		strictGump.pauseRequestGums( REQUESTID_1 );
		EasyMock.expectLastCall().once();
		strictGump.resumeRequestGums( REQUESTID_1 );
		EasyMock.expectLastCall().once();
		strictGump.pauseRequestGums( REQUESTID_1 );
		EasyMock.expectLastCall().once();
		strictGump.finishRequestGums( REQUESTID_1 );
		EasyMock.expectLastCall().once();

		EasyMock.replay( strictGump );

		/* Adding the job */
		this.ebJobManager.setMaxFails( "2" );
		JobSpec jobSpecs = createJobSpec( "", 2, ECHO_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );
		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		jobRequestEntry.addPeer( strictGump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		/* Delivering Gums to Job */
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, strictGumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		ebGridManager.hereIsGum( uaClient2, REQUESTID_1, strictGumpID );
		ebGridManager.ready( uaClient2.getGumSpec().getGumID() );

		/* Scheduling */
		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaEntry replicaEntry1 = this.schedulingHeuristic.getLatestScheduledReplica();
		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaEntry replicaEntry2 = this.schedulingHeuristic.getLatestScheduledReplica();

		ReplicaExecutorResult res1 = new ReplicaExecutorResult( replicaEntry1 );
		res1.setGumClient( uaClient1 );

		ReplicaExecutorResult res2 = new ReplicaExecutorResult( replicaEntry2 );
		res1.setGumClient( uaClient2 );

		this.schedulingHeuristic.replicaFailed( res1 );
		this.schedulingHeuristic.replicaFinished( res2 );

		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaEntry replicaEntry3 = this.schedulingHeuristic.getLatestScheduledReplica();

		ReplicaExecutorResult res3 = new ReplicaExecutorResult( replicaEntry3 );
		res1.setGumClient( uaClient1 );

		this.schedulingHeuristic.replicaFinished( res3 );

		/* Machines should have been disposed */
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );

		/* Verifying the Gum method invocation */
		EasyMock.verify( strictGump );
	}


	public void testPauseResume2() throws Exception {

		configureEnvironment( new EBSchedulerFacade(), new FakeEBReplicaExecutorFacade() );

		GumSpec gumSpec11 = new GumSpec();
		gumSpec11.putAttribute( GumSpec.ATT_NAME, "mockGum11" );
		GumSpec gumSpec22 = new GumSpec();
		gumSpec22.putAttribute( GumSpec.ATT_NAME, "mockGum22" );

		final Gum mockGum11 = getNiceMockGum();
		UserAgentClient uaClient1 = new UserAgentClient( mockGum11, gumSpec11 );
		final Gum mockGum22 = getNiceMockGum();
		UserAgentClient uaClient2 = new UserAgentClient( mockGum22, gumSpec22 );

		/* Reconfigure mock gump */
		Gump strictGump = this.createStrictMockGump();

		/* Adding the job */
		this.ebJobManager.setMaxReplicas( "2" );
		this.ebJobManager.setMaxFails( "1" );
		JobSpec jobSpecs = createJobSpec( "", 3, ECHO_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );
		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		jobRequestEntry.addPeer( strictGump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		/* Delivering Gums to Job */
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, strictGumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		ebGridManager.hereIsGum( uaClient2, REQUESTID_1, strictGumpID );
		ebGridManager.ready( uaClient2.getGumSpec().getGumID() );

		/* Scheduling */
		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaExecutorResult res1_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaExecutorResult res2_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		assertFalse( this.schedulingHeuristic.schedule() );

		this.schedulingHeuristic.replicaFinished( res1_1 );

		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaExecutorResult res3_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );

		this.schedulingHeuristic.replicaFailed( res2_1 );

		strictGump.pauseRequestGums( REQUESTID_1 );
		EasyMock.expectLastCall().once();
		EasyMock.replay( strictGump );

		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaExecutorResult res3_2 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );

		EasyMock.verify( strictGump );
		EasyMock.reset( strictGump );

		strictGump.disposeGum( REQUESTID_1, uaClient1.getGumSpec().getGumID() );
		EasyMock.expectLastCall().once();
		strictGump.finishRequestGums( REQUESTID_1 );
		EasyMock.expectLastCall().once();

		EasyMock.replay( strictGump );

		this.schedulingHeuristic.replicaFinished( res3_1 );
		this.schedulingHeuristic.replicaFinished( res3_2 );

		EasyMock.verify( strictGump );

		/* Machines should have been disposed */
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );

	}


	public void testPauseResume3() throws Exception {

		configureEnvironment( new EBSchedulerFacade(), new FakeEBReplicaExecutorFacade() );

		GumSpec gumSpec11 = new GumSpec();
		gumSpec11.putAttribute( GumSpec.ATT_NAME, "mockGum11" );
		GumSpec gumSpec22 = new GumSpec();
		gumSpec22.putAttribute( GumSpec.ATT_NAME, "mockGum22" );
		GumSpec gumSpec33 = new GumSpec();
		gumSpec33.putAttribute( GumSpec.ATT_NAME, "mockGum33" );

		final Gum mockGum11 = getNiceMockGum();
		UserAgentClient uaClient1 = new UserAgentClient( mockGum11, gumSpec11 );
		final Gum mockGum22 = getNiceMockGum();
		UserAgentClient uaClient2 = new UserAgentClient( mockGum22, gumSpec22 );
		final Gum mockGum33 = getNiceMockGum();
		UserAgentClient uaClient3 = new UserAgentClient( mockGum33, gumSpec33 );

		/* Reconfigure mock gump */
		Gump strictGump = this.createStrictMockGump();

		/* Adding the job */
		this.ebJobManager.setMaxReplicas( "1" );
		this.ebJobManager.setMaxFails( "1" );
		JobSpec jobSpecs = createJobSpec( "", 3, ECHO_COMMAND );
		schedulingHeuristic.addJob( JOBID_1, jobSpecs );
		JobRequestEntry jobRequestEntry = new JobRequestEntry( JOBID_1, REQUESTID_1 );
		jobRequestEntry.addPeer( strictGump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		/* Delivering Gums to Job */
		ebGridManager.hereIsGum( uaClient1, REQUESTID_1, strictGumpID );
		ebGridManager.ready( uaClient1.getGumSpec().getGumID() );

		ebGridManager.hereIsGum( uaClient2, REQUESTID_1, strictGumpID );
		ebGridManager.ready( uaClient2.getGumSpec().getGumID() );

		ebGridManager.hereIsGum( uaClient3, REQUESTID_1, strictGumpID );
		ebGridManager.ready( uaClient3.getGumSpec().getGumID() );

		strictGump.pauseRequestGums( REQUESTID_1 );
		EasyMock.expectLastCall().once();
		EasyMock.replay( strictGump );

		/* Scheduling */

		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaExecutorResult res1_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaExecutorResult res2_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );
		assertTrue( this.schedulingHeuristic.schedule() );
		ReplicaExecutorResult res3_1 = new ReplicaExecutorResult( this.schedulingHeuristic.getLatestScheduledReplica() );

		EasyMock.verify( strictGump );
		EasyMock.reset( strictGump );

		strictGump.unwantedGum( REQUESTID_1, uaClient1.getGumSpec().getGumID() );
		EasyMock.expectLastCall().once();
		strictGump.unwantedGum( REQUESTID_1, uaClient2.getGumSpec().getGumID() );
		EasyMock.expectLastCall().once();
		strictGump.finishRequestGums( REQUESTID_1 );
		EasyMock.expectLastCall().once();

		EasyMock.replay( strictGump );

		this.schedulingHeuristic.replicaFailed( res1_1 );
		this.schedulingHeuristic.replicaFailed( res2_1 );
		this.schedulingHeuristic.replicaFailed( res3_1 );

		EasyMock.verify( strictGump );

		/* Machines should have been disposed */
		assertTrue( ebGridManager.getAvailableGuMs( JOBID_1, 1 ).isEmpty() );

	}


	private void configureEnvironment( EBSchedulerFacade schedulerFacade,
										EBReplicaExecutorFacade ebReplicaExecutorFacade ) throws Exception {

		this.ebSchedulerFacade = schedulerFacade;
		this.ebReplicaExecutorFacade = ebReplicaExecutorFacade;

		this.ebSchedulerFacade.config( this.ebReplicaExecutorFacade );

		this.gridMachineConsumerImpl = new GumpClientImpl( ebSchedulerFacade );
		this.ebGridManager = (EBGridManager) this.ebSchedulerFacade.getEventProcessor().getGridManager();
		this.ebJobManager = this.ebSchedulerFacade.getEventProcessor().getJobManager();

		this.schedulingHeuristic = new FakeSchedulingHeuristic( this.ebJobManager, this.ebGridManager,
				this.ebReplicaExecutorFacade );

		this.ebReplicaExecutorFacade.config( this.ebSchedulerFacade );
	}


	private Gum getNiceMockGum() throws Exception {

		Gum newMockGum = EasyMock.createNiceMock( Gum.class );

		return newMockGum;
	}


	private Gump createMockGump() throws Exception {

		// Creating the mockGump
		Gump mockGump = EasyMock.createMock( Gump.class );
		EasyMock.reset( mockGump );

		// mockGump.pauseRequestGums( EasyMock.anyLong() );
		// EasyMock.expectLastCall().anyTimes();

		bindableMockGumpWrapper = new BindableMockGumpWrapper( mockGump );

		gumpID = new ObjectID( "MockGump", 42L, "rmi://localhost:" + rmiPort + "/MockGump" );

		Naming.rebind( gumpID.getObjectURL(), bindableMockGumpWrapper );

		return mockGump;

	}


	private Gump createStrictMockGump() throws Exception {

		// Creating the mockGump
		Gump mockGump = EasyMock.createMock( Gump.class );
		EasyMock.reset( mockGump );

		strictBindableMockGumpWrapper = new BindableMockGumpWrapper( mockGump );

		strictGumpID = new ObjectID( "MockGump", 42L, "rmi://localhost:" + rmiPort + "/MockGump" );

		Naming.rebind( strictGumpID.getObjectURL(), strictBindableMockGumpWrapper );

		return mockGump;

	}
}
