/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.test;

import java.io.File;
import java.rmi.Naming;
import java.rmi.Remote;
import java.rmi.registry.LocateRegistry;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.common.gum.FakeGumClient;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.gump.BindableMockGumpWrapper;
import org.ourgrid.common.gump.Gump;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.IOBlock;
import org.ourgrid.common.spec.IOEntry;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.common.spec.exception.JobSpecificationException;
import org.ourgrid.common.spec.exception.TaskSpecificationException;
import org.ourgrid.common.util.JavaFileUtil;
import org.ourgrid.common.util.TempFileManager;
import org.ourgrid.mygrid.replicaexecutor.test.FakeEBReplicaExecutorFacade;
import org.ourgrid.mygrid.scheduler.DataDiscovery;
import org.ourgrid.mygrid.scheduler.FakeEBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.JobRequestEntry;
import org.ourgrid.mygrid.scheduler.StorageAffinity;
import org.ourgrid.mygrid.scheduler.gridmanager.EBGridManager;
import org.ourgrid.mygrid.scheduler.gump.GumpClientImpl;
import org.ourgrid.mygrid.scheduler.jobmanager.EBJobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.TaskEntry;
import org.ourgrid.test.SchedulerFastTest;
import org.ourgrid.test.TestUtils;

/**
 * The test for <code>StorageAffinity</code>.
 */
public class StorageAffinityTest extends TestCase implements SchedulerFastTest {

	private StorageAffinity sa;

	private JobManager jobManager;

	private EBGridManager ebGridManager;

	private GumpClientImpl gridMachineConsumerImpl;

	private FakeEBReplicaExecutorFacade fakeEBReplicaExecutorFacade;

	private FakeEBSchedulerFacade fakeEBSchedulerFacade;

	private TaskSpec taskSpec;

	private TestUtils testUtils;

	private int rmiPort;

	private Remote bindableMockGumpWrapper;

	private ObjectID mockGumpObjectID1;

	private Gump gump;


	@Override
	protected void setUp() throws Exception {

		super.setUp();

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createMGProperties();

		Configuration.getInstance( Configuration.MYGRID );

		rmiPort = Integer.parseInt( testUtils.getProperty( TestUtils.RMI_PORT ) );
		try {
			LocateRegistry.createRegistry( rmiPort );
		} catch ( Exception e ) {
			LocateRegistry.getRegistry( rmiPort );
		}

		jobManager = new EBJobManager();
		fakeEBSchedulerFacade = new FakeEBSchedulerFacade();
		gridMachineConsumerImpl = new GumpClientImpl( fakeEBSchedulerFacade );
		ebGridManager = new EBGridManager( fakeEBSchedulerFacade, gridMachineConsumerImpl );
		fakeEBReplicaExecutorFacade = new FakeEBReplicaExecutorFacade();

		gump = createNiceMockGump();

		DataDiscovery.resetInstance();

		sa = new StorageAffinity( jobManager, ebGridManager, fakeEBReplicaExecutorFacade );
	}


	@Override
	protected void tearDown() throws Exception {

		super.tearDown();

		ebGridManager.shutdown();
		EventQueue<ShutdownResponseEvent> shutdownResponse = fakeEBSchedulerFacade.shutdown();
		shutdownResponse.blockingRemove();
		OurgridUnicastRemoteObject.unexportObject( this.gridMachineConsumerImpl, true );

		sa = null;

		try {
			Naming.unbind( "rmi://localhost:" + rmiPort + "/MockGump" );
			OurgridUnicastRemoteObject.unexportObject( bindableMockGumpWrapper, true );
		} catch ( Exception e ) {
		}

		DataDiscovery.resetInstance();
		Configuration.reset();

		testUtils.reset();
	}


	/**
	 * Creates a JobSpec with number of tasks and a required file specified.
	 * 
	 * @return A JobSpec with this requirements.
	 * @throws TaskSpecificationException
	 * @throws JobSpecificationException
	 */
	private JobSpec createJob( int numberOfTasks, File[ ] fileInTask ) throws TaskSpecificationException,
		JobSpecificationException {

		IOBlock initBlock = new IOBlock();
		for ( int i = 0; i < fileInTask.length; i++ ) {
			IOEntry ioEntry = new IOEntry( "STORE", fileInTask[i].getAbsolutePath(), fileInTask[i].getName() );
			initBlock.putEntry( ioEntry );
		}

		TaskSpec taskSpec = new TaskSpec( initBlock, "cd /tmp", null );
		List<TaskSpec> taskSpecList = new LinkedList<TaskSpec>();
		for ( int i = 0; i < numberOfTasks; i++ ) {
			taskSpecList.add( taskSpec );
		}

		JobSpec job = new JobSpec( "TestJob", "", taskSpecList );
		return job;
	}


	/**
	 * Tests the <code>getAffinity()</code> method with only one task and one
	 * gum.
	 * 
	 * @throws Exception If something wrong happens.
	 */
	public void testGetAffinityWithOneTaskAndOneGuM() throws Exception {

		Map<String,String> atts = new HashMap<String,String>();
		atts.put( GumSpec.ATT_NAME, "FakeGuM1" );
		atts.put( GumSpec.ATT_SITE, "localpeer" );

		GumSpec fakeGuMSpec = new GumSpec( atts );

		String tempDir = testUtils.getProperty( TestUtils.SYSTEM_TEMP_DIR );

		File file1 = TempFileManager.createTempFileWithBogusData( "sa", "test", new File( tempDir ), 256 );
		File file2 = TempFileManager.createTempFileWithBogusData( "sa", "test", new File( tempDir ), 512 );
		File file3 = TempFileManager.createTempFileWithBogusData( "sa", "test", new File( tempDir ), 1024 );

		GumClient gumexMock = new FakeGumClient( fakeGuMSpec, new File[ ] { new File( "test.StorageAffinityTest" ),
																			new File( "test3.StorageAffinityTest" ) },
				new long[ ] { file1.length(), file3.length() } );

		IOBlock initBlock = new IOBlock();
		IOEntry ioEntry = new IOEntry( "STORE", file1.getAbsolutePath(), "test.StorageAffinityTest" );
		initBlock.putEntry( ioEntry );
		ioEntry = new IOEntry( "STORE", file2.getAbsolutePath(), "test2.StorageAffinityTest" );
		initBlock.putEntry( ioEntry );
		ioEntry = new IOEntry( "STORE", file3.getAbsolutePath(), "test3.StorageAffinityTest" );
		initBlock.putEntry( "condition1", ioEntry );

		taskSpec = new TaskSpec( initBlock, "cd /tmp", null );
		TaskEntry task = new TaskEntry( taskSpec, 0, 0, 0, 0 );

		DataDiscovery.getInstance().putDDEntry( JavaFileUtil.getDigestRepresentation( file1 ),
				gumexMock.getStorageDirectory() + "/" + "test.StorageAffinityTest", file1.length(), gumexMock );
		DataDiscovery.getInstance().putDDEntry( JavaFileUtil.getDigestRepresentation( file3 ),
				gumexMock.getStorageDirectory() + "/" + "test3.StorageAffinityTest", file3.length(), gumexMock );

		long affinity = sa.getAffinity( task, gumexMock );

		assertEquals( file1.length() + file3.length(), affinity );

	}


	/**
	 * Tests the <code>schedule()</code> method with only one task and one
	 * gum.
	 * 
	 * @throws Exception If something wrong happens.
	 */
	public void testScheduleWithOneTaskAndOneGuM() throws Exception {

		IOBlock initBlock = new IOBlock();
		IOEntry ioEntry = new IOEntry( "", "", "test.StorageAffinityTest" );
		initBlock.putEntry( ioEntry );

		taskSpec = new TaskSpec( initBlock, "cd /tmp", null );
		new TaskEntry( taskSpec, 0, 0, 0, 0 );
		List<TaskSpec> taskSpecList = new LinkedList<TaskSpec>();
		taskSpecList.add( taskSpec );
		JobSpec job = new JobSpec( "TestJob", "", taskSpecList );

		File file1 = new File( ".mgstorage/test.StorageAffinityTest" );

		Map<String,String> atts = new HashMap<String,String>();

		atts.put( GumSpec.ATT_STORAGE_SHARED, "no" );
		atts.put( GumSpec.ATT_SITE, "lsd" );
		atts.put( GumSpec.ATT_NAME, "fakeGuMex1" );

		GumSpec fakeGuMSpec = new GumSpec( atts );
		GumClient fakeGuMex1 = new FakeGumClient( fakeGuMSpec, new File[ ] { file1 }, new long[ ] { 1024 } );

		int maxReplicas = 1;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.clonedJobList().size() == 0 );

		sa.addJob( 1, job );

		JobRequestEntry jobRequestEntry = new JobRequestEntry( 1, 1 );
		jobRequestEntry.addPeer( gump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );
		ebGridManager.hereIsGum( fakeGuMex1, 1, mockGumpObjectID1 );

		assertTrue( sa.schedule() );

		ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
		assertEquals( fakeGuMex1.getGumSpec().getGumID(), lastScheduledReplica.getGumID() );

		assertFalse( sa.schedule() );

	}


	/**
	 * Tests the <code>schedule()</code> method with three task and four gums.
	 * 
	 * @throws Exception If something wrong happens.
	 */
	public void testScheduleWithThreeTasksAndFourGuMs() throws Exception {

		String tempDir = testUtils.getProperty( TestUtils.SYSTEM_TEMP_DIR );

		File file2 = TempFileManager.createTempFileWithBogusData( "sa", "test", new File( tempDir ), 1024 );
		File file3 = TempFileManager.createTempFileWithBogusData( "sa", "test", new File( tempDir ), 2048 );

		File[ ] filesJob2 = { file2, file3 };

		JobSpec job1 = createJob( 10, filesJob2 );

		long size2 = 768;
		long size3 = 2048;

		Map<String,String> atts1 = new HashMap<String,String>();

		atts1.put( GumSpec.ATT_STORAGE_SHARED, "no" );
		atts1.put( GumSpec.ATT_SITE, "lsd" );
		atts1.put( GumSpec.ATT_NAME, "fakeGuMex1" );

		GumSpec fakeGuMSpec1 = new GumSpec( atts1 );

		Map<String,String> atts2 = new HashMap<String,String>();

		atts2.put( GumSpec.ATT_STORAGE_SHARED, "no" );
		atts2.put( GumSpec.ATT_SITE, "lsd" );
		atts2.put( GumSpec.ATT_NAME, "fakeGuMex2" );

		GumSpec fakeGuMSpec2 = new GumSpec( atts2 );

		Map<String,String> atts3 = new HashMap<String,String>();

		atts3.put( GumSpec.ATT_STORAGE_SHARED, "no" );
		atts3.put( GumSpec.ATT_SITE, "lsd" );
		atts3.put( GumSpec.ATT_NAME, "fakeGuMex3" );

		GumSpec fakeGuMSpec3 = new GumSpec( atts3 );

		Map<String,String> atts4 = new HashMap<String,String>();

		atts4.put( GumSpec.ATT_STORAGE_SHARED, "no" );
		atts4.put( GumSpec.ATT_SITE, "lsd" );
		atts4.put( GumSpec.ATT_NAME, "fakeGuMex4" );

		GumSpec fakeGuMSpec4 = new GumSpec( atts4 );

		GumClient gum1 = new FakeGumClient( fakeGuMSpec1, new File[ ] { file2 }, new long[ ] { size2 } );
		GumClient gum2 = new FakeGumClient( fakeGuMSpec2, new File[ ] { file2, file3 }, new long[ ] { size2, size3 } );
		GumClient gum3 = new FakeGumClient( fakeGuMSpec3, new File[ ] { file3 }, new long[ ] { size3 } );
		GumClient gum4 = new FakeGumClient( fakeGuMSpec4, new File[ ] {}, new long[ ] {} );

		DataDiscovery.getInstance().putDDEntry( JavaFileUtil.getDigestRepresentation( file2 ),
				gum1.getStorageDirectory() + "/" + file2.getName(), file2.length(), gum1 );
		DataDiscovery.getInstance().putDDEntry( JavaFileUtil.getDigestRepresentation( file2 ),
				gum2.getStorageDirectory() + "/" + file2.getName(), file2.length(), gum2 );
		DataDiscovery.getInstance().putDDEntry( JavaFileUtil.getDigestRepresentation( file3 ),
				gum2.getStorageDirectory() + "/" + file3.getName(), file3.length(), gum2 );
		DataDiscovery.getInstance().putDDEntry( JavaFileUtil.getDigestRepresentation( file3 ),
				gum3.getStorageDirectory() + "/" + file3.getName(), file3.length(), gum3 );
		DataDiscovery.getInstance().putDDEntry( JavaFileUtil.getDigestRepresentation( file2 ),
				gum4.getStorageDirectory() + "/" + file2.getName(), file2.length(), gum4 );

		int maxReplicas = 2;
		Configuration.getInstance()
				.setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, Integer.toString( maxReplicas ) );
		jobManager.setMaxReplicas( Integer.toString( maxReplicas ) );

		assertTrue( jobManager.clonedJobList().size() == 0 );

		sa.addJob( 1, job1 );

		JobRequestEntry jobRequestEntry;

		jobRequestEntry = new JobRequestEntry( 1, 1 );
		jobRequestEntry.addPeer( gump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );

		ebGridManager.hereIsGum( gum1, 1, mockGumpObjectID1 );
		ebGridManager.hereIsGum( gum2, 1, mockGumpObjectID1 );
		ebGridManager.hereIsGum( gum3, 1, mockGumpObjectID1 );
		ebGridManager.hereIsGum( gum4, 1, mockGumpObjectID1 );

		assertEquals( 4, ebGridManager.getAvailableAndValidGums().size() );

		assertTrue( sa.schedule() );
		ReplicaEntry lastReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
		assertEquals( 1, lastReplica.getTaskId() );
		assertEquals( gum2.getGumSpec().getGumID(), lastReplica.getGumID() );

		assertTrue( sa.schedule() );
		lastReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
		assertEquals( 2, lastReplica.getTaskId() );
		assertEquals( gum3.getGumSpec().getGumID(), lastReplica.getGumID() );

		assertTrue( sa.schedule() );
		lastReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
		assertEquals( 3, lastReplica.getTaskId() );
		assertEquals( gum1.getGumSpec().getGumID(), lastReplica.getGumID() );

		assertTrue( sa.schedule() );
		lastReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
		assertEquals( 4, lastReplica.getTaskId() );
		assertEquals( gum4.getGumSpec().getGumID(), lastReplica.getGumID() );

		// shouldn't be scheduled
		assertTrue( !sa.schedule() );
		lastReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
		// the last replica doesn't change
		assertEquals( 4, lastReplica.getTaskId() );
		assertEquals( gum4.getGumSpec().getGumID(), lastReplica.getGumID() );

	}


	private Gump createNiceMockGump() throws Exception {

		String peerName = "test.peername";
		String peerPort = "0";

		// Creating the mockGump
		Gump mockGump = EasyMock.createNiceMock( Gump.class );

		bindableMockGumpWrapper = new BindableMockGumpWrapper( mockGump );

		mockGumpObjectID1 = new ObjectID( "MockGump", (long) 42, "rmi://localhost:" + rmiPort + "/MockGump" );

		Naming.rebind( mockGumpObjectID1.getObjectURL(), bindableMockGumpWrapper );

		return mockGump;
	}
}
