/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.test;

import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.gump.Gump;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.common.spec.exception.JobSpecificationException;
import org.ourgrid.common.spec.exception.TaskSpecificationException;
import org.ourgrid.mygrid.replicaexecutor.test.FakeEBReplicaExecutorFacade;
import org.ourgrid.mygrid.scheduler.FakeEBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.Workqueue;
import org.ourgrid.mygrid.scheduler.gridmanager.EBGridManager;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManager;
import org.ourgrid.mygrid.scheduler.gump.GumpClientImpl;
import org.ourgrid.mygrid.scheduler.jobmanager.EBJobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.test.StressTest;

public class WorkqueueStressTest extends TestCase implements StressTest {

	private static final ObjectID OBJECT_ID_1 = new ObjectID( "Gambit", (long) 4242, "rmi://gambit:4242/XMen" );

	private JobManager jobManager;

	private GridManager ebGridManager;

	private GumpClientImpl gridMachineConsumerImpl;

	private FakeEBReplicaExecutorFacade fakeEBReplicaExecutorFacade;

	private Workqueue wkq;

	private Gump mockGump;

	private GumClient mockGumClient;

	private FakeEBSchedulerFacade fakeEBSchedulerFacade;

	private boolean verifyMocks;


	/**
	 * Constructor for WorkqueueStressTest.
	 * 
	 * @param name
	 */
	public WorkqueueStressTest( String name ) throws SecurityException {

		super( name );
	}


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();
		Configuration.getInstance( Configuration.MYGRID );
		verifyMocks = false;
		jobManager = new EBJobManager();
		fakeEBSchedulerFacade = new FakeEBSchedulerFacade();
		gridMachineConsumerImpl = new GumpClientImpl( fakeEBSchedulerFacade );
		ebGridManager = new EBGridManager( fakeEBSchedulerFacade, gridMachineConsumerImpl );
		fakeEBReplicaExecutorFacade = new FakeEBReplicaExecutorFacade();
		wkq = new Workqueue( jobManager, ebGridManager, fakeEBReplicaExecutorFacade );
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		ebGridManager.shutdown();
		EventQueue shutdownResponse = fakeEBSchedulerFacade.shutdown();
		shutdownResponse.blockingRemove();

		wkq = null;
		OurgridUnicastRemoteObject.unexportObject( this.gridMachineConsumerImpl, true );
		if ( verifyMocks ) {
			EasyMock.verify( mockGumClient );
			EasyMock.verify( mockGump );
		}

		Configuration.reset();
	}


	/**
	 * Creates a handful of jobs, submits them to workqueue, and check if the
	 * jobs are there.
	 * 
	 * @throws Exception when a called method throws an exception
	 */
	public void testAddJob() throws Exception {

		assertTrue( jobManager.jobList().size() == 0 );

		// creating the jobs
		// the jobs' attributes
		String[ ] jobAttribs = { "os==linux", "location==lsd", "university==ufcg" };
		int[ ] numTasks = { 1000, 1500, 1100 };
		int numberOfTasks = 0;
		for ( int i = 0; i < numTasks.length; i++ ) {
			numberOfTasks += numTasks[i];
		}

		JobSpec[ ] jobSpecs = new JobSpec[ jobAttribs.length ];
		for ( int i = 0; i < jobSpecs.length; i++ ) {
			jobSpecs[i] = createJobSpec( jobAttribs[i], numTasks[i] );
		}

		for ( int i = 0; i < jobSpecs.length; i++ ) {
			wkq.addJob( i + 1, jobSpecs[i] );
		}
		assertTrue( jobManager.jobList().size() == jobSpecs.length );

	}


	public void testGetNumberOfGuMsNeeded() throws RemoteException, TaskSpecificationException,
		JobSpecificationException {

		createOneMockGuMAndOneMockGuMP();
		ebGridManager.hereIsGum( mockGumClient, 1, OBJECT_ID_1 );

		int numTasks = 5;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, "0" );
		int maxReplicas = Integer.parseInt( Configuration.getInstance().getProperty(
				MyGridConfiguration.PROP_MAX_REPLICAS ) );
		JobSpec jobSpec = createJobSpec( "linux", numTasks );
		assertEquals( "The number of GridMachines needed should be " + numTasks + " * " + maxReplicas + ".", numTasks
				* maxReplicas, wkq.getNumberOfNeededGums( jobSpec ) );

		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, "-1" );
		maxReplicas = Integer
				.parseInt( Configuration.getInstance().getProperty( MyGridConfiguration.PROP_MAX_REPLICAS ) );
		jobSpec = createJobSpec( "linux", numTasks );
		assertEquals( "The number of GridMachines needed should be " + numTasks + " * " + maxReplicas + ".", numTasks
				* maxReplicas, wkq.getNumberOfNeededGums( jobSpec ) );

		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, "1" );
		maxReplicas = Integer
				.parseInt( Configuration.getInstance().getProperty( MyGridConfiguration.PROP_MAX_REPLICAS ) );
		jobSpec = createJobSpec( "linux", numTasks );
		assertEquals( "The number of GridMachines needed should be " + numTasks + " * " + maxReplicas + ".", numTasks
				* maxReplicas, wkq.getNumberOfNeededGums( jobSpec ) );

		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, "100000" );
		maxReplicas = Integer
				.parseInt( Configuration.getInstance().getProperty( MyGridConfiguration.PROP_MAX_REPLICAS ) );
		jobSpec = createJobSpec( "linux", numTasks );
		assertEquals( "The number of GridMachines needed should be " + numTasks + " * " + maxReplicas + ".", numTasks
				* maxReplicas, wkq.getNumberOfNeededGums( jobSpec ) );
	}


	public void testScheduleOneJob() throws ConfigException, RemoteException, TaskSpecificationException,
		JobSpecificationException {

		createOneMockGuMAndOneMockGuMP();
		ebGridManager.hereIsGum( mockGumClient, 1, OBJECT_ID_1 );

		int maxReplicas = 1;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.jobList().size() == 0 );
		String jobAttribs = "os==linux";
		int numTasks = 1000;

		JobSpec jobSpecs = createJobSpec( jobAttribs, numTasks );
		wkq.addJob( 1, jobSpecs );

		for ( int taskId = 1; taskId <= numTasks; taskId++ ) {
			assertTrue( wkq.schedule() );
			ebGridManager.ready( mockGumClient.getGumSpec().getGumID() );
			ReplicaEntry requiredReplica = new ReplicaEntry( 1, taskId, 1, new TaskSpec( null, "echo 0", null ) );
			ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
			assertReplicaId( requiredReplica, lastScheduledReplica );
		}
		assertTrue( !wkq.schedule() );

		verifyMocks = true;
	}


	public void testScheduleThreeJobs() throws ConfigException, RemoteException, TaskSpecificationException,
		JobSpecificationException {

		createOneMockGuMAndOneMockGuMP();
		ebGridManager.hereIsGum( mockGumClient, 1, OBJECT_ID_1 );

		int maxReplicas = 1;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.jobList().size() == 0 );
		String[ ] jobAttribs = { "os==linux", "location==lsd", "university==ufcg" };
		int[ ] numTasks = { 1000, 2000, 1600 };
		int numberOfTasks = 0;
		for ( int i = 0; i < numTasks.length; i++ ) {
			numberOfTasks += numTasks[i];
		}

		JobSpec[ ] jobSpecs = new JobSpec[ jobAttribs.length ];
		for ( int i = 0; i < jobSpecs.length; i++ ) {
			jobSpecs[i] = createJobSpec( jobAttribs[i], numTasks[i] );
		}

		for ( int i = 0; i < jobSpecs.length; i++ ) {
			wkq.addJob( i + 1, jobSpecs[i] );
		}

		for ( int jobId = 1; jobId <= jobAttribs.length; jobId++ ) {
			for ( int taskId = 1; taskId <= numTasks[jobId - 1]; taskId++ ) {
				assertTrue( wkq.schedule() );
				ebGridManager.ready( mockGumClient.getGumSpec().getGumID() );
				ReplicaEntry requiredReplica = new ReplicaEntry( 1, taskId, jobId, new TaskSpec( null, "echo 0", null ) );
				ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
				assertReplicaId( requiredReplica, lastScheduledReplica );
			}
		}
		assertTrue( !wkq.schedule() );

		verifyMocks = true;
	}


	public void testScheduleOneJobWithReplication() throws ConfigException, RemoteException,
		TaskSpecificationException, JobSpecificationException {

		createOneMockGuMAndOneMockGuMP();
		ebGridManager.hereIsGum( mockGumClient, 1, OBJECT_ID_1 );

		int maxReplicas = 150;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.jobList().size() == 0 );
		String jobAttribs = "os==linux";
		int numTasks = 100;

		JobSpec jobSpecs = createJobSpec( jobAttribs, numTasks );
		wkq.addJob( 1, jobSpecs );
		for ( int replicaId = 1; replicaId <= maxReplicas; replicaId++ ) {
			for ( int taskId = 1; taskId <= numTasks; taskId++ ) {
				assertTrue( wkq.schedule() );
				ebGridManager.ready( mockGumClient.getGumSpec().getGumID() );
				ReplicaEntry requiredReplica = new ReplicaEntry( replicaId, taskId, 1, new TaskSpec( null, "echo 0",
						null ) );
				ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
				assertReplicaId( requiredReplica, lastScheduledReplica );
			}
		}

		for ( int k = 0; k < 100; k++ ) {
			assertTrue( !wkq.schedule() );
		}

		verifyMocks = true;
	}


	public void testScheduleThreeJobsWithReplication() throws ConfigException, RemoteException,
		TaskSpecificationException, JobSpecificationException {

		createOneMockGuMAndOneMockGuMP();
		ebGridManager.hereIsGum( mockGumClient, 1, OBJECT_ID_1 );
		int maxReplicas = 50;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.jobList().size() == 0 );
		String[ ] jobAttribs = { "os==linux", "location==lsd", "university==ufcg", "location==lsd", "university==ufcg",
								"location==lsd", "university==ufcg", "location==lsd", "university==ufcg",
								"location==lsd", "university==ufcg", "location==lsd", "university==ufcg" };
		int[ ] numTasks = { 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50 };
		int numberOfTasks = 0;
		for ( int i = 0; i < numTasks.length; i++ ) {
			numberOfTasks += numTasks[i];
		}

		JobSpec[ ] jobSpecs = new JobSpec[ jobAttribs.length ];
		for ( int i = 0; i < jobSpecs.length; i++ ) {
			jobSpecs[i] = createJobSpec( jobAttribs[i], numTasks[i] );
		}

		for ( int i = 0; i < jobSpecs.length; i++ ) {
			wkq.addJob( i + 1, jobSpecs[i] );
		}

		for ( int replicaId = 1; replicaId <= maxReplicas; replicaId++ ) {
			for ( int jobId = 1; jobId <= jobAttribs.length; jobId++ ) {
				for ( int taskId = 1; taskId <= numTasks[jobId - 1]; taskId++ ) {
					assertTrue( wkq.schedule() );
					ebGridManager.ready( mockGumClient.getGumSpec().getGumID() );
					ReplicaEntry requiredReplica = new ReplicaEntry( replicaId, taskId, jobId, new TaskSpec( null,
							"echo 0", null ) );
					ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
					assertReplicaId( requiredReplica, lastScheduledReplica );
				}
			}
		}
		assertTrue( !wkq.schedule() );

		verifyMocks = true;
	}


	/**
	 * This test verifies if the heuristic scheduled the replicas in the correct
	 * order. These test simulates job adding during other jobs execution.
	 * 
	 * @throws RemoteException
	 * @throws TaskSpecificationException
	 * @throws JobSpecificationException
	 */
	public void testScheduleThreeJobsWithReplicationIntercalated() throws ConfigException, RemoteException,
		TaskSpecificationException, JobSpecificationException {

		createOneMockGuMAndOneMockGuMP();
		ebGridManager.hereIsGum( mockGumClient, 1, OBJECT_ID_1 );

		int maxReplicas = 3;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.jobList().size() == 0 );
		String[ ] jobAttribs = { "os==linux", "location==lsd", "university==ufcg" };
		int[ ] numTasks = { 50, 50, 50 };
		int numberOfTasks = 0;
		for ( int i = 0; i < numTasks.length; i++ ) {
			numberOfTasks += numTasks[i];
		}

		JobSpec[ ] jobSpecs = new JobSpec[ jobAttribs.length ];
		for ( int i = 0; i < jobSpecs.length; i++ ) {
			jobSpecs[i] = createJobSpec( jobAttribs[i], numTasks[i] );
		}

		ReplicaEntry[ ] replicasArray = new ReplicaEntry[ maxReplicas * numberOfTasks ];
		int[ ] numberOfTasksToScheduleArray = { 80, 50, 20, 50, 50, 20, 100, 30, 50 };
		int[ ] initialReplicaIdArray = { 1, 1, 2, 2, 3, 3, 1, 3, 3 };
		int[ ] initialTaskIdArray = { 1, 1, 31, 1, 1, 1, 1, 21, 1 };
		int[ ] jobIdArray = { 1, 2, 1, 2, 1, 2, 3, 2, 3 };
		boolean[ ] addJobArray = { true, true, false, false, false, false, true, false, false };

		int currentTasksScheduled = 0;
		for ( int x = 0; x < jobIdArray.length; x++ ) {
			if ( addJobArray[x] ) {
				wkq.addJob( jobIdArray[x], jobSpecs[jobIdArray[x] - 1] );
			}
			addReplicasToArray( replicasArray, maxReplicas, numberOfTasksToScheduleArray[x], currentTasksScheduled,
					initialReplicaIdArray[x], initialTaskIdArray[x], numTasks, jobIdArray[x] );
			for ( int replicaIndex = currentTasksScheduled; replicaIndex < (numberOfTasksToScheduleArray[x] + currentTasksScheduled); replicaIndex++ ) {
				assertTrue( wkq.schedule() );
				ebGridManager.ready( mockGumClient.getGumSpec().getGumID() );
				ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
				assertReplicaId( replicasArray[replicaIndex], lastScheduledReplica );
			}
			currentTasksScheduled += numberOfTasksToScheduleArray[x];
		}

		assertTrue( !wkq.schedule() );

		verifyMocks = true;
	}


	/**
	 * This test verifies if the heuristic scheduled the replicas in the correct
	 * order. These test simulates job adding during other jobs execution.
	 * 
	 * @throws TaskSpecificationException
	 * @throws JobSpecificationException
	 * @throws RemoteException
	 */
	public void testOtherScheduleThreeJobsWithReplicationIntercalated() throws ConfigException, RemoteException,
		TaskSpecificationException, JobSpecificationException {

		createOneMockGuMAndOneMockGuMP();
		ebGridManager.hereIsGum( mockGumClient, 1, OBJECT_ID_1 );

		int maxReplicas = 3;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.jobList().size() == 0 );
		String[ ] jobAttribs = { "os==linux", "location==lsd", "university==ufcg" };
		int[ ] numTasks = { 500, 500, 500 };
		int numberOfTasks = 0;
		for ( int i = 0; i < numTasks.length; i++ ) {
			numberOfTasks += numTasks[i];
		}

		JobSpec[ ] jobSpecs = new JobSpec[ jobAttribs.length ];
		for ( int i = 0; i < jobSpecs.length; i++ ) {
			jobSpecs[i] = createJobSpec( jobAttribs[i], numTasks[i] );
		}

		ReplicaEntry[ ] replicasArray = new ReplicaEntry[ maxReplicas * numberOfTasks ];
		int[ ] numberOfTasksToScheduleArray = { 800, 500, 200, 500, 500, 200, 1000, 300, 500 };
		int[ ] initialReplicaIdArray = { 1, 1, 2, 2, 3, 3, 1, 3, 3 };
		int[ ] initialTaskIdArray = { 1, 1, 301, 1, 1, 1, 1, 201, 1 };
		int[ ] jobIdArray = { 1, 2, 1, 2, 1, 2, 3, 2, 3 };
		boolean[ ] addJobArray = { true, true, false, false, false, false, true, false, false };

		int currentTasksScheduled = 0;
		for ( int x = 0; x < jobIdArray.length; x++ ) {
			if ( addJobArray[x] ) {
				wkq.addJob( jobIdArray[x], jobSpecs[jobIdArray[x] - 1] );
			}
			addReplicasToArray( replicasArray, maxReplicas, numberOfTasksToScheduleArray[x], currentTasksScheduled,
					initialReplicaIdArray[x], initialTaskIdArray[x], numTasks, jobIdArray[x] );
			for ( int replicaIndex = currentTasksScheduled; replicaIndex < (numberOfTasksToScheduleArray[x] + currentTasksScheduled); replicaIndex++ ) {
				assertTrue( wkq.schedule() );
				ebGridManager.ready( mockGumClient.getGumSpec().getGumID() );
				ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
				assertReplicaId( replicasArray[replicaIndex], lastScheduledReplica );
			}
			currentTasksScheduled += numberOfTasksToScheduleArray[x];
		}

		assertTrue( !wkq.schedule() );

		verifyMocks = true;
	}


	private void assertReplicaId( ReplicaEntry replica1, ReplicaEntry replica2 ) {

		assertEquals( replica1.getJobId(), replica2.getJobId() );
		assertEquals( replica1.getTaskId(), replica2.getTaskId() );
		assertEquals( replica1.getId(), replica2.getId() );
	}


	private void addReplicasToArray( ReplicaEntry[ ] replicasArray, int maxReplicas, int numberOfTasksToSchedule,
										int currentTasksScheduled, int initialReplicaId, int initialTaskId,
										int[ ] jobNumTasks, int jobId ) throws TaskSpecificationException {

		int replicaIndex = currentTasksScheduled;
		for ( int replicaId = initialReplicaId; replicaId <= maxReplicas; replicaId++ ) {
			for ( int taskId = initialTaskId; taskId <= jobNumTasks[jobId - 1]; taskId++ ) {
				replicasArray[replicaIndex] = new ReplicaEntry( replicaId, taskId, jobId, new TaskSpec( null, "echo 0",
						null ) );
				replicaIndex++;
				if ( replicaIndex == numberOfTasksToSchedule ) {
					return;
				}
			}
		}
	}


	/**
	 * Creates a job with <code>numTasks</code> identical tasks. Each task
	 * simply performs "echo 0".
	 * 
	 * @param expression The job attribute expression
	 * @param numTasks The number of tasks in the job
	 * @return The created job
	 * @throws TaskSpecificationException When the <code>TaskSpec</code> is
	 *         not valid
	 * @throws JobSpecificationException
	 */
	private JobSpec createJobSpec( String expression, int numTasks ) throws TaskSpecificationException,
		JobSpecificationException {

		List<TaskSpec> taskSpecList = new LinkedList<TaskSpec>();

		for ( int i = 0; i < numTasks; i++ ) {
			TaskSpec t = new TaskSpec( null, "echo 0", null );
			taskSpecList.add( t );
		}

		return new JobSpec( "TestJob", expression, taskSpecList );
	}


	private void createOneMockGuMAndOneMockGuMP() throws RemoteException {

		// Setting the GridMachines
		mockGumClient = EasyMock.createMock( GumClient.class );
		mockGumClient.getGumSpec().getGumID();
		EasyMock.expectLastCall().andReturn( "MockGuM1" ).anyTimes();
		mockGumClient.getGumSpec().getAttributes();
		Map<String,String> attribs = new HashMap<String,String>();
		attribs.put( "os", "linux" );
		attribs.put( "location", "lsd" );
		attribs.put( "university", "ufcg" );
		EasyMock.expectLastCall().andReturn( attribs ).anyTimes();

		// Creating the mockGump
		mockGump = EasyMock.createMock( Gump.class );
		mockGump.disposeGum( 1, new GumID( "MockGuM1", "0", "peername", "0" ) );
		EasyMock.expectLastCall().once();

		EasyMock.replay( mockGumClient );
		EasyMock.replay( mockGump );
	}
}
