/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.scheduler.test;

import java.rmi.Naming;
import java.rmi.Remote;
import java.rmi.registry.LocateRegistry;
import java.util.LinkedList;
import java.util.List;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.gump.BindableMockGumpWrapper;
import org.ourgrid.common.gump.Gump;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.TaskSpec;
import org.ourgrid.common.spec.exception.JobSpecificationException;
import org.ourgrid.common.spec.exception.TaskSpecificationException;
import org.ourgrid.mygrid.replicaexecutor.test.FakeEBReplicaExecutorFacade;
import org.ourgrid.mygrid.replicaexecutor.test.FakeTaskSpec;
import org.ourgrid.mygrid.scheduler.FakeEBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.JobRequestEntry;
import org.ourgrid.mygrid.scheduler.Workqueue;
import org.ourgrid.mygrid.scheduler.gridmanager.EBGridManager;
import org.ourgrid.mygrid.scheduler.gump.GumpClientImpl;
import org.ourgrid.mygrid.scheduler.jobmanager.EBJobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManager;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.test.SchedulerFastTest;
import org.ourgrid.test.TestUtils;

public class WorkqueueTest extends TestCase implements SchedulerFastTest {

	private static final String ECHO_COMMAND = "echo 0";

	private JobManager jobManager;

	private EBGridManager ebGridManager;

	private GumpClientImpl gridMachineConsumerImpl;

	private FakeEBReplicaExecutorFacade fakeEBReplicaExecutorFacade;

	private Workqueue wkq;

	private Gump mockGump;

	private GumClient mockGum;

	private FakeEBSchedulerFacade fakeEBSchedulerFacade;

	private boolean verifyMocks;

	private TestUtils testUtils;

	private int rmiPort;

	private ObjectID mockGumpObjectID1;

	private Remote boundableMockGumpWrapper;


	/**
	 * Constructor for WorkqueueTest.
	 * 
	 * @param name
	 */
	public WorkqueueTest( String name ) throws SecurityException {

		super( name );

	}


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();
		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createMGProperties();

		rmiPort = Integer.parseInt( testUtils.getProperty( TestUtils.RMI_PORT ) );
		try {
			LocateRegistry.createRegistry( rmiPort );
		} catch ( Exception e ) {
			LocateRegistry.getRegistry( rmiPort );
		}

		Configuration.getInstance( Configuration.MYGRID );
		verifyMocks = false;
		jobManager = new EBJobManager();
		fakeEBSchedulerFacade = new FakeEBSchedulerFacade();
		gridMachineConsumerImpl = new GumpClientImpl( fakeEBSchedulerFacade );
		ebGridManager = new EBGridManager( fakeEBSchedulerFacade, gridMachineConsumerImpl );
		fakeEBReplicaExecutorFacade = new FakeEBReplicaExecutorFacade();
		wkq = new Workqueue( jobManager, ebGridManager, fakeEBReplicaExecutorFacade );
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		ebGridManager.shutdown();
		EventQueue shutdownResponse = fakeEBSchedulerFacade.shutdown();
		shutdownResponse.blockingRemove();

		if ( verifyMocks ) {
			EasyMock.verify( mockGum );
		}

		wkq = null;
		OurgridUnicastRemoteObject.unexportObject( this.gridMachineConsumerImpl, true );

		try {
			Naming.unbind( mockGumpObjectID1.getObjectURL() );
			OurgridUnicastRemoteObject.unexportObject( boundableMockGumpWrapper, true );
		} catch ( Exception e ) {
		}

		Configuration.reset();
		testUtils.reset();

	}


	/**
	 * Creates a handful of jobs, submits them to workqueue, and check if the
	 * jobs are there.
	 * 
	 * @throws Exception when a called method throws an exception
	 */
	public void testAddJob() throws Exception {

		assertTrue( jobManager.clonedJobList().size() == 0 );

		// creating the jobs
		// the jobs' attributes
		String[ ] expressions = { "os==linux", "location==lsd", "location==lsd && os==linux" };
		int[ ] numTasks = { 3, 5, 10 };
		int numberOfTasks = 0;
		for ( int i = 0; i < numTasks.length; i++ ) {
			numberOfTasks += numTasks[i];
		}

		JobSpec[ ] jobSpecs = new JobSpec[ expressions.length ];
		for ( int i = 0; i < jobSpecs.length; i++ ) {
			jobSpecs[i] = createJobSpec( expressions[i], numTasks[i], ECHO_COMMAND );
		}

		for ( int i = 0; i < jobSpecs.length; i++ ) {
			wkq.addJob( i + 1, jobSpecs[i] );
		}

		assertEquals( expressions.length, jobManager.clonedJobList().size() );

	}


	public void testGetNumberOfGuMsNeeded() throws TaskSpecificationException, JobSpecificationException {

		int numTasks = 5;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, "0" );
		int maxReplicas = Integer.parseInt( Configuration.getInstance().getProperty(
				MyGridConfiguration.PROP_MAX_REPLICAS ) );
		JobSpec jobSpec = createJobSpec( "so=linux", numTasks, ECHO_COMMAND );
		assertEquals( "The number of GridMachines needed should be " + numTasks + " * " + maxReplicas + ".", numTasks
				* maxReplicas, wkq.getNumberOfNeededGums( jobSpec ) );

		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, "1" );
		maxReplicas = Integer
				.parseInt( Configuration.getInstance().getProperty( MyGridConfiguration.PROP_MAX_REPLICAS ) );
		jobSpec = createJobSpec( "so=linux", numTasks, ECHO_COMMAND );
		assertEquals( "The number of GridMachines needed should be " + numTasks + " * " + maxReplicas + ".", numTasks
				* maxReplicas, wkq.getNumberOfNeededGums( jobSpec ) );

		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, "100000" );
		maxReplicas = Integer
				.parseInt( Configuration.getInstance().getProperty( MyGridConfiguration.PROP_MAX_REPLICAS ) );
		jobSpec = createJobSpec( "so=linux", numTasks, ECHO_COMMAND );
		assertEquals( "The number of GridMachines needed should be " + numTasks + " * " + maxReplicas + ".", numTasks
				* maxReplicas, wkq.getNumberOfNeededGums( jobSpec ) );
	}


	public void testScheduleOneJob() throws Exception {

		createOneMockGuMAndOneMockGuMP();

		int maxReplicas = 1;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.clonedJobList().size() == 0 );
		String expression = "os==linux";
		int numTasks = 5;

		JobSpec jobSpecs = createJobSpec( expression, numTasks, ECHO_COMMAND );
		wkq.addJob( 1, jobSpecs );

		final long REQID1 = 1L;
		JobRequestEntry jobRequestEntry = new JobRequestEntry( 1, REQID1 );
		jobRequestEntry.addPeer( mockGump );
		ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );
		ebGridManager.hereIsGum( mockGum, REQID1, mockGumpObjectID1 );

		for ( int taskId = 1; taskId <= numTasks; taskId++ ) {
			assertTrue( wkq.schedule() );
			ebGridManager.ready( mockGum.getGumSpec().getGumID() );
			ReplicaEntry requiredReplica = new ReplicaEntry( 1, taskId, 1, FakeTaskSpec.getFakeTaskSpec() );
			ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
			assertReplicaId( requiredReplica, lastScheduledReplica );
		}
		assertTrue( !wkq.schedule() );

		verifyMocks = true;
	}


	public void testScheduleThreeJobs() throws Exception {

		createOneMockGuMAndOneMockGuMP();

		int maxReplicas = 1;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.clonedJobList().size() == 0 );
		String[ ] expressions = { "os==linux", "location==lsd", "university==ufcg" };

		int[ ] numTasks = { 5, 10, 15 };

		JobSpec[ ] jobSpecs = new JobSpec[ expressions.length ];

		for ( int i = 0; i < jobSpecs.length; i++ )
			jobSpecs[i] = createJobSpec( expressions[i], numTasks[i], ECHO_COMMAND );

		for ( int i = 0; i < jobSpecs.length; i++ )
			wkq.addJob( i + 1, jobSpecs[i] );

		for ( int jobId = 1; jobId <= jobSpecs.length; jobId++ ) {

			JobRequestEntry jobRequestEntry = new JobRequestEntry( jobId, jobId );
			jobRequestEntry.addPeer( mockGump );

			ebGridManager.jobRequestsEntryMap.put( jobId, jobRequestEntry );
			ebGridManager.hereIsGum( mockGum, jobId, mockGumpObjectID1 );

			for ( int taskId = 1; taskId <= numTasks[jobId - 1]; taskId++ ) {
				assertTrue( wkq.schedule() );
				ebGridManager.ready( mockGum.getGumSpec().getGumID() );
				ReplicaEntry requiredReplica = new ReplicaEntry( 1, taskId, jobId, FakeTaskSpec.getFakeTaskSpec() );
				ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
				assertReplicaId( requiredReplica, lastScheduledReplica );
			}
			ebGridManager.deleteGum( mockGum.getGumSpec().getGumID() );
		}
		assertTrue( !wkq.schedule() );

		verifyMocks = true;
	}


	public void testScheduleOneJobWithReplication() throws Exception {

		createOneMockGuMAndOneMockGuMP();

		int maxReplicas = 3;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.clonedJobList().size() == 0 );
		String expression = "os==linux";
		int numTasks = 5;

		JobSpec jobSpecs = createJobSpec( expression, numTasks, ECHO_COMMAND );
		wkq.addJob( 1, jobSpecs );
		for ( int replicaId = 1; replicaId <= maxReplicas; replicaId++ ) {

			JobRequestEntry jobRequestEntry = new JobRequestEntry( 1, 1 );
			jobRequestEntry.addPeer( mockGump );
			ebGridManager.jobRequestsEntryMap.put( 1, jobRequestEntry );
			ebGridManager.hereIsGum( mockGum, 1, mockGumpObjectID1 );

			for ( int taskId = 1; taskId <= numTasks; taskId++ ) {
				assertTrue( wkq.schedule() );
				ebGridManager.ready( mockGum.getGumSpec().getGumID() );
				ReplicaEntry requiredReplica = new ReplicaEntry( replicaId, taskId, 1, FakeTaskSpec.getFakeTaskSpec() );
				ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
				assertReplicaId( requiredReplica, lastScheduledReplica );
			}

			ebGridManager.deleteGum( mockGum.getGumSpec().getGumID() );

		}
		assertTrue( !wkq.schedule() );

		verifyMocks = true;
	}


	public void testScheduleThreeJobsWithReplication() throws Exception {

		createOneMockGuMAndOneMockGuMP();

		int maxReplicas = 3;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.clonedJobList().size() == 0 );
		String[ ] expressions = { "os==linux", "location==lsd", "university==ufcg" };
		int[ ] numTasks = { 5, 5, 5 };
		int numberOfTasks = 0;
		for ( int i = 0; i < numTasks.length; i++ ) {
			numberOfTasks += numTasks[i];
		}

		JobSpec[ ] jobSpecs = new JobSpec[ expressions.length ];
		for ( int i = 0; i < jobSpecs.length; i++ ) {
			jobSpecs[i] = createJobSpec( expressions[i], numTasks[i], ECHO_COMMAND );
		}

		for ( int i = 0; i < jobSpecs.length; i++ ) {
			wkq.addJob( i + 1, jobSpecs[i] );
		}

		for ( int replicaId = 1; replicaId <= maxReplicas; replicaId++ ) {
			for ( int jobId = 1; jobId <= expressions.length; jobId++ ) {

				JobRequestEntry jobRequestEntry = new JobRequestEntry( jobId, jobId );
				jobRequestEntry.addPeer( mockGump );
				ebGridManager.jobRequestsEntryMap.put( jobId, jobRequestEntry );
				ebGridManager.hereIsGum( mockGum, jobId, mockGumpObjectID1 );

				for ( int taskId = 1; taskId <= numTasks[jobId - 1]; taskId++ ) {
					assertTrue( wkq.schedule() );
					ebGridManager.ready( mockGum.getGumSpec().getGumID() );
					ReplicaEntry requiredReplica = new ReplicaEntry( replicaId, taskId, jobId, FakeTaskSpec
							.getFakeTaskSpec() );
					ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
					assertReplicaId( requiredReplica, lastScheduledReplica );
				}
				ebGridManager.deleteGum( mockGum.getGumSpec().getGumID() );
			}
		}
		assertTrue( !wkq.schedule() );

		verifyMocks = true;
	}


	/**
	 * This test verifies if the heuristic scheduled the replicas in the correct
	 * order. These test simulates job adding during other jobs execution. The
	 * order of the scheduled replicas must be: ( 1.1, 1.2, 1.3, 1.4, 1.5, 1.1',
	 * 1.2', 1.3', 2.1, 2.2, 2.3, 2.4, 2.5, 1.4', 1.5', 2.1', 2.2', 2.3', 2.4',
	 * 2.5', 1.1'', 1.2'', 1.3'', 1.4'', 1.5'', 2.1'', 2.2'', 3.1, 3.2, 3.3,
	 * 3.4, 3.5, 3.1', 3.2', 3.3', 3.4', 3.5', 2.3'', 2.4'', 2.5'', 3.1'',
	 * 3.2'', 3.3'', 3.4'', 3.5'' )
	 * 
	 * @throws Exception
	 */
	public void testScheduleThreeJobsWithReplicationIntercalated() throws Exception {

		createOneMockGuMAndOneMockGuMP();

		int maxReplicas = 3;
		Configuration.getInstance().setProperty( MyGridConfiguration.PROP_MAX_REPLICAS, maxReplicas + "" );
		jobManager.setMaxReplicas( maxReplicas + "" );

		assertTrue( jobManager.clonedJobList().size() == 0 );
		String[ ] expressions = { "os==linux", "location==lsd", "university==ufcg" };
		int[ ] numTasks = { 5, 5, 5 };
		int numberOfTasks = 0;
		for ( int i = 0; i < numTasks.length; i++ ) {
			numberOfTasks += numTasks[i];
		}

		JobSpec[ ] jobSpecs = new JobSpec[ expressions.length ];
		for ( int i = 0; i < jobSpecs.length; i++ ) {
			jobSpecs[i] = createJobSpec( expressions[i], numTasks[i], ECHO_COMMAND );
		}

		ReplicaEntry[ ] replicasArray = new ReplicaEntry[ maxReplicas * numberOfTasks ];
		int[ ] numberOfTasksToScheduleArray = { 8, 5, 2, 5, 5, 2, 10, 3, 5 };
		int[ ] initialReplicaIdArray = { 1, 1, 2, 2, 3, 3, 1, 3, 3 };
		int[ ] initialTaskIdArray = { 1, 1, 4, 1, 1, 1, 1, 3, 1 };
		int[ ] jobIdArray = { 1, 2, 1, 2, 1, 2, 3, 2, 3 };
		boolean[ ] addJobArray = { true, true, false, false, false, false, true, false, false };

		int currentTasksScheduled = 0;
		for ( int x = 0; x < jobIdArray.length; x++ ) {

			if ( addJobArray[x] ) {

				wkq.addJob( jobIdArray[x], jobSpecs[jobIdArray[x] - 1] );
			}
			addReplicasToArray( replicasArray, maxReplicas, numberOfTasksToScheduleArray[x], currentTasksScheduled,
					initialReplicaIdArray[x], initialTaskIdArray[x], numTasks, jobIdArray[x] );
			for ( int replicaIndex = currentTasksScheduled; replicaIndex < (numberOfTasksToScheduleArray[x] + currentTasksScheduled); replicaIndex++ ) {
				JobRequestEntry jobRequestEntry = new JobRequestEntry( jobIdArray[x], jobIdArray[x] );
				jobRequestEntry.addPeer( mockGump );
				ebGridManager.jobRequestsEntryMap.put( jobIdArray[x], jobRequestEntry );
				ebGridManager.hereIsGum( mockGum, jobIdArray[x], mockGumpObjectID1 );

				assertTrue( wkq.schedule() );
				ReplicaEntry lastScheduledReplica = fakeEBReplicaExecutorFacade.getLastScheduledReplica();
				assertReplicaId( replicasArray[replicaIndex], lastScheduledReplica );

				ebGridManager.deleteGum( mockGum.getGumSpec().getGumID() );
			}
			currentTasksScheduled += numberOfTasksToScheduleArray[x];
		}

		assertTrue( !wkq.schedule() );

		verifyMocks = true;
	}


	private void assertReplicaId( ReplicaEntry replica1, ReplicaEntry replica2 ) {

		assertEquals( replica1.getJobId(), replica2.getJobId() );
		assertEquals( replica1.getTaskId(), replica2.getTaskId() );
		assertEquals( replica1.getId(), replica2.getId() );
	}


	private void addReplicasToArray( ReplicaEntry[ ] replicasArray, int maxReplicas, int numberOfTasksToSchedule,
										int currentTasksScheduled, int initialReplicaId, int initialTaskId,
										int[ ] jobNumTasks, int jobId ) {

		int replicaIndex = currentTasksScheduled;
		for ( int replicaId = initialReplicaId; replicaId <= maxReplicas; replicaId++ ) {
			for ( int taskId = initialTaskId; taskId <= jobNumTasks[jobId - 1]; taskId++ ) {
				replicasArray[replicaIndex] = new ReplicaEntry( replicaId, taskId, jobId, FakeTaskSpec
						.getFakeTaskSpec() );
				replicaIndex++;
				if ( replicaIndex == numberOfTasksToSchedule ) {
					return;
				}
			}
		}
	}


	/**
	 * Creates a job with <code>numTasks</code> identical tasks. Each task
	 * simply performs "command".
	 * 
	 * @param expression The job attribute expression
	 * @param numTasks The number of tasks in the job
	 * @param command A test command
	 * @return The created job
	 * @throws TaskSpecificationException When the <code>TaskSpec</code> is
	 *         not valid
	 * @throws JobSpecificationException
	 */
	private JobSpec createJobSpec( String expression, int numTasks, String command ) throws TaskSpecificationException,
		JobSpecificationException {

		List<TaskSpec> taskSpecList = new LinkedList<TaskSpec>();

		for ( int i = 0; i < numTasks; i++ ) {
			TaskSpec t = new TaskSpec( null, command, null );
			taskSpecList.add( t );
		}

		return new JobSpec( "TestJob", expression, taskSpecList );
	}


	private void createOneMockGuMAndOneMockGuMP() throws Exception {

		mockGum = EasyMock.createMock( GumClient.class );

		GumSpec gumSpec1 = new GumSpec();
		gumSpec1.putAttribute( GumSpec.ATT_NAME, "mockGum1" );
		gumSpec1.putAttribute( "os", "linux" );
		gumSpec1.putAttribute( "location", "lsd" );
		gumSpec1.putAttribute( "university", "ufcg" );

		mockGum.getGumSpec();
		EasyMock.expectLastCall().andReturn( gumSpec1 ).anyTimes();
		mockGum.getStatus();
		EasyMock.expectLastCall().andReturn( GumStatus.UP ).anyTimes();

		// Creating the mockGump
		mockGump = EasyMock.createNiceMock( Gump.class );

		boundableMockGumpWrapper = new BindableMockGumpWrapper( mockGump );
		mockGumpObjectID1 = new ObjectID( "MockGump1", (long) 42, "rmi://localhost:" + rmiPort + "/MockGump1" );
		Naming.rebind( mockGumpObjectID1.getObjectURL(), boundableMockGumpWrapper );

		EasyMock.replay( mockGump );
		EasyMock.replay( mockGum );
	}

}
