/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.test;

import java.util.LinkedList;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.event.Event;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.test.CommonTest;
import org.ourgrid.test.threadServices.ThreadServices;

/**
 * Class desired to test the EventQueue class.
 */
public class EventQueueTest extends TestCase implements CommonTest {

	/**
	 * The event queue.
	 */
	private EventQueue eq;

	/**
	 * A mock event.
	 */
	private Event mockEvent;


	/**
	 * The constructor.
	 * 
	 * @param name Test name
	 */
	public EventQueueTest( String name ) {

		super( name );
	}


	/**
	 * Initializes the <code>EventQueue</code> and creates a
	 * <code>mockEvent</code>.
	 */
	protected void setUp() throws Exception {

		super.setUp();
		eq = new EventQueue();
		mockEvent = EasyMock.createMock( Event.class );
	}


	/**
	 * Tests the put and unblockingRemove operations.
	 */
	public void testPutAndUnblockingRemove() {

		// Returning null
		Event eventNull = eq.unblockingRemove();
		assertNull( eventNull );

		// One event
		eq.put( mockEvent );
		Event eventNotNull = eq.unblockingRemove();
		assertTrue( mockEvent == eventNotNull );
		eventNull = eq.unblockingRemove();
		assertNull( eventNull );

		// Two equal events
		eq.put( mockEvent );
		eq.put( mockEvent );
		eventNotNull = eq.unblockingRemove();
		assertTrue( mockEvent == eventNotNull );
		eventNotNull = eq.unblockingRemove();
		assertTrue( mockEvent == eventNotNull );
		eventNull = eq.unblockingRemove();
		assertNull( eventNull );
	}


	/**
	 * Tests the put and unblockingRemove operations using many threads.
	 */
	public void testPutAndUnblockingRemoveByManyThreads() {

		// One hundred events added by threads
		PutterThread putterThread;
		for ( int x = 0; x < 100; x++ ) {
			putterThread = new PutterThread( eq, mockEvent );
			putterThread.start();
		}

		ThreadServices.waitUntilWorkIsDone();

		assertEquals( 100, eq.size() );

		Event eventNotNull;
		for ( int x = 0; x < 100; x++ ) {
			eventNotNull = eq.unblockingRemove();
			assertTrue( eventNotNull == mockEvent );
		}

		Event eventNull = eq.unblockingRemove();
		assertNull( eventNull );
	}


	/**
	 * Tests the put and blockingRemove operations using threads.
	 */
	public void testPutAndBlockingRemove() {

		assertTrue( eq.size() == 0 );

		PutterThread putterThread = new PutterThread( eq, mockEvent );
		putterThread.start();

		Event eventNotNull = eq.blockingRemove();
		assertTrue( mockEvent == eventNotNull );
		Event eventNull = eq.unblockingRemove();
		assertNull( eventNull );

		// Three consecutive events
		PutterThread putterThread1 = new PutterThread( eq, mockEvent );
		putterThread1.start();
		PutterThread putterThread2 = new PutterThread( eq, mockEvent );
		putterThread2.start();
		PutterThread putterThread3 = new PutterThread( eq, mockEvent );
		putterThread3.start();
		eventNotNull = eq.blockingRemove();
		assertTrue( mockEvent == eventNotNull );
		eventNotNull = eq.blockingRemove();
		assertTrue( mockEvent == eventNotNull );
		eventNotNull = eq.blockingRemove();
		assertTrue( mockEvent == eventNotNull );
		eventNull = eq.unblockingRemove();
		assertNull( eventNull );

		// Three intercalated events
		putterThread1 = new PutterThread( eq, mockEvent );
		putterThread1.start();
		eventNotNull = eq.blockingRemove();
		assertTrue( mockEvent == eventNotNull );
		putterThread2 = new PutterThread( eq, mockEvent );
		putterThread2.start();
		eventNotNull = eq.blockingRemove();
		assertTrue( mockEvent == eventNotNull );
		putterThread3 = new PutterThread( eq, mockEvent );
		putterThread3.start();
		eventNotNull = eq.blockingRemove();
		assertTrue( mockEvent == eventNotNull );
		eventNull = eq.unblockingRemove();
		assertNull( eventNull );

	}


	/**
	 * Tests the put and blockingRemove operations using many threads.
	 */
	public void testPutAndBlockingRemoveByManyThreads() {

		assertTrue( eq.size() == 0 );

		// Number of events to be added by threads
		int numberOfEvents = 100;

		RemoverThread removerThread = new RemoverThread( eq, numberOfEvents );
		removerThread.start();

		PutterThread putterThread;
		for ( int x = 0; x < numberOfEvents; x++ ) {
			putterThread = new PutterThread( eq, mockEvent );
			putterThread.start();
		}

		ThreadServices.waitUntilWorkIsDone();

		assertEquals( 0, eq.size() );
		LinkedList eventsList = removerThread.getEventsList();
		assertEquals( numberOfEvents, eventsList.size() );

	}

}


/**
 * Thread to queue events.
 */
class PutterThread extends Thread {

	private EventQueue myEq;

	private Event myEvent;


	public PutterThread( EventQueue eq, Event event ) {

		super( "PutterThread" );
		myEq = eq;
		myEvent = event;
	}


	public void run() {

		myEq.put( myEvent );
	}

}


/**
 * Thread to remove a defined number of events from an <code>EventQueue</code>
 * using the <code>blockingRemove()</code>
 */
class RemoverThread extends Thread {

	private EventQueue myEq;

	private LinkedList myEvents;

	int myObjectsNumber;


	public RemoverThread( EventQueue eq, int objectsNumber ) {

		super( "RemoverThread" );
		myEq = eq;
		myObjectsNumber = objectsNumber;
		myEvents = new LinkedList();
	}


	public void run() {

		for ( int x = 0; x < myObjectsNumber; x++ ) {
			myEvents.addLast( myEq.blockingRemove() );
		}
	}


	public LinkedList getEventsList() {

		return myEvents;
	}

}
