/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui;

import java.io.File;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.LogConfiguration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.common.fd.AbstractFailureDetector;
import org.ourgrid.common.fd.Notifiable;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.PeerSpec;
import org.ourgrid.common.spec.main.CompilerException;
import org.ourgrid.common.spec.main.DescriptionFileCompile;
import org.ourgrid.common.url.MyGridURLProvider;
import org.ourgrid.mygrid.fd.MyGridFailureDetector;
import org.ourgrid.mygrid.fd.PeerStateNotifiable;
import org.ourgrid.mygrid.replicaexecutor.EBReplicaExecutorFacade;
import org.ourgrid.mygrid.scheduler.DataDiscovery;
import org.ourgrid.mygrid.scheduler.EBSchedulerFacade;
import org.ourgrid.mygrid.scheduler.PeerEntry;
import org.ourgrid.mygrid.scheduler.Scheduler;
import org.ourgrid.mygrid.scheduler.exception.DDInitException;
import org.ourgrid.mygrid.scheduler.exception.JobCannotBeRemovedException;
import org.ourgrid.mygrid.scheduler.exception.JobCouldNotBeCancelledException;
import org.ourgrid.mygrid.scheduler.exception.JobNotFoundException;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerEntry;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerListenerRemote;
import org.ourgrid.mygrid.scheduler.gump.GumpManager;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManagerListenerRemote;
import org.ourgrid.mygrid.ui.command.UIMessages;
import org.ourgrid.mygrid.ui.exception.CouldNotAddJobOutsideHomeMachineException;
import org.ourgrid.mygrid.ui.exception.MyGridIsAlreadyRunningException;
import org.ourgrid.mygrid.ui.exception.NotExecutingInHomeMachine;
import org.ourgrid.mygrid.ui.exception.StartingMyGridException;

public class MyGridUIManager extends OurgridUnicastRemoteObject implements UIManager {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( MyGridUIManager.class );

	private static final long serialVersionUID = 33L;

	private static UIManager uniqueInstance;

	private GumpManager gumpManager;

	private Scheduler scheduler;

	private Notifiable peerStateNotifiable;

	private MyGridFailureDetector failureDetector;

	private EBSchedulerFacade ebSchedulerFacade;

	private EBReplicaExecutorFacade ebReplicaExecutorFacade;


	private MyGridUIManager() throws RemoteException {

		isMyGridRunning();

		try {
			failureDetector = new MyGridFailureDetector( Configuration.QUERY_FREQ );
		} catch ( RemoteException e ) {
			LOG.error( "Error trying to create Failure Detector", e );
		}

	}


	/**
	 * Resets the singleton.
	 */
	public static synchronized void reset() {

		uniqueInstance = null;
	}


	public static synchronized UIManager getInstance() throws RemoteException {

		if ( uniqueInstance == null ) {

			initRMIRegistry();

			try {
				// tries to get a remote instance already bound by another
				// invocation of this method.
				uniqueInstance = (UIManager) Naming.lookup( MyGridURLProvider.mygridUIManager() );
			} catch ( MalformedURLException e ) {
			} catch ( NotBoundException e ) {
			} catch ( RemoteException e ) {
			}

			if ( uniqueInstance == null ) {

				uniqueInstance = new MyGridUIManager();

				try {
					Naming.rebind( MyGridURLProvider.mygridUIManager(), uniqueInstance );
				} catch ( MalformedURLException e ) {
					LOG.debug( "Error trying to rebind at URL " + MyGridURLProvider.mygridUIManager() );
				}
			}
		}

		return uniqueInstance;

	}


	private static void initRMIRegistry() {

		int registryPort = Integer.parseInt( Configuration.getInstance( Configuration.MYGRID ).getProperty(
				Configuration.PROP_PORT ) );

		System.setProperty( "java.rmi.server.hostname", Configuration.getInstance().getProperty(
				Configuration.PROP_EXTERNAL_NAME ) );

		try {
			LocateRegistry.createRegistry( registryPort );
		} catch ( RemoteException e ) {
			LOG.debug( "Cannot start RMI registry at port: " + registryPort );
		}

	}


	public void startMyGridService() throws NotExecutingInHomeMachine, MyGridIsAlreadyRunningException,
		StartingMyGridException, ConfigException {

		/**
		 * Security options.
		 */
		if ( Configuration.getInstance().isSecurityEnabled() ) {
			String keyStoreFilename = Configuration.getInstance().getProperty( Configuration.PROP_KEYSTORE_FILENAME );
			String keyStorePassword = Configuration.getInstance().getProperty( Configuration.PROP_KEYSTORE_PASSWORD );

			System.setProperty( "javax.net.ssl.keyStore", keyStoreFilename );
			System.setProperty( "javax.net.ssl.keyStorePassword", keyStorePassword );

			String trustStoreFilename = Configuration.getInstance()
					.getProperty( Configuration.PROP_TRUSTSTORE_FILENAME );
			String trustStorePassword = Configuration.getInstance()
					.getProperty( Configuration.PROP_TRUSTSTORE_PASSWORD );

			System.setProperty( "javax.net.ssl.trustStore", trustStoreFilename );
			System.setProperty( "javax.net.ssl.trustStorePassword", trustStorePassword );
		}

		if ( !isExecutingInHomeMachine() ) {
			throw new NotExecutingInHomeMachine();
		}

		if ( isMyGridRunning() ) {
			throw new MyGridIsAlreadyRunningException();
		}

		String logfile = Configuration.getInstance().getLogPath();

		Configuration.getInstance().getProperty( MyGridConfiguration.PROP_HEURISTIC );
		boolean ddFile = Configuration.getInstance().isEnabled( MyGridConfiguration.PROP_CACHE_DDFILE );
		try {
			DataDiscovery.initDataDiscovery( ddFile );
		} catch ( DDInitException e ) {
			LOG.error( "Error trying to init Data Discovery", e );
		}

		String parents = new File( logfile ).getParent();

		if ( parents != null ) {
			new File( parents ).mkdirs();
		}

		/** **** Configuring Log4j ******* */
		String logPropertiesfile = Configuration.getInstance().getLogPropertiesPath();
		LogConfiguration.configure( logPropertiesfile, logfile );

		try {
			// Creating facades
			this.ebSchedulerFacade = new EBSchedulerFacade();
			this.ebReplicaExecutorFacade = new EBReplicaExecutorFacade();

			// Configuring facades
			ebSchedulerFacade.config( ebReplicaExecutorFacade );
			ebReplicaExecutorFacade.config( ebSchedulerFacade );

			// Starting the facades
			ebSchedulerFacade.startEventProcessor();
			ebReplicaExecutorFacade.startEventProcessor();

		} catch ( ConfigException e ) {
			LOG.error( "Error trying to read properties from the MG.properties file", e );
		} catch ( RemoteException e ) {
			LOG.error( "Error trying to create remote objects", e );
		}

		if ( isMyGridRunning() ) {
			startFailureDetector();
			LOG.info( UIMessages.MYGRID_IS_UP_AND_RUNNING );
		} else {
			throw new StartingMyGridException();
		}

		loadDefaultGdf();

	}


	private void loadDefaultGdf() {

		String defaultGdf = MyGridConfiguration.getInstance().getProperty( MyGridConfiguration.PROP_DEFAULT_GDF );
		if ( defaultGdf != null ) {

			File defaultSDF = new File( defaultGdf );

			if ( !defaultSDF.exists() ) {
				LOG.info( "OurGrid Peer default SDF not found: " + defaultGdf );
				return;
			}

			try {
				List<PeerSpec> peerSpecs = DescriptionFileCompile.compileGDF( defaultGdf );
				this.setPeers( peerSpecs );
			} catch ( CompilerException e ) {
				LOG.info( "MyGrid could not compile default GDF!" );
			} catch ( RemoteException e ) {
				LOG.info( "MyGrid could not contact UIManager to load default GDF!" );
			}

		}

	}


	private void startFailureDetector() {

		peerStateNotifiable = new PeerStateNotifiable( gumpManager );
		try {

			failureDetector.startMonitoring();
			Naming.rebind( MyGridURLProvider.failureDetector(), failureDetector );

		} catch ( RemoteException e ) {
			LOG.error( "Cannot bind Peer Failure Detector Object.", e );
		} catch ( MalformedURLException e ) {
			LOG.error( "Cannot bind Peer Failure Detector Object.", e );
		}
	}


	public void stopMyGridService( boolean wait, boolean callExit ) throws MyGridIsNotRunningException {

		if ( isMyGridRunning() ) {
			shutdown( wait );
		} else {
			throw new MyGridIsNotRunningException();
		}

		if ( callExit ) {
			callExit();
		}
	}


	public void stopMyGridService( boolean callExit ) throws MyGridIsNotRunningException {

		stopMyGridService( true, callExit );
	}


	private void callExit() {

		System.exit( 0 );
	}


	private void shutdown( boolean wait ) {

		if ( this.ebSchedulerFacade != null ) {
			EventQueue<ShutdownResponseEvent> schEq = this.ebSchedulerFacade.shutdown();
			if ( wait ) {
				schEq.blockingRemove();
			}
		}

		if ( this.ebReplicaExecutorFacade != null ) {
			EventQueue<ShutdownResponseEvent> reEq = this.ebReplicaExecutorFacade.shutdown();
			if ( wait ) {
				reEq.blockingRemove();
			}
		}
	}


	public boolean isMyGridRunning() {

		try {
			findExportedObjects();
			return true;
		} catch ( Exception e ) {
			return false;
		}
	}


	private void findExportedObjects() throws MalformedURLException, RemoteException, NotBoundException {

		scheduler = (Scheduler) Naming.lookup( MyGridURLProvider.scheduler() );

		gumpManager = (GumpManager) Naming.lookup( MyGridURLProvider.gumpManager() );
	}


	public boolean isExecutingInHomeMachine() {

		String homeMachine = Configuration.getInstance().getProperty( Configuration.PROP_NAME );

		try {
			String localHost = InetAddress.getLocalHost().getHostName();
			String localHostCompleteName = InetAddress.getLocalHost().getCanonicalHostName();
			String localIp = InetAddress.getLocalHost().getHostAddress();

			if ( homeMachine == null ) {
				throw new UnknownHostException( "Cannot determine home machine name!" );
			}

			return localHost.equals( homeMachine ) || localIp.equals( homeMachine )
					|| localHostCompleteName.equals( homeMachine );

		} catch ( Exception e ) {
			return false;
		}
	}


	public Collection<JobEntry> getJobs( JobManagerListenerRemote newListener ) throws RemoteException {

		if ( !isMyGridRunning() && scheduler == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		return scheduler.jobList( newListener );
	}


	public void cancelJob( int jobId ) throws JobNotFoundException, RemoteException, JobCouldNotBeCancelledException {

		if ( !isMyGridRunning() && scheduler == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		scheduler.cancelJob( jobId );
	}


	public void waitForJob( int jobId ) throws RemoteException, JobNotFoundException {

		if ( !isMyGridRunning() && scheduler == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		scheduler.waitForJob( jobId );

	}


	public int addJob( JobSpec newJob ) throws RemoteException {

		if ( !isMyGridRunning() && scheduler == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		int jobId = scheduler.addJob( newJob );
		return jobId;
	}


	public void cleanFinishedJobs() throws RemoteException {

		if ( !isMyGridRunning() && scheduler == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		scheduler.cleanFinishedJobs();

	}


	public void removeJob( int jobId ) throws RemoteException, JobCannotBeRemovedException {

		if ( !isMyGridRunning() && scheduler == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		scheduler.removeJob( jobId );
	}


	public Collection<PeerEntry> getPeerEntries( GridManagerListenerRemote newListener ) throws RemoteException {

		if ( !isMyGridRunning() && gumpManager == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		return gumpManager.getPeerEntries( newListener );
	}


	public void setPeers( Collection<PeerSpec> peerSpecs ) throws RemoteException {

		if ( !isMyGridRunning() && gumpManager == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		gumpManager.setPeers( peerSpecs );

		updateFailureDetector( peerSpecs );

	}


	private void updateFailureDetector( Collection<PeerSpec> guMPSpecs ) throws RemoteException {

		/* Removing old registereds Peers in this Failure Detector */
		failureDetector.unregister( peerStateNotifiable );

		/* Registering new Peers in this Failure Detector */
		for ( Iterator iter = guMPSpecs.iterator(); iter.hasNext(); ) {
			PeerSpec peerSpec = (PeerSpec) iter.next();

			String rmiHost = "rmi://" + peerSpec.getName() + ":" + peerSpec.getPort() + "/"
					+ AbstractFailureDetector.BOUND_NAME;

			try {
				failureDetector.register( peerStateNotifiable, rmiHost, true );
			} catch ( MalformedURLException e ) {
				LOG.error( "Cannot register in MyGrid failure Detector the Peer: " + peerSpec.getURL(), e );
			}

		}

	}


	public Collection<GumSpec> getGumSpecs() throws RemoteException {

		if ( !isMyGridRunning() && scheduler == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		return scheduler.getGumSpecs();
	}


	public Collection<GridManagerEntry> getGumEntries( GridManagerListenerRemote listener ) throws RemoteException {

		if ( !isMyGridRunning() && scheduler == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		return scheduler.getGumEntries( listener );
	}


	@Deprecated
	public void setPeers( String descriptionFilePath ) throws RemoteException, CompilerException {

		List<PeerSpec> result = DescriptionFileCompile.compileGDF( descriptionFilePath );
		this.setPeers( result );
	}


	@Deprecated
	public int addJob( String descriptionFilePath ) throws CompilerException, RemoteException,
		CouldNotAddJobOutsideHomeMachineException {

		if ( !isExecutingInHomeMachine() ) {
			throw new CouldNotAddJobOutsideHomeMachineException();
		}

		JobSpec result = DescriptionFileCompile.compileJDF( descriptionFilePath );
		return this.addJob( result );
	}


	public void removeJobManagerListener( JobManagerListenerRemote listener ) throws RemoteException {

		if ( !isMyGridRunning() && scheduler == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		scheduler.removeJobManagerListener( listener );
	}


	public void removeGridManagerListener( GridManagerListenerRemote listener ) throws RemoteException {

		if ( !isMyGridRunning() && scheduler == null ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}

		scheduler.removeGridManagerListener( listener );
	}
}
