/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.Collection;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.PeerSpec;
import org.ourgrid.common.spec.main.CompilerException;
import org.ourgrid.common.spec.main.DescriptionFileCompile;
import org.ourgrid.mygrid.scheduler.PeerEntry;
import org.ourgrid.mygrid.scheduler.exception.JobCannotBeRemovedException;
import org.ourgrid.mygrid.scheduler.exception.JobCouldNotBeCancelledException;
import org.ourgrid.mygrid.scheduler.exception.JobNotFoundException;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerEntry;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerListenerRemote;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.JobManagerListenerRemote;
import org.ourgrid.mygrid.ui.exception.CouldNotAddJobOutsideHomeMachineException;
import org.ourgrid.mygrid.ui.exception.MyGridIsAlreadyRunningException;
import org.ourgrid.mygrid.ui.exception.NotExecutingInHomeMachine;
import org.ourgrid.mygrid.ui.exception.StartingMyGridException;

/**
 * Interface of Mygrid API.
 * <p>
 * To use the API, applications must add ourgrid.jar to their classpath. The
 * application command line must include a definition to the MGROOT variable
 * (i.e. add -DMGROOT=/mygrid/installation/directory to the Java VM command
 * line).
 * <p>
 * Before calling methods, an application must obtain a reference to Mygrid's UI
 * Manager by calling {@link MyGridUIManager#getInstance()}
 * <p>
 */
public interface UIManager extends Remote {

	/**
	 * Adds a job to be executed by Mygrid. The <code>JobSpec</code> obtained
	 * by compiling a JDF (job description file) with Ourgrid compiler.
	 * 
	 * @param theJob a job specification generated by the Ourgrid compiler from
	 *        a JDF
	 * @return the job ID
	 * @throws RemoteException if a communication problem occurs
	 * @throws CouldNotAddJobOutsideHomeMachineException it is only possible to
	 *         add a job when running on the same machine as Mygrid
	 */
	public int addJob( JobSpec theJob ) throws RemoteException, CouldNotAddJobOutsideHomeMachineException;


	/**
	 * Replaced by {@link UIManager#addJob(JobSpec)}.
	 * 
	 * @param descriptionFilePath the path of a JDF to be compiled
	 * @return the generated job ID
	 * @throws RemoteException if a communication problem occurs
	 * @throws CompilerException if compiler error occurs
	 * @throws CouldNotAddJobOutsideHomeMachineException it is only possible to
	 *         add a job when running on the same machine as Mygrid
	 */
	@Deprecated
	public int addJob( String descriptionFilePath ) throws RemoteException, CompilerException,
		CouldNotAddJobOutsideHomeMachineException;


	/**
	 * Asks Mygrid to cancel a job
	 * 
	 * @param jobId the job to cancel
	 * @throws JobNotFoundException the job does not exist
	 * @throws RemoteException if a communication problem occurs
	 * @throws JobCouldNotBeCancelledException if some error occurs when trying
	 *         to cancel the job
	 */
	public void cancelJob( int jobId ) throws JobNotFoundException, RemoteException, JobCouldNotBeCancelledException;


	/**
	 * Erases all jobs that are not running anymore
	 * 
	 * @throws RemoteException if a communication problem occurs
	 */
	public void cleanFinishedJobs() throws RemoteException;


	/**
	 * Obtains a collection with the gums that are currently in use by Mygrid.
	 * Includes an option of becoming a listener of Mygrid's Grid Manager to
	 * receive updates about the state of machines. This listener will be
	 * notified whenever a new machine is received or currently owned machines
	 * are disposed.
	 * 
	 * @param newListener a reference to the listener that is interested on
	 *        receiving updates about machines
	 * @return a collection with the gums that are currently in use by Mygrid
	 * @throws RemoteException if a communication problem occurs
	 */
	public Collection<GridManagerEntry> getGumEntries( GridManagerListenerRemote newListener ) throws RemoteException;


	/**
	 * Obtains a collection with the gums that are currently in use by Mygrid
	 * 
	 * @return a collection with the gums that are currently in use by Mygrid
	 * @throws RemoteException if a communication problem occurs
	 */
	public Collection<GumSpec> getGumSpecs() throws RemoteException;


	/**
	 * Obtains a collection with the jobs that are currently added to Mygrid, no
	 * matter the jobs states. Includes an option of becoming a listener of
	 * Mygrid's Job Manager to receive updates about the state of jobs. This
	 * listener will be notified whenever there is a change on the state of
	 * jobs.
	 * 
	 * @param newListener a reference to the listener that is interested on
	 *        receiving updates about jobs
	 * @return a collection with the jobs currently added to Mygrid
	 * @throws RemoteException if a communication problem occurs
	 */
	public Collection<JobEntry> getJobs( JobManagerListenerRemote newListener ) throws RemoteException;


	/**
	 * Obtains a collection with the peers that are currently set on Mygrid.
	 * Includes an option of becoming a listener of Mygrid's Grid Manager to
	 * receive updates about the state of peers. This listener will be notified
	 * whenever there is a change on the state of peers.
	 * 
	 * @param newListener a reference to the listener that is interested on
	 *        receiving updates about peers
	 * @return a collection with the peers currently set on Mygrid
	 * @throws RemoteException if a communication problem occurs
	 */
	public Collection<PeerEntry> getPeerEntries( GridManagerListenerRemote newListener ) throws RemoteException;


	/**
	 * Checks if Mygrid is running
	 * 
	 * @return true if Mygrid is running
	 * @throws RemoteException if a communication problem occurs
	 */
	public boolean isMyGridRunning() throws RemoteException;


	/**
	 * Unregisters a Grid Manager listener
	 * 
	 * @param listener listener a reference to the listener that will be
	 *        unregistered
	 * @throws RemoteException if a communication problem occurs
	 */
	public void removeGridManagerListener( GridManagerListenerRemote listener ) throws RemoteException;


	/**
	 * Erases a specific job if it is not running.
	 * 
	 * @param jobId id of the job to be removed
	 * @throws RemoteException if a communication problem occurs
	 * @throws JobCannotBeRemovedException if the jobs is still running
	 */
	public void removeJob( int jobId ) throws RemoteException, JobCannotBeRemovedException;


	/**
	 * Unregisters a Job Manager listener
	 * 
	 * @param listener a reference to the listener that will be unregistered
	 * @throws RemoteException if a communication problem occurs
	 */
	public void removeJobManagerListener( JobManagerListenerRemote listener ) throws RemoteException;


	/**
	 * Adds a list of peers that will work as a machine provider to Mygrid. A
	 * list of <code>PeerSpec</code> is obtained by compiling a SDF (site
	 * description file) with Ourgrid compiler.
	 * 
	 * @see DescriptionFileCompile#compileSDF(String)
	 * @param peerSpecs a collection of peer specification generated by the
	 *        compiler
	 * @throws RemoteException if a communication problem occurs
	 */
	public void setPeers( Collection<PeerSpec> peerSpecs ) throws RemoteException;


	/**
	 * Replaced by {@link UIManager#setPeers(Collection)}.
	 * 
	 * @param descriptionFilePath the path to a SDF to be compiled
	 * @throws RemoteException if a communication problem occurs
	 * @throws CompilerException if compiler error occurs
	 */
	@Deprecated
	public void setPeers( String descriptionFilePath ) throws RemoteException, CompilerException;


	/**
	 * Start MyGrid service. The service will be initiated on the same JVM of
	 * the calling application
	 * 
	 * @throws RemoteException if a communication problem occurs
	 * @throws NotExecutingInHomeMachine if Mygrid is starting on a different
	 *         machine than that specified in the Mygrid configuration file
	 * @throws MyGridIsAlreadyRunningException if Mygrid is already running
	 * @throws StartingMyGridException if Mygrid was unable to start correctly
	 * @throws ConfigException if a configuration problem exists
	 */
	public void startMyGridService() throws RemoteException, NotExecutingInHomeMachine,
		MyGridIsAlreadyRunningException, StartingMyGridException, ConfigException;


	/**
	 * Stop MyGrid service. This method does waits for MyGrid to shutdown
	 * nicely.
	 * 
	 * @param callExit if is desired to call System.exit() terminating the JVM
	 *        (which may be the same of a user's application)
	 * @throws RemoteException if a communication problem occurs
	 * @throws MyGridIsNotRunningException if trying to stop a Mygrid that was
	 *         not started
	 */
	public void stopMyGridService( boolean callExit ) throws RemoteException, MyGridIsNotRunningException;


	/**
	 * Stop MyGrid service. This method presents an option of waiting Mygrid to
	 * stop nicely or forcing Mygrid to stop.
	 * 
	 * @param wait if is desired to wait for Mygrid to stop nicely.
	 * @param callExit if is desired to call System.exit() terminating the JVM
	 *        (which may be the same of a user's application)
	 * @throws RemoteException if a communication problem occurs
	 * @throws MyGridIsNotRunningException if trying contact a Mygrid that is
	 *         not running
	 */
	public void stopMyGridService( boolean wait, boolean callExit ) throws RemoteException, MyGridIsNotRunningException;


	/**
	 * The caller will block until the specified job finishes, fails or is
	 * cancelled.
	 * 
	 * @param jobId the job to wait
	 * @throws RemoteException if a communication problem occurs
	 * @throws JobNotFoundException if the job does not exist
	 */
	public void waitForJob( int jobId ) throws RemoteException, JobNotFoundException;
}
