/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.mygrid.ui.command.UIMessages;

/**
 * This class provides a wizard tool to create or import an 'mg.properties'
 * file.
 */
public class Wizard {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( Wizard.class );

	/**
	 * Reader used when importing MG.properties.
	 */
	private BufferedReader in;


	/**
	 * Main constructor. It doesn't start the wizard.
	 */
	public Wizard() {

		in = new BufferedReader( new InputStreamReader( System.in ) );

	}


	/**
	 * Effectively starts the wizard, asking whether the user wants to import or
	 * create a new MG.properties file.
	 */
	public void start() {

		System.out.println( UIMessages.MG_PROPERTIES_NOT_FOUND );
		System.out.println( UIMessages.CREATEING_NEW_MG_PROP );
		createMGProperties();
		copyMGLogProperties();
	}


	/**
	 * Copy mygrid.log.properties from Mygrid Aplication root to
	 * user.home/.mygrid dir
	 */
	private void copyMGLogProperties() {

		String mgroot = Configuration.getInstance().getProperty( Configuration.MGROOT );
		String source = mgroot + File.separator
				+ Configuration.getInstance().getProperty( MyGridConfiguration.PROP_LOG_PROPERTIES_FILE );
		String dest = MyGridConfiguration.LOG_PROPERTIES_FILENAME_DESTINATION;

		InputStream in = null;
		OutputStream out = null;
		try {
			in = new FileInputStream( source );
			out = new FileOutputStream( dest );

			// Transfer bytes from in to out
			byte[ ] buf = new byte[ 1024 ];
			int len;
			while ( (len = in.read( buf )) > 0 ) {
				out.write( buf, 0, len );
			}
		} catch ( Exception e ) {
			System.out.println( UIMessages.ERROR_COPYING_LOG_PROP );
			// e.printStackTrace();
		} finally {
			try {
				if ( in != null )
					in.close();
				if ( out != null )
					out.close();
			} catch ( Exception e ) {
				// no do
			}
		}

	}


	/**
	 * Creates a new MG.properties, based on user's answers about some
	 * questions. For each question, there is a default answer that will be used
	 * if the user doesn't supply an answer.
	 */
	private void createMGProperties() {

		BufferedWriter out = null;

		try {

			Configuration config = Configuration.getInstance( Configuration.MYGRID );

			String[ ][ ] MGsettings = getDefaultMGProperties( config );

			/* Defines questions of the quiz. */
			String[ ] questions = {
									"Home machine name: ",
									"RMI port number: ",
									"Scheduling heuristic (workqueue/storageaffinity): ",
									"How many replicas of a task should Mygrid try to run simultaneously? ",
									"How many times should Mygrid try to execute a given task before considering it has failed? ",
									"How many times may a machine fail within a job before it is blacklisted? ",
									"Save Data Discovery history (for use with Storage Affinity only)? "

			};

			/* Asks pre-defined questions and gets user's answers. */
			for ( int i = 0; i < questions.length; i++ ) {
				System.out.print( i + 1 + "/" + questions.length + ": " + questions[i] + "[" + MGsettings[i][1] + "] " );
				String response = in.readLine();
				if ( response != null && !response.equals( "" ) ) {
					MGsettings[i][1] = response;
				}
			}

			/*
			 * Effectively creates the new MG.properties file, using user's
			 * answers.
			 */
			File MGproperties = new File( MyGridConfiguration.PROPERTIES_FILENAME );
			out = new BufferedWriter( new FileWriter( MGproperties ) );
			for ( int i = 0; i < questions.length; i++ ) {
				out.write( MGsettings[i][0] + " = " + MGsettings[i][1] );
				out.write( System.getProperty( "line.separator" ) );
			}

		} catch ( IOException e ) {

			System.out.println( "Wizard failed trying to read from standard input." );
			LOG.error( "Wizard failed trying to read from standard input." );
		} finally {

			if ( out != null ) {
				try {
					out.close();
				} catch ( IOException e1 ) {
					// If this throws a exception nothing can be done.
				}
			}
		}
	}


	/**
	 * @param config
	 * @return
	 */
	private String[ ][ ] getDefaultMGProperties( Configuration config ) {

		/* Defines MG.properties' default contents. */
		String[ ][ ] MGsettings = new String[ 9 ][ 2 ];

		MGsettings[0][0] = MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_NAME;
		MGsettings[0][1] = config.getProperty( MyGridConfiguration.PROP_NAME );

		MGsettings[1][0] = MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_PORT;
		MGsettings[1][1] = config.getProperty( MyGridConfiguration.PROP_PORT );

		MGsettings[2][0] = MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_HEURISTIC;
		MGsettings[2][1] = config.getProperty( MyGridConfiguration.PROP_HEURISTIC );

		MGsettings[3][0] = MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_MAX_REPLICAS;
		MGsettings[3][1] = config.getProperty( MyGridConfiguration.PROP_MAX_REPLICAS );

		MGsettings[4][0] = MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_MAX_FAILS;
		MGsettings[4][1] = config.getProperty( MyGridConfiguration.PROP_MAX_FAILS );

		MGsettings[5][0] = MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_MAX_BL_FAILS;
		MGsettings[5][1] = config.getProperty( MyGridConfiguration.PROP_MAX_BL_FAILS );

		/* Added by Hunter Hagewood 22/03/2005 */
		MGsettings[6][0] = MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_CACHE_DDFILE;
		MGsettings[6][1] = config.getProperty( MyGridConfiguration.PROP_CACHE_DDFILE );

		return MGsettings;
	}

}
