/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui.command;

import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.LogConfiguration;
import org.ourgrid.mygrid.ui.MyGridUIManager;
import org.ourgrid.mygrid.ui.UIManager;

/**
 * Implements the main of Command Pattern. All MyGrid commands must pass here.
 */
public class CommandManager {

	/**
	 * Option to indicate that the command must not initialize the RMI registry
	 */
	private static final Object NO_REGISTRY_OPTION = "-reg";

	private Map<String,Command> mygridCommands;

	private UIManager UImanager = null;

	public static final String MYGRID_STARTER = "start";

	public static final String MYGRID_STOPPER = "stop";

	public static final String MYGRID_STATUS = "status";

	public static final String MYGRID_SETGRID = "setgrid";

	public static final String MYGRID_ADDJOB = "addjob";

	public static final String MYGRID_CANCELJOB = "canceljob";

	public static final String MYGRID_WAITFORJOB = "waitforjob";

	public static final String MYGRID_CONFIG = "config";

	public static final String MYGRID_GUI = "gui";

	public static final String MYGRID_CLEANER = "clean";


	/**
	 * Default constructor
	 * 
	 * @param startUIManager if the command will use a component's
	 *        <code>UIManager</code> the value should be true. This value
	 *        should be set to false only if the command does not need to
	 *        communicate with the component's daemon.
	 * @throws RemoteException if a communication error occurs when getting the
	 *         reference to the <code>UIManager</code>
	 */
	public CommandManager( boolean startUIManager ) throws RemoteException {

		if ( startUIManager ) {
			this.UImanager = MyGridUIManager.getInstance();
		}

		initCommandMap();
	}


	public CommandManager( UIManager UIManager ) {

		this.UImanager = UIManager;

		initCommandMap();
	}


	/**
	 * Init Log Configuration
	 */
	private static void initLog() {

		/** Configuring log * */
		Configuration.getInstance( Configuration.MYGRID );
		String logfile = Configuration.getInstance().getLogPath();
		String logPropertiesfile = Configuration.getInstance().getLogPropertiesPath();
		LogConfiguration.configure( logPropertiesfile, logfile );

	}


	/**
	 * it starts the map with MyGridCommands.
	 */
	public void initCommandMap() {

		mygridCommands = new HashMap<String,Command>();

		addNewCommand( MYGRID_STOPPER, new Stopper( this.UImanager ) );
		addNewCommand( MYGRID_STARTER, new Starter( this.UImanager ) );
		addNewCommand( MYGRID_STATUS, new Status( this.UImanager ) );
		addNewCommand( MYGRID_SETGRID, new GridSetter( this.UImanager ) );
		addNewCommand( MYGRID_ADDJOB, new JobAdder( this.UImanager ) );
		addNewCommand( MYGRID_CANCELJOB, new JobCanceller( this.UImanager ) );
		addNewCommand( MYGRID_WAITFORJOB, new JobExpecter( this.UImanager ) );
		addNewCommand( MYGRID_CONFIG, new EnvironmentConfigurator() );
		addNewCommand( MYGRID_GUI, new GUIStarter() );
		addNewCommand( MYGRID_CLEANER, new JobCleaner( this.UImanager ) );
	}


	public void addNewCommand( String command, Command gridCommand ) {

		mygridCommands.put( command, gridCommand );
	}


	/**
	 * Method that checks what <code>Command</code> will be able to execute
	 * the command in parameter "param". It will verify in it's map of commands,
	 * and then, call his run method.
	 * 
	 * @param params This parameter must contain the Command string
	 *        representation, followed by it's parameters.
	 * @return true if the JVM must be kept alive after execution
	 * @throws InvalidCommandException
	 */
	public int executeCommand( String[ ] params ) throws InvalidCommandException {

		Command gridCommand = mygridCommands.get( params[0] );

		if ( gridCommand == null ) {
			throw new InvalidCommandException( params[0] );
		}

		String[ ] parameters = discardFirstParameter( params );

		if ( gridCommand.configureEnvironment() ) {
			checkConfiguration();
		}

		return gridCommand.run( parameters );
	}


	/**
	 * Method that invoke a command checker <code>EnvironmentConfigurator</code>
	 * to validate the environment
	 */
	private void checkConfiguration() {

		Command gridCommand = mygridCommands.get( MYGRID_CONFIG );
		gridCommand.run( new String[ ] {} );
	}


	/**
	 * Discard the first parameter from this String[]. This parameter represents
	 * the Command from line command.
	 * 
	 * @param params An String array with all parameter that came from line
	 *        command, including the specified MyGrid Command.
	 * @return All other parameters except the MyGrid Command
	 */
	private static String[ ] discardFirstParameter( String[ ] params ) {

		String[ ] parameters = new String[ params.length - 1 ];

		for ( int i = 1; i < params.length; i++ ) {
			parameters[i - 1] = params[i];
		}

		return parameters;
	}


	public static void main( String[ ] args ) {

		String[ ] params;
		boolean startUIManager;

		initLog();
		try {
			if ( args[0].trim().equals( NO_REGISTRY_OPTION ) ) {
				params = discardFirstParameter( args );
				startUIManager = false;
			} else {
				params = args;
				startUIManager = true;
			}

			CommandManager commandManager = new CommandManager( startUIManager );
			int errorCode = commandManager.executeCommand( params );

			if ( errorCode != ErrorCode.START_CODE ) {
				System.exit( errorCode );

			}

		} catch ( InvalidCommandException e ) {
			System.exit( ErrorCode.FAIL );
		} catch ( Exception e1 ) {
			System.out.println( "Could not execute MyGrid Command " + args[0] + " " + e1.getMessage() );
			System.exit( ErrorCode.FAIL );
		}
	}
}
