/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui.command;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.mygrid.scheduler.PeerEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.TaskEntry;
import org.ourgrid.mygrid.ui.MyGridUIManager;
import org.ourgrid.mygrid.ui.UIManager;

/**
 * Description: This class is responsible to print mygrid current job and gums
 * state. This can be made by getting an instance from <i>Status </i> and using
 * its API or through <i>mygrid status </i> script.
 * 
 * @version 1.0 Created on 03/08/2004
 */
public class Status implements Command {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( Status.class );

	private Map<String,String> arguments;

	private UIManager manager;


	public Status( UIManager manager ) {

		this.manager = manager;
	}


	/**
	 * Main method. This method should be called with no command line arguments
	 */
	public int run( String[ ] params ) {

		Configuration.getInstance( Configuration.MYGRID );

		try {
			printGridStatus( params );
			return ErrorCode.OK;
		} catch ( RemoteException re ) {
			LOG.error( "Could not print the grid status: " + re.getMessage() );
			System.err.println( "Could not print the grid status: " + re.getMessage() );
		}

		return ErrorCode.COULD_NOT_PRINT_GRID_STATUS;
	}


	/**
	 * Print's this class command line usage.
	 */
	public static void printUsage() {

		System.err.println( "Usage: java -classpath YOUR_CLASS_PATH Status" );

	}


	/**
	 * Prints MyGrid status
	 * 
	 * @throws RemoteException If, for some reason, MyGrid could not be
	 *         contacted.
	 */
	public void printGridStatus( String args[] ) throws RemoteException {

		LOG.info( "Starting to print grid status." );
		System.out.println( "Starting to print grid status." );

		parseArgs( args );

		if ( !MyGridUIManager.getInstance().isMyGridRunning() ) {
			throw new RemoteException( UIMessages.MYGRID_IS_NOT_RUNNING );
		}
		System.out.println( UIMessages.MYGRID_IS_UP_AND_RUNNING );

		printPeers();
		printReceivedGums();
		printJobs();
	}


	/**
	 * Prints the collection of gums from the GuMManager.
	 * 
	 * @throws RemoteException If MyGrid is not up and running.
	 */
	private void printReceivedGums() throws RemoteException {

		Collection<GumSpec> gumsCollection = manager.getGumSpecs();

		if ( gumsCollection == null ) {
			System.err.println( "Grid Machine collection is invalid!" );
		} else {
			boolean any = false;

			System.out.print( "\n\nReceived Grid Machines:" );

			Iterator<GumSpec> itGums = gumsCollection.iterator();

			while ( itGums.hasNext() ) {
				any = true;
				printGumID( itGums.next() );
			}

			if ( !any ) {
				System.out.println( " none" );
			}

		}

	}


	/**
	 * Prints the collection of gumps
	 * 
	 * @throws RemoteException If MyGrid is not up and running.
	 */
	private void printPeers() throws RemoteException {

		Collection<PeerEntry> peersCollection = manager.getPeerEntries( null );

		if ( peersCollection == null ) {
			System.err.println( "Peers collection is invalid!" );
		} else {
			System.out.println( "\n\nPeers:" );
			for ( PeerEntry peer : peersCollection ) {
				printPeerSpec( peer );
			}
		}
	}


	/**
	 * Prints the list of jobs
	 * 
	 * @throws RemoteException If MyGrid is not up and running.
	 */
	private void printJobs() throws RemoteException {

		Collection<JobEntry> jobList = manager.getJobs( null );

		if ( jobList == null ) {

			System.err.println( "Job list is invalid!" );

		} else {

			System.out.print( "\n\nJobs:" );

			Iterator<JobEntry> itJobs = jobList.iterator();
			boolean any = false;
			while ( itJobs.hasNext() ) {
				JobEntry j = itJobs.next();
				if ( shouldBePrinted( j ) ) {
					any = true;
					printJob( j );
				}
			}

			if ( !any ) {
				System.out.println( " none" );
			} else {
				System.out.println();
			}
			System.out.println();
		}
	}


	/**
	 * Parses command line arguments.
	 * 
	 * @param args Arguments array
	 */
	private void parseArgs( String[ ] args ) {

		arguments = new HashMap<String,String>();

		for ( int i = 0; i < args.length; i++ ) {

			if ( args[i].startsWith( "-l=" ) ) {

				arguments.put( "LABEL", args[i].substring( 3 ) );

			} else if ( args[i].startsWith( "-s=" ) ) {

				arguments.put( "STATUS", args[i].substring( 3 ) );

			} else if ( args[i].startsWith( "-i=" ) ) {

				arguments.put( "JOBID", args[i].substring( 3 ) );

			}
		}
	}


	/**
	 * Verifies if a given should be printed, based in the command line
	 * arguments.
	 * 
	 * @param j The job to be tested.
	 * @return True if the job should be printed, false otherwise.
	 */
	private boolean shouldBePrinted( JobEntry j ) {

		if ( arguments.get( "JOBID" ) != null ) {
			int jobid;
			try {
				jobid = Integer.parseInt( arguments.get( "JOBID" ) );
			} catch ( NumberFormatException e ) {
				return false;
			}

			if ( j.getId() != jobid ) {
				return false;
			}

		}

		if ( arguments.get( "LABEL" ) != null ) {

			if ( !j.getLabel().equalsIgnoreCase( arguments.get( "LABEL" ) ) ) {
				return false;
			}
		}

		if ( arguments.get( "STATUS" ) != null ) {

			if ( !j.getStatus().toString().equalsIgnoreCase( arguments.get( "STATUS" ) ) ) {
				return false;
			}

		}
		return true;
	}


	/**
	 * Prints a job.
	 * 
	 * @param aJob The job to be printed.
	 */
	private void printJob( JobEntry aJob ) {

		System.out.println( "\n\t" + "Job " + aJob.getId() + ": " + aJob.getLabel() + " ["
				+ aJob.getStatus().toString() + "]" );

		Iterator<TaskEntry> itTasks = aJob.getTasks().iterator();
		while ( itTasks.hasNext() ) {

			printTask( itTasks.next() );
		}
	}


	/**
	 * Prints a task.
	 * 
	 * @param aTask The task to be printed.
	 */
	private void printTask( TaskEntry aTask ) {

		System.out.println( "\t\t" + "Task " + aTask.getId() + ": [" + aTask.getState() + "]" );

		Iterator<ReplicaEntry> itReplicas = aTask.getReplicas().iterator();
		while ( itReplicas.hasNext() ) {

			printReplica( itReplicas.next() );
		}
	}


	/**
	 * Prints a replica
	 * 
	 * @param aReplica The replica to be printed.
	 */
	private void printReplica( ReplicaEntry aReplica ) {

		System.out.println( "\t\t\t" + "Replica " + aReplica.getId() + ": [" + aReplica.getStatus().toString()
				+ "] - assigned to " + aReplica.getGumID() );

	}


	/**
	 * Prints a GuMSpec
	 * 
	 * @param guMSpec The GuMSpec to be printed.
	 */
	private void printGumID( GumSpec guMSpec ) {

		System.out.print( "\n\t" + guMSpec.getGumID() );

	}


	/**
	 * Prints a GuMPSpec
	 * 
	 * @param peerEntry The GuMPSpec to be printed.
	 */
	private void printPeerSpec( PeerEntry peerEntry ) {

		System.out.println( "\n" + peerEntry.getPeerSpec().getLabel() + " => " + peerEntry.getPeerSpec().getURL()
				+ " [ " + peerEntry.getStatus() + " ]" );

	}


	/**
	 * @see org.ourgrid.mygrid.ui.command.Command#configureEnvironment()
	 */
	public boolean configureEnvironment() {

		return true;
	}
}
