/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui.gui;

import java.rmi.RemoteException;
import java.util.Collection;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.PeerSpec;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.PeerEntry;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;
import org.ourgrid.mygrid.ui.UIManager;
import org.ourgrid.mygrid.ui.gui.event.GUIAllocateNotifyEvent;
import org.ourgrid.mygrid.ui.gui.event.GUIDeleteGumNotifyEvent;
import org.ourgrid.mygrid.ui.gui.event.GUIHereIsGumNotifyEvent;
import org.ourgrid.mygrid.ui.gui.event.GUIMyGridDownNotifyEvent;
import org.ourgrid.mygrid.ui.gui.event.GUIPeerAliveNotifyEvent;
import org.ourgrid.mygrid.ui.gui.event.GUIPeerIsDeadNotifyEvent;
import org.ourgrid.mygrid.ui.gui.event.GUIReadyNotifyEvent;
import org.ourgrid.mygrid.ui.gui.event.GUIRemoveJobNotifyEvent;
import org.ourgrid.mygrid.ui.gui.event.GUISetPeersNotifyEvent;
import org.ourgrid.mygrid.ui.gui.event.GuiJobCanceledNotifyEvent;
import org.ourgrid.mygrid.ui.gui.event.GuiJobStateChangedNotifyEvent;
import org.ourgrid.mygrid.ui.gui.event.GuiReplicaStateChangedNotifyEvent;

/**
 * Facade used to contact MyGrid's GUI.
 */
public class EBGuiFacade implements JobManagerListener, GridManagerListener {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( EBGuiFacade.class );

	/**
	 * The EventEngine
	 */
	private GuiEventEngine guiEventEngine;

	private SimpleSynchronizedJobManager jobManager;

	private SimpleSynchronizedGridManager gridManager;

	private MyGridGUI gui;

	private GUIListenerImpl listener;

	private UIManager manager;


	/**
	 * Creates a new <code>GuiFacade</code>
	 * 
	 * @throws RemoteException
	 */
	public EBGuiFacade() throws RemoteException {

	}


	public void startEventProcessor() {

		guiEventEngine.startProcessing();
	}


	public void jobCreated( JobEntry jobEntry ) {

		this.guiEventEngine.putEvent( new GuiJobStateChangedNotifyEvent( jobEntry, this.jobManager ) );

	}


	public void replicaStateChanged( ReplicaEntry replicaEntry, ExecutionStatus jobState, ExecutionStatus taskState,
										int numberOfTaskFails ) {

		this.guiEventEngine.putEvent( new GuiReplicaStateChangedNotifyEvent( replicaEntry, jobState, taskState,
				numberOfTaskFails, this.jobManager ) );
	}


	public void removeJob( int jobId ) {

		this.guiEventEngine.putEvent( new GUIRemoveJobNotifyEvent( jobId, this.jobManager ) );
	}


	public void jobCanceled( int jobId ) {

		this.guiEventEngine.putEvent( new GuiJobCanceledNotifyEvent( jobManager, jobId ) );
	}


	public void hereIsGum( GumSpec spec, int jobID, ObjectID peerBrokerAccessID ) {

		this.guiEventEngine.putEvent( new GUIHereIsGumNotifyEvent( spec, jobID, peerBrokerAccessID, this.gridManager ) );
	}


	public void setPeers( Collection<PeerSpec> peerSpecs ) {

		this.guiEventEngine.putEvent( new GUISetPeersNotifyEvent( peerSpecs, this.gridManager ) );
	}


	public void deleteGum( GumID gumID ) {

		this.guiEventEngine.putEvent( new GUIDeleteGumNotifyEvent( gumID, this.gridManager ) );
	}


	public void allocate( GumID gumID, ReplicaEntry replica ) {

		this.guiEventEngine.putEvent( new GUIAllocateNotifyEvent( gumID, replica, this.gridManager ) );
	}


	public void gumIsReady( GumID gumID ) {

		this.guiEventEngine.putEvent( new GUIReadyNotifyEvent( gumID, this.gridManager ) );
	}


	public void peerAlive( PeerSpec peerSpec ) {

		this.guiEventEngine.putEvent( new GUIPeerAliveNotifyEvent( peerSpec, this.gridManager ) );
	}


	public void peerIsDead( PeerSpec peerSpec ) {

		this.guiEventEngine.putEvent( new GUIPeerIsDeadNotifyEvent( peerSpec, this.gridManager ) );
	}


	public void myGridDown() {

		this.guiEventEngine.putEvent( new GUIMyGridDownNotifyEvent( listener ) );
	}


	public MyGridGUI getGUI() {

		return gui;
	}


	public void config( UIManager manager, GUIListenerImpl listener ) throws RemoteException, ConfigException {

		this.listener = listener;
		this.manager = manager;

		Configuration.getInstance( Configuration.MYGRID );

		this.jobManager = new SimpleSynchronizedJobManager();
		this.gridManager = new SimpleSynchronizedGridManager();
		this.gui = new MyGridGUI( this, jobManager, gridManager );
		this.guiEventEngine = new GuiEventEngine( gui );

		Collection<JobEntry> jobs = manager.getJobs( listener );
		Collection<GridManagerEntry> gums = manager.getGumEntries( listener );
		Collection<PeerEntry> peers = manager.getPeerEntries( listener );

		this.jobManager.createEntries( jobs );
		this.gridManager.createEntries( gums, peers );

		this.gui.refreshGridTab();
		this.gui.refreshJobsTab();
		this.gui.refreshReceivedGumsTab();
	}


	public void shutdown() {

		try {
			manager.removeGridManagerListener( this.listener );
			manager.removeJobManagerListener( this.listener );
		} catch ( RemoteException e ) {
			LOG.error( "Could not shutdown GUIFacade: " + e );
		}
	}
}
