/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui.gui;

import java.awt.Color;
import java.awt.GridLayout;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTree;
import javax.swing.border.LineBorder;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeSelectionModel;

import org.ourgrid.common.spec.PeerSpec;

/**
 * This class generate the GuMP Manager, it has a instance of MyGridUIServices
 * to obtain informations to create table.<br>
 * Implements the abstract methods of AbstractTableModel to instantiate JTable.
 */
public class GridPanel implements TreeSelectionListener, Serializable {

	private static final long serialVersionUID = 33L;

	private final static String LINE_SEPARATOR = System.getProperty( "line.separator" );

	private static final String PEER_NAME = "Name";

	private static final String PEER_URL = "URL";

	private static final String PEER_PORT = "Port";

	private static final String PEER_LABEL = "Label";

	private static final String PEER_STATE = "Status";

	/** JPanel */
	private JPanel masterPanel;

	/** Properties Area */
	private JTextArea propertiesArea;

	/** JTree */
	private JTree tree;

	/** Tree's Scroll */
	private JScrollPane scroll;

	/** PropertiesArea's Scroll */
	private JScrollPane propertiesScroll;

	/** DefaultTreeModel */
	private GridTreeModel gtm;

	private final SimpleSynchronizedGridManager simpleGridManager;


	/**
	 * Creates a new <code>GridPanel</code>
	 */
	public GridPanel( SimpleSynchronizedGridManager simpleGridManager ) {

		this.simpleGridManager = simpleGridManager;
		masterPanel = new JPanel();
		masterPanel.setLayout( new GridLayout( 1, 2 ) );
		masterPanel.setDoubleBuffered( true );

		gtm = new GridTreeModel();

		// Tree
		configureTree( gtm.getTreeModel() );

		// Scroll
		configureScroll();

		// Properties Area
		configurePropertiesArea();
	}


	private void configurePropertiesArea() {

		propertiesArea = new JTextArea();
		propertiesArea.setBorder( new LineBorder( Color.BLACK ) );
		propertiesArea.setEditable( false );
		propertiesScroll = new JScrollPane( propertiesArea );
		masterPanel.add( propertiesScroll );
		masterPanel.setDoubleBuffered( true );
	}


	private void configureScroll() {

		scroll = new JScrollPane( tree );
		scroll.setViewportView( tree );
		masterPanel.add( scroll );
	}


	private void configureTree( DefaultTreeModel treeModel ) {

		tree = new JTree( treeModel );
		tree.setLargeModel( true );
		tree.getSelectionModel().setSelectionMode( TreeSelectionModel.SINGLE_TREE_SELECTION );
		tree.addTreeSelectionListener( this );
		tree.setCellRenderer( new GridTreeCellRenderer() );
		tree.setDoubleBuffered( true );
		tree.setOpaque( true );
		tree.setBorder( new LineBorder( Color.BLACK ) );
	}


	/** Return JPanel with JScrollPane */
	public JPanel getPanel() {

		return masterPanel;
	}


	public void refreshTree() {

		valueChanged( null );

		synchronized ( gtm.getTree() ) {
			gtm.refreshTree( simpleGridManager.getPeers() );
		}
	}


	/**
	 * Clean the table
	 */
	public void clean() {

		propertiesArea.repaint();
		tree.repaint();
		scroll.repaint();
	}


	/**
	 * @see javax.swing.event.TreeSelectionListener#valueChanged(javax.swing.event.TreeSelectionEvent)
	 */
	public void valueChanged( TreeSelectionEvent e ) {

		PeerNode node = (PeerNode) tree.getLastSelectedPathComponent();

		if ( node == null )
			return;

		if ( node.getSpec() == null ) {
			propertiesArea.setText( "" );
			return;
		}

		showPropertiesOfGuMP( node );
	}


	/**
	 * Search the GuMP properties and put in properties Area.
	 * 
	 * @param node GuMP label in tree.
	 * @throws RemoteException
	 */
	private void showPropertiesOfGuMP( PeerNode node ) {

		PeerSpec peerSpec = node.getSpec();

		Map<String,String> p = new HashMap<String,String>();
		p.put( PEER_NAME, peerSpec.getName() );
		p.put( PEER_PORT, String.valueOf( peerSpec.getPort() ) );
		p.put( PEER_LABEL, peerSpec.getLabel() );
		p.put( PEER_URL, peerSpec.getURL() );
		p.put( PEER_STATE, node.getState().toString() );
		writeGuMPInTextArea( p );
	}


	/**
	 * Put all properties in properties area.
	 * 
	 * @param p Properties.
	 */
	private void writeGuMPInTextArea( Map p ) {

		propertiesArea.setText( "" );

		StringBuffer sb = new StringBuffer();
		sb.append( PEER_NAME + ": " + p.get( PEER_NAME ) );
		sb.append( LINE_SEPARATOR );
		sb.append( PEER_LABEL + ": " + p.get( PEER_LABEL ) );
		sb.append( LINE_SEPARATOR );
		sb.append( PEER_PORT + ": " + p.get( PEER_PORT ) );
		sb.append( LINE_SEPARATOR );
		sb.append( PEER_URL + ": " + p.get( PEER_URL ) );
		sb.append( LINE_SEPARATOR );
		sb.append( PEER_STATE + ": " + p.get( PEER_STATE ) );

		propertiesArea.setText( sb.toString() );
		propertiesArea.repaint();
	}
}
