/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui.gui;

import org.ourgrid.common.event.EventEngine;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.mygrid.ui.gui.event.GUIActionEvent;
import org.ourgrid.mygrid.ui.gui.event.GUIEventType;

/**
 * This class is responsible for receiving events that come from the
 * <code>EBSchedulerFacade</code> and process them.
 */
public class GuiEventEngine implements EventEngine<GUIActionEvent> {

	/**
	 * Serial UID
	 */
	private static final long serialVersionUID = 33L;

	/**
	 * A <CODE>EventQueue</CODE> that is filled by <CODE>putEvent()</CODE>
	 * method and provides the events to <CODE>run()</CODE> method consuming.
	 */
	private EventQueue<GUIActionEvent> eventQueue;

	/**
	 * Thread that performs <CODE>Event</CODE> consuming.
	 */
	private Thread myThread;

	/** Indicates if this <CODE>EventEngine</CODE> is alive. */
	private boolean isAlive;

	/**
	 * Logger to store log messages
	 */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( GuiEventEngine.class );

	private MyGridGUI gui;


	/**
	 * Creates a new <CODE>PeerEventEngine</CODE>.
	 */
	public GuiEventEngine( MyGridGUI gui ) {

		this.gui = gui;
		this.eventQueue = new EventQueue<GUIActionEvent>();
		this.isAlive = false;
		this.myThread = new Thread( this, "GuiEventEngine" );
	}


	/**
	 * Begin consuming of events. Consuming means remove events from the queue
	 * and process (call the <CODE>process()</CODE> method) them.
	 */
	public void startProcessing() {

		this.isAlive = true;
		this.myThread.start();
	}


	/**
	 * Inserts a <CODE>Event</CODE>- to be processed - into a internal <CODE>
	 * EventQueue</CODE>.
	 * 
	 * @param event A <CODE>Event</CODE> to be processed.
	 */
	public void putEvent( GUIActionEvent event ) {

		this.eventQueue.put( event );
	}


	/**
	 * Consumes events from internal <CODE>EventQueue</CODE>. In other words,
	 * it removes events from the queue and process (call the <CODE>process()
	 * </CODE> method) them.
	 */
	public void run() {

		whileBlock: while ( true ) {
			try {
				GUIEventType eventType = eventQueue.blockingRemove().process();

				switch ( eventType ) {
					case GUMS:
						gui.refreshReceivedGumsTab();
						break;
					case PEERS:
						gui.refreshGridTab();
						break;
					case JOB:
						gui.refreshJobsTab();
						break;
					case MYGRIDDOWN:
						gui.myGridStopOccured();
						break whileBlock;
					default:
						break;
				}

				gui.refreshSelf();
			} catch ( Exception e ) {
				LOG.error( "Error occurred while processing events", e );
			}
		}

		System.exit( 0 );
	}


	/**
	 * @see org.ourgrid.common.event.EBSyncShutdownable#isAlive()
	 */
	public boolean isAlive() {

		return this.isAlive;
	}
}
