/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui.gui;

import java.awt.GridLayout;
import java.util.Map;

import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableColumn;

import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.util.BidimensionalArray;
import org.ourgrid.mygrid.scheduler.gridmanager.GridManagerEntry;
import org.ourgrid.mygrid.scheduler.jobmanager.ReplicaEntry;

/**
 * This class generate the Scheduler GuM Manager Table, it has a instance of
 * MyGridUIServices to obtain informations to create table. <br>
 * Implements the abstract methods of AbstractTableModel to instantiate JTable.
 */
public class GumTable extends AbstractTableModel {

	private static final long serialVersionUID = 33L;

	private static final int COLUMN_NUMBER = 0;

	private static final int COLUMN_NAME = 1;

	private static final int COLUMN_SITE = 2;

	private static final int COLUMN_CPU = 3;

	private static final int COLUMN_MEM = 4;

	private static final int COLUMN_TYPE = 5;

	private static final int COLUMN_REPLICA = 6;

	private static final int MIN_SIZE = 7;

	/** Array of column names */
	private String[ ] columnNames = { "#", "Name", "Site", "CPU", "MEM", "Type", "Running" };

	/** Bidimensional Array */
	private BidimensionalArray<String> data;

	/** JTree */
	private final JTable table;

	private SimpleSynchronizedGridManager simpleGridManager;

	private int rowCount;


	public GumTable( SimpleSynchronizedGridManager simpleGridManager ) {

		this.simpleGridManager = simpleGridManager;
		data = new BidimensionalArray<String>( MIN_SIZE );
		table = new JTable( this );
		table.setEnabled( false );
		table.setColumnSelectionAllowed( false );
		table.setCellSelectionEnabled( false );
		table.setRowSelectionAllowed( false );

		rowCount = 0;
	}


	/**
	 * Return JPanel with JScrollPane
	 */
	public JPanel getPanel() {

		JPanel panel = new JPanel();
		panel.setLayout( new GridLayout( 1, 1 ) );

		TableColumn numberColumn = table.getColumnModel().getColumn( 0 );
		numberColumn.setMaxWidth( 30 );
		numberColumn.setMinWidth( 30 );
		numberColumn.setPreferredWidth( 30 );

		JScrollPane scroll = new JScrollPane( table );
		panel.add( scroll );
		return panel;
	}


	/**
	 * Get Map of GuMs and put in table
	 */
	public void refreshGums() {

		synchronized ( data ) {
			refreshTable( this.simpleGridManager.getGums() );
		}
	}


	/**
	 * Repaints and updates the data table
	 */
	private void refreshData() {

		table.repaint();
		super.fireTableDataChanged();
	}


	/**
	 * Refreshes information in data table, reseting the old Gum list, and
	 * updating with this new one.
	 */
	private void refreshTable( Map<GumID,GridManagerEntry> gums ) {

		data = new BidimensionalArray<String>( MIN_SIZE );

		int i = 1;
		for ( GridManagerEntry entry : gums.values() ) {
			updateTable( entry, i );
			i++;
		}
		rowCount = gums.size();

		refreshData();
	}


	private void updateTable( GridManagerEntry entry, int gumNumber ) {

		final GumSpec gumSpec = entry.getGumSpec();
		GumID gumID = gumSpec.getGumID();

		String name = gumID.getGumName();
		String site = gumSpec.getAttribute( GumSpec.ATT_SITE );
		String cpu = gumSpec.getAttribute( GumSpec.ATT_PROCESSOR_FAMILY );
		String mem = gumSpec.getAttribute( GumSpec.ATT_MEM );
		String os = gumSpec.getAttribute( GumSpec.ATT_OS );

		data.insertAtColumn( COLUMN_NUMBER, gumNumber + "" );

		if ( name != null ) {
			data.insertAtColumn( COLUMN_NAME, name );
		} else {
			data.insertAtColumn( COLUMN_NAME, "?" );
		}

		if ( site != null ) {
			data.insertAtColumn( COLUMN_SITE, site );
		} else {
			data.insertAtColumn( COLUMN_SITE, "?" );
		}

		if ( cpu != null ) {
			data.insertAtColumn( COLUMN_CPU, cpu );
		} else {
			data.insertAtColumn( COLUMN_CPU, "?" );
		}

		if ( mem != null ) {
			data.insertAtColumn( COLUMN_MEM, mem );
		} else {
			data.insertAtColumn( COLUMN_MEM, "?" );
		}

		if ( os != null ) {
			data.insertAtColumn( COLUMN_TYPE, gumSpec.getAttribute( GumSpec.ATT_OS ) );
		} else {
			data.insertAtColumn( COLUMN_TYPE, "?" );
		}

		ReplicaEntry replica = entry.getReplica();
		if ( replica != null ) {
			data.insertAtColumn( COLUMN_REPLICA, replica.toString() );
		} else {
			data.insertAtColumn( COLUMN_REPLICA, "?" );
		}
	}


	/** Returns number of columns */
	public int getColumnCount() {

		return columnNames.length;
	}


	/**
	 * @see javax.swing.table.TableModel#getRowCount()
	 */
	public int getRowCount() {

		return rowCount;
	}


	/**
	 * Return object in rowIndex and columnIndex
	 * 
	 * @param rowIndex Row number
	 * @param columnIndex Column number
	 */
	public Object getValueAt( int rowIndex, int columnIndex ) {

		synchronized ( data ) {
			try {
				String value = data.get( rowIndex, columnIndex );
				return value == null ? "" : value;
			} catch ( ArrayIndexOutOfBoundsException e ) {
				return "";
			}
		}
	}


	@Override
	public String getColumnName( int column ) {

		synchronized ( data ) {
			return columnNames[column];
		}
	}


	@Override
	public boolean isCellEditable( int row, int col ) {

		return false;
	}


	public BidimensionalArray<String> getData() {

		return data;
	}


	public JTable getTable() {

		return table;
	}

}
