/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.rmi.RemoteException;
import java.util.Properties;

import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTree;
import javax.swing.border.LineBorder;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeSelectionModel;

import org.ourgrid.mygrid.scheduler.ExecutionStatus;

/**
 * This class contains SchedulerTreeModel and create a JTree with it.
 */
public class JobsPanel implements TreeSelectionListener {

	/** Keys to map properties to show in Job/Task/Replica properties area */
	private final static String LINE_SEPARATOR = System.getProperty( "line.separator" );

	private static final String REPLICA = "REPLICA";

	private static final String TASK = "TASK";

	private static final String JOB = "JOB";

	private static final String ASSIGNED_TO = "ASSIGNED TO";

	private static final String STATE = "STATE";

	private static final String INIT = "INIT";

	private static final String REMOTE = "REMOTE";

	private static final String FINAL = "FINAL";

	private static final String TOTAL_REPLICAS = "TOTAL REPLICAS";

	private static final String RUNNING_REPLICAS = "RUNNING REPLICAS";

	private static final String UNSTARTED_REPLICAS = "UNSTARTED REPLICAS";

	private static final String FINISHED_REPLICAS = "FINISHED REPLICAS";

	private static final String ABORTED_REPLICAS = "ABORTED REPLICAS";

	private static final String FAILED_REPLICAS = "FAILED REPLICAS";

	private static final String CANCELED_REPLICAS = "CANCELED REPLICAS";

	private static final String ACTUAL_FAILS = "ACTUAL FAILS";

	private static final String TASKSPEC = "TASKSPEC";

	private static final String ERROR_CAUSE = "ERROR CAUSE";

	private static final String REQUIREMENTS = "REQUIREMENTS";

	private static final String NUMBER_OF_TASKS = "NUMBER OF TASKS";

	private static final String NUMBER_OF_GUMS = "NUMBER OF GUMS";

	private static final String DETAILED_ERROR_CAUSE = "DETAILED ERROR CAUSE";

	private static final String UNSTARTED_TASKS = "UNSTARTED TASKS";

	/** Panel thats adds JTree */
	private JPanel masterPanel;

	private JPanel centerPanel;

	/** Jobs TreeModel */
	private JobsTreeModel jtm;

	/** JTree */
	private JTree tree;

	/** JScrollPane to formate JTree */
	private JScrollPane scroll;

	/** Properties Area */
	private JTextArea propertiesArea;

	/** PropertiesArea's Scroll */
	private JScrollPane propertiesScroll;

	private SimpleSynchronizedJobManager jobManager;


	/**
	 * Creates the JobPanel
	 */
	public JobsPanel( SimpleSynchronizedJobManager syncJobManager ) {

		this.jobManager = syncJobManager;
		masterPanel = new JPanel();
		masterPanel.setLayout( new BorderLayout() );
		masterPanel.setDoubleBuffered( true );

		centerPanel = new JPanel();
		centerPanel.setLayout( new BorderLayout() );
		centerPanel.setDoubleBuffered( true );

		jtm = new JobsTreeModel();

		tree = new JTree( jtm.getTreeModel() );
		tree.setCellRenderer( new JobsTreeCellRenderer() );
		tree.setLargeModel( true );
		tree.getSelectionModel().setSelectionMode( TreeSelectionModel.SINGLE_TREE_SELECTION );
		tree.addTreeSelectionListener( this );
		tree.setBorder( new LineBorder( Color.BLACK ) );
		tree.setDoubleBuffered( true );
		tree.setOpaque( true );

		scroll = new JScrollPane( tree );
		scroll.setViewportView( tree );
		scroll.setMinimumSize( new Dimension( 300, 100 ) );
		scroll.setPreferredSize( new Dimension( 300, 100 ) );
		scroll.setDoubleBuffered( true );
		centerPanel.add( scroll, BorderLayout.WEST );

		propertiesArea = new JTextArea();
		propertiesArea.setBorder( new LineBorder( Color.BLACK ) );
		propertiesArea.setEditable( false );
		propertiesScroll = new JScrollPane( propertiesArea );
		propertiesScroll.setDoubleBuffered( true );

		centerPanel.add( propertiesScroll, BorderLayout.CENTER );
		masterPanel.add( centerPanel, BorderLayout.CENTER );
		masterPanel.setDoubleBuffered( true );

	}


	/** Returns JTree's Panel */
	public JPanel getPanel() {

		return masterPanel;
	}


	/** Delegate refreshTree method of SchedulerTreeModel */
	public void refreshTree() {

		valueChanged( null );

		synchronized ( jtm.getTreeModel() ) {
			jtm.refreshTree( jobManager.getJobs() );
		}
	}


	public void clean() {

		propertiesArea.repaint();
		tree.repaint();
		scroll.repaint();
	}


	/**
	 * @see javax.swing.event.TreeSelectionListener#valueChanged(javax.swing.event.TreeSelectionEvent)
	 */
	public void valueChanged( TreeSelectionEvent e ) {

		DefaultMutableTreeNode node = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();

		if ( node == null ) {
			propertiesArea.setText( "" );
			return;
		}

		String nodeInfo = (String) node.getUserObject();

		if ( nodeInfo.equals( "Jobs" ) ) {
			propertiesArea.setText( "Click on a job for information" );
			return;
		}

		if ( nodeInfo.startsWith( "Task" ) ) {
			String jobInfo = (String) ((DefaultMutableTreeNode) node.getParent()).getUserObject();
			showPropertiesOfTask( nodeInfo, jobInfo );
		} else if ( nodeInfo.startsWith( "Replica" ) ) {
			DefaultMutableTreeNode taskNode = (DefaultMutableTreeNode) node.getParent();
			String taskInfo = (String) taskNode.getUserObject();
			DefaultMutableTreeNode jobNode = (DefaultMutableTreeNode) taskNode.getParent();
			String jobInfo = (String) jobNode.getUserObject();
			showPropertiesOfReplica( nodeInfo, taskInfo, jobInfo );
		} else {
			showPropertiesOfJob( nodeInfo );
		}
	}


	/**
	 * Search properties of a job and put in properties area.
	 * 
	 * @param jobLabel Job Label in tree.
	 * @throws RemoteException
	 */
	private void showPropertiesOfJob( String jobLabel ) {

		int id = getJobId( jobLabel );

		SimpleJobEntry jobEntry = jobManager.getJobByID( id );
		if ( jobEntry != null ) {
			Properties p = new Properties();
			p.put( JOB, String.valueOf( jobEntry.getJobID() ) );
			p.put( REQUIREMENTS, jobEntry.getJobSpec().getRequirements() );
			p.put( STATE, jobEntry.getState().toString() );
			p.put( NUMBER_OF_TASKS, String.valueOf( jobEntry.getTasks().size() ) );

			int unstartedReplicas = 0;
			int runningReplicas = 0;
			int finishedReplicas = 0;
			int abortedReplicas = 0;
			int canceledReplicas = 0;
			int failedReplicas = 0;
			int unstartedTasks = 0;

			for ( SimpleTaskEntry simpleTaskEntry : jobEntry.getTasks().values() ) {

				if ( simpleTaskEntry.getTaskState().equals( ExecutionStatus.UNSTARTED ) ) {
					unstartedTasks++;
				}

				for ( SimpleReplicaEntry simpleReplicaEntry : simpleTaskEntry.getReplicas().values() ) {
					switch ( simpleReplicaEntry.getReplicaState() ) {
						case UNSTARTED:
							unstartedReplicas++;
							break;
						case RUNNING:
							runningReplicas++;
							break;
						case FINISHED:
							finishedReplicas++;
							break;
						case ABORTED:
							abortedReplicas++;
							break;
						case FAILED:
							failedReplicas++;
							break;
						case CANCELLED:
							canceledReplicas++;
							break;
					}
				}
			}

			p.put( UNSTARTED_TASKS, String.valueOf( unstartedTasks ) );
			p.put( UNSTARTED_REPLICAS, String.valueOf( unstartedReplicas ) );
			p.put( RUNNING_REPLICAS, String.valueOf( runningReplicas ) );
			p.put( FINISHED_REPLICAS, String.valueOf( finishedReplicas ) );
			p.put( ABORTED_REPLICAS, String.valueOf( abortedReplicas ) );
			p.put( FAILED_REPLICAS, String.valueOf( failedReplicas ) );
			p.put( CANCELED_REPLICAS, String.valueOf( canceledReplicas ) );

			if ( !jobEntry.getState().equals( ExecutionStatus.CANCELLED ) ) {
				p.put( NUMBER_OF_GUMS, String.valueOf( runningReplicas + unstartedReplicas ) );
			} else {
				p.put( NUMBER_OF_GUMS, String.valueOf( 0 ) );
			}

			writeJobInTextArea( p );

			return;
		}

		propertiesArea.setText( "" );
	}


	/**
	 * Search the Task properties and put in properties area.
	 * 
	 * @param taskLabel Task label in tree.
	 * @param jobLabel Job Label in tree.
	 * @throws RemoteException
	 */
	private void showPropertiesOfTask( String taskLabel, String jobLabel ) {

		int id = getTaskId( taskLabel );
		int jobId = getJobId( jobLabel );

		SimpleJobEntry jobEntry = jobManager.getJobByID( jobId );
		if ( jobEntry != null ) {
			SimpleTaskEntry task = jobEntry.getTaskByID( id );
			if ( task != null ) {
				Properties p = new Properties();
				p.put( TASK, String.valueOf( task.getTaskID() ) );
				p.put( JOB, String.valueOf( task.getJobID() ) );
				p.put( STATE, task.getTaskState().toString() );
				p.put( TOTAL_REPLICAS, String.valueOf( task.getReplicas().size() ) );
				p.put( RUNNING_REPLICAS, String.valueOf( task.getNumberOfRunningReplicas() ) );
				p.put( UNSTARTED_REPLICAS, String.valueOf( task.getNumberOfUnstartedReplicas() ) );
				p.put( ACTUAL_FAILS, String.valueOf( task.getActualFails() ) );
				p.put( TASKSPEC, task.getTaskSpec().toString() );

				writeTaskInTextArea( p );

				return;
			}
		}

		propertiesArea.setText( "" );
	}


	/**
	 * Search the Replica properties and put in properties Area.
	 * 
	 * @param replicaLabel Replica label in tree.
	 * @param taskLabel Task label in tree.
	 * @param jobLabel Job label in tree.
	 * @throws RemoteException
	 */
	private void showPropertiesOfReplica( String replicaLabel, String taskLabel, String jobLabel ) {

		int id = getReplicaId( replicaLabel );
		int taskId = getTaskId( taskLabel );
		int jobId = getJobId( jobLabel );

		SimpleJobEntry jobEntry = jobManager.getJobByID( jobId );
		if ( jobEntry != null ) {
			SimpleTaskEntry task = jobEntry.getTaskByID( taskId );
			if ( task != null ) {
				SimpleReplicaEntry replica = task.getReplicaByID( id );
				if ( replica != null ) {
					Properties p = new Properties();
					p.put( REPLICA, String.valueOf( replica.getReplicaID() ) );
					p.put( TASK, String.valueOf( replica.getTaskID() ) );
					p.put( JOB, String.valueOf( replica.getJobID() ) );
					p.put( ASSIGNED_TO, replica.getGumID().toString() );
					p.put( STATE, replica.getReplicaState().toString() );

					if ( replica.getReplicaState().equals( ExecutionStatus.FAILED ) ) {
						if ( replica.getResult().getErrorCause() != null ) {
							if ( replica.getResult().getErrorCause().getMessage() != null ) {
								String errorCause = replica.getResult().getErrorCause().getSimpleMessage();
								p.put( ERROR_CAUSE, errorCause );

								String detailedErrorCause = replica.getResult().getErrorCause().getMessage();
								p.put( DETAILED_ERROR_CAUSE, detailedErrorCause );
							}
						}
					} else if ( replica.getReplicaState().equals( ExecutionStatus.FINISHED ) ) {

						p.put( INIT, new Long( replica.getResult().getInitData().getEndTime()
								- replica.getResult().getInitData().getStartTime() ).toString() );
						p.put( REMOTE, new Long( replica.getResult().getRemoteData().getEndTime()
								- replica.getResult().getRemoteData().getStartTime() ).toString() );
						p.put( FINAL, new Long( replica.getResult().getFinalData().getEndTime()
								- replica.getResult().getFinalData().getStartTime() ).toString() );

					}

					writeReplicaInTextArea( p );

					return;
				}
			}
		}

		propertiesArea.setText( "" );
	}


	/**
	 * Put all properties in properties area.
	 * 
	 * @param p Properties.
	 */
	private void writeReplicaInTextArea( Properties p ) {

		propertiesArea.setText( "" );

		StringBuffer sb = new StringBuffer();
		sb.append( " JOB: " + p.getProperty( JOB ) );
		sb.append( LINE_SEPARATOR + " TASK: " + p.getProperty( TASK ) );
		sb.append( LINE_SEPARATOR + " REPLICA: " + p.getProperty( REPLICA ) );
		sb.append( LINE_SEPARATOR + " ASSIGNED TO: " + p.getProperty( ASSIGNED_TO ) );
		sb.append( LINE_SEPARATOR + " STATE: " + p.getProperty( STATE ) );

		if ( p.getProperty( ERROR_CAUSE ) != null ) {
			sb.append( LINE_SEPARATOR + " -----" + LINE_SEPARATOR + " ERROR CAUSE: " + LINE_SEPARATOR
					+ p.getProperty( ERROR_CAUSE ) );
		}

		if ( p.getProperty( DETAILED_ERROR_CAUSE ) != null ) {
			sb.append( LINE_SEPARATOR + "" + LINE_SEPARATOR + " ADDITIONAL INFORMATION: " + LINE_SEPARATOR
					+ p.getProperty( DETAILED_ERROR_CAUSE ) );
		}

		if ( p.get( INIT ) != null ) {

			sb.append( LINE_SEPARATOR + "-----" + LINE_SEPARATOR + " EXECUTION TIMES: " );
			sb.append( LINE_SEPARATOR + "    INIT PHASE: " + p.get( INIT ) + " ms." );
			sb.append( LINE_SEPARATOR + "    REMOTE PHASE: " + p.get( REMOTE ) + " ms." );
			sb.append( LINE_SEPARATOR + "    FINAL PHASE: " + p.get( FINAL ) + " ms." );

		}

		propertiesArea.setText( sb.toString() );
	}


	/**
	 * Put all properties in properties
	 * area.rid.mygrid.ui.gui.JobsPanel.getJobId(JobsPanel.java:436)
	 * 
	 * @param p Properties.
	 */
	private void writeTaskInTextArea( Properties p ) {

		propertiesArea.setText( "" );

		StringBuffer sb = new StringBuffer();
		sb.append( " JOB: " + p.getProperty( JOB ) );
		sb.append( LINE_SEPARATOR + " TASK: " + p.getProperty( TASK ) );
		sb.append( LINE_SEPARATOR + " TOTAL REPLICAS: " + p.getProperty( TOTAL_REPLICAS ) );
		sb.append( LINE_SEPARATOR + " RUNNING REPLICAS: " + p.getProperty( RUNNING_REPLICAS ) );
		sb.append( LINE_SEPARATOR + " UNSTARTED REPLICAS: " + p.getProperty( UNSTARTED_REPLICAS ) );
		sb.append( LINE_SEPARATOR + " ACTUAL FAILS: " + p.getProperty( ACTUAL_FAILS ) );
		sb.append( LINE_SEPARATOR + " STATE: " + p.getProperty( STATE ) );
		sb.append( LINE_SEPARATOR + "-----" + LINE_SEPARATOR + " TASK SPEC:" );
		sb.append( LINE_SEPARATOR + p.getProperty( TASKSPEC ) );

		propertiesArea.setText( sb.toString() );
	}


	/**
	 * Put all properties in properties area.
	 * 
	 * @param p Properties.
	 */
	private void writeJobInTextArea( Properties p ) {

		propertiesArea.setText( "" );

		StringBuffer sb = new StringBuffer();
		sb.append( " Job	: " + p.getProperty( JOB ) );

		String requirements = p.getProperty( REQUIREMENTS );
		if ( requirements == null || requirements.equals( "" ) ) {
			requirements = "unspecified";
		}

		sb.append( LINE_SEPARATOR + " REQUIREMENTS: " + requirements );
		sb.append( LINE_SEPARATOR + " STATE: " + p.getProperty( STATE ) );
		sb.append( LINE_SEPARATOR + " NUMBER OF TASKS: " + p.getProperty( NUMBER_OF_TASKS ) );
		sb.append( LINE_SEPARATOR + " UNSTARTED OF TASKS: " + p.getProperty( UNSTARTED_TASKS ) );
		sb.append( LINE_SEPARATOR + "----" + LINE_SEPARATOR + "INFO:" + LINE_SEPARATOR );
		sb.append( LINE_SEPARATOR + " UNSTARTED REPLICAS: " + p.getProperty( UNSTARTED_REPLICAS ) );
		sb.append( LINE_SEPARATOR + " RUNNING REPLICAS: " + p.getProperty( RUNNING_REPLICAS ) );
		sb.append( LINE_SEPARATOR + " FINISHED REPLICAS: " + p.getProperty( FINISHED_REPLICAS ) );
		sb.append( LINE_SEPARATOR + " ABORTED REPLICAS: " + p.getProperty( ABORTED_REPLICAS ) );
		sb.append( LINE_SEPARATOR + " FAILED REPLICAS: " + p.getProperty( FAILED_REPLICAS ) );
		sb.append( LINE_SEPARATOR + " CANCELED REPLICAS: " + p.getProperty( CANCELED_REPLICAS ) );
		sb
				.append( LINE_SEPARATOR + "" + LINE_SEPARATOR + " CURRENT NUMBER OF GUMS: "
						+ p.getProperty( NUMBER_OF_GUMS ) );

		propertiesArea.setText( sb.toString() );
	}


	/**
	 * Returns the id of a Replica.
	 * 
	 * @param replicaLabel Label thats showed in the tree.
	 * @return Integer with the replica id.
	 */
	private int getReplicaId( String replicaLabel ) {

		int i = 8;
		while ( replicaLabel.charAt( i ) != ':' )
			i++;
		String idString = replicaLabel.substring( 8, i );
		return Integer.parseInt( idString );
	}


	/**
	 * Returns the id of a Task.
	 * 
	 * @param taskLabel Label that's showed in the tree.
	 * @return Integer with the task id.
	 */
	private int getTaskId( String taskLabel ) {

		int i = 5;
		while ( taskLabel.charAt( i ) != ':' )
			i++;
		String idString = taskLabel.substring( 5, i );
		return Integer.parseInt( idString );
	}


	/**
	 * Returns the id of a Job.
	 * 
	 * @param jobLabel Label that's showed in the tree.
	 * @return Integer with the job id.
	 */
	private int getJobId( String jobLabel ) {

		int i = 4;
		while ( jobLabel.charAt( i ) != ':' )
			i++;
		String idString = jobLabel.substring( 4, i );
		return Integer.parseInt( idString );
	}


	/**
	 * Returns the job id of the selected job in tree.
	 * 
	 * @return job id, -1 if selected node is not a job.
	 */
	public int getJobIdToCancel() {

		DefaultMutableTreeNode node = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();

		if ( node == null )
			return -1;

		String nodeInfo = (String) node.getUserObject();

		if ( !nodeInfo.startsWith( "Job" ) || nodeInfo.startsWith( "Jobs" ) )
			return -1;

		return getJobId( nodeInfo );
	}

}
