/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui.gui;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;

import org.ourgrid.common.id.GumID;

/**
 * This class generate the Job Tree. It access a </code>SynchronizedEBJobManager</code>
 * to the necessary information. It also has maps to store all jobs, tasks and
 * replicas that have already been inserted in tree to prevent exhausting
 * algorithms to verify if jobs, tasks and replicas already had been inserted in
 * the tree.
 */
public class JobsTreeModel {

	/** DefaultTreeNode */
	private DefaultTreeModel treeModel;

	/** The Root Node */
	private DefaultMutableTreeNode jobRoot;

	/** All jobs that already had been added in Tree */
	private Map<String,DefaultMutableTreeNode> mapNodeJobs;

	/** All tasks that already had been added in Tree */
	private Map<String,DefaultMutableTreeNode> mapNodeTasks;

	/** All replicas that already had been added in Tree */
	private Map<String,DefaultMutableTreeNode> mapNodeReplicas;


	/**
	 * Creates a new instance of LocalGuMPTree
	 */
	public JobsTreeModel() {

		jobRoot = new DefaultMutableTreeNode( "Jobs" );
		treeModel = new DefaultTreeModel( jobRoot );

		mapNodeJobs = new HashMap<String,DefaultMutableTreeNode>();
		mapNodeTasks = new HashMap<String,DefaultMutableTreeNode>();
		mapNodeReplicas = new HashMap<String,DefaultMutableTreeNode>();
	}


	/** Returns Tree Model to create JTree */
	public DefaultTreeModel getTreeModel() {

		return treeModel;
	}


	/**
	 * Get job's list from MyGridUIServices and update the tree
	 */
	public void refreshTree( Map<Integer,SimpleJobEntry> jobsList ) {

		updateTree( jobsList );
	}


	private void updateTree( Map<Integer,SimpleJobEntry> allJobs ) {

		Set<String> knownJobs = new HashSet<String>( mapNodeJobs.keySet() );

		// Removing jobs that were deleted
		synchronized ( treeModel ) {
			for ( String jobID : knownJobs ) {
				int id = Integer.parseInt( jobID );

				// Job has been removed
				if ( !allJobs.containsKey( id ) ) {
					DefaultMutableTreeNode jobNode = mapNodeJobs.remove( jobID );
					treeModel.removeNodeFromParent( jobNode );

				}
			}
		}

		// Updating known jobs.
		for ( SimpleJobEntry jobEntry : allJobs.values() ) {
			DefaultMutableTreeNode jobNode = updateJobNode( jobEntry );

			for ( SimpleTaskEntry taskEntry : jobEntry.getTasks().values() ) {
				DefaultMutableTreeNode taskNode = updateTaskNode( jobNode, taskEntry );

				for ( SimpleReplicaEntry replicaEntry : taskEntry.getReplicas().values() ) {
					updateReplicaNode( taskNode, replicaEntry );
				}
			}
		}
	}


	private DefaultMutableTreeNode updateJobNode( SimpleJobEntry jobEntry ) {

		DefaultMutableTreeNode jobNode = mapNodeJobs.get( jobEntry.getJobID() + "" );
		String jobInfo = getJobInfo( jobEntry );

		if ( jobNode == null ) {
			jobNode = new DefaultMutableTreeNode( jobInfo );
			treeModel.insertNodeInto( jobNode, jobRoot, jobRoot.getChildCount() );
			mapNodeJobs.put( jobEntry.getJobID() + "", jobNode );
		} else {
			jobNode.setUserObject( jobInfo );
		}
		return jobNode;
	}


	private DefaultMutableTreeNode updateTaskNode( DefaultMutableTreeNode jobNode, SimpleTaskEntry taskEntry ) {

		DefaultMutableTreeNode taskNode = mapNodeTasks.get( taskEntry.getJobID() + "." + taskEntry.getTaskID() );
		String taskInfo = getTaskInfo( taskEntry );

		if ( taskNode == null ) {
			taskNode = new DefaultMutableTreeNode( taskInfo );
			treeModel.insertNodeInto( taskNode, jobNode, jobNode.getChildCount() );
			mapNodeTasks.put( taskEntry.getJobID() + "." + taskEntry.getTaskID(), taskNode );
		} else {
			taskNode.setUserObject( taskInfo );
		}
		return taskNode;
	}


	private void updateReplicaNode( DefaultMutableTreeNode taskNode, SimpleReplicaEntry replicaEntry ) {

		DefaultMutableTreeNode replicaNode = mapNodeReplicas.get( replicaEntry.getJobID() + "."
				+ replicaEntry.getTaskID() + "." + replicaEntry.getReplicaID() );
		String replicaInfo = getReplicaInfo( replicaEntry );

		if ( replicaNode == null ) {
			replicaNode = new DefaultMutableTreeNode( replicaInfo );
			treeModel.insertNodeInto( replicaNode, taskNode, taskNode.getChildCount() );
			mapNodeReplicas.put( replicaEntry.getJobID() + "." + replicaEntry.getTaskID() + "."
					+ replicaEntry.getReplicaID(), replicaNode );
		} else {
			replicaNode.setUserObject( replicaInfo );
		}
	}


	private String getReplicaInfo( SimpleReplicaEntry replicaEntry ) {

		GumID gumID = replicaEntry.getGumID();

		if ( gumID != null ) {
			return "Replica " + replicaEntry.getReplicaID() + ":"
					+ formatStatus( gumID.getGumName(), replicaEntry.getReplicaState().toString(), true );
		}

		return "Replica " + replicaEntry.getReplicaID() + ":"
				+ formatStatus( "", replicaEntry.getReplicaState().toString(), true );
	}


	private String getTaskInfo( SimpleTaskEntry taskEntry ) {

		return "Task " + taskEntry.getTaskID() + ":" + formatStatus( "", taskEntry.getTaskState().toString(), false );
	}


	private String getJobInfo( SimpleJobEntry jobEntry ) {

		return jobEntry.getJobSpec().getLabel() != null ? "Job " + jobEntry.getJobID() + ": "
				+ jobEntry.getJobSpec().getLabel() + "  [" + jobEntry.getState().toString() + "]" : "Job "
				+ jobEntry.getJobID() + ": [" + jobEntry.getState().toString() + "]";
	}


	/**
	 * Return a formated string with task status
	 * 
	 * @param gumName the grid machine name that is running the task
	 * @param status the status of the task
	 * @param isReplica tells if the task running is a replica
	 * @return A string with the formated task status
	 */
	private static String formatStatus( String gumName, String status, boolean isReplica ) {

		String statusFormated = "";
		if ( status.equals( "Running" ) && isReplica )
			statusFormated = " [" + status + "] at: " + gumName;
		else
			statusFormated = " [" + status + "]";
		return statusFormated;
	}


	/** This method is friendly because test needs a reference to Node Root */
	public DefaultMutableTreeNode getGuMRoot() {

		return jobRoot;
	}

}
