/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.mygrid.ui.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.GridLayout;
import java.awt.Toolkit;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.filechooser.FileFilter;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.PeerSpec;
import org.ourgrid.common.spec.main.DescriptionFileCompile;
import org.ourgrid.mygrid.scheduler.exception.JobCouldNotBeCancelledException;
import org.ourgrid.mygrid.ui.MyGridUIManager;

/**
 * MyGrid GUI, this class contains the main method.
 */

public class MyGridGUI implements Serializable {

	/**
	 * Serial Version
	 */
	private static final long serialVersionUID = 33L;

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( MyGridGUI.class );

	/** String constants */
	private static final String JOB_COULD_NOT_BE_CANCELLED = "Job could not be cancelled.";

	private static final String ERROR_JOB = "Select a job in job tree.";

	private static final String ERROR_ON_ABOUT = "Error on reading about contents.";

	private static final String UNEXPECTED_ERROR = "Unexpected error: ";

	private static final String COMMUNICATION_PROBLEM = "Communication problem. MyGrid is not running!";

	private static final String ERROR_OCCURRED = "Error occurred";

	private static final String GUMS = "Received Machines";

	private static final String GUM_PROVIDERS = "Peers";

	private static final String JOBS = "Jobs";

	private static final String SETTINGS = "Configuration";

	private static final String ABOUT = "About";

	private static final String CANCELJOB_BUTTON = "Cancel Job";

	private static final String ADDJOB_BUTTON = "Add Job";

	private static final String SETGRID_BUTTON = "Set Grid";

	private static final String LINE_SEPARATOR = System.getProperty( "line.separator" );

	private static final String FILE_SEPARATOR = System.getProperty( "file.separator" );

	private static final String MYGRID_TITLE = "MyGrid " + Configuration.VERSION;

	private static final String CLEANJOB_BUTTON = "Clean Job";

	private static final String CLEANALLJOB_BUTTON = "Clean All";

	private final String mgroot = Configuration.getInstance( Configuration.MYGRID ).getProperty( Configuration.MGROOT );

	/** About Contents file path */
	private final String ABOUT_CONTENTS_FILE = mgroot + FILE_SEPARATOR + "resources" + FILE_SEPARATOR + "ABOUT";

	/** Icon file path */
	private final String ICON_PATH = mgroot + FILE_SEPARATOR + "resources/images" + FILE_SEPARATOR + "icon.png";

	/** JFrame */
	private JFrame frame;

	/**
	 * Stores different sections of the GUI, MyGrid config section, Local GuMP
	 * section, GuMCache section...
	 */
	private JTabbedPane tabbedPane;

	/** This Panel contains all the other panels of this GUI */
	private JPanel masterPanel;

	/** This Panel contains the buttons of start and stop MyGrid */
	private JPanel masterButtonsPanel;

	/** This Panel contains SchedulerGuMManager section */
	private JPanel schedulerGuMManagerSection;

	/** This Panel contains GuMPManager section */
	private JPanel gumpManagerSection;

	/** This panel contains Jobs section */
	private JPanel jobsSection;

	/** This panel contains About section */
	private JPanel aboutSection;

	/** MyGridSection */
	private SettingsPanel mygridSection;

	/** GuMCacheTable */
	private GumTable gumsTable;

	/** GuMPManagerTable */
	private GridPanel gridPanel;

	/** SchedulerPanel */
	private JobsPanel jobsPanel;

	private JButton setGridButton;

	private JButton addJobButton;

	private JButton cancelJobButton;

	/** Opened file */
	private File openedFile = null;

	private SimpleSynchronizedGridManager syncGridManager;

	private SimpleSynchronizedJobManager syncJobManager;

	private JButton cleanJobButton;

	private JButton cleanAllJobsButton;


	public MyGridGUI( final EBGuiFacade facade, SimpleSynchronizedJobManager jobManager,
						SimpleSynchronizedGridManager gridManager ) throws ConfigException {

		this.syncJobManager = jobManager;
		this.syncGridManager = gridManager;

		/**
		 * Security options.
		 */
		if ( Configuration.getInstance().isSecurityEnabled() ) {

			String trustStoreFilename = Configuration.getInstance()
					.getProperty( Configuration.PROP_TRUSTSTORE_FILENAME );
			String trustStorePassword = Configuration.getInstance()
					.getProperty( Configuration.PROP_TRUSTSTORE_PASSWORD );

			System.setProperty( "javax.net.ssl.trustStore", trustStoreFilename );
			System.setProperty( "javax.net.ssl.trustStorePassword", trustStorePassword );

		}

		frame = new JFrame();
		frame.setTitle( MYGRID_TITLE );
		frame.setSize( 740, 512 );
		frame.setLocationRelativeTo( null );
		frame.setResizable( true );
		frame.setIconImage( Toolkit.getDefaultToolkit().createImage( ICON_PATH ) );
		frame.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );

		// Window close event
		frame.addWindowListener( new WindowAdapter() {

			@Override
			public void windowClosing( WindowEvent e ) {

				facade.shutdown();
			}
		} );

		initMasterPanel();
		initMasterButtonsPanel();
		initTabbedPane();

	}


	/** Delegates the show() method */
	public void show() {

		frame.setVisible( true );
	}


	/** Starts the Master Panel */
	private void initMasterPanel() {

		masterPanel = new JPanel();
		masterPanel.setLayout( new BorderLayout() );
		masterPanel.setDoubleBuffered( true );
		frame.setContentPane( masterPanel );
	}


	/** Starts the Master Buttons Panel and configuring it */
	private void initMasterButtonsPanel() {

		masterButtonsPanel = new JPanel();

		setGridButton = new JButton( SETGRID_BUTTON );
		setGridButton.setToolTipText( "Set a list of Peers to be used by MyGrid" );

		addJobButton = new JButton( ADDJOB_BUTTON );
		addJobButton.setToolTipText( "Adds a Job to be executed by MyGrid" );

		cancelJobButton = new JButton( CANCELJOB_BUTTON );
		cancelJobButton.setToolTipText( "Cancels the Job currenty selected" );

		cleanJobButton = new JButton( CLEANJOB_BUTTON );
		cleanJobButton.setToolTipText( "Removes information on the selected Job from MyGrid" );

		cleanAllJobsButton = new JButton( CLEANALLJOB_BUTTON );
		cleanAllJobsButton.setToolTipText( "Removes information on all the Jobs that have ended" );

		masterButtonsPanel.add( setGridButton );
		masterButtonsPanel.add( addJobButton );
		masterButtonsPanel.add( cancelJobButton );
		masterButtonsPanel.add( cleanJobButton );
		masterButtonsPanel.add( cleanAllJobsButton );

		masterPanel.add( masterButtonsPanel, BorderLayout.SOUTH );

		addJobButton.addActionListener( new java.awt.event.ActionListener() {

			public void actionPerformed( java.awt.event.ActionEvent evt ) {

				addJobActionPerformed();
			}
		} );

		cancelJobButton.addActionListener( new java.awt.event.ActionListener() {

			public void actionPerformed( java.awt.event.ActionEvent evt ) {

				cancelJobActionPerformed();
			}
		} );

		setGridButton.addActionListener( new java.awt.event.ActionListener() {

			public void actionPerformed( java.awt.event.ActionEvent evt ) {

				setGridActionPerformed();
			}
		} );

		cleanJobButton.addActionListener( new java.awt.event.ActionListener() {

			public void actionPerformed( java.awt.event.ActionEvent evt ) {

				cleanActionPerformed();
			}
		} );

		cleanAllJobsButton.addActionListener( new java.awt.event.ActionListener() {

			public void actionPerformed( java.awt.event.ActionEvent evt ) {

				cleanAllActionPerformed();
			}
		} );
	}


	/** Inits the Tabbed Pane adding and configuting tabs */
	private void initTabbedPane() {

		tabbedPane = new JTabbedPane();

		masterPanel.add( tabbedPane, BorderLayout.CENTER );

		// MyGrid Panel
		mygridSection = new SettingsPanel();
		tabbedPane.addTab( SETTINGS, mygridSection.getPanel() );

		// Jobs Section
		jobsPanel = new JobsPanel( syncJobManager );
		jobsSection = jobsPanel.getPanel();
		tabbedPane.addTab( JOBS, jobsSection );

		// GuMPManager and LocalGuMP
		gumpManagerSection = new JPanel();
		gumpManagerSection.setLayout( new GridLayout( 1, 1 ) );

		gridPanel = new GridPanel( syncGridManager );

		gumpManagerSection.add( gridPanel.getPanel() );
		tabbedPane.addTab( GUM_PROVIDERS, gumpManagerSection );

		// Scheduler GuM Manager
		schedulerGuMManagerSection = new JPanel();
		schedulerGuMManagerSection.setLayout( new GridLayout( 1, 1 ) );
		gumsTable = new GumTable( syncGridManager );
		schedulerGuMManagerSection.add( gumsTable.getPanel() );
		tabbedPane.addTab( GUMS, schedulerGuMManagerSection );

		// about
		aboutSection = new JPanel();
		aboutSection.setLayout( new GridLayout( 1, 1 ) );
		aboutSection.add( createAbout() );
		aboutSection.setBackground( Color.LIGHT_GRAY );
		tabbedPane.addTab( ABOUT, aboutSection );
	}


	/**
	 * Creates a JPanel containing the About text.
	 * 
	 * @return A JPanel containing the About text.
	 */
	private JPanel createAbout() {

		JPanel aboutPanel = new JPanel();
		aboutPanel.setBackground( Color.LIGHT_GRAY );
		JTextArea aboutText = new JTextArea( getAboutContents() );
		aboutText.setEditable( false );
		aboutText.setBackground( Color.LIGHT_GRAY );
		aboutPanel.add( aboutText );
		return aboutPanel;

	}


	/**
	 * When addJob button is clicked
	 */
	void addJobActionPerformed() {

		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setCurrentDirectory( new File( "." ) );

		fileChooser.setFileFilter( new FileFilter() {

			@Override
			public boolean accept( File f ) {

				return f.getName().toLowerCase().endsWith( ".jdf" ) || f.isDirectory();
			}


			@Override
			public String getDescription() {

				return "Job Description File (*.jdf)";
			}
		} );

		openedFile = this.getSelectedFile( fileChooser );

		if ( openedFile == null )
			return;

		new Thread( new UIManagerAddJobContacter() ).start();
	}


	/**
	 * When setGrid button is clicked
	 */
	void setGridActionPerformed() {

		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setCurrentDirectory( new File( "." ) );

		fileChooser.setFileFilter( new FileFilter() {

			@Override
			public boolean accept( File f ) {

				return f.getName().toLowerCase().endsWith( ".gdf" ) || f.isDirectory();
			}


			@Override
			public String getDescription() {

				return "Grid Description File (*.gdf)";
			}
		} );

		openedFile = this.getSelectedFile( fileChooser );

		if ( openedFile == null )
			return;

		new Thread( new UIManagerSetGridContacter() ).start();
	}


	/**
	 * Return Selected file in file chooser
	 * 
	 * @param fileChooser The FileChooser
	 */
	private File getSelectedFile( JFileChooser fileChooser ) {

		int result = fileChooser.showOpenDialog( this.frame );
		if ( result == JFileChooser.APPROVE_OPTION )
			return fileChooser.getSelectedFile();
		return null;
	}


	/**
	 * When cancel job button is clicked
	 */
	void cancelJobActionPerformed() {

		int jobId = jobsPanel.getJobIdToCancel();

		if ( jobId == -1 ) {
			JOptionPane.showMessageDialog( frame, ERROR_JOB, ERROR_OCCURRED, JOptionPane.ERROR_MESSAGE );
		} else {
			int option = JOptionPane.showConfirmDialog( frame, "Really Cancel Job " + jobId + "?",
					"Confirm Cancel Job", JOptionPane.YES_NO_OPTION );

			if ( option == JOptionPane.YES_OPTION ) {
				new Thread( new UIManagerCancelJobContacter( jobId ) ).start();
			}
		}
	}


	void cleanAllActionPerformed() {

		new Thread( new UIManagerCleanAllContacter() ).start();
	}


	void cleanActionPerformed() {

		int jobId = jobsPanel.getJobIdToCancel();

		if ( jobId == -1 ) {
			JOptionPane.showMessageDialog( frame, ERROR_JOB, ERROR_OCCURRED, JOptionPane.ERROR_MESSAGE );
		} else {
			new Thread( new UIManagerCleanJobContacter( jobId ) ).start();
		}
	}


	/**
	 * Shows the user a message when an exception occurs.
	 * 
	 * @param e The exception occurred.
	 */
	private void showErrorToUser( Exception e ) {

		String message = getFriedlyMessage( e );
		JOptionPane.showMessageDialog( this.frame, message, ERROR_OCCURRED, JOptionPane.ERROR_MESSAGE );

	}


	/**
	 * Returns a user-friendly error message.
	 * 
	 * @param e The exception occurred.
	 * @return An user-friendly error message according to the exception.
	 */
	private String getFriedlyMessage( Exception e ) {

		if ( e instanceof RemoteException ) {
			return COMMUNICATION_PROBLEM;
		}

		if ( e instanceof JobCouldNotBeCancelledException ) {
			return JOB_COULD_NOT_BE_CANCELLED;
		}

		return UNEXPECTED_ERROR + LINE_SEPARATOR + e.getMessage();
	}


	/**
	 * Reads the about contents from a file.
	 * 
	 * @return A string containg the about.
	 */
	private String getAboutContents() {

		String aboutContents = "";
		BufferedReader br = null;
		try {
			br = new BufferedReader( new FileReader( ABOUT_CONTENTS_FILE ) );
			String line = null;
			while ( (line = br.readLine()) != null ) {
				aboutContents += line + LINE_SEPARATOR;
			}
			br.close();
		} catch ( IOException e ) {
			return ERROR_ON_ABOUT;
		} finally {
			try {
				if ( br != null ) {
					br.close();
				}
			} catch ( IOException e ) {
				LOG.debug( "Failed to close stream on Exception." );
			}
		}
		return aboutContents;
	}


	public void refreshJobsTab() {

		jobsPanel.refreshTree();
	}


	public void refreshReceivedGumsTab() {

		gumsTable.refreshGums();
	}


	public void refreshGridTab() {

		gridPanel.refreshTree();
	}


	public void refreshSelf() {

		synchronized ( frame ) {
			frame.repaint();
		}
	}


	public void myGridStopOccured() {

		JOptionPane.showMessageDialog( frame, "MyGrid is down, GUI is closing", "MyGrid Down",
				JOptionPane.WARNING_MESSAGE );
		frame.dispose();
	}

	private class UIManagerAddJobContacter implements Runnable {

		public void run() {

			try {
				JobSpec spec = DescriptionFileCompile.compileJDF( openedFile.getAbsolutePath() );
				MyGridUIManager.getInstance().addJob( spec );
			} catch ( Exception re ) {
				showErrorToUser( re );
			}
		}
	}

	private class UIManagerSetGridContacter implements Runnable {

		public void run() {

			try {
				List<PeerSpec> specs = DescriptionFileCompile.compileGDF( openedFile.getAbsolutePath() );
				MyGridUIManager.getInstance().setPeers( specs );
			} catch ( Exception e ) {
				showErrorToUser( e );
			}
		}
	}

	private class UIManagerCancelJobContacter implements Runnable {

		private final int jobId;


		public UIManagerCancelJobContacter( int jobId ) {

			this.jobId = jobId;
		}


		public void run() {

			try {
				MyGridUIManager.getInstance().cancelJob( jobId );
			} catch ( Exception e ) {
				showErrorToUser( e );
			}
		}
	}

	private class UIManagerCleanJobContacter implements Runnable {

		private final int jobId;


		public UIManagerCleanJobContacter( int jobId ) {

			this.jobId = jobId;
		}


		public void run() {

			try {
				MyGridUIManager.getInstance().removeJob( jobId );
			} catch ( Exception e ) {
				showErrorToUser( e );
			}
		}
	}

	private class UIManagerCleanAllContacter implements Runnable {

		public void run() {

			try {
				MyGridUIManager.getInstance().cleanFinishedJobs();
			} catch ( Exception e ) {
				showErrorToUser( e );
			}
		}
	}

}
