/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.community;

import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.common.spec.RequestSpec.RequestSource;
import org.ourgrid.common.url.PeerURLProvider;
import org.ourgrid.common.url.URLProvider;
import org.ourgrid.corepeer.CorePeer;
import org.ourgrid.mygrid.scheduler.gridmanager.GumResponseEntry;
import org.ourgrid.peer.community.communicationlayer.CommunicationLayerImpl;
import org.ourgrid.peer.community.nof.AccountingManager;
import org.ourgrid.peer.community.obtainer.EBCommunityObtainerFacade;
import org.ourgrid.peer.ebgum.EBGum;
import org.ourgrid.peer.ebgum.EBGumCreator;
import org.ourgrid.peer.manager.EBPeerManagerFacade;
import org.ourgrid.peer.manager.allocation.AllocationStatus;
import org.ourgrid.peer.manager.status.StatusEntry;
import org.ourgrid.peer.ui.PeerUIManager;
import org.ourgrid.peer.ui.PeerUIManager.Notifiables;

public class RemoteAccessImpl extends OurgridUnicastRemoteObject implements RemoteAccess {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	private static final long serialVersionUID = 33L;

	/**
	 * This peer's ID.
	 */
	private ObjectID peerID;

	/**
	 * Know peers list.
	 */
	private Collection<ObjectID> peerList;

	/**
	 * A map with all known peers on the community, associated with an URL.
	 */
	private Map<String,ObjectID> knownPeersbyURL;

	/**
	 * Logger to store log messages
	 */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( RemoteAccessImpl.class );

	/**
	 * The community obtainer facade.
	 */
	private transient EBCommunityObtainerFacade ebCommunityObtainerFacade;

	private EBPeerManagerFacade ebPeerManagerFacade;


	/**
	 * Creates a peerImpl
	 * 
	 * @param peerID This peer's ID
	 * @param ebCommunityObtainerFacade This peer's community obtainer
	 * @param ebPeerManagerFacade This peer's community proxy
	 * @throws RemoteException If couldn't create the peer.
	 */
	public RemoteAccessImpl( ObjectID peerID, EBCommunityObtainerFacade ebCommunityObtainerFacade,
								EBPeerManagerFacade ebPeerManagerFacade ) throws RemoteException {

		super();
		this.peerID = peerID;
		this.ebCommunityObtainerFacade = ebCommunityObtainerFacade;
		this.ebPeerManagerFacade = ebPeerManagerFacade;

	}


	public ObjectID getObjectID() throws RemoteException {

		return peerID;
	}


	/**
	 * @see org.ourgrid.peer.community.RemoteAccess#hereIsPeerList(Collection)
	 */
	public void hereIsPeerList( Collection<ObjectID> peerListReceived ) throws RemoteException {

		LOG.info( "Peer list received with " + peerListReceived.size() + " peers." );
		if ( this.peerList == null )
			this.peerList = new LinkedList<ObjectID>();
		if ( knownPeersbyURL == null )
			knownPeersbyURL = new HashMap<String,ObjectID>();

		synchronized ( peerListReceived ) {

			// Removing this peer if it's in receivedList
			peerListReceived.remove( this.peerID );

			// Delete peers which come with the same URL but different peerIDs
			for ( ObjectID pID : peerListReceived ) {
				ObjectID p = knownPeersbyURL.get( pID.getObjectURL() );
				if ( p != null ) {
					if ( !p.getIdValue().equals( pID.getIdValue() ) ) {
						this.peerList.remove( p );
						knownPeersbyURL.remove( p.getObjectURL() );
					}
				}
				knownPeersbyURL.put( pID.getObjectURL(), pID );
			}

			Collection<ObjectID> newPeerList = getPreviouslyUnknownPeers( this.peerList, peerListReceived );
			Collection<ObjectID> absentPeers = getAbsentPeers( this.peerList, peerListReceived );

			verifyAndRemoveAbsent( this.peerList, absentPeers );

			newPeerList.addAll( this.peerList );

			this.peerList = newPeerList;

			CommunicationLayerImpl.getInstance().hereIsPeerList( this.peerList );

		}

	}


	private void verifyAndRemoveAbsent( Collection<ObjectID> peerList, Collection<ObjectID> absentPeers ) {

		try {

			Collection<StatusEntry> remoteGums = PeerUIManager.getInstance().getGums( AllocationStatus.IN_USE,
					RequestSource.FROM_REMOTE_PEER );

			for ( Iterator<ObjectID> iter = absentPeers.iterator(); iter.hasNext(); ) {
				ObjectID absentPeer = iter.next();
				String absentPeerName = URLProvider.getHost( absentPeer.getObjectURL() );
				String absentPeerPort = new Integer( URLProvider.getPort( absentPeer.getObjectURL() ) ).toString();

				boolean remove = true;
				for ( StatusEntry entry : remoteGums ) {

					String gumPeerName = entry.getGumSpec().getAttribute( GumSpec.ATT_PEERNAME );
					String gumPeerPort = entry.getGumSpec().getAttribute( GumSpec.ATT_PEERPORT );

					if ( absentPeerName.equals( gumPeerName ) && absentPeerPort.equals( gumPeerPort ) ) {
						remove = false;
					}

				}

				if ( !remove )
					iter.remove();

			}

			peerList.removeAll( absentPeers );

		} catch ( RemoteException e ) {
			LOG.error( "Can't get remote gum list.", e );
		} catch ( URISyntaxException e ) {
			LOG.error( "Can't get remote gum list.", e );
		}

	}


	private Collection<ObjectID> getPreviouslyUnknownPeers( Collection<ObjectID> oldList, Collection<ObjectID> list ) {

		if ( oldList.size() == 0 )
			return list;

		Collection<ObjectID> unknownPeers = new LinkedList<ObjectID>( list );

		unknownPeers.removeAll( oldList );

		return unknownPeers;
	}


	private Collection<ObjectID> getAbsentPeers( Collection<ObjectID> oldList, Collection<ObjectID> list ) {

		Collection<ObjectID> absent = new LinkedList<ObjectID>( oldList );

		absent.removeAll( list );

		return absent;
	}


	/**
	 * @return Return the know peer list.
	 */
	public Collection<ObjectID> getPeerList() {

		return peerList;
	}


	/*
	 * Used only by tests.
	 */
	public void setPeerList( List<ObjectID> peerList ) {

		this.peerList = peerList;
	}


	/**
	 * Sets this peer's ID.
	 * 
	 * @param peerID The peer's ID
	 */
	public void setPeerID( ObjectID peerID ) {

		this.peerID = peerID;
	}


	/**
	 * Joins the OurGrid Community.
	 * 
	 * @throws RemoteException If could not contact the core peer.
	 */
	public void joinCommunity() throws RemoteException {

		CorePeer corePeer = null;
		try {

			PeerUIManager.getInstance().registerNewMonitorable( Notifiables.COREPEER,
					PeerURLProvider.corePeerFailureDetector() );

			LOG.info( "Trying to (re)join the OurGrid Community" );

			String corepeerURL = PeerURLProvider.corePeer();
			corePeer = (CorePeer) Naming.lookup( corepeerURL );
			peerList = corePeer.hereIAm( peerID );
			CommunicationLayerImpl.getInstance().hereIsPeerList( peerList );

			LOG.info( "OurGrid Community was sucessfully (re) joined." );

		} catch ( MalformedURLException e ) {
			// Nothing to do
		} catch ( NotBoundException e ) {
			LOG.info( "Corepeer could not be contacted. Is it UP?" );

		}

	}


	public void hereIsGum( GumResponseEntry gumResponseEntry ) throws RemoteException {

		EBGum ebGum = EBGumCreator.createGuM( gumResponseEntry.getGum(), gumResponseEntry.getGumSpec() );
		ebCommunityObtainerFacade.hereIsGum( ebGum, gumResponseEntry.getObjectID(), gumResponseEntry.getRequestID() );
	}


	public void requestGums( RequestSpec requestSpec ) throws RemoteException {

		AccountingManager.getInstance().registerPeerIdForRequest( requestSpec.getID(), requestSpec.getObjectID() );
		if ( requestSpec.getRequiredGums() == 0 ) {
			throw new RemoteException( "Invalid number of required gums." );
		}
		requestSpec.setAsFromRemotePeer();
		ebPeerManagerFacade.requestGums( requestSpec );
	}


	public void finishRequestGums( long requestID ) throws RemoteException {

		ebPeerManagerFacade.finishRequestGums( requestID );

		AccountingManager.getInstance().removeRequest( requestID );
	}


	public void disposeGum( long requestID, GumID gumID ) throws RemoteException {

		AccountingManager.getInstance().finishAccounting( gumID, requestID, true );

		ebPeerManagerFacade.disposeGum( requestID, gumID );
	}


	public void pauseRequestGums( long requestID ) throws RemoteException {

		ebPeerManagerFacade.pauseRequestGums( requestID );
	}


	public void resumeRequestGums( long requestID ) throws RemoteException {

		ebPeerManagerFacade.resumeRequestGums( requestID );
	}


	public void unwantedGum( long requestID, GumID gumID ) throws RemoteException {

		AccountingManager.getInstance().finishAccounting( gumID, requestID, true );

		this.ebPeerManagerFacade.unwantedGum( requestID, gumID );
	}

}
