/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.community.communicationlayer;

import java.rmi.Naming;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.PeerConfiguration;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.peer.community.RemoteAccess;

public class CommunicationLayerImpl implements CommunicationLayer {

	/**
	 * This peer id.
	 */
	private ObjectID peerID;

	/**
	 * Logger to store log messages
	 */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( CommunicationLayerImpl.class );

	/**
	 * The know peer list.
	 */
	private Collection<ObjectID> peerList = new Vector<ObjectID>();

	/**
	 * The singleton instance.
	 */
	private static CommunicationLayerImpl instance = null;

	/**
	 * The threadPool used by the wannaGuMs broadcast.
	 */
	private ThreadPoolExecutor threadPool;

	private class RequestSender implements Runnable {

		private ObjectID peerID;

		private RequestSpec requestSpec;


		public RequestSender( ObjectID peerID, RequestSpec requestSpec ) {

			this.peerID = peerID;
			this.requestSpec = requestSpec;

		}


		public void run() {

			try {

				RemoteAccess cl = getPeerRef( CommunicationLayerImpl.getInstance().getPeerID() );
				RequestSpec requestSpecToPropagate = new RequestSpec( requestSpec.getID(), requestSpec
						.getRequirements(), requestSpec.getRequiredGums(), cl );
				requestSpecToPropagate.setAsFromRemotePeer();
				((RemoteAccess) Naming.lookup( peerID.getObjectURL() )).requestGums( requestSpecToPropagate );

			} catch ( Exception e ) {
				LOG.error( "Could not send wannaGuMs request to peer " + peerID.getName() + " at "
						+ peerID.getObjectURL() + ". Cause: " + e.getClass().getName() );
			}

		}
	}

	private class FinishRequestSender implements Runnable {

		private ObjectID peerID;

		private long requestID;


		public FinishRequestSender( ObjectID peerID, long requestID ) {

			this.peerID = peerID;
			this.requestID = requestID;

		}


		public void run() {

			try {

				((RemoteAccess) Naming.lookup( peerID.getObjectURL() )).finishRequestGums( requestID );

			} catch ( Exception e ) {
				LOG.error( "Could not send noMoreGuMs request to peer " + peerID.getName() + ". Cause: "
						+ e.getClass().getName() );
			}

		}
	}

	private class PauseRequestSender implements Runnable {

		private ObjectID peerID;

		private long requestID;


		public PauseRequestSender( ObjectID peerID, long requestID ) {

			this.peerID = peerID;
			this.requestID = requestID;

		}


		public void run() {

			try {

				((RemoteAccess) Naming.lookup( peerID.getObjectURL() )).pauseRequestGums( requestID );

			} catch ( Exception e ) {
				LOG.error( "Could not send suspendWannaGuMs request to peer " + peerID.getName() + ". Cause: "
						+ e.getClass().getName() );
			}

		}
	}

	private class ResumeRequestSender implements Runnable {

		private ObjectID peerID;

		private long requestID;


		public ResumeRequestSender( ObjectID peerID, long requestID ) {

			this.peerID = peerID;
			this.requestID = requestID;

		}


		public void run() {

			try {

				((RemoteAccess) Naming.lookup( peerID.getObjectURL() )).resumeRequestGums( requestID );

			} catch ( Exception e ) {
				LOG.error( "Could not send resumeWannaGuMs request to peer " + peerID.getName() + ". Cause: "
						+ e.getClass().getName() );
			}

		}
	}

	private class DisposeGumSender implements Runnable {

		private ObjectID peerID;

		private GumID gumID;

		private long requestID;


		public DisposeGumSender( ObjectID peerID, long requestID, GumID gumID ) {

			this.peerID = peerID;
			this.requestID = requestID;
			this.gumID = gumID;
		}


		public void run() {

			try {
				((RemoteAccess) Naming.lookup( peerID.getObjectURL() )).disposeGum( requestID, gumID );
			} catch ( Exception e ) {
				LOG.error( "Could not send resumeWannaGuMs request to peer " + peerID.getName() + ". Cause: "
						+ e.getClass().getName() );
			}
		}
	}

	private class UnwantedGumSender implements Runnable {

		private ObjectID peerID;

		private GumID gumID;

		private long requestID;


		public UnwantedGumSender( ObjectID peerID, long requestID, GumID gumID ) {

			this.peerID = peerID;
			this.requestID = requestID;
			this.gumID = gumID;
		}


		public void run() {

			try {
				((RemoteAccess) Naming.lookup( peerID.getObjectURL() )).unwantedGum( requestID, gumID );
			} catch ( Exception e ) {
				LOG.error( "Could not send resumeWannaGuMs request to peer " + peerID.getName() + ". Cause: "
						+ e.getClass().getName() );
			}
		}
	}


	/**
	 * @return The singleton intance.
	 */
	public synchronized static final CommunicationLayerImpl getInstance() {

		if ( instance == null ) {
			instance = new CommunicationLayerImpl();
		}

		return instance;

	}


	/**
	 * Creates the Communication Layer.
	 */
	protected CommunicationLayerImpl() {

		this.threadPool = new ThreadPoolExecutor( 1, Integer.MAX_VALUE, 60, TimeUnit.SECONDS,
				new SynchronousQueue<Runnable>() );
	}


	/**
	 * Configures the ConcreteCommunicationLayer.
	 */
	public void config( RemoteAccess remoteAccess ) {

		try {
			this.peerID = remoteAccess.getObjectID();
		} catch ( RemoteException e ) {
			// Must never be thrown, remoteaccess is local to this class.
		}

		if ( Configuration.getInstance().isEnabled( PeerConfiguration.PROP_JOIN_COMMUNITY ) ) {
			try {
				remoteAccess.joinCommunity();
			} catch ( Exception e ) {
				LOG.error( "Could not join OurGrid Community. Cause: " + e.getMessage() );
			}
		}

	}


	/**
	 * @see org.ourgrid.peer.community.communicationlayer.CommunicationLayer#getPeerID()
	 */
	public ObjectID getPeerID() {

		return peerID;
	}


	/**
	 * @see org.ourgrid.peer.community.communicationlayer.CommunicationLayer#hereIsPeerList(Collection)
	 */
	public void hereIsPeerList( Collection<ObjectID> peerList ) {

		LOG.info( "Peer list received with " + peerList.size() + " peers." );
		synchronized ( this.peerList ) {
			this.peerList = peerList;
			if ( this.peerList.contains( peerID ) ) {
				peerList.remove( peerID );
			}
		}

	}


	/**
	 * @param peerID
	 */
	private RemoteAccess getPeerRef( ObjectID peerID ) throws RemoteException {

		try {
			return (RemoteAccess) Naming.lookup( peerID.getObjectURL() );
		} catch ( Exception e ) {
			throw new RemoteException( "Could not contact peer: " + e );
		}
	}


	/**
	 * @see org.ourgrid.peer.community.communicationlayer.CommunicationLayer#sendRequest(RequestSpec)
	 */
	public void sendRequest( RequestSpec requestSpec ) {

		synchronized ( peerList ) {

			for ( ObjectID peerID : peerList ) {
				if ( !peerID.equals( this.peerID ) ) {
					threadPool.execute( new RequestSender( peerID, requestSpec ) );
				}
			}
		}
	}


	/**
	 * @see org.ourgrid.peer.community.communicationlayer.CommunicationLayer#finishRequest(long)
	 */
	public void finishRequest( long requestID ) {

		synchronized ( peerList ) {

			Iterator<ObjectID> it = peerList.iterator();
			while ( it.hasNext() ) {
				ObjectID peerID = it.next();

				if ( !peerID.equals( this.peerID ) ) {
					threadPool.execute( new FinishRequestSender( peerID, requestID ) );
				}

			}

		}

	}


	/**
	 * @see org.ourgrid.peer.community.communicationlayer.CommunicationLayer#pauseRequest(long)
	 */
	public void pauseRequest( long requestID ) {

		synchronized ( peerList ) {

			Iterator<ObjectID> it = peerList.iterator();
			while ( it.hasNext() ) {
				ObjectID peerID = it.next();

				if ( !peerID.equals( this.peerID ) ) {
					threadPool.execute( new PauseRequestSender( peerID, requestID ) );
				}

			}

		}

	}


	/**
	 * @see org.ourgrid.peer.community.communicationlayer.CommunicationLayer#resumeRequest(long)
	 */
	public void resumeRequest( long requestID ) {

		synchronized ( peerList ) {

			Iterator<ObjectID> it = peerList.iterator();
			while ( it.hasNext() ) {
				ObjectID peerID = it.next();

				if ( !peerID.equals( this.peerID ) ) {
					threadPool.execute( new ResumeRequestSender( peerID, requestID ) );
				}

			}

		}

	}


	/**
	 * @see org.ourgrid.peer.community.communicationlayer.CommunicationLayer#disposeOf(GumID,
	 *      long, ObjectID)
	 */
	public void disposeOf( GumID gumID, long requestID, ObjectID peerID ) {

		threadPool.execute( new DisposeGumSender( peerID, requestID, gumID ) );
	}


	/**
	 * @see org.ourgrid.peer.community.communicationlayer.CommunicationLayer#unwantedGum(ObjectID,
	 *      long, GumID)
	 */
	public void unwantedGum( ObjectID gumOwnerID, long requestID, GumID gumID ) {

		threadPool.execute( new UnwantedGumSender( gumOwnerID, requestID, gumID ) );
	}
}
