/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.community.nof;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.Map.Entry;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.PeerConfiguration;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;

/**
 * <p>
 * Stores the reputation of all known peers. It coordinates the peer's
 * strategies for calculating and maintaining his knowledge about the other
 * peer's reputations. Each peer maintains a map with a balance for each other
 * known peer calculated as a function of the past bilateral interactions
 * between this and each other known peer.
 * </p>
 * <p>
 * Each peer accounts both for each favor that it makes and each favor that is
 * made to him.
 * </p>
 */
public class AccountingManager {

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( AccountingManager.class );

	/**
	 * The default value of the reputation of an unknown peer.
	 */
	public static final double DEFAULT_UNKNOWN_PEER_BALANCE = 0.0;

	/**
	 * This factor is a timer scale.
	 */
	public static final double FACTOR = 1000;

	/**
	 * Maps the associate a requestId with the PeerId. Only for foreign
	 * requests.
	 */
	private Map<Long,ObjectID> peerIdForRequestId;

	/**
	 * Times of usage start
	 */
	private Map<String,Long> initialTimes;

	/**
	 * Current ranking of peers reputations
	 */
	private Map<ObjectID,Double> knownPeersBalances;

	/**
	 * Timer for frequently save the ranking in file.
	 */
	private Timer saveRankingTimer;

	/**
	 * Singleton instance.
	 */
	private static AccountingManager uniqueInstance = null;


	private AccountingManager() {

		knownPeersBalances = new HashMap<ObjectID,Double>();
		initialTimes = new HashMap<String,Long>();

		this.saveRankingTimer = new Timer();

		/*
		 * keeps saving the ranking in file. note that the first execution is
		 * done after OurConfig.getTimeBetweenRankingSaves() seconds.
		 */
		Configuration config = Configuration.getInstance();
		int timeout = Integer.parseInt( config.getProperty( PeerConfiguration.PROP_SAVING_INTERVAL ) );
		this.saveRankingTimer.scheduleAtFixedRate( new AccountingMaintainer( this ), timeout, timeout );

		this.loadRankingFromFile( config.getProperty( PeerConfiguration.PROP_RANKINGFILE ) );

		this.peerIdForRequestId = new HashMap<Long,ObjectID>();

		Runtime.getRuntime().addShutdownHook( new Thread( new AccountingMaintainer( this ), "AccountShutdownSaver" ) );
	}


	public void startAccounting( GumID gumID, long requestID ) {

		final ObjectID peerIdForRequest = getPeerIdForRequest( requestID );

		if ( peerIdForRequest != null )
			this.startAccounting( gumID, peerIdForRequest );
	}


	/**
	 * <p>
	 * Informs, with accounting purpose, that the usage of a gum has started.
	 * Note that it is used both when providing or receiving a machine.
	 * </p>
	 * 
	 * @param gumID The internet name of GridMachine just being used.
	 * @param peerID The peer's ID that is receiving or providing the machine.
	 */
	public void startAccounting( GumID gumID, ObjectID peerID ) {

		/*
		 * A peer can receive and/or provide many different machines, thats why
		 * we make a new id to be used as our key
		 */
		String key = generateKey( gumID, peerID.getIdValue().toString() );

		/*
		 * hold the current time as the moment the GridMachine started to be
		 * used
		 */
		Long startTime = new Long( System.currentTimeMillis() );
		synchronized ( initialTimes ) {
			initialTimes.put( key, startTime );
		}
	}


	public double finishAccounting( GumID gumID, long requestID, boolean isProviding ) {

		return finishAccounting( gumID, getPeerIdForRequest( requestID ), isProviding );
	}


	public double finishAccounting( GumID gumID, ObjectID peerID, boolean isProviding ) {

		String key = generateKey( gumID, peerID.getIdValue().toString() );

		/*
		 * get the moment that the GridMachine started to be used as the
		 * accounting
		 */
		Long initialTimeObj;
		synchronized ( initialTimes ) {
			initialTimeObj = initialTimes.remove( key );
		}

		/*
		 * if there is not an initial time for the given GridMachine it was not
		 * being used. That is why it is an illegal account.
		 */
		if ( initialTimeObj == null ) {
			LOG.error( "Tried to account grid machine <" + gumID + "> to peer [" + peerID
					+ "] but the machine was not being used" );
			return 0.0;
		}

		/* calculate the used time */
		double initialTime = initialTimeObj.longValue();
		double finalTime = System.currentTimeMillis();
		double usedTime = finalTime - initialTime;

		synchronized ( knownPeersBalances ) {
			/* calculate the new balance */
			double newBalance = calculateNewBalance( peerID, usedTime, isProviding );

			/* update the balance of the peer */
			knownPeersBalances.put( peerID, newBalance );
		}
		return usedTime;

	}


	/**
	 * Calculates the new balance based on the use of a given <code>GuM</code>.
	 * 
	 * @param usedTime The used time of this <code>GuM</code>.
	 * @param peerID The peer that is using or providind this <code>GuM</code>.
	 * @param isProviding If the peer is providing the <code>GuM</code>.
	 * @return The new balance based on the use of a given <code>GuM</code>.
	 */
	protected double calculateNewBalance( ObjectID peerID, double usedTime, boolean isProviding ) {

		double currentBalance = getPeerBalance( peerID );
		if ( isProviding )
			return providerAccount( usedTime, currentBalance );

		return consumerAccount( usedTime, currentBalance );
	}


	/**
	 * <p>
	 * Gets the balance of a peer represented by its ID.
	 * </p>
	 * 
	 * @param peerID The ID of peer whose balance is wished.
	 * @return The balance of the peer reprented by its ID
	 */
	public double getPeerBalance( ObjectID peerID ) {

		/* if the peer is know get its balance */
		if ( isAKnownPeer( peerID ) ) {

			Double balance = knownPeersBalances.get( peerID );
			return balance.doubleValue();
		}
		return DEFAULT_UNKNOWN_PEER_BALANCE;
	}


	public List<Entry<ObjectID,Double>> getKnownPeers() {

		List<Entry<ObjectID,Double>> entries = new LinkedList<Entry<ObjectID,Double>>( knownPeersBalances.entrySet() );

		Collections.sort( entries, new EntryComparator() );

		return entries;

	}


	/**
	 * <p>
	 * Verifies if the peer is a known one.
	 * </p>
	 * 
	 * @param peerID The peer is
	 * @return true if the peer is known, false otherwise.
	 */
	protected boolean isAKnownPeer( ObjectID peerID ) {

		if ( peerID != null ) {

			return knownPeersBalances.containsKey( peerID );

		}

		return false;
	}


	/**
	 * <p>
	 * As two GridMachines of two different consumers can have the same id we
	 * need to have a unique key given a Grid Machine ID and a Consumer ID
	 * </p>
	 * 
	 * @param gumID The GridMachine ID.
	 * @param consumerID The consumer ID
	 * @return A new unique ID from the GridMachine ID and consumer ID.
	 */
	private String generateKey( GumID gumID, String consumerID ) {

		return gumID + consumerID;
	}


	/**
	 * <p>
	 * Gets a Map containing all the known peer balances.
	 * </p>
	 * 
	 * @return A Map containing all the known peer balances.
	 */
	public Map<ObjectID,Double> getKnownPeerBalances() {

		return new HashMap<ObjectID,Double>( knownPeersBalances );
	}


	/**
	 * <p>
	 * Saves the current ranking to a file. Currently, this file is written as a
	 * Properties file.
	 * </p>
	 * 
	 * @param fileName The name of the file to write into.
	 * @throws IOException If any problem happens while writing the file.
	 */
	public synchronized void saveRankingInFile( String fileName ) throws IOException {

		if ( fileName != null ) {

			ObjectOutputStream obj = null;

			try {

				obj = new ObjectOutputStream( new FileOutputStream( fileName ) );

				obj.writeObject( getKnownPeerBalances() );

			} finally {

				if ( obj != null ) {
					obj.close();
				}
			}
		}

	}


	/**
	 * <p>
	 * Loads the ranking from a file.
	 * </p>
	 * 
	 * @param fileName The name of the file to read from.
	 */
	public void loadRankingFromFile( String fileName ) {

		try {

			ObjectInputStream obj = new ObjectInputStream( new FileInputStream( fileName ) );

			Map<ObjectID,Double> savedBalances = (Map<ObjectID,Double>) obj.readObject();
			knownPeersBalances = savedBalances;
			obj.close();

		} catch ( Exception ioe ) {
			LOG.debug( "Could not find the ranking file " + fileName );
		}

	}


	/**
	 * Gets the peer's id of the peer that make the request identified by the
	 * given requestID
	 * 
	 * @param requestID The request ID.
	 * @return the peerID or null if the request isn't associate with any
	 *         peerID.
	 */
	public synchronized ObjectID getPeerIdForRequest( long requestID ) {

		Long key = new Long( requestID );

		ObjectID peerID = peerIdForRequestId.get( key );

		return peerID;
	}


	/**
	 * Associates an requestID with the peer's id of the peer that has made that
	 * request. If already exists an peerID associate with the given requestId,
	 * nothing is done.
	 * 
	 * @param requestID The unique identifier for the request.
	 * @param peerID the unique identifier for the peer that makes the request.
	 */
	public synchronized void registerPeerIdForRequest( long requestID, ObjectID peerID ) {

		if ( !peerIdForRequestId.containsKey( requestID ) ) {
			LOG.debug( "Request ID " + requestID + " was registered  to Peer " + peerID.getName() );
			peerIdForRequestId.put( requestID, peerID );
		}
	}


	/**
	 * Removed a request from the
	 * 
	 * @param requestID
	 */
	public synchronized void removeRequest( long requestID ) {

		peerIdForRequestId.remove( requestID );
	}


	/**
	 * Recover the unique instance of this entity.
	 * 
	 * @return Single AccountingManager instance.
	 */
	public synchronized static AccountingManager getInstance() {

		if ( uniqueInstance == null ) {
			uniqueInstance = new AccountingManager();
		}
		return uniqueInstance;
	}


	/**
	 * <p>
	 * Calculates the new balance given the grid machine used time and the
	 * current balance of a consumer peer. Only positives balance are allowed.
	 * That is, if the new consumer balance will not be lower than 0 (zero). The
	 * new balance is obtained from currentConsumerBalance - usedTime /
	 * SimpleAccontant.FACTOR.
	 * </p>
	 * 
	 * @param usedTime The time a GridMachine was used.
	 * @param currentConsumerBalance The current balance of a peer
	 * @return The new balance of the peer.
	 * @throws IllegalArgumentException Thrown when the given used time is
	 *         negative.
	 */
	public double providerAccount( double usedTime, double currentConsumerBalance ) throws IllegalArgumentException {

		if ( usedTime < 0 ) {
			LOG.error( "Invalid used time received. <usedTime>=" + usedTime + "." );

			throw new IllegalArgumentException( "Invalid used time. It must be a non negative value." );
		}

		double newConsumerBalance = currentConsumerBalance - usedTime / FACTOR;

		/* ensures that will exists only positive balances */
		newConsumerBalance = ((newConsumerBalance > 0) ? newConsumerBalance : 0);

		LOG.debug( "The new value of the consumer balance is " + newConsumerBalance + "." );

		return newConsumerBalance;
	}


	/**
	 * <p>
	 * Calculates the new balance given the used time of a GridMachine and the
	 * current balance of a provider peer. The new balance is obtained from
	 * currentConsumerBalance + usedTime / SimpleAccontant.FACTOR.
	 * </p>
	 * 
	 * @param usedTime The time a GridMachine was used.
	 * @param currentProviderBalance The current balance of a peer
	 * @return The new balance of the peer.
	 * @throws IllegalArgumentException Thrown when the given used time is not
	 *         positive.
	 */
	public double consumerAccount( double usedTime, double currentProviderBalance ) throws IllegalArgumentException {

		if ( usedTime < 0 ) {
			LOG.error( "Invalid used time received. <usedTime>=" + usedTime + "." );

			throw new IllegalArgumentException( "Invalid used time. It must be non negative." );
		}

		double newProviderBalance = currentProviderBalance + usedTime / FACTOR;

		/* ensures that will exists only positive balances */
		newProviderBalance = ((newProviderBalance > 0) ? newProviderBalance : 0);

		LOG.debug( "The new value of the provider balance is " + newProviderBalance + "." );

		return newProviderBalance;
	}


	/**
	 * Reseting the know instance of AccountManager.
	 */
	public static void reset() {

		uniqueInstance = null;
	}
}
