/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.community.nof;

import java.io.File;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.PeerConfiguration;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.test.PeerFastTest;
import org.ourgrid.test.TestUtils;

public class AccountingManagerImplTest extends TestCase implements PeerFastTest {

	protected double newBalance;

	private AccountingManager accountingManager;

	protected ObjectID peerID1;

	protected ObjectID unknownPeer;

	protected GumID gumID;

	protected ObjectID peerID2;

	private TestUtils testUtils;

	private double negativeUsedTime;

	private double positiveUsedTime;

	private double positiveBalance;

	private double negativeBalance;

	private double zeroBalance;


	/**
	 * Constructor for AccountingManagerTest.
	 */
	public AccountingManagerImplTest() {

		unknownPeer = new ObjectID( "reallyUnknownPeer", new Long( 891233782 ), "rmi:/" );
		peerID1 = new ObjectID( "peerID-1", new Long( 12322 ), "rmi:/" );
		peerID2 = new ObjectID( "peerID-2", new Long( 1032 ), "rmi:/" );
		gumID = new GumID( "gridMachineName", "0", "peername", "0" );

		newBalance = 360;

		positiveUsedTime = 200;
		negativeUsedTime = -200;

		positiveBalance = 1200;
		negativeBalance = -200;
		zeroBalance = 0;
	}


	protected void setUp() throws Exception {

		Configuration.reset();
		Configuration.getInstance( Configuration.PEER );

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createOGProperties();

		File f = new File( Configuration.getInstance().getProperty( PeerConfiguration.PROP_RANKINGFILE ) );
		if ( f.exists() ) {
			f.delete();
		}

		accountingManager = AccountingManager.getInstance();
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		File f = new File( Configuration.getInstance().getProperty( PeerConfiguration.PROP_RANKINGFILE ) );
		if ( f.exists() ) {
			f.delete();
		}

		testUtils.reset();
		Configuration.reset();
	}


	public void testGetUnknownPeerBalance() {

		/*
		 * verifies if the balance of a unknown peer is equal to the default
		 * value for unknown peer
		 */
		assertEquals( " 1", AccountingManager.DEFAULT_UNKNOWN_PEER_BALANCE, accountingManager
				.getPeerBalance( unknownPeer ), 0 );
	}


	public void testStartAccouting() {

		/* starts the grid machine usage */
		accountingManager.startAccounting( gumID, peerID2 );
		accountingManager.finishAccounting( gumID, peerID2, true );
		assertTrue( "Is a known peer? ", accountingManager.isAKnownPeer( peerID2 ) );

	}


	public void testGetPeerBalance() {

		/*
		 * accounts a GridMachine usage of a peer and verifies if its balance is
		 * different from a balance of a unknown peer
		 */

		/* starts the GridMachine usage and sleep */
		accountingManager.startAccounting( gumID, peerID2 );

		// the peer balance must no change until a finishProviderGMAccounting()
		assertEquals( "1", AccountingManager.DEFAULT_UNKNOWN_PEER_BALANCE, accountingManager
				.getPeerBalance( unknownPeer ) );

		try {
			Thread.sleep( 1000 );
		} catch ( InterruptedException e ) {
			e.printStackTrace();
		}
		/* finish the GridMachine usage and verify if the account is as expected */
		double usedTime = accountingManager.finishAccounting( gumID, peerID2, true );

		assertTrue( usedTime > 0.0 );

	}


	public void testFinishProviderGMAccountingOfANotStartedAccounting() {

		/*
		 * attempting to finish a GridMachine usage of a GridMachine whose usage
		 * has not been started
		 */
		assertEquals( "1", 0.0, accountingManager.finishAccounting( gumID, peerID1, true ), 0 );
	}


	public void testGetPeerIdForRequest() {

		/*
		 * Tests the usage of the map RequestIdxPeerID
		 */
		ObjectID the_PeerID = new ObjectID( "the_PeerID", new Long( 2732 ), null );

		// if there's no peer id associate for that request return ""
		assertEquals( null, accountingManager.getPeerIdForRequest( 189787 ) );

		// associate the specified PeerId for the requestID
		accountingManager.registerPeerIdForRequest( 189787, the_PeerID );
		// assert that the association was done.
		assertEquals( the_PeerID, accountingManager.getPeerIdForRequest( 189787 ) );
		// try to associate an existent request for other PeerId.
		accountingManager.registerPeerIdForRequest( 189787, new ObjectID( "another_PeerID", new Long( 2342 ), null ) );
		// As the requestId was already associate to a PeerId, keep it.
		assertEquals( the_PeerID, accountingManager.getPeerIdForRequest( 189787 ) );

		accountingManager.removeRequest( 189787 );

		assertNull( accountingManager.getPeerIdForRequest( 189787 ) );

	}


	/**
	 * Tests the <code>providerAccount()</code> method.
	 */
	public void testProviderAccount() {

		/*
		 * ensures that the new balance is accounted as currentBalance -
		 * usedTime / SimpleAccountant.FACTOR
		 */
		assertEquals( "1", positiveBalance - positiveUsedTime / AccountingManager.FACTOR, accountingManager
				.providerAccount( positiveUsedTime, positiveBalance ), 0 );

		/* ensures that the new balance will not be lower than zero */
		assertEquals( "2", zeroBalance, accountingManager.providerAccount( positiveUsedTime, negativeBalance ), 0 );
		assertEquals( "3", zeroBalance, accountingManager.providerAccount( positiveUsedTime, zeroBalance ), 0 );

		/* verifies if a negative used time is an invalid argument */
		try {
			accountingManager.providerAccount( negativeUsedTime, positiveBalance );
			fail( "An IllegalArgumentException should be thrown." );
		} catch ( IllegalArgumentException ex ) {
			// this line has to be executed!
		}
	}


	/**
	 * Tests the <code>consumerAccount()</code> method.
	 */
	public void testConsumerAccount() {

		/*
		 * ensures that the new balance is accounted as currentBalance +
		 * usedTime / SimpleAccountant.FACTOR
		 */
		assertEquals( "1", positiveBalance + positiveUsedTime / AccountingManager.FACTOR, accountingManager
				.consumerAccount( positiveUsedTime, positiveBalance ), 0 );

		/* ensures that the new balance will not be lower than zero */
		assertEquals( "2", zeroBalance, accountingManager.consumerAccount( positiveUsedTime, negativeBalance ), 0 );
		assertEquals( "3", positiveUsedTime / AccountingManager.FACTOR, accountingManager.consumerAccount(
				positiveUsedTime, zeroBalance ), 0 );

		/* verifies if a negative used time is an invalid argument */
		try {
			accountingManager.consumerAccount( negativeUsedTime, positiveBalance );
			fail( "An IllegalArgumentException should be thrown." );
		} catch ( IllegalArgumentException ex ) {
			// this line has to be executed!
		}
	}

}
