/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.community.nof;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.ourgrid.common.id.ObjectID;
import org.ourgrid.peer.manager.request.RequestEntry;

/**
 * <p>
 * Chooses always the request with the greatest balance among the ones passed.
 * This <code>Accountant</code> strategy takes no other characteristics of the
 * consumers into account when choosing which request to fullfill.
 * </p>
 */
public class Arbitrator {

	/**
	 * Choose the request with less allocated gums
	 * 
	 * @param requests A collection of requests to be considered in the decision
	 * @return The request with the lowest number of allocated gums
	 */
	public static RequestEntry chooseRequestToReceiveGum( Collection<RequestEntry> requests ) {

		RequestEntry chosenRequest = null;

		if ( !requests.isEmpty() ) {
			chosenRequest = requests.iterator().next();
			for ( RequestEntry entry : requests )
				if ( entry.getGums().size() < chosenRequest.getGums().size() )
					chosenRequest = entry;
		}
		return chosenRequest;
	}


	/**
	 * Choose the request with most allocated gums
	 * 
	 * @param requests A collection of requests to be considered in the decision
	 * @return The request with the largest number of allocated gums
	 */
	public static RequestEntry chooseRequestToLoseGum( Collection<RequestEntry> requests ) {

		RequestEntry chosenRequest = null;

		if ( !requests.isEmpty() ) {
			chosenRequest = requests.iterator().next();
			for ( RequestEntry entry : requests )
				if ( entry.getGums().size() > chosenRequest.getGums().size() )
					chosenRequest = entry;
		}
		return chosenRequest;
	}


	/**
	 * Choose a request made by a community consumer that deserves to receive a
	 * gum
	 * 
	 * @param requests A collection of requests to be considered in the decision
	 * @return The request with the lowest number of gums, which belongs to the
	 *         consumer with the highest NoF balance
	 */
	public static RequestEntry chooseRemoteRequestToReceiveGum( Collection<RequestEntry> requests ) {

		RequestEntry choosenRequest = null;

		double entryBalance;
		double balance;

		ObjectID entryPeerID;
		ObjectID choosenPeerID = null;

		if ( !requests.isEmpty() ) {
			RequestEntry firstEntry = requests.iterator().next();
			choosenPeerID = firstEntry.getSpec().getObjectID();
			balance = AccountingManager.getInstance().getPeerBalance( choosenPeerID );

			// Looking for the peer with the highest balance
			for ( RequestEntry entry : requests ) {
				entryPeerID = entry.getSpec().getObjectID();
				entryBalance = AccountingManager.getInstance().getPeerBalance( entryPeerID );

				if ( entryBalance > balance ) {
					choosenPeerID = entryPeerID;
					balance = entryBalance;
				}
			}

			List<RequestEntry> peerRequests = new LinkedList<RequestEntry>();

			// Iterate again to filter the requests which belong to the peer
			// with highest balance
			for ( RequestEntry entry : requests ) {
				if ( entry.getSpec().getObjectID().equals( choosenPeerID ) )
					peerRequests.add( entry );
			}

			// Look for the request with the lowest number of grid machines
			choosenRequest = chooseRequestToReceiveGum( peerRequests );
		}

		return choosenRequest;
	}


	/**
	 * Choose a request made by a community consumer that should have a gum
	 * preempted
	 * 
	 * @param requests A collection of requests to be considered in the decision
	 * @return The request with the largest number of gums, which belongs to the
	 *         consumer with the lowest NoF balance
	 */
	public static RequestEntry chooseRemoteRequestToLoseGum( Collection<RequestEntry> requests ) {

		RequestEntry choosenRequest = null;

		if ( !requests.isEmpty() ) {
			ObjectID entryPeerID;
			ObjectID choosenPeerID = null;
			double entryBalance;
			double balance;

			RequestEntry firstEntry = requests.iterator().next();
			choosenPeerID = firstEntry.getSpec().getObjectID();
			balance = AccountingManager.getInstance().getPeerBalance( choosenPeerID );

			// Looking for the peer with the lowest balance
			for ( RequestEntry entry : requests ) {
				entryPeerID = entry.getSpec().getObjectID();
				entryBalance = AccountingManager.getInstance().getPeerBalance( entryPeerID );

				if ( entryBalance < balance ) {
					choosenPeerID = entryPeerID;
					balance = entryBalance;
				}
			}

			List<RequestEntry> peerRequests = new LinkedList<RequestEntry>();

			// Iterate again to filter the requests which belong to the peer
			// with highest balance
			for ( RequestEntry entry : requests ) {
				if ( entry.getSpec().getObjectID().equals( choosenPeerID ) )
					peerRequests.add( entry );
			}

			// Look for the request with the largest number of machines
			choosenRequest = chooseRequestToLoseGum( peerRequests );
		}

		return choosenRequest;
	}
}
