/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.community.nof;

import java.io.File;
import java.rmi.RemoteException;
import java.util.Collection;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.PeerConfiguration;
import org.ourgrid.common.exception.RequestAlreadyExistsException;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.common.spec.RequestSpec.RequestSource;
import org.ourgrid.peer.manager.FakeGumpClient;
import org.ourgrid.peer.manager.request.RequestEntry;
import org.ourgrid.peer.manager.request.RequestManager;
import org.ourgrid.test.PeerFastTest;
import org.ourgrid.test.TestUtils;

public class ArbitratorTest extends TestCase implements PeerFastTest {

	private GumID gumID;

	private ObjectID peer1ID;

	private ObjectID peer2ID;

	private ObjectID peer3ID;

	private AccountingManager accountingManager;

	/**
	 * Used to create requests.
	 */
	private RequestManager requestManager;

	private TestUtils testUtils;


	public ArbitratorTest() {

		this.peer1ID = new ObjectID( "peer1ID", 1l, null );
		this.peer2ID = new ObjectID( "peer2ID", 2l, null );
		this.peer3ID = new ObjectID( "peer3ID", 3l, null );

		this.gumID = new GumID( "gumID", "0", "peername", "0" );
	}


	protected void setUp() throws Exception {

		super.setUp();

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createOGProperties();

		Configuration.createInstance( Configuration.PEER );

		File rankingFile = new File( Configuration.getInstance().getProperty( PeerConfiguration.PROP_RANKINGFILE ) );

		rankingFile.delete();

		AccountingManager.reset();
		RequestManager.reset();

		this.accountingManager = AccountingManager.getInstance();
		this.requestManager = RequestManager.getInstance();
	}


	protected void tearDown() throws Exception {

		super.tearDown();

		Configuration.reset();

		testUtils.reset();

	}


	private Collection<RequestEntry> createRequestsCollection() throws RequestAlreadyExistsException, RemoteException {

		RequestSpec request1 = new RequestSpec( 100, "", 10, new FakeGumpClient( peer1ID ) );
		RequestSpec request2 = new RequestSpec( 200, "", 10, new FakeGumpClient( peer2ID ) );
		RequestSpec request3 = new RequestSpec( 300, "", 10, new FakeGumpClient( peer3ID ) );

		request1.setAsFromRemotePeer();
		request2.setAsFromRemotePeer();
		request3.setAsFromRemotePeer();
		this.requestManager.createRequestEntry( request1 );
		this.requestManager.createRequestEntry( request2 );
		this.requestManager.createRequestEntry( request3 );

		return requestManager
				.getRequestsThatNeedGums( requestManager.getAliveRequests( RequestSource.FROM_REMOTE_PEER ) );
	}


	private void setupConsumersRanking() throws InterruptedException {

		this.accountingManager.startAccounting( this.gumID, peer1ID );
		this.accountingManager.finishAccounting( this.gumID, peer1ID, false );

		this.accountingManager.startAccounting( this.gumID, peer2ID );
		Thread.sleep( 100 );
		this.accountingManager.finishAccounting( this.gumID, peer2ID, false );

		this.accountingManager.startAccounting( this.gumID, peer3ID );
		this.accountingManager.finishAccounting( this.gumID, peer3ID, false );
	}


	public void testChooseRequestWhereAllHaveTheSameRanking() throws RequestAlreadyExistsException, RemoteException {

		Collection<RequestEntry> requests = this.createRequestsCollection();
		// as the requesters ranking is not set up all of them have the same
		// ranking.

		// if all have the same ranking the first got is choosen
		RequestEntry choosenRequest = Arbitrator.chooseRemoteRequestToReceiveGum( requests );
		// get the first request
		RequestEntry firstRequest = requests.iterator().next();

		assertTrue( firstRequest == choosenRequest );
	}


	public void testChooseBestRankedRequest() throws Exception {

		Collection<RequestEntry> requests = this.createRequestsCollection();
		// associate the request id with peer id
		this.accountingManager.registerPeerIdForRequest( 100, this.peer1ID );
		this.accountingManager.registerPeerIdForRequest( 200, this.peer2ID );
		this.accountingManager.registerPeerIdForRequest( 300, this.peer3ID );
		// configure the consumers ranking
		this.setupConsumersRanking();

		// the best ranked peer must be choosen
		RequestEntry choosenRequest = Arbitrator.chooseRemoteRequestToReceiveGum( requests );

		RequestEntry secondRequest = null;

		for ( Object request : requests ) {
			RequestEntry aRequest = (RequestEntry) request;
			if ( aRequest.getSpec().getID() == 200 ) {
				secondRequest = aRequest;
			}
		}
		assertTrue( secondRequest == choosenRequest );
	}
}
