/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.community.obtainer;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.ourgrid.common.gump.EBGump;
import org.ourgrid.common.gump.EBGumpClient;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.peer.community.communicationlayer.CommunicationLayer;
import org.ourgrid.peer.community.communicationlayer.CommunicationLayerImpl;
import org.ourgrid.peer.community.nof.AccountingManager;
import org.ourgrid.peer.ebgum.EBGum;
import org.ourgrid.peer.ui.UIManager;
import org.ourgrid.peer.ui.PeerUIManager.Notifiables;

/**
 * This entity is responsible for, in this peer, ask the community for
 * <code>GridMachine</code>s. Also, At the beginning of the request process,
 * it will broadcast a wannaGuM message for a request, and when receive answers
 * (hereIsGuM) will pass a valid OurGridMachine abstraction to the PeerManager
 * and start the counters for the usage time for the respective GuM.
 */
public class EBCommunityObtainer implements EBGump, EBGumpClient {

	/**
	 * This map will store the pair machineName/ObjectID of the machine's owner
	 */
	private Map<GumID,ObjectID> machineOwners;

	/**
	 * Maps request ids to gums currently accounting for.
	 */
	private Map<Long,Set<GumID>> requestID2gumID;

	/**
	 * The layer use to communicate with other peers.
	 */
	private CommunicationLayer communicationLayer;

	/**
	 * The consumer that all the OurGuMs will be given.
	 */
	private EBGumpClient consumer;

	/**
	 * UIManager used to get failure detector.
	 */
	private UIManager uimanager;

	/**
	 * Logger to store log messages
	 */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( EBCommunityObtainer.class );


	public EBCommunityObtainer( EBGumpClient consumer, UIManager uimanager ) {

		this.consumer = consumer;
		this.machineOwners = new HashMap<GumID,ObjectID>();
		this.requestID2gumID = new HashMap<Long,Set<GumID>>();
		this.communicationLayer = CommunicationLayerImpl.getInstance();
		this.uimanager = uimanager;
	}


	protected EBCommunityObtainer( EBGumpClient consumer, CommunicationLayer communication, UIManager uimanager ) {

		this( consumer, uimanager );
		this.uimanager = uimanager;
		this.communicationLayer = communication;
	}


	/**
	 * @see org.ourgrid.common.gump.EBGump#requestGums(RequestSpec)
	 */
	public void requestGums( RequestSpec requestSpec ) {

		AccountingManager.getInstance().registerPeerIdForRequest( requestSpec.getID(), requestSpec.getObjectID() );

		this.communicationLayer.sendRequest( requestSpec );
	}


	/**
	 * @see org.ourgrid.common.gump.EBGump#finishRequestGums(long)
	 */
	public void finishRequestGums( long requestID ) {

		this.communicationLayer.finishRequest( requestID );

		/*
		 * Finishing accounting
		 */
		HashSet<ObjectID> ownerIDs = new HashSet<ObjectID>();
		Set<GumID> gumIDs = requestID2gumID.remove( requestID );
		if ( gumIDs != null ) {
			for ( GumID gumID : gumIDs ) {
				ObjectID ownerID = machineOwners.remove( gumID );
				if ( ownerID != null ) {
					ownerIDs.add( ownerID );
					AccountingManager.getInstance().finishAccounting( gumID, ownerID, false );
				}
			}
		}

		removePeersFromFD( ownerIDs );
		AccountingManager.getInstance().removeRequest( requestID );
	}


	/**
	 * @see org.ourgrid.common.gump.EBGump#disposeGum(long, GumID)
	 */
	public void disposeGum( long requestID, GumID gumID ) {

		ObjectID peerID = machineOwners.remove( gumID );
		if ( peerID != null ) {
			AccountingManager.getInstance().finishAccounting( gumID, peerID, false );

			ArrayList<ObjectID> ownerIDs = new ArrayList<ObjectID>();
			ownerIDs.add( peerID );
			removePeersFromFD( ownerIDs );

			this.communicationLayer.disposeOf( gumID, requestID, peerID );
		}

		Set<GumID> gumIDs = requestID2gumID.get( requestID );

		if ( gumIDs != null ) {
			gumIDs.remove( gumID );
		}
	}


	/**
	 * @see org.ourgrid.common.gump.EBGump#pauseRequestGums(long)
	 */
	public void pauseRequestGums( long requestID ) {

		this.communicationLayer.pauseRequest( requestID );
	}


	/**
	 * @see org.ourgrid.common.gump.EBGump#resumeRequestGums(long)
	 */
	public void resumeRequestGums( long requestID ) {

		this.communicationLayer.resumeRequest( requestID );
	}


	/**
	 * @see org.ourgrid.common.gump.EBGumpClient#hereIsGum(EBGum,
	 *      org.ourgrid.common.id.ObjectID, long)
	 */
	public void hereIsGum( EBGum receivedGum, ObjectID gumOwnerObjectID, long requestID ) {

		GumID gumID = receivedGum.getID();

		this.machineOwners.put( gumID, gumOwnerObjectID );
		/*
		 * Maps request to gums
		 */
		Set<GumID> gumIDs = requestID2gumID.get( requestID );
		if ( gumIDs == null ) {
			gumIDs = new HashSet<GumID>();
		}

		gumIDs.add( gumID );
		requestID2gumID.put( requestID, gumIDs );
		this.consumer.hereIsGum( receivedGum, gumOwnerObjectID, requestID );
		AccountingManager.getInstance().startAccounting( gumID, gumOwnerObjectID );

		try {
			uimanager.registerNewMonitorable( Notifiables.CONSUMER, gumOwnerObjectID );
			LOG.debug( "Added new monitorable: " + gumOwnerObjectID.getObjectURL() );
		} catch ( RemoteException exception ) {
			LOG.error( "Could not add new monitorable: " + gumOwnerObjectID.getObjectURL() + ": "
					+ exception.getMessage(), exception );
		}
	}


	/**
	 * @see org.ourgrid.common.gump.EBGump#unwantedGum(long,
	 *      org.ourgrid.common.id.GumID)
	 */
	public void unwantedGum( long requestID, GumID gumID ) {

		ObjectID gumOwnerID = this.machineOwners.remove( gumID );

		if ( gumOwnerID != null ) {
			AccountingManager.getInstance().finishAccounting( gumID, gumOwnerID, false );

			ArrayList<ObjectID> ownerIDs = new ArrayList<ObjectID>();
			ownerIDs.add( gumOwnerID );
			removePeersFromFD( ownerIDs );

			this.communicationLayer.unwantedGum( gumOwnerID, requestID, gumID );
		}

		Set<GumID> gumIDs = requestID2gumID.get( requestID );

		if ( gumIDs != null ) {
			gumIDs.remove( gumID );
		}
	}


	/**
	 * Unregistering peers from failure detector if we have no more gums from
	 * these.
	 * 
	 * @param peerIDs IDs to verify.
	 */
	private void removePeersFromFD( Collection<ObjectID> peerIDs ) {

		for ( ObjectID peerID : peerIDs ) {
			if ( !machineOwners.containsValue( peerID ) ) {
				try {
					uimanager.unregisterNewMonitorable( Notifiables.CONSUMER, peerID );
				} catch ( RemoteException e ) {
				}
				LOG.debug( "Peer " + peerID.getObjectURL() + " was successfully unregistered from failure detector." );
			}
		}
	}
}
