/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.community.obtainer;

import java.net.URISyntaxException;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.gump.EBGumpClient;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.peer.community.communicationlayer.CommunicationLayer;
import org.ourgrid.peer.community.nof.AccountingManager;
import org.ourgrid.peer.ebgum.EBGum;
import org.ourgrid.peer.ebgum.FakeEBGum;
import org.ourgrid.peer.manager.FakeGumpClient;
import org.ourgrid.peer.ui.UIManager;
import org.ourgrid.peer.ui.PeerUIManager.Notifiables;
import org.ourgrid.test.PeerFastTest;
import org.ourgrid.test.TestUtils;

public class EBCommunityObtainerTest extends TestCase implements PeerFastTest {

	private EBCommunityObtainer ebCommunityObtainer;

	protected EBGumpClient mockConsumer;

	protected CommunicationLayer mockCommunicationLayer;

	private long reqID;

	private String requeriments;

	private EBGum ebGum;

	private TestUtils testUtils;

	private GumSpec fakeGuMSpec;

	private UIManager mockUIManager;


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();
		Configuration.getInstance( Configuration.PEER );

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createOGProperties();

		this.reqID = 10;
		this.requeriments = "os = linux";

		Map<String,String> attrMap = new HashMap<String,String>();
		attrMap.put( GumSpec.ATT_NAME, "ourGuMTest" );

		fakeGuMSpec = new GumSpec( attrMap );

		ebGum = new FakeEBGum( fakeGuMSpec );

		// creates mocks
		this.mockConsumer = EasyMock.createMock( EBGumpClient.class );
		this.mockCommunicationLayer = EasyMock.createMock( CommunicationLayer.class );
		this.mockUIManager = EasyMock.createMock( UIManager.class );

		// creates EBCommunityObtainer
		this.ebCommunityObtainer = new EBCommunityObtainer( mockConsumer, mockCommunicationLayer, mockUIManager );
		AccountingManager.reset();
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		testUtils.reset();
		Configuration.reset();
		AccountingManager.reset();
	}


	public void testWannaGuMs() throws RemoteException {

		RequestSpec request1 = new RequestSpec( reqID, requeriments, 3, new FakeGumpClient( "rmi://localhost.test1:0" ) );
		this.mockCommunicationLayer.sendRequest( request1 );
		EasyMock.replay( mockCommunicationLayer );

		this.getEBCommunityObtainer().requestGums( request1 );
		EasyMock.verify( mockCommunicationLayer );

	}


	public void testNoMoreGuMs() throws Exception {

		this.mockCommunicationLayer.finishRequest( reqID );
		EasyMock.replay( mockCommunicationLayer );

		this.getEBCommunityObtainer().finishRequestGums( reqID );
		EasyMock.verify( mockCommunicationLayer );
	}


	public void testSuspendWannaGuMs() throws RemoteException {

		this.mockCommunicationLayer.pauseRequest( reqID );
		EasyMock.replay( mockCommunicationLayer );

		this.getEBCommunityObtainer().pauseRequestGums( reqID );
		EasyMock.verify( mockCommunicationLayer );
	}


	public void testResumeWannaGuMs() throws RemoteException {

		this.mockCommunicationLayer.resumeRequest( reqID );
		EasyMock.replay( mockCommunicationLayer );

		this.getEBCommunityObtainer().resumeRequestGums( reqID );
		EasyMock.verify( mockCommunicationLayer );
	}


	public void testHereIsGuMAndDisposeOf() throws InterruptedException, RemoteException {

		ObjectID peerID = new ObjectID( "PeerID", new Long( 123 ), "rmi:/" );

		this.mockConsumer.hereIsGum( ebGum, peerID, reqID );
		EasyMock.expectLastCall().once();
		EasyMock.replay( mockConsumer );

		this.mockCommunicationLayer.disposeOf( ebGum.getID(), reqID, peerID );
		EasyMock.expectLastCall().once();
		EasyMock.replay( mockCommunicationLayer );

		double initialBalance = AccountingManager.getInstance().getPeerBalance( peerID );
		this.getEBCommunityObtainer().hereIsGum( ebGum, peerID, reqID );

		Thread.sleep( 500 );
		this.getEBCommunityObtainer().disposeGum( reqID, ebGum.getID() );
		double finalAccount = AccountingManager.getInstance().getPeerBalance( peerID );
		assertTrue( finalAccount > initialBalance );
		EasyMock.verify( mockCommunicationLayer );
		EasyMock.verify( mockConsumer );

		// we are not intended to verify the mockProvider
	}


	public void testHereIsGuMAndUnwanted() throws InterruptedException, RemoteException {

		ObjectID peerID = new ObjectID( "PeerID", new Long( 123 ), "rmi:/" );

		this.mockConsumer.hereIsGum( ebGum, peerID, reqID );
		EasyMock.expectLastCall().once();
		EasyMock.replay( mockConsumer );

		this.mockCommunicationLayer.unwantedGum( peerID, reqID, ebGum.getID() );
		EasyMock.expectLastCall().once();
		EasyMock.replay( mockCommunicationLayer );

		double initialBalance = AccountingManager.getInstance().getPeerBalance( peerID );
		this.getEBCommunityObtainer().hereIsGum( ebGum, peerID, reqID );

		Thread.sleep( 500 );
		this.getEBCommunityObtainer().unwantedGum( reqID, ebGum.getID() );
		double finalAccount = AccountingManager.getInstance().getPeerBalance( peerID );
		assertTrue( finalAccount > initialBalance );
		EasyMock.verify( mockCommunicationLayer );
		EasyMock.verify( mockConsumer );
	}


	public void testHereIsGuMAndFinishRequest() throws InterruptedException, RemoteException {

		ObjectID peerID = new ObjectID( "PeerID", new Long( 123 ), "rmi:/" );

		Map<String,String> attrMap = new HashMap<String,String>();
		attrMap.put( GumSpec.ATT_NAME, "gum1" );
		GumSpec spec1 = new GumSpec( attrMap );
		EBGum ebGum1 = new FakeEBGum( spec1 );

		Map<String,String> attrMap2 = new HashMap<String,String>();
		attrMap2.put( GumSpec.ATT_NAME, "gum2" );
		GumSpec spec2 = new GumSpec( attrMap2 );
		EBGum ebGum2 = new FakeEBGum( spec2 );

		Map<String,String> attrMap3 = new HashMap<String,String>();
		attrMap3.put( GumSpec.ATT_NAME, "gum3" );
		GumSpec spec3 = new GumSpec( attrMap3 );
		EBGum ebGum3 = new FakeEBGum( spec3 );

		this.mockConsumer.hereIsGum( ebGum1, peerID, reqID );
		EasyMock.expectLastCall().once();

		this.mockConsumer.hereIsGum( ebGum2, peerID, reqID );
		EasyMock.expectLastCall().once();

		this.mockConsumer.hereIsGum( ebGum3, peerID, reqID );
		EasyMock.expectLastCall().once();
		EasyMock.replay( mockConsumer );

		this.mockCommunicationLayer.finishRequest( reqID );
		EasyMock.expectLastCall().once();
		EasyMock.replay( mockCommunicationLayer );

		double initialBalance = AccountingManager.getInstance().getPeerBalance( peerID );
		this.getEBCommunityObtainer().hereIsGum( ebGum1, peerID, reqID );
		this.getEBCommunityObtainer().hereIsGum( ebGum2, peerID, reqID );
		this.getEBCommunityObtainer().hereIsGum( ebGum3, peerID, reqID );

		Thread.sleep( 500 );
		this.getEBCommunityObtainer().finishRequestGums( reqID );
		double finalAccount = AccountingManager.getInstance().getPeerBalance( peerID );
		assertTrue( finalAccount > initialBalance );
		EasyMock.verify( mockCommunicationLayer );
		EasyMock.verify( mockConsumer );
	}


	public void testPeerRemovalFromFD() throws InterruptedException, RemoteException, URISyntaxException {

		long reqID2 = reqID + 1;
		long reqID3 = reqID + 1;

		ObjectID peerID1 = new ObjectID( "PeerID", new Long( 123 ), "rmi://localhost:9090/Peer" );
		ObjectID peerID2 = new ObjectID( "PeerID2", new Long( 1234 ), "rmi://localhost2:9092/Peer" );
		ObjectID peerID3 = new ObjectID( "PeerID3", new Long( 12345 ), "rmi://localhost3:9093/Peer" );

		Map<String,String> attrMap = new HashMap<String,String>();
		attrMap.put( GumSpec.ATT_NAME, "gum1" );
		GumSpec spec1 = new GumSpec( attrMap );
		EBGum ebGum1 = new FakeEBGum( spec1 );

		Map<String,String> attrMap2 = new HashMap<String,String>();
		attrMap2.put( GumSpec.ATT_NAME, "gum2" );
		GumSpec spec2 = new GumSpec( attrMap2 );
		EBGum ebGum2 = new FakeEBGum( spec2 );

		Map<String,String> attrMap3 = new HashMap<String,String>();
		attrMap3.put( GumSpec.ATT_NAME, "gum3" );
		GumSpec spec3 = new GumSpec( attrMap3 );
		EBGum ebGum3 = new FakeEBGum( spec3 );

		Map<String,String> attrMap4 = new HashMap<String,String>();
		attrMap4.put( GumSpec.ATT_NAME, "gum4" );
		GumSpec spec4 = new GumSpec( attrMap4 );
		EBGum ebGum4 = new FakeEBGum( spec4 );

		Map<String,String> attrMap5 = new HashMap<String,String>();
		attrMap5.put( GumSpec.ATT_NAME, "gum5" );
		GumSpec spec5 = new GumSpec( attrMap5 );
		EBGum ebGum5 = new FakeEBGum( spec5 );

		Map<String,String> attrMap6 = new HashMap<String,String>();
		attrMap6.put( GumSpec.ATT_NAME, "gum6" );
		GumSpec spec6 = new GumSpec( attrMap6 );
		EBGum ebGum6 = new FakeEBGum( spec6 );

		this.mockUIManager.registerNewMonitorable( Notifiables.CONSUMER, peerID1 );
		EasyMock.expectLastCall().times( 2 );
		this.mockUIManager.registerNewMonitorable( Notifiables.CONSUMER, peerID2 );
		EasyMock.expectLastCall().times( 2 );
		this.mockUIManager.registerNewMonitorable( Notifiables.CONSUMER, peerID3 );
		EasyMock.expectLastCall().times( 2 );
		EasyMock.replay( mockUIManager );

		// The dispose and unwanted call should not cause a unregister
		this.getEBCommunityObtainer().hereIsGum( ebGum1, peerID1, reqID );
		this.getEBCommunityObtainer().hereIsGum( ebGum2, peerID1, reqID );
		this.getEBCommunityObtainer().hereIsGum( ebGum3, peerID2, reqID2 );
		this.getEBCommunityObtainer().hereIsGum( ebGum4, peerID2, reqID2 );
		this.getEBCommunityObtainer().hereIsGum( ebGum5, peerID3, reqID3 );
		this.getEBCommunityObtainer().hereIsGum( ebGum6, peerID3, reqID3 );
		this.getEBCommunityObtainer().disposeGum( reqID, ebGum1.getID() );
		this.getEBCommunityObtainer().unwantedGum( reqID3, ebGum5.getID() );
		EasyMock.verify( mockUIManager );

		// The next dispose and unwanted call SHOULD cause an unregister.
		EasyMock.reset( mockUIManager );
		this.mockUIManager.unregisterNewMonitorable( Notifiables.CONSUMER, (peerID1) );
		EasyMock.expectLastCall().once();
		this.mockUIManager.unregisterNewMonitorable( Notifiables.CONSUMER, (peerID3) );
		EasyMock.expectLastCall().once();
		EasyMock.replay( mockUIManager );

		this.getEBCommunityObtainer().disposeGum( reqID, ebGum2.getID() );
		this.getEBCommunityObtainer().unwantedGum( reqID3, ebGum6.getID() );
		EasyMock.verify( mockUIManager );

		// Finish request calls should cause an unregister.
		EasyMock.reset( mockUIManager );
		this.mockUIManager.unregisterNewMonitorable( Notifiables.CONSUMER, (peerID2) );
		EasyMock.expectLastCall().once();
		EasyMock.replay( mockUIManager );

		this.ebCommunityObtainer.finishRequestGums( reqID2 );
		EasyMock.verify( mockUIManager );
	}


	public void testPeerRemovalFromFD2() throws InterruptedException, RemoteException, URISyntaxException {

		long reqID2 = reqID + 1;

		ObjectID peerID1 = new ObjectID( "PeerID", new Long( 123 ), "rmi://localhost:9090/Peer" );

		Map<String,String> attrMap = new HashMap<String,String>();
		attrMap.put( GumSpec.ATT_NAME, "gum1" );
		GumSpec spec1 = new GumSpec( attrMap );
		EBGum ebGum1 = new FakeEBGum( spec1 );

		Map<String,String> attrMap2 = new HashMap<String,String>();
		attrMap2.put( GumSpec.ATT_NAME, "gum2" );
		GumSpec spec2 = new GumSpec( attrMap2 );
		EBGum ebGum2 = new FakeEBGum( spec2 );

		Map<String,String> attrMap3 = new HashMap<String,String>();
		attrMap3.put( GumSpec.ATT_NAME, "gum3" );
		GumSpec spec3 = new GumSpec( attrMap3 );
		EBGum ebGum3 = new FakeEBGum( spec3 );

		Map<String,String> attrMap4 = new HashMap<String,String>();
		attrMap4.put( GumSpec.ATT_NAME, "gum4" );
		GumSpec spec4 = new GumSpec( attrMap4 );
		EBGum ebGum4 = new FakeEBGum( spec4 );

		this.mockUIManager.registerNewMonitorable( Notifiables.CONSUMER, peerID1 );
		EasyMock.expectLastCall().times( 4 );
		EasyMock.replay( mockUIManager );

		// The finishrequest call should not cause a unregister
		this.getEBCommunityObtainer().hereIsGum( ebGum1, peerID1, reqID );
		this.getEBCommunityObtainer().hereIsGum( ebGum2, peerID1, reqID );
		this.getEBCommunityObtainer().hereIsGum( ebGum3, peerID1, reqID2 );
		this.getEBCommunityObtainer().hereIsGum( ebGum4, peerID1, reqID2 );
		this.ebCommunityObtainer.finishRequestGums( reqID2 );
		EasyMock.verify( mockUIManager );

		// The next dispose call should not cause an unregister.
		EasyMock.reset( mockUIManager );
		EasyMock.replay( mockUIManager );

		this.getEBCommunityObtainer().disposeGum( reqID, ebGum1.getID() );
		EasyMock.verify( mockUIManager );

		// This dispode call should cause an unregister because its the last gum
		// from peer.
		EasyMock.reset( mockUIManager );
		this.mockUIManager.unregisterNewMonitorable( Notifiables.CONSUMER, (peerID1) );
		EasyMock.expectLastCall().once();
		EasyMock.replay( mockUIManager );

		this.getEBCommunityObtainer().disposeGum( reqID, ebGum2.getID() );
		EasyMock.verify( mockUIManager );
	}


	/**
	 * @return
	 */
	protected EBCommunityObtainer getEBCommunityObtainer() {

		return this.ebCommunityObtainer;
	}

}
