/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.ebgum;

import java.io.File;
import java.util.Map;

import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.InvalidationResponseEvent;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.event.ShutdownEventEngine;
import org.ourgrid.common.event.ShutdownRequestEvent;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.peer.manager.PeerEventEngine;
import org.ourgrid.peer.manager.gummanager.events.GumChangeFilePermissionsEvent;
import org.ourgrid.peer.manager.gummanager.events.GumCreatePlaypenEvent;
import org.ourgrid.peer.manager.gummanager.events.GumCreateStorageDirectoryEvent;
import org.ourgrid.peer.manager.gummanager.events.GumFileExistsEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetFileEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetRemoteFileDigestEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetRemoteFileLastModificationEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetRemoteFileLengthEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetResultEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetStatusEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetStorageDirectoryEvent;
import org.ourgrid.peer.manager.gummanager.events.GumPutFileEvent;
import org.ourgrid.peer.manager.gummanager.events.GumRemoteExecuteEvent;
import org.ourgrid.peer.manager.gummanager.events.GumStartReplicaEvent;
import org.ourgrid.peer.manager.gummanager.events.GumStopReplicaEvent;

/**
 * This Facade is used as the only way to communicate with the
 * <code>EBGridMachine</code>. All methods are called on this facade, and
 * then it generates the respective event for each method call and put them into
 * the <code>EventQueue</code>, so they will be processed by the
 * <code>EBGridMachine</code>
 */
public class EBGumFacade implements EBGum {

	private ShutdownEventEngine eventProcessor;

	protected EBGum ebGridMachine;

	/** Logger to store log messages. */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( EBGumFacade.class );


	/**
	 * Default constructor that receives the <code>EBGridMachine</code> that
	 * will receive the events created by this facade
	 * 
	 * @param ebGumImpl The <code>EBGridMachine</code> that will be called
	 *        through this facade
	 */
	public EBGumFacade( EBGumImpl ebGumImpl ) {

		this.eventProcessor = new PeerEventEngine( "EBGumFacade" );
		this.ebGridMachine = ebGumImpl;
	}


	/**
	 * Used only for tests.
	 * 
	 * @param engine The fake engine to be used.
	 */
	public EBGumFacade( EBGum gum, ShutdownEventEngine engine ) {

		this.ebGridMachine = gum;
		this.eventProcessor = engine;

	}


	public void start() {

		if ( !eventProcessor.isAlive() ) {
			this.eventProcessor.startProcessing();
		}
	}


	public EventQueue<ShutdownResponseEvent> stop() {

		EventQueue<ShutdownResponseEvent> responseQueue = new EventQueue<ShutdownResponseEvent>();
		if ( eventProcessor.isAlive() ) {
			eventProcessor.putEvent( new ShutdownRequestEvent( responseQueue, eventProcessor ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new ShutdownResponseEvent() );
		}
		return responseQueue;
	}


	public boolean isStarted() {

		return this.eventProcessor.isAlive();
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#remoteExecute(EventQueue, String,
	 *      String, Map)
	 */
	public void remoteExecute( EventQueue<ResponseEvent> responseQueue, String remoteDir, String command,
								Map<String,String> environmentVars ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumRemoteExecuteEvent( ebGridMachine, responseQueue, remoteDir, command,
					environmentVars ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#putFile(EventQueue, String, byte[],
	 *      boolean)
	 */
	public void putFile( EventQueue<ResponseEvent> responseQueue, String remoteFile, byte[ ] data, boolean firstSlice ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumPutFileEvent( responseQueue, ebGridMachine, remoteFile, data,
					firstSlice ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getFile(EventQueue, String, long, int)
	 */
	public void getFile( EventQueue<ResponseEvent> responseQueue, String filePath, long filePos, int size ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumGetFileEvent( responseQueue, ebGridMachine, filePath, filePos, size ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getID()
	 */
	public GumID getID() {

		return this.ebGridMachine.getID();
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getResult(EventQueue, ExecutorHandle)
	 */
	public void getResult( EventQueue<ResponseEvent> responseQueue, ExecutorHandle handle ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumGetResultEvent( ebGridMachine, responseQueue, handle ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#fileExists(EventQueue, File)
	 */
	public void fileExists( EventQueue<ResponseEvent> responseQueue, File file ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumFileExistsEvent( ebGridMachine, responseQueue, file ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#changeFilePermissions(EventQueue, File,
	 *      int)
	 */
	public void changeFilePermissions( EventQueue<ResponseEvent> responseQueue, File file, int permissions ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumChangeFilePermissionsEvent( responseQueue, ebGridMachine, file,
					permissions ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getRemoteFileLength(EventQueue, File)
	 */
	public void getRemoteFileLength( EventQueue<ResponseEvent> responseQueue, File file ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumGetRemoteFileLengthEvent( ebGridMachine, responseQueue, file ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getRemoteFileLastModification(EventQueue,
	 *      File, boolean)
	 */
	public void getRemoteFileLastModification( EventQueue<ResponseEvent> responseQueue, File remoteFile,
												boolean normalized ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumGetRemoteFileLastModificationEvent( ebGridMachine, responseQueue,
					remoteFile, normalized ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#createPlaypen(EventQueue, int)
	 */
	public void createPlaypen( EventQueue<ResponseEvent> responseQueue, int size ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumCreatePlaypenEvent( ebGridMachine, responseQueue, size ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#createStorageDirectory(org.ourgrid.common.event.EventQueue)
	 */
	public void createStorageDirectory( EventQueue<ResponseEvent> responseQueue ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumCreateStorageDirectoryEvent( ebGridMachine, responseQueue ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getStorageDirectory(EventQueue)
	 */
	public void getStorageDirectory( EventQueue<ResponseEvent> responseQueue ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumGetStorageDirectoryEvent( ebGridMachine, responseQueue ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getStatus(EventQueue)
	 */
	public void getStatus( EventQueue<ResponseEvent> responseQueue ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumGetStatusEvent( ebGridMachine, responseQueue ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getGumSpec()
	 */
	public GumSpec getGumSpec() {

		return ebGridMachine.getGumSpec();
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getRemoteFileDigest(EventQueue, File)
	 */
	public void getRemoteFileDigest( EventQueue<ResponseEvent> responseQueue, File remoteFile ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumGetRemoteFileDigestEvent( ebGridMachine, responseQueue, remoteFile ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#startReplica(EventQueue)
	 */
	public void startReplica( EventQueue<ResponseEvent> responseQueue ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumStartReplicaEvent( ebGridMachine, responseQueue ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#stopReplica(EventQueue)
	 */
	public void stopReplica( EventQueue<ResponseEvent> responseQueue ) {

		if ( eventProcessor.isAlive() ) {
			this.eventProcessor.putEvent( new GumStopReplicaEvent( ebGridMachine, responseQueue ) );
		} else {
			LOG.debug( "EventProcessor is NOT ALIVE" );
			responseQueue.put( new InvalidationResponseEvent() );
		}
	}
}
