/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.ebgum;

import java.io.File;
import java.rmi.RemoteException;
import java.util.Map;

import org.ourgrid.common.event.DummyEvent;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.common.event.ThrowableEvent;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.peer.manager.gummanager.events.GumCreatePlaypenResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumCreateStorageDirectoryResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumFileExistsResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetFileResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetRemoteFileDigestResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetRemoteFileLastModificationResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetRemoteFileLengthResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetResultResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetStatusResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetStorageDirectoryResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumRemoteExecuteResponseEvent;

public class EBGumImpl implements EBGum {

	/** The <code>GridMachine</code> to which the methods will be forwarded. */
	protected Gum gum;

	private GumSpec gumSpec;

	private GumID gumID;


	public EBGumImpl( Gum gum, GumSpec gumSpec ) {

		this.gum = gum;
		this.gumSpec = gumSpec;
		this.gumID = gumSpec.getGumID();
	}


	/**
	 * This constructor should only be used by tests.
	 */
	protected EBGumImpl( GumSpec gumSpec ) {

		this.gumSpec = gumSpec;
		this.gumID = gumSpec.getGumID();
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#remoteExecute(org.ourgrid.common.event.EventQueue,
	 *      java.lang.String, java.lang.String, java.util.Map)
	 */
	public void remoteExecute( EventQueue<ResponseEvent> responseQueue, String remoteDir, String command,
								Map<String,String> environmentVars ) {

		try {

			ExecutorHandle remoteExecute = this.gum.remoteExecute( remoteDir, command, environmentVars );

			GumRemoteExecuteResponseEvent responseEvent = new GumRemoteExecuteResponseEvent( remoteExecute );

			responseQueue.put( responseEvent );

		} catch ( Exception e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#putFile(EventQueue, String, byte[],
	 *      boolean)
	 */
	public void putFile( EventQueue<ResponseEvent> responseQueue, String remoteFile, byte[ ] data, boolean firstSlice ) {

		try {
			this.gum.putFile( remoteFile, data, firstSlice );
			responseQueue.put( new DummyEvent() );
		} catch ( Exception e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getFile(org.ourgrid.common.event.EventQueue,
	 *      java.lang.String, long, int)
	 */
	public void getFile( EventQueue<ResponseEvent> responseQueue, String filePath, long filePos, int size ) {

		try {

			byte[ ] file = this.gum.getFile( filePath, filePos, size );

			GumGetFileResponseEvent responseEvent = new GumGetFileResponseEvent( file );

			responseQueue.put( responseEvent );
		} catch ( Exception e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getID()
	 */
	public GumID getID() {

		return this.gumID;
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getResult(EventQueue, ExecutorHandle)
	 */
	public void getResult( EventQueue<ResponseEvent> responseQueue, ExecutorHandle handle ) {

		try {
			GumGetResultResponseEvent responseEvent = new GumGetResultResponseEvent( gum, handle );
			responseQueue.put( responseEvent );
		} catch ( Exception e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#fileExists(org.ourgrid.common.event.EventQueue,
	 *      java.io.File)
	 */
	public void fileExists( EventQueue<ResponseEvent> responseQueue, File file ) {

		try {
			boolean fileExists = this.gum.fileExists( file );
			GumFileExistsResponseEvent responseEvent = new GumFileExistsResponseEvent( fileExists );

			responseQueue.put( responseEvent );
		} catch ( Exception e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#changeFilePermissions(EventQueue, File,
	 *      int)
	 */
	public void changeFilePermissions( EventQueue<ResponseEvent> responseQueue, File file, int permissions ) {

		try {
			this.gum.changeFilePermissions( file, permissions );
			responseQueue.put( new DummyEvent() );
		} catch ( Exception e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getRemoteFileLength(org.ourgrid.common.event.EventQueue,
	 *      java.io.File)
	 */
	public void getRemoteFileLength( EventQueue<ResponseEvent> responseQueue, File file ) {

		try {
			long remoteFileLength = this.gum.getRemoteFileLength( file );

			GumGetRemoteFileLengthResponseEvent responseEvent = new GumGetRemoteFileLengthResponseEvent(
					remoteFileLength );
			responseQueue.put( responseEvent );
		} catch ( Exception e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getRemoteFileLastModification(org.ourgrid.common.event.EventQueue,
	 *      java.io.File, boolean)
	 */
	public void getRemoteFileLastModification( EventQueue<ResponseEvent> responseQueue, File remoteFile,
												boolean normalized ) {

		try {
			long remoteFileLastModification = this.gum.getRemoteFileLastModification( remoteFile, normalized );

			GumGetRemoteFileLastModificationResponseEvent responseEvent = new GumGetRemoteFileLastModificationResponseEvent(
					remoteFileLastModification );

			responseQueue.put( responseEvent );
		} catch ( Exception e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#createPlaypen(org.ourgrid.common.event.EventQueue,
	 *      int)
	 */
	public void createPlaypen( EventQueue<ResponseEvent> responseQueue, int size ) {

		try {

			String createPlaypen = this.gum.createPlaypen( size );

			GumCreatePlaypenResponseEvent responseEvent = new GumCreatePlaypenResponseEvent( createPlaypen );

			responseQueue.put( responseEvent );
		} catch ( Exception e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#createStorageDirectory(org.ourgrid.common.event.EventQueue)
	 */
	public void createStorageDirectory( EventQueue<ResponseEvent> responseQueue ) {

		try {
			String createStorageDirectory = this.gum.createStorageDirectory();

			GumCreateStorageDirectoryResponseEvent responseEvent = new GumCreateStorageDirectoryResponseEvent(
					createStorageDirectory );

			responseQueue.put( responseEvent );
		} catch ( Exception e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getStorageDirectory(org.ourgrid.common.event.EventQueue)
	 */
	public void getStorageDirectory( EventQueue<ResponseEvent> responseQueue ) {

		try {
			String storageDirectory = this.gum.getStorageDirectory();

			GumGetStorageDirectoryResponseEvent responseEvent = new GumGetStorageDirectoryResponseEvent(
					storageDirectory );

			responseQueue.put( responseEvent );
		} catch ( RemoteException e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getStatus(org.ourgrid.common.event.EventQueue)
	 */
	public void getStatus( EventQueue<ResponseEvent> responseQueue ) {

		try {

			GumStatus status = this.gum.getStatus();

			GumGetStatusResponseEvent responseEvent = new GumGetStatusResponseEvent( status );

			responseQueue.put( responseEvent );
		} catch ( RemoteException e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getGumSpec()
	 */
	public GumSpec getGumSpec() {

		return this.gumSpec;
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#getRemoteFileDigest(org.ourgrid.common.event.EventQueue,
	 *      java.io.File)
	 */
	public void getRemoteFileDigest( EventQueue<ResponseEvent> responseQueue, File remoteFile ) {

		try {
			String remoteFileDigest = this.gum.getRemoteFileDigest( remoteFile );

			GumGetRemoteFileDigestResponseEvent responseEvent = new GumGetRemoteFileDigestResponseEvent(
					remoteFileDigest );

			responseQueue.put( responseEvent );
		} catch ( Exception e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}

	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#startReplica(org.ourgrid.common.event.EventQueue)
	 */
	public void startReplica( EventQueue<ResponseEvent> responseQueue ) {

		try {
			this.gum.startReplica();
			responseQueue.put( new DummyEvent() );
		} catch ( RemoteException e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}
	}


	/**
	 * @see org.ourgrid.peer.ebgum.EBGum#stopReplica(org.ourgrid.common.event.EventQueue)
	 */
	public void stopReplica( EventQueue<ResponseEvent> responseQueue ) {

		try {
			this.gum.stopReplica();
			responseQueue.put( new DummyEvent() );
		} catch ( RemoteException e ) {
			ThrowableEvent throwableEvent = new ThrowableEvent( e );
			responseQueue.put( throwableEvent );
		}
	}


	public EventQueue<ShutdownResponseEvent> stop() {

		return null;
	}


	public void start() {

	}
}
