/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.ebgum;

import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.exception.UnableToCreatePlaypenException;
import org.ourgrid.common.exception.UnableToCreateStorageException;
import org.ourgrid.common.exception.UnableToDigestFileException;
import org.ourgrid.common.executor.ExecutorException;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.ExecutorResult;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.peer.manager.gummanager.events.GumCreatePlaypenResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumCreateStorageDirectoryResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumFileExistsResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetFileResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetRemoteFileDigestResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetRemoteFileLastModificationResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetRemoteFileLengthResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetResultResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetStatusResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumGetStorageDirectoryResponseEvent;
import org.ourgrid.peer.manager.gummanager.events.GumRemoteExecuteResponseEvent;
import org.ourgrid.test.GumPreCommitTest;

public class EBGumTest extends TestCase implements GumPreCommitTest {

	/* The EBGridMachine to be tested */
	EBGumImpl ebGum;

	Gum mockGuM;

	private Map specAttributesMap;

	private String specAttributeName;

	private GumSpec spec;

	private GumID gumID;


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();

		spec = new GumSpec();
		specAttributeName = "FakeGridMachineClient";

		spec.putAttribute( GumSpec.ATT_NAME, specAttributeName );
		spec.putAttribute( GumSpec.ATT_OS, "Linux" );

		specAttributesMap = spec.getAttributes();

		gumID = new GumID( spec );

		mockGuM = EasyMock.createMock( Gum.class );

		EasyMock.replay( mockGuM );
		this.ebGum = new EBGumImpl( mockGuM, spec );

		EasyMock.reset( mockGuM );
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
	}


	public void testRemoteExecuteEventQueueStringStringMap() throws RemoteException, ExecutorException {

		EventQueue<ResponseEvent> responseQueue = new EventQueue<ResponseEvent>();
		HashMap<String,String> testHashMap = new HashMap<String,String>();

		ExecutorHandle mockExecutorHandle = EasyMock.createMock( ExecutorHandle.class );

		EasyMock.expect( mockGuM.remoteExecute( "remoteDir", "command", testHashMap ) ).andReturn( mockExecutorHandle )
				.once();

		EasyMock.replay( mockGuM );

		this.ebGum.remoteExecute( responseQueue, "remoteDir", "command", testHashMap );
		ResponseEvent re = responseQueue.unblockingRemove();

		assertTrue( re instanceof GumRemoteExecuteResponseEvent );
		assertEquals( re.getResponse(), mockExecutorHandle );

		EasyMock.verify( mockGuM );
	}


	public void testPutFile() throws RemoteException, IOException {

		EventQueue<ResponseEvent> responseQueue = new EventQueue<ResponseEvent>();
		byte[ ] testByteArray = new byte[ 0 ];

		mockGuM.putFile( "", testByteArray, false );

		EasyMock.expectLastCall().once();
		EasyMock.replay( mockGuM );

		this.ebGum.putFile( responseQueue, "", testByteArray, false );

		EasyMock.verify( mockGuM );
	}


	public void testGetFile() throws RemoteException, IOException {

		EventQueue<ResponseEvent> responseQueue = new EventQueue<ResponseEvent>();

		byte[ ] testByteArray = new byte[ 0 ];

		EasyMock.expect( mockGuM.getFile( "filePath", 0l, 0 ) ).andReturn( testByteArray ).once();
		EasyMock.replay( mockGuM );

		this.ebGum.getFile( responseQueue, "filePath", 0, 0 );

		ResponseEvent re = responseQueue.unblockingRemove();

		assertTrue( re instanceof GumGetFileResponseEvent );
		assertEquals( re.getResponse(), testByteArray );

		EasyMock.verify( mockGuM );

	}


	public void testGetAttributes() throws RemoteException {

		assertEquals( specAttributesMap, ebGum.getGumSpec().getAttributes() );
	}


	public void testGetID() {

		EasyMock.replay( mockGuM );

		assertEquals( gumID, this.ebGum.getID() );

		EasyMock.verify( mockGuM );
	}


	public void testGetResult() throws RemoteException, ExecutorException {

		ExecutorHandle executorHandle = EasyMock.createMock( ExecutorHandle.class );
		EventQueue<ResponseEvent> responseQueue = new EventQueue<ResponseEvent>();
		ExecutorResult executorResult = EasyMock.createMock( ExecutorResult.class );

		EasyMock.expect( mockGuM.getResult( executorHandle ) ).andReturn( executorResult ).once();
		EasyMock.replay( mockGuM );

		this.ebGum.getResult( responseQueue, executorHandle );

		ResponseEvent re = responseQueue.unblockingRemove();

		assertTrue( re instanceof GumGetResultResponseEvent );
		assertEquals( re.getResponse(), executorResult );

		EasyMock.verify( mockGuM );
	}


	public void testFileExists() throws RemoteException {

		EventQueue<ResponseEvent> responseQueue = new EventQueue<ResponseEvent>();
		File testFile = new File( "." );

		EasyMock.expect( mockGuM.fileExists( testFile ) ).andReturn( false ).once();
		EasyMock.replay( mockGuM );

		this.ebGum.fileExists( responseQueue, testFile );

		ResponseEvent re = responseQueue.unblockingRemove();

		assertTrue( re instanceof GumFileExistsResponseEvent );
		assertEquals( re.getResponse(), false );

		EasyMock.verify( mockGuM );
	}


	public void testChangeFilePermissions() throws RemoteException, IOException {

		EventQueue<ResponseEvent> responseQueue = new EventQueue<ResponseEvent>();
		File testFile = new File( "." );

		mockGuM.changeFilePermissions( testFile, 0 );

		EasyMock.expectLastCall().once();
		EasyMock.replay( mockGuM );

		this.ebGum.changeFilePermissions( responseQueue, testFile, 0 );

		EasyMock.verify( mockGuM );
	}


	public void testGetRemoteFileLength() throws RemoteException, IOException {

		EventQueue<ResponseEvent> responseQueue = new EventQueue<ResponseEvent>();
		File testFile = new File( "." );

		EasyMock.expect( mockGuM.getRemoteFileLength( testFile ) ).andReturn( 0l ).once();
		EasyMock.replay( mockGuM );

		this.ebGum.getRemoteFileLength( responseQueue, testFile );

		ResponseEvent re = responseQueue.unblockingRemove();

		assertTrue( re instanceof GumGetRemoteFileLengthResponseEvent );
		assertEquals( re.getResponse(), 0l );

		EasyMock.verify( mockGuM );
	}


	public void testGetRemoteFileLastModification() throws RemoteException, IOException {

		EventQueue<ResponseEvent> responseQueue = new EventQueue<ResponseEvent>();
		File testFile = new File( "." );

		EasyMock.expect( mockGuM.getRemoteFileLastModification( testFile, true ) ).andReturn( 0l ).once();
		EasyMock.replay( mockGuM );

		this.ebGum.getRemoteFileLastModification( responseQueue, testFile, true );

		ResponseEvent re = responseQueue.unblockingRemove();

		assertTrue( re instanceof GumGetRemoteFileLastModificationResponseEvent );
		assertEquals( re.getResponse(), 0l );

		EasyMock.verify( mockGuM );
	}


	public void testCreatePlaypen() throws RemoteException, UnableToCreatePlaypenException {

		EventQueue<ResponseEvent> responseQueue = new EventQueue<ResponseEvent>();

		EasyMock.expect( mockGuM.createPlaypen( 1000 ) ).andReturn( "playpen" ).once();
		EasyMock.replay( mockGuM );

		this.ebGum.createPlaypen( responseQueue, 1000 );

		ResponseEvent re = responseQueue.unblockingRemove();

		assertTrue( re instanceof GumCreatePlaypenResponseEvent );
		assertEquals( re.getResponse(), "playpen" );

		EasyMock.verify( mockGuM );
	}


	public void testCreateStorageDirectory() throws RemoteException, UnableToCreateStorageException {

		EventQueue responseQueue = new EventQueue();

		EasyMock.expect( mockGuM.createStorageDirectory() ).andReturn( "storage" ).once();
		EasyMock.replay( mockGuM );

		this.ebGum.createStorageDirectory( responseQueue );

		ResponseEvent re = (ResponseEvent) responseQueue.unblockingRemove();

		assertTrue( re instanceof GumCreateStorageDirectoryResponseEvent );
		assertEquals( re.getResponse(), "storage" );

		EasyMock.verify( mockGuM );
	}


	public void testGetStorageDirectory() throws RemoteException {

		EventQueue responseQueue = new EventQueue();

		EasyMock.expect( mockGuM.getStorageDirectory() ).andReturn( "storage" ).once();
		EasyMock.replay( mockGuM );

		this.ebGum.getStorageDirectory( responseQueue );

		ResponseEvent re = (ResponseEvent) responseQueue.unblockingRemove();

		assertTrue( re instanceof GumGetStorageDirectoryResponseEvent );
		assertEquals( re.getResponse(), "storage" );

		EasyMock.verify( mockGuM );
	}


	public void testGetStatus() throws RemoteException {

		EventQueue responseQueue = new EventQueue();

		EasyMock.expect( mockGuM.getStatus() ).andReturn( GumStatus.READY ).once();
		EasyMock.replay( mockGuM );

		this.ebGum.getStatus( responseQueue );

		ResponseEvent re = (ResponseEvent) responseQueue.unblockingRemove();

		assertTrue( re instanceof GumGetStatusResponseEvent );
		assertEquals( re.getResponse(), GumStatus.READY );

		EasyMock.verify( mockGuM );

	}


	public void testGetRemoteFileDigest() throws RemoteException, UnableToDigestFileException {

		EventQueue responseQueue = new EventQueue();

		File file = new File( "." );

		EasyMock.expect( mockGuM.getRemoteFileDigest( file ) ).andReturn( "getRemoteFileDigest" ).once();
		EasyMock.replay( mockGuM );

		this.ebGum.getRemoteFileDigest( responseQueue, file );

		ResponseEvent re = (ResponseEvent) responseQueue.unblockingRemove();

		assertTrue( re instanceof GumGetRemoteFileDigestResponseEvent );
		assertEquals( re.getResponse(), "getRemoteFileDigest" );

	}
}
