/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.ebgum;

import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.Map;

import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.InvalidationResponseEvent;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.exception.UnableToCreatePlaypenException;
import org.ourgrid.common.exception.UnableToCreateStorageException;
import org.ourgrid.common.exception.UnableToDigestFileException;
import org.ourgrid.common.executor.ExecutorException;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.ExecutorResult;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.security.MGSecureClientSocketFactory;
import org.ourgrid.common.security.MGSecureServerSocketFactory;
import org.ourgrid.peer.manager.EBPeerManagerFacade;

/**
 * It will be the remote object to be contacted by <code>GumClient</code>. It
 * represents a proxy to make possible remote connections between a
 * <code>GumClient</code> and the <code>EBGumFacade</code>. The peer holds
 * a proxy to every Mygrid-UserAgent connection, so we have an in-band
 * connection between the two components.
 */
public class GumProxy extends OurgridUnicastRemoteObject implements InvalidatableProxy, Gum {

	/**
	 * Serial identification of the class. It need to be changed only if the
	 * class interface is changed.
	 */
	public static final long serialVersionUID = 33L;

	/**
	 * A remote reference to a <code>EBGum</code> that will be the object
	 * responsible for remote calls.
	 */
	private EBGum ebGum;

	private transient ExecutorHandle handle;

	/** Informs a client that this proxy was invalidated. */
	private boolean invalid = false;

	private EBPeerManagerFacade peerManagerFacade;

	private Collection<EventQueue> blockedEventQueues;

	private GumID gumID;

	/** Logger to store log messages. */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( GumProxy.class );

	private final long requestID;


	/**
	 * Default constructor
	 * 
	 * @param gridMachineFacade the <code>EBGum</code> that this proxy will
	 *        forward calls
	 * @param peerManagerFacade a reference to the peer manager, useful to
	 *        report when a gum is lost
	 * @param requestID
	 * @throws RemoteException
	 */
	public GumProxy( EBGum gridMachineFacade, EBPeerManagerFacade peerManagerFacade, long requestID )
		throws RemoteException {

		super();
		this.ebGum = gridMachineFacade;
		this.requestID = requestID;
		this.gumID = ebGum.getGumSpec().getGumID();
		this.peerManagerFacade = peerManagerFacade;
		blockedEventQueues = Collections.synchronizedCollection( new LinkedList<EventQueue>() );
	}


	/**
	 * Constructor with SecureSocketFactories.
	 * 
	 * @param gridMachineFacade
	 * @param peerManagerFacade
	 * @param secureClientSocketFactory
	 * @param secureServerSocketFactory
	 * @param securePort
	 * @throws RemoteException
	 */
	public GumProxy( EBGum gridMachineFacade, EBPeerManagerFacade peerManagerFacade, long requestID,
						MGSecureClientSocketFactory secureClientSocketFactory,
						MGSecureServerSocketFactory secureServerSocketFactory, int securePort ) throws RemoteException {

		super( securePort, secureClientSocketFactory, secureServerSocketFactory );
		this.ebGum = gridMachineFacade;
		this.peerManagerFacade = peerManagerFacade;
		this.requestID = requestID;
		this.gumID = ebGum.getGumSpec().getGumID();
		blockedEventQueues = Collections.synchronizedCollection( new LinkedList<EventQueue>() );
	}


	/**
	 * Called when a remote exception occurs.
	 */
	private void lostGum() {

		LOG.debug( gumID + " was lost" );
		this.peerManagerFacade.lostGum( this.gumID, this.requestID );
	}


	/**
	 * Test if the Gum is invalid. When this happen, the Gum will throw a
	 * exception to signalize it new state.
	 * 
	 * @throws RemoteException If the GuM is invalid.
	 */
	private synchronized void testInvalidation() throws RemoteException {

		if ( invalid ) {
			throw new RemoteException( "Invalidated Gum: The machine has been allocated to another user" );
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#putFile(String, byte[], boolean)
	 */
	public void putFile( String remoteFile, byte[ ] data, boolean firstSlice ) throws RemoteException, IOException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().putFile( eq, remoteFile, data, firstSlice );

		ResponseEvent re = waitResponse( eq );

		final Object response = re.getResponse();

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof IOException )
				throw (IOException) response;
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getFile(String, long, int)
	 */
	public byte[ ] getFile( String filePath, long filePos, int size ) throws RemoteException, IOException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().getFile( eq, filePath, filePos, size );

		ResponseEvent re = waitResponse( eq );

		final Object response = re.getResponse();

		if ( response instanceof byte[ ] ) {
			return (byte[ ]) response;
		}

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof IOException )
				throw (IOException) response;
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
		}

		return null; // This might never happen.
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#createPlaypen( int )
	 */
	public String createPlaypen( int size ) throws RemoteException, UnableToCreatePlaypenException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().createPlaypen( eq, size );

		ResponseEvent re = waitResponse( eq );

		final Object response = re.getResponse();

		if ( response instanceof String ) {
			return (String) response;
		}

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof UnableToCreatePlaypenException )
				throw (UnableToCreatePlaypenException) response;
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
		}

		return null; // This might never happen.

	}


	/**
	 * @see org.ourgrid.common.gum.Gum#createStorageDirectory()
	 */
	public String createStorageDirectory() throws RemoteException, UnableToCreateStorageException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().createStorageDirectory( eq );

		ResponseEvent re = waitResponse( eq );

		final Object response = re.getResponse();

		if ( response instanceof String ) {
			return (String) response;
		}

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof UnableToCreateStorageException )
				throw (UnableToCreateStorageException) response;
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
		}

		return null; // This might never happen.
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getStorageDirectory()
	 */
	public String getStorageDirectory() throws RemoteException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().getStorageDirectory( eq );

		ResponseEvent re = waitResponse( eq );

		final Object response = re.getResponse();

		if ( response instanceof String ) {
			return (String) response;
		}

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
		}

		return null; // This might never happen.

	}


	/**
	 * @see org.ourgrid.common.gum.Gum#fileExists(File)
	 */
	public boolean fileExists( File file ) throws RemoteException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().fileExists( eq, file );

		ResponseEvent re = waitResponse( eq );

		final Object response = re.getResponse();

		if ( response instanceof Boolean ) {
			return ((Boolean) response).booleanValue();
		}

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
		}

		return false; // This might never happen.

	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getRemoteFileLength(File)
	 */
	public long getRemoteFileLength( File remoteFile ) throws RemoteException, IOException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().getRemoteFileLength( eq, remoteFile );

		ResponseEvent re = waitResponse( eq );

		final Object response = re.getResponse();

		if ( response instanceof Long ) {
			return ((Long) response).longValue();
		}

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
			if ( response instanceof IOException ) {
				throw (IOException) response;
			}
		}

		return -1L; // This might never happen.

	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getRemoteFileLastModification(File,
	 *      boolean)
	 */
	public long getRemoteFileLastModification( File remoteFile, boolean normalized ) throws RemoteException,
		IOException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().getRemoteFileLastModification( eq, remoteFile, normalized );

		ResponseEvent re = waitResponse( eq );

		final Object response = re.getResponse();

		if ( response instanceof Long ) {
			return ((Long) response).longValue();
		}

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof IOException ) {
				throw (IOException) response;
			}
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
		}

		return -1L; // This might never happen.

	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getRemoteFileDigest(java.io.File)
	 */
	public String getRemoteFileDigest( File remoteFile ) throws RemoteException, UnableToDigestFileException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().getRemoteFileDigest( eq, remoteFile );

		ResponseEvent re = waitResponse( eq );

		final Object response = re.getResponse();

		if ( response instanceof String ) {
			return (String) response;
		}

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof UnableToDigestFileException ) {
				throw (UnableToDigestFileException) response;
			}
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
		}

		return null; // This might never happen.
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#changeFilePermissions(java.io.File, int)
	 */
	public void changeFilePermissions( File file, int permissions ) throws RemoteException, IOException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().changeFilePermissions( eq, file, permissions );

		ResponseEvent re = waitResponse( eq );

		final Object response = re.getResponse();

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
			if ( response instanceof IOException ) {
				throw (IOException) response;
			}
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#remoteExecute(String, String, Map)
	 */
	public ExecutorHandle remoteExecute( String remoteDir, String command, Map<String,String> envVars )
		throws RemoteException, ExecutorException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().remoteExecute( eq, remoteDir, command, envVars );

		ResponseEvent re = waitResponse( eq );

		final Object response = re.getResponse();

		if ( response instanceof ExecutorHandle ) {
			this.handle = (ExecutorHandle) response;
			return this.handle;
		}

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof ExecutorException ) {
				throw (ExecutorException) response;
			}
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
		}

		return null;
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getResult(ExecutorHandle)
	 */
	public ExecutorResult getResult( ExecutorHandle handle ) throws RemoteException, ExecutorException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().getResult( eq, handle );

		ResponseEvent re = waitResponse( eq );

		testInvalidation();

		Object response = re.getResponse();

		testInvalidation();

		if ( response instanceof ExecutorResult ) {
			return (ExecutorResult) response;
		}

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
			if ( response instanceof ExecutorException ) {
				throw (ExecutorException) response;
			}
		}

		return null;

	}


	private synchronized EBGum getGridMachine() throws RemoteException {

		if ( ebGum == null ) {
			throw new RemoteException();
		}
		return ebGum;
	}


	/**
	 * Makes this proxy invalid. First, it sets the proxy as invalid, then puts
	 * a invalidation event on every blocked event queue. Its Stops replica
	 * currently executing on the gum
	 */
	public synchronized void invalidate() {

		this.invalid = true;

		EBGum ebGumRef = ebGum;

		ebGum = null;

		synchronized ( blockedEventQueues ) {
			for ( EventQueue<ResponseEvent> eq : blockedEventQueues ) {
				eq.put( new InvalidationResponseEvent() );
			}
		}

		if ( ebGumRef != null ) {
			EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
			ebGumRef.stopReplica( eq );
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#getStatus()
	 */
	public GumStatus getStatus() {

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		ResponseEvent re = null;
		try {
			testInvalidation();
			getGridMachine().getStatus( eq );
			re = waitResponse( eq );
		} catch ( RemoteException e ) {
			return GumStatus.CONTACTING;
		}

		if ( re != null && re.getResponse() instanceof GumStatus ) {
			return (GumStatus) re.getResponse();
		}

		return GumStatus.CONTACTING;

	}


	/**
	 * @see org.ourgrid.common.gum.Gum#startReplica()
	 */
	public void startReplica() throws RemoteException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().startReplica( eq );

		ResponseEvent re = waitResponse( eq );

		Object response = re.getResponse();

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
		}
	}


	/**
	 * @see org.ourgrid.common.gum.Gum#stopReplica()
	 */
	public void stopReplica() throws RemoteException {

		testInvalidation();

		EventQueue<ResponseEvent> eq = new EventQueue<ResponseEvent>();
		getGridMachine().stopReplica( eq );

		ResponseEvent re = waitResponse( eq );

		Object response = re.getResponse();

		if ( response instanceof Throwable ) {
			LOG.debug( "Proxy Error on proxy " + this, (Throwable) response );
			if ( response instanceof RemoteException ) {
				this.lostGum();
				throw (RemoteException) response;
			}
		}
	}


	/**
	 * Block until a response is returned by the Gum
	 * 
	 * @throws RemoteException
	 */
	private ResponseEvent waitResponse( EventQueue<ResponseEvent> eq ) throws RemoteException {

		blockedEventQueues.add( eq );
		ResponseEvent re = eq.blockingRemove();
		blockedEventQueues.remove( eq );

		testInvalidation();

		return re;
	}
}
