/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.ebgum;

import java.io.File;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.exception.UnavailableGumException;
import org.ourgrid.common.executor.ExecutorHandle;
import org.ourgrid.common.executor.ExecutorResult;
import org.ourgrid.common.gum.GumClient;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.util.TempFileManager;
import org.ourgrid.gridmachine.useragent.UserAgentClient;
import org.ourgrid.peer.manager.EBPeerManagerFacade;
import org.ourgrid.peer.manager.FakeEventEngine;
import org.ourgrid.test.PeerPreCommitTest;
import org.ourgrid.test.TestUtils;

/**
 * The <code>GumProxy</code> test. It will try to execute all steps from the
 * proxy creation, to the <code>EBGum<code> and the GridMachine.
 */
public class GumProxyTest extends TestCase implements PeerPreCommitTest {

	private static final long REQID_1 = 1;

	private TestUtils testUtils;

	private List gumSpecs;

	private String createdSDF;

	private GumClient gumClient;

	private File fileTest;

	private File tmpDir;

	private GumProxy proxyOfLocalPeer, proxyOfRemotePeer;

	private EBPeerManagerFacade pmFacade2;

	private GumSpec gumSpec;


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();

		testUtils = new TestUtils();
		testUtils.load();

		gumSpecs = testUtils.getGumSpecs( 1, TestUtils.TYPE_UALINUX );

		createdSDF = testUtils.createSDF( gumSpecs );

		gumSpec = (GumSpec) gumSpecs.get( 0 );

		String command = "uaadmin init " + createdSDF;
		Configuration configuration = Configuration.getInstance( Configuration.PEER );
		String mgroot = (new File( configuration.getProperty( Configuration.MGROOT ) )).getCanonicalPath();
		Process process = Runtime.getRuntime().exec( mgroot + File.separator + "bin" + File.separator + command );
		assertEquals( 0, process.waitFor() );

		EBGumFacade ebGumFacadeLocalPeer = EBGumCreator.createGuM( gumSpec );
		ebGumFacadeLocalPeer.start();

		EBPeerManagerFacade nicePmFacade = new EBPeerManagerFacade( new FakeEventEngine() );

		proxyOfLocalPeer = new GumProxy( ebGumFacadeLocalPeer, nicePmFacade, REQID_1 );

		EBGumImpl ebGuMRemotePeer = new EBGumImpl( proxyOfLocalPeer, gumSpec );

		EBGumFacade ebGumFacadeRemotePeer = new EBGumFacade( ebGuMRemotePeer );
		ebGumFacadeRemotePeer.start();

		pmFacade2 = EasyMock.createMock( EBPeerManagerFacade.class );

		proxyOfRemotePeer = new GumProxy( ebGumFacadeRemotePeer, pmFacade2, REQID_1 );

		gumClient = new UserAgentClient( proxyOfRemotePeer, gumSpec );

		System.out.println( "Chosen GuM -> " + gumSpec.getAttributes() );

		tmpDir = new File( testUtils.getProperty( TestUtils.SYSTEM_TEMP_DIR ) );
		fileTest = new File( tmpDir + File.separator + "temp-000" );

		fileTest.delete();
	}


	protected void tearDown() throws Exception {

		super.tearDown();

		String command = "uaadmin stop " + createdSDF;
		Configuration configuration = Configuration.getInstance( Configuration.PEER );
		String mgroot = (new File( configuration.getProperty( Configuration.MGROOT ) )).getCanonicalPath();
		Process process = Runtime.getRuntime().exec( mgroot + File.separator + "bin" + File.separator + command );
		assertEquals( 0, process.waitFor() );

	}


	public void testInvalidateProxyOfRemotePeer() throws Exception {

		gumClient.getGumSpec().getGumID();
		gumClient.fileExists( new File( "testFile1" ) );
		proxyOfRemotePeer.invalidate();
		gumClient.getGumSpec().getGumID();
		try {
			gumClient.fileExists( new File( "testFile1" ) );
			fail();
		} catch ( UnavailableGumException e ) {
			// The proxy was invalidated.
		}
		assertEquals( GumStatus.CONTACTING, gumClient.getStatus() );
	}


	public void testInvalidateProxyOfLocalPeer() throws Exception {

		gumClient.getGumSpec().getGumID();
		gumClient.fileExists( new File( "testFile1" ) );
		proxyOfLocalPeer.invalidate();
		gumClient.getGumSpec().getGumID();
		try {
			gumClient.fileExists( new File( "testFile1" ) );
			fail();
		} catch ( UnavailableGumException e ) {
			// The proxy was invalidated.
		}
		assertEquals( GumStatus.CONTACTING, gumClient.getStatus() );
	}


	public void testInvalidateProxyOfLocalPeer2() throws Exception {

		System.out.println( "testInvalidate2! Remote executing...." );
		gumClient.startReplica();
		ExecutorHandle handle = gumClient.remoteExecute( "/tmp", "sleep 1000", null );

		pmFacade2.lostGum( EasyMock.eq( gumSpec.getGumID() ), EasyMock.eq( REQID_1 ) );
		EasyMock.expectLastCall().once();

		EasyMock.replay( pmFacade2 );

		System.out.println( "testInvalidate2! Getting result in new Thread" );
		ExecutorService exec = Executors.newSingleThreadExecutor();
		Future<ExecutorResult> future = exec.submit( new ResultThread( gumClient, handle ) );

		// This should be here because we want the previous thread to block
		// before invalidating
		// the proxy.
		Thread.sleep( 2000 );

		proxyOfLocalPeer.invalidate();

		try {
			future.get();
			fail( "Should throw UnavailableGumException" );
		} catch ( Exception e ) {
			e.printStackTrace();
			assertTrue( e.getCause() instanceof UnavailableGumException );
		}

		EasyMock.verify( pmFacade2 );
	}

	private class ResultThread implements Callable<ExecutorResult> {

		private GumClient client;

		private ExecutorHandle handle;


		public ResultThread( GumClient client, ExecutorHandle handle ) {

			this.client = client;
			this.handle = handle;
		}


		public ExecutorResult call() throws Exception {

			System.out.println( "client: " + client + ", handle: " + handle );
			return client.getResult( handle );
		}
	}


	public void testCommands() throws Exception {

		this.remoteExecute();
		this.remoteExecuteWithTaskFailure();
		this.createPlaypen();
		this.getID();
		this.existsAndGetRemoteFileLength();
		this.storageDirectory();
		this.getStatus();
	}


	private void remoteExecute() throws Exception {

		System.out.println( "Remote executing...." );
		gumClient.startReplica();
		ExecutorHandle handle = gumClient.remoteExecute( "/tmp", "echo oi", null );

		System.out.println( "Getting result..." );
		ExecutorResult result = gumClient.getResult( handle );
		assertEquals( 0, result.getExitValue() );
		gumClient.stopReplica();
	}


	private void remoteExecuteWithTaskFailure() throws Exception {

		System.out.println( "Remote executing...." );
		gumClient.startReplica();

		System.out.println( "Getting result..." );
		ExecutorHandle handle = gumClient.remoteExecute( "/tmp", "exit 1", null );

		ExecutorResult result = gumClient.getResult( handle );

		assertEquals( 1, result.getExitValue() );
		assertEquals( GumStatus.UP, gumClient.getStatus() );
		gumClient.stopReplica();
	}


	private void createPlaypen() throws Exception {

		System.out.println( "Creating playpen...." );
		gumClient.startReplica();
		gumClient.createPlaypen( 1000 );

		File tmp = TempFileManager.createTempFileWithBogusData( "tmp", "file", tmpDir, 128 );

		System.out.println( "Putting a file..." );

		gumClient.putFile( tmp.getAbsolutePath(), "test" );

		gumClient.getFile( "/etc/passwd", fileTest.getAbsolutePath() );

		fileTest.deleteOnExit();

		gumClient.stopReplica();

	}


	private void getID() {

		assertEquals( gumClient.getGumSpec().getGumID(), ((GumSpec) gumSpecs.get( 0 )).getGumID() );

	}


	private void existsAndGetRemoteFileLength() throws Exception {

		File tmpDir = new File( testUtils.getProperty( TestUtils.SYSTEM_TEMP_DIR ) );
		File tmpFile = TempFileManager.createTempFileWithBogusData( "mygrid", "test", tmpDir, 128 );

		gumClient.startReplica();

		String playpen = gumClient.createPlaypen( 1000 );
		File newFile = new File( playpen + File.separator + "tmpFile" );
		gumClient.putFile( tmpFile.getAbsolutePath(), "tmpFile" );
		assertTrue( gumClient.fileExists( newFile ) );
		assertEquals( 128L, gumClient.getRemoteFileLength( newFile ) );

		gumClient.stopReplica();
	}


	private void storageDirectory() throws Exception {

		gumClient.startReplica();

		assertEquals( "/tmp", gumClient.getStorageDirectory() );

		gumClient.stopReplica();
	}


	private void getStatus() {

		assertEquals( GumStatus.UP, gumClient.getStatus() );
	}
}
