/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager;

import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ShutdownEventEngine;
import org.ourgrid.common.event.ShutdownRequestEvent;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.common.exception.IllegalStateException;
import org.ourgrid.common.gump.EBGump;
import org.ourgrid.common.gump.EBGumpClient;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.peer.ebgum.EBGum;
import org.ourgrid.peer.event.DisposeOfGumNotifyEvent;
import org.ourgrid.peer.event.FinishRequestGumsNotifyEvent;
import org.ourgrid.peer.event.HereIsGumNotifyEvent;
import org.ourgrid.peer.event.PauseRequestGumsEvent;
import org.ourgrid.peer.event.RequestGumsEvent;
import org.ourgrid.peer.event.ResumeRequestGumsRequestEvent;

/**
 * Description: Basic facade for ourgrid peer modules.
 */
public abstract class EBPeerFacade implements EBGump, EBGumpClient {

	protected EBGump ebGump;

	protected EBGumpClient ebGumpClient;

	protected ShutdownEventEngine eventProcessor;

	/** indicates if the event processor has started */
	private boolean started;


	public EBPeerFacade() {

		eventProcessor = new PeerEventEngine( "PeerFacade" );
	}


	/**
	 * Shuts down the facade.
	 */
	public synchronized EventQueue<ShutdownResponseEvent> shutdown() {

		EventQueue<ShutdownResponseEvent> responseQueue = new EventQueue<ShutdownResponseEvent>();
		ShutdownRequestEvent shutdownRequestEvent = new ShutdownRequestEvent( responseQueue, eventProcessor );
		if ( eventProcessor.isAlive() ) {
			eventProcessor.putEvent( shutdownRequestEvent );
		} else {
			responseQueue.put( new ShutdownResponseEvent() );
		}
		return responseQueue;
	}


	/**
	 * Starts the thread
	 */
	public synchronized void startEventProcessor() {

		if ( this.started == true ) {
			throw new IllegalStateException();
		}
		eventProcessor.startProcessing();
		this.started = true;
	}


	/**
	 * @see org.ourgrid.common.gump.EBGump#resumeRequestGums(long)
	 */
	public void resumeRequestGums( long requestID ) {

		eventProcessor.putEvent( new ResumeRequestGumsRequestEvent( ebGump, requestID ) );
	}


	/**
	 * @see org.ourgrid.common.gump.EBGump#pauseRequestGums(long)
	 */
	public void pauseRequestGums( long requestID ) {

		eventProcessor.putEvent( new PauseRequestGumsEvent( ebGump, requestID ) );
	}


	/**
	 * @see org.ourgrid.common.gump.EBGump#requestGums(RequestSpec)
	 */
	public void requestGums( RequestSpec requestSpec ) {

		eventProcessor.putEvent( new RequestGumsEvent( ebGump, requestSpec ) );
	}


	/**
	 * @see org.ourgrid.common.gump.EBGump#finishRequestGums(long)
	 */
	public void finishRequestGums( long requestID ) {

		eventProcessor.putEvent( new FinishRequestGumsNotifyEvent( ebGump, requestID ) );
	}


	/**
	 * @see org.ourgrid.common.gump.EBGump#disposeGum(long, GumID)
	 */
	public void disposeGum( long requestID, GumID gumID ) {

		eventProcessor.putEvent( new DisposeOfGumNotifyEvent( ebGump, requestID, gumID ) );
	}


	/**
	 * @see org.ourgrid.common.gump.EBGumpClient#hereIsGum(EBGum, ObjectID,
	 *      long)
	 */
	public void hereIsGum( EBGum receivedGum, ObjectID gumOwnerObjectID, long requestID ) {

		eventProcessor.putEvent( new HereIsGumNotifyEvent( ebGumpClient, receivedGum, gumOwnerObjectID, requestID ) );
	}
}
