/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager;

import java.net.URISyntaxException;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import org.ourgrid.common.exception.RequestAlreadyExistsException;
import org.ourgrid.common.fd.AbstractFailureDetector;
import org.ourgrid.common.gum.GumStateFD;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.gump.EBGump;
import org.ourgrid.common.gump.EBGumpClient;
import org.ourgrid.common.gump.GumpClient;
import org.ourgrid.common.gump.GumpClientThreadManager;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.matcher.Matcher;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.common.spec.RequestSpec.RequestSource;
import org.ourgrid.common.url.URLProvider;
import org.ourgrid.mygrid.scheduler.gridmanager.GumResponseEntry;
import org.ourgrid.peer.community.nof.AccountingManager;
import org.ourgrid.peer.community.nof.Arbitrator;
import org.ourgrid.peer.community.obtainer.EBCommunityObtainerFacade;
import org.ourgrid.peer.ebgum.EBGum;
import org.ourgrid.peer.ebgum.GumProxy;
import org.ourgrid.peer.manager.allocation.AllocationEntry;
import org.ourgrid.peer.manager.allocation.AllocationManager;
import org.ourgrid.peer.manager.allocation.AllocationEntry.GumSource;
import org.ourgrid.peer.manager.exception.RequestDoesNotExistException;
import org.ourgrid.peer.manager.gummanager.EBGumManagerImpl;
import org.ourgrid.peer.manager.request.RequestEntry;
import org.ourgrid.peer.manager.request.RequestManager;
import org.ourgrid.peer.ui.UIManager;
import org.ourgrid.peer.ui.PeerUIManager.Notifiables;

/**
 * The <code>EBPeerManager</code> is responsible to manage the peer resources
 * (Gums). It is important to distribute the gums to the correct request and
 * keep the grid machines distributed steadily.
 */
public class EBPeerManager implements EBGump, EBGumpClient {

	protected RequestManager requestManager;

	protected AllocationManager allocationManager;

	protected EBGumManagerImpl ebGumManager;

	protected EBCommunityObtainerFacade ebCommunityObtainer;

	/**
	 * A ThreadPool responsible to manage threads to invoke methods on the
	 * external consumers.
	 */
	private GumpClientThreadManager consumerThreadManager;

	/**
	 * Logger to store log messages.
	 */
	private static final transient org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( EBPeerManager.class );

	/**
	 * Object id used to identify this peer to other peers.
	 */
	private ObjectID remoteAccessID;

	/**
	 * Object id used to identify this peer to local brokers.
	 */
	private ObjectID brokerAccessID;

	private UIManager uiManager;


	/**
	 * Default constructor.
	 * 
	 * @param consumerThreadManager a thread manager to control the operations
	 *        done in separate threads (for example: delivering a machine to to
	 *        a remote consumer)
	 * @param remoteAccessID the identification of the remote object used by
	 *        remote peer to access this peer
	 * @param brokerAccessID the identification of the remote object used by
	 *        local brokers to access this peer
	 */
	public EBPeerManager( GumpClientThreadManager consumerThreadManager, ObjectID remoteAccessID,
							ObjectID brokerAccessID, UIManager uiManager ) {

		this( RequestManager.getInstance(), AllocationManager.getInstance(), consumerThreadManager, uiManager );
		this.remoteAccessID = remoteAccessID;
		this.brokerAccessID = brokerAccessID;
	}


	/**
	 * This constructor should be directly used only by tests.
	 * 
	 * @param reqManager <code>RequestManager</code> created by the test
	 * @param allocationManager <code>AllocationManager</code> created by the
	 *        test
	 * @param consumerThreadManager <code>GumpClientThreadManager</code>
	 *        created by the test
	 */
	protected EBPeerManager( RequestManager reqManager, AllocationManager allocationManager,
								GumpClientThreadManager consumerThreadManager, UIManager uiManager ) {

		this.requestManager = reqManager;
		this.allocationManager = allocationManager;
		this.consumerThreadManager = consumerThreadManager;
		this.uiManager = uiManager;
	}


	/**
	 * Configure internal modules.
	 */
	public void config( EBCommunityObtainerFacade communityObtainerFacade, EBGumManagerImpl gumManager ) {

		this.ebCommunityObtainer = communityObtainerFacade;
		this.ebGumManager = gumManager;
	}


	public void requestGums( RequestSpec requestSpec ) throws RequestAlreadyExistsException {

		this.requestManager.createRequestEntry( requestSpec );

		registerClientMonitorable( requestSpec );

		// broadcast the request to the entire community
		if ( requestSpec.isFromALocalBroker() ) {
			LOG.info( "Local Peer needs " + requestSpec.getRequiredGums() + " gums. Request: " + requestSpec.getID() );
			ebCommunityObtainer.requestGums( requestSpec );
		} else {
			LOG.info( "Remote Peer needs " + requestSpec.getRequiredGums() + " gums. Request: " + requestSpec.getID()
					+ ". Peer:  " + requestSpec.getObjectID() );
		}

		// try to serve the request with local deallocated gums
		try {
			respondAndRedistribute( requestSpec.getID() );
		} catch ( RequestDoesNotExistException e ) {
			LOG.error( e.getMessage() );
		}
	}


	/**
	 * @param newRequestId
	 * @throws RequestDoesNotExistException
	 */
	private void respondAndRedistribute( final long newRequestId ) throws RequestDoesNotExistException {

		respondRequest( newRequestId );

		if ( this.requestManager.requestNeedMoreGums( newRequestId ) ) {
			redistribute( newRequestId );
		}
	}


	/**
	 * Makes the necessary redistributions: 1. Tries to get gums allocated to
	 * remote requests 2. Distribute the local gums evenly among local requests
	 * 3. Distributes the local gums among remote requests, according to the NOF
	 * balances
	 * 
	 * @param newRequestId the newly arrived request
	 * @throws RequestDoesNotExistException
	 */
	private void redistribute( long newRequestId ) throws RequestDoesNotExistException {

		if ( this.requestManager.requestIsLocal( newRequestId ) ) {
			this.takeGumsFromRemoteRequests( newRequestId );
			this.redistributeLocalGumsForLocalRequests( newRequestId );
		} else {
			this.redistributeLocalGumsForRemoteRequests( newRequestId );
		}
	}


	/**
	 * Takes the gums which are allocated to remote request and give them the
	 * newly arrived local request
	 * 
	 * @param newRequestId local request which has just arrived
	 * @throws RequestDoesNotExistException
	 */
	private void takeGumsFromRemoteRequests( long newRequestId ) throws RequestDoesNotExistException {

		Collection<RequestEntry> targetRequests = requestManager
				.getRequestsWithAllocatedGums( RequestSource.FROM_REMOTE_PEER );

		while ( this.requestManager.requestNeedMoreGums( newRequestId ) && !targetRequests.isEmpty() ) {
			RequestEntry requestToLoseGum = Arbitrator.chooseRemoteRequestToLoseGum( targetRequests );
			boolean gumDonated = false;
			for ( GumID gum : requestToLoseGum.getGums() ) {
				if ( gumMatchesWithRequest( gum, newRequestId ) ) {
					this.preemptGum( gum );
					this.deliverGum( newRequestId, gum );
					gumDonated = true;
					LOG.debug( "Preempting gum: " + gum + " from request " + requestToLoseGum
							+ " and giving to request " + newRequestId );
					break;
				}
			}
			if ( !gumDonated ) {
				targetRequests.remove( requestToLoseGum );
			}
		}
	}


	private void registerClientMonitorable( RequestSpec requestSpec ) {

		try {
			uiManager.registerNewMonitorable( Notifiables.CONSUMER, requestSpec.getObjectID() );
		} catch ( RemoteException e ) {
			LOG.info( "Cannot register notifiable at address " + requestSpec.getObjectID().getObjectURL() );
		}
	}


	protected void redistributeLocalGumsForLocalRequests( long newRequestId ) throws RequestDoesNotExistException {

		Collection<GumpClient> localConsumers = this.requestManager.getConsumers( RequestSource.FROM_LOCAL_BROKER );

		if ( localConsumers.size() <= 1 ) {
			// there is no redistribution because there is only one consumer
			return;
		}

		int nLocalGums = this.allocationManager.getNumberOfLocalGums();

		int gumsDeservedByConsumer = (nLocalGums / localConsumers.size()) < 1	? 1
																				: (nLocalGums / localConsumers.size());

		GumpClient newConsumer = this.requestManager.getRequestSpec( newRequestId ).getGumpClient();

		int receiverInbalance = gumsDeservedByConsumer
				- this.requestManager.getGums( newConsumer, GumSource.LOCAL ).size();

		if ( receiverInbalance > 0 ) {
			Iterator<GumpClient> iter = localConsumers.iterator();
			while ( iter.hasNext() && this.requestManager.requestNeedMoreGums( newRequestId ) ) {

				GumpClient donator = iter.next();

				if ( !newConsumer.equals( donator ) ) {
					int donatorInbalance = this.requestManager.getGums( donator, GumSource.LOCAL ).size()
							- gumsDeservedByConsumer;
					if ( donatorInbalance > 0 ) {
						/*
						 * This consumer has more gums that it deserves, we will
						 * now take gums from it while its balance is positive
						 * or while the receiver request needs gums.
						 */
						Collection<RequestEntry> donatorRequests = this.requestManager.getRequests( donator );
						if ( !donatorRequests.isEmpty() )
							makeDonations( newRequestId, donatorRequests, donatorInbalance, receiverInbalance );
					}
				}
			}
		}
	}


	protected void redistributeLocalGumsForRemoteRequests( long newRequestId ) throws RequestDoesNotExistException {

		Collection<GumpClient> remoteConsumers = requestManager.getConsumers( RequestSource.FROM_REMOTE_PEER );

		if ( remoteConsumers.size() <= 1 ) {
			// there is no redistribution because there is only one consumer
			return;
		}

		List<DistributionHelpEntry> distributionList = new LinkedList<DistributionHelpEntry>();

		DistributionHelpEntry newConsumerDistributionEntry = null;

		for ( GumpClient consumerEntry : remoteConsumers ) {
			Collection<RequestEntry> consumerRequests = requestManager.getRequests( consumerEntry );
			if ( !consumerRequests.isEmpty() ) {
				double consumerBalance = AccountingManager.getInstance().getPeerBalance(
						consumerRequests.iterator().next().getSpec().getObjectID() );
				DistributionHelpEntry newDistributionEntry = new DistributionHelpEntry( consumerBalance, requestManager
						.getGums( consumerEntry ).size(), consumerRequests );
				distributionList.add( newDistributionEntry );
				if ( this.requestManager.getRequestSpec( newRequestId ).getGumpClient().equals( consumerEntry ) ) {
					newConsumerDistributionEntry = newDistributionEntry;
				}
			}
		}

		this.computeCommunityRedistribution( distributionList, this.allocationManager.getLocalDonated().size() );

		if ( newConsumerDistributionEntry != null ) {
			int receiverInbalance = newConsumerDistributionEntry.getGumsToReceive()
					- newConsumerDistributionEntry.getGumsCurrentlyAllocated();
			if ( receiverInbalance > 0 ) {
				// Choose the gums that will be preempted and delivered to the
				// request
				Iterator<DistributionHelpEntry> it1 = distributionList.iterator();
				while ( it1.hasNext() && this.requestManager.requestNeedMoreGums( newRequestId ) ) {
					DistributionHelpEntry distEntry = it1.next();
					int inbalance = distEntry.getGumsCurrentlyAllocated() - distEntry.getGumsToReceive();
					if ( inbalance > 0 ) {
						// This consumer has more gums that it deserves, we will
						// now takes gums from it
						// while its balance is positive or while the receiver
						// request needs gums.
						if ( !distEntry.getRequests().isEmpty() ) {
							makeDonations( newRequestId, distEntry.getRequests(), inbalance, receiverInbalance );
						}
					}
				}
			}
		}
	}


	/**
	 * Calculates the new distribution of local gums allocated to the community
	 * according to Network of Favors balances.
	 * 
	 * @param redistributionList a list containing information about consumers.
	 *        It will be modified by this method.
	 * @param numberOfGumsInUseByCommunity the total number of gums currently
	 *        allocated to the community
	 */
	protected void computeCommunityRedistribution( List<DistributionHelpEntry> redistributionList,
													int numberOfGumsInUseByCommunity ) {

		double totalBalance = 0;

		LOG.debug( "Computing redistribution for " + numberOfGumsInUseByCommunity + " machines and "
				+ redistributionList.size() + " consumer(s)" );

		Collections.sort( redistributionList, DistributionHelpEntry.getBalanceComparator() );

		ListIterator<DistributionHelpEntry> it = redistributionList.listIterator();

		while ( it.hasNext() ) {
			DistributionHelpEntry helper = it.next();
			totalBalance += helper.getNofBalance();
		}

		int assignedGums = 0;
		it = redistributionList.listIterator();
		while ( it.hasNext() ) {
			int consumerDeservedGums;
			DistributionHelpEntry helper = it.next();
			assignedGums += consumerDeservedGums = Math.round( (float) (numberOfGumsInUseByCommunity
					* helper.getNofBalance() / totalBalance) );
			helper.setGumsToReceive( consumerDeservedGums );
			LOG.debug( "Assigning " + consumerDeservedGums + ", gums to " + helper.getNofBalance() );
		}

		// We may have to adjust something due to a rounding issue
		int diff = assignedGums - numberOfGumsInUseByCommunity;

		if ( diff != 0 ) {
			if ( diff > 0 ) {
				// We have to take some some gums out from someone
				while ( diff > 0 && it.hasPrevious() ) {
					DistributionHelpEntry helper = it.previous();
					int gumsToReceive = helper.getGumsToReceive();
					if ( gumsToReceive > 0 ) {
						diff--;
						helper.setGumsToReceive( --gumsToReceive );
					}
				}
			} else {
				// We have to give some more gums to someone
				it = redistributionList.listIterator();
				while ( diff < 0 && it.hasNext() ) {
					DistributionHelpEntry helper = it.next();
					helper.setGumsToReceive( helper.getGumsToReceive() + 1 );
					diff++;
				}
			}
		}
	}


	private void makeDonations( long newRequestId, Collection<RequestEntry> donatorRequests, int donatorInbalance,
								int receiverInbalance ) throws RequestDoesNotExistException {

		do {
			RequestEntry requestToLoseGum = Arbitrator.chooseRequestToLoseGum( donatorRequests );

			boolean donationDone = false;
			for ( GumID gum : requestToLoseGum.getGums() ) {
				if ( gumMatchesWithRequest( gum, newRequestId ) ) {
					preemptGum( gum );
					deliverGum( newRequestId, gum );
					donationDone = true;
					donatorInbalance--;
					receiverInbalance--;
					break;
				}
			}
			if ( !donationDone ) {
				donatorRequests.remove( requestToLoseGum );
			}
		} while ( !donatorRequests.isEmpty() && donatorInbalance > 0 && receiverInbalance > 0
				&& this.requestManager.requestNeedMoreGums( newRequestId ) );
	}


	private boolean gumMatchesWithRequest( GumID gumID, long requestId ) throws RequestDoesNotExistException {

		AllocationEntry allocationEntry = allocationManager.getAllocationEntry( gumID );
		String requirements = this.requestManager.getRequestSpec( requestId ).getRequirements();
		return Matcher.match( requirements, allocationEntry.getEBGum().getGumSpec().getAttributes() )
				&& this.requestManager.requestWantsGum( requestId, gumID );
	}


	/**
	 * Tries to serve a request with local deallocated gums.
	 * 
	 * @param newRequestId the request to be served
	 * @throws RequestDoesNotExistException
	 */
	protected void respondRequest( long newRequestId ) throws RequestDoesNotExistException {

		Collection<AllocationEntry> idleGums = this.allocationManager.getLocalIdle();

		Iterator<AllocationEntry> idleGumsItr = idleGums.iterator();
		while ( idleGumsItr.hasNext() && this.requestManager.requestNeedMoreGums( newRequestId ) ) {
			AllocationEntry allocationEntry = idleGumsItr.next();
			if ( gumMatchesWithRequest( allocationEntry.getEBGum().getID(), newRequestId ) ) {
				LOG.debug( "Delivering idle gum: " + allocationEntry.getEBGum().getID() + " to request: "
						+ newRequestId );
				this.deliverGum( newRequestId, allocationEntry.getEBGum().getID() );
			}
		}
	}


	/**
	 * @see org.ourgrid.common.gump.Gump#finishRequestGums(long)
	 */
	public void finishRequestGums( long requestID ) {

		try {
			boolean isLocal = this.requestManager.requestIsLocal( requestID );
			Collection<GumID> allocatedGums = this.requestManager.getRequestGums( requestID );

			LOG.debug( allocatedGums.size() + " gums allocated to request " + requestID + ": " + allocatedGums );

			/*
			 * the request is deleted here to prevent that it gets gums during a
			 * reallocation
			 */
			ObjectID consumerID = this.requestManager.getRequestSpec( requestID ).getObjectID();
			this.requestManager.deleteRequest( requestID );

			if ( isLocal ) {
				LOG.info( "Local Peer needs no more gums for request " + requestID );
				// Unregistering consumer from FD
				if ( !this.requestManager.getConsumersIDs( RequestSource.FROM_LOCAL_BROKER ).contains( consumerID ) ) {
					try {
						uiManager.unregisterNewMonitorable( Notifiables.CONSUMER, consumerID );
					} catch ( RemoteException e ) {
					}
				}

				this.ebCommunityObtainer.finishRequestGums( requestID );
			} else {
				LOG.info( "Remote Peer needs no more gums for request " + requestID );
			}

			for ( GumID gumId : allocatedGums ) {
				AllocationEntry allocationEntry = this.allocationManager.getAllocationEntry( gumId );
				LOG.debug( "Will deallocate gum: " + gumId + " AllocationEntry: " + allocationEntry );

				if ( allocationEntry != null ) {
					if ( allocationEntry.isALocalGum() ) {
						this.ebGumManager.invalidate( gumId );
						this.allocationManager.deallocateGum( gumId );
						reallocateGum( gumId );
					} else {
						this.ebGumManager.removeProxy( gumId );
						this.allocationManager.deallocateAndRemove( gumId );
					}
				}
			}
		} catch ( RequestDoesNotExistException e ) {
			LOG.error( e.getMessage() );
		}
	}


	private void reallocateGum( GumID gumID ) {

		RequestEntry request = chooseRequestToReceiveGum( gumID );

		if ( request != null ) {
			try {
				deliverGum( request.getSpec().getID(), gumID );
				LOG.debug( "The chosen request is " + request + "." );
			} catch ( RequestDoesNotExistException e ) {
				LOG.error( e.getMessage() );
			}
		}
	}


	/**
	 * @see org.ourgrid.common.gump.Gump#disposeGum(long, GumID)
	 */
	public void disposeGum( long requestID, GumID gumID ) {

		AllocationEntry allocationEntry = this.allocationManager.getAllocationEntry( gumID );

		// a gum has been disposed without previous allocation
		if ( allocationEntry == null ) {
			LOG.error( "Trying to disposed a gum that has not been allocated before ( " + gumID + " ) " );
			return;
		}

		if ( allocationEntry.getRequestID() != requestID ) {
			LOG.error( "Trying to disposed a gum ( " + gumID + " ) that has not been allocated to the requestID ( "
					+ requestID + " ) " );
			return;
		}

		if ( allocationEntry.isALocalGum() ) {
			disposeOfLocalGum( requestID, gumID );
		} else {
			this.ebCommunityObtainer.disposeGum( requestID, gumID );
			discardGum( requestID, gumID );
		}
		LOG.debug( "The grid machine <" + gumID + "> was disposed." );
	}


	private void discardGum( long requestID, GumID gumID ) {

		this.ebGumManager.removeProxy( gumID );
		this.allocationManager.deallocateAndRemove( gumID );
		this.requestManager.removeGumFromRequest( requestID, gumID );
	}


	private void disposeOfLocalGum( long requestID, GumID gumID ) {

		this.ebGumManager.invalidate( gumID );
		this.allocationManager.deallocateGum( gumID );
		this.requestManager.removeGumFromRequest( requestID, gumID );
		reallocateGum( gumID );
	}


	public void lostGum( GumID gumID, long requestID ) {

		AllocationEntry allocationEntry = this.allocationManager.getAllocationEntry( gumID );

		if ( allocationEntry != null ) {

			if ( allocationEntry.getRequestID() == requestID ) {

				if ( allocationEntry.isALocalGum() ) {
					LOG.debug( "Gum lost is local. Will advice failure detector" );
					this.adviceFailureDetector( gumID );
				} else {
					LOG.debug( "Gum lost is remote. Will clean stuff" );
					this.gumWasLost( gumID );
				}
			}
		}
	}


	public void adviceFailureDetector( GumID gumID ) {

		try {
			String monitorableURL = URLProvider.getFullURL( gumID.getUrl(), AbstractFailureDetector.BOUND_NAME );
			uiManager.failureDetectorAdvice( monitorableURL, new GumStateFD( GumStatus.CONTACTING ) );
		} catch ( RemoteException e ) {
			LOG.error( "Could not get access to PeerUIManager" );
		} catch ( URISyntaxException e ) {
			LOG.error( "Invalid gum URL: " + gumID.getUrl() );
		}
	}


	public void gumWasLost( GumID gumID ) {

		AllocationEntry allocationEntry = this.allocationManager.getAllocationEntry( gumID );

		if ( allocationEntry != null ) {

			final long requestID = allocationEntry.getRequestID();

			try {
				this.discardGum( requestID, gumID );
				respondAndRedistribute( requestID );
			} catch ( RequestDoesNotExistException e ) {
				LOG.error( e.getMessage() );
			}
		}
	}


	// org.ourgrid.common.gump.GumpClient behavior

	/**
	 * @see org.ourgrid.common.gump.GumpClient#hereIsGum(GumResponseEntry)
	 */
	public void hereIsGum( EBGum receivedGum, ObjectID gumOwnerObjectID, long requestID ) {

		final GumID gumID = receivedGum.getID();
		LOG.info( "Gum " + gumID + " received from peer " + gumOwnerObjectID.getObjectURL() + " for request "
				+ requestID );

		boolean delivered = false;

		try {

			boolean requestIsLocal = this.requestManager.requestIsLocal( requestID );

			if ( requestIsLocal ) {

				if ( this.allocationManager.gumExists( gumID ) ) {
					// same gums has been received twice
					AllocationEntry allocationEntry = this.allocationManager.getAllocationEntry( gumID );

					// if the gum is already allocated to the right request,
					// keep it allocated.
					if ( allocationEntry.isAllocatedToRequest( requestID ) )
						return;

					// remove gum from old request
					this.ebGumManager.invalidate( gumID );
					this.discardGum( allocationEntry.getRequestID(), gumID );
				}

				this.allocationManager.gumReceived( receivedGum, receivedGum.getGumSpec() );
				delivered = this.deliverGum( requestID, gumID );
			}

		} catch ( RequestDoesNotExistException e ) {
			LOG.error( "Gum could not be delivered to a non existent request. Disposing..." );
			this.ebCommunityObtainer.finishRequestGums( requestID );
		}

		if ( !delivered ) {
			this.allocationManager.deallocateAndRemove( gumID );
			this.ebCommunityObtainer.disposeGum( requestID, gumID );
		}
	}


	/**
	 * When a machine is set at <code>EBGumManagerImpl</code> at it state is
	 * up, a hereIsLocalGuM is received at this entity.
	 */
	public void hereIsLocalGum( EBGum ebGum, GumSpec gumSpec ) {

		AllocationEntry allocationEntry;

		allocationEntry = this.allocationManager.gumReceived( ebGum, gumSpec );

		if ( allocationEntry != null ) {

			LOG.debug( "The grid machine <" + ebGum.getID() + "> was received." );

			RequestEntry request = chooseRequestToReceiveGum( ebGum.getID() );
			if ( request != null ) {
				try {
					this.deliverGum( request.getSpec().getID(), ebGum.getID() );
					LOG.debug( "The chosen request is " + request + "." );
				} catch ( RequestDoesNotExistException e ) {
					LOG.debug( "Could not deliver gum to request: " + e.getMessage() );
				}
			}
		} else {
			LOG.debug( "The grid machine <" + ebGum.getID() + "> was already received." );
		}
	}


	/**
	 * @see org.ourgrid.common.gump.Gump#pauseRequestGums(long)
	 */
	public void pauseRequestGums( long requestID ) {

		try {
			if ( !this.requestManager.requestIsPaused( requestID ) ) {
				this.requestManager.pauseRequestGums( requestID );

				if ( this.requestManager.requestIsLocal( requestID ) ) {
					LOG.info( "Asked to pause LOCAL request " + requestID );
					ebCommunityObtainer.pauseRequestGums( requestID );
				} else {
					LOG.info( "Asked to pause REMOTE request " + requestID );
				}
			}
		} catch ( RequestDoesNotExistException e ) {
			LOG.error( e.getMessage() );
		}
	}


	/**
	 * @see org.ourgrid.common.gump.Gump#resumeRequestGums(long)
	 */
	public void resumeRequestGums( long requestID ) {

		try {
			if ( this.requestManager.requestIsPaused( requestID ) ) {
				this.requestManager.resumeRequestGums( requestID );

				respondAndRedistribute( requestID );

				if ( this.requestManager.requestIsLocal( requestID ) ) {
					LOG.info( "Asked to resume LOCAL request " + requestID );
					ebCommunityObtainer.resumeRequestGums( requestID );
				} else {
					LOG.info( "Asked to resume REMOTE request " + requestID );
				}
			}
		} catch ( RequestDoesNotExistException e ) {
			LOG.error( e.getMessage() );
		}
	}


	/**
	 * Finds a (local or community) request that is satisfied by the grid
	 * machine.
	 * 
	 * @param gumID the identification of the grid machine
	 */
	private RequestEntry chooseRequestToReceiveGum( GumID gumID ) {

		Collection<RequestEntry> requests;
		Collection<RequestEntry> requestsWantingGums;
		Collection<RequestEntry> satisfiedRequests;

		RequestEntry chosenRequest = null;

		// First, look for local requests
		requests = this.requestManager.getAliveRequests( RequestSource.FROM_LOCAL_BROKER );
		requestsWantingGums = this.requestManager.getRequestsThatNeedGums( requests );
		satisfiedRequests = this.getRequestsSatisfiedByGum( requestsWantingGums, gumID );

		chosenRequest = Arbitrator.chooseRequestToReceiveGum( satisfiedRequests );

		if ( chosenRequest == null ) {
			// If no local request need gums, consider remote requests
			requests = this.requestManager.getAliveRequests( RequestSource.FROM_REMOTE_PEER );
			requestsWantingGums = this.requestManager.getRequestsThatNeedGums( requests );
			satisfiedRequests = this.getRequestsSatisfiedByGum( requestsWantingGums, gumID );
			chosenRequest = Arbitrator.chooseRemoteRequestToReceiveGum( satisfiedRequests );
		}

		return chosenRequest;

	}


	/**
	 * Finds all the <code>RequestEntry</code>s at the specified list that is
	 * satisfied by the grid machine attributes.
	 * 
	 * @param requests The <code>Collection</code> containing the requests.
	 * @param gumID The identification of the grid machine
	 * @return The <code>Collection</code> containing all the
	 *         <code>RequestEntry</code> satisfied by the specified attributes
	 *         or <code>null</code> if no request was found.
	 */
	private Collection<RequestEntry> getRequestsSatisfiedByGum( Collection<RequestEntry> requests, GumID gumID ) {

		AllocationEntry allocationEntry = AllocationManager.getInstance().getAllocationEntry( gumID );
		Collection<RequestEntry> result = new LinkedList<RequestEntry>();

		for ( RequestEntry entry : requests ) {

			try {
				if ( gumMatchesWithRequest( allocationEntry.getGumSpec().getGumID(), entry.getSpec().getID() ) ) {
					result.add( entry );
				}
			} catch ( RequestDoesNotExistException e ) {
			}

		}

		return result;
	}


	/**
	 * Delivers a grid machine for the specified request id. This method will
	 * update all information for keep track of the machine and keep the request
	 * status.
	 * 
	 * @param requestID identification of the request that will receive the gum
	 * @param gumID identification of the gum that will be delivered
	 * @throws RequestDoesNotExistException
	 */
	private boolean deliverGum( long requestID, GumID gumID ) throws RequestDoesNotExistException {

		final boolean requestIsLocal = this.requestManager.requestIsLocal( requestID );

		AllocationEntry allocationEntry = this.allocationManager.getAllocationEntry( gumID );

		if ( allocationEntry == null ) {
			return false;
		}

		EBGum ebGum = allocationEntry.getEBGum();

		GumProxy gumProxy;

		try {
			gumProxy = this.ebGumManager.createAndSetProxy( ebGum, requestID );
			LOG.debug( "Created new Proxy (" + gumID + "): " + gumProxy );
		} catch ( RemoteException e ) {
			LOG.error( "Error creating Proxy " + gumID + ". Cause: " + e.getClass().getName() + ", " + e.getMessage() );
			return false;
		}

		GumResponseEntry gumResponseEntry;
		if ( requestIsLocal ) {
			gumResponseEntry = new GumResponseEntry( gumProxy, allocationEntry.getGumSpec(), requestID,
					this.brokerAccessID );
		} else {
			gumResponseEntry = new GumResponseEntry( gumProxy, allocationEntry.getGumSpec(), requestID,
					this.remoteAccessID );
		}

		this.allocationManager.allocateGumToRequest( gumID, requestID, requestIsLocal );

		this.requestManager.hereIsGumToRequest( requestID, gumID );

		this.consumerThreadManager.createAndExecuteDelivererThread( this.requestManager.getRequestSpec( requestID )
				.getObjectID(), gumResponseEntry );

		return true;
	}


	/**
	 * Preempts a grid machine, that is, deallocate the grid machine from a
	 * request.
	 * 
	 * @param gumID the identification of the grid machine to be preempted
	 */
	protected void preemptGum( GumID gumID ) {

		AllocationEntry allocationEntry = this.allocationManager.getAllocationEntry( gumID );

		if ( allocationEntry != null ) {
			// invalidate the gum and clean it.
			this.ebGumManager.invalidate( gumID );

			long requestID = allocationEntry.getRequestID();
			allocationManager.deallocateGum( allocationEntry.getEBGum().getID() );
			requestManager.removeGumFromRequest( requestID, allocationEntry.getEBGum().getID() );
		} else {
			LOG.debug( "The grid machine <" + gumID + "> doesn't exist." );
		}
	}


	public void consumerIsDead( ObjectID consumerID ) {

		LOG.debug( "Consumer " + consumerID + " is dead" );

		Collection<RequestEntry> requests = this.requestManager.getConsumerRequests( consumerID );

		for ( RequestEntry request : requests ) {
			finishRequestGums( request.getSpec().getID() );
		}

		this.requestManager.deleteConsumerRequests( consumerID );
	}


	/**
	 * Notifies the request that the given gum is not wanted.
	 * 
	 * @param requestID ID of the request to notify.
	 * @param gumID ID of the gum that is not wanted.
	 */
	public void unwantedGum( long requestID, GumID gumID ) {

		AllocationEntry allocationEntry = this.allocationManager.getAllocationEntry( gumID );

		// a gum has been disposed without previous allocation
		if ( allocationEntry == null ) {
			LOG.error( "Gum was not allocated before ( " + gumID + " ) " );
			return;
		}

		try {
			if ( allocationEntry.isALocalGum() ) {
				LOG.debug( "Receiving unwantedGum for LOCAL gum: " + gumID + ". Request: " + requestID );
				this.requestManager.unwantedGum( requestID, gumID );
			} else {
				LOG.debug( "Receiving unwantedGum for REMOTE gum: " + gumID + ". Request: " + requestID );
				this.ebCommunityObtainer.unwantedGum( requestID, gumID );
			}

			if ( allocationEntry.getRequestID() == requestID ) {
				if ( allocationEntry.isALocalGum() ) {
					this.disposeOfLocalGum( requestID, gumID );
				} else {
					this.discardGum( requestID, gumID );
				}
			} else {
				LOG.error( "Could not dispose gum: " + gumID + ". It was not allocated to request " + requestID );
			}
			LOG.debug( "Gum (" + gumID + ") is unwanted for request (" + requestID + ")" );
		} catch ( RequestDoesNotExistException e ) {
			LOG.error( e.getMessage() );
		}
	}
}
