/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager;

import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.event.Event;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.common.exception.IllegalStateException;
import org.ourgrid.common.gump.RemoteFakeGumpClient;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.peer.event.DisposeOfGumNotifyEvent;
import org.ourgrid.peer.event.FinishRequestGumsNotifyEvent;
import org.ourgrid.peer.event.RequestGumsEvent;
import org.ourgrid.test.PeerFastTest;
import org.ourgrid.test.TestUtils;

/**
 * The EBPeerManagerFacade test.
 */
public class EBPeerManagerFacadeTest extends TestCase implements PeerFastTest {

	private EBPeerManagerFacade ebPeerManagerFacade;

	private FakeEventEngine fakeEventProcessor;

	private TestUtils testUtils;


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();

		Configuration.reset();

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createMGProperties();

		Configuration config = Configuration.getInstance( Configuration.PEER );
		config.setProperty( Configuration.PROP_PORT, testUtils.getProperty( TestUtils.RMI_PORT ) + "" );

		int registryPort = Integer.parseInt( testUtils.getProperty( TestUtils.RMI_PORT ) );

		try {
			// Starting registry
			LocateRegistry.createRegistry( registryPort );

		} catch ( RemoteException e ) {
			LocateRegistry.getRegistry( registryPort );
		}

		fakeEventProcessor = new FakeEventEngine();
		ebPeerManagerFacade = new EBPeerManagerFacade( fakeEventProcessor );
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		Configuration.reset();
		testUtils.reset();
	}


	/**
	 * Verifies if the <code>JICWannaGuMsRequestEvent</code> is been received
	 * by the <code>FakeEventEngine</code> using the facade.
	 * 
	 * @throws RemoteException
	 */
	public void testWannaGuMs() throws RemoteException {

		RequestSpec requestSpec = new RequestSpec( 1, "test", 10, new RemoteFakeGumpClient(
				"rmi://localhost:3000/Test1" ) );
		ebPeerManagerFacade.requestGums( requestSpec );
		Event event = fakeEventProcessor.getEvent();
		assertTrue( event instanceof RequestGumsEvent );
	}


	/**
	 * Verifies if the <code>JICNoMoreGuMsNotifyEvent</code> is been received
	 * by the <code>FakeEventEngine</code> using the facade.
	 */
	public void testNoMoreGuMs() {

		ebPeerManagerFacade.finishRequestGums( 12345 );
		Event event = fakeEventProcessor.getEvent();
		assertTrue( event instanceof FinishRequestGumsNotifyEvent );
	}


	/**
	 * Verifies if the <code>JICDisposeOfGuMNotifyEvent</code> is been
	 * received by the <code>FakeEventEngine</code> using the facade.
	 */
	public void testDisposeOf() {

		GumID gumID = new GumID( TestUtils.FAKE_MACHINE_NAME, "0", "peername", "0" );
		ebPeerManagerFacade.disposeGum( 12345, gumID );
		Event event = fakeEventProcessor.getEvent();
		assertTrue( event instanceof DisposeOfGumNotifyEvent );
	}


	/**
	 * Tests the start processing.
	 */
	public void testStartProcessing() {

		assertFalse( fakeEventProcessor.isAlive() );
		ebPeerManagerFacade.startEventProcessor();
		assertTrue( fakeEventProcessor.isAlive() );
		try {
			ebPeerManagerFacade.startEventProcessor();
			fail( "A exception should be raised before this point because the startEventProcessor() method must be invoked only once." );
		} catch ( IllegalStateException ise ) {
		}
	}


	/**
	 * Tests the shutdown.
	 */
	public void testShutdown() {

		/* FakeEventEngine is running */
		ebPeerManagerFacade.startEventProcessor();
		EventQueue queue = ebPeerManagerFacade.shutdown();
		assertEquals( 0, queue.size() );

		/* FakeEventEngine is already stopped */
		assertFalse( fakeEventProcessor.isAlive() );
		queue = ebPeerManagerFacade.shutdown();
		assertEquals( 1, queue.size() );
		assertTrue( queue.unblockingRemove() instanceof ShutdownResponseEvent );
	}
}
