/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager;

import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.exception.RequestAlreadyExistsException;
import org.ourgrid.common.gump.FakeEBGump;
import org.ourgrid.common.gump.GumpClient;
import org.ourgrid.common.gump.GumpClientThreadManager;
import org.ourgrid.common.gump.RemoteFakeGumpClient;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.peer.community.nof.AccountingManager;
import org.ourgrid.peer.community.obtainer.EBCommunityObtainerFacade;
import org.ourgrid.peer.ebgum.EBGum;
import org.ourgrid.peer.ebgum.EBGumCreator;
import org.ourgrid.peer.ebgum.EBGumFacade;
import org.ourgrid.peer.ebgum.FakeGumFacade;
import org.ourgrid.peer.manager.allocation.AllocationEntry;
import org.ourgrid.peer.manager.allocation.AllocationManager;
import org.ourgrid.peer.manager.allocation.GumAlreadyAllocatedException;
import org.ourgrid.peer.manager.allocation.AllocationEntry.GumSource;
import org.ourgrid.peer.manager.exception.RequestDoesNotExistException;
import org.ourgrid.peer.manager.gummanager.EBGumManagerImpl;
import org.ourgrid.peer.manager.request.RequestManager;
import org.ourgrid.peer.ui.UIManager;
import org.ourgrid.peer.ui.PeerUIManager.Notifiables;
import org.ourgrid.test.PeerPreCommitTest;
import org.ourgrid.test.SpecialCareTest;
import org.ourgrid.test.TestUtils;
import org.ourgrid.test.threadServices.ThreadPoolServices;
import org.ourgrid.test.threadServices.ThreadServices;

public class EBPeerManagerTest extends TestCase implements PeerPreCommitTest, SpecialCareTest {

	private static final long reqID_1 = 1L;
	
	private static final long reqID_2 = 2L;

	private ObjectID objectID1;

	private ObjectID objectID2;

	private ObjectID objectID3;

	private ObjectID objectID4;

	/**
	 * The Allocation.
	 */
	protected AllocationManager allocationManager;

	/**
	 * The EBRequestManager.
	 */
	protected RequestManager requestManager;

	/**
	 * The object to be tested.
	 */
	protected EBPeerManager ebPeerManager;

	/**
	 * A broker consumer.
	 */
	private GumpClient fakeBroker;

	private GumpClient localFakeConsumer;

	/**
	 * A community fake consumer.
	 */
	private RemoteFakeGumpClient communityFakeConsumer;

	/**
	 * These are fake local grid machines.
	 */
	private EBGumFacade localGum1, localGum2, localGum3, localGum4, localGum5, localGum6;

	/**
	 * These are fake community grid machines.
	 */
	private EBGum communityGum1, communityGum2, communityGum3, communityGum4;

	private TestUtils testUtils;

	private GumID gumID2, gumID3, gumID4, gumID5, gumID6;

	private String peerName;

	private String peerPort;

	private AccountingManager accountingManager;

	private int rmiPort;

	private ObjectID communityFakeConsumerObjectID;

	private ObjectID fakeBrokerObjectID;

	private ObjectID localFakeConsumerObjectID;

	private FakeEBGump communityObtainer;

	private EBCommunityObtainerFacade communityObtainerFacade;

	private ObjectID fakeBrokerObjectID2;

	private RemoteFakeGumpClient fakeBroker2;

	private EBGumManagerImpl ebGumManager;

	private UIManager mockUIManager;


	@Override
	protected void setUp() throws Exception {

		super.setUp();

		Configuration.reset();

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createMGProperties();
		testUtils.createOGProperties();

		rmiPort = testUtils.getRMIPort();
		Configuration config = Configuration.getInstance( Configuration.PEER );
		config.setProperty( Configuration.PROP_PORT, rmiPort + "" );

		try {
			LocateRegistry.createRegistry( rmiPort );
		} catch ( Exception e ) {
			LocateRegistry.getRegistry( rmiPort );
		}

		objectID1 = new ObjectID( "Community1", (long) 1111, "rmi://localhost:" + rmiPort + "/FakeGum1" );
		objectID2 = new ObjectID( "Community2", (long) 2222, "rmi://localhost:" + rmiPort + "/FakeGum2" );
		objectID3 = new ObjectID( "Community3", (long) 3333, "rmi://localhost:" + rmiPort + "/FakeGum3" );
		objectID4 = new ObjectID( "Community4", (long) 4444, "rmi://localhost:" + rmiPort + "/FakeGum4" );

		peerName = "test.peername";
		peerPort = "0";

		String communityPeerName = "another peer";
		String communityPeerPort = "0";

		config.setProperty( Configuration.PROP_EXTERNAL_NAME, peerName );
		config.setProperty( Configuration.PROP_PORT, peerPort );

		// creates the fake objects
		this.localGum1 = new FakeGumFacade( new GumSpec( generateAttributeMap( "linux", "256", "localgum1", "1000",
				peerName, peerPort ) ) );
		this.localGum2 = new FakeGumFacade( new GumSpec( generateAttributeMap( "windows", "1024", "localgum2", "1000",
				peerName, peerPort ) ) );
		this.localGum3 = new FakeGumFacade( new GumSpec( generateAttributeMap( "windows", "1024", "localgum3", "1000",
				peerName, peerPort ) ) );
		this.localGum4 = new FakeGumFacade( new GumSpec( generateAttributeMap( "linux", "512", "localgum4", "1000",
				peerName, peerPort ) ) );
		this.localGum5 = new FakeGumFacade( new GumSpec( generateAttributeMap( "linux", "1024", "localgum5", "1000",
				peerName, peerPort ) ) );
		this.localGum6 = new FakeGumFacade( new GumSpec( generateAttributeMap( "linux", "256", "localgum6", "1000",
				peerName, peerPort ) ) );

		this.communityGum1 = new FakeGumFacade( new GumSpec( generateAttributeMap( "linux", "512", "communitygum1",
				"1000", communityPeerName, communityPeerPort ) ) );
		this.communityGum2 = new FakeGumFacade( new GumSpec( generateAttributeMap( "windows", "512", "communitygum2",
				"1000", communityPeerName, communityPeerPort ) ) );
		this.communityGum3 = new FakeGumFacade( new GumSpec( generateAttributeMap( "windows", "512", "communitygum3",
				"1000", communityPeerName, communityPeerPort ) ) );
		this.communityGum4 = new FakeGumFacade( new GumSpec( generateAttributeMap( "windows", "512", "communitygum4",
				"1000", communityPeerName, communityPeerPort ) ) );

		this.communityFakeConsumerObjectID = new ObjectID( "CommunityFakeConsumerObjectID", (long) 8976,
				"rmi://localhost:" + rmiPort + "/CommunityFakeConsumer" );
		this.communityFakeConsumer = new RemoteFakeGumpClient( communityFakeConsumerObjectID );

		this.fakeBrokerObjectID = new ObjectID( "FakeGumClient", (long) 89761, "rmi://localhost:" + rmiPort
				+ "/FakeBroker" );
		this.fakeBrokerObjectID2 = new ObjectID( "FakeGumClient2", (long) 89762, "rmi://localhost:" + rmiPort
				+ "/FakeBroker" );

		this.localFakeConsumerObjectID = new ObjectID( "FakeGumClient", (long) 89763, "rmi://localhost:" + rmiPort
				+ "/FakeConsumer" );

		fakeBroker = new RemoteFakeGumpClient( fakeBrokerObjectID );
		fakeBroker2 = new RemoteFakeGumpClient( fakeBrokerObjectID2 );
		localFakeConsumer = new RemoteFakeGumpClient( localFakeConsumerObjectID );

		Naming.rebind( communityFakeConsumerObjectID.getObjectURL(), communityFakeConsumer );
		Naming.rebind( localFakeConsumerObjectID.getObjectURL(), localFakeConsumer );
		Naming.rebind( fakeBrokerObjectID.getObjectURL(), fakeBroker );

		AllocationManager.reset();
		this.allocationManager = AllocationManager.getInstance();

		RequestManager.reset();
		this.requestManager = RequestManager.getInstance();

		EBPeerManagerFacade peerManagerFacadeMock = EasyMock.createNiceMock( EBPeerManagerFacade.class );

		GumpClientThreadManager consumerThreadManager = new GumpClientThreadManager( peerManagerFacadeMock );

		this.mockUIManager = EasyMock.createMock( UIManager.class );
		this.ebPeerManager = new EBPeerManager( requestManager, allocationManager, consumerThreadManager, mockUIManager );

		communityObtainer = new FakeEBGump();
		this.communityObtainerFacade = new EBCommunityObtainerFacade();

		ebGumManager = new EBGumManagerImpl( this.ebPeerManager, peerManagerFacadeMock );

		this.ebPeerManager.config( this.communityObtainerFacade, ebGumManager );

		AccountingManager.reset();
		accountingManager = AccountingManager.getInstance();

		communityObtainer.reset();
	}


	/**
	 * Generates an attribute map of a Gum.
	 * 
	 * @param os Gum OS.
	 * @param mem Gum available memory
	 * @param gumName Gum name.
	 * @param gumPort Gum port.
	 * @param peerName Peer name.
	 * @param peerPort Peer port.
	 * @return The attributes map.
	 */
	private Map<String,String> generateAttributeMap( String os, String mem, String gumName, String gumPort,
														String peername, String peerport ) {

		Map<String,String> attrib1 = new HashMap<String,String>();
		attrib1.put( "os", os );
		attrib1.put( "mem", mem );
		attrib1.put( GumSpec.ATT_NAME, gumName );
		attrib1.put( GumSpec.ATT_PORT, gumPort );
		attrib1.put( GumSpec.ATT_PEERNAME, peername );
		attrib1.put( GumSpec.ATT_PEERPORT, peerport );
		return attrib1;
	}


	/*
	 * @see TestCase#tearDown()
	 */
	@Override
	protected void tearDown() throws Exception {

		super.tearDown();
		AllocationManager.reset();
		RequestManager.reset();
		Configuration.reset();
		testUtils.reset();

		Naming.unbind( communityFakeConsumerObjectID.getObjectURL() );
		OurgridUnicastRemoteObject.unexportObject( communityFakeConsumer, true );

		Naming.unbind( localFakeConsumerObjectID.getObjectURL() );
		OurgridUnicastRemoteObject.unexportObject( localFakeConsumer, true );

		Naming.unbind( fakeBrokerObjectID.getObjectURL() );
		OurgridUnicastRemoteObject.unexportObject( fakeBroker, true );

		communityObtainer.reset();
	}


	/**
	 * Verifies if the <code>hereIsGuM()</code> called by
	 * <code>EBLocalGuMPFacade</code> when there isn't another request is
	 * working correctly.
	 */
	public void testHereIsGuMOnlyWithRequestMadeByLocalGuMP() {

		// receives the gum for request
		this.ebPeerManager.hereIsLocalGum( this.localGum1, this.localGum1.getGumSpec() );

		// check the map
		assertEquals( 1, this.allocationManager.getLocalIdle().size() );
	}


	/**
	 * Verifies if the <code>hereIsGuM()</code> is working correctly when the
	 * request is already fulfilled or doesn't exist.
	 * 
	 * @throws RequestAlreadyExistsException Indicates that a request already
	 *         exists.
	 * @throws RequestDoesNotExistException
	 */
	public void testHereIsGuMForCommunityGuMsWhenRequestDoesntExistOrFulfilled() throws RequestAlreadyExistsException,
		RequestDoesNotExistException {

		// creates other requests

		RequestSpec request = new RequestSpec( 1, "", 1, this.localFakeConsumer );
		request.setAsFromLocalBroker();
		this.requestManager.createRequestEntry( request );

		// receives the local GuM1 for request
		this.ebPeerManager.hereIsLocalGum( this.localGum1, this.localGum1.getGumSpec() );
		// check the maps
		assertEquals( 1, this.requestManager.getRequestGums( 1 ).size() );

		assertTrue( this.allocationManager.getAllocationEntry( localGum1.getID() ).getRequestID() == 1 );

		// ensures that this request doesn't need more GuMs
		assertFalse( this.requestManager.requestNeedMoreGums( 1 ) );

		// receives the community GuM1 for fulfilled request, the peer will
		// accept the gums, that should be disposed by the broker.

		this.ebPeerManager.hereIsGum( this.communityGum1, objectID1, 1 );
		// check the maps
		assertEquals( 2, this.requestManager.getRequestGums( 1 ).size() );
		assertNotNull( this.allocationManager.getAllocationEntry( communityGum1.getID() ) );
		assertFalse( this.allocationManager.getRemoteInUse().isEmpty() );

		// receives the community GuM2 for inexistent request
		this.ebPeerManager.hereIsGum( this.communityGum2, objectID2, 987 );
		assertNull( this.allocationManager.getAllocationEntry( communityGum2.getID() ) );
		assertTrue( this.allocationManager.getRemoteIdle().isEmpty() );
	}


	/**
	 * Verifies if the <code>hereIsGuM()</code> called by
	 * <code>EBLocalGuMPFacade</code> when there is more requests is working
	 * correctly.
	 * 
	 * @throws RequestAlreadyExistsException Indicates that a request already
	 *         exists.
	 * @throws RequestDoesNotExistException
	 */
	public void testHereIsGuMForLocalGuMs() throws RequestAlreadyExistsException, RequestDoesNotExistException {

		// creates other requests
		RequestSpec request1 = new RequestSpec( 2, "os == linux", 2, this.localFakeConsumer );
		RequestSpec request2 = new RequestSpec( 3, "os == windows", 1, this.localFakeConsumer );
		RequestSpec request3 = new RequestSpec( 1, "os == windows", 1, this.localFakeConsumer );

		request1.setAsFromLocalBroker();
		request2.setAsFromLocalBroker();
		request3.setAsFromLocalBroker();

		this.requestManager.createRequestEntry( request1 );
		this.requestManager.createRequestEntry( request2 );
		this.requestManager.createRequestEntry( request3 );

		// receives the GuM1 for request
		this.ebPeerManager.hereIsLocalGum( this.localGum1, this.localGum1.getGumSpec() ); // linux
		// check the maps
		assertEquals( 1, this.requestManager.getRequestGums( request1.getID() ).size() );
		assertEquals( 2, this.allocationManager.getAllocationEntry( localGum1.getID() ).getRequestID() );

		// receives the GuM2 for request
		ebPeerManager.hereIsLocalGum( this.localGum2, this.localGum2.getGumSpec() ); // windows
		// check the maps
		assertEquals( 1, this.requestManager.getRequestGums( request2.getID() ).size() );
		assertEquals( 3, this.allocationManager.getAllocationEntry( localGum2.getID() ).getRequestID() );

		// receives the GuM3 for request
		ebPeerManager.hereIsLocalGum( this.localGum3, this.localGum3.getGumSpec() ); // windows
		// check the maps
		assertEquals( 1, this.requestManager.getRequestGums( request3.getID() ).size() );
		assertEquals( 1, this.allocationManager.getAllocationEntry( localGum3.getID() ).getRequestID() );

		// receives the GuM4 for request
		this.ebPeerManager.hereIsLocalGum( this.localGum4, this.localGum4.getGumSpec() ); // linux
		// check the maps
		assertEquals( 2, this.requestManager.getRequestGums( request1.getID() ).size() );
		assertEquals( 2, this.allocationManager.getAllocationEntry( localGum4.getID() ).getRequestID() );
	}


	/**
	 * Verifies if the <code>hereIsGuM()</code> called by community is working
	 * correctly.
	 * 
	 * @throws RequestAlreadyExistsException Indicates that a request already
	 *         exists.
	 * @throws RequestDoesNotExistException
	 */
	public void testHereIsGuMForCommunityGuMs() throws RequestAlreadyExistsException, RequestDoesNotExistException {

		// creates other requests

		RequestSpec request1 = new RequestSpec( 1, "os == windows", 1, this.localFakeConsumer );
		RequestSpec request2 = new RequestSpec( 2, "os == windows", 1, this.communityFakeConsumer );
		RequestSpec request3 = new RequestSpec( 3, "os == linux", 2, this.localFakeConsumer );

		request1.setAsFromLocalBroker();
		request2.setAsFromRemotePeer();
		request3.setAsFromLocalBroker();
		this.requestManager.createRequestEntry( request1 );
		this.requestManager.createRequestEntry( request2 );
		this.requestManager.createRequestEntry( request3 );

		// receives the community GuM1 for request 3
		this.ebPeerManager.hereIsGum( this.communityGum1, objectID1, request3.getID() ); // linux
		// check the maps
		assertEquals( 1, this.requestManager.getRequestGums( request3.getID() ).size() );
		assertEquals( this.allocationManager.getAllocationEntry( communityGum1.getID() ).getRequestID(), request3
				.getID() );

		// receives the community GuM2 for request 1
		this.ebPeerManager.hereIsGum( this.communityGum2, objectID2, request1.getID() ); // windows
		// check the maps
		assertEquals( 1, this.requestManager.getRequestGums( request1.getID() ).size() );
		assertEquals( request1.getID(), this.allocationManager.getAllocationEntry( communityGum2.getID() )
				.getRequestID() );

		// receives the community GuM3 for community request 2
		this.ebPeerManager.hereIsGum( this.communityGum3, objectID3, request2.getID() ); // windows
		// check the maps
		assertTrue( this.requestManager.getRequestGums( request2.getID() ).isEmpty() );
		assertNull( this.allocationManager.getAllocationEntry( communityGum3.getID() ) );
		assertEquals( 0, this.allocationManager.getRemoteIdle().size() );

		// receives the community GuM4 for request 3
		this.ebPeerManager.hereIsGum( this.communityGum4, objectID4, request3.getID() ); // linux
		// check the maps
		assertEquals( 2, this.requestManager.getRequestGums( request3.getID() ).size() );
		assertEquals( request3.getID(), this.allocationManager.getAllocationEntry( communityGum4.getID() )
				.getRequestID() );
	}


	/**
	 * Verifies the <code>wannaGuMs()</code> and <code>hereIsGuM()</code>
	 * methods overloaded.
	 * 
	 * @throws RequestAlreadyExistsException Indicates that a request already
	 *         exists.
	 * @throws RequestDoesNotExistException
	 */
	public void testOverloadWannaGuMsHereIsGuM() throws RequestAlreadyExistsException, RequestDoesNotExistException {

		final int NUMBER_OF_GUMS = 100;
		// creates gums
		EBGum gums[] = this.createFakeGuMs( NUMBER_OF_GUMS );

		RequestSpec request1 = new RequestSpec( 123, "", NUMBER_OF_GUMS / 2, fakeBroker );
		this.ebPeerManager.requestGums( request1 );

		// receives the gums from local gump
		for ( int i = 0; i < NUMBER_OF_GUMS; i++ ) {
			this.ebPeerManager.hereIsLocalGum( gums[i], gums[i].getGumSpec() );
		}

		// check the number of gums
		assertEquals( NUMBER_OF_GUMS / 2, this.requestManager.getRequestGums( request1.getID() ).size() );

		// ask for gums
		RequestSpec request2 = new RequestSpec( 321, "", NUMBER_OF_GUMS / 2, fakeBroker );
		this.ebPeerManager.requestGums( request2 );

		// check the number of gums
		assertEquals( NUMBER_OF_GUMS / 2, this.requestManager.getRequestGums( request2.getID() ).size() );
	}


	/**
	 * Configure the initial configuration for <code>wannaGuMs()</code> tests.
	 * 
	 * @throws GumAlreadyAllocatedException
	 */
	private void configureForWannaGuMs() {

		// adding machines to allocationManager
		Map<String,String> attrib = generateAttributeMap( "linux", "512", "gum#2", "1000", peerName, peerPort );
		GumSpec gumSpec = new GumSpec( attrib );
		gumID2 = gumSpec.getGumID();

		allocationManager.gumReceived( new FakeGumFacade( gumSpec ), gumSpec );

		attrib = generateAttributeMap( "linux", "512", "gum#3", "1000", peerName, peerPort );
		gumSpec = new GumSpec( attrib );
		gumID3 = gumSpec.getGumID();

		allocationManager.gumReceived( new FakeGumFacade( gumSpec ), gumSpec );

		attrib = generateAttributeMap( "linux", "512", "gum#4", "1000", peerName, peerPort );
		gumSpec = new GumSpec( attrib );
		gumID4 = gumSpec.getGumID();

		allocationManager.gumReceived( new FakeGumFacade( gumSpec ), gumSpec );

		attrib = generateAttributeMap( "windows", "512", "gum#5", "1000", peerName, peerPort );
		gumSpec = new GumSpec( attrib );
		gumID5 = gumSpec.getGumID();

		allocationManager.gumReceived( new FakeGumFacade( gumSpec ), gumSpec );

		attrib = generateAttributeMap( "linux", "512", "gum#6", "1000", peerName, peerPort );
		gumSpec = new GumSpec( attrib );
		gumID6 = gumSpec.getGumID();

		allocationManager.gumReceived( new FakeGumFacade( gumSpec ), gumSpec );
	}


	/**
	 * Creates and returns an array containing
	 * <code>FakeGridMachineClient</code>s.
	 * 
	 * @param numberOfGuMs The number of gums to be created.
	 * @return The array with gums.
	 */
	private EBGum[ ] createFakeGuMs( int numberOfGuMs ) {

		EBGum gums[] = new FakeGumFacade[ numberOfGuMs ];
		Map<String,String> attrib = new HashMap<String,String>();
		for ( int i = 0; i < numberOfGuMs; i++ ) {
			attrib.put( GumSpec.ATT_NAME, "GuM" + i );
			attrib.put( GumSpec.ATT_PORT, "1000" );
			attrib.put( GumSpec.ATT_PEERNAME, peerName );
			attrib.put( GumSpec.ATT_PEERPORT, peerPort );
			gums[i] = new FakeGumFacade( new GumSpec( new HashMap<String,String>( attrib ) ) );
		}
		return gums;

	}


	/**
	 * Tests the basic functions os <code>wannaGuMs()</code> method.
	 * 
	 * @throws RequestAlreadyExistsException Indicates that a request already
	 *         exists.
	 * @throws RequestDoesNotExistException
	 */
	public void testWannaGuMs1() throws RequestAlreadyExistsException, RequestDoesNotExistException {

		// configures the community fake consumer
		this.communityFakeConsumer.setExpectedHereIsGum( 1 );

		// auxiliar method
		configureForWannaGuMs();

		assertFalse( requestManager.requestExists( 1 ) );

		RequestSpec request1 = new RequestSpec( 1, "os == linux", 10, fakeBroker );
		request1.setAsFromLocalBroker();
		ebPeerManager.requestGums( request1 );
		this.ebPeerManager.hereIsGum( this.communityGum1, objectID1, 1 );
		assertTrue( requestManager.requestExists( request1.getID() ) );
		assertEquals( 10, requestManager.getRequestSpec( request1.getID() ).getRequiredGums() );
		assertEquals( fakeBroker, requestManager.getRequestSpec( request1.getID() ).getGumpClient() );
		assertEquals( "os == linux", requestManager.getRequestSpec( request1.getID() ).getRequirements() );

		// verify if the correct machines was delivered
		assertTrue( allocationManager.getAllocationEntry( gumID2 ).isAllocated() );
		assertTrue( allocationManager.getAllocationEntry( gumID3 ).isAllocated() );
		assertTrue( allocationManager.getAllocationEntry( gumID6 ).isAllocated() );
		assertFalse( allocationManager.getAllocationEntry( gumID5 ).isAllocated() ); // attrib
		// doesn't match
		assertTrue( allocationManager.getAllocationEntry( gumID4 ).isAllocated() );
		assertTrue( allocationManager.getAllocationEntry( communityGum1.getID() ).isAllocated() );

		assertEquals( 5, requestManager.getRequestGums( request1.getID() ).size() );

		RequestSpec request2 = new RequestSpec( 2, "", 10, this.communityFakeConsumer );
		request2.setAsFromRemotePeer();
		ebPeerManager.requestGums( request2 );

		// This gum shouldn't be allocated for this request, since both are
		// remote.
		this.ebPeerManager.hereIsGum( this.communityGum2, objectID2, 2 );

		// verify if the correct machine was delivered
		assertEquals( 1, requestManager.getRequestGums( 2 ).size() );

		// check the community fake consumer
		ThreadPoolServices.waitUntilWorkIsDone();
		this.communityFakeConsumer.verify();

	}


	/**
	 * Tests if the <code>wannaGuMs()</code> method is delivering the correct
	 * number of GuMs.
	 * 
	 * @throws RequestAlreadyExistsException Indicates that a request already
	 *         exists.
	 * @throws RequestDoesNotExistException
	 */
	public void testWannaGuMs2() throws RequestAlreadyExistsException, RequestDoesNotExistException {

		// auxiliar method
		configureForWannaGuMs();

		assertFalse( requestManager.requestExists( 1 ) );
		RequestSpec requestSpec = new RequestSpec( 1, "os == linux", 2, fakeBroker );
		ebPeerManager.requestGums( requestSpec );
		assertTrue( requestManager.requestExists( 1 ) );

		// verify if the correct machines was delivered
		assertFalse( allocationManager.getAllocationEntry( gumID5 ).isAllocated() );
		assertEquals( 2, requestManager.getRequestGums( 1 ).size() );
	}


	/**
	 * Tests the preemption.
	 * 
	 * @throws RequestAlreadyExistsException
	 * @throws RequestAlreadyExistsException Indicates that a request already
	 *         exists.
	 * @throws RequestDoesNotExistException s
	 */
	public void testWannaGuMs3() throws RequestAlreadyExistsException, RequestDoesNotExistException {

		// auxiliar method
		configureForWannaGuMs();

		// creating a fake community request
		RequestSpec request1 = new RequestSpec( 56, "os == linux", 4, this.communityFakeConsumer );
		request1.setAsFromRemotePeer();
		requestManager.createRequestEntry( request1 );
		allocationManager.allocateGumToRequest( gumID2, 56, request1.isFromALocalBroker() );
		allocationManager.allocateGumToRequest( gumID6, 56, request1.isFromALocalBroker() );
		requestManager.hereIsGumToRequest( 56, gumID2 );
		requestManager.hereIsGumToRequest( 56, gumID6 );

		assertEquals( 2, allocationManager.getLocalDonated().size() );

		// Asking for gums for a local consumer
		RequestSpec request2 = new RequestSpec( 1, "os == linux", 10, fakeBroker );
		ebPeerManager.requestGums( request2 );

		// verify if the correct machines was delivered
		assertEquals( 4, requestManager.getRequestGums( 1 ).size() );
		assertEquals( 1, allocationManager.getAllocationEntry( gumID6 ).getRequestID() );
		assertEquals( 1, allocationManager.getAllocationEntry( gumID3 ).getRequestID() );
		assertEquals( 1, allocationManager.getAllocationEntry( gumID2 ).getRequestID() );
		assertEquals( 1, allocationManager.getAllocationEntry( gumID4 ).getRequestID() );
		assertFalse( allocationManager.getAllocationEntry( gumID5 ).isAllocated() );
	}


	/**
	 * This test verifies if the gums are been deallocated correctly.
	 * 
	 * @throws RequestAlreadyExistsException If the request id already exists.
	 */
	public void testWannaGuMs4() throws RequestAlreadyExistsException {

		// configures the community fake consumer
		this.communityFakeConsumer.setExpectedHereIsGum( 3 );

		// auxiliar method
		configureForWannaGuMs();

		RequestSpec request1 = new RequestSpec( 1, "", 3, this.communityFakeConsumer );
		request1.setAsFromRemotePeer();
		this.ebPeerManager.requestGums( request1 );
		assertEquals( 3, allocationManager.getLocalDonated().size() );

		RequestSpec request2 = new RequestSpec( 2, "", 2, fakeBroker );
		request2.setAsFromLocalBroker();
		this.ebPeerManager.requestGums( request2 );

		assertEquals( 2, allocationManager.getLocalInUse().size() );
		assertEquals( 3, allocationManager.getLocalDonated().size() );

		// check the community fake consumer
		ThreadPoolServices.waitUntilWorkIsDone();
		this.communityFakeConsumer.verify();
	}


	public void testPauseResume1() throws RequestAlreadyExistsException, RequestDoesNotExistException {

		RequestSpec request1 = new RequestSpec( 1, "os == linux", 10, fakeBroker );
		ebPeerManager.requestGums( request1 );
		assertTrue( requestManager.requestNeedMoreGums( 1 ) );
		ebPeerManager.pauseRequestGums( 1 );
		assertFalse( requestManager.requestNeedMoreGums( 1 ) );
		ebPeerManager.resumeRequestGums( 1 );
		assertTrue( requestManager.requestNeedMoreGums( 1 ) );

		RequestSpec request2 = new RequestSpec( 2, "os == windows", 10, this.communityFakeConsumer );
		ebPeerManager.requestGums( request2 );
		assertTrue( requestManager.requestNeedMoreGums( 2 ) );
		ebPeerManager.pauseRequestGums( 2 );
		assertFalse( requestManager.requestNeedMoreGums( 2 ) );
		ebPeerManager.resumeRequestGums( 2 );
		assertTrue( requestManager.requestNeedMoreGums( 2 ) );
	}


	public void testPauseResume2() throws RequestAlreadyExistsException, RequestDoesNotExistException {

		ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );
		ebPeerManager.hereIsLocalGum( localGum3, localGum3.getGumSpec() );

		RequestSpec request1 = new RequestSpec( 1, "", 5, fakeBroker );
		request1.setAsFromLocalBroker();
		ebPeerManager.requestGums( request1 );

		assertEquals( 3, this.requestManager.getRequestGums( request1.getID() ).size() );
		assertEquals( 3, this.allocationManager.getLocalInUse().size() );
		assertEquals( 0, this.allocationManager.getLocalDonated().size() );

		assertTrue( this.requestManager.requestNeedMoreGums( request1.getID() ) );

		ebPeerManager.pauseRequestGums( request1.getID() );

		assertFalse( this.requestManager.requestNeedMoreGums( request1.getID() ) );

		ebPeerManager.hereIsLocalGum( localGum4, localGum4.getGumSpec() );
		ebPeerManager.hereIsLocalGum( localGum5, localGum5.getGumSpec() );

		assertEquals( 3, this.requestManager.getRequestGums( request1.getID() ).size() );
		assertEquals( 3, this.allocationManager.getLocalInUse().size() );
		assertEquals( 2, this.allocationManager.getLocalIdle().size() );

		ebPeerManager.resumeRequestGums( request1.getID() );

		assertEquals( 5, this.requestManager.getRequestGums( request1.getID() ).size() );
		assertEquals( 5, this.allocationManager.getLocalInUse().size() );
		assertEquals( 0, this.allocationManager.getLocalIdle().size() );

		// false as it received 5 gums
		assertFalse( requestManager.requestNeedMoreGums( request1.getID() ) );

		RequestSpec request2 = new RequestSpec( 2, "", 5, fakeBroker2 );
		request2.setAsFromLocalBroker();
		ebPeerManager.requestGums( request2 );

		assertEquals( 3, this.requestManager.getRequestGums( request1.getID() ).size() );
		assertEquals( 2, this.requestManager.getRequestGums( request2.getID() ).size() );
		assertEquals( 5, this.allocationManager.getLocalInUse().size() );
		assertEquals( 0, this.allocationManager.getLocalIdle().size() );

		assertTrue( requestManager.requestNeedMoreGums( request1.getID() ) );

		ebPeerManager.pauseRequestGums( request2.getID() );

		ebPeerManager.finishRequestGums( request1.getID() );

		assertEquals( 2, this.requestManager.getRequestGums( request2.getID() ).size() );
		assertEquals( 2, this.allocationManager.getLocalInUse().size() );
		assertEquals( 3, this.allocationManager.getLocalIdle().size() );

		ebPeerManager.resumeRequestGums( request2.getID() );

		assertEquals( 5, this.requestManager.getRequestGums( request2.getID() ).size() );
		assertEquals( 5, this.allocationManager.getLocalInUse().size() );
		assertEquals( 0, this.allocationManager.getLocalIdle().size() );

		// false as it received 5 gums
		assertFalse( requestManager.requestNeedMoreGums( request2.getID() ) );

	}


	/**
	 * Verifies if the disposeOf() method for a non existent requestID is
	 * working correctly.
	 * 
	 * @throws RequestAlreadyExistsException If the request id already exists.
	 */
	public void testDisposeOfForNonExistentRequestID() throws RequestAlreadyExistsException {

		// auxiliar method
		this.configureForWannaGuMs();

		// check the maps
		assertFalse( this.allocationManager.getAllocationEntry( gumID2 ).isAllocated() );
		// dispose for a non existent request
		this.ebPeerManager.disposeGum( 999, gumID2 );
		// check the maps
		assertFalse( this.allocationManager.getAllocationEntry( gumID2 ).isAllocated() );

		// allocate the gum and dispose from a non existent id

		RequestSpec request1 = new RequestSpec( 1, "", 10, fakeBroker );
		this.ebPeerManager.requestGums( request1 );
		// check the maps
		assertTrue( this.allocationManager.getAllocationEntry( gumID2 ).isAllocated() );
		// dispose for a non existent request
		this.ebPeerManager.disposeGum( 999, gumID2 );
		// check the maps
		assertTrue( this.allocationManager.getAllocationEntry( gumID2 ).isAllocated() );
	}


	/**
	 * Verifies if the disposed local gums are been distributed correctly.
	 * 
	 * @throws RequestAlreadyExistsException If the request id already exists.
	 * @throws RequestDoesNotExistException
	 */
	public void testDisposeOfForLocalGuMs() throws RequestAlreadyExistsException, RequestDoesNotExistException {

		// configures the community fake consumer
		this.communityFakeConsumer.setExpectedHereIsGum( 1 );

		// auxiliar method
		this.configureForWannaGuMs();

		// gets all the gums
		RequestSpec request1 = new RequestSpec( 1, "", 10, fakeBroker );
		RequestSpec request2 = new RequestSpec( 2, "", 1, fakeBroker );
		RequestSpec request3 = new RequestSpec( 3, "", 1, this.communityFakeConsumer );
		request3.setAsFromRemotePeer();

		this.ebPeerManager.requestGums( request1 );
		this.ebPeerManager.requestGums( request2 );
		this.ebPeerManager.requestGums( request3 );
		// check if the local gums were received
		assertEquals( 5, this.requestManager.getRequestGums( 1 ).size() );
		assertEquals( 5, this.allocationManager.getLocalInUse().size() );
		assertEquals( 5, this.ebGumManager.getProxies().size() );

		// dispose before noMoreGuMs
		this.ebPeerManager.disposeGum( 1, gumID3 );

		// check if the gum came back
		assertEquals( 4, this.requestManager.getRequestGums( 1 ).size() );
		assertEquals( 1, this.requestManager.getRequestGums( 2 ).size() );
		assertEquals( 5, this.allocationManager.getLocalInUse().size() );
		assertEquals( 0, this.allocationManager.getLocalDonated().size() );

		// pause request gums
		this.ebPeerManager.pauseRequestGums( 1 );

		// dispose after noMoreGuMs
		this.ebPeerManager.disposeGum( 1, gumID3 );

		assertTrue( this.requestManager.getRequestGums( 2 ).contains( gumID3 ) );
		assertEquals( 5, this.allocationManager.getLocalInUse().size() );
		assertEquals( 0, this.allocationManager.getLocalDonated().size() );

		this.ebPeerManager.disposeGum( 1, gumID4 );
		assertTrue( this.requestManager.getRequestGums( 2 ).contains( gumID3 ) );
		assertTrue( this.requestManager.getRequestGums( 3 ).contains( gumID4 ) );
		assertEquals( 4, this.allocationManager.getLocalInUse().size() );
		assertEquals( 1, this.allocationManager.getLocalDonated().size() );

		this.ebPeerManager.disposeGum( 1, gumID6 );
		this.ebPeerManager.disposeGum( 1, gumID5 );
		assertEquals( 1, this.allocationManager.getLocalDonated().size() );

		// no more gums
		this.ebPeerManager.finishRequestGums( 1 );

		// check the maps
		try {
			this.requestManager.getRequestGums( 1 );
			fail( "Request does not exist, exception should be thrown" );
		} catch ( RequestDoesNotExistException e ) {

		}

		assertEquals( 1, this.allocationManager.getLocalInUse().size() );
		assertEquals( 1, this.allocationManager.getLocalDonated().size() );

		// dispose a gum as another request
		this.ebPeerManager.disposeGum( 1, gumID3 );
		this.ebPeerManager.disposeGum( 1, gumID4 );
		// check the maps
		assertTrue( this.requestManager.getRequestGums( 2 ).contains( gumID3 ) );
		assertTrue( this.requestManager.getRequestGums( 3 ).contains( gumID4 ) );
		assertEquals( 1, this.allocationManager.getLocalInUse().size() );
		assertEquals( 1, this.allocationManager.getLocalDonated().size() );
		assertEquals( 2, this.ebGumManager.getProxies().size() );

		// dispose a gum as another request
		this.ebPeerManager.disposeGum( 1, gumID3 );
		this.ebPeerManager.disposeGum( 1, gumID4 );
		// check the maps
		assertTrue( this.requestManager.getRequestGums( 2 ).contains( gumID3 ) );
		assertTrue( this.requestManager.getRequestGums( 3 ).contains( gumID4 ) );
		assertEquals( 1, this.allocationManager.getLocalInUse().size() );
		assertEquals( 1, this.allocationManager.getLocalDonated().size() );
		assertEquals( 2, this.ebGumManager.getProxies().size() );

		// check the community fake consumer
		ThreadPoolServices.waitUntilWorkIsDone();
		this.communityFakeConsumer.verify();
	}


	/**
	 * Verifies if the disposed community gums are been distributed correctly.
	 * 
	 * @throws RequestAlreadyExistsException If the request id already exists.
	 * @throws RequestDoesNotExistException
	 */
	public void testDisposeOfForCommunityGuMs() throws RequestAlreadyExistsException, RequestDoesNotExistException {

		// gives a community grid machine to a local request

		RequestSpec request1 = new RequestSpec( 1, "", 10, fakeBroker );
		RequestSpec request2 = new RequestSpec( 2, "", 10, fakeBroker );
		RequestSpec request3 = new RequestSpec( 123, "", 2, fakeBroker );
		this.ebPeerManager.requestGums( request1 );
		this.ebPeerManager.requestGums( request2 );
		this.ebPeerManager.requestGums( request3 );
		this.ebPeerManager.hereIsGum( this.communityGum1, objectID1, request3.getID() );
		// check the maps
		assertTrue( this.requestManager.getRequestGums( request3.getID() ).contains( communityGum1.getID() ) );
		assertEquals( 1, this.allocationManager.getRemoteInUse().size() );

		// dispose a gum as another request
		this.ebPeerManager.disposeGum( request1.getID(), communityGum1.getID() );
		this.ebPeerManager.disposeGum( request2.getID(), communityGum1.getID() );

		// check the maps
		assertTrue( this.requestManager.getRequestGums( request3.getID() ).contains( communityGum1.getID() ) );
		assertEquals( request1.getID(), this.allocationManager.getRemoteInUse().size() );

		// check the maps
		assertTrue( this.allocationManager.getRemoteIdle().isEmpty() );
		// dispose
		this.ebPeerManager.disposeGum( request3.getID(), communityGum1.getID() );
		// check the maps
		assertTrue( this.allocationManager.getRemoteIdle().isEmpty() );
	}


	/**
	 * Verifies the <code>lostGuM()</code> method called by local and
	 * community requests tries to provide another gums.
	 * 
	 * @throws RequestAlreadyExistsException If the request id already exists.
	 * @throws RequestDoesNotExistException
	 */
	public void testGumWasLost() throws RequestAlreadyExistsException, RequestDoesNotExistException {

		/* LOCAL REQUEST */
		// initial state
		RequestSpec request1 = new RequestSpec( 123, "", 2, fakeBroker );
		this.ebPeerManager.requestGums( request1 );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsGum( this.communityGum1, objectID1, 123 );
		this.ebPeerManager.hereIsLocalGum( localGum4, localGum4.getGumSpec() );

		assertEquals( 2, ebGumManager.getProxies().size() );

		// check the maps

		assertTrue( this.requestManager.getRequestGums( 123 ).contains( localGum1.getID() ) );
		assertTrue( this.requestManager.getRequestGums( 123 ).contains( communityGum1.getID() ) );
		assertFalse( this.requestManager.getRequestGums( 123 ).contains( localGum4.getID() ) );
		assertTrue( allocationManager.getAllocationEntry( localGum1.getID() ).isAllocated() );
		assertTrue( allocationManager.getAllocationEntry( communityGum1.getID() ).isAllocated() );
		assertFalse( allocationManager.getAllocationEntry( localGum4.getID() ).isAllocated() );
		assertFalse( this.requestManager.requestNeedMoreGums( 123 ) );

		this.ebPeerManager.gumWasLost( localGum1.getID() );

		assertEquals( 2, ebGumManager.getProxies().size() );

		// check the maps
		assertFalse( this.requestManager.getRequestGums( 123 ).contains( localGum1.getID() ) );
		assertTrue( this.requestManager.getRequestGums( 123 ).contains( communityGum1.getID() ) );
		assertTrue( this.requestManager.getRequestGums( 123 ).contains( localGum4.getID() ) );
		assertNull( allocationManager.getAllocationEntry( localGum1.getID() ) );
		assertTrue( allocationManager.getAllocationEntry( communityGum1.getID() ).isAllocated() );
		assertTrue( allocationManager.getAllocationEntry( localGum4.getID() ).isAllocated() );
		assertFalse( this.requestManager.requestNeedMoreGums( 123 ) );

		this.ebPeerManager.gumWasLost( communityGum1.getID() );

		assertEquals( 1, ebGumManager.getProxies().size() );

		// check the maps
		assertFalse( this.requestManager.getRequestGums( 123 ).contains( localGum1.getID() ) );
		assertFalse( this.requestManager.getRequestGums( 123 ).contains( communityGum1.getID() ) );
		assertTrue( this.requestManager.getRequestGums( 123 ).contains( localGum4.getID() ) );
		assertNull( allocationManager.getAllocationEntry( localGum1.getID() ) );
		assertTrue( allocationManager.getAllocationEntry( localGum4.getID() ).isAllocated() );
		assertNull( allocationManager.getAllocationEntry( communityGum1.getID() ) );
		assertTrue( this.requestManager.requestNeedMoreGums( 123 ) );

		this.ebPeerManager.gumWasLost( localGum4.getID() );

		assertEquals( 0, ebGumManager.getProxies().size() );

		// no more gums because the other request have to be attended
		this.ebPeerManager.finishRequestGums( 123 );

		/* COMMUNITY REQUEST */
		// configures the community fake consumer
		this.communityFakeConsumer.setExpectedHereIsGum( 3 );

		// initial state
		RequestSpec request2 = new RequestSpec( 521, "", 3, this.communityFakeConsumer );
		request2.setAsFromRemotePeer();
		this.ebPeerManager.requestGums( request2 );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum3, localGum3.getGumSpec() );

		// check the maps
		assertEquals( 3, ebGumManager.getProxies().size() );

		assertTrue( this.requestManager.getRequestGums( 521 ).contains( localGum2.getID() ) );
		assertTrue( this.requestManager.getRequestGums( 521 ).contains( localGum1.getID() ) );
		assertTrue( this.requestManager.getRequestGums( 521 ).contains( localGum3.getID() ) );
		assertTrue( allocationManager.getAllocationEntry( localGum1.getID() ).isAllocated() );
		assertTrue( allocationManager.getAllocationEntry( localGum2.getID() ).isAllocated() );
		assertTrue( allocationManager.getAllocationEntry( localGum3.getID() ).isAllocated() );
		assertFalse( this.requestManager.requestNeedMoreGums( 521 ) );

		// call the method
		this.ebPeerManager.gumWasLost( localGum3.getID() );

		// check the maps
		assertEquals( 2, ebGumManager.getProxies().size() );

		assertTrue( this.requestManager.getRequestGums( 521 ).contains( localGum1.getID() ) );
		assertTrue( this.requestManager.getRequestGums( 521 ).contains( localGum2.getID() ) );
		assertFalse( this.requestManager.getRequestGums( 521 ).contains( localGum3.getID() ) );
		assertTrue( allocationManager.getAllocationEntry( localGum1.getID() ).isAllocated() );
		assertTrue( allocationManager.getAllocationEntry( localGum2.getID() ).isAllocated() );
		assertNull( allocationManager.getAllocationEntry( localGum3.getID() ) );
		assertTrue( this.requestManager.requestNeedMoreGums( 521 ) );
		// call the method
		this.ebPeerManager.gumWasLost( localGum1.getID() );

		// check the maps
		assertEquals( 1, ebGumManager.getProxies().size() );

		assertFalse( this.requestManager.getRequestGums( 521 ).contains( localGum1.getID() ) );
		assertTrue( this.requestManager.getRequestGums( 521 ).contains( localGum2.getID() ) );
		assertFalse( this.requestManager.getRequestGums( 521 ).contains( localGum3.getID() ) );
		assertNull( allocationManager.getAllocationEntry( localGum1.getID() ) );
		assertTrue( allocationManager.getAllocationEntry( localGum2.getID() ).isAllocated() );
		assertNull( allocationManager.getAllocationEntry( localGum3.getID() ) );
		// call the method
		this.ebPeerManager.gumWasLost( localGum2.getID() );

		// check the maps
		assertEquals( 0, ebGumManager.getProxies().size() );

		assertFalse( this.requestManager.getRequestGums( 521 ).contains( localGum1.getID() ) );
		assertFalse( this.requestManager.getRequestGums( 521 ).contains( localGum2.getID() ) );
		assertFalse( this.requestManager.getRequestGums( 521 ).contains( localGum3.getID() ) );
		assertNull( allocationManager.getAllocationEntry( localGum1.getID() ) );
		assertNull( allocationManager.getAllocationEntry( localGum2.getID() ) );
		assertNull( allocationManager.getAllocationEntry( localGum3.getID() ) );

		// check the community fake consumer
		ThreadPoolServices.waitUntilWorkIsDone();
		this.communityFakeConsumer.verify();
	}


	public void testGumWasLost2() throws RemoteException, RequestAlreadyExistsException {

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );
		RemoteFakeGumpClient consumer2 = new RemoteFakeGumpClient( "consumer2" );

		RequestSpec request1 = new RequestSpec( reqID_1, "", 2, consumer1 );
		RequestSpec request2 = new RequestSpec( reqID_2, "", 2, consumer2 );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );

		this.ebPeerManager.requestGums( request1 );

		this.ebPeerManager.hereIsLocalGum( localGum3, localGum3.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum4, localGum4.getGumSpec() );

		this.ebPeerManager.requestGums( request2 );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );

		this.ebPeerManager.gumWasLost( localGum1.getGumSpec().getGumID() );

		assertEquals( 1, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );

		this.ebPeerManager.gumWasLost( localGum2.getGumSpec().getGumID() );

		assertEquals( 1, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
	}


	public void testLocalGumIsDeadWithNotAllocatedGums() {

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum3, localGum3.getGumSpec() );

		AllocationEntry entry1 = this.allocationManager.getAllocationEntry( localGum1.getGumSpec().getGumID() );
		AllocationEntry entry2 = this.allocationManager.getAllocationEntry( localGum2.getGumSpec().getGumID() );
		AllocationEntry entry3 = this.allocationManager.getAllocationEntry( localGum3.getGumSpec().getGumID() );

		assertFalse( entry1.isAllocated() );
		assertEquals( -1L, entry1.getRequestID() );

		assertFalse( entry2.isAllocated() );
		assertEquals( -1L, entry1.getRequestID() );

		assertFalse( entry3.isAllocated() );
		assertEquals( -1L, entry1.getRequestID() );

		this.ebPeerManager.gumWasLost( localGum1.getGumSpec().getGumID() );
		assertNull( this.allocationManager.getAllocationEntry( localGum1.getGumSpec().getGumID() ) );

		this.ebPeerManager.gumWasLost( localGum2.getGumSpec().getGumID() );
		assertNull( this.allocationManager.getAllocationEntry( localGum2.getGumSpec().getGumID() ) );

		this.ebPeerManager.gumWasLost( localGum3.getGumSpec().getGumID() );
		assertNull( this.allocationManager.getAllocationEntry( localGum3.getGumSpec().getGumID() ) );

		// test with a non existent gum
		this.ebPeerManager.gumWasLost( localGum4.getGumSpec().getGumID() );

	}


	/**
	 * The <code>redistributeGuMsByConsumer()</code> test.
	 * 
	 * @throws RequestAlreadyExistsException If the request id already exists.
	 * @throws RemoteException
	 */
	public void testRedistributeGums1() throws RequestAlreadyExistsException, RemoteException {

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );
		RemoteFakeGumpClient consumer2 = new RemoteFakeGumpClient( "consumer2" );
		RemoteFakeGumpClient consumer3 = new RemoteFakeGumpClient( "consumer3" );

		RequestSpec request1 = new RequestSpec( 1, "", 5, consumer1 );
		this.ebPeerManager.requestGums( request1 );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum3, localGum3.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum4, localGum4.getGumSpec() );

		assertEquals( 3, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );

		RequestSpec request2 = new RequestSpec( 2, "", 3, consumer2 );

		this.ebPeerManager.requestGums( request2 );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );

		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );

		RequestSpec request3 = new RequestSpec( 3, "", 3, consumer2 );
		this.ebPeerManager.requestGums( request3 );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		// consumer3 did not asked for gums yet
		assertEquals( 0, requestManager.getGums( consumer3, GumSource.LOCAL ).size() );

		RequestSpec request4 = new RequestSpec( 4, "", 2, consumer3 );
		this.ebPeerManager.requestGums( request4 );

		assertEquals( 1, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer3, GumSource.LOCAL ).size() );
	}


	/**
	 * The <code>redistributeGuMsByConsumer()</code> atest.
	 * 
	 * @throws RequestAlreadyExistsException If the request id already exists.
	 * @throws RemoteException
	 */
	public void testRedistributeGums2() throws RequestAlreadyExistsException, RemoteException {

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );
		RemoteFakeGumpClient consumer2 = new RemoteFakeGumpClient( "consumer2" );
		RemoteFakeGumpClient consumer3 = new RemoteFakeGumpClient( "consumer3" );

		RequestSpec request1 = new RequestSpec( 1, "", 5, consumer1 );
		this.ebPeerManager.requestGums( request1 );

		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer3, GumSource.LOCAL ).size() );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsGum( this.communityGum1, objectID1, 1 );
		this.ebPeerManager.hereIsLocalGum( localGum4, localGum4.getGumSpec() );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumer1, GumSource.REMOTE ).size() );
		assertEquals( 0, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer3, GumSource.LOCAL ).size() );

		RequestSpec request2 = new RequestSpec( 2, "", 3, consumer2 );
		this.ebPeerManager.requestGums( request2 );

		assertEquals( 1, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumer1, GumSource.REMOTE ).size() );
		assertEquals( 1, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer3, GumSource.LOCAL ).size() );

		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );

		assertEquals( 1, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumer1, GumSource.REMOTE ).size() );
		assertEquals( 2, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer3, GumSource.LOCAL ).size() );

		RequestSpec request3 = new RequestSpec( 3, "", 3, consumer3 );
		this.ebPeerManager.requestGums( request3 );

		assertEquals( 1, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumer1, GumSource.REMOTE ).size() );
		assertEquals( 1, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumer3, GumSource.LOCAL ).size() );

	}


	/**
	 * The <code>redistributeGuMsByConsumer()</code> atest.
	 * 
	 * @throws RequestAlreadyExistsException If the request id already exists.
	 * @throws RemoteException
	 */
	public void testRedistributeGums3() throws RequestAlreadyExistsException, RemoteException {

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );
		RemoteFakeGumpClient consumer2 = new RemoteFakeGumpClient( "consumer2" );

		RequestSpec request1 = new RequestSpec( 1, "", 5, consumer1 );
		this.ebPeerManager.requestGums( request1 );

		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum3, localGum3.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum4, localGum4.getGumSpec() );

		assertEquals( 3, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );

		RequestSpec request2 = new RequestSpec( 2, "", 3, consumer2 );
		this.ebPeerManager.requestGums( request2 );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );

		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );

		RequestSpec request3 = new RequestSpec( 3, "", 3, consumer2 );
		this.ebPeerManager.requestGums( request3 );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );

		RequestSpec request4 = new RequestSpec( 4, "", 7, consumer2 );
		this.ebPeerManager.requestGums( request4 );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );

	}


	public void testComputeCommunityRedistribution() {

		List<DistributionHelpEntry> testList1 = new Vector<DistributionHelpEntry>();

		DistributionHelpEntry h0 = new DistributionHelpEntry( 4.0, 0, null );
		DistributionHelpEntry h1 = new DistributionHelpEntry( 4.0, 0, null );
		DistributionHelpEntry h2 = new DistributionHelpEntry( 3.0, 0, null );
		DistributionHelpEntry h3 = new DistributionHelpEntry( 3.0, 0, null );
		DistributionHelpEntry h4 = new DistributionHelpEntry( 2.0, 0, null );
		DistributionHelpEntry h5 = new DistributionHelpEntry( 2.0, 0, null );
		DistributionHelpEntry h6 = new DistributionHelpEntry( 1.0, 0, null );
		DistributionHelpEntry h7 = new DistributionHelpEntry( 1.0, 0, null );
		DistributionHelpEntry h8 = new DistributionHelpEntry( 0.0, 4, null );
		DistributionHelpEntry h9 = new DistributionHelpEntry( 0.0, 4, null );

		int[ ] sequence0 = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
		int[ ] sequence1 = { 1, 1, 0, 0, 0, 0, 0, 0, 0, 0 };
		int[ ] sequence2 = { 2, 2, 1, 1, 1, 1, 0, 0, 0, 0 };
		int[ ] sequence3 = { 3, 3, 3, 3, 2, 2, 1, 0, 0, 0 };
		int[ ] sequence4 = { 6, 6, 5, 5, 3, 3, 1, 1, 0, 0 };
		int[ ] sequence5 = { 9, 9, 6, 6, 4, 4, 2, 2, 0, 0 };

		testList1.add( h0 );
		testList1.add( h1 );
		testList1.add( h2 );
		testList1.add( h3 );
		testList1.add( h4 );
		testList1.add( h5 );
		testList1.add( h6 );
		testList1.add( h7 );
		testList1.add( h8 );
		testList1.add( h9 );

		// Distribute 0 machines
		this.ebPeerManager.computeCommunityRedistribution( testList1, 0 );
		assertTrue( assertDistributionSequence( testList1, sequence0 ) );

		// Distribute 2 machines
		this.ebPeerManager.computeCommunityRedistribution( testList1, 2 );
		assertTrue( assertDistributionSequence( testList1, sequence1 ) );

		// Distribute 8 machines
		this.ebPeerManager.computeCommunityRedistribution( testList1, 8 );
		assertTrue( assertDistributionSequence( testList1, sequence2 ) );

		// Distribute 17 machines
		this.ebPeerManager.computeCommunityRedistribution( testList1, 17 );
		assertTrue( assertDistributionSequence( testList1, sequence3 ) );

		// Distribute 30 machines
		this.ebPeerManager.computeCommunityRedistribution( testList1, 30 );
		assertTrue( assertDistributionSequence( testList1, sequence4 ) );

		// Distribute 42 machines
		this.ebPeerManager.computeCommunityRedistribution( testList1, 42 );
		assertTrue( assertDistributionSequence( testList1, sequence5 ) );

	}


	private boolean assertDistributionSequence( List<DistributionHelpEntry> distribution, int[ ] sequence ) {

		assertEquals( distribution.size(), sequence.length );

		int i = 0;
		for ( DistributionHelpEntry helper : distribution )
			if ( helper.getGumsToReceive() != sequence[i++] )
				return false;

		return true;
	}


	public void testRedistributeGums4() throws RequestAlreadyExistsException, InterruptedException, RemoteException {

		ObjectID peerID1 = new ObjectID( "consumer1", new Long( "123" ), "rmi://consumer1" );
		ObjectID peerID2 = new ObjectID( "consumer2", new Long( "234" ), "rmi://consumer2" );
		ObjectID peerID_FR = new ObjectID( "consumerFR", new Long( "456" ), "rmi://consumerFR" );

		RemoteFakeGumpClient localConsumer1 = new RemoteFakeGumpClient( "localConsumer" );
		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( peerID1 );
		RemoteFakeGumpClient consumer2 = new RemoteFakeGumpClient( peerID2 );
		RemoteFakeGumpClient consumerFR = new RemoteFakeGumpClient( peerID_FR );

		RequestSpec request1_1 = new RequestSpec( 111, "", 1, consumer1 );
		request1_1.setAsFromRemotePeer();

		RequestSpec request1_2 = new RequestSpec( 222, "", 3, consumer1 );
		request1_2.setAsFromRemotePeer();

		RequestSpec request2_1 = new RequestSpec( 333, "", 2, consumer2 );
		request2_1.setAsFromRemotePeer();

		RequestSpec request2_2 = new RequestSpec( 444, "", 2, consumer2 );
		request2_2.setAsFromRemotePeer();

		RequestSpec request_local_1 = new RequestSpec( 555, "", 2, localConsumer1 );
		request_local_1.setAsFromLocalBroker();

		RequestSpec requestFR_1 = new RequestSpec( 666, "", 3, consumerFR );
		requestFR_1.setAsFromRemotePeer();

		RequestSpec requestFR_2 = new RequestSpec( 777, "", 2, consumerFR );
		requestFR_2.setAsFromRemotePeer();

		RequestSpec requestFR_3 = new RequestSpec( 888, "", 1, consumerFR );
		requestFR_3.setAsFromRemotePeer();

		accountingManager.startAccounting( localGum1.getID(), peerID1 );
		Thread.sleep( 98 ); // balance will be 0.1
		accountingManager.finishAccounting( localGum1.getID(), peerID1, false );
		accountingManager.startAccounting( localGum2.getID(), peerID2 );
		Thread.sleep( 298 ); // balance will be 0.3
		accountingManager.finishAccounting( localGum2.getID(), peerID2, false );

		/* Set gums */
		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum3, localGum3.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum4, localGum4.getGumSpec() );

		/* Free rider will get 3 machines */
		accountingManager.registerPeerIdForRequest( requestFR_1.getID(), peerID_FR );
		this.ebPeerManager.requestGums( requestFR_1 );

		assertEquals( 3, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

		/* Peer with balance 0.1 will get all machines, with two requests */
		accountingManager.registerPeerIdForRequest( request1_1.getID(), peerID1 );
		this.ebPeerManager.requestGums( request1_1 );

		accountingManager.registerPeerIdForRequest( request1_2.getID(), peerID1 );
		this.ebPeerManager.requestGums( request1_2 );

		assertEquals( 4, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

		/* Peer with the highest balance will ask for machines */
		accountingManager.registerPeerIdForRequest( request2_1.getID(), peerID2 );
		this.ebPeerManager.requestGums( request2_1 );

		accountingManager.registerPeerIdForRequest( request2_2.getID(), peerID2 );
		this.ebPeerManager.requestGums( request2_2 );

		assertEquals( 1, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 3, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

		/* Local request */
		this.ebPeerManager.requestGums( request_local_1 );

		assertEquals( 2, requestManager.getGums( localConsumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

		// The free-rider peer should get no gums
		accountingManager.registerPeerIdForRequest( requestFR_2.getID(), peerID_FR );
		this.ebPeerManager.requestGums( requestFR_2 );

		assertEquals( 2, requestManager.getGums( localConsumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

		accountingManager.registerPeerIdForRequest( requestFR_3.getID(), peerID_FR );
		this.ebPeerManager.requestGums( requestFR_3 );

		assertEquals( 2, requestManager.getGums( localConsumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

	}


	public void testRedistributeGums5() throws RequestAlreadyExistsException, InterruptedException, RemoteException {

		ObjectID peerID1 = new ObjectID( "consumer1", new Long( "123" ), "rmi://consumer1" );
		ObjectID peerID2 = new ObjectID( "consumer2", new Long( "234" ), "rmi://consumer2" );
		ObjectID peerID_FR = new ObjectID( "consumerFR", new Long( "456" ), "rmi://consumerFR" );

		RemoteFakeGumpClient localConsumer1 = new RemoteFakeGumpClient( "localConsumer" );
		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( peerID1 );
		RemoteFakeGumpClient consumer2 = new RemoteFakeGumpClient( peerID2 );
		RemoteFakeGumpClient consumerFR = new RemoteFakeGumpClient( peerID_FR );

		// Create requests
		RequestSpec request1_1 = new RequestSpec( 111, "os==linux && mem >= 512", 3, consumer1 );
		request1_1.setAsFromRemotePeer();
		accountingManager.registerPeerIdForRequest( request1_1.getID(), peerID1 );

		RequestSpec request1_2 = new RequestSpec( 222, "os==linux", 2, consumer1 );
		request1_2.setAsFromRemotePeer();
		accountingManager.registerPeerIdForRequest( request1_2.getID(), peerID1 );

		RequestSpec request2_1 = new RequestSpec( 333, "os==linux && mem >= 1024", 2, consumer2 );
		request2_1.setAsFromRemotePeer();
		accountingManager.registerPeerIdForRequest( request2_1.getID(), peerID2 );

		RequestSpec request2_2 = new RequestSpec( 444, "mem >= 512", 3, consumer2 );
		request2_2.setAsFromRemotePeer();
		accountingManager.registerPeerIdForRequest( request2_2.getID(), peerID2 );

		RequestSpec request_local_1 = new RequestSpec( 555, "os=linux", 2, localConsumer1 );
		request_local_1.setAsFromLocalBroker();

		RequestSpec requestFR_1 = new RequestSpec( 666, "", 3, consumerFR );
		requestFR_1.setAsFromRemotePeer();
		accountingManager.registerPeerIdForRequest( requestFR_1.getID(), peerID_FR );

		RequestSpec requestFR_2 = new RequestSpec( 777, "", 2, consumerFR );
		requestFR_2.setAsFromRemotePeer();
		accountingManager.registerPeerIdForRequest( requestFR_2.getID(), peerID_FR );

		RequestSpec requestFR_3 = new RequestSpec( 888, "", 1, consumerFR );
		requestFR_3.setAsFromRemotePeer();
		accountingManager.registerPeerIdForRequest( requestFR_3.getID(), peerID_FR );

		accountingManager.startAccounting( localGum1.getID(), peerID1 );
		Thread.sleep( 98 ); // balance will be 0.1
		accountingManager.finishAccounting( localGum1.getID(), peerID1, false );
		accountingManager.startAccounting( localGum2.getID(), peerID2 );
		Thread.sleep( 298 ); // balance will be 0.3
		accountingManager.finishAccounting( localGum2.getID(), peerID2, false );

		// Set 5 gums
		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum3, localGum3.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum4, localGum4.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum5, localGum5.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum6, localGum6.getGumSpec() );

		// Consumer 1 will get 2 machines
		this.ebPeerManager.requestGums( request1_1 );

		assertEquals( 0, requestManager.getGums( localConsumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

		// Consumer 1 will get 2 more machines
		this.ebPeerManager.requestGums( request1_2 );

		assertEquals( 0, requestManager.getGums( localConsumer1, GumSource.LOCAL ).size() );
		assertEquals( 4, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

		// Consumer 2 will get 1 machine and consumer 1 will lose 1
		this.ebPeerManager.requestGums( request2_1 );

		assertEquals( 0, requestManager.getGums( localConsumer1, GumSource.LOCAL ).size() );
		assertEquals( 3, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

		// Free rider will get 1 non-allocated machine
		this.ebPeerManager.requestGums( requestFR_1 );

		assertEquals( 0, requestManager.getGums( localConsumer1, GumSource.LOCAL ).size() );
		assertEquals( 3, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 1, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

		// Consumer 2 will get 2 machines. Free-rider will lose one. Consumer 1
		// will lose 1 machines
		this.ebPeerManager.requestGums( request2_2 );

		assertEquals( 0, requestManager.getGums( localConsumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 3, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

		// Local consumer will get 2 linux machines
		this.ebPeerManager.requestGums( request_local_1 );

		assertEquals( 2, requestManager.getGums( localConsumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 3, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );

		// Free-rider will try to be more machines will no success
		this.ebPeerManager.requestGums( requestFR_2 );
		this.ebPeerManager.requestGums( requestFR_3 );

		assertEquals( 2, requestManager.getGums( localConsumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 3, requestManager.getGums( consumer2, GumSource.LOCAL ).size() );
		assertEquals( 0, requestManager.getGums( consumerFR, GumSource.LOCAL ).size() );
	}


	public void testGumsAreAllocatedAfterRequestExists() throws RemoteException, RequestAlreadyExistsException {

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );

		RequestSpec request1 = new RequestSpec( reqID_1, "", 2, consumer1 );

		this.ebPeerManager.requestGums( request1 );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
	}


	public void testUnwantedGumsAreNotAllocated() throws RemoteException, RequestAlreadyExistsException,
		RequestDoesNotExistException {

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );

		RequestSpec request1 = new RequestSpec( reqID_1, "", 2, consumer1 );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );
		assertEquals( 2, allocationManager.getLocalIdle().size() );

		this.ebPeerManager.requestGums( request1 );
		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, ebGumManager.getProxies().size() );
		assertEquals( 0, allocationManager.getLocalIdle().size() );

		this.ebPeerManager.unwantedGum( reqID_1, localGum1.getID() );
		this.ebPeerManager.unwantedGum( reqID_1, localGum2.getID() );

		assertEquals( 2, allocationManager.getLocalIdle().size() );
		assertEquals( 0, ebGumManager.getProxies().size() );
		assertTrue( requestManager.getUnwantedGumsForRequest( reqID_1 ).contains( localGum1.getID() ) );
		assertTrue( requestManager.getUnwantedGumsForRequest( reqID_1 ).contains( localGum2.getID() ) );
		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
	}


	public void testUnwantedGumsWithDifferentRequest() throws RemoteException, RequestAlreadyExistsException,
		RequestDoesNotExistException {

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );

		RequestSpec request1 = new RequestSpec( reqID_1, "", 2, consumer1 );

		this.ebPeerManager.requestGums( request1 );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );

		this.ebPeerManager.preemptGum( localGum1.getID() );
		this.ebPeerManager.preemptGum( localGum2.getID() );

		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );

		this.ebPeerManager.unwantedGum( reqID_1, localGum1.getID() );
		this.ebPeerManager.unwantedGum( reqID_1, localGum2.getID() );

		assertTrue( requestManager.getUnwantedGumsForRequest( reqID_1 ).contains( localGum1.getID() ) );
		assertTrue( requestManager.getUnwantedGumsForRequest( reqID_1 ).contains( localGum2.getID() ) );
	}


	public void testLocalGumIsDeadAndUnwantedGum() throws RemoteException, RequestAlreadyExistsException,
		RequestDoesNotExistException {

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );

		RequestSpec request1 = new RequestSpec( reqID_1, "", 2, consumer1 );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );

		this.ebPeerManager.requestGums( request1 );
		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );

		this.ebPeerManager.unwantedGum( reqID_1, localGum1.getID() );
		this.ebPeerManager.unwantedGum( reqID_1, localGum2.getID() );
		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );

		assertTrue( requestManager.getUnwantedGumsForRequest( reqID_1 ).contains( localGum1.getID() ) );
		assertTrue( requestManager.getUnwantedGumsForRequest( reqID_1 ).contains( localGum2.getID() ) );

		this.ebPeerManager.gumWasLost( localGum1.getGumSpec().getGumID() );
		this.ebPeerManager.gumWasLost( localGum2.getGumSpec().getGumID() );

		assertTrue( requestManager.getUnwantedGumsForRequest( reqID_1 ).contains( localGum1.getID() ) );
		assertTrue( requestManager.getUnwantedGumsForRequest( reqID_1 ).contains( localGum2.getID() ) );
	}


	public void testRemoteUnwantedGum() throws RemoteException, RequestAlreadyExistsException, MalformedURLException,
		NotBoundException, RequestDoesNotExistException {

		this.communityObtainerFacade.config( communityObtainer, this.ebPeerManager );
		this.communityObtainerFacade.startEventProcessor();

		communityObtainer.reset();
		communityObtainer.setExpectedUnwantedGum( 1 );
		communityObtainer.setExpectedRequestGums( 1 );
		communityObtainer.setExpectedPauseRequestGums( 0 );
		communityObtainer.setExpectedDisposeGum( 0 );

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );

		GumSpec gumSpec = new GumSpec();
		gumSpec.putAttribute( GumSpec.ATT_NAME, "remoteg" );
		gumSpec.putAttribute( GumSpec.ATT_PEERNAME, "remote" );

		gumSpec.putAttribute( GumSpec.ATT_PORT, "1111" );
		gumSpec.putAttribute( GumSpec.ATT_PEERPORT, "1112" );
		gumSpec.putAttribute( GumSpec.ATT_TYPE, GumSpec.TYPE_FAKE_GUM );

		EBGum remoteGum = EBGumCreator.createGuM( gumSpec );

		RequestSpec request1 = new RequestSpec( reqID_1, "", 2, consumer1 );
		this.ebPeerManager.requestGums( request1 );

		this.allocationManager.gumReceived( remoteGum, gumSpec );
		this.requestManager.hereIsGumToRequest( reqID_1, remoteGum.getID() );
		this.allocationManager.allocateGumToRequest( remoteGum.getID(), reqID_1, request1.isFromALocalBroker() );

		assertEquals( 1, requestManager.getGums( consumer1, GumSource.REMOTE ).size() );

		this.ebPeerManager.unwantedGum( reqID_1, remoteGum.getID() );

		assertEquals( 0, requestManager.getGums( consumer1, GumSource.REMOTE ).size() );
		// The request manager is the local entity, so the remote request
		// manager must contais the unwanted gum
		assertFalse( requestManager.getUnwantedGumsForRequest( reqID_1 ).contains( remoteGum.getID() ) );

		ThreadPoolServices.waitUntilWorkIsDone();
		ThreadServices.waitUntilWorkIsDone();

		communityObtainer.verify();

		communityObtainer.reset();
		communityObtainer.setExpectedFinishRequestGums( 1 );
		this.ebPeerManager.finishRequestGums( reqID_1 );

		// Verifying if the community obtainer is called when the request is
		// finished
		ThreadPoolServices.waitUntilWorkIsDone();
		ThreadServices.waitUntilWorkIsDone();

		communityObtainer.verify();
		this.communityObtainerFacade.shutdown();
	}


	public void testConsumerIsDead() throws RemoteException, RequestAlreadyExistsException {

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );

		RequestSpec request1 = new RequestSpec( reqID_1, "", 2, consumer1 );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );

		this.ebPeerManager.requestGums( request1 );
		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, ebGumManager.getProxies().size() );

		this.ebPeerManager.consumerIsDead( consumer1.getObjectID() );
		assertEquals( 0, requestManager.getConsumerRequests( consumer1.getObjectID() ).size() );
		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 0, ebGumManager.getProxies().size() );
	}


	public void testFinishRequest() throws RemoteException, RequestAlreadyExistsException {

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );

		this.mockUIManager.registerNewMonitorable( Notifiables.CONSUMER, consumer1.getObjectID() );
		EasyMock.expectLastCall().times( 2 );
		EasyMock.replay( mockUIManager );

		RequestSpec request1 = new RequestSpec( reqID_1, "", 2, consumer1 );
		RequestSpec request2 = new RequestSpec( reqID_2, "", 2, consumer1 );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );
		this.ebPeerManager.hereIsLocalGum( localGum2, localGum2.getGumSpec() );

		this.ebPeerManager.requestGums( request1 );
		assertEquals( 2, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, ebGumManager.getProxies().size() );

		this.ebPeerManager.requestGums( request2 );
		this.ebPeerManager.finishRequestGums( reqID_1 );

		EasyMock.verify( mockUIManager );
		EasyMock.reset( mockUIManager );

		this.mockUIManager.unregisterNewMonitorable( Notifiables.CONSUMER, consumer1.getObjectID() );
		EasyMock.expectLastCall().times( 1 );
		EasyMock.replay( mockUIManager );

		this.ebPeerManager.finishRequestGums( reqID_2 );
		EasyMock.verify( mockUIManager );
	}
	

	public void testConsumerIsDeadWithRemoteGum() throws RemoteException, RequestAlreadyExistsException,
		MalformedURLException, NotBoundException, RequestDoesNotExistException {

		this.communityObtainerFacade.config( communityObtainer, this.ebPeerManager );
		this.communityObtainerFacade.startEventProcessor();

		communityObtainer.reset();
		communityObtainer.setExpectedRequestGums( 1 );
		communityObtainer.setExpectedFinishRequestGums( 1 );

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );

		GumSpec gumSpec = new GumSpec();
		gumSpec.putAttribute( GumSpec.ATT_NAME, "remoteg" );
		gumSpec.putAttribute( GumSpec.ATT_PEERNAME, "remote" );

		gumSpec.putAttribute( GumSpec.ATT_PORT, "1111" );
		gumSpec.putAttribute( GumSpec.ATT_PEERPORT, "1112" );
		gumSpec.putAttribute( GumSpec.ATT_TYPE, GumSpec.TYPE_FAKE_GUM );

		GumSpec gumSpec2 = new GumSpec();
		gumSpec2.putAttribute( GumSpec.ATT_NAME, "remoteg2" );
		gumSpec2.putAttribute( GumSpec.ATT_PEERNAME, "remote2" );

		gumSpec2.putAttribute( GumSpec.ATT_PORT, "1112" );
		gumSpec2.putAttribute( GumSpec.ATT_PEERPORT, "1212" );
		gumSpec2.putAttribute( GumSpec.ATT_TYPE, GumSpec.TYPE_FAKE_GUM );

		EBGum remoteGum = EBGumCreator.createGuM( gumSpec );
		this.ebGumManager.createAndSetProxy( remoteGum, reqID_1 );

		EBGum remoteGum2 = EBGumCreator.createGuM( gumSpec2 );
		this.ebGumManager.createAndSetProxy( remoteGum2, reqID_1 );

		RequestSpec request1 = new RequestSpec( reqID_1, "", 2, consumer1 );
		this.ebPeerManager.requestGums( request1 );

		this.allocationManager.gumReceived( remoteGum, gumSpec );
		this.requestManager.hereIsGumToRequest( reqID_1, remoteGum.getID() );
		this.allocationManager.allocateGumToRequest( remoteGum.getID(), reqID_1, request1.isFromALocalBroker() );

		this.allocationManager.gumReceived( remoteGum2, gumSpec2 );
		this.requestManager.hereIsGumToRequest( reqID_1, remoteGum2.getID() );
		this.allocationManager.allocateGumToRequest( remoteGum2.getID(), reqID_1, request1.isFromALocalBroker() );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.REMOTE ).size() );
		assertEquals( 2, allocationManager.getRemoteGums().size() );
		assertEquals( 2, ebGumManager.getProxies().size() );

		this.ebPeerManager.consumerIsDead( consumer1.getObjectID() );

		assertEquals( 0, requestManager.getGums( consumer1, GumSource.REMOTE ).size() );
		assertEquals( 0, allocationManager.getRemoteGums().size() );
		assertEquals( 0, ebGumManager.getProxies().size() );

		ThreadPoolServices.waitUntilWorkIsDone();
		ThreadServices.waitUntilWorkIsDone();

		communityObtainer.verify();
	}


	public void testConsumerIsDead3() throws RemoteException, RequestAlreadyExistsException,
		RequestDoesNotExistException, MalformedURLException, NotBoundException {

		this.communityObtainerFacade.config( communityObtainer, this.ebPeerManager );
		this.communityObtainerFacade.startEventProcessor();

		communityObtainer.reset();
		communityObtainer.setExpectedRequestGums( 1 );
		communityObtainer.setExpectedFinishRequestGums( 1 );

		RemoteFakeGumpClient consumer1 = new RemoteFakeGumpClient( "consumer1" );
		RequestSpec request1 = new RequestSpec( reqID_1, "", 3, consumer1 );

		GumSpec gumSpec = new GumSpec();
		gumSpec.putAttribute( GumSpec.ATT_NAME, "remoteg" );
		gumSpec.putAttribute( GumSpec.ATT_PEERNAME, "remote" );

		gumSpec.putAttribute( GumSpec.ATT_PORT, "1111" );
		gumSpec.putAttribute( GumSpec.ATT_PEERPORT, "1112" );
		gumSpec.putAttribute( GumSpec.ATT_TYPE, GumSpec.TYPE_FAKE_GUM );

		EBGumFacade remoteEBGumFacade = EBGumCreator.createGuM( gumSpec );

		GumSpec gumSpec2 = new GumSpec();
		gumSpec2.putAttribute( GumSpec.ATT_NAME, "remoteg2" );
		gumSpec2.putAttribute( GumSpec.ATT_PEERNAME, "remote2" );

		gumSpec2.putAttribute( GumSpec.ATT_PORT, "1112" );
		gumSpec2.putAttribute( GumSpec.ATT_PEERPORT, "1212" );
		gumSpec2.putAttribute( GumSpec.ATT_TYPE, GumSpec.TYPE_FAKE_GUM );

		EBGumFacade remoteEBGumFacade2 = EBGumCreator.createGuM( gumSpec2 );

		this.ebPeerManager.requestGums( request1 );

		this.allocationManager.gumReceived( remoteEBGumFacade, gumSpec );
		this.requestManager.hereIsGumToRequest( reqID_1, remoteEBGumFacade.getID() );
		this.allocationManager.allocateGumToRequest( remoteEBGumFacade.getID(), reqID_1, request1.isFromALocalBroker() );

		this.allocationManager.gumReceived( remoteEBGumFacade2, gumSpec2 );
		this.requestManager.hereIsGumToRequest( reqID_1, remoteEBGumFacade2.getID() );
		this.allocationManager.allocateGumToRequest( remoteEBGumFacade2.getID(), reqID_1, request1.isFromALocalBroker() );

		this.ebPeerManager.hereIsLocalGum( localGum1, localGum1.getGumSpec() );

		assertEquals( 2, requestManager.getGums( consumer1, GumSource.REMOTE ).size() );
		assertEquals( 1, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );
		assertEquals( 2, allocationManager.getRemoteGums().size() );
		assertEquals( 1, allocationManager.getLocalGums().size() );

		this.ebPeerManager.consumerIsDead( consumer1.getObjectID() );

		assertEquals( 0, requestManager.getGums( consumer1, GumSource.REMOTE ).size() );
		assertEquals( 0, allocationManager.getRemoteGums().size() );
		assertEquals( 0, requestManager.getConsumerRequests( consumer1.getObjectID() ).size() );
		assertEquals( 0, requestManager.getGums( consumer1, GumSource.LOCAL ).size() );

		assertFalse( remoteEBGumFacade.isStarted() );
		assertFalse( remoteEBGumFacade2.isStarted() );

		// Gum should not be deleted
		assertEquals( 1, allocationManager.getNumberOfLocalGums() );

		ThreadPoolServices.waitUntilWorkIsDone();
		ThreadServices.waitUntilWorkIsDone();

		communityObtainer.verify();
	}
	
	public void testHereIsGumOfSameMachine() throws Exception {
		
		// creates other requests
		RequestSpec request1 = new RequestSpec( 1, "os == linux", 2, this.localFakeConsumer );
		RequestSpec request2 = new RequestSpec( 2, "os == linux", 2, this.localFakeConsumer );

		this.requestManager.createRequestEntry( request1 );
		this.requestManager.createRequestEntry( request2 );
		
		this.ebPeerManager.hereIsGum( this.communityGum1, objectID1, request1.getID() );
		
		assertEquals( 1, this.requestManager.getRequestGums( request1.getID() ).size() );
		assertEquals( 0, this.requestManager.getRequestGums( request2.getID() ).size() );
		assertEquals( this.allocationManager.getAllocationEntry( communityGum1.getID() ).getRequestID(), request1
				.getID() );
		
		this.ebPeerManager.hereIsGum( this.communityGum1, objectID1, request1.getID() );

		assertEquals( 1, this.requestManager.getRequestGums( request1.getID() ).size() );
		assertEquals( 0, this.requestManager.getRequestGums( request2.getID() ).size() );
		assertEquals( this.allocationManager.getAllocationEntry( communityGum1.getID() ).getRequestID(), request1
				.getID() );
		
		this.ebPeerManager.hereIsGum( this.communityGum1, objectID1, request2.getID() );

		assertEquals( 0, this.requestManager.getRequestGums( request1.getID() ).size() );
		assertEquals( 1, this.requestManager.getRequestGums( request2.getID() ).size() );
		assertEquals( this.allocationManager.getAllocationEntry( communityGum1.getID() ).getRequestID(), request2
				.getID() );
	}
}
