/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager;

import java.io.Serializable;

import org.ourgrid.common.event.ActionEvent;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ShutdownEventEngine;
import org.ourgrid.common.event.ShutdownResponseEvent;

/**
 * This class is responsible for receiving events that come probably from a
 * <code>EBPeerManagerFacade</code> and process them.
 * 
 * @see EBPeerFacade
 */
public class PeerEventEngine implements ShutdownEventEngine, Serializable {

	private static final long serialVersionUID = 33L;

	/**
	 * A <CODE>EventQueue</CODE> that is filled by <CODE>putEvent()</CODE>
	 * method and provides the events to <CODE>run()</CODE> method consuming.
	 */
	private EventQueue<ActionEvent> eventQueue;

	/**
	 * Indicates <CODE>run()</CODE> method must shutdown.
	 */
	private boolean mustShutdown;

	/**
	 * When the processing (event consuming) shutdowns, this queue receives a
	 * <CODE>ShutdownNotifyEvent</CODE>. Such event indicates that shutdown
	 * is done.
	 */
	private EventQueue<ShutdownResponseEvent> shutdownEventQueue;

	/**
	 * Thread that performs <CODE>Event</CODE> consuming.
	 */
	private Thread myThread;

	/** Indicates if this <CODE>EventEngine</CODE> is alive. */
	private boolean isAlive;

	/** Indicates if event process time should be logged. */
	private boolean experiment = false;

	/**
	 * Logger to store log messages
	 */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( PeerEventEngine.class );


	/**
	 * Creates a new <CODE>PeerEventEngine</CODE>.
	 * 
	 * @param threadName a name for the event processor thread
	 */
	public PeerEventEngine( String threadName ) {

		this.eventQueue = new EventQueue<ActionEvent>();
		this.mustShutdown = false;
		this.isAlive = false;
		this.myThread = new Thread( this, threadName );
	}


	/**
	 * Begin consuming of events. Consuming means remove events from the queue
	 * and process (call the <CODE>process()</CODE> method) them.
	 */
	public void startProcessing() {

		this.isAlive = true;
		this.myThread.start();
	}


	/**
	 * Inserts a <CODE>Event</CODE>- to be processed - into a internal <CODE>
	 * EventQueue</CODE>.
	 * 
	 * @param event A <CODE>Event</CODE> to be processed.
	 */
	public void putEvent( ActionEvent event ) {

		this.eventQueue.put( event );
	}


	/**
	 * Consumes events from internal <CODE>EventQueue</CODE>. In other words,
	 * it removes events from the queue and process (call the <CODE>process()
	 * </CODE> method) them.
	 */
	public void run() {

		while ( !mustShutdown ) {
			ActionEvent actionEvent = eventQueue.blockingRemove();
			long t = 0;

			if ( experiment ) {
				t = System.currentTimeMillis();
			}

			try {
				actionEvent.process();
			} catch ( Exception e ) {
				LOG.error( "WARNING !!! " + e.getClass().getName() + ", " + e.getMessage() );
			}

			if ( experiment ) {
				LOG.debug( "Event " + actionEvent.getClass().getName() + ".process() duration: [ "
						+ (System.currentTimeMillis() - t) + " ms ]" );
			}
		}

		this.isAlive = false;
		this.shutdownEventQueue.put( new ShutdownResponseEvent() );
	}


	/**
	 * Marks this <code>Runnable</code> to be shutdown. It saves the parameter
	 * <CODE>EventQueue</CODE> to put a <CODE>ShutdownNotifyEvent</CODE>
	 * when the shutdown is done.
	 * 
	 * @param eq A <CODE>EventQueue</CODE> that receives a <CODE>
	 *        ShutdownNotifyEvent</CODE> when the shutdown is completed.
	 */
	public void shutdown( EventQueue<ShutdownResponseEvent> eq ) {

		this.shutdownEventQueue = eq;
		this.mustShutdown = true;
	}


	/**
	 * @see org.ourgrid.common.event.EBSyncShutdownable#isAlive()
	 */
	public boolean isAlive() {

		return this.isAlive;
	}

}
