/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.RequestEvent;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.event.ShutdownRequestEvent;
import org.ourgrid.common.event.ShutdownResponseEvent;
import org.ourgrid.test.PeerPreCommitTest;
import org.ourgrid.test.TestUtils;
import org.ourgrid.test.threadServices.ThreadServices;

/**
 * Tests the <code>PeerEventEngine</code> functions.
 */
public class PeerEventEngineTest extends TestCase implements PeerPreCommitTest {

	private PeerEventEngine peerEventProcessor;

	private TestUtils testUtils;


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createOGProperties();

		Configuration.getInstance( Configuration.PEER );

		peerEventProcessor = new PeerEventEngine( "PeerTest" );
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		Configuration.reset();
		testUtils.reset();
	}


	/**
	 * Tests if the events are been received by the <code>PeerEventEngine</code>.
	 */
	public void testAll() {

		EventQueue eventQueue = new EventQueue();
		RequestEvent requestEvent = new PeerTestRequestEvent( eventQueue );

		// Put Event, StartProcessing and isAlive
		peerEventProcessor.putEvent( requestEvent );
		ThreadServices.waitUntilWorkIsDone();

		Object response = eventQueue.unblockingRemove();
		assertNull( response );
		assertFalse( peerEventProcessor.isAlive() );

		peerEventProcessor.startProcessing();
		ThreadServices.waitUntilWorkIsDone();
		assertTrue( peerEventProcessor.isAlive() );
		response = eventQueue.unblockingRemove();
		assertNotNull( response );
		assertTrue( response instanceof PeerTestResponseEvent );

		int numberOfEventQueues = 10000;
		EventQueue[ ] eventQueues = new EventQueue[ numberOfEventQueues ];
		for ( int k = 0; k < eventQueues.length; k++ ) {
			eventQueues[k] = new EventQueue();
			RequestEvent re = new PeerTestRequestEvent( eventQueues[k] );
			peerEventProcessor.putEvent( re );
		}

		for ( int k = 0; k < eventQueues.length; k++ ) {
			Object responseEvent = eventQueues[k].blockingRemove();
			assertNotNull( responseEvent );
			assertTrue( responseEvent instanceof PeerTestResponseEvent );
		}

		// Shutdown, isAlive
		assertTrue( peerEventProcessor.isAlive() );
		peerEventProcessor.putEvent( new ShutdownRequestEvent( eventQueue, peerEventProcessor ) );
		ThreadServices.waitUntilWorkIsDone();
		assertFalse( peerEventProcessor.isAlive() );
		response = eventQueue.unblockingRemove();
		assertTrue( response instanceof ShutdownResponseEvent );
	}

	/**
	 * PeerTestRequestEvent This Event is used only for testing, it only puts a
	 * PeerTestResponseEvent on its ResponseQueue in order to test if the
	 * <code>process</code> method was invoked.
	 */
	private class PeerTestRequestEvent implements RequestEvent {

		private static final long serialVersionUID = 33L;

		/** An EventQueue */
		private EventQueue responseQueue;


		/**
		 * Constructs the Event.
		 * 
		 * @param responseQueue The EventQueue where the PeerTestResponseEvent
		 *        will be put.
		 */
		PeerTestRequestEvent( EventQueue responseQueue ) {

			this.responseQueue = responseQueue;
		}


		/**
		 * Puts a PeerTestResponseEvent on the responseQueue
		 * 
		 * @see org.ourgrid.common.event.ActionEvent#process()
		 */
		public void process() {

			responseQueue.put( new PeerTestResponseEvent() );
		}

	}

	/**
	 * PeerTestResponseEvent This Event is used only for testing. It is put on a
	 * EventQueue when the PeerTestRequestEvent's <code>process</code> method
	 * is invoked.
	 */
	private class PeerTestResponseEvent implements ResponseEvent {

		private static final long serialVersionUID = 33L;


		/**
		 * Returns Boolean.True.
		 * 
		 * @return Boolean.True.
		 * @see org.ourgrid.common.event.ResponseEvent#getResponse()
		 */
		public Object getResponse() {

			return Boolean.TRUE;
		}

	}

}
