/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager.allocation;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.peer.ebgum.EBGum;
import org.ourgrid.peer.manager.exception.RequestDoesNotExistException;
import org.ourgrid.peer.manager.request.RequestManager;

/**
 * This class represents an allocation in the peer. It indicates who is using
 * this machine and if the machine is local or from the Community.
 */
public class AllocationEntry {

	private long requestID = -1;

	private EBGum ebGum;

	private AllocationStatus status;

	private GumSource gumSource;

	private GumSpec gumSpec;


	// Constructor
	public AllocationEntry( EBGum ebGum, GumSpec gumSpec ) {

		this.ebGum = ebGum;
		this.gumSpec = gumSpec;
		this.status = AllocationStatus.IDLE;
		this.gumSource = getGumSource();
	}


	public AllocationStatus getStatus() {

		return this.status;
	}


	public GumSpec getGumSpec() {

		return this.gumSpec;
	}


	public boolean isAllocated() {

		return getStatus().isAllocated();
	}


	public boolean isAllocatedToRequest( long requestID ) {

		return getStatus().isAllocated() && this.requestID == requestID;
	}


	public boolean isALocalGum() {

		return this.gumSource.isLocal();
	}


	public boolean isARemoteGum() {

		return this.gumSource.isRemote();
	}


	public void allocate( long requestID, boolean requestIsLocal ) {

		if ( requestIsLocal ) {
			this.status = AllocationStatus.IN_USE;
		} else {
			this.status = AllocationStatus.DONATED;
		}

		this.requestID = requestID;
	}


	protected void deallocate() {

		this.requestID = -1;
		this.status = AllocationStatus.IDLE;
	}


	public EBGum getEBGum() {

		return this.ebGum;
	}


	public long getRequestID() {

		return this.requestID;
	}


	@Override
	public String toString() {

		return this.ebGum.getID().toString();
	}

	public enum GumSource {
		// indicates the gum source (from local peer or from remote ones)
		LOCAL, REMOTE;

		public boolean isLocal() {

			return (this.equals( LOCAL ));
		}


		public boolean isRemote() {

			return (this.equals( REMOTE ));
		}
	}


	public GumSource getGumSource() {

		String thisPeerName = Configuration.getInstance().getProperty( Configuration.PROP_EXTERNAL_NAME );
		String thisPeerPort = Configuration.getInstance().getProperty( Configuration.PROP_PORT );

		String gumPeerName = ebGum.getGumSpec().getAttribute( GumSpec.ATT_PEERNAME );
		String gumPeerPort = ebGum.getGumSpec().getAttribute( GumSpec.ATT_PEERPORT );

		if ( gumPeerName.equalsIgnoreCase( thisPeerName ) && gumPeerPort.equals( thisPeerPort ) ) {
			return GumSource.LOCAL;
		}
		return GumSource.REMOTE;
	}

}
