/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager.allocation;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.peer.ebgum.EBGum;
import org.ourgrid.peer.manager.allocation.AllocationEntry.GumSource;

/**
 * Keeps track of all the machines available to the peer.
 */
public class AllocationManager {

	private static AllocationManager singleInstance;

	private Map<GumID,AllocationEntry> remoteMachines;

	private Map<GumID,AllocationEntry> localMachines;

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( AllocationManager.class );


	public static AllocationManager getInstance() {

		if ( singleInstance == null ) {
			singleInstance = new AllocationManager();
		}
		return singleInstance;
	}


	protected AllocationManager() {

		this.localMachines = new HashMap<GumID,AllocationEntry>();
		this.remoteMachines = new HashMap<GumID,AllocationEntry>();
	}


	/**
	 * A new gum has been received
	 * 
	 * @param ebGum The event-base gum object representing the grid machine
	 * @param gumSpec The grid machine specification
	 * @return The newly created allocation, or an already existent, if any.
	 */
	public AllocationEntry gumReceived( EBGum ebGum, GumSpec gumSpec ) {

		GumID gumID = gumSpec.getGumID();
		ebGum.start();

		ebGum.stopReplica( new EventQueue<ResponseEvent>() );

		AllocationEntry allocationEntry = getAllocationEntry( gumID );

		if ( allocationEntry == null ) {

			allocationEntry = new AllocationEntry( ebGum, gumSpec );

			if ( allocationEntry.isALocalGum() )
				this.localMachines.put( gumID, allocationEntry );
			else
				this.remoteMachines.put( gumID, allocationEntry );

			LOG.debug( "New gum: " + gumID + " has been received" );
		}

		return allocationEntry;
	}


	public boolean gumExists( GumID gumID ) {

		return localMachines.containsKey( gumID ) || remoteMachines.containsKey( gumID );
	}
	

	/**
	 * A gum has to be disposed.
	 * 
	 * @param gumID the Id of the gum to be disposed.
	 */
	public void deallocateAndRemove( GumID gumID ) {

		AllocationEntry allocationEntry = getAllocationEntry( gumID );

		if ( allocationEntry != null ) {

			EBGum ebGum = allocationEntry.getEBGum();

			if ( ebGum != null ) {
				ebGum.stop();
			}

			if ( allocationEntry.isALocalGum() )
				this.localMachines.remove( gumID );
			else
				this.remoteMachines.remove( gumID );

			LOG.debug( "Gum: " + gumID + " has been removed" );
		}
	}


	public void allocateGumToRequest( GumID gumID, long requestID, boolean requestIsLocal ) {

		AllocationEntry allocationEntry;

		allocationEntry = this.getAllocationEntry( gumID );

		if ( allocationEntry != null ) {
			allocationEntry.allocate( requestID, requestIsLocal );
		}
	}


	/**
	 * Deallocates a gum from the current user. If the machine isn't allocate or
	 * doesn't exist at the allocations map the operation won't be concluded.
	 * 
	 * @param gumID The GridMachine name.
	 */
	public void deallocateGum( GumID gumID ) {

		AllocationEntry allocationEntry;

		allocationEntry = this.getAllocationEntry( gumID );

		if ( allocationEntry != null ) {
			allocationEntry.deallocate();
		}
	}


	public Collection<AllocationEntry> getAllocationEntries( GumSource gumSource, AllocationStatus status ) {

		Collection<AllocationEntry> allocationList;

		if ( gumSource.isLocal() ) {
			allocationList = getLocalGums();
		} else {
			allocationList = getRemoteGums();
		}

		Collection<AllocationEntry> allocationEntries = new LinkedList<AllocationEntry>();
		for ( AllocationEntry allocationEntry : allocationList ) {

			if ( allocationEntry.getStatus().equals( status ) ) {

				allocationEntries.add( allocationEntry );
			}
		}

		return allocationEntries;
	}


	public int getNumberOfLocalGums() {

		return localMachines.size();
	}


	/**
	 * @returns All local machines which are IDLE, DONATED or IN_USE.
	 */
	public Collection<AllocationEntry> getLocalGums() {

		return new LinkedList<AllocationEntry>( localMachines.values() );
	}


	/**
	 * @return All local machines with AllocationStatus.IN_USE status.
	 */
	public Collection<AllocationEntry> getLocalInUse() {

		return getAllocationEntries( GumSource.LOCAL, AllocationStatus.IN_USE );
	}


	/**
	 * @return All local machines with AllocationStatus.DONATED status.
	 */
	public Collection<AllocationEntry> getLocalDonated() {

		return getAllocationEntries( GumSource.LOCAL, AllocationStatus.DONATED );
	}


	/**
	 * @return All local machines with AllocationStatus.IDLE status.
	 */
	public Collection<AllocationEntry> getLocalIdle() {

		return getAllocationEntries( GumSource.LOCAL, AllocationStatus.IDLE );
	}


	/**
	 * @returns All remote machines which are IDLE or IN_USE by this peer.
	 */
	public Collection<AllocationEntry> getRemoteGums() {

		return new LinkedList<AllocationEntry>( remoteMachines.values() );
	}


	/**
	 * @return All remote machines with AllocationStatus.IN_USE status.
	 */
	public Collection<AllocationEntry> getRemoteInUse() {

		return getAllocationEntries( GumSource.REMOTE, AllocationStatus.IN_USE );
	}


	/**
	 * @return All remote machines with AllocationStatus.IDLE status.
	 */
	public Collection<AllocationEntry> getRemoteIdle() {

		return getAllocationEntries( GumSource.REMOTE, AllocationStatus.IDLE );
	}


	public AllocationEntry getAllocationEntry( GumID gumID ) {

		AllocationEntry allocationEntry;

		allocationEntry = this.localMachines.get( gumID );
		if ( allocationEntry == null ) {
			allocationEntry = this.remoteMachines.get( gumID );
		}
		return allocationEntry;
	}


	public static void reset() {

		singleInstance = null;
	}
}
