/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager.allocation;

import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.peer.ebgum.EBGum;
import org.ourgrid.peer.ebgum.FakeGumFacade;
import org.ourgrid.peer.manager.FakeEventEngine;
import org.ourgrid.peer.manager.FakeGumpClient;
import org.ourgrid.peer.manager.request.RequestManager;
import org.ourgrid.test.PeerFastTest;
import org.ourgrid.test.TestUtils;

public class AllocationManagerTest extends TestCase implements PeerFastTest {

	private static final int REMOTEREQUEST1 = 1254;

	private static final int LOCALREQUEST1 = 12346;

	private TestUtils testUtils;

	private Map<String,String> attributes, attributes2;

	private EBGum gridMachine;

	private EBGum cgridMachine;

	private EBGum cgum;

	private FakeGumFacade ebGum;

	private RequestSpec localRequest;

	private RequestSpec remoteRequest;


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();
		AllocationManager.reset();

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createOGProperties();

		Configuration.getInstance( Configuration.PEER );
		attributes = new HashMap<String,String>();
		attributes.put( GumSpec.ATT_NAME, "Gum1" );
		attributes.put( GumSpec.ATT_PORT, "0" );
		attributes.put( GumSpec.ATT_TYPE, GumSpec.TYPE_FAKE_GUM );

		attributes.put( GumSpec.ATT_PEERNAME, Configuration.getInstance()
				.getProperty( Configuration.PROP_EXTERNAL_NAME ) );
		attributes.put( GumSpec.ATT_PEERPORT, Configuration.getInstance().getProperty( Configuration.PROP_PORT ) );

		GumSpec gumSpec1 = new GumSpec( attributes );
		gridMachine = new FakeGumFacade( gumSpec1 );
		ebGum = new FakeGumFacade( gridMachine, new FakeEventEngine() );

		attributes2 = new HashMap<String,String>();
		attributes2.put( GumSpec.ATT_NAME, "Gum2" );
		attributes2.put( GumSpec.ATT_PORT, "0" );
		attributes2.put( GumSpec.ATT_TYPE, GumSpec.TYPE_FAKE_GUM );
		attributes2.put( GumSpec.ATT_PEERNAME, "externalpeer" );
		attributes2.put( GumSpec.ATT_PEERPORT, "6666" );

		GumSpec gumSpec2 = new GumSpec( attributes2 );
		cgridMachine = new FakeGumFacade( gumSpec2 );
		cgum = new FakeGumFacade( cgridMachine, new FakeEventEngine() );

		remoteRequest = new RequestSpec( REMOTEREQUEST1, "", 1, new FakeGumpClient( "" ) );
		remoteRequest.setAsFromRemotePeer();
		
		RequestManager.getInstance().createRequestEntry( remoteRequest );
		localRequest = new RequestSpec( LOCALREQUEST1, "", 1, new FakeGumpClient( "" ) );
		RequestManager.getInstance().createRequestEntry( localRequest );

	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		Configuration.reset();
		AllocationManager.reset();
		RequestManager.reset();
		testUtils.reset();
	}


	/**
	 * Tests the allocations and deallocations to local gums. It allocates the
	 * gum to local and for community request.
	 */
	public void testAllocateAndDeallocateLocalGuM() {

		assertFalse( ebGum.isStarted() );
		getAllocationManager().gumReceived( ebGum, ebGum.getGumSpec() );
		assertTrue( ebGum.isStarted() );

		// check the maps
		assertTrue( getAllocationManager().getLocalInUse().isEmpty() );
		assertTrue( getAllocationManager().getLocalDonated().isEmpty() );
		assertEquals( 1, getAllocationManager().getLocalIdle().size() );

		// allocates the gum to local
		getAllocationManager().allocateGumToRequest( ebGum.getID(), LOCALREQUEST1, localRequest.isFromALocalBroker() );
		
		// check the maps
		assertTrue( getAllocationManager().getAllocationEntry( ebGum.getID() ).isAllocated() );
		assertTrue( getAllocationManager().getAllocationEntry( ebGum.getID() ).isAllocatedToRequest( LOCALREQUEST1 ) );
		
		assertEquals( 1, getAllocationManager().getLocalInUse().size() );
		assertTrue( getAllocationManager().getLocalDonated().isEmpty() );
		assertTrue( getAllocationManager().getLocalIdle().isEmpty() );

		// deallocates the gum
		getAllocationManager().deallocateGum( ebGum.getID() );

		// check the maps
		assertFalse( getAllocationManager().getAllocationEntry( ebGum.getID() ).isAllocated() );
		assertEquals( 1, getAllocationManager().getLocalIdle().size() );
		assertTrue( ebGum.isStarted() );
	}


	/**
	 * Tests the allocations and deallocations to community gums. It allocates
	 * the gum to local request. ATTENTION: THIS TEST DO NOT PERMIT ALLOCATE A
	 * GUM FROM COMMUNITY FOR COMMUNITY.
	 */
	public void testAllocateAndDeallocateCommunityGuM() {

		// COMMUNITY gum was received

		getAllocationManager().gumReceived( cgum, cgum.getGumSpec() );
		// check the maps
		assertTrue( getAllocationManager().getLocalInUse().isEmpty() );
		assertTrue( getAllocationManager().getLocalDonated().isEmpty() );
		assertTrue( getAllocationManager().getRemoteInUse().isEmpty() );
		assertTrue( getAllocationManager().getLocalIdle().isEmpty() );
		assertEquals( 1, getAllocationManager().getRemoteIdle().size() );
	}


	/**
	 * Tests the getAllocationManager() receive and dispose functions.
	 */
	public void testReceivingAndDisposing() {

		assertNull( getAllocationManager().getAllocationEntry( ebGum.getID() ) );
		assertNotNull( getAllocationManager().gumReceived( ebGum, ebGum.getGumSpec() ) );
		assertTrue( ebGum.isStarted() );
		assertNotNull( getAllocationManager().getAllocationEntry( ebGum.getID() ) );
		getAllocationManager().allocateGumToRequest( ebGum.getID(), 1254, remoteRequest.isFromALocalBroker() );
		getAllocationManager().deallocateAndRemove( ebGum.getID() );
		assertFalse( ebGum.isStarted() );
		assertNull( getAllocationManager().getAllocationEntry( ebGum.getID() ) );
	}
	
	public void testExists() throws Exception {
		
		assertFalse( getAllocationManager().gumExists( ebGum.getID() ) );
		assertFalse( getAllocationManager().gumExists( cgridMachine.getID() ) );	                                                                    
		
		getAllocationManager().gumReceived( ebGum, ebGum.getGumSpec() );
		
		assertTrue( getAllocationManager().gumExists( ebGum.getID() ) );
		assertFalse( getAllocationManager().gumExists( cgridMachine.getID() ) );
		
		getAllocationManager().gumReceived( cgridMachine, cgridMachine.getGumSpec() );
		
		assertTrue( getAllocationManager().gumExists( ebGum.getID() ) );
		assertTrue( getAllocationManager().gumExists( cgridMachine.getID() ) );
		
		getAllocationManager().deallocateAndRemove( cgridMachine.getID() );
		
		assertTrue( getAllocationManager().gumExists( ebGum.getID() ) );
		assertFalse( getAllocationManager().gumExists( cgridMachine.getID() ) );
		
		getAllocationManager().deallocateAndRemove( ebGum.getID() );
		
		assertFalse( getAllocationManager().gumExists( ebGum.getID() ) );
		assertFalse( getAllocationManager().gumExists( cgridMachine.getID() ) );
	}


	protected AllocationManager getAllocationManager() {

		return AllocationManager.getInstance();
	}

}
