/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager.gummanager;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.peer.ebgum.EBGum;
import org.ourgrid.peer.ebgum.GumProxy;
import org.ourgrid.peer.manager.EBPeerManager;
import org.ourgrid.peer.manager.EBPeerManagerFacade;

/**
 * A <code>EBGumManagerImpl</code> controls GuMs that can be used by the peer.
 * Dealing with local GuMs, it is notified about states change at a GridMachine
 * that it knows and also notifies the <code>EBPeerManager</code>. When
 * dealing with received GuMs from another peer, it will create a Proxy and hold
 * information about its state. This is done by methods that create and remove
 * proxies. When the Grid Machine is lost, EBGumManagerImpl MUST receive an
 * invalidate call so it will remove the proxy.
 */
public class EBGumManagerImpl {

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( EBGumManagerImpl.class );

	/** A Map of gums associating a gum name with a gumEntry. */
	private Map<GumID,GumManagerEntry> gumManagerEntries;

	/** A Map of gums associating a gum name with a gumProxy. */
	protected Map<GumID,GumProxy> gumProxies;

	private Collection<GumSpec> gumSpecList;

	private EBPeerManager peerManager;

	private EBPeerManagerFacade peerManagerFacade;


	/**
	 * Default constructor of the <code>EBGumManagerImpl</code>.
	 */
	public EBGumManagerImpl( EBPeerManager peerManager, EBPeerManagerFacade peerManagerFacade ) {

		this.peerManager = peerManager;
		gumSpecList = new ArrayList<GumSpec>();
		gumManagerEntries = new HashMap<GumID,GumManagerEntry>();
		gumProxies = new ConcurrentHashMap<GumID,GumProxy>();
		this.peerManagerFacade = peerManagerFacade;
	}


	public void notifyGumStateChange( GumID gumID, GumStatus gumState ) {

		GumManagerEntry gumManagerEntry = gumManagerEntries.get( gumID );

		if ( gumManagerEntry == null ) {
			return; // We dont know this grid machine.
		}

		LOG.info( "Gum state change: " + gumID + ", " + gumState );

		if ( gumState.equals( GumStatus.CONTACTING ) || gumState.equals( GumStatus.OWNER ) ) {
			gumManagerEntry.setStatus( gumState );
			invalidate( gumID );
			peerManager.gumWasLost( gumID );
			LOG.info( "Gum is now down: " + gumManagerEntry.getGumName() );
		} else if ( gumState.equals( GumStatus.UP ) ) {
			gumManagerEntry.setStatus( gumState );

			if ( gumManagerEntry.isUp() ) {
				this.peerManager.hereIsLocalGum( gumManagerEntry.getEBGumFacade(), gumManagerEntry.getGumSpec() );
				LOG.info( "Gum is now UP: " + gumManagerEntry.getGumName() + " >>> " + gumID );
			}
		} else {
			LOG.error( "Unexpected state in this context :" + gumState );
		}
	}


	/**
	 * Sets the collection of gums to be managed.
	 * 
	 * @param gumSpecList A collection of <code>GuMSpec</code>s that
	 *        represents the gums to be managedgumManagerEntries.
	 */
	public void setGums( Collection<GumSpec> gumSpecList ) {

		this.gumSpecList = gumSpecList;

		deallocateOldGuMs();

		gumManagerEntries = new HashMap<GumID,GumManagerEntry>();

		Iterator<GumSpec> gumSpecListIt = gumSpecList.iterator();

		while ( gumSpecListIt.hasNext() ) {
			GumSpec gumSpec = gumSpecListIt.next();

			String peerName = Configuration.getInstance().getProperty( Configuration.PROP_EXTERNAL_NAME );
			String peerPort = Configuration.getInstance().getProperty( Configuration.PROP_PORT );

			gumSpec.putAttribute( GumSpec.ATT_PEERNAME, peerName );
			gumSpec.putAttribute( GumSpec.ATT_PEERPORT, peerPort );

			GumManagerEntry gumEntry = new GumManagerEntry( gumSpec );

			gumManagerEntries.put( gumSpec.getGumID(), gumEntry );

		}
		LOG.info( "A collection of GuMs has been set." );
	}


	private synchronized void deallocateOldGuMs() {

		Iterator<GumManagerEntry> gumIt = gumManagerEntries.values().iterator();
		while ( gumIt.hasNext() ) {
			GumManagerEntry gridMachineEntry = gumIt.next();
			peerManager.gumWasLost( gridMachineEntry.getGumID() );
		}
		LOG.info( "Old GuMs deallocated." );
	}


	/**
	 * Sets the proxy associated to the given EBGum.
	 * 
	 * @param requestID
	 * @throws RemoteException Was not possible create this proxy due to a
	 *         communication failure with the received Gum.
	 * @return The proxy that was just set.
	 */
	public GumProxy createAndSetProxy( EBGum ebGuM, long requestID ) throws RemoteException {

		final GumID gumID = ebGuM.getGumSpec().getGumID();

		invalidate( gumID );

		GumProxy proxy = new GumProxy( ebGuM, this.peerManagerFacade, requestID );
		gumProxies.put( gumID, proxy );

		return proxy;
	}


	/**
	 * Invalidates the proxy related to the given EBGridMachine.
	 * 
	 * @param gumID EBGridMachine which the related proxy must be invalidated
	 */
	public void invalidate( GumID gumID ) {

		GumProxy proxy = removeProxy( gumID );
		if ( proxy != null ) {
			proxy.invalidate();
		}
	}


	public GumProxy removeProxy( GumID gumID ) {

		GumProxy proxy = gumProxies.remove( gumID );

		return proxy;
	}


	public Map<GumID,GumProxy> getProxies() {

		return gumProxies;
	}


	public Collection<GumSpec> getOfflineGums() {

		List<GumSpec> offlineGumsList = new LinkedList<GumSpec>();
		Iterator<GumManagerEntry> itGums = gumManagerEntries.values().iterator();
		GumManagerEntry gumManagerEntry;

		while ( itGums.hasNext() ) {
			gumManagerEntry = itGums.next();
			if ( gumManagerEntry.getStatus().equals( GumStatus.CONTACTING ) ) {
				offlineGumsList.add( gumManagerEntry.getGumSpec() );
			}
		}

		return offlineGumsList;

	}


	public Collection<GumSpec> getOnlineGums() {

		List<GumSpec> onlineGumsList = new LinkedList<GumSpec>();
		Iterator<GumManagerEntry> itGums = gumManagerEntries.values().iterator();
		GumManagerEntry gumManagerEntry;

		while ( itGums.hasNext() ) {
			gumManagerEntry = itGums.next();
			if ( gumManagerEntry.getStatus().equals( GumStatus.UP ) ) {
				onlineGumsList.add( gumManagerEntry.getGumSpec() );
			}
		}
		return onlineGumsList;
	}


	public Collection<GumSpec> getOwnerGums() {

		List<GumSpec> ownerGumsList = new LinkedList<GumSpec>();
		Iterator<GumManagerEntry> itGums = gumManagerEntries.values().iterator();
		GumManagerEntry gumManagerEntry;

		while ( itGums.hasNext() ) {
			gumManagerEntry = itGums.next();

			if ( gumManagerEntry.getStatus().equals( GumStatus.OWNER ) ) {
				ownerGumsList.add( gumManagerEntry.getGumSpec() );
			}
		}
		return ownerGumsList;
	}


	protected Collection<GumSpec> getGumSpecs() {

		return gumSpecList;
	}
}
