/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager.gummanager;

import java.io.Serializable;
import java.net.MalformedURLException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.Map;

import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.peer.ebgum.EBGumCreator;
import org.ourgrid.peer.ebgum.EBGumFacade;

/**
 * GumManagerEntry It represents a GridMachine and its state in the Peer.
 */
public class GumManagerEntry implements Comparable<GumManagerEntry>, Serializable {

	/** Logger (log4j) object to store log events */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( GumManagerEntry.class );

	private static final long serialVersionUID = 33L;

	private GumStatus status;

	/** Represents if the GuM is allocated to some consumer */
	private boolean allocated;

	/** The EBGridMachine */
	private EBGumFacade ebGumFacade;

	/**
	 * <CODE>GuMSpec</CODE> that describes the <CODE>GridMachine</CODE> of
	 * this Entry.
	 */
	private GumSpec spec;


	/**
	 * The GUM will not be contacted during creation. Its initial status will be
	 * off-line and CONTACTING. The failure detector will notify the GumManager
	 * about the this GumManagerEntry state change
	 * 
	 * @param gumSpec GUM specification represented by this GumManagerEntry
	 */
	public GumManagerEntry( GumSpec gumSpec ) {

		this.spec = gumSpec;
		this.ebGumFacade = null;
		this.status = GumStatus.CONTACTING;
		this.allocated = false;
	}


	/**
	 * Returns the GridMachine's attributes.
	 * 
	 * @return The GridMachine's attributes.
	 */
	public Map<String,String> getAttributes() {

		return this.spec.getAttributes();
	}


	/**
	 * Returns the EBGridMachine.
	 * 
	 * @return The EBGridMachine or null if it could not be created.
	 */
	public EBGumFacade getEBGumFacade() {

		if ( this.ebGumFacade == null ) {
			Throwable error = null;
			try {
				this.ebGumFacade = EBGumCreator.createGuM( spec );
				this.status = GumStatus.UP;
				return this.ebGumFacade;
			} catch ( RemoteException e ) {
				error = e;
			} catch ( MalformedURLException e ) {
				error = e;
			} catch ( NotBoundException e ) {
				error = e;
			}

			this.status = GumStatus.CONTACTING;
			this.ebGumFacade = null;
			LOG.error( "Error creating ebGumFacade to gum: " + spec.getGumID() + ": ", error );
		}
		return this.ebGumFacade;
	}


	/**
	 * Returns the GridMachine name.
	 * 
	 * @return The GridMachine name.
	 */
	public String getGumName() {

		return this.getGumSpec().getGumID().getGumName();
	}


	public GumID getGumID() {

		return this.getGumSpec().getGumID();
	}


	/**
	 * Returns if the GridMachine is available or not. The GuM is available when
	 * it is <i>status </i> and it is <i>not allocated </i>.
	 * 
	 * @return <code>True</code> if it is available ( <i>status </i> and
	 *         <i>not allocated </i>). <code>False</code> otherwise.
	 */
	public boolean isAvailable() {

		return this.isUp() && !this.isAllocated();
	}


	/**
	 * Returns if the GridMachine is allocated or not.
	 * 
	 * @return <code>True</code> if the GuM is allocated. <code>False</code>
	 *         otherwise.
	 */
	public boolean isAllocated() {

		return this.allocated;
	}


	/**
	 * Returns if the GridMachine is status (working) or not.
	 * 
	 * @return <code>True</code> id the GuM is status. <code>False</code>
	 *         otherwise.
	 */
	public boolean isUp() {

		return this.getStatus().equals( GumStatus.UP ) && (ebGumFacade != null);
	}


	/**
	 * Returns if the GridMachine is status (working) or not.
	 * 
	 * @return <code>True</code> id the GuM is status. <code>False</code>
	 *         otherwise.
	 */
	public boolean isDown() {

		return this.getStatus().equals( GumStatus.CONTACTING );
	}


	/**
	 * Allocates the GridMachine - sets its <code>allocated</code> state to
	 * <code>true</code>.
	 */
	public void allocate() {

		this.allocated = true;
	}


	/**
	 * Deallocates the grid machine and stops the module's event processor
	 */
	public void deallocate() {

		this.allocated = false;
	}


	public void setStatus( GumStatus newState ) {

		this.status = newState;

		switch ( this.status ) {
			case CONTACTING:
				this.deallocate();
				this.ebGumFacade = null;
				break;
			case OWNER:
				this.deallocate();
				break;
			case UP:
				// Machine was down and is now UP, must reset the EBGum
				// reference
				this.deallocate();
				this.ebGumFacade = null;
				getEBGumFacade();
				break;
		}
	}


	/**
	 * Returns a String that describes this object.
	 * 
	 * @return A String that contains the GuM name and its attributes.
	 */
	public String toString() {

		return this.getGumName() + "\t" + this.spec.getAttributes();
	}


	/**
	 * Compares the name of Entries.
	 * 
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 * @param o Object to be compared to. It must be a <CODE>GridMachineEntry</CODE>.
	 * @return 0 if equals, 1 otherwise.
	 */
	public int compareTo( GumManagerEntry o ) {

		GumManagerEntry guME = o;
		if ( guME.getGumName().equals( this.getGumName() ) ) {
			return 0;
		}
		return 1;
	}


	/**
	 * Returns the <CODE>GuMSpec</CODE> that describes the <CODE>GridMachine</CODE>
	 * of this Entry.
	 * 
	 * @return The GridMachine Spec object.
	 */
	public GumSpec getGumSpec() {

		return this.spec;
	}


	public GumStatus getStatus() {

		return this.status;
	}
}
