/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager.gummanager;

import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.peer.ebgum.FakeGumFacade;
import org.ourgrid.test.PeerFastTest;

public class GumManagerEntryTest extends TestCase implements PeerFastTest {

	/** The name of GuM that will be used in the test. */
	private static final String GRID_MACHINE_NAME = "FakeGuM";

	/** The attributes of the GuM that will be used in the test. */
	private static Map<String,String> ATTRIBUTES;

	/**
	 * The construction of the attributes.
	 */
	static {
		ATTRIBUTES = new HashMap<String,String>();
		ATTRIBUTES.put( GumSpec.ATT_OS, "linux" );
		ATTRIBUTES.put( GumSpec.ATT_SITE, "lsd" );
		ATTRIBUTES.put( GumSpec.ATT_NAME, GRID_MACHINE_NAME );
		ATTRIBUTES.put( GumSpec.ATT_TYPE, GumSpec.TYPE_FAKE_GUM );
	}

	/** The object tested. */
	GumManagerEntry gridMachineEntry;

	private GumSpec guMSpec;


	/**
	 * Constructor for GridMachineEntryTest.
	 * 
	 * @param arg0
	 */
	public GumManagerEntryTest( String arg0 ) {

		super( arg0 );
	}


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();

		Configuration.getInstance( Configuration.MYGRID );

		guMSpec = new GumSpec( ATTRIBUTES );

		gridMachineEntry = new GumManagerEntry( guMSpec );
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		Configuration.reset();
	}


	/**
	 * Tests the <code>getAttributes()</code> method.
	 */
	public void testGetAttributes() {

		assertEquals( ATTRIBUTES, gridMachineEntry.getAttributes() );
	}


	/**
	 * Tests the <code>getGridMachine()</code> method.
	 */
	public void testGetGridMachine() {

		assertTrue( new FakeGumFacade( guMSpec ).equals( gridMachineEntry.getEBGumFacade() ) );
	}


	/**
	 * Tests the <code>isAvailable()</code> method.
	 */
	public void testIsAvailable() {

		// is the GridMachine is up and is not allocated, it is available.
		assertEquals( gridMachineEntry.isUp() && !gridMachineEntry.isAllocated(), gridMachineEntry.isAvailable() );

		// GridMachine starts down
		assertFalse( gridMachineEntry.isAvailable() );
		// allocates the GridMachine - it should not be available anymore.
		gridMachineEntry.setStatus( GumStatus.UP );
		gridMachineEntry.allocate();
		assertFalse( gridMachineEntry.isAvailable() );
		// deallocates the GridMachine - it should be available again.
		gridMachineEntry.deallocate();
		assertTrue( gridMachineEntry.isAvailable() );
		assertTrue( gridMachineEntry.isAvailable() );
		assertEquals( gridMachineEntry.isUp() && !gridMachineEntry.isAllocated(), gridMachineEntry.isAvailable() );
		gridMachineEntry.allocate();
		gridMachineEntry.allocate();
		assertFalse( gridMachineEntry.isAvailable() );
		assertFalse( gridMachineEntry.isAvailable() );
		gridMachineEntry.deallocate();
		gridMachineEntry.deallocate();
		assertTrue( gridMachineEntry.isAvailable() );

		assertTrue( gridMachineEntry.isAvailable() );
		// sets the GridMachine down - it should not be available anymore.
		gridMachineEntry.setStatus( GumStatus.CONTACTING );
		assertFalse( gridMachineEntry.isAvailable() );
		// sets the GridMachine up - it should be available again.
		gridMachineEntry.setStatus( GumStatus.UP );
		assertTrue( gridMachineEntry.isAvailable() );
		assertTrue( gridMachineEntry.isAvailable() );
		assertEquals( gridMachineEntry.isUp() && !gridMachineEntry.isAllocated(), gridMachineEntry.isAvailable() );
		gridMachineEntry.setStatus( GumStatus.CONTACTING );
		gridMachineEntry.setStatus( GumStatus.CONTACTING );
		assertFalse( gridMachineEntry.isAvailable() );
		assertFalse( gridMachineEntry.isAvailable() );
		gridMachineEntry.setStatus( GumStatus.UP );
		gridMachineEntry.setStatus( GumStatus.UP );
		assertTrue( gridMachineEntry.isAvailable() );
	}


	/**
	 * Tests the <code>up()</code>, <code>down()</code> and
	 * <code>isUp()</code> methods.
	 */
	public void testUpDownIsUp() {

		// when the GridMachineEntry is created, it is down.
		assertFalse( gridMachineEntry.isUp() );
		assertTrue( gridMachineEntry.isDown() );
		// the GridMachine is set down - its isUp() should return false.
		gridMachineEntry.setStatus( GumStatus.CONTACTING );
		assertFalse( gridMachineEntry.isUp() );
		assertTrue( gridMachineEntry.isDown() );
		gridMachineEntry.setStatus( GumStatus.UP );
		assertTrue( gridMachineEntry.isUp() );
		assertTrue( gridMachineEntry.isUp() );
		assertEquals( gridMachineEntry.isUp() && !gridMachineEntry.isAllocated(), gridMachineEntry.isUp() );
		gridMachineEntry.setStatus( GumStatus.CONTACTING );
		gridMachineEntry.setStatus( GumStatus.CONTACTING );
		assertFalse( gridMachineEntry.isUp() );
		assertTrue( gridMachineEntry.isDown() );
		assertFalse( gridMachineEntry.isUp() );
		assertTrue( gridMachineEntry.isDown() );
		gridMachineEntry.setStatus( GumStatus.UP );
		gridMachineEntry.setStatus( GumStatus.UP );
		assertTrue( gridMachineEntry.isUp() );
	}


	/**
	 * Tests the <code>allocate()</code>, <code>deallocate()</code> and
	 * <code>isAllocated()</code> methods.
	 */
	public void testAllocateDeallocateIsAllocated() {

		// when the GridMachineEntry is created, it is deallocated.
		assertFalse( gridMachineEntry.isAllocated() );
		gridMachineEntry.allocate();
		assertTrue( gridMachineEntry.isAllocated() );
		gridMachineEntry.deallocate();
		assertFalse( gridMachineEntry.isAllocated() );

		assertFalse( gridMachineEntry.isAllocated() );
		gridMachineEntry.allocate();
		gridMachineEntry.allocate();
		assertTrue( gridMachineEntry.isAllocated() );
		assertTrue( gridMachineEntry.isAllocated() );
		gridMachineEntry.deallocate();
		gridMachineEntry.deallocate();
		assertFalse( gridMachineEntry.isAllocated() );
	}

}
