/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager.request;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Set;

import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.peer.manager.allocation.AllocationEntry;
import org.ourgrid.peer.manager.allocation.AllocationManager;
import org.ourgrid.peer.manager.allocation.AllocationEntry.GumSource;

/**
 * Represents a request made by some consumer. It is used by the peer to store
 * information about a request. The instances of <code>RequestEntry</code> are
 * managed by the <code>RequestManager</code>
 * 
 * @see org.ourgrid.peer.manager.request.RequestManager
 */
public class RequestEntry {

	/**
	 * Contains the names of the grid machines that are currently allocated to
	 * this request.
	 */
	private Collection<GumID> gums;

	/**
	 * The current status of the request
	 */
	private RequestStatus status;

	/**
	 * The request specification contains all the information about the request
	 * provided by the consumer.
	 * 
	 * @see org.ourgrid.common.spec.RequestSpec
	 */
	private RequestSpec requestSpec;

	/**
	 * Gums that are not wanted by this request.
	 */
	private Set<GumID> unwantedGums;


	/**
	 * Default constructor.
	 * 
	 * @param requestSpec the request specification provided by the consumer
	 */
	public RequestEntry( RequestSpec requestSpec ) {

		this.requestSpec = requestSpec;
		this.gums = new LinkedHashSet<GumID>();
		this.unwantedGums = new HashSet<GumID>();
		this.status = RequestStatus.ALIVE;
	}


	public RequestSpec getSpec() {

		return this.requestSpec;
	}


	public Collection<GumID> getGums() {

		return new LinkedList<GumID>( gums );
	}


	public Collection<GumID> getGums( GumSource source ) {

		Collection<GumID> allGums = new LinkedList<GumID>();
		for ( GumID gumID : gums ) {

			AllocationEntry allocationEntry = AllocationManager.getInstance().getAllocationEntry( gumID );
			if ( allocationEntry != null && allocationEntry.getGumSource().equals( source ) ) {
				allGums.add( gumID );
			}
		}
		return allGums;
	}


	public void unwantedGum( GumID gumID ) {

		this.unwantedGums.add( gumID );
		this.gums.remove( gumID );
	}


	public boolean isGumUnwanted( GumID gumID ) {

		return this.unwantedGums.contains( gumID );
	}


	public Set<GumID> getUnwantedGums() {

		return new HashSet<GumID>( unwantedGums );
	}


	public int numberOfAllocatedGums() {

		return gums.size();
	}


	public RequestStatus getStatus() {

		return this.status;
	}


	public void setStatus( RequestStatus status ) {

		this.status = status;
	}


	public boolean isPaused() {

		return this.getStatus().isPaused();
	}


	public int getNeededGums() {

		if ( this.numberOfAllocatedGums() >= this.getSpec().getRequiredGums() )
			return 0;
		return this.getSpec().getRequiredGums() - this.numberOfAllocatedGums();
	}


	public boolean needMoreGums() {

		return this.getStatus().isAlive() && this.getSpec().getRequiredGums() > this.numberOfAllocatedGums();
	}


	public void disposeOfGum( GumID gumID ) {

		this.gums.remove( gumID );
	}


	public void hereIsGum( GumID gumID ) {

		this.gums.add( gumID );
	}


	public void pauseRequest() {

		setStatus( RequestStatus.PAUSED );
	}


	public void resumeRequest() {

		setStatus( RequestStatus.ALIVE );
	}


	public String toString() {

		return this.getSpec().toString() + " Status: " + getStatus();
	}

	public enum RequestStatus {

		PAUSED, ALIVE;

		public boolean isAlive() {

			return this.equals( ALIVE );
		}


		public boolean isPaused() {

			return this.equals( PAUSED );
		}
	}
}
