/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager.request;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.peer.manager.FakeGumpClient;
import org.ourgrid.test.PeerFastTest;
import org.ourgrid.test.TestUtils;

/**
 * Test for OurRequest class.
 */
public class RequestEntryTest extends TestCase implements PeerFastTest {

	/**
	 * The request object used in tests.
	 */
	private RequestEntry request;

	private TestUtils testUtils;


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createMGProperties();

		Configuration.getInstance( Configuration.PEER );

		RequestSpec requestSpec = new RequestSpec( 12345, "os == linux", 3, new FakeGumpClient(
				"rmi://localhost:3000/Test1" ) );
		requestSpec.setAsFromLocalBroker();
		request = new RequestEntry( requestSpec );
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		Configuration.reset();
		testUtils.reset();
	}


	/**
	 * Verifies if the <code>suspendWannaGuMsWasReceived()</code> and
	 * <code>resumeWannaGuMsWasReceived()</code> methods are working
	 * correctly.
	 */
	public void testSuspendResumeWannaGuMs() {

		// before a noMoreGuMs()
		assertTrue( request.needMoreGums() );
		request.pauseRequest();
		assertFalse( request.needMoreGums() );
		request.resumeRequest();
		assertTrue( request.needMoreGums() );
	}


	/**
	 * Tests if the grid machines are been correctly associated.
	 */
	public void testIsWithGridMachine() {

		GumID gumID1 = new GumID( "gm1", "0", "peername", "0" );
		GumID gumID2 = new GumID( "gm2", "0", "peername", "0" );
		GumID gumID3 = new GumID( "gm3", "0", "peername", "0" );

		request.hereIsGum( gumID1 );
		request.hereIsGum( gumID2 );
		assertTrue( request.getGums().contains( gumID2 ) );
		assertTrue( request.getGums().contains( gumID1 ) );
		request.disposeOfGum( gumID1 );
		assertFalse( request.getGums().contains( gumID1 ) );
		request.hereIsGum( gumID3 );
		request.disposeOfGum( gumID2 );
		assertFalse( request.getGums().contains( gumID2 ) );
	}


	/**
	 * Tests all the cases that grid machines are lost, disposed or received.
	 */
	public void testAll() {

		GumID gumID1 = new GumID( "gm1", "0", "peername", "0" );
		GumID gumID2 = new GumID( "gm2", "0", "peername", "0" );
		GumID gumID3 = new GumID( "gm3", "0", "peername", "0" );

		assertEquals( 3, request.getSpec().getRequiredGums() );
		request.hereIsGum( gumID1 );
		request.hereIsGum( gumID2 );
		request.hereIsGum( gumID3 );
		assertEquals( 3, request.getGums().size() );
		assertFalse( request.needMoreGums() );

		request.disposeOfGum( gumID1 );
		assertEquals( 2, request.getGums().size() );
		assertTrue( request.needMoreGums() );

		request.disposeOfGum( gumID2 );
		assertEquals( 1, request.getGums().size() );
		assertTrue( request.needMoreGums() );

		request.hereIsGum( gumID1 );
		request.hereIsGum( gumID2 );
		assertFalse( request.needMoreGums() );
		assertEquals( 3, request.getGums().size() );
	}


	public void testUnwantedGum() {

		GumID gumID1 = new GumID( "gm1", "0", "peername", "0" );
		GumID gumID2 = new GumID( "gm2", "0", "peername", "0" );

		assertFalse( this.request.isGumUnwanted( gumID1 ) );
		assertFalse( this.request.isGumUnwanted( gumID2 ) );

		this.request.unwantedGum( gumID1 );
		assertTrue( this.request.getUnwantedGums().contains( gumID1 ) );
		assertTrue( this.request.isGumUnwanted( (gumID1) ) );
		assertFalse( this.request.isGumUnwanted( gumID2 ) );

		this.request.unwantedGum( gumID2 );
		assertTrue( this.request.getUnwantedGums().contains( gumID2 ) );
		assertEquals( 2, this.request.getUnwantedGums().size() );
		assertTrue( this.request.isGumUnwanted( gumID1 ) );
		assertTrue( this.request.isGumUnwanted( gumID2 ) );
	}
}
