/*
 * Copyright (c) 2002-2004 Universidade Federal de Campina Grande
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ourgrid.peer.manager.request;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.ourgrid.common.exception.RequestAlreadyExistsException;
import org.ourgrid.common.gump.GumpClient;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.common.spec.RequestSpec.RequestSource;
import org.ourgrid.peer.manager.allocation.AllocationEntry.GumSource;
import org.ourgrid.peer.manager.exception.RequestDoesNotExistException;

/**
 * Keeps track of all the requests and manage them.
 */
public class RequestManager {

	private Map<Long,RequestEntry> requests;

	private static RequestManager instance = null;

	private static final transient org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( RequestManager.class );


	public static RequestManager getInstance() {

		if ( instance == null ) {
			instance = new RequestManager();
		}

		return instance;
	}


	private RequestManager() {

		// this map must be sorted
		this.requests = new LinkedHashMap<Long,RequestEntry>();
	}


	/**
	 * Creates and stores a request.
	 */
	public RequestEntry createRequestEntry( RequestSpec requestSpec ) throws RequestAlreadyExistsException {

		RequestEntry requestEntry;

		if ( this.requestExists( requestSpec.getID() ) ) {
			throw new RequestAlreadyExistsException( "Already exists a request with ID <" + requestSpec.getID() + ">" );
		}

		requestEntry = new RequestEntry( requestSpec );

		requests.put( requestSpec.getID(), requestEntry );

		return requestEntry;
	}


	public boolean requestExists( long requestID ) {

		return this.requests.containsKey( requestID );
	}


	/**
	 * Get the <code>GumpClient</code> of the given request id.
	 * 
	 * @throws RequestDoesNotExistException
	 */
	public GumpClient getGumpClient( long requestID ) throws RequestDoesNotExistException {

		RequestEntry requestEntry;

		requestEntry = this.getRequestEntry( requestID );

		return requestEntry.getSpec().getGumpClient();

	}


	/**
	 * Informs that this request was canceled by its consumer.
	 */
	public void deleteRequest( long requestID ) {

		this.requests.remove( requestID );
	}


	/**
	 * Indicates that a request does not want more gums as far. The request was
	 * paused.
	 */
	public void pauseRequestGums( long requestID ) throws RequestDoesNotExistException {

		RequestEntry requestEntry = this.getRequestEntry( requestID );

		requestEntry.pauseRequest();
	}


	/**
	 * Indicates that a request needs gums again.
	 */
	public void resumeRequestGums( long requestID ) throws RequestDoesNotExistException {

		RequestEntry requestEntry = this.getRequestEntry( requestID );

		requestEntry.resumeRequest();
	}


	/**
	 * Refresh this request about the <code>hereIsGuM()</code> command.
	 * 
	 * @throws RequestDoesNotExistException
	 */
	public void hereIsGumToRequest( long requestID, GumID gumID ) throws RequestDoesNotExistException {

		RequestEntry requestEntry = this.getRequestEntry( requestID );

		requestEntry.hereIsGum( gumID );
	}


	/**
	 * Refresh this request about the <code>dispose</code> instruction.
	 * 
	 */
	public void removeGumFromRequest( long requestID, GumID gumID ) {

		RequestEntry requestEntry;
		try {
			requestEntry = this.getRequestEntry( requestID );
			requestEntry.disposeOfGum( gumID );
			
		} catch ( RequestDoesNotExistException e ) {
			LOG.debug( "Trying to remove a gum from a nonexistent request" );
		}
		
		LOG.debug( "Gum: " + gumID + " removed from request " + requestID );
	}


	private RequestEntry getRequestEntry( long requestID ) throws RequestDoesNotExistException {

		RequestEntry request = requests.get( requestID );
		if ( request == null ) {
			throw new RequestDoesNotExistException( requestID );
		}
		return request;
	}


	public Collection<RequestEntry> getAliveRequests( RequestSource source ) {

		Collection<RequestEntry> allRequests = this.requests.values();
		Collection<RequestEntry> chosenRequests = new LinkedList<RequestEntry>();

		for ( RequestEntry request : allRequests )
			if ( request.getSpec().getSource().equals( source ) && request.getStatus().isAlive() )
				chosenRequests.add( request );

		return chosenRequests;
	}


	public Collection<RequestEntry> getRequestsWithAllocatedGums( RequestSource source ) {

		Collection<RequestEntry> allRequests = this.requests.values();
		Collection<RequestEntry> chosenRequests = new LinkedList<RequestEntry>();

		for ( RequestEntry request : allRequests )
			if ( request.getSpec().getSource().equals( source ) && request.numberOfAllocatedGums() > 0 )
				chosenRequests.add( request );

		return chosenRequests;
	}


	/**
	 * Recovers a <code>Collection</code> containing all
	 * <code>GumpClient</code> local <code>RequestEntry</code>
	 * 
	 * @return The collection containing the local requests.
	 */
	public Collection<RequestEntry> getRequests( GumpClient gumpClient ) {

		Collection<RequestEntry> allRequests = this.requests.values();
		Collection<RequestEntry> result = new LinkedList<RequestEntry>();

		for ( RequestEntry request : allRequests ) {
			if ( request.getSpec().getGumpClient().equals( gumpClient ) ) {
				result.add( request );
			}
		}

		return result;
	}


	/**
	 * Recovers a <code>Collection</code> containing all <code>PeerID</code>
	 * local <code>RequestEntry</code>
	 * 
	 * @return The collection containing the requests from PeerID.
	 */
	public Collection<RequestEntry> getConsumerRequests( ObjectID peerID ) {

		Collection<RequestEntry> allRequests = this.requests.values();
		Collection<RequestEntry> result = new LinkedList<RequestEntry>();

		for ( RequestEntry request : allRequests ) {

			ObjectID objID;

			objID = request.getSpec().getObjectID();
			if ( peerID != null && objID.equals( peerID ) ) {
				result.add( request );
			}
		}

		return result;

	}


	/**
	 * Recovers a <code>Collection</code> containing all <code>PeerID</code>
	 * local <code>RequestEntry</code>
	 * 
	 * @return The collection containing the requests from PeerID.
	 */
	public Collection<Long> getConsumerRequestsIDs( ObjectID consumerID ) {

		Collection<Long> result = new LinkedList<Long>();

		for ( RequestEntry request : this.requests.values() ) {

			if ( request.getSpec().getObjectID().equals( consumerID ) ) {
				result.add( new Long( request.getSpec().getID() ) );
			}
		}

		return result;

	}


	public Collection<GumID> getGums( GumpClient gumpClient, GumSource source ) {

		Collection<RequestEntry> requestEntries = getRequests( gumpClient );
		Collection<GumID> result = new LinkedList<GumID>();
		for ( RequestEntry request : requestEntries ) {
			result.addAll( request.getGums( source ) );
		}

		return result;
	}


	public Collection<GumID> getGums( GumpClient gumpClient ) {

		Collection<GumID> gums = getGums( gumpClient, GumSource.LOCAL );
		gums.addAll( getGums( gumpClient, GumSource.REMOTE ) );

		return gums;

	}


	public Collection<GumID> getGums( ObjectID peerID ) {

		Collection<GumID> gums = getGums( peerID, GumSource.LOCAL );
		gums.addAll( getGums( peerID, GumSource.REMOTE ) );

		return gums;
	}


	public Collection<GumID> getGums( ObjectID peerID, GumSource source ) {

		Collection<RequestEntry> requestEntries = getConsumerRequests( peerID );
		Collection<GumID> result = new LinkedList<GumID>();
		for ( RequestEntry request : requestEntries ) {
			result.addAll( request.getGums( source ) );
		}

		return result;
	}


	/**
	 * Gets a Collection containing all consumers (local or remote)
	 * 
	 * @param source Choose if the list to be returned will be composed by local
	 *        or remote consumers
	 * @return A collection with the selected consumers
	 */
	public Collection<GumpClient> getConsumers( RequestSource source ) {

		Collection<GumpClient> consumers = new LinkedList<GumpClient>();

		for ( RequestEntry request : this.requests.values() ) {
			GumpClient consumer = request.getSpec().getGumpClient();

			if ( request.getSpec().getSource().equals( source ) && !consumers.contains( consumer ) )
				consumers.add( consumer );
		}

		return consumers;
	}


	/**
	 * Gets a Collection containing all Id's of current consumers
	 * 
	 * @param source Choose if the list to be returned will be composed by local
	 *        or remote consumers
	 * @return A collection with the selected consumers Id's
	 */
	public Collection<ObjectID> getConsumersIDs( RequestSource source ) {

		Collection<ObjectID> result = new LinkedList<ObjectID>();

		for ( RequestEntry request : this.requests.values() ) {
			ObjectID consumerId = request.getSpec().getObjectID();

			if ( request.getSpec().getSource().equals( source ) && !result.contains( consumerId ) )
				result.add( consumerId );
		}

		return result;
	}


	/**
	 * Selects the requests that still need Gums.
	 * 
	 * @return The collection containing the selected requests
	 */
	public Collection<RequestEntry> getRequestsThatNeedGums( Collection<RequestEntry> requestList ) {

		Collection<RequestEntry> requestsThatNeedGums = new LinkedList<RequestEntry>();

		for ( RequestEntry request : requestList ) {

			if ( request.needMoreGums() ) {
				requestsThatNeedGums.add( request );
			}
		}

		return requestsThatNeedGums;
	}


	public void unwantedGum( long requestID, GumID gumID ) throws RequestDoesNotExistException {

		RequestEntry request = this.getRequestEntry( requestID );

		request.unwantedGum( gumID );
	}


	public Set<GumID> getUnwantedGumsForRequest( long requestID ) throws RequestDoesNotExistException {

		RequestEntry request = this.getRequestEntry( requestID );

		return request.getUnwantedGums();
	}


	public static void reset() {

		instance = null;
	}


	public Collection<GumID> getRequestGums( long requestID ) throws RequestDoesNotExistException {

		RequestEntry request = this.getRequestEntry( requestID );

		return request.getGums();
	}


	public boolean requestIsLocal( long requestID ) throws RequestDoesNotExistException {

		RequestEntry request = this.getRequestEntry( requestID );

		return request.getSpec().isFromALocalBroker();
	}


	public void deleteConsumerRequests( ObjectID consumerID ) {

		Iterator<Entry<Long,RequestEntry>> it = this.requests.entrySet().iterator();

		while ( it.hasNext() ) {
			Entry<Long,RequestEntry> entry = it.next();
			if ( entry.getValue().getSpec().getObjectID().equals( consumerID ) ) {
				it.remove();
			}
		}
	}


	public RequestSpec getRequestSpec( long requestID ) throws RequestDoesNotExistException {

		RequestEntry request = this.getRequestEntry( requestID );

		return request.getSpec();
	}


	public boolean requestWantsGum( long requestID, GumID gumID ) throws RequestDoesNotExistException {

		RequestEntry request = this.getRequestEntry( requestID );

		return !request.isGumUnwanted( gumID );
	}


	public boolean requestNeedMoreGums( long requestId ) throws RequestDoesNotExistException {

		RequestEntry request = this.getRequestEntry( requestId );

		return request.needMoreGums();
	}


	public int getRequestNeededGums( long requestId ) throws RequestDoesNotExistException {

		RequestEntry request = this.getRequestEntry( requestId );

		return request.getNeededGums();
	}


	public boolean requestIsPaused( long requestId ) throws RequestDoesNotExistException {

		RequestEntry request = this.getRequestEntry( requestId );

		return request.isPaused();
	}


	public boolean requestHasGum( long requestId, GumID gumID ) throws RequestDoesNotExistException {

		RequestEntry request = this.getRequestEntry( requestId );

		return request.getGums().contains( gumID );
	}
}
