/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager.request;

import junit.framework.TestCase;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.exception.RequestAlreadyExistsException;
import org.ourgrid.common.id.GumID;
import org.ourgrid.common.spec.RequestSpec;
import org.ourgrid.common.spec.RequestSpec.RequestSource;
import org.ourgrid.peer.manager.FakeGumpClient;
import org.ourgrid.peer.manager.exception.RequestDoesNotExistException;
import org.ourgrid.test.PeerFastTest;
import org.ourgrid.test.TestUtils;

/**
 * The test for <code>EBRequestManager</code> class.
 */
public class RequestManagerTest extends TestCase implements PeerFastTest {

	private static final int INVALIDREDID1 = 123450;

	private static final int REQID3 = 987654321;

	private static final int REQID2 = 12345;

	private static final int REQID1 = 1987523;

	/** The EBRequestManager object used in tests. */
	private RequestManager requestManager;

	/** Some fake consumers. */
	private FakeGumpClient firstFakeGmc;

	private FakeGumpClient secondFakeGmc;

	private FakeGumpClient thirdFakeGmc;

	private TestUtils testUtils;

	private RequestSpec requestSpec1;

	private RequestSpec requestSpec3;

	private RequestSpec requestSpec2;


	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {

		super.setUp();

		testUtils = new TestUtils();
		testUtils.load();
		testUtils.createMGProperties();

		firstFakeGmc = new FakeGumpClient( "1" );
		secondFakeGmc = new FakeGumpClient( "2" );
		thirdFakeGmc = new FakeGumpClient( "3" );

		Configuration.getInstance( Configuration.PEER );

		requestManager = RequestManager.getInstance();

		requestSpec1 = new RequestSpec( REQID1, "os == linux && ram == 256", 10, firstFakeGmc );
		requestSpec1.setAsFromLocalBroker();
		requestManager.createRequestEntry( requestSpec1 ); // Local Request

		// Community Requests
		requestSpec3 = new RequestSpec( REQID2, "os == linux && ram == 256", 10, secondFakeGmc );
		requestManager.createRequestEntry( requestSpec3 );
		requestSpec2 = new RequestSpec( REQID3, "os == windows || ram == 128", 2, thirdFakeGmc );
		requestManager.createRequestEntry( requestSpec2 );
	}


	/*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		super.tearDown();
		Configuration.reset();
		testUtils.reset();
		RequestManager.reset();
	}


	/**
	 * Verifies if the requests are been created correctly.
	 */
	public void testCreateAndContainRequest() {

		assertTrue( requestManager.requestExists( REQID2 ) );
		assertTrue( requestManager.requestExists( REQID3 ) );
		assertFalse( requestManager.requestExists( INVALIDREDID1 ) );
		try {
			RequestSpec requestSpec4 = new RequestSpec( REQID2, "ram == 64", 10, firstFakeGmc );
			requestSpec4.setAsFromRemotePeer();
			requestManager.createRequestEntry( requestSpec4 );
			fail( "A RequestAlreadyExistsException should be thrown." );
		} catch ( RequestAlreadyExistsException e ) {
			// this line has to be executed!
		}
	}// /**


	/**
	 * Verifies if consumers are been correctly associated.
	 * 
	 * @throws RequestDoesNotExistException
	 */
	public void testGetConsumerFromRequestID() throws RequestDoesNotExistException {

		assertTrue( requestManager.getGumpClient( REQID2 ) == secondFakeGmc );
		assertTrue( requestManager.getGumpClient( REQID1 ) == firstFakeGmc );
		assertTrue( requestManager.getGumpClient( REQID3 ) == thirdFakeGmc );
	}


	public void testNoMoreGuMsForRequest() throws RequestDoesNotExistException {

		assertTrue( requestManager.requestNeedMoreGums( REQID2 ) );
		requestManager.deleteRequest( REQID2 );
		try {
			requestManager.requestNeedMoreGums( REQID2 );
			fail( "The request does not exist, an exception should be thrown" );
		} catch ( RequestDoesNotExistException e ) {
		}
	}


	/**
	 * Verifies if the <code>suspendWannaGuMsForRequest()</code> and
	 * <code>resumeWannaGuMsForRequest()</code> methods are changing the
	 * request.
	 * 
	 * @throws RequestDoesNotExistException
	 */
	public void testPauseResumeRequest() throws RequestDoesNotExistException {

		assertTrue( requestManager.requestNeedMoreGums( REQID2 ) );
		requestManager.pauseRequestGums( REQID2 );
		assertFalse( requestManager.requestNeedMoreGums( REQID2 ) );
		requestManager.resumeRequestGums( REQID2 );
		assertTrue( requestManager.requestNeedMoreGums( REQID2 ) );
	}


	/**
	 * Verifies if the <code>hereIsGuMForRequest()</code>,
	 * <code>disposeOfForRequest()</code> and <code>lostGuMForRequest()</code>
	 * methods are changing the request.
	 * 
	 * @throws RequestDoesNotExistException
	 * @throws RequestDoesNotExistException
	 */
	public void testMethodCallsForRequest() throws RequestDoesNotExistException {

		// does hereIsGuM for requests

		GumID gumID1 = new GumID( "gm1", "0", "peername", "0" );
		GumID gumID2 = new GumID( "gm2", "0", "peername", "0" );
		GumID gumID3 = new GumID( "gm3", "0", "peername", "0" );
		GumID gumID4 = new GumID( "gm4", "0", "peername", "0" );

		requestManager.hereIsGumToRequest( requestSpec2.getID(), gumID1 );
		requestManager.hereIsGumToRequest( requestSpec2.getID(), gumID2 );
		requestManager.hereIsGumToRequest( requestSpec3.getID(), gumID3 );

		assertTrue( requestManager.requestHasGum( requestSpec2.getID(), gumID1 ) );
		assertTrue( requestManager.requestHasGum( requestSpec2.getID(), gumID2 ) );
		assertFalse( requestManager.requestHasGum( requestSpec2.getID(), gumID3 ) );
		assertFalse( requestManager.requestHasGum( requestSpec2.getID(), gumID4 ) );

		requestManager.removeGumFromRequest( requestSpec2.getID(), gumID2 );
		assertTrue( requestManager.requestHasGum( requestSpec2.getID(), gumID1 ) );
		assertFalse( requestManager.requestHasGum( requestSpec2.getID(), gumID2 ) );

		requestManager.hereIsGumToRequest( requestSpec2.getID(), gumID4 ); // now,
																			// the
																			// request
																			// need
																			// a
																			// machine
		assertTrue( requestManager.requestHasGum( requestSpec2.getID(), gumID4 ) );

		requestManager.removeGumFromRequest( requestSpec2.getID(), gumID1 );
		assertFalse( requestManager.requestHasGum( requestSpec2.getID(), gumID1 ) );

		requestManager.deleteRequest( requestSpec2.getID() );

		try {
			requestManager.hereIsGumToRequest( requestSpec2.getID(), gumID2 );
			fail( "RequestDoesNotExistException should be thrown" );
		} catch ( RequestDoesNotExistException e ) {
		}
	}


	/**
	 * Tests the initial map contents.
	 */
	public void testInitialConditions() {

		assertEquals( 3, requestManager.getAliveRequests( RequestSource.FROM_LOCAL_BROKER ).size() );
	}

}
