/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.manager.status;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Map;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.RequestSpec.RequestSource;
import org.ourgrid.mygrid.scheduler.event.SimpleResponseEvent;
import org.ourgrid.peer.community.nof.AccountingManager;
import org.ourgrid.peer.manager.allocation.AllocationEntry;
import org.ourgrid.peer.manager.allocation.AllocationManager;
import org.ourgrid.peer.manager.allocation.AllocationStatus;
import org.ourgrid.peer.manager.allocation.AllocationEntry.GumSource;
import org.ourgrid.peer.manager.gummanager.EBGumManagerImpl;
import org.ourgrid.peer.manager.request.RequestManager;

public class EBStatusProviderImpl implements EBStatusProvider {

	private AllocationManager allocationManager;

	private AccountingManager accountingManager;

	private EBGumManagerImpl gumManager;

	private RequestManager requestManager;


	public EBStatusProviderImpl( EBGumManagerImpl gumManager ) {

		Configuration.getInstance( Configuration.PEER );

		this.allocationManager = AllocationManager.getInstance();
		this.gumManager = gumManager;
		this.accountingManager = AccountingManager.getInstance();
		this.requestManager = RequestManager.getInstance();

	}


	public void getLocalGums( EventQueue<ResponseEvent<Collection<StatusEntry>>> responseQueue ) {

		Collection<StatusEntry> gums = new LinkedList<StatusEntry>();
		gums.addAll( createStatusEntries( allocationManager.getLocalGums() ) );
		gums.addAll( createOfflineStatusEntry( gumManager.getOfflineGums() ) );

		responseQueue.put( new SimpleResponseEvent<Collection<StatusEntry>>( gums ) );

	}


	public void getRemoteGums( EventQueue<ResponseEvent<Collection<StatusEntry>>> responseQueue ) {

		responseQueue.put( new SimpleResponseEvent<Collection<StatusEntry>>( createStatusEntries( allocationManager
				.getRemoteGums() ) ) );

	}


	public void getGums( EventQueue<ResponseEvent<Collection<StatusEntry>>> responseQueue,
							AllocationStatus allocationStatus, RequestSource source ) {

		switch ( allocationStatus ) {
			case DONATED:
				responseQueue.put( new SimpleResponseEvent<Collection<StatusEntry>>(
						createStatusEntries( allocationManager.getLocalDonated() ) ) );
				break;
			case IDLE:
				responseQueue.put( new SimpleResponseEvent<Collection<StatusEntry>>(
						createStatusEntries( allocationManager.getLocalIdle() ) ) );
				break;
			case IN_USE:
				if ( source.isLocal() ) {
					responseQueue.put( new SimpleResponseEvent<Collection<StatusEntry>>(
							createStatusEntries( allocationManager.getLocalInUse() ) ) );
				} else {
					responseQueue.put( new SimpleResponseEvent<Collection<StatusEntry>>(
							createStatusEntries( allocationManager.getRemoteInUse() ) ) );
				}
				break;
			case CONTACTING:
				responseQueue.put( new SimpleResponseEvent<Collection<StatusEntry>>(
						createOfflineStatusEntry( gumManager.getOfflineGums() ) ) );
				break;
			case OWNER:
				responseQueue.put( new SimpleResponseEvent<Collection<StatusEntry>>( createOwnerStatusEntry( gumManager
						.getOwnerGums() ) ) );
				break;
			default:
				responseQueue.put( new SimpleResponseEvent<Collection<StatusEntry>>( new LinkedList<StatusEntry>() ) );
		}

	}


	public void getPeerbalance( EventQueue<ResponseEvent<Map<ObjectID,Double>>> responseQueue ) {

		responseQueue.put( new SimpleResponseEvent<Map<ObjectID,Double>>( accountingManager.getKnownPeerBalances() ) );

	}


	private Collection<StatusEntry> createStatusEntries( Collection<AllocationEntry> allocationEntry ) {

		Collection<StatusEntry> statusEntries = new LinkedList<StatusEntry>();

		for ( AllocationEntry entry : allocationEntry ) {
			statusEntries.add( new StatusEntry( entry.getEBGum().getGumSpec(), entry.getStatus() ) );
		}

		return statusEntries;

	}


	private Collection<StatusEntry> createOfflineStatusEntry( Collection<GumSpec> gumSpecList ) {

		Collection<StatusEntry> statusEntries = new LinkedList<StatusEntry>();

		for ( GumSpec gumSpec : gumSpecList ) {
			statusEntries.add( new StatusEntry( gumSpec, AllocationStatus.CONTACTING ) );
		}

		return statusEntries;

	}


	private Collection<StatusEntry> createOwnerStatusEntry( Collection<GumSpec> gumSpecList ) {

		Collection<StatusEntry> statusEntries = new LinkedList<StatusEntry>();

		for ( GumSpec gumSpec : gumSpecList ) {
			statusEntries.add( new StatusEntry( gumSpec, AllocationStatus.OWNER ) );
		}

		return statusEntries;

	}


	public void getConsumers( EventQueue<ResponseEvent<Collection<ConsumerStatus>>> responseQueue, RequestSource source ) {

		Collection<ObjectID> consumers = requestManager.getConsumersIDs( source );
		Collection<ConsumerStatus> statusCollection = new ArrayList<ConsumerStatus>();

		for ( ObjectID consumerId : consumers ) {
			int localGums = requestManager.getGums( consumerId, GumSource.LOCAL ).size();
			int remoteGums = requestManager.getGums( consumerId, GumSource.REMOTE ).size();
			statusCollection.add( new ConsumerStatus( consumerId, localGums, remoteGums ) );
		}

		responseQueue.put( new SimpleResponseEvent<Collection<ConsumerStatus>>( statusCollection ) );
	}
}
