/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.ui;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.spec.RequestSpec.RequestSource;
import org.ourgrid.peer.manager.EBPeerManagerFacade;
import org.ourgrid.peer.manager.allocation.AllocationStatus;
import org.ourgrid.peer.manager.status.StatusEntry;
import org.ourgrid.peer.ui.listener.PeerReceivedCommunityGumsStateListener;
import org.ourgrid.peer.ui.listener.PeerSetterStateListener;
import org.ourgrid.peer.ui.listener.PeerStarterStateListener;
import org.ourgrid.peer.ui.listener.PeerStopperStateListener;

public class ObserverDispatcherImpl implements ObserverDispatcher {

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( ObserverDispatcherImpl.class );

	/**
	 * List of all MyGrid listeners
	 */
	private ArrayList<Object> allListeners;

	private static final long serialVersionUID = 33L;

	private ArrayList<PeerSetterStateListener> peerSetterListeners;

	private ArrayList<PeerStopperStateListener> peerStopperListeners;

	private ArrayList<PeerStarterStateListener> peerStarterListeners;

	private ArrayList<PeerReceivedCommunityGumsStateListener> peerReceivedCommunityGumsListeners;

	private EBPeerManagerFacade ebPeermanagerFacade;


	public ObserverDispatcherImpl( EBPeerManagerFacade ebPeerManagerFacade ) {

		super();
		this.ebPeermanagerFacade = ebPeerManagerFacade;
		allListeners = new ArrayList<Object>();
		peerSetterListeners = new ArrayList<PeerSetterStateListener>();
		peerStopperListeners = new ArrayList<PeerStopperStateListener>();
		peerReceivedCommunityGumsListeners = new ArrayList<PeerReceivedCommunityGumsStateListener>();
		peerStarterListeners = new ArrayList<PeerStarterStateListener>();
	}


	/*
	 * Peer Setter Methods
	 */
	public void addPeerSetterStateListener( PeerSetterStateListener listener ) {

		peerSetterListeners.add( listener );
		allListeners.add( listener );
		listener.setGumsOcurred();
	}


	public void removePeerSetterStateListener( PeerSetterStateListener listener ) {

		peerSetterListeners.remove( listener );
		allListeners.remove( listener );
		notifyNoMoreListeners();
	}


	public void notifyPeerSetterStateListeners() {

		ArrayList<PeerSetterStateListener> setterListenersClone;
		synchronized ( this ) {
			setterListenersClone = (ArrayList<PeerSetterStateListener>) this.peerSetterListeners.clone();
		}
		for ( Iterator<PeerSetterStateListener> iter = setterListenersClone.iterator(); iter.hasNext(); ) {
			PeerSetterStateListener listener = iter.next();
			listener.setGumsOcurred();

		}
	}


	public void addPeerStopperStateListener( PeerStopperStateListener listener ) {

		peerStopperListeners.add( listener );
		allListeners.add( listener );
	}


	public void removePeerStopperStateListener( PeerStopperStateListener listener ) {

		peerStopperListeners.remove( listener );
		allListeners.remove( listener );
		notifyNoMoreListeners();
	}


	public void notifyPeerStopperStateListeners() {

		ArrayList<PeerStopperStateListener> stopperListenersClone;
		synchronized ( this ) {
			stopperListenersClone = (ArrayList<PeerStopperStateListener>) this.peerStopperListeners.clone();
		}
		for ( Iterator<PeerStopperStateListener> iter = stopperListenersClone.iterator(); iter.hasNext(); ) {
			PeerStopperStateListener listener = iter.next();
			listener.stopperEventOcurred();

		}
	}


	public void addPeerReceivedCommunityGumsStateListener( PeerReceivedCommunityGumsStateListener listener ) {

		peerReceivedCommunityGumsListeners.add( listener );
		allListeners.add( listener );

		EventQueue<ResponseEvent<Collection<StatusEntry>>> responseQueue = new EventQueue<ResponseEvent<Collection<StatusEntry>>>();
		ebPeermanagerFacade.getGums( responseQueue, AllocationStatus.IN_USE, RequestSource.FROM_REMOTE_PEER );
		Collection<StatusEntry> gumNameList = responseQueue.blockingRemove().getResponse();

		listener.receivedCommunityGumEventOcurred( gumNameList );

	}


	public void removePeerReceivedCommunityGumsStateListener( PeerReceivedCommunityGumsStateListener listener ) {

		peerReceivedCommunityGumsListeners.remove( listener );
		allListeners.remove( listener );
		notifyNoMoreListeners();
	}


	public void notifyPeerReceivedCommunityGumsStateListeners( Collection communityGums ) {

		ArrayList<PeerReceivedCommunityGumsStateListener> receivedListenersClone;
		synchronized ( this ) {
			receivedListenersClone = (ArrayList<PeerReceivedCommunityGumsStateListener>) this.peerReceivedCommunityGumsListeners
					.clone();
		}
		for ( Iterator<PeerReceivedCommunityGumsStateListener> iter = receivedListenersClone.iterator(); iter.hasNext(); ) {
			PeerReceivedCommunityGumsStateListener listener = iter.next();
			listener.receivedCommunityGumEventOcurred( communityGums );
		}
	}


	/**
	 * Verifies if there is no more listeners subscribed in this dispatcher. If
	 * it's true, the MyGridManager will be notified, and can take a decision
	 * about to shutdown the virtual machine or not.
	 */
	private void notifyNoMoreListeners() {

		if ( allListeners.isEmpty() ) {
			try {
				PeerUIManager.getInstance().noMoreListeners();
			} catch ( RemoteException e ) {
				LOG.debug( "Could not contact the Peer UI Manager" );
			}
		}

	}


	/**
	 * Verifies if there is no more listeners subscribed in this dispatcher.
	 * 
	 * @return Return true if this Dispatcher still have any listener, returns
	 *         false otherwise.
	 */
	public boolean stillHaveListeners() {

		return !allListeners.isEmpty();
	}


	/**
	 * @see org.ourgrid.peer.ui.ObserverDispatcher#addPeerStarterStateListener(org.ourgrid.peer.ui.listener.PeerStarterStateListener)
	 */
	public void addPeerStarterStateListener( PeerStarterStateListener listener ) {

		peerStarterListeners.add( listener );
		allListeners.add( listener );
		try {
			if ( PeerUIManager.getInstance().isPeerRunning() )
				listener.starterEventOcurred();
		} catch ( RemoteException e ) {
			LOG.debug( "Could not contact the Peer UI Manager" );
		}
	}


	/**
	 * @see org.ourgrid.peer.ui.ObserverDispatcher#removePeerStarterStateListener(org.ourgrid.peer.ui.listener.PeerStarterStateListener)
	 */
	public void removePeerStarterStateListener( PeerStarterStateListener listener ) {

		peerStarterListeners.remove( listener );
		allListeners.remove( listener );
		notifyNoMoreListeners();
	}


	/**
	 * @see org.ourgrid.peer.ui.ObserverDispatcher#notifyPeerStarterStateListeners()
	 */
	public void notifyPeerStarterStateListeners() {

		ArrayList<PeerStarterStateListener> starterListenersClone;
		synchronized ( this ) {
			starterListenersClone = (ArrayList<PeerStarterStateListener>) this.peerStarterListeners.clone();
		}
		for ( Iterator<PeerStarterStateListener> iter = starterListenersClone.iterator(); iter.hasNext(); ) {
			PeerStarterStateListener listener = iter.next();
			listener.starterEventOcurred();
		}
	}

}
