/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.ui;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.Map;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.LogConfiguration;
import org.ourgrid.common.config.PeerConfiguration;
import org.ourgrid.common.event.EventQueue;
import org.ourgrid.common.event.ResponseEvent;
import org.ourgrid.common.fd.AbstractFailureDetector;
import org.ourgrid.common.fd.GenericState;
import org.ourgrid.common.gump.BrokerAccessImpl;
import org.ourgrid.common.gump.Gump;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.rmi.OurgridUnicastRemoteObject;
import org.ourgrid.common.security.MGSecureClientSocketFactory;
import org.ourgrid.common.security.MGSecureServerSocketFactory;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.RequestSpec.RequestSource;
import org.ourgrid.common.spec.main.CommonCompiler;
import org.ourgrid.common.spec.main.CompilerException;
import org.ourgrid.common.url.PeerURLProvider;
import org.ourgrid.common.url.URLProvider;
import org.ourgrid.peer.community.RemoteAccess;
import org.ourgrid.peer.community.RemoteAccessImpl;
import org.ourgrid.peer.community.communicationlayer.CommunicationLayerImpl;
import org.ourgrid.peer.community.obtainer.EBCommunityObtainerFacade;
import org.ourgrid.peer.fd.ConsumerStateNotifiable;
import org.ourgrid.peer.fd.CorePeerStateNotifiable;
import org.ourgrid.peer.fd.GumStateNotifiable;
import org.ourgrid.peer.fd.PeerFailureDetector;
import org.ourgrid.peer.manager.EBPeerManagerFacade;
import org.ourgrid.peer.manager.allocation.AllocationStatus;
import org.ourgrid.peer.manager.status.StatusEntry;
import org.ourgrid.peer.manager.status.StatusProvider;
import org.ourgrid.peer.manager.status.StatusProviderImpl;
import org.ourgrid.peer.ui.command.ErrorCode;
import org.ourgrid.peer.ui.command.UIMessages;
import org.ourgrid.peer.ui.exception.PeerUIException;
import org.ourgrid.peer.ui.gui.PeerGUI;

public class PeerUIManager extends OurgridUnicastRemoteObject implements UIManager {

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( PeerUIManager.class );

	private static final long serialVersionUID = 33L;

	private static UIManager uniqueInstance;

	private ObserverDispatcher observerDispatcher;

	private EBPeerManagerFacade ebPeerManagerFacade;

	private EBCommunityObtainerFacade ebCommunityObtainerFacade;

	private PeerFailureDetector failureDetector;

	public enum Notifiables {
		COREPEER, CONSUMER, GUM;
	}

	private GumStateNotifiable gumStateNotifiable;

	private CorePeerStateNotifiable corePeerStateNotifiable;

	private ConsumerStateNotifiable consumerStateNotifiable;

	private StatusProvider statusProvider;

	private Gump brokerAccess;

	private Gump remoteAccess;

	private ObjectID remoteAccessID;

	private ObjectID brokerAccessID;


	private PeerUIManager() throws RemoteException {

		super();
		try {
			failureDetector = new PeerFailureDetector( Configuration.QUERY_FREQ );
			gumStateNotifiable = new GumStateNotifiable();
			corePeerStateNotifiable = new CorePeerStateNotifiable();
			consumerStateNotifiable = new ConsumerStateNotifiable();
		} catch ( RemoteException e ) {
			LOG.error( "Error trying to create Failure Detector", e );
		}
	}


	public static UIManager getInstance() throws RemoteException {

		if ( uniqueInstance == null ) {

			Configuration.getInstance( Configuration.PEER );
			String logfile = Configuration.getInstance().getLogPath();
			String xmlLogfile = Configuration.getInstance().getXMLLogPath();

			String parents = new File( logfile ).getParent();

			if ( parents != null ) {
				(new File( parents )).mkdirs();
			}
			// Configuring Log4j
			String logPropertiesfile = Configuration.getInstance().getLogPropertiesPath();
			LogConfiguration.configure( logPropertiesfile, logfile, xmlLogfile );

			try {
				// tries to get a remote instance already bound by another
				// invocation of this method.
				uniqueInstance = (UIManager) Naming.lookup( PeerURLProvider.peerUIManager() );
			} catch ( MalformedURLException e ) {
			} catch ( NotBoundException e ) {
			} catch ( RemoteException e ) {
			}

			if ( uniqueInstance == null ) {
				uniqueInstance = new PeerUIManager();

				try {
					Naming.rebind( PeerURLProvider.peerUIManager(), uniqueInstance );
				} catch ( MalformedURLException e ) {
					LOG.debug( "Error trying to rebind at URL " + PeerURLProvider.peerUIManager() );
				}
			}
		}
		return uniqueInstance;
	}


	/**
	 * @see org.ourgrid.peer.ui.UIManager#startPeerService()
	 */
	public void startPeerService() throws RemoteException, MalformedURLException, NumberFormatException,
		ConfigException {

		Configuration.getInstance( Configuration.PEER );

		Configuration.getInstance().setProperty( Configuration.PROP_STARTED_TIME, new Date().toString() );

		/**
		 * Security options.
		 */
		if ( Configuration.getInstance().isSecurityEnabled() ) {

			String keyStoreFilename = Configuration.getInstance().getProperty( Configuration.PROP_KEYSTORE_FILENAME );
			String keyStorePassword = Configuration.getInstance().getProperty( Configuration.PROP_KEYSTORE_PASSWORD );

			System.setProperty( "javax.net.ssl.keyStore", keyStoreFilename );
			System.setProperty( "javax.net.ssl.keyStorePassword", keyStorePassword );

			String trustStoreFilename = Configuration.getInstance()
					.getProperty( Configuration.PROP_TRUSTSTORE_FILENAME );
			String trustStorePassword = Configuration.getInstance()
					.getProperty( Configuration.PROP_TRUSTSTORE_PASSWORD );

			System.setProperty( "javax.net.ssl.trustStore", trustStoreFilename );
			System.setProperty( "javax.net.ssl.trustStorePassword", trustStorePassword );

		}

		// Configuring Facades
		ebPeerManagerFacade = new EBPeerManagerFacade();
		ebCommunityObtainerFacade = new EBCommunityObtainerFacade();
		observerDispatcher = new ObserverDispatcherImpl( ebPeerManagerFacade );
		startFailureDetector();

		createBrokerAccess();
		createRemoteAccess();

		ebPeerManagerFacade.config( ebCommunityObtainerFacade, remoteAccessID, brokerAccessID, this );
		ebCommunityObtainerFacade.config( ebPeerManagerFacade, this );

		ebPeerManagerFacade.startEventProcessor();
		ebCommunityObtainerFacade.startEventProcessor();

		bindBrokerAccess();
		bindRemoteAccess();
		bindStatusProvider();

		CommunicationLayerImpl communicationLayer = CommunicationLayerImpl.getInstance();
		communicationLayer.config( (RemoteAccess) remoteAccess );

		observerDispatcher.notifyPeerStarterStateListeners();
		LOG.info( "OurGrid Peer " + Configuration.VERSION + " is Up and Running!" );

		String propDefaultSDF = Configuration.getInstance().getProperty( PeerConfiguration.PROP_DEFAULT_SDF );

		if ( propDefaultSDF != null )
			loadDefaultSDF( propDefaultSDF );

	}


	/**
	 * Loads the configurated default SDF.
	 * 
	 * @param propDefaultSDF The default SDF.
	 */
	private void loadDefaultSDF( String propDefaultSDF ) {

		File defaultSDF = new File( propDefaultSDF );
		if ( !defaultSDF.exists() ) {
			LOG.info( "OurGrid Peer default SDF not found: " + propDefaultSDF );
			return;
		}

		try {
			setLocalGums( propDefaultSDF );
		} catch ( PeerUIException e ) {
			LOG.info( "OurGrid Peer could not contact UIManager to load default SDF!" );
		} catch ( CompilerException e ) {
			LOG.info( "OurGrid Peer could not compile default SDF!" );
		} catch ( RemoteException e ) {
			LOG.info( "OurGrid Peer could not contact UIManager to load default SDF!" );
		}

	}


	/**
	 * Starts a StatusProvider RMI Object and publish it
	 * 
	 * @throws RemoteException
	 * @throws MalformedURLException
	 */
	private void bindStatusProvider() throws RemoteException, MalformedURLException {

		this.statusProvider = new StatusProviderImpl( ebPeerManagerFacade );
		Naming.rebind( PeerURLProvider.statusProvider(), statusProvider );
		LOG.debug( "The StatusProvider RMI object was successfully published." );
	}


	/**
	 * Starts Failure Detector
	 * 
	 * @throws MalformedURLException
	 * @throws RemoteException
	 */
	private void startFailureDetector() throws RemoteException, MalformedURLException {

		gumStateNotifiable.setFacade( ebPeerManagerFacade );
		consumerStateNotifiable.setFacade( ebPeerManagerFacade );
		failureDetector.startMonitoring();
		Naming.rebind( PeerURLProvider.peerFailureDetector(), failureDetector );
	}


	/**
	 * @see org.ourgrid.peer.ui.UIManager#stopPeerService()
	 */
	public void stopPeerService() throws PeerUIException {

		if ( isPeerRunning() ) {

			shutdown();
			observerDispatcher.notifyPeerStopperStateListeners();
			failureDetector.shutdown();

		} else {
			throw new PeerUIException( "Peer is not running!", null );
		}

		notifyPeerIsDown();
	}


	/**
	 * @see org.ourgrid.peer.ui.UIManager#startGUIService()
	 */
	public void startGUIService() throws RemoteException {

		PeerGUI.getInstance().show();
	}


	/**
	 * @see org.ourgrid.peer.ui.UIManager#setLocalGums(Collection)
	 */
	public void setLocalGums( Collection<GumSpec> guMSpecs ) throws PeerUIException {

		if ( isPeerRunning() ) {

			ebPeerManagerFacade.setGums( guMSpecs );

			observerDispatcher.notifyPeerSetterStateListeners();

			updateFailureDetector( guMSpecs );

		} else {
			throw new PeerUIException( "Peer is not running!", null );
		}

	}


	private void updateFailureDetector( Collection<GumSpec> guMSpecs ) {

		/* Removing old registereds Gums in this Failure Detector */
		try {
			failureDetector.unregister( gumStateNotifiable );
		} catch ( RemoteException e1 ) {
			LOG.error( "Can not unregister GumStateNotifiable from PeerFailureDetector.", e1 );
		}

		/* Registering new Gums in this Failure Detector */
		for ( Iterator<GumSpec> iter = guMSpecs.iterator(); iter.hasNext(); ) {

			GumSpec gumSpec = iter.next();

			String rmiHost = "rmi://" + gumSpec.getAttribute( GumSpec.ATT_NAME ) + ":"
					+ gumSpec.getAttribute( GumSpec.ATT_PORT ) + "/" + AbstractFailureDetector.BOUND_NAME;

			try {
				this.registerNewMonitorable( Notifiables.GUM, rmiHost );
			} catch ( RemoteException e ) {
				LOG.error( "Cannot register gum " + gumSpec.getAttribute( GumSpec.ATT_NAME ), e );
			}

		}

	}


	/**
	 * @throws RemoteException
	 * @see org.ourgrid.peer.ui.UIManager#setLocalGums(java.lang.String)
	 */
	public void setLocalGums( String descriptionFilePath ) throws PeerUIException, CompilerException, RemoteException {

		if ( !PeerUIManager.getInstance().isPeerRunning() ) {
			throw new RemoteException( UIMessages.PEER_IS_NOT_RUNNING );
		}

		CommonCompiler compiler = new CommonCompiler();

		compiler.compile( descriptionFilePath, CommonCompiler.SDF_TYPE );

		@SuppressWarnings("unchecked")
		Collection<GumSpec> gumsList = compiler.getResult();

		if ( gumsList != null ) {

			setLocalGums( gumsList );

		} else {

			throw new CompilerException( "OurGrid error: description file is empty!" );

		}
	}


	/**
	 * @see org.ourgrid.peer.ui.UIManager#getObserverDispatcher()
	 */
	public ObserverDispatcher getObserverDispatcher() {

		return observerDispatcher;
	}


	private void notifyPeerIsDown() {

		if ( !getObserverDispatcher().stillHaveListeners() ) {
			callExit();
		}

	}


	/**
	 * @see org.ourgrid.peer.ui.UIManager#noMoreListeners()
	 */
	public void noMoreListeners() {

		if ( !isPeerRunning() ) {
			callExit();
		}

	}


	/**
	 * Method used to shutdown this JVM (calling System.exit(0)).
	 */
	private void callExit() {

		System.exit( ErrorCode.OK );
	}


	/**
	 * @see org.ourgrid.peer.ui.UIManager#isPeerRunning()
	 */
	public boolean isPeerRunning() {

		try {
			findExportedObjects();
			return true;
		} catch ( Exception e ) {
			return false;
		}
	}


	private void findExportedObjects() throws MalformedURLException, RemoteException, NotBoundException {

		Naming.lookup( PeerURLProvider.localAccess() );

		if ( Configuration.getInstance().isEnabled( PeerConfiguration.PROP_JOIN_COMMUNITY ) ) {
			Naming.lookup( PeerURLProvider.remoteAccess() );
		}
	}


	private void shutdown() {

		if ( ebPeerManagerFacade != null ) {
			ebPeerManagerFacade.shutdown();
			LOG.info( "PeerManager shutdown!" );
		}

		if ( ebCommunityObtainerFacade != null ) {
			ebCommunityObtainerFacade.shutdown();
			LOG.info( "CommunityObtainer shutdown!" );
		}
		LOG.info( "Peer now is going to stop!" );
	}


	public Collection<StatusEntry> getGums( AllocationStatus allocationStatus, RequestSource source )
		throws RemoteException {

		EventQueue<ResponseEvent<Collection<StatusEntry>>> responseQueue = new EventQueue<ResponseEvent<Collection<StatusEntry>>>();

		ebPeerManagerFacade.getGums( responseQueue, allocationStatus, source );

		return responseQueue.blockingRemove().getResponse();
	}


	public Map<ObjectID,Double> getPeerBalances() throws RemoteException {

		EventQueue<ResponseEvent<Map<ObjectID,Double>>> responseQueue = new EventQueue<ResponseEvent<Map<ObjectID,Double>>>();

		ebPeerManagerFacade.getPeerBalance( responseQueue );

		return responseQueue.blockingRemove().getResponse();
	}


	/**
	 * @see org.ourgrid.peer.ui.UIManager#registerNewMonitorable(org.ourgrid.peer.ui.PeerUIManager.Notifiables,
	 *      ObjectID)
	 */
	public void registerNewMonitorable( Notifiables notifiableType, ObjectID objectID ) throws RemoteException {

		try {
			if ( objectID != null ) {
				String url = URLProvider.getFullURL( objectID.getObjectURL(), AbstractFailureDetector.BOUND_NAME );
				registerNewMonitorable( notifiableType, url );

				if ( notifiableType.equals( Notifiables.CONSUMER ) ) {
					consumerStateNotifiable.newMonitorable( objectID );
				}
			} else {
				LOG.error( "Unable to register ObjectID because it is null" );
			}
		} catch ( URISyntaxException e ) {
			LOG.error( "Unable to define url", e );
		}
	}


	/**
	 * @see org.ourgrid.peer.ui.UIManager#unregisterNewMonitorable(org.ourgrid.peer.ui.PeerUIManager.Notifiables,
	 *      org.ourgrid.common.id.ObjectID)
	 */
	public void unregisterNewMonitorable( Notifiables notifiableType, ObjectID objectID ) {

		try {
			if ( objectID != null ) {
				String url = URLProvider.getFullURL( objectID.getObjectURL(), AbstractFailureDetector.BOUND_NAME );
				unregisterNewMonitorable( notifiableType, url );

				if ( notifiableType.equals( Notifiables.CONSUMER ) ) {
					consumerStateNotifiable.removeMonitorable( objectID );
				}
			} else {
				LOG.error( "Unable to unregister ObjectID because it is null" );
			}
		} catch ( URISyntaxException e ) {
			LOG.error( "Unable to define url", e );
		} catch ( RemoteException e ) {
		}
	}


	/**
	 * @see org.ourgrid.peer.ui.UIManager#registerNewMonitorable(org.ourgrid.peer.ui.PeerUIManager.Notifiables,
	 *      String)
	 */
	public void registerNewMonitorable( Notifiables notifiableType, String monitorableAddress ) throws RemoteException {

		switch ( notifiableType ) {
			case GUM: {
				try {
					LOG.info( "Registering new Gum: " + monitorableAddress + " to failure detector" );
					this.failureDetector.register( this.gumStateNotifiable, monitorableAddress, true );
				} catch ( Exception e ) {
					LOG.error( "Cannot register GumNotifiable at " + monitorableAddress, e );
				}
				break;
			}

			case COREPEER: {
				try {
					LOG.info( "Registering new Core Peer: " + monitorableAddress + " to failure detector" );
					this.failureDetector.register( this.corePeerStateNotifiable, monitorableAddress, true );
				} catch ( Exception e ) {
					LOG.error( "Cannot register CorePeerNotifiable at " + monitorableAddress, e );
				}
				break;
			}

			case CONSUMER: {
				try {
					LOG.info( "Registering new Consumer: " + monitorableAddress + " to failure detector" );
					this.failureDetector.register( this.consumerStateNotifiable, monitorableAddress, true );
				} catch ( Exception e ) {
					LOG.error( "Cannot register PeerNotifiable at " + monitorableAddress, e );
				}
				break;
			}

			default:
				break;
		}

	}


	/**
	 * Creates the gump used locally to access the peer.
	 * 
	 * @throws RemoteException
	 * @throws ConfigException
	 * @throws NumberFormatException
	 */
	private void createBrokerAccess() throws NumberFormatException, ConfigException, RemoteException {

		this.brokerAccessID = new ObjectID( "BrokerAccessImpl", new Long( 0 ), PeerURLProvider.localAccess() );

		if ( Configuration.getInstance().isSecurityEnabled() ) {
			int securePort = Integer
					.parseInt( Configuration.getInstance().getProperty( Configuration.PROP_SECURE_PORT ) );
			MGSecureServerSocketFactory secureServerSocketFactory = MGSecureServerSocketFactory.getInstance();
			MGSecureClientSocketFactory secureClientSocketFactory = MGSecureClientSocketFactory.getInstance();
			brokerAccess = new BrokerAccessImpl( brokerAccessID, ebPeerManagerFacade, secureClientSocketFactory,
					secureServerSocketFactory, securePort );
		} else {
			brokerAccess = new BrokerAccessImpl( brokerAccessID, ebPeerManagerFacade );
		}
		LOG.debug( "LocalAccessGump was successfully created." );
	}


	/**
	 * Binds the Gump used locally to access the peer.
	 * 
	 * @throws MalformedURLException
	 * @throws RemoteException
	 */
	private void bindBrokerAccess() throws RemoteException, MalformedURLException {

		Naming.rebind( PeerURLProvider.localAccess(), brokerAccess );
		LOG.debug( "BrokerAccess was successfully bound at: " + PeerURLProvider.localAccess() );
	}


	/**
	 * Creates the gump used locally to access the peer.
	 * 
	 * @throws RemoteException
	 */
	private void createRemoteAccess() throws RemoteException {

		this.remoteAccessID = loadPeerID();
		remoteAccess = new RemoteAccessImpl( remoteAccessID, ebCommunityObtainerFacade, ebPeerManagerFacade );
	}


	/**
	 * Binds the gump used locally to access the peer.
	 * 
	 * @throws MalformedURLException
	 * @throws RemoteException
	 */
	private void bindRemoteAccess() throws RemoteException, MalformedURLException {

		Naming.rebind( PeerURLProvider.remoteAccess(), remoteAccess );
		LOG.debug( "RemoteAccess was sucessfully bound at: " + PeerURLProvider.remoteAccess() );
	}


	/**
	 * Loads from file, if already exists, or creates this peer id.
	 * 
	 * @return This peer ID.
	 */
	private ObjectID loadPeerID() {

		LOG.debug( "Creating the ObjectID..." );

		File f = new File( Configuration.getInstance().getProperty( PeerConfiguration.PROP_IDFILE ) );

		String peerName = Configuration.getInstance().getProperty( PeerConfiguration.PROP_LABEL );

		Long idValue = null;

		if ( f.exists() && f.canRead() ) {

			LOG.debug( "Loading ObjectID from file " + f.getAbsolutePath() );

			try {

				ObjectInputStream obj = new ObjectInputStream( new FileInputStream( f ) );
				idValue = (Long) obj.readObject();
				obj.close();
				LOG.debug( "ObjectID sucessfuly loaded from file " + f.getAbsolutePath() );
			} catch ( Exception e ) {
				LOG.debug( "Could not load ObjectID from file " + f.getAbsolutePath() + " Cause: " + e.getMessage() );
			}

		}

		if ( idValue == null ) {

			LOG.debug( "Creating a new ObjectID..." );

			idValue = choosePeerId();

			try {

				LOG.debug( "Saving ObjectID to file " + f.getAbsolutePath() );

				ObjectOutputStream obj = new ObjectOutputStream( new FileOutputStream( f ) );
				obj.writeObject( idValue );
				obj.flush();
				obj.close();

				LOG.debug( "ObjectID sucessfully saved on file " + f.getAbsolutePath() );

			} catch ( Exception e ) {

				LOG.debug( "Could not save ObjectID on file " + f.getAbsolutePath() + " Cause: " + e.getMessage() );

			}

		}

		return new ObjectID( peerName, idValue, PeerURLProvider.peerExternalURL() );
	}


	/**
	 * @return A random choosed 64 bit number.
	 */
	private Long choosePeerId() {

		return new Long( ((long) (Math.random() * Long.MAX_VALUE)) );
	}


	/**
	 * @see org.ourgrid.peer.ui.UIManager#unregisterNewMonitorable(org.ourgrid.peer.ui.PeerUIManager.Notifiables,
	 *      String)
	 */
	public void unregisterNewMonitorable( Notifiables notifiableType, String monitorableAddress )
		throws RemoteException {

		switch ( notifiableType ) {
			case GUM: {
				try {
					LOG.info( "Unregistering GumNotifiable at " + monitorableAddress + " from failure detector" );
					this.failureDetector.unregister( this.gumStateNotifiable, monitorableAddress );
				} catch ( Exception e ) {
					LOG.error( "Cannot unregister GumNotifiable at " + monitorableAddress, e );
				}
				break;
			}

			case COREPEER: {
				try {
					LOG.info( "Unregistering CorePeerNotifiable at " + monitorableAddress + " from failure detector" );
					this.failureDetector.unregister( this.corePeerStateNotifiable, monitorableAddress );
				} catch ( Exception e ) {
					LOG.error( "Cannot unregister CorePeerNotifiable at " + monitorableAddress, e );
				}
				break;
			}

			case CONSUMER: {
				try {
					LOG.info( "Unregistering ConsumerNotifiable at " + monitorableAddress + " from failure detector" );
					this.failureDetector.unregister( this.consumerStateNotifiable, monitorableAddress );
				} catch ( Exception e ) {
					LOG.error( "Cannot unregister PeerNotifiable at " + monitorableAddress, e );
				}
				break;
			}

			default:
				break;
		}

	}


	public void failureDetectorAdvice( String monitorableAddress, GenericState newState ) throws RemoteException {

		this.failureDetector.setMonitorableState( monitorableAddress, newState );
	}
}
