/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.ui.command;

import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.util.HashMap;
import java.util.Map;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.mygrid.ui.command.ErrorCode;
import org.ourgrid.mygrid.ui.command.InvalidCommandException;
import org.ourgrid.peer.ui.PeerUIManager;
import org.ourgrid.peer.ui.UIManager;

/**
 * Implements the main of Command Pattern. All Peer commands must pass here.
 */
public class CommandManager {

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( CommandManager.class );

	private Map<String,Command> peerCommands;

	private static final String PEER_STARTER = "start";

	private static final String PEER_STOPPER = "stop";

	private static final String PEER_STATUS = "status";

	private static final String PEER_SETGUMS = "setgums";

	private UIManager manager;


	public CommandManager() {

		peerCommands = new HashMap<String,Command>();
		initCommandMap();
		initRMIRegistry();
		try {
			manager = PeerUIManager.getInstance();
		} catch ( RemoteException e ) {
			LOG.debug( "Could not get instance to peer UI manager" );
		}
	}


	private void initRMIRegistry() {

		int registryPort = Integer.parseInt( Configuration.getInstance( Configuration.PEER ).getProperty(
				Configuration.PROP_PORT ) );

		System.setProperty( "java.rmi.server.hostname", Configuration.getInstance().getProperty(
				Configuration.PROP_EXTERNAL_NAME ) );

		try {
			// Starting registry
			LocateRegistry.createRegistry( registryPort );
		} catch ( RemoteException e ) {
			// TODO LOG
		}
	}


	/**
	 * it starts the map with PeerCommands.
	 */
	private void initCommandMap() {

		peerCommands.put( PEER_STOPPER, new PeerStopper() );
		peerCommands.put( PEER_STARTER, new PeerStarter() );
		peerCommands.put( PEER_STATUS, new PeerStatus() );
		peerCommands.put( PEER_SETGUMS, new PeerSetter() );
	}


	/**
	 * Method that checks what <code>Command</code> will be able to execute
	 * the command in parameter "param". It will verify in it's map of commands,
	 * and then, call his run method.
	 * 
	 * @param params This parameter must contain the Command string
	 *        representation, followed by it's parameters.
	 * @throws InvalidCommandException 
	 */
	public void executeCommand( String[ ] params ) throws InvalidCommandException {

		String command = params[0];

		String[ ] parameters = discardCommandParameter( params );
		Command gridCommand = peerCommands.get( command );

		if ( gridCommand == null ) {
			throw new InvalidCommandException( params[0] );
		}

		gridCommand.run( parameters, manager );

	}


	/**
	 * Discard the first parameter from this String[]. This parameter represents
	 * the Command from line command.
	 * 
	 * @param params An String array with all parameter that came from line
	 *        comand, including the especified Peer Command.
	 * @return All other parameters except the Peer Command
	 */
	private String[ ] discardCommandParameter( String[ ] params ) {

		String[ ] parameters = new String[ params.length - 1 ];

		for ( int i = 1; i < params.length; i++ ) {
			parameters[i - 1] = params[i];
		}

		return parameters;
	}


	public static void main( String[ ] args ) {

		try {
			new CommandManager().executeCommand( args );
		} catch ( InvalidCommandException e ) {
			System.exit( ErrorCode.FAIL );
		} catch ( Exception e1 ) {
			System.out.println( "Could not execute Peer Command " + args[0] + " " + e1.getMessage() );
			System.exit( ErrorCode.FAIL );
		}
	}

}
