/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.ui.command;

import java.rmi.RemoteException;
import java.text.ParseException;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.RequestSpec.RequestSource;
import org.ourgrid.common.url.URLProvider;
import org.ourgrid.peer.manager.allocation.AllocationStatus;
import org.ourgrid.peer.manager.status.StatusEntry;
import org.ourgrid.peer.ui.PeerUIManager;
import org.ourgrid.peer.ui.UIManager;
import org.ourgrid.peer.ui.exception.PeerUIException;

public class PeerStatus implements Command {

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( PeerStatus.class );


	/**
	 * Main method. This method should be called with no command line arguments
	 */
	public void run( String[ ] params, UIManager manager ) {

		try {

			if ( !manager.isPeerRunning() ) {
				System.out.println( UIMessages.PEER_IS_NOT_RUNNING );
				System.exit( ErrorCode.PEER_IS_NOT_RUNNING );

			} else {

				System.out.println( UIMessages.PEER_IS_UP_RUNNING );

				printGumList();
				printAccounting();

				System.exit( ErrorCode.OK );

			}

		} catch ( PeerUIException re ) {
			System.out.println( UIMessages.PEER_IS_NOT_RUNNING );
			System.exit( ErrorCode.PEER_IS_NOT_RUNNING );
		} catch ( RemoteException e ) {
			System.err.println( e.getMessage() );
			System.exit( ErrorCode.FAIL );
		}

	}


	private void printGumList() throws PeerUIException {

		Collection<StatusEntry> localGums = null;
		Collection<StatusEntry> remoteGums = null;
		Collection<StatusEntry> donatedGums = null;
		Collection<StatusEntry> idleGums = null;
		Collection<StatusEntry> offlineGums = null;
		Collection<StatusEntry> ownerGums = null;

		try {

			localGums = PeerUIManager.getInstance().getGums( AllocationStatus.IN_USE, RequestSource.FROM_LOCAL_BROKER );
			idleGums = PeerUIManager.getInstance().getGums( AllocationStatus.IDLE, RequestSource.FROM_LOCAL_BROKER );
			remoteGums = PeerUIManager.getInstance().getGums( AllocationStatus.IN_USE, RequestSource.FROM_REMOTE_PEER );
			donatedGums = PeerUIManager.getInstance().getGums( AllocationStatus.DONATED,
					RequestSource.FROM_LOCAL_BROKER );
			offlineGums = PeerUIManager.getInstance().getGums( AllocationStatus.CONTACTING,
					RequestSource.FROM_LOCAL_BROKER );
			ownerGums = PeerUIManager.getInstance().getGums( AllocationStatus.OWNER, RequestSource.FROM_LOCAL_BROKER );

		} catch ( RemoteException e ) {
			LOG.debug( "Could not contact the Peer UI Manager: " + e.getMessage() );
			throw new PeerUIException( e );
		}

		System.out.println( "\n" + UIMessages.LOCAL_MACHINES );

		// printing idle local gums
		printGums( idleGums, AllocationStatus.IDLE );

		// printing local gums in use by local brokers
		printGums( localGums, AllocationStatus.IN_USE );

		// printing gums donated to the community
		printGums( donatedGums, AllocationStatus.DONATED );

		// printing local offline gums
		printGums( offlineGums, AllocationStatus.CONTACTING );

		// printing local owner gums
		printGums( ownerGums, AllocationStatus.OWNER );

		System.out.println( "\n" + UIMessages.REMOTE_MACHINES );

		// printing gums from community in use by local brokers
		printGums( remoteGums, AllocationStatus.IN_USE );
	}


	private void printGums( Collection<StatusEntry> gumsList, AllocationStatus status ) {

		if ( gumsList != null ) {

			Iterator it = gumsList.iterator();

			while ( it.hasNext() ) {
				System.out.println( it.next() + " [ " + status + " ]" );
			}
		}
	}


	/**
	 * Prints network of favors accounting info.
	 * 
	 * @throws PeerUIException If cannot contact the peer.
	 * @throws ParseException
	 */
	private void printAccounting() throws PeerUIException {

		Map<ObjectID,Double> peerBalances;

		try {
			peerBalances = PeerUIManager.getInstance().getPeerBalances();

			if ( peerBalances == null ) {

				System.err.println( "Peer Balances map is invalid!" );

			} else {

				boolean any = false;

				System.out.print( "\nNetwork of Favors Accounting:\n" );

				Iterator<ObjectID> itPeers = peerBalances.keySet().iterator();

				if ( itPeers.hasNext() ) {
					System.out.println( "\n\t\t Peer Id \t\t\t\t Peer Balance" );
				}

				while ( itPeers.hasNext() ) {
					any = true;

					ObjectID peerID = itPeers.next();
					printPeerBalance( peerID, peerBalances.get( peerID ) );
				}

				if ( !any ) {
					System.out.print( " none\n" );
				}

			}

		}

		catch ( Exception e ) {
			System.err.println( "Peer can not print balance.\n" );
			throw new PeerUIException( e );
		}
	}


	private void printPeerBalance( ObjectID peerId, Double balance ) throws Exception {

		String peerUrl = peerId.getObjectURL();
		String peerhost = URLProvider.getHost( peerUrl );
		int peerPort = URLProvider.getPort( peerUrl );
		System.out.println( "\n\t" + peerhost + ":" + peerPort + "\t\t\t" + balance + "\n" );

	}
}
