/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.peer.ui.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.GridLayout;
import java.awt.Toolkit;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.rmi.RemoteException;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.filechooser.FileFilter;

import org.ourgrid.common.config.ConfigException;
import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.spec.main.CompilerException;
import org.ourgrid.peer.ui.PeerUIManager;
import org.ourgrid.peer.ui.UIManager;
import org.ourgrid.peer.ui.exception.PeerUIException;
import org.ourgrid.peer.ui.listener.PeerGUIListenerImpl;

/**
 * Peer GUI, this class contains the main method.
 */

public class PeerGUI {

	/**
	 * Logger to store log messages
	 */
	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger.getLogger( PeerGUI.class );

	/* Listener object */
	private PeerGUIListenerImpl guiListener = null;

	private static PeerGUI uniqueInstance = null;

	/** String constants */

	private static final String ERROR_ON_ABOUT = "Error on reading about contents.";

	private static final String UNEXPECTED_ERROR = "Unexpected error: ";

	private static final String COMMUNICATION_PROBLEM = "Communication problem. Peer is not running!";

	private static final String ERROR_OCCURRED = "Error occurred";

	private static final String GUMS = "Grid Machines";

	private static final String CONF = "Settings";

	private static final String ABOUT = "About";

	private static final String SETGUMS_BUTTON = "Set Gums";

	private static final String STOP_BUTTON = "Stop";

	private static final String START_BUTTON = "Start";

	private static final String PEER_TITLE = "OurGrid Peer ";

	private static final String LINE_SEPARATOR = System.getProperty( "line.separator" );

	private static final String FILE_SEPARATOR = System.getProperty( "file.separator" );

	private final String ogroot = Configuration.getInstance( Configuration.PEER ).getProperty( Configuration.MGROOT );

	/** About Contents file path */
	private final String ABOUT_CONTENTS_FILE = ogroot + FILE_SEPARATOR + "resources" + FILE_SEPARATOR + "ABOUT";

	/** Icon file path */
	private final String ICON_PATH = ogroot + FILE_SEPARATOR + "resources/images" + FILE_SEPARATOR + "icon.png";

	/** JFrame */
	private JFrame frame;

	/**
	 * Stores diferent sections of the GUI, Peer config section, Local GuMP
	 * section, GuMCache section...
	 */
	private JTabbedPane tabbedPane;

	/** This Panel contains all the other panels of this GUI */
	private JPanel masterPanel;

	/** This Panel contains the buttons of start and stop MyGrid */
	private JPanel masterButtonsPanel;

	/** This panel contains Doctor section */
	private JPanel doctorSection;

	/** This panel contains About section */
	private JPanel aboutSection;

	/** Settings Panel */
	private SettingsPanel settingsPanel;

	/** Grid Machines Panel */
	private GumPanel gridMachinePanel;

	/** Statistics Panel */
	// private StatisticsPanel statisticsPanel;

	/** PeerUIServices */
	private UIManager uiManager;

	/** Buttons */
	private JButton startButton;

	private JButton stopButton;

	private JButton setGumsButton;

	/** Opened file */
	private File openedFile = null;


	/**
	 * Constructor
	 * 
	 * @throws RemoteException
	 */
	private PeerGUI() throws RemoteException {

		/**
		 * Security options.
		 */
		try {
			if ( Configuration.getInstance( Configuration.PEER ).isSecurityEnabled() ) {

				String trustStoreFilename = Configuration.getInstance().getProperty(
						Configuration.PROP_TRUSTSTORE_FILENAME );
				String trustStorePassword = Configuration.getInstance().getProperty(
						Configuration.PROP_TRUSTSTORE_PASSWORD );

				System.setProperty( "javax.net.ssl.trustStore", trustStoreFilename );
				System.setProperty( "javax.net.ssl.trustStorePassword", trustStorePassword );

			}
		} catch ( ConfigException e ) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		uiManager = PeerUIManager.getInstance();

		frame = new JFrame();
		frame.setTitle( PEER_TITLE + Configuration.VERSION );
		frame.setSize( 640, 512 );
		frame.setLocationRelativeTo( null );
		frame.setResizable( true );
		frame.setIconImage( Toolkit.getDefaultToolkit().createImage( ICON_PATH ) );

		initMasterPanel();
		initMasterButtonsPanel();
		initTabbedPanel();

		stoppedPeer();
		subscribeNewListener();
	}


	public static PeerGUI getInstance() throws RemoteException {

		if ( uniqueInstance == null ) {
			uniqueInstance = new PeerGUI();
		}

		return uniqueInstance;

	}


	/**
	 * Subscribe a new listener into observerDispatcher
	 */
	private void subscribeNewListener() {

		try {
			guiListener = new PeerGUIListenerImpl( this );

		} catch ( Exception e ) {
			LOG.error( "Error trying to bind PeerObservrDispatcher remote object. ", e );
		}

		try {
			PeerUIManager.getInstance().getObserverDispatcher().addPeerStarterStateListener( guiListener );
			PeerUIManager.getInstance().getObserverDispatcher().addPeerSetterStateListener( guiListener );
			PeerUIManager.getInstance().getObserverDispatcher().addPeerStopperStateListener( guiListener );
			PeerUIManager.getInstance().getObserverDispatcher().addPeerReceivedCommunityGumsStateListener( guiListener );
		} catch ( RemoteException e ) {
			LOG.error( "Cannot subscribe under ObserverDispatcher", e );
		}

	}


	/**
	 * Unsubscribe a new listener into observerDispatcher
	 */
	private void unSubscribeNewListener() {

		try {
			PeerUIManager.getInstance().getObserverDispatcher().removePeerStarterStateListener( guiListener );
			PeerUIManager.getInstance().getObserverDispatcher().removePeerSetterStateListener( guiListener );
			PeerUIManager.getInstance().getObserverDispatcher().removePeerStopperStateListener( guiListener );
			PeerUIManager.getInstance().getObserverDispatcher().removePeerReceivedCommunityGumsStateListener(
					guiListener );
		} catch ( RemoteException e ) {
			LOG.error( "Cannot unsubscribe under ObserverDispatcher", e );
		}

	}


	/** Delegates the show() method */
	public void show() {

		frame.setVisible( true );
	}


	/** Starts the Master Panel */
	private void initMasterPanel() {

		masterPanel = new JPanel();
		masterPanel.setLayout( new BorderLayout() );
		frame.getContentPane().add( masterPanel );
	}


	/** Starts the Master Buttons Panel and configuring it */
	private void initMasterButtonsPanel() {

		masterButtonsPanel = new JPanel();

		startButton = new JButton( START_BUTTON );
		stopButton = new JButton( STOP_BUTTON );
		setGumsButton = new JButton( SETGUMS_BUTTON );

		masterButtonsPanel.add( startButton );
		masterButtonsPanel.add( stopButton );
		masterButtonsPanel.add( setGumsButton );

		masterPanel.add( masterButtonsPanel, BorderLayout.SOUTH );

		// Events
		startButton.addActionListener( new java.awt.event.ActionListener() {

			public void actionPerformed( java.awt.event.ActionEvent evt ) {

				startButtonActionPerformed();
			}
		} );

		stopButton.addActionListener( new java.awt.event.ActionListener() {

			public void actionPerformed( java.awt.event.ActionEvent evt ) {

				stopButtonActionPerformed();
			}
		} );

		setGumsButton.addActionListener( new java.awt.event.ActionListener() {

			public void actionPerformed( java.awt.event.ActionEvent evt ) {

				setGridActionPerformed();
			}
		} );

		// Window close event
		frame.addWindowListener( new WindowAdapter() {

			public void windowClosing( WindowEvent e ) {

				super.windowClosing( e );
				unSubscribeNewListener();
			}

		} );

	}


	/** Inits the Tabbed Pane adding and configuting tabs */
	private void initTabbedPanel() {

		tabbedPane = new JTabbedPane();

		masterPanel.add( tabbedPane, BorderLayout.CENTER );

		// Settings Panel
		settingsPanel = SettingsPanel.getInstance();
		tabbedPane.addTab( CONF, settingsPanel.getPanel() );

		JPanel gumsSection = new JPanel();
		gumsSection.setLayout( new GridLayout( 1, 1 ) );
		gridMachinePanel = GumPanel.getInstance();
		gumsSection.add( gridMachinePanel.getPanel() );
		tabbedPane.addTab( GUMS, gridMachinePanel.getPanel() );

		// Doctor
		doctorSection = new JPanel();
		doctorSection.setLayout( new GridLayout( 1, 1 ) );

		// gumsPanel.setDoctorPanel(doctorPanel);

		// Statistics Panel
		// statisticsPanel = StatisticsPanel.getInstance();
		// tabbedPane.addTab(GUMS, statisticsPanel.getPanel());

		// about Panel
		aboutSection = new JPanel();
		aboutSection.setLayout( new GridLayout( 1, 1 ) );
		aboutSection.add( createAbout() );
		aboutSection.setBackground( Color.LIGHT_GRAY );
		tabbedPane.addTab( ABOUT, aboutSection );
	}


	/**
	 * Creates a JPanel containing the About text.
	 * 
	 * @return A JPanel containing the About text.
	 */
	private JPanel createAbout() {

		JPanel aboutPanel = new JPanel();
		aboutPanel.setBackground( Color.LIGHT_GRAY );
		JTextArea aboutText = new JTextArea( getAboutContents() );
		aboutText.setEditable( false );
		aboutText.setBackground( Color.LIGHT_GRAY );
		aboutPanel.add( aboutText );
		return aboutPanel;

	}


	/** When peer is started */
	public void startedPeer() {

		startButton.setEnabled( false );
		stopButton.setEnabled( true );
		setGumsButton.setEnabled( true );
	}


	/** When peer is stoped */
	public void stoppedPeer() {

		startButton.setEnabled( true );
		stopButton.setEnabled( false );
		setGumsButton.setEnabled( false );
	}


	/**
	 * When start button is clicked
	 */
	private void startButtonActionPerformed() {

		try {
			uiManager.startPeerService();
		} catch ( Exception e ) {
			LOG.debug( "Could not start peer service" );
		}
	}


	/**
	 * When stop button is clicked
	 */
	private void stopButtonActionPerformed() {

		try {
			uiManager.stopPeerService();
		} catch ( PeerUIException e1 ) {
			if ( !(e1.getCause() instanceof java.rmi.UnmarshalException) ) {
				showErrorToUser( e1 );
			}
		} catch ( RemoteException e ) {
			LOG.debug( "Could not stop peer service" );
		}
	}


	/**
	 * When setGums button is clicked
	 */
	private void setGridActionPerformed() {

		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setCurrentDirectory( new File( "." ) );

		fileChooser.setFileFilter( new FileFilter() {

			public boolean accept( File f ) {

				return f.getName().toLowerCase().endsWith( ".sdf" ) || f.isDirectory();
			}


			public String getDescription() {

				return "Site Description File (*.sdf)";
			}
		} );

		openedFile = this.getSelectedFile( fileChooser );

		if ( openedFile == null )
			return;

		try {
			PeerUIManager.getInstance().setLocalGums( openedFile.getAbsolutePath() );
		} catch ( PeerUIException uie ) {
			showErrorToUser( uie );
		} catch ( RemoteException re ) {
			showErrorToUser( re );
		} catch ( CompilerException ce ) {
			showErrorToUser( ce );
		}

	}


	/**
	 * Return Selected file in file chooser
	 * 
	 * @param fileChooser The FileChooser
	 */
	private File getSelectedFile( JFileChooser fileChooser ) {

		int result = fileChooser.showOpenDialog( this.frame );
		if ( result == JFileChooser.APPROVE_OPTION )
			return fileChooser.getSelectedFile();
		return null;
	}


	public void refreshGridTab() {

		try {
			if ( uiManager.isPeerRunning() ) {
				gridMachinePanel.refreshTree();
			} else {
				gridMachinePanel.clean();
			}
		} catch ( RemoteException e ) {
			LOG.debug( "Could not access peer service" );
		}
	}


	/**
	 * Shows the user a message when an exception occurs.
	 * 
	 * @param e The exception occurred.
	 */
	private void showErrorToUser( Exception e ) {

		String message = getFriendlyMessage( e );
		JOptionPane.showMessageDialog( this.frame, message, ERROR_OCCURRED, JOptionPane.ERROR_MESSAGE );

	}


	/**
	 * Returns a user-friendly error message.
	 * 
	 * @param e The exception occurred.
	 * @return An user-friendly error message according to the exception.
	 */
	private String getFriendlyMessage( Exception e ) {

		if ( e instanceof RemoteException ) {
			return COMMUNICATION_PROBLEM;
		}

		return UNEXPECTED_ERROR + LINE_SEPARATOR + e.getMessage();
	}


	/**
	 * Main
	 * 
	 * @param args Arguments
	 */
	public static void main( String[ ] args ) {

		PeerGUI ogg;
		try {
			ogg = new PeerGUI();
			ogg.show();
		} catch ( RemoteException e ) {
			LOG.debug( "Cannot create a new Peer GUI" );
		}

	}


	/**
	 * Reads the about contents from a file.
	 * 
	 * @return A string containg the about.
	 */
	private String getAboutContents() {

		String aboutContents = "";
		BufferedReader br = null;
		try {
			br = new BufferedReader( new FileReader( ABOUT_CONTENTS_FILE ) );
			String line = null;
			while ( (line = br.readLine()) != null ) {
				aboutContents += line + LINE_SEPARATOR;
			}
			br.close();
		} catch ( IOException e ) {
			return ERROR_ON_ABOUT;
		} finally {
			try {
				if ( br != null ) {
					br.close();
				}
			} catch ( IOException e ) {
				LOG.debug( "Failed to close stream on Exception." );
			}
		}
		return aboutContents;
	}

}
