/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.Properties;

import org.ourgrid.common.config.PeerConfiguration;
import org.ourgrid.common.spec.PeerSpec;

/**
 * <code>LocalPeer</code> provides a instance to create and execute a local
 * peer (that will not join community). Used at MyGrid tests.
 */
public class LocalPeer {

	private static final String NO = "no";

	private static final String DEFAULT_NAME = "localhost";

	private String OGROOT;

	private Properties properties;

	private int port;

	private File backupPeerProperties;

	private File originalPeerProperties;

	private static final int ERROR = -1;

	private boolean backupIsCreated;

	private File gumListFile;

	private static final long DEFAULT_SLEEP_TIME = 10000;

	private static final long DEFAULT_SLEEP_TIMEINTERVAL = 1000;


	/**
	 * LocalPeer constructor
	 * 
	 * @param port Port to be used
	 */
	public LocalPeer( int port ) {

		properties = new Properties();

		properties.setProperty( "peer." + PeerConfiguration.PROP_JOIN_COMMUNITY, NO );
		properties.setProperty( "peer." + PeerConfiguration.PROP_NAME, DEFAULT_NAME );
		properties.setProperty( "peer." + PeerConfiguration.PROP_PORT, Integer.toString( port ) );

		OGROOT = System.getProperty( "OGROOT" );

		if ( OGROOT == null ) {
			OGROOT = System.getProperty( "MGROOT" );
		}

		this.port = port;

	}


	private void createLocalPeerProperties() {

		originalPeerProperties = new File( OGROOT + "/peer.properties" );
		backupPeerProperties = new File( OGROOT + "/peer.properties.not_local" );
		if ( backupPeerProperties.exists() ) {
			backupPeerProperties.delete();
		}

		if ( originalPeerProperties.exists() ) {
			originalPeerProperties.renameTo( backupPeerProperties );
		}

		try {
			FileOutputStream outputStream = new FileOutputStream( originalPeerProperties );
			properties.store( outputStream, "Local " + originalPeerProperties );
			outputStream.close();
			this.backupIsCreated = true;
		} catch ( Exception e ) {
			System.err.println( "Could not create a new " + originalPeerProperties + "." );
			e.printStackTrace();
		}
	}


	/**
	 * Sets a configuration property, the peer must not be running.
	 * 
	 * @param key Key to be setted
	 * @param value Value of the key
	 */
	public void setProperty( String key, String value ) {

		if ( isRunning() ) {
			throw new IllegalStateException( "Peer already running" );
		}
		properties.setProperty( key, value );
	}


	/**
	 * Get the local peer GuMPSpec
	 * 
	 * @return The local peer GuMPSpec
	 */
	public PeerSpec getGuMPSpec() {

		return new PeerSpec( properties.getProperty( "peer." + PeerConfiguration.PROP_NAME ), port );
	}


	/**
	 * Starts the peer execution with setted configurations. Its assure that
	 * peer will start by making a occupied waiting.
	 */
	public void start() {

		int result = ERROR;

		String[ ] peerCommandStart = { "bin/peer", "start" };

		try {
			createLocalPeerProperties();
			result = executeCommand( peerCommandStart );
		} catch ( Exception e ) {
			restoreBackup();
			throw new Error( "Could not start peer, a " + e.getClass() + " has been throw" );
		}

		if ( result != 0 ) {
			throw new Error( "Could not start peer, script has failed" );
		}

	}


	/**
	 * @param command Command to be executed
	 * @return Return value of the process
	 * @throws Exception Throws a exception when it cannot execute
	 */
	private int executeCommand( String[ ] command ) throws Exception {

		Process process;
		Runtime exec = Runtime.getRuntime();
		process = exec.exec( command );
		process.waitFor();
		return process.exitValue();
	}


	/**
	 * Stops the peer execution
	 */
	public void stop() {

		int result = ERROR;

		String[ ] peerCommandStop = { "bin/peer", "stop" };

		try {
			result = executeCommand( peerCommandStop );
		} catch ( Exception e ) {
			throw new Error( "Could not stop peer, a " + e.getClass() + " has been throw" );
		} finally {
			if ( gumListFile != null ) {
				stopUAs( gumListFile );
			}
			restoreBackup();
		}

		if ( result != 0 ) {
			throw new Error( "Could not stop peer, script has failed" );
		}

	}


	/**
	 * Creates a temporary GDF file to be used with LocalPeer as GuMP
	 * 
	 * @param tempDirectory tempDirectory to be used
	 * @return Path of temporary GDF File
	 * @throws IOException If cannot write to file
	 */
	public String generateTempGDF( String tempDirectory ) throws IOException {

		File tempGDF = File.createTempFile( "mygrid-test", ".gdf", new File( tempDirectory ) );

		BufferedWriter writer = new BufferedWriter( new FileWriter( tempGDF ) );

		writer.write( "Peer : " );
		writer.newLine();
		writer.write( " label : LocalPeer" );
		writer.newLine();
		writer.write( " name : " + properties.getProperty( "peer." + PeerConfiguration.PROP_NAME ) );
		writer.newLine();
		writer.write( "port : " + port );
		writer.newLine();

		writer.close();

		tempGDF.deleteOnExit();

		return tempGDF.toString();
	}


	/**
	 * Restores a backup file
	 */
	private void restoreBackup() {

		if ( backupIsCreated ) {
			originalPeerProperties.delete();
			backupPeerProperties.renameTo( originalPeerProperties );
		}
	}


	/**
	 * Tests if the <code>LocalPeer</code> is running
	 * 
	 * @return True if its running
	 */
	public boolean isRunning() {

		int result = ERROR;

		String[ ] peerCommandStatus = { "bin/peer", "status" };

		long expectedTime = 0;

		try {
			do {
				result = executeCommand( peerCommandStatus );
				expectedTime += 500;
				Thread.sleep( DEFAULT_SLEEP_TIMEINTERVAL );
				expectedTime += DEFAULT_SLEEP_TIMEINTERVAL;
			} while ( result != 0 && expectedTime < DEFAULT_SLEEP_TIME );
			return (result == 0 ? true : false);
		} catch ( Exception e ) {
			return false;
		}
	}


	/**
	 * Set a list of Gums to be used at peer, also starts the UserAgents
	 * 
	 * @param GuMList
	 */
	public boolean setGuMs( File GuMList ) {

		try {
			String[ ] uaAdminCommand = { "bin/uaadmin", "restart", GuMList.getCanonicalFile().toString() };
			System.out.println( "Running: " + Arrays.toString( uaAdminCommand ) );
			executeCommand( uaAdminCommand );
		} catch ( Exception e ) {
			return false;
		}

		int result = ERROR;

		try {
			String[ ] peerCommandSetGuMs = { "bin/peer", "setgums", GuMList.getCanonicalFile().toString() };
			System.out.println( "Running: " + Arrays.toString( peerCommandSetGuMs ) );
			result = executeCommand( peerCommandSetGuMs );

		} catch ( Exception e ) {
			return false;
		}

		if ( result != 0 ) {
			return false;
		}

		this.gumListFile = GuMList;

		return true;

	}


	/**
	 * Stops the UAs
	 * 
	 * @param GuMList Description file with the GuMs
	 * @return True if it succeds
	 */
	public boolean stopUAs( File GuMList ) {

		try {
			String[ ] uaAdminCommand = { "bin/uaadmin", "stop", GuMList.getCanonicalFile().toString() };
			executeCommand( uaAdminCommand );
			return true;
		} catch ( Exception e ) {
			return false;
		}
	}

}
