/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test;

import java.io.File;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.Vector;

import junit.framework.TestSuite;
import junit.runner.TestCollector;

/**
 * Description: Creates a test suite composed of tests of a specified type.
 * 
 * @version 1.0 Created on 05/07/2004
 */
public class SelectiveTestCollector extends TestSuite implements TestCollector {

	/**
	 * This class represents the type of tests that should be collected.
	 */
	private Class selectiveTestType;

	/**
	 * Dir to search for test classes
	 */
	private String basedir;


	/**
	 * Creates a new SelectiveTestCollector with a test type to collect. It uses
	 * the $mygrid.output.dir as basedir to search.
	 * 
	 * @param selectiveTestType Type of tests that should be collected
	 */
	public SelectiveTestCollector( Class selectiveTestType ) {

		this.selectiveTestType = selectiveTestType;
		this.basedir = System.getProperty( "ourgrid.output.dir" );
	}


	/**
	 * Creates a new SelectiveTestCollector with a test type to collect and a
	 * basedir to search.
	 * 
	 * @param selectiveTestType Type of tests that should be collected
	 * @param basedir Basedir to search for test classes.
	 */
	public SelectiveTestCollector( Class selectiveTestType, String basedir ) {

		this.selectiveTestType = selectiveTestType;
		this.basedir = basedir;
	}


	/**
	 * Returns a enumeration with all tests of the type
	 * <code>selectiveTestType</code> that where found inside the
	 * <code>basedir</code> and all it's sub-dirs
	 * 
	 * @return Enumeration that matches with requirements
	 */
	public Enumeration collectTests() {

		Vector tests = collectTests( new File( basedir ) );

		Collections.sort( tests, new Comparator() {

			public int compare( Object o1, Object o2 ) {

				Class c1 = (Class) o1;
				Class c2 = (Class) o2;

				boolean b1 = isSpecialTestCase( c1 );
				boolean b2 = isSpecialTestCase( c2 );

				if ( b1 && !b2 ) {
					return -1;
				}

				if ( !b1 && b2 ) {
					return 1;
				}

				return c1.getName().compareTo( c2.getName() );
			}

		} );

		return tests.elements();

	}


	/**
	 * Do the real job of collect the tests os type
	 * <code>selectiveTestType</code> inside a directory. Calls itselft
	 * recursivelly to search all sub-dirs.
	 * 
	 * @param directory A directory to search for tests classes
	 * @return A <code>Vector</code> with all tests classes inside
	 *         <code>directory</code> and its sub-dirs.
	 */
	private Vector<Class> collectTests( File directory ) {

		File files[] = directory.listFiles();
		Vector<Class> tests = new Vector<Class>();

		for ( int i = 0; i < files.length; i++ ) {

			if ( files[i].isFile() ) {

				if ( files[i].getName().endsWith( ".class" ) ) {
					String fullPathName = files[i].getAbsolutePath();
					String split[] = fullPathName.split( "/classes/" );
					String className = "";

					try {
						className = split[1].replace( '/', '.' );
						className = className.substring( 0, className.lastIndexOf( '.' ) );

					} catch ( NoSuchElementException e ) {
						System.err.println( "Invalid class path found: " + className );
					}

					try {
						Class testClass = Class.forName( className );
						if ( shouldBeAdded( testClass ) ) {

							tests.add( testClass );
						}

					} catch ( ClassNotFoundException cnfe ) {
						System.err.println( "Invalid class name: " + className );
					}
				}
			} else {
				tests.addAll( collectTests( files[i] ) );
			}
		}
		return tests;
	}


	/**
	 * Informs if a specified <code>testCass</code> is of type
	 * <code>selectiveTestType</code>
	 * 
	 * @param testClass A class to be tested.
	 * @return True if <code>testClass</code> is of type
	 *         <code>selectiveTestType</code> and false otherwise.
	 */
	private boolean shouldBeAdded( Class testClass ) {

		return isAssignableFrom( testClass, selectiveTestType );
	}


	/**
	 * Informs if a specified test class is an expecial care test.
	 * 
	 * @param testClass The class to be tested
	 * @return True if this testclass is a special test case, false otherwise.
	 */
	protected boolean isSpecialTestCase( Class testClass ) {

		return isAssignableFrom( testClass, SpecialCareTest.class );

	}


	/**
	 * Tells if a given test class is assignable from a given base class
	 * 
	 * @param testClass the test class
	 * @param baseClass the base class
	 * @return true if a given test class is assignable from a given base class,
	 *         false otherwise.
	 */
	private boolean isAssignableFrom( Class testClass, Class baseClass ) {

		if ( testClass.isInterface() ) {

			return false;

		}

		Class interfaces[] = testClass.getInterfaces();
		for ( int j = 0; j < interfaces.length; j++ ) {

			if ( baseClass.isAssignableFrom( interfaces[j] ) ) {

				return true;

			}

		}

		return false;

	}

}
