/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.Enumeration;

import junit.framework.Test;
import junit.framework.TestSuite;
import junit.textui.ResultPrinter;

/**
 * Description: This classes creates, using a
 * <code>SelectiveTestCollector</code> and runs, using a
 * <code>junit.runner,TestRunner</code>, a test suite. <br>
 * The test suite is created based on the tests type that should be executed, as
 * described in <code>org.ourgrid.test.SelectiveTestCollector.</code> <br>
 * The tests can be executed with a <code>junit.textui.TestRunner</code> or a
 * <code>junit.swingui.TestRunner</code>. This selection depends on the value
 * of the envionment variable <code>ourgrid.test.type</code>.
 * 
 * @version 1.0 Created on 05/07/2004 TODO Write unit tests for this class
 */
public class SelectiveTestsRunner {

	/**
	 * A <code>SelectiveTestCollector</code> to create the test suite.
	 */
	protected SelectiveTestCollector testCollector;

	/**
	 * Value of the environment variable <code>ourgrid.test.type</code> that
	 * Indicates that the tests should be executed with a
	 * <code>junit.textui.TestRunner</code>
	 */
	public static String TEXTRUNNER = "TEXTUI";

	/**
	 * Value of the environment variable <code>ourgrid.test.type</code> that
	 * Indicates that the tests should be executed with a
	 * <code>junit.swingui.TestRunner</code>
	 */
	public static String SWINGRUNNER = "SWINGUI";

	/**
	 * Indicates the type os TestRunner that will be used to run the tests.
	 */
	protected String runnerType = null;

	/**
	 * The interface used to collect the test classes.
	 */
	private String selectiveClassName;

	/**
	 * @author Alexandre N?brega Duarte - alex@dsc.ufcg.edu.br Description: A
	 *         class that encapsulates the execution of a test suite using a
	 *         <code>junit.swingui.TestRunner</code>
	 * @version 1.0 Created on 06/07/2004
	 */
	protected class SwingRunner extends junit.swingui.TestRunner {

		/**
		 * Creates a SwingRunner
		 */
		public SwingRunner() {

			super();
		}


		/**
		 * Runs the test suite using a <code>junit.swingui.TestRunner</code>
		 * 
		 * @param suite Test suite to be executed
		 */
		public void runTests( Test suite ) {

			fFrame = createUI( "Swing Runner" );
			fFrame.pack();
			fFrame.setVisible( true );
			runTest( suite );
		}

	}

	/**
	 * @author Alexandre N?brega Duarte - alex@dsc.ufcg.edu.br Description: A
	 *         class that encapsulates the execution of a test suite using a
	 *         <code>junit.textui.TestRunner</code>
	 * @version 1.0 Created on 06/07/2004
	 */
	protected class TextRunner extends junit.textui.TestRunner {

		/**
		 * Creates a SwingRunner
		 */
		public TextRunner() {

			super();
		}


		public TextRunner( ResultPrinter rp ) {

			super( rp );
		}


		/**
		 * Runs the test suite using a <code>junit.textui.TestRunner</code>
		 * 
		 * @param suite Test suite to be executed
		 */
		public void runTests( Test suite ) {

			doRun( suite );
		}

	}


	/**
	 * Creates a SelectiveTestsRunner setting the test type to be executed and
	 * the basedir to search for test classes.
	 * 
	 * @param selectiveClassName Full class name of the test type that should be
	 *        executed. E.g. <code>org.ougrid.tests.FastTest</code>
	 * @throws ClassNotFoundException If the <code>selectiveClassName</code>
	 *         didn't represented a valid class name.
	 */
	public SelectiveTestsRunner( String selectiveClassName ) throws ClassNotFoundException {

		this.selectiveClassName = selectiveClassName;
		runnerType = System.getProperty( "ourgrid.test.type" );
		testCollector = new SelectiveTestCollector( Class.forName( selectiveClassName ) );

	}


	/**
	 * Runs the test suite created using a <code>SelectiveTestCollector</code>
	 * The tests can be executed with a <code>junit.textui.TestRunner</code>
	 * or a <code>junit.swingui.TestRunner</code>. This selection depends on
	 * the value of the envionment variable <code>ourgrid.test.type</code>.
	 */
	public void runTests() {

		if ( runnerType != null && runnerType.trim().equals( SelectiveTestsRunner.SWINGRUNNER ) ) {
			new SwingRunner().runTests( getTestSuite() );
		} else {
			PrintStream ps;
			try {
				ps = new PrintStream( new FileOutputStream( new File( "TEST-OUTPUT-" + selectiveClassName + ".txt" ) ) );
				ResultPrinter rp = new ResultPrinter( ps );
				new TextRunner( rp ).runTests( getTestSuite() );
			} catch ( FileNotFoundException e ) {
				e.printStackTrace();
				new TextRunner().runTests( getTestSuite() );
			}
		}
	}


	/**
	 * Asks the <code>SelectiveTestCollector</code> to create the test suite.
	 * 
	 * @return A test suite where all tests are of type
	 *         <code>selectiveTestType</code>
	 */
	protected Test getTestSuite() {

		Enumeration e = testCollector.collectTests();
		TestSuite suite = new TestSuite();
		while ( e.hasMoreElements() ) {
			Class c = (Class) e.nextElement();
			suite.addTestSuite( c );
			System.out.println( c.getName() );
		}
		return suite;
	}


	/**
	 * Runs a suite of tests in which all tests implements an interface
	 * specified as a command line argument.
	 * 
	 * @param args
	 */
	public static void main( String args[] ) {

		if ( args.length < 1 ) {
			System.err.println( "Usage: java SelectiveTestsRunner fully.qualified.test.type" );
			System.err.println( "Example: java SelectiveTestsRunner org.ourgrid.test.TestModuleFastTests" );
			System.exit( ErrorCode.FAIL );
		} else {

			try {
				SelectiveTestsRunner testRunner = new SelectiveTestsRunner( args[0] );
				testRunner.runTests();
			} catch ( ClassNotFoundException cnfe ) {
				cnfe.printStackTrace();
				System.exit( ErrorCode.FAIL );
			}
		}
	}
}
