/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.rmi.registry.Registry;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.CorePeerConfiguration;
import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.config.PeerConfiguration;
import org.ourgrid.common.config.UserAgentConfiguration;
import org.ourgrid.common.exception.NotEnoughGumSpecsException;
import org.ourgrid.common.gum.Gum;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.main.CompilerException;
import org.ourgrid.common.spec.main.DescriptionFileCompile;
import org.ourgrid.gridmachine.services.GumSpecServices;
import org.ourgrid.mygrid.ui.MyGridUIManager;

/**
 * This class provides several methods to help during the test/development phase
 * of the software. It eases the customization of the unit tests.
 */
public class TestUtils {

	public static final String NONEXISTENT_MACHINE = "nonexistent.machine";

	public static final String SYSTEM_TEMP_DIR = "system.temp.dir";

	public static final String MG_HOMEMACHINE = "mg.homemachine";

	public static final String MG_PORT = "mg.port";

	public static final String MYGUMP_MACHINE = "mygump.machine";

	public static final String KEYSTORE_FILENAME = "keystore.filename";

	public static final String KEYSTORE_PASSWORD = "keystore.password";

	public static final String TYPE_UALINUX = UserAgentConfiguration.TYPE_UA_LINUX;

	public static final String TYPE_UAWINDOWS = UserAgentConfiguration.TYPE_UA_WINDOWS;

	public static final String TYPE_SWAN = UserAgentConfiguration.TYPE_UA_SWAN;

	public static final String TYPE_LOCALHOST = "localhost";

	public static final String MYGUMP_DEFAULT_PLAYPENROOT = "mygump.default.playpenroot";

	public static final String MYGUMP_DEFAULT_PORT = "mygump.default.port";

	public static final String TRUSTSTORE_FILENAME = "truststore.filename";

	public static final String TRUSTSTORE_PASSWORD = "truststore.password";

	public static final String INVALID_TRUSTSTORE_FILENAME = "invalid.truststore.filename";

	public static final String INVALID_TRUSTSTORE_PASSWORD = "invalid.truststore.password";

	public static final String MYGUMP_DEFAULT_STORAGEDIR = "mygump.default.storagedir";

	public static final String MYGUMP_DEFAULT_REMEXEC = "mygump.default.remexec";

	public static final String MYGUMP_DEFAULT_COPYFROM = "mygump.default.copyfrom";

	public static final String MYGUMP_DEFAULT_COPYTO = "mygump.default.copyto";

	public static final String OG_PORT = "peer.port";

	public static final String OG_PORT_II = "peer.port.II";

	public static final String RMI_PORT = "rmi.port";

	public static final String GUI_PORT = "gui.port";

	public static final String FAKE_PORT = "fake.port";

	public static final String FAKE_MACHINE_NAME = "fake.machine";

	public static final String FAKE_FILE = "fake.file";

	public static final String NOT_WRITABLE_FILE = "protected.file";

	public static final String REMOTE_PROTECTED_FILE = "remote.protected.file";

	public static final String REMOTE_FAKE_FILE = "remote.fake.file";

	public static final String REMOTE_TEMP_DIR = "remote.temp.dir";

	private static final String[ ] UA_GUMSPEC_ATT_ORDER = { Configuration.PROP_NAME, Configuration.PROP_PORT,
															UserAgentConfiguration.ATT_PLAYPEN_ROOT,
															UserAgentConfiguration.ATT_STORAGE_DIR,
															UserAgentConfiguration.ATT_REM_EXEC,
															UserAgentConfiguration.ATT_COPY_TO,
															UserAgentConfiguration.ATT_COPY_FROM };

	private Properties testProperties;

	private static final String DEFAULT_TEST_PROPERTIES_FILENAME = "test" + File.separatorChar + "resources"
			+ File.separatorChar + "test.properties";

	private boolean ogPropertiesWasCreated = false;

	private boolean corepeerPropertiesWasCreated = false;

	private boolean mgPropertiesWasCreated = false;

	private static final String GUMDEFAULTS = "GUMDEFAULTS";

	private static final String PEER = "PEER";

	public static final String PROTECTED_DIR = "protected.dir";

	public static final String REMOTE_PROTECTED_DIR = "remote.protected.dir";

	public static final String REMOTE_FILE = "remote.file";

	private static final long TESTUA_TIMEOUT = 2000;

	private static final String SH = "sh";

	private static final String BIN = "bin" + File.separator;

	public static final String MYGRID_COMMAND = BIN + "mygrid";

	public static final String PEER_COMMAND = BIN + "peer";

	public static final String CORE_PEER_COMMAND = BIN + "corepeer";

	public static final String COREPEER_PORT = "corepeer.port";


	/**
	 * Loads the tests properties from the default file.
	 */
	public void load() {

		load( DEFAULT_TEST_PROPERTIES_FILENAME );
	}


	/**
	 * Loads the tests properties from the specified file.
	 * 
	 * @param fileName The file name.
	 */
	public void load( String fileName ) {

		this.testProperties = new Properties();

		FileInputStream fileInputStream;

		try {
			fileInputStream = new FileInputStream( fileName );
			testProperties.load( fileInputStream );
		} catch ( IOException e ) {
			System.err.println( "Could not load test properties." );
			e.printStackTrace();
			System.exit( ErrorCode.COULD_NOT_LOAD_TEST_PROPERTIES );
		}

	}


	/**
	 * Resets all data.
	 */
	public void reset() {

		testProperties = new Properties();
		if ( mgPropertiesWasCreated ) {
			if ( (new File( getMGPropertiesFileName() + "~" )).exists() ) {
				restoreTo( getMGPropertiesFileName() + "~", getMGPropertiesFileName() );
			} else {
				(new File( getMGPropertiesFileName() )).delete();
			}
			this.mgPropertiesWasCreated = false;
		}
		if ( ogPropertiesWasCreated ) {
			if ( (new File( getOGPropertiesFileName() + "~" )).exists() ) {
				restoreTo( getOGPropertiesFileName() + "~", getOGPropertiesFileName() );
			} else {
				(new File( getOGPropertiesFileName() )).delete();
			}
			this.ogPropertiesWasCreated = false;
		}

		if ( corepeerPropertiesWasCreated ) {
			if ( (new File( getCorepeerPropertiesFileName() + "~" )).exists() ) {
				restoreTo( getCorepeerPropertiesFileName() + "~", getCorepeerPropertiesFileName() );
			} else {
				(new File( getCorepeerPropertiesFileName() )).delete();
			}
			this.corepeerPropertiesWasCreated = false;
		}
	}


	/**
	 * Restores the file1 to file2.
	 * 
	 * @param file1 The file that will replace the file2.
	 * @param file2 The file that will be replaced by file1.
	 */
	private void restoreTo( String file1, String file2 ) {

		File a = new File( file1 );
		File b = new File( file2 );
		a.renameTo( b );
	}


	/**
	 * Returns the property specified.
	 * 
	 * @param propertyKey The property key.
	 * @return Returns the property value or <code>null</code> case the
	 *         property was not previously specified.
	 */
	public String getProperty( String propertyKey ) {

		return testProperties.getProperty( propertyKey );
	}


	/**
	 * Sets a property.
	 * 
	 * @param propertyKey The property key.
	 * @param propertyValue The property value.
	 */
	public void setProperty( String propertyKey, String propertyValue ) {

		testProperties.setProperty( propertyKey, propertyValue );
	}


	/**
	 * Returns a List containing GuMSpecs that were defined as usable in the
	 * test.properties file.
	 * 
	 * @param numberOfGums The number of GuMSpecs desired.
	 * @param type The type of the GuMSpecs (if they will be UALinux, UAWindows,
	 *        etc.)
	 * @return A List of GuMSpecs.
	 * @throws NotEnoughGumSpecsException Case the number of GuMSpecs available
	 *         is less than the number of GuMSpecs required.
	 */
	public List<GumSpec> getGumSpecs( int numberOfGums, String type ) throws NotEnoughGumSpecsException {

		List<GumSpec> gumSpecsList = new Vector<GumSpec>();
		GumSpec currentGumSpec;
		Enumeration keys = testProperties.keys();
		while ( gumSpecsList.size() < numberOfGums && keys.hasMoreElements() ) {
			String currentkey = keys.nextElement().toString();
			if ( currentkey.startsWith( MYGUMP_MACHINE + "." + type ) ) {
				currentGumSpec = generateGuMSpec( type, getProperty( currentkey ) );
				gumSpecsList.add( currentGumSpec );
			}
		}

		if ( gumSpecsList.size() < numberOfGums )
			throw new NotEnoughGumSpecsException( "Only " + gumSpecsList.size() + " GuM(s) available" );

		return gumSpecsList;
	}


	/**
	 * Tests if all UA in a GUM are alive. It does it only once.
	 * 
	 * @param GuMSpecsList List of GuMs
	 * @return
	 */
	public boolean testUA( List GuMSpecsList ) {

		return testUA( GuMSpecsList, TESTUA_TIMEOUT );
	}


	/**
	 * Tests if all UA in a GUM are alive. It does it only timeout milisecs.
	 * 
	 * @param GuMSpecsList List of GuMs
	 * @return
	 */
	public boolean testUA( List GuMSpecsList, long timeout ) {

		boolean GuMsAlive = false;
		long timeexpected = 0;

		try {
			while ( (timeexpected < timeout) && !(GuMsAlive) ) {
				GuMsAlive = true;
				timeexpected += TESTUA_TIMEOUT;
				Iterator GuMSpecsListIter = GuMSpecsList.iterator();
				Thread.sleep( TESTUA_TIMEOUT );
				while ( GuMSpecsListIter.hasNext() ) {
					GumSpec currentGuMSpec = (GumSpec) GuMSpecsListIter.next();

					Gum uaGuM;
					try {
						uaGuM = GumSpecServices.createGum( currentGuMSpec );

						if ( uaGuM.getStatus() == GumStatus.CONTACTING ) {
							GuMsAlive = false;
						} else {
							GuMsAlive = true;
						}

					} catch ( Exception e ) {
						GuMsAlive = false;
					}
				}
			}
		} catch ( InterruptedException e ) {
			// Nothing to be done

		}

		return GuMsAlive;
	}


	/**
	 * Generates a GuMSpec for LocalMachine.
	 * 
	 * @param type The GuM type.
	 * @return GuMSpec of the localmachine.
	 */
	public GumSpec getLocalGuMSpec( String type ) {

		GumSpec gumSpec;
		try {
			gumSpec = getGumSpecs( 1, TYPE_LOCALHOST ).get( 0 );
		} catch ( NotEnoughGumSpecsException e ) {
			gumSpec = new GumSpec();
			gumSpec.putAttribute( Configuration.PROP_NAME, "localhost" );
			gumSpec.putAttribute( Configuration.PROP_PORT, Integer.toString( Registry.REGISTRY_PORT ) );
		}
		gumSpec.putAttribute( UserAgentConfiguration.ATT_TYPE, type );
		return gumSpec;
	}


	/**
	 * Generates a GuMSpec from a String in the GuMSpecAttOrder format.
	 * 
	 * @param type GuMSpec type.
	 * @param attMachineString String of attributes.
	 */
	private GumSpec generateGuMSpec( String type, String attMachineString ) {

		Map<String,String> guMSpecAtt = new HashMap<String,String>( getDefaultGuMSpecAtt() );

		guMSpecAtt.put( UserAgentConfiguration.ATT_TYPE, type );

		String[ ] guMSpecAttOrder = null;

		guMSpecAttOrder = UA_GUMSPEC_ATT_ORDER;

		// Removes the "{}"
		attMachineString = attMachineString.substring( 1, attMachineString.length() - 1 );

		StringTokenizer buildString = new StringTokenizer( attMachineString, "," );

		int index = 0;

		while ( buildString.hasMoreTokens() )
			guMSpecAtt.put( guMSpecAttOrder[index++], ((String) buildString.nextElement()).trim() );

		return new GumSpec( guMSpecAtt );
	}


	/**
	 * Creates a default GuMSpec attribute's map common to all GuMs.
	 * 
	 * @return A default attribute map for all GuMs.
	 */
	private Map<String,String> getDefaultGuMSpecAtt() {

		Map<String,String> defaultGuMSpecAtt = new HashMap<String,String>();

		if ( testProperties.containsKey( MYGUMP_DEFAULT_PLAYPENROOT ) )
			defaultGuMSpecAtt.put( UserAgentConfiguration.ATT_PLAYPEN_ROOT, getProperty( MYGUMP_DEFAULT_PLAYPENROOT ) );
		if ( testProperties.containsKey( MYGUMP_DEFAULT_PORT ) )
			defaultGuMSpecAtt.put( Configuration.PROP_PORT, getProperty( MYGUMP_DEFAULT_PORT ) );
		if ( testProperties.containsKey( MYGUMP_DEFAULT_STORAGEDIR ) )
			defaultGuMSpecAtt.put( UserAgentConfiguration.ATT_STORAGE_DIR, getProperty( MYGUMP_DEFAULT_STORAGEDIR ) );
		if ( testProperties.containsKey( MYGUMP_DEFAULT_REMEXEC ) )
			defaultGuMSpecAtt.put( UserAgentConfiguration.ATT_REM_EXEC, getProperty( MYGUMP_DEFAULT_REMEXEC ) );
		if ( testProperties.containsKey( MYGUMP_DEFAULT_COPYFROM ) )
			defaultGuMSpecAtt.put( UserAgentConfiguration.ATT_COPY_FROM, getProperty( MYGUMP_DEFAULT_COPYFROM ) );
		if ( testProperties.containsKey( MYGUMP_DEFAULT_COPYTO ) )
			defaultGuMSpecAtt.put( UserAgentConfiguration.ATT_COPY_TO, getProperty( MYGUMP_DEFAULT_COPYTO ) );

		return defaultGuMSpecAtt;
	}


	/**
	 * Returns the number of GuMSpecs available according to the test.properties
	 * file.
	 * 
	 * @param type The type of the GuMSpecs.
	 * @return The number of GuMSpecs available from the type specified.
	 */
	public int getNumberOfGuMSpecsAvailable( String type ) {

		int numberofgums = 0;
		Enumeration keys = testProperties.keys();

		while ( keys.hasMoreElements() ) {
			String currentkey = keys.nextElement().toString();
			if ( currentkey.startsWith( MYGUMP_MACHINE + "." + type + "." ) ) {
				numberofgums++;
			}

		}

		return numberofgums;
	}


	/**
	 * Generates a GDF and returns the full path to it.
	 * 
	 * @param guMSpecs A list of GuMSpecs that will compose the Grid Description
	 *        File.
	 * @return The full path to the GDF created.
	 * @throws IOException Case file cannot be created (disk full or wrong
	 *         setted temp directory)
	 */
	public String createSDF( List guMSpecs ) throws IOException {

		File tempSDF = File.createTempFile( "mygrid-test", ".sdf", new File( testProperties
				.getProperty( SYSTEM_TEMP_DIR ) ) );

		BufferedWriter writer = new BufferedWriter( new FileWriter( tempSDF ) );

		writer.write( GUMDEFAULTS + ": " );
		writer.newLine();

		if ( testProperties.containsKey( MYGUMP_DEFAULT_PORT ) ) {
			writer.write( " port : " + testProperties.getProperty( MYGUMP_DEFAULT_PORT ) );
			writer.newLine();
		}
		if ( testProperties.containsKey( MYGUMP_DEFAULT_PLAYPENROOT ) ) {
			writer.write( " playpenroot : " + testProperties.getProperty( MYGUMP_DEFAULT_PLAYPENROOT ) );
			writer.newLine();
		}
		if ( testProperties.containsKey( MYGUMP_DEFAULT_STORAGEDIR ) ) {
			writer.write( " storagedir : " + testProperties.getProperty( MYGUMP_DEFAULT_STORAGEDIR ) );
			writer.newLine();
		}
		if ( testProperties.containsKey( MYGUMP_DEFAULT_REMEXEC ) ) {
			writer.write( " remexec : " + testProperties.getProperty( MYGUMP_DEFAULT_REMEXEC ) );
			writer.newLine();
		}
		if ( testProperties.containsKey( MYGUMP_DEFAULT_COPYTO ) ) {
			writer.write( " copyto : " + testProperties.getProperty( MYGUMP_DEFAULT_COPYTO ) );
			writer.newLine();
		}
		if ( testProperties.containsKey( MYGUMP_DEFAULT_COPYFROM ) ) {
			writer.write( " copyfrom : " + testProperties.getProperty( MYGUMP_DEFAULT_COPYFROM ) );
		}
		writer.newLine();

		Iterator iterGuMSpecs = guMSpecs.iterator();
		while ( iterGuMSpecs.hasNext() ) {
			GumSpec gum = (GumSpec) iterGuMSpecs.next();
			writer.write( "GUM : " );
			writer.newLine();
			Set keySet = gum.getAttributes().keySet();
			Iterator iterAtts = keySet.iterator();
			while ( iterAtts.hasNext() ) {
				String Att = (String) iterAtts.next();
				writer.write( " " + Att + " : " + gum.getAttribute( Att ) );
				writer.newLine();
			}

		}

		writer.close();

		tempSDF.deleteOnExit();

		return tempSDF.toString();
	}


	/**
	 * Returns the path to the MG.properties file.
	 * 
	 * @return A String containing the path to the MG.properties file.
	 */
	public String getMGPropertiesFileName() {

		return MyGridConfiguration.PROPERTIES_FILENAME;
	}


	/**
	 * Creates a MG.properties file in the
	 * <code>getMGPropertiesFileName()</code> location. It will backup the old
	 * MG.properties file to be restored when <code>this.reset()</code> is
	 * invoked.
	 */
	public void createMGProperties() {

		this.mgPropertiesWasCreated = true;
		createProperties( getMGPropertiesFileName(), getMGPropertiesFileName() + "~", "mg." );
	}


	private void createProperties( String originalFileName, String backUpFileName, String prefix ) {

		File prop = new File( originalFileName );
		File propBackup = new File( backUpFileName );
		prop.renameTo( propBackup );
		Properties props = new Properties();

		try {
			loadPropertiesInto( props, prefix );
			FileOutputStream outputStream = new FileOutputStream( originalFileName );
			props.store( outputStream, "New " + originalFileName );
			outputStream.close();
		} catch ( Exception e ) {
			System.err.println( "Could not create a new " + backUpFileName + "." );
			e.printStackTrace();
		}
	}


	/**
	 * Loads the properties refering to MG.properties from the
	 * <code>DEFAULT_TEST_PROPERTIES_FILENAME</code>.
	 * 
	 * @param props Where the properties will be loaded into.
	 * @param prefix The prefix of the properties that will be loaded into.
	 */
	private void loadPropertiesInto( Properties props, String prefix ) {

		Iterator iterator = testProperties.keySet().iterator();
		while ( iterator.hasNext() ) {
			String propertyKey = (String) iterator.next();
			if ( propertyKey.startsWith( prefix ) ) {
				props.setProperty( propertyKey, testProperties.getProperty( propertyKey ) );
			}
		}
	}


	/**
	 * Returns the path to the peer.properties file.
	 * 
	 * @return A String containing the path to the peer.properties file.
	 */
	public String getOGPropertiesFileName() {

		return PeerConfiguration.PROPERTIES_FILENAME;
	}


	/**
	 * Creates an peer.properties file in the
	 * <code>getOGPropertiesFileName()</code> location. It will backup the old
	 * peer.properties file to be restored when <code>this.reset()</code> is
	 * invoked.
	 */
	public void createOGProperties() {

		this.ogPropertiesWasCreated = true;
		createProperties( getOGPropertiesFileName(), getOGPropertiesFileName() + "~", "peer." );
	}


	/**
	 * Returns the path to the corepeer.properties file.
	 * 
	 * @return A String containing the path to the peer.properties file.
	 */
	public String getCorepeerPropertiesFileName() {

		return CorePeerConfiguration.PROPERTIES_FILENAME;
	}


	/**
	 * Creates an corepeer.properties file in the
	 * <code>getCorepeerPropertiesFileName()</code> location. It will backup
	 * the old corepeer.properties file to be restored when
	 * <code>this.reset()</code> is invoked.
	 */
	public void createCorepeerProperties() {

		this.corepeerPropertiesWasCreated = true;
		createProperties( getCorepeerPropertiesFileName(), getCorepeerPropertiesFileName() + "~", "corepeer." );
	}


	/**
	 * @throws IOException
	 */
	public String createLocalPeerGDF() throws IOException {

		File tempGDF = File.createTempFile( "ourgrid-test-", ".gdf", new File( testProperties
				.getProperty( SYSTEM_TEMP_DIR ) ) );

		BufferedWriter writer = new BufferedWriter( new FileWriter( tempGDF ) );

		writer.write( PEER + ": " );
		writer.newLine();
		writer.write( " label : local peer" );
		writer.newLine();
		writer.write( " name : 127.0.0.1" );
		writer.newLine();
		if ( testProperties.containsKey( OG_PORT ) ) {
			writer.write( " port : " + testProperties.getProperty( OG_PORT ) );
			writer.newLine();
		} else {
			writer.write( " port : 10991" );
			writer.newLine();
		}

		writer.close();

		tempGDF.deleteOnExit();

		return tempGDF.toString();
	}


	/**
	 * Runs a command.
	 * 
	 * @param commands The command.
	 * @param timeWaiting The time to sleep after the command was performed -
	 *        not recommended to use.
	 * @throws Exception Case there's any problem.
	 */
	public static void runCommand( String[ ] commands, long timeWaiting ) throws Exception {

		String command = "";
		for ( int k = 0; k < commands.length; k++ ) {
			command += commands[k] + " ";
		}

		System.out.println( "Executing: " + command );

		Runtime exec = Runtime.getRuntime();
		Process process = exec.exec( commands );

		StreamGobbler streamGobblerStd = new StreamGobbler( process.getInputStream(), "STD" );
		streamGobblerStd.start();
		StreamGobbler streamGobblerErr = new StreamGobbler( process.getErrorStream(), "ERROR" );
		streamGobblerErr.start();

		int exitValue = 0;

		try {

			exitValue = process.waitFor();
		} catch ( Exception e ) {
			e.printStackTrace();
		}

		if ( exitValue != 0 ) {
			System.err.println( "Finished (with error): " + command );
			throw new Exception( command + "exit with value " + exitValue );
		}

		Thread.sleep( timeWaiting );
		System.out.println( "Finished: " + command );
	}


	public boolean checkFileContent( String expectedContent, File file ) throws Exception {

		BufferedReader bufferedReader = new BufferedReader( new FileReader( file ) );
		String line = bufferedReader.readLine();
		String contentInFile = "";
		while ( line != null ) {
			contentInFile += line;
			line = bufferedReader.readLine();
		}
		System.out.println( "expected: " + expectedContent );
		System.out.println( "obtained: " + contentInFile );
		bufferedReader.close();
		return expectedContent.equals( contentInFile );
	}


	/**
	 * Executes a command with its arguments.
	 * 
	 * @param command The command to be executed.
	 * @param args Its arguments.
	 * @throws Exception Case the command cannot be executed.
	 */
	public static void executeCommand( String command, String[ ] args ) throws Exception {

		String[ ] commandArray = new String[ args.length + 2 ];
		commandArray[0] = SH;
		commandArray[1] = command;
		for ( int k = 0; k < args.length; k++ ) {
			commandArray[k + 2] = args[k];
		}
		TestUtils.runCommand( commandArray, 0 );
	}


	/**
	 * Wait until mygrid is up with a busy-wait call.
	 */
	public void waitUntilMyGridIsUp() {

		int retryCount = 0;
		while ( !isMyGridRunning() && retryCount < 60 ) {
			retryCount++;
			try {
				Thread.sleep( 1000 );
			} catch ( InterruptedException e ) {
				// Thread interrupt, ignore.
			}
		}

		if ( !isMyGridRunning() )
			throw new RuntimeException( "MyGrid is not Running" );

	}


	/**
	 * Wait until mygrid is down with a busy-wait call.
	 */
	public void waitUntilMyGridIsDown() {

		int retryCount = 0;
		while ( isMyGridRunning() && retryCount < 60 ) {
			retryCount++;
			try {
				Thread.sleep( 1000 );
			} catch ( InterruptedException e ) {
				// Thread interrupt, ignore.
			}
		}
		if ( isMyGridRunning() )
			throw new RuntimeException( "MyGrid still Running" );
	}


	private boolean isMyGridRunning() {

		boolean result;
		try {
			result = MyGridUIManager.getInstance().isMyGridRunning();
		} catch ( Exception e ) {
			result = false;
		}
		return result;
	}


	public JobSpec getSimpleJobSpec() throws CompilerException {

		return DescriptionFileCompile.compileJDF( getProperty( "job.simple" ) );
	}


	public void restartMygrid() throws Exception {

		TestUtils.executeCommand( TestUtils.MYGRID_COMMAND, new String[ ] { "stop" } );

		MyGridUIManager.reset();

		TestUtils.executeCommand( TestUtils.MYGRID_COMMAND, new String[ ] { "start" } );
		waitUntilMyGridIsUp();
	}


	public void stopMygrid() throws Exception {

		TestUtils.executeCommand( TestUtils.MYGRID_COMMAND, new String[ ] { "stop" } );

		MyGridUIManager.reset();
	}


	public int getRMIPort() {

		return Integer.parseInt( getProperty( TestUtils.RMI_PORT ) );
	}


	public int guiPort() {

		final String property = getProperty( TestUtils.GUI_PORT );
		System.out.println( "pro: " + property );
		return Integer.parseInt( property );
	}

}


class StreamGobbler extends Thread {

	InputStream is;

	String type;


	StreamGobbler( InputStream is, String type ) {

		this.is = is;
		this.type = type;
	}


	public void run() {

		try {
			InputStreamReader isr = new InputStreamReader( is );
			BufferedReader br = new BufferedReader( isr );
			String line = null;
			while ( (line = br.readLine()) != null )
				System.out.println( type + "> " + line );
		} catch ( IOException ioe ) {
			ioe.printStackTrace();
		}
	}
}
