package org.ourgrid.test.functionaltests;

import java.util.HashMap;

import junit.framework.TestCase;

import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.peer.manager.allocation.AllocationStatus;
import org.ourgrid.test.FunctionalStressTest;
import org.ourgrid.test.functionaltests.busywait.ConditionWaiter;
import org.ourgrid.test.functionaltests.busywait.JobWaiterRunnable;
import org.ourgrid.test.functionaltests.busywait.MyGridGumsCondition;
import org.ourgrid.test.functionalunits.CorePeerUnit;
import org.ourgrid.test.functionalunits.FunctionalUnitsControl;
import org.ourgrid.test.functionalunits.MyGridUnit;
import org.ourgrid.test.functionalunits.PeerUnit;
import org.ourgrid.test.functionalunits.UserAgentUnit;
import org.ourgrid.test.threadServices.ThreadServices;

/**
 * This test will group all functional tests that are related to redistribution
 * and community requests. Each test must describe a scenario and also start and
 * stop all functional test units that will be used.
 */
public class NetworkOfFavoursFunctionalTests extends TestCase implements FunctionalStressTest {

	/**
	 * Reset the <code>FunctionTestsControl</code> to clear any properties
	 * configuration that was used before.
	 * 
	 * @throws Exception
	 * @see junit.framework.TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {

		FunctionalUnitsControl.getInstance().destroyAllRemainingUnits();
		FunctionalUnitsControl.getInstance().reset();
	}


	/**
	 * Test scenario:<br> - MyGrid local and Peer local. Description:<br> -
	 * Test if a Gum is donated from a peer to another in a simple scenario.
	 * Functional tests units:<br> - Core peer,<br> - Peer A,<br> - Peer B,<br> -
	 * MyGrid.<br>
	 * Test scenario workflow:<br> - Start mygrid,<br> - Set mygrid with Peer
	 * A,<br> - Set Peer A corepeer,<br> - Set Peer B corepeer,<br> - Start
	 * Peer A,<br> - Start Peer B,<br> - Set Gums at Peer B,<br> - Add job at
	 * MyGrid,<br> - Wait until MyGrid job is finished,<br> - Clear scenario.
	 * 
	 * @throws Exception If any exception is thrown the test fails.
	 */
	public void testMyGridPeerInteraction() throws Exception {

		MyGridUnit testMygrid = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "TestMyGrid" );
		PeerUnit testPeer = FunctionalUnitsControl.getInstance().createANewPeerUnit( "TestPeer" );

		testPeer.start();

		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 1 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}
		testPeer.setGums( userAgentUnits );

		testMygrid.start();
		testMygrid.setPeer( testPeer );

		int jobId = testMygrid.addJob( FunctionalUnitsControl.JOB_ECHO );

		new Thread( new JobWaiterRunnable( jobId, testMygrid ) ).start();
		ThreadServices.waitUntilWorkIsDone();
		assertEquals( ExecutionStatus.FINISHED, testMygrid.getJobStatus( jobId ) );
	}


	/**
	 * Test scenario:<br> - Peer A, Peer B, MyGrid at Peer A, Gums at Peer B.
	 * Description:<br> - Test if a Gum is donated from a peer to another in a
	 * simple scenario. Functional tests units:<br> - Core peer,<br> - Peer A,<br> -
	 * Peer B,<br> - MyGrid.<br>
	 * Test scenario workflow:<br> - Start mygrid,<br> - Set mygrid with Peer
	 * A,<br> - Set Peer A corepeer,<br> - Set Peer B corepeer,<br> - Start
	 * Peer A,<br> - Start Peer B,<br> - Set Gums at Peer B,<br> - Add job at
	 * MyGrid,<br> - Wait until MyGrid job is finished,<br> - Clear scenario.
	 * 
	 * @throws Exception If any exception is thrown the test fails.
	 */
	public void testPeerPeerCorepeerInteraction() throws Exception {

		MyGridUnit testMygrid = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "TestMyGrid" );
		PeerUnit testPeerA = FunctionalUnitsControl.getInstance().createANewPeerUnit( "TestPeerA" );
		PeerUnit testPeerB = FunctionalUnitsControl.getInstance().createANewPeerUnit( "TestPeerB" );
		CorePeerUnit testCorePeer = FunctionalUnitsControl.getInstance().createANewCorePeerUnit( "TestCorePeerA" );

		testCorePeer.start();

		testPeerA.setCorePeer( testCorePeer );
		testPeerB.setCorePeer( testCorePeer );

		testPeerA.start();
		testPeerB.start();

		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 1 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}
		testPeerB.setGums( userAgentUnits );

		testMygrid.start();
		testMygrid.setPeer( testPeerA );

		int jobId = testMygrid.addJob( FunctionalUnitsControl.JOB_ECHO );

		new Thread( new JobWaiterRunnable( jobId, testMygrid ) ).start();
		ThreadServices.waitUntilWorkIsDone();

		assertEquals( ExecutionStatus.FINISHED, testMygrid.getJobStatus( jobId ) );
	}


	/**
	 * Test scenario:<br> - Peer A, Peer B, MyGrid at Peer A, Gums at Peer B,
	 * MyGrid at Peer B. Description:<br> - Tests if a local broker will have
	 * preference over a remote broker. Functional tests units:<br> - Core
	 * peer,<br> - Peer A,<br> - Peer B,<br> - MyGrid A.<br> - MyGrid B.<br>
	 * Test scenario workflow:<br> - Start MyGrid A,<br> - Set MyGrid A with
	 * Peer A,<br> - Start MyGrid B,<br> - Set MyGrid B with Peer B,<br> -
	 * Set Peer A corepeer,<br> - Set Peer B corepeer,<br> - Start Peer A,<br> -
	 * Start Peer B,<br> - Set Gums at Peer B,<br> - Add job at MyGrid A,<br> -
	 * Wait until MyGrid A job is running,<br> - Add job at MyGrid B,<br> -
	 * Wait until MyGrid B job is running,<br> - Verify that MyGrid A job is
	 * failed, - Wait until MyGrid B job is finished,<br> - Wait until MyGrid A
	 * job is finished,<br> - Clear scenario.
	 * 
	 * @throws Exception If any exception is thrown the test fails.
	 */
	public void testLocalPreemption() throws Exception {

		MyGridUnit testMygridA = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "TestMyGridA" );
		MyGridUnit testMygridB = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "TestMyGridB" );
		PeerUnit testPeerA = FunctionalUnitsControl.getInstance().createANewPeerUnit( "TestPeerA" );
		PeerUnit testPeerB = FunctionalUnitsControl.getInstance().createANewPeerUnit( "TestPeerB" );
		CorePeerUnit testCorePeer = FunctionalUnitsControl.getInstance().createANewCorePeerUnit( "TestCorePeerA" );

		testCorePeer.start();

		testPeerA.setCorePeer( testCorePeer );
		testPeerB.setCorePeer( testCorePeer );

		testPeerA.start();
		testPeerB.start();

		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 2 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}
		testPeerB.setGums( userAgentUnits );

		testMygridA.start();
		testMygridA.setPeer( testPeerA );

		testMygridB.start();
		testMygridB.setPeer( testPeerB );

		int jobIdA = testMygridA.addJob( FunctionalUnitsControl.JOB_MORE_SLEEP );

		// Wait until MygridA has both gums
		ConditionWaiter.waitForCondition( new MyGridGumsCondition( testMygridA, 2 ) );
		assertEquals( 2, testMygridA.getRunningGums().size() );

		int jobIdB = testMygridB.addJob( FunctionalUnitsControl.JOB_SLEEP );

		// Wait until MygridB has both gums, also make sure MyGrid A has not
		// finished its job
		ConditionWaiter.waitForCondition( new MyGridGumsCondition( testMygridB, 2 ) );
		assertFalse( testMygridA.getJobStatus( jobIdA ).isFinished() );
		assertEquals( 2, testMygridB.getRunningGums().size() );

		// Wait until MygridA has no gums
		ConditionWaiter.waitForCondition( new MyGridGumsCondition( testMygridA, 0 ) );
		assertEquals( 0, testMygridA.getRunningGums().size() );
		assertFalse( testMygridA.getJobStatus( jobIdA ).isFinished() );

		Thread waiterThreadB = new Thread( new JobWaiterRunnable( jobIdB, testMygridB ) );
		waiterThreadB.start();
		ThreadServices.waitUntilThreadHasFinished( waiterThreadB );
		assertTrue( testMygridB.getJobStatus( jobIdB ).isFinished() );

		// Wait until MygridB has no gums
		ConditionWaiter.waitForCondition( new MyGridGumsCondition( testMygridB, 0 ) );
		assertEquals( 0, testMygridB.getRunningGums().size() );

		// Wait until MygridA has both gums again, and until its job finishes.
		ConditionWaiter.waitForCondition( new MyGridGumsCondition( testMygridA, 2 ) );
		assertEquals( 2, testMygridA.getRunningGums().size() );
	}


	/**
	 * Test scenario:<br> - Peer, two local MyGrids. Description:<br> - Tests
	 * if the peer do a local redistribution. Functional tests units:<br> -
	 * Peer,<br> - MyGrid A.<br> - MyGrid B.<br>
	 * Test scenario workflow:<br> - Start MyGrid A,<br> - Set MyGrid A with
	 * Peer,<br> - Start MyGrid B,<br> - Set MyGrid B with Peer,<br> - Start
	 * Peer,<br> - Set Gums at Peer,<br> - Add job at MyGrid A,<br> - Wait
	 * until MyGrid A job is running,<br> - Add job at MyGrid B,<br> - Wait
	 * until MyGrid B job is running,<br> - Verify that MyGrid A has the same
	 * number of gums than MyGrid B, - Clear scenario.
	 * 
	 * @throws Exception If any exception is thrown the test fails.
	 */
	public void testPeerMyGridLocalAndMyGridLocal() throws Exception {

		MyGridUnit testMygridA = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "TestMyGridA" );
		MyGridUnit testMygridB = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "TestMyGridB" );
		PeerUnit testPeer = FunctionalUnitsControl.getInstance().createANewPeerUnit( "TestPeer" );

		testPeer.start();
		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 4 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}
		testPeer.setGums( userAgentUnits );

		testMygridA.start();
		testMygridA.setPeer( testPeer );

		testMygridB.start();
		testMygridB.setPeer( testPeer );

		int jobIdA = testMygridA.addJob( FunctionalUnitsControl.JOB_MORE_SLEEP );
		ConditionWaiter.waitForCondition( new MyGridGumsCondition( testMygridA, 4 ) );
		assertEquals( testMygridA.getRunningGums().size(), 4 );
		assertEquals( 4, testPeer.gumsInState( AllocationStatus.IN_USE ).size() );

		testMygridB.addJob( FunctionalUnitsControl.JOB_MORE_SLEEP );
		ConditionWaiter.waitForCondition( new MyGridGumsCondition( testMygridB, 2 ) );
		assertEquals( testMygridB.getRunningGums().size(), 2 );
		assertFalse( testMygridA.getJobStatus( jobIdA ).isFinished() );
		assertEquals( testMygridA.getRunningGums().size(), 2 );

		assertEquals( 4, testPeer.gumsInState( AllocationStatus.IN_USE ).size() );
	}


	/**
	 * Test scenario:<br> - A Community with 3 Peers, 3 MyGrids for each peer.
	 * Description:<br> - Tests if the remote peer do a local redistribution to
	 * the . Functional tests units:<br> - Peers: A, B and C,<br> - MyGrids:
	 * A, B and C.<br>
	 * Test scenario workflow:<br> - Start Core Peer - Start Peers A and B -
	 * Associate User Agents to respective peers - Start MyGrid B and associate
	 * to the Peer B - Add Job from MyGrid B (to receive machines from Peer A) -
	 * Wait until all jobs associated to MyGrid B is done. - Start Peer C -
	 * Start MyGrid C and associate to the Peer C - Add Job from MyGrid C (to
	 * receive machines from Peer A) - Wait until MyGrid C gets 4 machines -
	 * Start MyGrid A and associate to the Peer A (that have a considerable
	 * value into Network of Favors on Peer B) - Wait until MyGrid A gets 4
	 * machines - Verify if MyGrid A have 4 machines and MyGrid C have 0
	 * machines - Wait while finish jobs in MyGrid A - Clear scenario.
	 * 
	 * @throws Exception If any exception is thrown the test fails.
	 */
	public void testNetworkOfFavorsRedistributions() throws Exception {

		CorePeerUnit corePeer = FunctionalUnitsControl.getInstance().createANewCorePeerUnit( "CorePeer" );

		PeerUnit peerA = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerA" );
		PeerUnit peerB = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerB" );
		PeerUnit peerC = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerC" );

		MyGridUnit mygridA = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "MyGridA" );
		MyGridUnit mygridB = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "MyGridB" );
		MyGridUnit mygridC = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "MyGridC" );

		System.out.println( "[DEBUG] Starting Core Peer" );

		corePeer.start();

		System.out.println( "[DEBUG] Setting Peers A & B" );

		peerA.setCorePeer( corePeer );
		peerB.setCorePeer( corePeer );
		peerC.setCorePeer( corePeer );

		System.out.println( "[DEBUG] Starting Peers A & B" );

		peerA.start();
		peerB.start();

		System.out.println( "[DEBUG] Starting 4 User Agents on Peer A" );

		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 4 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}
		peerA.setGums( userAgentUnits );

		System.out.println( "[DEBUG] Starting MyGrid B" );

		mygridB.start();
		mygridB.setPeer( peerB );

		int jobB = mygridB.addJob( FunctionalUnitsControl.JOB_SLEEP );

		System.out.println( "[DEBUG] Waiting MyGrid B receive 4 machines" );

		ConditionWaiter.waitForCondition( new MyGridGumsCondition( mygridB, 4 ) );
		assertEquals( mygridB.getRunningGums().size(), 4 );

		System.out.println( "[DEBUG] Waiting Job finish on MyGrid B" );

		new Thread( new JobWaiterRunnable( jobB, mygridB ) ).start();
		ThreadServices.waitUntilWorkIsDone();
		assertEquals( mygridB.getJobStatus( jobB ), ExecutionStatus.FINISHED );
		assertEquals( mygridB.getRunningGums().size(), 0 );

		System.out.println( "[DEBUG] Stopping User Agents on Peer A" );

		peerA.unsetGums();
		assertEquals( peerA.gumsInState( AllocationStatus.CONTACTING ).size(), 0 );
		assertEquals( peerA.gumsInState( AllocationStatus.DONATED ).size(), 0 );
		assertEquals( peerA.gumsInState( AllocationStatus.IDLE ).size(), 0 );
		assertEquals( peerA.gumsInState( AllocationStatus.IN_USE ).size(), 0 );

		assertNotNull( peerB.getRankingForPeer( peerA ) );

		System.out.println( "[DEBUG] Starting 4 User Agents on Peer B" );

		UserAgentUnit[ ] userAgentUnits2 = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 4 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits2 ) {
			userAgentUnit.start();
		}
		peerB.setGums( userAgentUnits2 );

		System.out.println( "[DEBUG] Starting Peer C" );

		peerC.start();

		System.out.println( "[DEBUG] Starting MyGrid C" );

		mygridC.start();
		mygridC.setPeer( peerC );

		assertNull( peerB.getRankingForPeer( peerC ) );

		int jobC = mygridC.addJob( FunctionalUnitsControl.JOB_MORE_SLEEP );

		System.out.println( "[DEBUG] Waiting MyGrid C receive 4 machines" );

		ConditionWaiter.waitForCondition( new MyGridGumsCondition( mygridC, 4 ) );
		assertEquals( mygridC.getRunningGums().size(), 4 );

		System.out.println( "[DEBUG] Starting MyGrid A" );

		mygridA.start();
		mygridA.setPeer( peerA );
		int jobA = mygridA.addJob( FunctionalUnitsControl.JOB_SLEEP );

		System.out.println( "[DEBUG] Waiting MyGrid A receive 4 machines" );

		ConditionWaiter.waitForCondition( new MyGridGumsCondition( mygridA, 4 ) );
		assertEquals( 4, mygridA.getRunningGums().size() );
		assertEquals( mygridC.getRunningGums().size(), 0 );

		System.out.println( "[DEBUG] Waiting Job finish on MyGrid A" );

		new Thread( new JobWaiterRunnable( jobA, mygridA ) ).start();
		ThreadServices.waitUntilWorkIsDone();
		assertEquals( mygridA.getJobStatus( jobA ), ExecutionStatus.FINISHED );
		assertFalse( mygridC.getJobStatus( jobC ).isFinished() );

		System.out.println( "[DEBUG] Waiting MyGrid C receive 4 machines" );

		ConditionWaiter.waitForCondition( new MyGridGumsCondition( mygridC, 4 ) );
		assertEquals( mygridC.getRunningGums().size(), 4 );

		System.out.println( "[DEBUG] Test finished" );
	}


	public void testNetworkOfFavorsRedistributions2() throws Exception {

		System.out.println( "[DEBUG] Creating Units" );

		CorePeerUnit corePeer = FunctionalUnitsControl.getInstance().createANewCorePeerUnit( "CorePeer" );

		PeerUnit peerA = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerA" );
		PeerUnit peerB = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerB" );
		PeerUnit peerC = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerC" );

		MyGridUnit mygridB = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "MyGridB" );
		MyGridUnit mygridC = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "MyGridC" );

		System.out.println( "[DEBUG] Starting Core Peer" );

		corePeer.start();

		System.out.println( "[DEBUG] Rankings Config" );

		peerA.setCorePeer( corePeer );
		peerB.setCorePeer( corePeer );
		peerC.setCorePeer( corePeer );

		peerA.setId( 11 );
		peerB.setId( 22 );
		peerC.setId( 33 );

		HashMap<PeerUnit,Double> rankings = new HashMap<PeerUnit,Double>();
		rankings.put( peerB, 70.0 );
		rankings.put( peerC, 30.0 );
		peerA.setRankingsFile( rankings );

		System.out.println( "[DEBUG] Starting Peer A" );

		peerA.start();

		System.out.println( "[DEBUG] Starting 10 UserAgents" );

		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 10 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}
		peerA.setGums( userAgentUnits );
		assertEquals( 10, peerA.gumsInState( AllocationStatus.IDLE ).size() );

		System.out.println( "[DEBUG] Starting Peer B and C" );

		peerB.start();
		peerC.start();

		mygridB.start();
		mygridB.setPeer( peerB );

		System.out.println( "[DEBUG] Starting MyGrid B and C" );

		mygridC.start();
		mygridC.setPeer( peerC );

		System.out.println( "[DEBUG] Adding Job at MyGrid C" );

		int jobC = mygridC.addJob( FunctionalUnitsControl.JOB_MORE_SLEEP );

		System.out.println( "[DEBUG] Waiting till MyGrid C has all Gums" );

		ConditionWaiter.waitForCondition( new MyGridGumsCondition( mygridC, 10 ) );
		assertEquals( mygridC.getRunningGums().size(), 10 );
		assertEquals( 10, peerA.gumsInState( AllocationStatus.DONATED ).size() );

		System.out.println( "[DEBUG] Adding Job at MyGrid B" );

		mygridB.addJob( FunctionalUnitsControl.JOB_MORE_SLEEP );

		System.out.println( "[DEBUG] Waiting till MyGrid B has 7 Gums" );

		ConditionWaiter.waitForCondition( new MyGridGumsCondition( mygridB, 7 ) );
		assertEquals( mygridB.getRunningGums().size(), 7 );
		assertFalse( mygridC.getJobStatus( jobC ).isFinished() );
		assertEquals( mygridC.getRunningGums().size(), 3 );

		assertEquals( 10, peerA.gumsInState( AllocationStatus.DONATED ).size() );
	}


	public void testNetworkOfFavorsRedistributions3() throws Exception {

		System.out.println( "[DEBUG] Creating Units" );

		CorePeerUnit corePeer = FunctionalUnitsControl.getInstance().createANewCorePeerUnit( "CorePeer" );

		PeerUnit peerA = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerA" );
		PeerUnit peerB = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerB" );
		PeerUnit peerC = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerC" );

		MyGridUnit mygridB = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "MyGridB" );
		MyGridUnit mygridC = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "MyGridC" );

		System.out.println( "[DEBUG] Starting Core Peer" );

		corePeer.start();

		System.out.println( "[DEBUG] Rankings Config" );

		peerA.setCorePeer( corePeer );
		peerB.setCorePeer( corePeer );
		peerC.setCorePeer( corePeer );

		peerA.setId( 11 );
		peerB.setId( 22 );
		peerC.setId( 33 );

		HashMap<PeerUnit,Double> rankings = new HashMap<PeerUnit,Double>();
		rankings.put( peerB, 50.0 );
		rankings.put( peerC, 50.0 );
		peerA.setRankingsFile( rankings );

		System.out.println( "[DEBUG] Starting Peer A" );

		peerA.start();

		System.out.println( "[DEBUG] Starting 10 UserAgents" );

		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 10 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}
		peerA.setGums( userAgentUnits );
		assertEquals( 10, peerA.gumsInState( AllocationStatus.IDLE ).size() );

		System.out.println( "[DEBUG] Starting Peer B and C" );

		peerB.start();
		peerC.start();

		mygridB.start();
		mygridB.setPeer( peerB );

		System.out.println( "[DEBUG] Starting MyGrid B and C" );

		mygridC.start();
		mygridC.setPeer( peerC );

		System.out.println( "[DEBUG] Adding Job at MyGrid C" );

		int jobC = mygridC.addJob( FunctionalUnitsControl.JOB_MORE_SLEEP );

		System.out.println( "[DEBUG] Waiting till MyGrid C has all Gums" );

		ConditionWaiter.waitForCondition( new MyGridGumsCondition( mygridC, 10 ) );
		assertEquals( mygridC.getRunningGums().size(), 10 );
		assertEquals( 10, peerA.gumsInState( AllocationStatus.DONATED ).size() );

		System.out.println( "[DEBUG] Adding Job at MyGrid B" );

		mygridB.addJob( FunctionalUnitsControl.JOB_MORE_SLEEP );

		System.out.println( "[DEBUG] Waiting till MyGrid B has 5 Gums" );

		ConditionWaiter.waitForCondition( new MyGridGumsCondition( mygridB, 5 ) );
		assertEquals( mygridB.getRunningGums().size(), 5 );
		assertFalse( mygridC.getJobStatus( jobC ).isFinished() );
		assertEquals( mygridC.getRunningGums().size(), 5 );

		assertEquals( 10, peerA.gumsInState( AllocationStatus.DONATED ).size() );
	}
}
