/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test.functionalunits;

import java.io.File;
import java.io.IOException;
import java.net.Inet4Address;
import java.net.UnknownHostException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;
import java.util.Properties;

import org.ourgrid.common.config.CorePeerConfiguration;
import org.ourgrid.common.url.CorePeerURLProvider;
import org.ourgrid.corepeer.ui.UIManager;

/**
 * <code>CorePeerUnit</code> will represents a peer that will be used to
 * execute functional tests of OurGrid. A default creation of a
 * <code>CorePeerUnit</code> will find and use a disponible port, address and
 * any other detail that isn't needed to a test. Also is possible interact
 * between functional tests unit to permit a easy way to create and mantain any
 * test. This class will execute scripts to control the corepeer.
 */
public class CorePeerUnit implements FunctionalTestUnit {

	/** The core peer address. */
	private String address;

	/** This core peer name. */
	private String name;

	private String hostname;

	private int corePeerPort;

	private UIManager uiManager;

	private Properties properties;

	private RemoteTestServices remoteTestServices;


	/**
	 * Default <code>CorePeerUnit</code> constructor that will find a
	 * disponible port, address and configuration to this peer initialization.
	 * 
	 * @param name This CorePeer name. Its only used to identify this
	 *        <code>CorePeerUnit</code>.
	 * @throws IOException Its not possible to create this test control file.
	 */
	CorePeerUnit( String name ) throws Exception {

		corePeerPort = FunctionalUnitsControl.getAvaliablePort();

		this.name = name;

		try {
			hostname = Inet4Address.getLocalHost().getCanonicalHostName();
		} catch ( UnknownHostException e ) {
			hostname = "localhost";
		}

		address = createAddress( hostname, corePeerPort );

		properties = new Properties();
		properties.put( CorePeerConfiguration.PREFIX + CorePeerConfiguration.PROP_NAME, hostname );
		properties
				.put( CorePeerConfiguration.PREFIX + CorePeerConfiguration.PROP_PORT, Integer.toString( corePeerPort ) );
	}


	/**
	 * Create the address to be used by this class.
	 * 
	 * @param name Hostname to be used.
	 * @param port Port to be used.
	 * @return The RMI address of this object.
	 */
	private String createAddress( String name, int port ) {

		return "rmi://" + name + ":" + port;
	}


	/**
	 * @see FunctionalTestUnit#getName()
	 */
	public String getName() {

		return name;
	}


	/**
	 * Get the hostname of the core peer.
	 * 
	 * @return The hostname of the core peer.
	 */
	public String getHostname() {

		return hostname;
	}


	/**
	 * Get the port of the core peer.
	 * 
	 * @return The port of the core peer.
	 */
	public int getPort() {

		return corePeerPort;
	}


	/**
	 * @see FunctionalTestUnit#getDefaultPropertiesFile()
	 */
	public String getDefaultPropertiesFile() {

		return FunctionalUnitsControl.COREPEER_PROPERTIES;
	}


	/**
	 * Gets the address that this corepeer is published.
	 * 
	 * @return The address that this corepeer is published.
	 */
	public String getAddress() {

		return address;
	}


	/**
	 * Starts this <code>CorePeerUnit</code>.
	 * 
	 * @throws IOException If cannot save properties file.
	 */
	public void start() throws Exception {

		if ( isRunning() ) {
			throw new Exception( "Already started" );
		}

		FunctionalUnitsControl.getInstance().registerControlFile( this, properties );
		FunctionalUnitsControl.getInstance().setControlFile( this );

		ProcessBuilder pbuilder = new ProcessBuilder( (new File( "." )).getAbsolutePath() + File.separator
				+ FunctionalUnitsControl.COREPEER_SCRIPT, "start" );
		pbuilder.directory( new File( "." ) );
		Process corePeerProcess = pbuilder.start();
		if ( corePeerProcess.waitFor() != 0 )
			throw new IOException( "CorePeer cannot be started." );

		uiManager = (UIManager) Naming.lookup( address + "/" + CorePeerURLProvider.COREPEER_UI_MANAGER );

		String url = FunctionalTestsURLProvider.getRemoteTestServicesURL( getHostname(), getPort() );
		this.remoteTestServices = (RemoteTestServices) Naming.lookup( url );

		waitUntilWorkIsDone();
		if ( !isRunning() ) {
			throw new Exception( "This CorePeerUnit was not started" );
		}
	}


	/**
	 * Stops this <code>CorePeerUnit</code>.
	 * 
	 * @throws Exception
	 */
	public void stop() throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This CorePeerUnit was not started" );
		}

		Naming.unbind( FunctionalTestsURLProvider.getRemoteTestServicesURL( getHostname(), getPort() ) );
		try {
			uiManager.stopCorePeerService();
		} catch ( UnmarshalException e ) {
			// expected when stopping a service
		}

		remoteTestServices = null;
		uiManager = null;
	}


	/**
	 * Verify if this corepeer is running.
	 * 
	 * @return True if this corepeer is running.
	 * @throws Exception
	 */
	public boolean isRunning() {

		if ( uiManager == null || remoteTestServices == null ) {
			return false;
		}

		try {
			return uiManager.isCorePeerRunning();
		} catch ( RemoteException e ) {
			return false;
		}
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#destroy()
	 */
	public void destroy() {

	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#waitUntilWorkIsDone()
	 */
	public void waitUntilWorkIsDone() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This CorePeerUnit was not started" );
		}

		remoteTestServices.waitUntilServiceHasStopped();
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#stillWaiting()
	 */
	public boolean stillIdle() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This CorePeerUnit was not started" );
		}

		return remoteTestServices.stillIdle();
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#kill()
	 */
	public void kill() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This CorePeerUnit was not started" );
		}

		try {
			remoteTestServices.killService();
		} catch ( UnmarshalException e ) {
		}

		FunctionalUnitsControl.getInstance().unregisterControlFile( this );
	}
}
