/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test.functionalunits;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.ServerSocket;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Random;

import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.util.JavaFileUtil;

/**
 * <code>FunctionalUnitsControl</code> is a singleton to control any logic
 * that will be used by functional tests like the scripts to be called or
 * collections to be controlled.
 */
public class FunctionalUnitsControl {

	/** CorePeer script property. */
	public static final String COREPEER_SCRIPT = "bin/corepeer";

	/** CorePeer properties file. */
	public static final String COREPEER_PROPERTIES = "corepeer.properties";

	/** UserAgent properties file. */
	public static final String USERAGENT_PROPERTIES = "ua.properties";

	/** Suffix of original properties file. */
	private static final String ORIGINAL_SUFFIX = ".orig";

	/** Peer properties file. */
	public static final String PEER_PROPERTIES = "peer.properties";

	/** Peer script property. */
	public static final String PEER_SCRIPT = "bin/peer";

	/** Mygrid properties file. */
	public static final String MYGRID_PROPERTIES = MyGridConfiguration.ROOT_DIR + "/mg.properties";

	/** Mygrid script property. */
	public static final String MYGRID_SCRIPT = "bin/mygrid";

	/** Uaadmin script property. */
	public static final String UAADMIN_SCRIPT = "bin/uaadmin";

	/** Uaadmin deploy dir. */
	public static final String UAADMIN_DEPLOYDIR = System.getProperty( "user.home" ) + File.separator + ".ourgrid";

	/** Storage directory property. */
	public static final String STORAGE_DIRECTORY_PROPERTY = "/tmp/.ogfunctionalstorage";

	/** Where the storage really is. */
	public static final String STORAGE_DIRECTORY = System.getProperty( "user.home" ) + File.separator
			+ ".ogfunctionalstorage";

	/** Simple job execution file. It performs an echo */
	public static final File JOB_ECHO = new File( ClassLoader.getSystemResource( "test/resources/EchoJob.jdf" )
			.getFile() );

	/** Simple job execution file. It performs many many echos */
	public static final File JOB_ECHO_MANY_TASKS = new File( ClassLoader.getSystemResource(
			"test/resources/EchoJob2.jdf" ).getFile() );

	/** Simple job execution file. It performs sleeps */
	public static final File JOB_SLEEP = new File( ClassLoader.getSystemResource( "test/resources/SleepJob.jdf" )
			.getFile() );

	/** Simple job execution file. It performs sleeps more times */
	public static final File JOB_MORE_SLEEP = new File( ClassLoader.getSystemResource(
			"test/resources/MoreSleepJob.jdf" ).getFile() );

	/** Simple job execution file. It performs sleeps more times */
	public static final File JOB_REQUIREMENTS_WINDOWS = new File( ClassLoader.getSystemResource(
			"test/resources/JobWindows.jdf" ).getFile() );

	/** Simple job execution file. It stores a file */
	public static final File JOB_STORE = new File( ClassLoader
			.getSystemResource( "test/resources/SimpleStorageJob.jdf" ).getFile() );

	/** empty sdf */
	public static final File EMPTY_SDF = new File( ClassLoader.getSystemResource( "test/resources/Empty.sdf" )
			.getFile() );

	/** Unique instance of this <code>FunctionalTestControl</code>. */
	private static FunctionalUnitsControl instance;

	/** Map to hold the property file of each <code>FunctionalTestUnit</code>. */
	private Map<FunctionalTestUnit,String> unit2properties = new HashMap<FunctionalTestUnit,String>();


	/**
	 * Returns a avaliable port at this server. This is done by trying to get a
	 * avaliable port to a server socket at the range of 2000 to 30000.
	 * 
	 * @return An avaliable port to use at this host.
	 */
	static int getAvaliablePort() {

		return getAvaliablePort( 2000, 30000 );
	}


	/**
	 * Returns a avaliable port at this server. This is done by trying to get a
	 * avaliable port to a server socket.
	 * 
	 * @param minBound Minimal port value to obtain.
	 * @param maxBound Maximum port value to obtain.
	 * @return An avaliable port to use at this host.
	 */
	static int getAvaliablePort( int minBound, int maxBound ) {

		Random random = new Random();
		int port = minBound + random.nextInt( maxBound - minBound );
		boolean avail = false;
		while ( avail == false ) {
			try {
				ServerSocket serverSocket = new ServerSocket( port );
				serverSocket.close();
				avail = true;
			} catch ( Exception e ) {
				port = minBound + random.nextInt( maxBound - minBound );
			}
		}
		return port;
	}


	/**
	 * Register a control file at the <code>FunctionalUnitsControl</code> to
	 * permit the control of this entity.
	 * 
	 * @param functionalTestUnit Test unit to control.
	 * @param properties Properties of this test unit.
	 * @throws IOException When its not possible to write this properties file.
	 */
	void registerControlFile( FunctionalTestUnit functionalTestUnit, Properties properties ) throws IOException {

		String propertiesFile = functionalTestUnit.getDefaultPropertiesFile() + "." + functionalTestUnit.getName();
		FileOutputStream fos = new FileOutputStream( new File( propertiesFile ) );
		properties.store( fos, "Functional core peer test: " + functionalTestUnit.getName() );
		unit2properties.put( functionalTestUnit, propertiesFile );
	}


	/**
	 * Unregistess a control file at the <code>FunctionalUnitsControl</code>.
	 * 
	 * @param functionalTestUnit Test unit to control.
	 */
	void unregisterControlFile( FunctionalTestUnit functionalTestUnit ) {

		String propertiesFile = functionalTestUnit.getDefaultPropertiesFile() + "." + functionalTestUnit.getName();

		new File( propertiesFile ).delete();
		unit2properties.remove( functionalTestUnit );
	}


	/**
	 * Sets the current property file to be used by this functional test
	 * control.
	 * 
	 * @param functionalTestUnit Functional test control that will hold the
	 *        property file.
	 * @throws IOException An exception is throw when it cannot set a file.
	 */
	synchronized void setControlFile( FunctionalTestUnit functionalTestUnit ) throws IOException {

		String controlFile = unit2properties.get( functionalTestUnit );
		JavaFileUtil.copyFile( controlFile, functionalTestUnit.getDefaultPropertiesFile() );
	}


	/**
	 * Creates the unique instance of this object. It also backup files to be
	 * restorade later.
	 */
	private FunctionalUnitsControl() {

		createBackupPropertiesFile();
	}


	/**
	 * Method to obtain the unique instance of this class.
	 * 
	 * @return The unique instance of this class.
	 */
	public static FunctionalUnitsControl getInstance() {

		notUsingAspectsException();

		if ( instance == null )
			instance = new FunctionalUnitsControl();
		return instance;
	}


	/**
	 * Calls the default <code>CorePeerUnit</code> constructor that will find
	 * an available port, address and configuration to this peer initialization.
	 * 
	 * @param name This CorePeer name. Its only used to identify this
	 *        <code>CorePeerUnit</code>.
	 * @throws Exception Its not possible to create this test control file.
	 */
	public CorePeerUnit createANewCorePeerUnit( String name ) throws Exception {

		return new CorePeerUnit( name );
	}


	/**
	 * Calls the default <code>PeerUnit</code> constructor that will find an
	 * available port, address and configuration to this peer initialization.
	 * 
	 * @param name This Peer name. Its only used to identify this
	 *        <code>PeerUnit</code>.
	 * @throws Exception Its not possible to create this test control file.
	 */
	public PeerUnit createANewPeerUnit( String name ) throws Exception {

		return new PeerUnit( name );
	}


	/**
	 * Calls the default <code>MyGridUnit</code> constructor that will find an
	 * available port, address and configuration to this peer initialization.
	 * 
	 * @param name This MyGrid name. Its only used to identify this
	 *        <code>MyGridUnit</code>.
	 * @throws Exception Its not possible to create this test control file.
	 */
	public MyGridUnit createANewMyGridUnit( String name ) throws Exception {

		return new MyGridUnit( name );
	}


	/**
	 * Creates a number of <code>UserAgentUnit</code> by calling the
	 * defaultconstructor that will find an available port, address and
	 * configuration to this peer initialization.
	 * 
	 * @param numberOfUnits Number of units to be created.
	 * @return The units created
	 * @throws Exception Its not possible to create this test control file.
	 */
	public UserAgentUnit[ ] createANumberOfUserAgentUnits( int numberOfUnits ) throws Exception {

		UserAgentUnit[ ] userAgentUnits = new UserAgentUnit[ numberOfUnits ];
		for ( int i = 0; i < userAgentUnits.length; i++ ) {
			userAgentUnits[i] = new UserAgentUnit( "UserAgentUnit" + i );
		}

		return userAgentUnits;
	}


	/**
	 * Stops and destroys any remaining test units.
	 * 
	 * @throws Exception If any of the units cannot be destroyed.
	 */
	public void destroyAllRemainingUnits() throws Exception {

		Iterator iterator = unit2properties.keySet().iterator();

		while ( iterator.hasNext() ) {
			FunctionalTestUnit functionalTestUnit = (FunctionalTestUnit) iterator.next();

			if ( functionalTestUnit.isRunning() ) { // Do it only if it has not
													// been stopped yet.
				functionalTestUnit.stop();
			}

			functionalTestUnit.destroy();

			String controlFile = unit2properties.get( functionalTestUnit );
			(new File( controlFile )).delete();

			iterator.remove();
		}
	}


	/**
	 * Restores the previous state of this FunctionalTestControl by reseting it
	 * instance and also restoring the original control files.
	 */
	public void reset() {

		restoreControlFiles();
		JavaFileUtil.deleteDir( STORAGE_DIRECTORY );
		instance = null;
	}


	/**
	 * Creates the backup of original properties files.
	 */
	private static void createBackupPropertiesFile() {

		File originalFile;

		originalFile = new File( COREPEER_PROPERTIES );
		createBackupFile( originalFile );

		originalFile = new File( PEER_PROPERTIES );
		createBackupFile( originalFile );

		originalFile = new File( MYGRID_PROPERTIES );
		createBackupFile( originalFile );

	}


	/**
	 * Creates a backup of a property file.
	 * 
	 * @param originalFile Property file to be backuped.
	 */
	private static void createBackupFile( File originalFile ) {

		if ( originalFile.exists() ) {
			String backupFilename = null;
			try {
				backupFilename = originalFile.getAbsolutePath() + ORIGINAL_SUFFIX;
				JavaFileUtil.copyFile( originalFile.getAbsolutePath(), backupFilename );
			} catch ( IOException e ) {
				// Tried our best...
			} finally {
				if ( backupFilename != null ) {
					File backupFile = new File( backupFilename );
					backupFile.deleteOnExit();
				}
			}
		}
	}


	/**
	 * Restores the backup of original properties files.
	 */
	private static void restoreControlFiles() {

		restoreCorePeerControlFile();
		restorePeerControlFile();
		restoreMygridControlFile();
	}


	/**
	 * Restores the original control file of core peer.
	 */
	private static void restoreCorePeerControlFile() {

		File originalFile = new File( COREPEER_PROPERTIES );
		File backupFile = new File( originalFile.getAbsolutePath() + ORIGINAL_SUFFIX );
		restoreBackupFile( originalFile, backupFile );
	}


	/**
	 * Restores the original control file of peer.
	 */
	private static void restorePeerControlFile() {

		File originalFile = new File( PEER_PROPERTIES );
		File backupFile = new File( originalFile.getAbsolutePath() + ORIGINAL_SUFFIX );
		restoreBackupFile( originalFile, backupFile );
	}


	/**
	 * Restores the original control file of peer.
	 */
	private static void restoreMygridControlFile() {

		File originalFile = new File( MYGRID_PROPERTIES );
		File backupFile = new File( originalFile.getAbsolutePath() + ORIGINAL_SUFFIX );
		restoreBackupFile( originalFile, backupFile );
	}


	/**
	 * Restores a backup file.
	 * 
	 * @param originalFile Original file to be restored.
	 * @param backupFile Backup file to be restored.
	 */
	private static void restoreBackupFile( File originalFile, File backupFile ) {

		if ( backupFile.exists() ) {
			try {
				JavaFileUtil.copyFile( backupFile.getAbsolutePath(), originalFile.getAbsolutePath() );
			} catch ( IOException e ) {
				// Tried our best...
			} finally {
				backupFile.delete();
			}
		}
	}


	/**
	 * Contacts every functional unit created and waits until they have blocked
	 * (stopped working). What this method does is it goes around all the units
	 * waiting until they have blocked. When a unit blocks only a external event
	 * may wake up, so it is waiting for this event, After all units have
	 * blocked, we go around them asking if they are still blocked meaning that
	 * no other woke it up since it blocked. When all units block and confirm
	 * that they are still blocked no external event can wake any of them up
	 * (since they are all blocked), meaning that the system is idle.
	 */
	public void waitUntilSystemIsIdle() throws Exception {

		waitUntilEntitiesAreIdle( this.unit2properties.keySet() );
	}


	/**
	 * Contacts given functional units and waits until they have blocked
	 * (stopped working). What this method does is it goes around the units
	 * waiting until they have blocked. When a unit blocks only a external event
	 * may wake up, so it is waiting for this event, After all units have
	 * blocked, we go around them asking if they are still blocked meaning that
	 * no other woke it up since it blocked. When all units block and confirm
	 * that they are still blocked no external event can wake any of them up
	 * (since they are all blocked), meaning that the system is idle.
	 */
	private void waitUntilEntitiesAreIdle( Collection<FunctionalTestUnit> functionalTestUnits ) throws Exception {

		notUsingAspectsException();

		long sleep = 1000;
		boolean systemIdle = false;
		while ( !systemIdle ) {
			// First run, will wait until every unit has stopped.
			for ( FunctionalTestUnit functionalTestUnit : functionalTestUnits ) {
				functionalTestUnit.waitUntilWorkIsDone();
			}

			try {
				Thread.sleep( sleep );
			} catch ( InterruptedException e ) {
			}

			systemIdle = true;

			// Second run, will verify if units did not wake up.
			for ( FunctionalTestUnit functionalTestUnit : functionalTestUnits ) {
				if ( !functionalTestUnit.stillIdle() ) {
					systemIdle = false;
					break;
				}
			}

			try {
				Thread.sleep( sleep );
			} catch ( InterruptedException e ) {
			}

			sleep += sleep * Math.random();
		}
	}


	/**
	 * @throws Exception Always
	 */
	static void notUsingAspectsException() {

		throw new RuntimeException( "This method may only be used after compilation with aspects" );
	}
}
