/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test.functionalunits;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.Inet4Address;
import java.net.UnknownHostException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;
import java.util.Collection;
import java.util.List;
import java.util.Properties;

import org.ourgrid.common.config.MyGridConfiguration;
import org.ourgrid.common.spec.JobSpec;
import org.ourgrid.common.spec.PeerSpec;
import org.ourgrid.common.spec.main.DescriptionFileCompile;
import org.ourgrid.common.url.MyGridURLProvider;
import org.ourgrid.mygrid.scheduler.ExecutionStatus;
import org.ourgrid.mygrid.scheduler.jobmanager.JobEntry;
import org.ourgrid.mygrid.ui.UIManager;

/**
 * <code>MyGridUnit</code> will represents a mygrid that will be used to
 * execute functional tests of OurGrid. A default creation of a
 * <code>MyGridUnit</code> will find and use a disponible port, address and
 * any other detail that isn't needed to a test. Also is possible interact
 * between functional tests unit to permit a easy way to create and mantain any
 * test. This class will execute scripts to control the MyGrid.
 */
public class MyGridUnit implements FunctionalTestUnit {

	private int mygridPort;

	private String name;

	private String hostname;

	private String address;

	private Properties properties;

	private UIManager uiManager;

	private RemoteTestServices remoteTestServices;


	/**
	 * Default <code>MyGridUnit</code> constructor that will find a disponible
	 * port, address and configuration to this MyGrid initialization.
	 * 
	 * @throws Exception
	 */
	MyGridUnit( String name ) throws Exception {

		mygridPort = FunctionalUnitsControl.getAvaliablePort();

		this.name = name;

		try {
			hostname = Inet4Address.getLocalHost().getCanonicalHostName();
		} catch ( UnknownHostException e ) {
			hostname = "localhost";
		}

		address = createAddress( hostname, mygridPort );

		properties = new Properties();
		properties.put( MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_NAME, hostname );
		properties.put( MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_PORT, Integer.toString( mygridPort ) );
		properties.put( MyGridConfiguration.PREFIX + MyGridConfiguration.PROP_PERSISTJOBID, "no" );
	}


	/**
	 * Create the address to be used by this class.
	 * 
	 * @param name Hostname to be used.
	 * @param port Port to be used.
	 * @return The RMI address of this object.
	 */
	private String createAddress( String name, int port ) {

		return "rmi://" + name + ":" + port;
	}


	/**
	 * Gets the address that this peer is published.
	 * 
	 * @return The address that this peer is published.
	 */
	public String getAddress() {

		return address;
	}


	/**
	 * @see FunctionalTestUnit#getName()
	 */
	public String getName() {

		return name;
	}


	/**
	 * Get the hostname of the peer.
	 * 
	 * @return The hostname of the peer.
	 */
	public String getHostname() {

		return hostname;
	}


	/**
	 * Get the port of the mygrid.
	 * 
	 * @return The port of the mygrid.
	 */
	public int getPort() {

		return mygridPort;
	}


	/**
	 * @see FunctionalTestUnit#getDefaultPropertiesFile()
	 */
	public String getDefaultPropertiesFile() {

		return FunctionalUnitsControl.MYGRID_PROPERTIES;
	}


	/**
	 * Set the peer that this MyGrid will contact.
	 * 
	 * @param peer <code>PeerUnit</code> object to be used as test.
	 * @throws Exception
	 */
	public void setPeer( PeerUnit peer ) throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This MyGridUnit was not started" );
		}

		String gdf = createGDF( peer );
		List<PeerSpec> peerSpecs = DescriptionFileCompile.compileGDF( gdf );

		uiManager.setPeers( peerSpecs );
	}


	/**
	 * Creates a GDF to be used to set this peer.
	 * 
	 * @param peer Peer that will be used by this GDF.
	 * @return The GDF filepath.
	 * @throws IOException If its not possible to generate a GDF.
	 */
	private String createGDF( PeerUnit peer ) throws IOException {

		File actualGDF = File.createTempFile( "ourgrid-functional-test-", ".gdf" );
		BufferedWriter writer = new BufferedWriter( new FileWriter( actualGDF ) );

		writer.write( "peer:" );
		writer.newLine();

		writer.write( " label : " + peer.getName() );
		writer.newLine();

		writer.write( " name : " + peer.getHostname() );
		writer.newLine();

		writer.write( " port : " + peer.getPort() );
		writer.newLine();

		writer.close();

		actualGDF.deleteOnExit();

		return actualGDF.getAbsolutePath();
	}


	/**
	 * Starts this <code>MyGridUnit</code>.
	 */
	public void start() throws Exception {

		if ( isRunning() ) {
			throw new Exception( "This MyGridUnit is already started" );
		}

		FunctionalUnitsControl.getInstance().registerControlFile( this, properties );
		FunctionalUnitsControl.getInstance().setControlFile( this );

		ProcessBuilder pbuilder = new ProcessBuilder( (new File( "." )).getAbsolutePath() + File.separator
				+ FunctionalUnitsControl.MYGRID_SCRIPT, "start" );
		pbuilder.directory( new File( "." ) );
		Process peerProcess = pbuilder.start();
		if ( peerProcess.waitFor() != 0 )
			throw new IOException( "Mygrid cannot be started." );

		uiManager = (UIManager) Naming.lookup( address + "/" + MyGridURLProvider.MYGRID_MANAGER );

		String url = FunctionalTestsURLProvider.getRemoteTestServicesURL( getHostname(), getPort() );
		this.remoteTestServices = (RemoteTestServices) Naming.lookup( url );

		waitUntilWorkIsDone();
		if ( !isRunning() ) {
			throw new Exception( "This MyGridUnit was not started" );
		}
	}


	/**
	 * Stops this <code>MyGridUnit</code>.
	 */
	public void stop() throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This MyGridUnit was not started" );
		}

		Naming.unbind( FunctionalTestsURLProvider.getRemoteTestServicesURL( getHostname(), getPort() ) );
		try {
			uiManager.stopMyGridService( false, true );
		} catch ( UnmarshalException e ) {
			// expected when stopping a service
		}

		remoteTestServices = null;
		uiManager = null;
	}


	/**
	 * Verify if this mygrid is running.
	 * 
	 * @return True if this mygrid is running.
	 */
	public boolean isRunning() {

		if ( uiManager == null || remoteTestServices == null ) {
			return false;
		}

		try {
			return uiManager.isMyGridRunning();
		} catch ( RemoteException e ) {
			return false;
		}
	}


	/**
	 * Add a job to this MyGrid.
	 * 
	 * @param job Job to be added to this MyGrid.
	 * @return Specified jobId of this job.
	 */
	public int addJob( File job ) throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This MyGridUnit was not started" );
		}

		JobSpec theJob = DescriptionFileCompile.compileJDF( job.getAbsolutePath() );
		return uiManager.addJob( theJob );
	}


	/**
	 * Verify the status of an job.
	 * 
	 * @param jobId JobId to verify the status.
	 * @return The status of this job.
	 * @throws Exception
	 */
	public ExecutionStatus getJobStatus( int jobId ) throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This MyGridUnit was not started" );
		}

		for ( JobEntry jobEntry : uiManager.getJobs( null ) ) {
			if ( jobEntry.getId() == jobId ) {
				return jobEntry.getStatus();
			}
		}

		throw new Exception( "No such job: " + jobId );
	}


	/**
	 * This method will get the running gums at this MyGrid at a map that
	 * associates each jobId (if any that is running) with a collection of
	 * running gums.
	 * 
	 * @return A map that associates each jobId (if any that is running) with a
	 *         collection of running gums.
	 */
	public Collection getRunningGums() throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This MyGridUnit was not started" );
		}
		return uiManager.getGumSpecs();
	}


	/**
	 * Waits for a job to finish.
	 * 
	 * @param jobid Id of the job to wait.
	 * @throws Exception
	 */
	public void waitForJob( int jobid ) throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This MyGridUnit was not started" );
		}

		uiManager.waitForJob( jobid );
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#destroy()
	 */
	public void destroy() {

	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#waitUntilWorkIsDone()
	 */
	public void waitUntilWorkIsDone() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This MyGridUnit was not started" );
		}

		remoteTestServices.waitUntilServiceHasStopped();
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#stillWaiting()
	 */
	public boolean stillIdle() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This MyGridUnit was not started" );
		}

		return remoteTestServices.stillIdle();
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#kill()
	 */
	public void kill() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This MyGridUnit was not started" );
		}

		try {
			remoteTestServices.killService();
		} catch ( UnmarshalException e ) {
		}

		FunctionalUnitsControl.getInstance().unregisterControlFile( this );
	}
}
