/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test.functionalunits;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Inet4Address;
import java.net.UnknownHostException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Properties;

import org.ourgrid.common.config.Configuration;
import org.ourgrid.common.config.PeerConfiguration;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.spec.RequestSpec.RequestSource;
import org.ourgrid.common.url.PeerURLProvider;
import org.ourgrid.peer.manager.allocation.AllocationStatus;
import org.ourgrid.peer.manager.status.StatusEntry;
import org.ourgrid.peer.ui.UIManager;
import org.ourgrid.test.functionaltests.busywait.ConditionWaiter;
import org.ourgrid.test.functionaltests.busywait.PeerGumsInStateCondition;

/**
 * <code>PeerUnit</code> will represents a peer that will be used to execute
 * functional tests of OurGrid. A default creation of a <code>PeerUnit</code>
 * will find and use a disponible port, address and any other detail that isn't
 * needed to a test. Also is possible interact between functional tests unit to
 * permit a easy way to create and mantain any test. This class will execute
 * scripts to control the peer.
 */
public class PeerUnit implements FunctionalTestUnit {

	private int peerPort;

	private String name;

	private String hostname;

	private String address;

	private Properties properties;

	private UIManager uiManager;

	private UserAgentUnit[ ] currentGums;

	private File rankingsFile;

	private File idFile;

	private Long peerID;

	private RemoteTestServices remoteTestServices;


	/**
	 * Default <code>PeerUnit</code> constructor that will find a disponible
	 * port, address and configuration to this peer initialization.
	 * 
	 * @throws IOException Its not possible to create this test control file.
	 */
	PeerUnit( String name ) throws Exception {

		peerPort = FunctionalUnitsControl.getAvaliablePort();

		this.name = name;

		try {
			hostname = Inet4Address.getLocalHost().getCanonicalHostName();
		} catch ( UnknownHostException e ) {
			hostname = "localhost";
		}

		address = createAddress( hostname, peerPort );

		properties = new Properties();
		properties.put( PeerConfiguration.PREFIX + PeerConfiguration.PROP_NAME, hostname );
		properties.put( PeerConfiguration.PREFIX + PeerConfiguration.PROP_PORT, Integer.toString( peerPort ) );
		properties.put( PeerConfiguration.PREFIX + PeerConfiguration.PROP_LABEL, name );
		properties.setProperty( PeerConfiguration.PREFIX + PeerConfiguration.PROP_JOIN_COMMUNITY, "no" );

		this.rankingsFile = new File( PeerConfiguration.ROOT_DIR + File.separator + hostname + "_" + peerPort
				+ ".rankings.dat" );

		properties.put( PeerConfiguration.PREFIX + PeerConfiguration.PROP_RANKINGFILE, rankingsFile.getAbsolutePath() );

		this.idFile = new File( PeerConfiguration.ROOT_DIR + File.separator + ".peerid-" + hostname + ":" + peerPort );
	}


	/**
	 * Create the address to be used by this class.
	 * 
	 * @param name Hostname to be used.
	 * @param port Port to be used.
	 * @return The RMI address of this object.
	 */
	private String createAddress( String name, int port ) {

		return "rmi://" + name + ":" + port;
	}


	/**
	 * Gets the address that this peer is published.
	 * 
	 * @return The address that this peer is published.
	 */
	public String getAddress() {

		return address;
	}


	/**
	 * @see FunctionalTestUnit#getName()
	 */
	public String getName() {

		return name;
	}


	/**
	 * Get the hostname of the peer.
	 * 
	 * @return The hostname of the peer.
	 */
	public String getHostname() {

		return hostname;
	}


	/**
	 * Get the port of the peer.
	 * 
	 * @return The port of the peer.
	 */
	public int getPort() {

		return peerPort;
	}


	/**
	 * @see FunctionalTestUnit#getDefaultPropertiesFile()
	 */
	public String getDefaultPropertiesFile() {

		return FunctionalUnitsControl.PEER_PROPERTIES;
	}


	/**
	 * Set the core peer that this peer will contact.
	 * 
	 * @param corePeer <code>CorePeerUnit</code> object to be used as test.
	 * @throws Exception
	 */
	public void setCorePeer( CorePeerUnit corePeer ) throws Exception {

		if ( isRunning() ) {
			throw new Exception( "Core peer can only be set before peer is running" );
		}

		properties
				.setProperty( PeerConfiguration.PREFIX + PeerConfiguration.PROP_COREPEER_NAME, corePeer.getHostname() );
		properties.setProperty( PeerConfiguration.PREFIX + PeerConfiguration.PROP_COREPEER_PORT, Integer
				.toString( corePeer.getPort() ) );
		properties.setProperty( PeerConfiguration.PREFIX + PeerConfiguration.PROP_JOIN_COMMUNITY, "yes" );
	}


	// /**
	// * Set the gums to be used by this <code>PeerUnit</code>. This method will
	// block until all gums
	// * are IDLE.
	// *
	// * @param numGums
	// * Number of gums to be setted.
	// *
	// * @return Spec of the gums that have been set.
	// *
	// * @throws Exception
	// */
	// public List<GumSpec> setGum(int numGums) throws Exception {
	// if (!isRunning()) {
	// throw new Exception("This PeerUnit was not started");
	// }
	//        
	// if (gumsBeenSet()) {
	// throw new Exception("Gums have already beem set at this peer");
	// }
	//        
	// ArrayList<GumSpec> myGums = new ArrayList<GumSpec>(numGums);
	// for (int i = 0; i < numGums; i++) {
	// GumSpec myGumspec = new GumSpec();
	// myGumspec.putAttribute(Configuration.PROP_NAME, hostname);
	// myGumspec.putAttribute(Configuration.PROP_PORT,
	// Integer.toString(FunctionalUnitsControl.getAvaliablePort()));
	// myGumspec.putAttribute(UserAgentConfiguration.ATT_TYPE,
	// UserAgentConfiguration.TYPE_UA_LINUX);
	// myGumspec.putAttribute(UserAgentConfiguration.ATT_STORAGE_DIR,
	// FunctionalUnitsControl.STORAGE_DIRECTORY_PROPERTY);
	// myGumspec.putAttribute(UserAgentConfiguration.ATT_STORAGE_SIZE, 500+"");
	// myGums.add(myGumspec);
	// }
	//		
	// String currentSDF = createSDF(myGums);
	//		
	// uaadminInitSDF(currentSDF);
	//			
	// uiManager.setLocalGums(new File(currentSDF).getAbsolutePath());
	//		
	//        
	// currentGums = myGums;
	//		
	// return myGums;
	// }

	/**
	 * Set the gums to be used by this <code>PeerUnit</code>. This method
	 * will block until all gums are IDLE.
	 * 
	 * @param userAgentUnits Units to be set at this peer.
	 * @throws Exception If gums could not ber set.
	 */
	public void setGums( UserAgentUnit... userAgentUnits ) throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This PeerUnit was not started" );
		}

		if ( gumsBeenSet() ) {
			throw new Exception( "Gums have already beem set at this peer" );
		}
		LinkedList<GumSpec> myGums = new LinkedList<GumSpec>();

		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			myGums.add( userAgentUnit.getGumSpec() );
		}

		uiManager.setLocalGums( myGums );

		ConditionWaiter.waitForCondition( new PeerGumsInStateCondition( this, userAgentUnits.length,
				AllocationStatus.IDLE ) );
		currentGums = userAgentUnits;
	}


	/**
	 * Verifies if gums have been set.
	 * 
	 * @return True if this peer has gums set
	 */
	public boolean gumsBeenSet() {

		return currentGums != null;
	}


	/**
	 * Starts this <code>PeerUnit</code>.
	 * 
	 * @throws IOException
	 */
	public void start() throws Exception {

		if ( isRunning() ) {
			throw new Exception( "This PeerUnit was already started" );
		}

		FunctionalUnitsControl.getInstance().registerControlFile( this, properties );
		FunctionalUnitsControl.getInstance().setControlFile( this );

		(new File( Configuration.getInstance( Configuration.PEER ).getProperty( PeerConfiguration.PROP_IDFILE ) ))
				.delete();

		ProcessBuilder pbuilder = new ProcessBuilder( (new File( "." )).getAbsolutePath() + File.separator
				+ FunctionalUnitsControl.PEER_SCRIPT, "start" );

		pbuilder.directory( new File( "." ) );
		Process peerProcess = pbuilder.start();
		if ( peerProcess.waitFor() != 0 )
			throw new IOException( "Peer cannot be started." );

		uiManager = (UIManager) Naming.lookup( address + "/" + PeerURLProvider.PEER_UI_MANAGER );

		String url = FunctionalTestsURLProvider.getRemoteTestServicesURL( getHostname(), getPort() );
		this.remoteTestServices = (RemoteTestServices) Naming.lookup( url );

		waitUntilWorkIsDone();
		if ( !isRunning() ) {
			throw new Exception( "This PeerUnit was not started" );
		}

		if ( this.peerID == null ) {
			this.peerID = loadIDFromFile();
		}
	}


	/**
	 * Get the id value from file
	 * 
	 * @return Id value.
	 * @throws Exception
	 */
	private Long loadIDFromFile() throws Exception {

		ObjectInputStream obj = null;
		try {
			obj = new ObjectInputStream( new FileInputStream( this.idFile ) );
			return (Long) obj.readObject();
		} finally {
			if ( obj != null ) {
				obj.close();
			}
		}
	}


	/**
	 * Stops this <code>PeerUnit</code>.
	 * 
	 * @throws IOException
	 */
	public void stop() throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This PeerUnit was not started" );
		}

		Naming.unbind( FunctionalTestsURLProvider.getRemoteTestServicesURL( getHostname(), getPort() ) );
		try {
			uiManager.stopPeerService();
		} catch ( UnmarshalException e ) {
			// expected when stopping a service
		}

		remoteTestServices = null;
		uiManager = null;
	}


	/**
	 * Unsets gums being used at this PeerUnit.
	 * 
	 * @throws Exception
	 */
	public void unsetGums() throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "Cannot unsetgums on a peer that is not running" );
		}

		// TODO: remove this very ugly sollution
		uiManager.setLocalGums( FunctionalUnitsControl.EMPTY_SDF.getAbsolutePath() );
		currentGums = null;
	}


	/**
	 * Verify if this peer is running.
	 * 
	 * @return True if this peer is running.
	 * @throws IOException
	 */
	public boolean isRunning() {

		if ( uiManager == null || remoteTestServices == null ) {
			return false;
		}

		try {
			return uiManager.isPeerRunning();
		} catch ( RemoteException e ) {
			return false;
		}
	}


	/**
	 * Verify what gums are in a certain state at this peer.
	 * 
	 * @param state State of Gums to be returned
	 * @return The gums that are up at this peer.
	 * @throws Exception
	 */
	public Collection<StatusEntry> gumsInState( AllocationStatus state ) throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This PeerUnit was not started" );
		}

		return uiManager.getGums( state, RequestSource.FROM_LOCAL_BROKER );
	}


	/**
	 * @see FunctionalTestUnit#destroy()
	 */
	public void destroy() {

		this.rankingsFile.delete();
		this.idFile.delete();
	}


	/**
	 * Gets the idfile being used by this test peer.
	 * 
	 * @return ID file.
	 */
	public File getIDFile() {

		return idFile;
	}


	/**
	 * Gets the rankings being used by this test peer.
	 * 
	 * @return ID file.
	 */
	public File getRankingsFile() {

		return rankingsFile;
	}


	/**
	 * Sets the ID value of peer
	 * 
	 * @param peerID new ID
	 */
	public void setId( long peerID ) throws Exception {

		if ( isRunning() ) {
			throw new Exception( "Can only set id berfoe the peer is running" );
		}

		ObjectOutputStream obj = null;

		try {
			obj = new ObjectOutputStream( new FileOutputStream( this.idFile ) );
			obj.writeObject( peerID );
			obj.flush();

			this.peerID = peerID;
		} finally {
			if ( obj != null ) {
				obj.close();
			}
		}

	}


	/**
	 * Sets the ranking file of Network of Favors
	 * 
	 * @param rankings Map containning the relations between Peers and value of
	 *        balance
	 * @throws FileNotFoundException if file does not exists
	 * @throws IOException if occurred a problem while manipulate the file
	 *         object
	 */
	public void setRankingsFile( Map<PeerUnit,Double> rankings ) throws Exception {

		if ( isRunning() ) {
			throw new Exception( "Can only set rankings before the peer is running." );
		}

		Map<ObjectID,Double> realRankingsMap = new HashMap<ObjectID,Double>();

		for ( PeerUnit peerUnit : rankings.keySet() ) {
			realRankingsMap.put( createObjectId( peerUnit ), rankings.get( peerUnit ) );
		}

		ObjectOutputStream obj = null;

		try {
			obj = new ObjectOutputStream( new FileOutputStream( this.rankingsFile ) );
			obj.writeObject( realRankingsMap );
		} finally {
			if ( obj != null ) {
				obj.close();
			}
		}
	}


	/**
	 * Get this peerID
	 * 
	 * @return ID of this peer unit
	 */
	public long getPeerID() {

		return this.peerID;
	}


	/**
	 * Getts the ranking of a given peer.
	 * 
	 * @param peerUnit Peer to get reanking from.
	 * @return Value of the ranking.
	 * @throws Exception
	 */
	public Double getRankingForPeer( PeerUnit peerUnit ) throws RemoteException {

		Map<ObjectID,Double> rankingsMap = uiManager.getPeerBalances();
		Double doubleValue = rankingsMap.get( createObjectId( peerUnit ) );
		return doubleValue;
	}


	/**
	 * Based on a <code>PeerUnit</code> this method will create its respective
	 * <code>ObjectID</code>
	 * 
	 * @param peerUnit Peer unit to create object id from.
	 * @return The <code>ObjectID</code> that represents this peer.
	 */
	private ObjectID createObjectId( PeerUnit peerUnit ) {

		return new ObjectID( peerUnit.getName(), peerUnit.getPeerID(), peerUnit.getAddress() + "/"
				+ PeerURLProvider.REMOTE_ACCESS );
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#waitUntilWorkIsDone()
	 */
	public void waitUntilWorkIsDone() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This PeerUnit was not started" );
		}

		remoteTestServices.waitUntilServiceHasStopped();
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#stillWaiting()
	 */
	public boolean stillIdle() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This PeerUnit was not started" );
		}

		return remoteTestServices.stillIdle();
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#kill()
	 */
	public void kill() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This PeerUnit was not started" );
		}

		try {
			remoteTestServices.killService();
		} catch ( UnmarshalException e ) {
		}

		FunctionalUnitsControl.getInstance().unregisterControlFile( this );
	}
}
