/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test.functionalunits;

import java.io.File;
import java.rmi.Naming;
import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;

import org.ourgrid.peer.manager.allocation.AllocationStatus;
import org.ourgrid.test.FunctionalStressTest;

/**
 * Test for <code>PeerUnit</code>.
 */
public class PeerUnitTest extends TestCase implements FunctionalStressTest {

	protected void tearDown() throws Exception {

		FunctionalUnitsControl.getInstance().destroyAllRemainingUnits();
		FunctionalUnitsControl.getInstance().reset();
	}


	public void testStopAStoppedPeer() throws Exception {

		PeerUnit peer = FunctionalUnitsControl.getInstance().createANewPeerUnit( "test peer 1" );
		assertFalse( peer.isRunning() );

		try {
			peer.stop();
			fail();
		} catch ( Exception e ) {
		}

		peer.destroy();
	}


	public void testTestPeerExecution() throws Exception {

		PeerUnit peer = FunctionalUnitsControl.getInstance().createANewPeerUnit( "test peer 1" );
		assertFalse( peer.isRunning() );
		peer.start();
		try {
			assertNotNull( Naming.lookup( peer.getAddress() ) );
		} catch ( Exception e ) {
			fail( "The core peer is not initialized. Exception: " + e.getMessage() );
		}
		assertTrue( peer.isRunning() );
		peer.stop();
		assertFalse( peer.isRunning() );
		peer.destroy();
	}


	public void testTestPeerExecutionWithSetGums() throws Exception {

		PeerUnit peer = FunctionalUnitsControl.getInstance().createANewPeerUnit( "test peer 1" );
		assertFalse( peer.isRunning() );
		peer.start();
		try {
			assertNotNull( Naming.lookup( peer.getAddress() ) );
		} catch ( Exception e ) {
			fail( "The core peer is not initialized. Exception: " + e.getMessage() );
		}
		assertTrue( peer.isRunning() );

		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 5 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}

		peer.setGums( userAgentUnits );
		assertEquals( 5, peer.gumsInState( AllocationStatus.IDLE ).size() );
		peer.stop();
		assertFalse( peer.isRunning() );
		peer.destroy();
	}


	public void testIfDirectoriesAreDeleted() throws Exception {

		CorePeerUnit testCorePeer = FunctionalUnitsControl.getInstance().createANewCorePeerUnit( "New test core peer" );
		testCorePeer.start();

		PeerUnit peer = FunctionalUnitsControl.getInstance().createANewPeerUnit( "test peer 1" );
		PeerUnit testPeer2 = FunctionalUnitsControl.getInstance().createANewPeerUnit( "test peer 2" );
		peer.setCorePeer( testCorePeer );
		testPeer2.setCorePeer( testCorePeer );

		peer.start();
		assertTrue( peer.getIDFile().exists() );

		testPeer2.start();

		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 5 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}

		peer.setGums( userAgentUnits );
		assertEquals( 5, peer.gumsInState( AllocationStatus.IDLE ).size() );

		MyGridUnit testMyGrid = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "MyGrid1" );
		testMyGrid.start();
		testMyGrid.setPeer( testPeer2 );

		int jobID = testMyGrid.addJob( FunctionalUnitsControl.JOB_STORE );
		testMyGrid.waitForJob( jobID );

		peer.stop();
		assertFalse( peer.isRunning() );

		while ( !peer.getRankingsFile().exists() );

		peer.destroy();

		assertFalse( peer.getIDFile().exists() );
		assertFalse( peer.getRankingsFile().exists() );

		testMyGrid.stop();
		testMyGrid.destroy();
		testCorePeer.stop();
		testCorePeer.destroy();
		testPeer2.stop();
		testPeer2.destroy();

		FunctionalUnitsControl.getInstance().destroyAllRemainingUnits();
		assertFalse( new File( FunctionalUnitsControl.STORAGE_DIRECTORY ).exists() );
	}


	public void testLoadedPeerIDValueFromFile() throws Exception {

		PeerUnit peer = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerA" );

		peer.setId( 1234567890 );
		peer.start();

		System.out.println( "[DEBUG] Verifying PeerID from file " + peer.getIDFile().getAbsolutePath() );
		assertEquals( peer.getPeerID(), 1234567890 );
	}


	public void testLoadedPeerBalancesValueFromFile() throws Exception {

		PeerUnit peerA = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerA" );
		PeerUnit peerB = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerB" );
		PeerUnit peerC = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerC" );

		peerA.setId( 123 );
		peerB.setId( 1234 );
		peerC.setId( 12345 );

		Map<PeerUnit,Double> rankingsMap = new HashMap<PeerUnit,Double>();

		rankingsMap.put( peerB, 12345.0 );
		rankingsMap.put( peerC, 1234.0 );

		peerA.setRankingsFile( rankingsMap );

		peerA.start();
		peerB.start();
		peerC.start();

		assertEquals( peerA.getRankingForPeer( peerB ), 12345.0 );
		assertEquals( peerA.getRankingForPeer( peerC ), 1234.0 );

		PeerUnit nonexistent = FunctionalUnitsControl.getInstance().createANewPeerUnit( "IDoNotExist=(" );
		nonexistent.setId( 30091985 );
		assertNull( peerA.getRankingForPeer( nonexistent ) );
	}


	public void testAnotherSetGums() throws Exception {

		PeerUnit peerA = FunctionalUnitsControl.getInstance().createANewPeerUnit( "PeerA" );
		peerA.start();

		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 5 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}
		peerA.setGums( userAgentUnits );

		assertEquals( peerA.gumsInState( AllocationStatus.IDLE ).size(), 5 );

		peerA.unsetGums();

		assertEquals( peerA.gumsInState( AllocationStatus.CONTACTING ).size(), 0 );
		assertEquals( peerA.gumsInState( AllocationStatus.DONATED ).size(), 0 );
		assertEquals( peerA.gumsInState( AllocationStatus.IDLE ).size(), 0 );
		assertEquals( peerA.gumsInState( AllocationStatus.IN_USE ).size(), 0 );
	}
}
