/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test.functionalunits;

import java.rmi.Naming;

import junit.framework.TestCase;

import org.ourgrid.test.FunctionalStressTest;

public class RemoteTestServicesImplTest extends TestCase implements FunctionalStressTest {

	private MyGridUnit myGridUnit;

	private PeerUnit peerUnit;


	public void setUp() throws Exception {

		this.myGridUnit = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "TestMyGridA" );
		this.myGridUnit.start();
		assertTrue( this.myGridUnit.isRunning() );

		this.peerUnit = FunctionalUnitsControl.getInstance().createANewPeerUnit( "TestPeerA" );
		this.peerUnit.start();
		assertTrue( this.peerUnit.isRunning() );
	}


	public void tearDown() throws Exception {

		FunctionalUnitsControl.getInstance().destroyAllRemainingUnits();

		assertFalse( this.myGridUnit.isRunning() );
		assertFalse( this.peerUnit.isRunning() );

		FunctionalUnitsControl.getInstance().reset();
	}


	public void testIfEverytingIsStartingOk() throws Exception {

		MyGridUnit m = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "M1" );
		PeerUnit p = FunctionalUnitsControl.getInstance().createANewPeerUnit( "P1" );
		CorePeerUnit cp = FunctionalUnitsControl.getInstance().createANewCorePeerUnit( "CP1" );
		UserAgentUnit ua = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 1 )[0];

		m.start();
		p.start();
		cp.start();
		ua.start();

		Naming.lookup( FunctionalTestsURLProvider.getRemoteTestServicesURL( m.getHostname(), m.getPort() ) );
		Naming.lookup( FunctionalTestsURLProvider.getRemoteTestServicesURL( p.getHostname(), p.getPort() ) );
		Naming.lookup( FunctionalTestsURLProvider.getRemoteTestServicesURL( cp.getHostname(), cp.getPort() ) );
		Naming.lookup( FunctionalTestsURLProvider.getRemoteTestServicesURL( ua.getHostname(), ua.getPort() ) );
	}


	public void testWaitUntilOneServiceHasStopped() throws Exception {

		int jobId = myGridUnit.addJob( FunctionalUnitsControl.JOB_SLEEP );
		myGridUnit.waitUntilWorkIsDone();
		assertTrue( myGridUnit.stillIdle() );

		myGridUnit.getJobStatus( jobId );
		assertFalse( myGridUnit.stillIdle() );

		myGridUnit.waitUntilWorkIsDone();
		assertTrue( myGridUnit.stillIdle() );
	}


	public void testIfOneServiceWakesUpAnother() throws Exception {

		myGridUnit.setPeer( peerUnit );

		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 5 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}
		peerUnit.setGums( userAgentUnits );

		FunctionalUnitsControl.getInstance().waitUntilSystemIsIdle();
		assertTrue( myGridUnit.stillIdle() );
		assertTrue( peerUnit.stillIdle() );

		myGridUnit.addJob( FunctionalUnitsControl.JOB_SLEEP );
		assertTrue( !myGridUnit.stillIdle() || !peerUnit.stillIdle() );

		FunctionalUnitsControl.getInstance().waitUntilSystemIsIdle();
		assertTrue( myGridUnit.stillIdle() );
		assertTrue( peerUnit.stillIdle() );
	}


	public void testWaitUntilMultipleServicesHasStopped() throws Exception {

		myGridUnit.setPeer( peerUnit );
		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 5 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}
		peerUnit.setGums( userAgentUnits );

		FunctionalUnitsControl.getInstance().waitUntilSystemIsIdle();
		assertTrue( myGridUnit.stillIdle() );
		assertTrue( peerUnit.stillIdle() );

		int jobId = myGridUnit.addJob( FunctionalUnitsControl.JOB_SLEEP );

		assertFalse( myGridUnit.stillIdle() );
		myGridUnit.waitForJob( jobId );

		assertFalse( myGridUnit.stillIdle() );
		assertFalse( peerUnit.stillIdle() );

		FunctionalUnitsControl.getInstance().waitUntilSystemIsIdle();
		assertTrue( myGridUnit.stillIdle() );
		assertTrue( peerUnit.stillIdle() );
	}


	public void testSimpleWait1() throws Exception {

		UserAgentUnit[ ] userAgentUnits = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 5 );
		for ( UserAgentUnit userAgentUnit : userAgentUnits ) {
			userAgentUnit.start();
		}
		peerUnit.setGums( userAgentUnits );

		FunctionalUnitsControl.getInstance().waitUntilSystemIsIdle();
		assertTrue( peerUnit.stillIdle() );
	}


	public void testSimpleWait2() throws Exception {

		myGridUnit.setPeer( peerUnit );

		FunctionalUnitsControl.getInstance().waitUntilSystemIsIdle();
		assertTrue( myGridUnit.stillIdle() );
	}


	public void testKillService() throws Exception {

		MyGridUnit m = FunctionalUnitsControl.getInstance().createANewMyGridUnit( "M1" );
		PeerUnit p = FunctionalUnitsControl.getInstance().createANewPeerUnit( "P1" );
		CorePeerUnit cp = FunctionalUnitsControl.getInstance().createANewCorePeerUnit( "CP1" );
		UserAgentUnit ua = FunctionalUnitsControl.getInstance().createANumberOfUserAgentUnits( 1 )[0];

		m.start();
		p.start();
		cp.start();
		ua.start();

		m.kill();
		p.kill();
		cp.kill();
		ua.kill();

		assertFalse( m.isRunning() );
		assertFalse( p.isRunning() );
		assertFalse( cp.isRunning() );
		assertFalse( ua.isRunning() );

		try {
			Naming.lookup( FunctionalTestsURLProvider.getRemoteTestServicesURL( m.getHostname(), m.getPort() ) );
			fail();
		} catch ( Exception e ) {
		}

		try {
			Naming.lookup( FunctionalTestsURLProvider.getRemoteTestServicesURL( p.getHostname(), p.getPort() ) );
			fail();
		} catch ( Exception e ) {
		}

		try {
			Naming.lookup( FunctionalTestsURLProvider.getRemoteTestServicesURL( cp.getHostname(), cp.getPort() ) );
			fail();
		} catch ( Exception e ) {
		}

		try {
			Naming.lookup( FunctionalTestsURLProvider.getRemoteTestServicesURL( ua.getHostname(), ua.getPort() ) );
			fail();
		} catch ( Exception e ) {
		}

		try {
			m.stop();
			fail();
		} catch ( Exception e ) {
		}

		try {
			p.stop();
			fail();
		} catch ( Exception e ) {
		}

		try {
			cp.stop();
			fail();
		} catch ( Exception e ) {
		}

		try {
			ua.stop();
			fail();
		} catch ( Exception e ) {
		}
	}
}
