/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test.functionalunits;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.Inet4Address;
import java.net.UnknownHostException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;

import org.ourgrid.common.config.UserAgentConfiguration;
import org.ourgrid.common.gum.GumStatus;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.common.url.UserAgentURLProvider;
import org.ourgrid.common.util.JavaFileUtil;
import org.ourgrid.gridmachine.useragent.ui.UIManager;

/**
 * <code>UserAgentUnit</code> will represents a peer that will be used to
 * execute functional tests of OurGrid. A default creation of a
 * <code>UserAgentUnit</code> will find and use a disponible port, address and
 * any other detail that isn't needed to a test. Also is possible interact
 * between functional tests unit to permit a easy way to create and mantain any
 * test. This class will execute scripts to control the corepeer.
 */
public class UserAgentUnit implements FunctionalTestUnit {

	/** The core peer address. */
	private String address;

	/** This core peer name. */
	private String name;

	private String hostname;

	private int corePeerPort;

	private UIManager uiManager;

	private Properties properties;

	private RemoteTestServices remoteTestServices;

	private GumSpec gumSpec;

	private File sdf;


	/**
	 * Default <code>UserAgentUnit</code> constructor that will find a
	 * disponible port, address and configuration to this peer initialization.
	 * 
	 * @param name This UserAgentUnit name. Its only used to identify this
	 *        <code>UserAgentUnit</code>.
	 * @throws IOException Its not possible to create this test control file.
	 */
	UserAgentUnit( String name ) throws Exception {

		corePeerPort = FunctionalUnitsControl.getAvaliablePort();

		this.name = name;

		try {
			hostname = Inet4Address.getLocalHost().getCanonicalHostName();
		} catch ( UnknownHostException e ) {
			hostname = "localhost";
		}

		address = createAddress( hostname, corePeerPort );

		properties = new Properties();
		properties.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.PROP_NAME, hostname );
		properties.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.PROP_PORT, Integer
				.toString( corePeerPort ) );
		properties.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.ATT_TYPE,
				UserAgentConfiguration.TYPE_UA_LINUX );
		properties.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.ATT_STORAGE_DIR,
				FunctionalUnitsControl.STORAGE_DIRECTORY_PROPERTY );
		properties.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.ATT_STORAGE_SIZE, 500 + "" );
		properties.put( UserAgentConfiguration.PREFIX + UserAgentConfiguration.PROP_IDLENESS_DETECTOR, "no" );

		gumSpec = new GumSpec();
		gumSpec.putAttribute( UserAgentConfiguration.PROP_NAME, hostname );
		gumSpec.putAttribute( UserAgentConfiguration.PROP_PORT, Integer.toString( corePeerPort ) );
		gumSpec.putAttribute( UserAgentConfiguration.ATT_TYPE, UserAgentConfiguration.TYPE_UA_LINUX );
		gumSpec
				.putAttribute( UserAgentConfiguration.ATT_STORAGE_DIR,
						FunctionalUnitsControl.STORAGE_DIRECTORY_PROPERTY );
		gumSpec.putAttribute( UserAgentConfiguration.ATT_STORAGE_SIZE, 500 + "" );

		createSDF();
	}


	private void createSDF() throws IOException {

		sdf = File.createTempFile( "ourgrid-functional-test-", ".sdf" );
		BufferedWriter writer = new BufferedWriter( new FileWriter( sdf ) );

		writer.write( "gumdefaults:" );
		writer.newLine();

		writer.write( " remexec : ssh -x $machine $command " );
		writer.newLine();

		writer.write( " copyto : scp $localfile $machine:$remotefile" );
		writer.newLine();

		writer.write( " copyfrom : scp $machine:$remotefile $localfile" );
		writer.newLine();

		writer.write( " type : ualinux" );
		writer.newLine();

		writer.write( "GUM : " );
		writer.newLine();
		Set keySet = gumSpec.getAttributes().keySet();
		Iterator iterAtts = keySet.iterator();
		while ( iterAtts.hasNext() ) {
			String Att = (String) iterAtts.next();
			writer.write( " " + Att + " : " + gumSpec.getAttribute( Att ) );
			writer.newLine();
		}

		writer.close();

		sdf.deleteOnExit();
	}


	/**
	 * Create the address to be used by this class.
	 * 
	 * @param name Hostname to be used.
	 * @param port Port to be used.
	 * @return The RMI address of this object.
	 */
	private String createAddress( String name, int port ) {

		return "rmi://" + name + ":" + port;
	}


	/**
	 * @see FunctionalTestUnit#getName()
	 */
	public String getName() {

		return name;
	}


	/**
	 * Get the hostname of the core peer.
	 * 
	 * @return The hostname of the core peer.
	 */
	public String getHostname() {

		return hostname;
	}


	/**
	 * Get the port of the core peer.
	 * 
	 * @return The port of the core peer.
	 */
	public int getPort() {

		return corePeerPort;
	}


	/**
	 * @see FunctionalTestUnit#getDefaultPropertiesFile()
	 */
	public String getDefaultPropertiesFile() {

		return FunctionalUnitsControl.USERAGENT_PROPERTIES;
	}


	/**
	 * Gets the address that this corepeer is published.
	 * 
	 * @return The address that this corepeer is published.
	 */
	public String getAddress() {

		return address;
	}


	/**
	 * Get's the <code>GumSpec</code> representing this user agent unit.
	 * 
	 * @return <code>GumSpec</code> representing this user agent unit.
	 */
	public GumSpec getGumSpec() {

		return gumSpec;
	}


	/**
	 * Starts this <code>UserAgentUnit</code>.
	 * 
	 * @throws IOException If cannot save properties file.
	 */
	public void start() throws Exception {

		if ( isRunning() ) {
			throw new Exception( "Already started" );
		}

		FunctionalUnitsControl.getInstance().registerControlFile( this, properties );
		FunctionalUnitsControl.getInstance().setControlFile( this );

		ProcessBuilder pbuilderUaadmin = new ProcessBuilder( (new File( "." )).getAbsolutePath() + File.separator
				+ FunctionalUnitsControl.UAADMIN_SCRIPT, "restart", sdf.getAbsolutePath() );

		pbuilderUaadmin.directory( new File( "." ) );
		Process uaadminProcess = pbuilderUaadmin.start();
		if ( uaadminProcess.waitFor() != 0 )
			throw new IOException( "UserAgent cannot be started." );

		uiManager = (UIManager) Naming.lookup( address + "/" + UserAgentURLProvider.USERAGENT_UIMANAGER );

		String url = FunctionalTestsURLProvider.getRemoteTestServicesURL( getHostname(), getPort() );
		this.remoteTestServices = (RemoteTestServices) Naming.lookup( url );

		waitUntilWorkIsDone();
		if ( !isRunning() ) {
			throw new Exception( "This UserAgentUnit was not started" );
		}
	}


	/**
	 * Stops this <code>UserAgentUnit</code>.
	 * 
	 * @throws Exception
	 */
	public void stop() throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This UserAgentUnit was not started" );
		}

		Naming.unbind( FunctionalTestsURLProvider.getRemoteTestServicesURL( getHostname(), getPort() ) );
		try {
			uiManager.stopUserAgentService();
		} catch ( UnmarshalException e ) {
			// expected when stopping a service
		}

		remoteTestServices = null;
		uiManager = null;
	}


	/**
	 * Verify if this corepeer is running.
	 * 
	 * @return True if this corepeer is running.
	 * @throws Exception
	 */
	public boolean isRunning() {

		if ( uiManager == null || remoteTestServices == null ) {
			return false;
		}

		try {
			uiManager.statusUserAgentService();
			return true;
		} catch ( RemoteException e ) {
			return false;
		}
	}


	/**
	 * Get's the current status of this useragetunit.
	 * 
	 * @return Status of the unit represented by a <code>GumStatus</code>
	 * @throws Exception If unable to acquire status.
	 */
	public GumStatus getGumStatus() throws Exception {

		if ( !isRunning() ) {
			throw new Exception( "This UserAgentUnit was not started" );
		}

		return uiManager.statusUserAgentService();
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#destroy()
	 */
	public void destroy() {

		JavaFileUtil.deleteDir( FunctionalUnitsControl.UAADMIN_DEPLOYDIR + File.separator
				+ gumSpec.getAttribute( GumSpec.ATT_NAME ) + "_" + gumSpec.getAttribute( GumSpec.ATT_PORT ) );
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#waitUntilWorkIsDone()
	 */
	public void waitUntilWorkIsDone() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This UserAgentUnit was not started" );
		}

		remoteTestServices.waitUntilServiceHasStopped();
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#stillWaiting()
	 */
	public boolean stillIdle() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This UserAgentUnit was not started" );
		}

		return remoteTestServices.stillIdle();
	}


	/*
	 * @see org.ourgrid.test.functionalunits.FunctionalTestUnit#kill()
	 */
	public void kill() throws Exception {

		FunctionalUnitsControl.notUsingAspectsException();

		if ( !isRunning() ) {
			throw new Exception( "This UserAgentUnit was not started" );
		}

		try {
			remoteTestServices.killService();
		} catch ( UnmarshalException e ) {
		}

		FunctionalUnitsControl.getInstance().unregisterControlFile( this );
	}
}
