/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test.threadServices;

import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

/**
 * This class represents an object and the threads waiting for it.
 */
public class MonitoredObject {

	/**
	 * The object being monitored (an object that has received a wait call).
	 */
	private Object monitored;

	/**
	 * The threads monitoring the object.
	 */
	private Map<String,Thread> monitoringThreads = Collections.synchronizedMap( new TreeMap<String,Thread>() );


	/**
	 * Constructor.
	 * 
	 * @param o The object being monitored.
	 */
	public MonitoredObject( Object o ) {

		this.monitored = o;

	}


	/**
	 * Constructor.
	 * 
	 * @param o The object being monitored (where a wait was called).
	 * @param t The thread waiting for the object.
	 */
	public MonitoredObject( Object o, Thread t ) {

		this.monitored = o;
		this.monitoringThreads.put( t.getName(), t );
	}


	/**
	 * Verifies if an object is monitored by a certain Thread.
	 * 
	 * @param t The Thread we want to know if it is monitored.
	 * @return true if while t was running, a wait on a certain object was
	 *         called.
	 */
	public boolean isMonitoredBy( Thread t ) {

		return this.monitoringThreads.containsValue( t );
	}


	/**
	 * Adds a certain Thread to the list of threads waiting for an object.
	 * 
	 * @param t The Thread.
	 * @return true, if t was not monitoring this object, and false otherwise.
	 */
	public boolean addThread( Thread t ) {

		Object o = this.monitoringThreads.put( t.getName(), t );
		return o != null;
	}


	/**
	 * Gets the list of threads waiting for an object notify or notifyAll.
	 * 
	 * @return the list of threads waiting for an object notify or notifyAll.
	 */
	public Iterator getMonitoringThreadsIterator() {

		return this.monitoringThreads.values().iterator();
	}


	/**
	 * Gets the Map storing threads waiting for an object notify or notifyAll.
	 * 
	 * @return the Map storing threads waiting for an object notify or
	 *         notifyAll.
	 */
	public Map<String,Thread> getMonitoringThreads() {

		return this.monitoringThreads;
	}


	/**
	 * Gets a String that represents this object.
	 * 
	 * @return the textual representation of this object.
	 */
	public String toString() {

		return "Obj:" + this.monitored + " Threads.size:" + this.monitoringThreads.size();
	}
}
