/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test.threadServices;

/**
 * Class used for tests only. This is a thread that will always run, unless
 * blocked or stopped manually.
 * 
 * @author Flavio Vinicius Diniz de Figueiredo - flaviov@lsd.ufcg.edu.br
 */
public class RunnerWaiterRunnable implements Runnable {

	/**
	 * Avalaible the statuses fot a <code>RunnerWaiterRunnable</code>.
	 */
	public enum Status {
		NOTSTARTED, RUNNING, WAITING, STOPPED
	};

	private boolean mustShutdown;

	private boolean mustWait;

	private Object lock;

	private Status status;

	private Thread runnerThread;


	/**
	 * Builda a new <code>RunnerWaiterRunnable</code>
	 */
	public RunnerWaiterRunnable() {

		this.mustShutdown = false;
		this.mustWait = false;
		this.lock = new Object();
		this.status = Status.NOTSTARTED;
		this.runnerThread = new Thread( this, getClass().getName() );
	}


	/**
	 * Tells this <code>RunnerWaiterRunnable</code> to start running.
	 */
	public synchronized void startRunning() {

		this.status = Status.RUNNING;
		this.runnerThread.start();
	}


	/**
	 * Tells this <code>RunnerWaiterRunnable</code> to wait until told
	 * otherwise.
	 */
	public synchronized void setWaiting() {

		if ( mustWait ) {
			throw new IllegalThreadStateException( "Already waiting" );
		}

		this.mustWait = true;
	}


	/**
	 * Tells this <code>RunnerWaiterRunnable</code> to stop waiting.
	 */
	public void unsetWaiting() {

		synchronized ( lock ) {
			this.mustWait = false;
			lock.notify();
		}
	}


	/**
	 * Informs that this runnable must stop. This method implies on waking the
	 * runnable if it is waiting.
	 */
	public synchronized void setMustStop() {

		if ( this.mustShutdown ) {
			throw new IllegalThreadStateException( "Already shutting down" );
		}

		if ( this.mustWait ) {
			unsetWaiting();
		}

		this.mustShutdown = true;
	}


	/**
	 * Gets the current status of the SleeperWaker.
	 * 
	 * @return The status.
	 */
	public Status getStatus() {

		return this.status;
	}


	/**
	 * Get's the tread that is actually running.
	 * 
	 * @return Thread
	 */
	public Thread getActualRunnerThread() {

		return this.runnerThread;
	}


	public void run() {

		while ( !mustShutdown ) {
			while ( mustWait ) {
				synchronized ( lock ) {
					try {
						this.status = Status.WAITING;
						lock.wait();
					} catch ( InterruptedException e ) {
						e.printStackTrace();
					}
				}
			}
			this.status = Status.RUNNING;
		}

		this.status = Status.STOPPED;
	}
}
