/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test.threadServices;

import org.ourgrid.common.event.Shutdownable;

/**
 * This class is used to perform tests on the threadServices package using
 * several threads.
 */
public class ShutdownableRunnable implements Shutdownable, Runnable {

	/**
	 * A debug variable.
	 */
	public static final boolean DEBUG = false;

	/**
	 * Signals when the scheduler must shutdown
	 */
	protected boolean mustShutdown = false;

	/**
	 * Indicates if there is anything to do for this Runnable.
	 */
	private boolean isThereAnythingToDo = false;

	/**
	 * The thread related with this Runnable.
	 */
	private ShutdownableThread myThread;

	/**
	 * A static counter to identify the threads being created.
	 */
	private static int count = 0;

	/**
	 * Indicates if this Thread also creates other Threads in its run method.
	 */
	private boolean isThereInner = false;


	/**
	 * Constructor of this runnable, creating its Thread object.
	 */
	public ShutdownableRunnable() {

		super();
		myThread = new ShutdownableThread( this, "Shutd" + (count++) );
	}


	/**
	 * Constructor with a boolean indicating if in this Runnable we should
	 * create other Threads.
	 * 
	 * @param withInner Indicates if in this Runnable run method other threads
	 *        should be created.
	 */
	public ShutdownableRunnable( boolean withInner ) {

		super();
		this.isThereInner = withInner;
		myThread = new ShutdownableThread( this, "Shutd" + (count++) );
	}


	/**
	 * Starts the thread related with this Runnable.
	 */
	public void startShutdownable() {

		this.myThread.start();
	}


	/**
	 * Marks the scheduler thread to be shutdown.
	 */
	public synchronized void shutdown() {

		mustShutdown = true;
		notify();
	}


	/**
	 * Prints some messages and creates InnerRunnable objects depending on the
	 * situation.
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run() {

		this.println( "MyRunnable" );
		int k = 0;
		int j = 0;
		this.println( "Must shut:" + mustShutdown );
		while ( !mustShutdown ) {
			// If there is nothing to do, 3 InnerRunnable objects are created
			// and started.
			// Then, <code>thisThread</code> will wait for the <code>this</code>
			// object.
			if ( !isThereAnythingToDo ) {
				if ( isThereInner && j < 3 ) {
					new InnerRunnable( this, "|myInner" + j + "|" );
					j++;
				}
				try {
					synchronized ( this ) {
						if ( !mustShutdown ) {
							this.myThread.waitFor( this );
						}
					}
					this.println( "-->WAITED on thread " + this.myThread + " must shutdown:" + this.mustShutdown );
				} catch ( InterruptedException ie ) {
					this.println( "Thread was interrupted:" + Thread.currentThread() );
				}
			} else {
				System.out.println( "::Thread:" + Thread.currentThread() + " run:" + k );
				if ( k % 5 == 0 ) {
					this.setIsThereAnythingToDo( false );
				}
			}
			k++;
		}
		this.myThread.setState( ShutdownableThread.DEAD );
	}


	/**
	 * Gets the flag value that indicates if there is something to do.
	 * 
	 * @return Returns true if there is anything to do and false, otherwise.
	 */
	public boolean isThereAnythingToDo() {

		return isThereAnythingToDo;
	}


	/**
	 * Sets the flag that indicates if there is something to do.
	 * 
	 * @param isThere The nothingToDo to set.
	 */
	public void setIsThereAnythingToDo( boolean isThere ) {

		synchronized ( this ) {
			this.isThereAnythingToDo = isThere;
			this.notifyAll();
		}
	}


	/**
	 * Gets the thread related with this Runnable.
	 * 
	 * @return Returns the myThread.
	 */
	public ShutdownableThread getMyThread() {

		return myThread;
	}


	/**
	 * General method that prints the string parameter on the standard output.
	 * 
	 * @param str String to be printed.
	 */
	public void println( String str ) {

		if ( DEBUG ) {
			System.out.println( str );
		}
	}
}
