/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.test.threadServices;

import org.ourgrid.common.event.Shutdownable;

/**
 * This class is used just to test the auxiliary classes used by the aspects.
 */
public class ShutdownableThread extends Thread implements Shutdownable {

	/**
	 * A DEBUG constant.
	 */
	public static final boolean DEBUG = false;

	/**
	 * A constant used to represent a Thread not yet created.
	 */
	public static final int NOT_CREATED = 1;

	/**
	 * A Thread that has been created. (Entered the constructor) The isAlive()
	 * method from Thread returns true, if it is in this state and also when the
	 * thread is DEAD.
	 */
	public static final int NEW = 2;

	/**
	 * A Thread that has entered the run method and has not yet left it. For
	 * some situations described in the other states, the thread can stop
	 * running for some time.
	 */
	public static final int RUNNING = 3;

	/**
	 * The Thread has been created and a start method has been called on it.
	 */
	public static final int STARTED = 4;

	/**
	 * A Thread that has leaved the run method. The isAlive() method from Thread
	 * returns true, if it is in this state and also when the thread is NEW.
	 */
	public static final int DEAD = 5;

	/**
	 * A thread is NOT_RUNNABLE when: <br> * Its sleep method is invoked.<br> *
	 * The thread calls the wait method to wait for a specific condition to be
	 * satisfied. <br> * The thread is blocking on I/O. <br>
	 * This variable will be used when we cannot indicate any of these or when
	 * the difference between them does not matter to the application.
	 */
	public static final int NOT_RUNNABLE = 6;

	/**
	 * The thread calls the wait method to wait for a specific condition to be
	 * satisfied. After this call to wait returns, this thread will be probably
	 * running.
	 */
	public static final int NOT_RUNNABLE_WAIT = 7;

	/**
	 * The thread has called the sleep method. After this call returns, the
	 * thread will be running.
	 */
	public static final int NOT_RUNNABLE_SLEEP = 8;

	/**
	 * A notify method has been called on an object whose lock was waited by
	 * this thread. However, we are not sure if this thread will run now.
	 */
	public static final int POSSIBLY_NOTIFIED = 9;

	/**
	 * An int representing this thread state.
	 */
	private int state;

	/**
	 * The Runnable related with this Thread.
	 */
	private Runnable myRunnable;


	/**
	 * Constructor.
	 * 
	 * @param target The Runnable related with this Thread.
	 */
	public ShutdownableThread( Runnable target ) {

		super( target );
		this.myRunnable = target;
		this.setState( ShutdownableThread.NEW );
	}


	/**
	 * Constructor.
	 * 
	 * @param target The Runnable related with this Thread.
	 * @param name The Thread name.
	 */
	public ShutdownableThread( Runnable target, String name ) {

		super( target, name );
		this.myRunnable = target;
		this.setState( ShutdownableThread.NEW );
	}


	/**
	 * Constructor.
	 * 
	 * @param group The Thread group.
	 * @param target The Runnable related with this thread.
	 * @param name The Thread name.
	 */
	public ShutdownableThread( ThreadGroup group, Runnable target, String name ) {

		super( group, target, name );
		this.myRunnable = target;
		this.setState( ShutdownableThread.NEW );
	}


	/**
	 * Invokes shutdown on this thread's runnable, if it is Shutdownable.
	 */
	public void shutdown() {

		if ( this.myRunnable instanceof Shutdownable ) {
			((Shutdownable) this.myRunnable).shutdown();
		}

	}


	/**
	 * Returns the textual representation of this Thread state.
	 * 
	 * @return a String representing the thread state, followed by the code of
	 *         the state. If there is not a string representation for the code,
	 *         just its code is returned.
	 */
	public String getStateString() {

		return getStateString( this.getStateValue() );
	}


	/**
	 * Gets the string representation of a certain state.
	 * 
	 * @param state The state value.
	 * @return a string representing <code>state</code>.
	 */
	public static String getStateString( int state ) {

		switch ( state ) {
			case ShutdownableThread.NOT_CREATED:
				return "NOT_CREATED (State Code:" + ShutdownableThread.NOT_CREATED + ")";

			case ShutdownableThread.NEW:
				return "NEW (State Code:" + ShutdownableThread.NEW + ")";

			case ShutdownableThread.STARTED:
				return "STARTED (State Code:" + ShutdownableThread.STARTED + ")";

			case ShutdownableThread.RUNNING:
				return "RUNNING (State Code:" + ShutdownableThread.RUNNING + ")";

			case ShutdownableThread.NOT_RUNNABLE:
				return "NOT_RUNNABLE (State Code:" + ShutdownableThread.NOT_RUNNABLE + ")";

			case ShutdownableThread.NOT_RUNNABLE_SLEEP:
				return "NOT_RUNNABLE_SLEEP (State Code:" + ShutdownableThread.NOT_RUNNABLE_SLEEP + ")";

			case ShutdownableThread.NOT_RUNNABLE_WAIT:
				return "NOT_RUNNABLE_WAIT (State Code:" + ShutdownableThread.NOT_RUNNABLE_WAIT + ")";

			case ShutdownableThread.POSSIBLY_NOTIFIED:
				return "POSSIBLY_NOTIFIED (State Code:" + ShutdownableThread.POSSIBLY_NOTIFIED + ")";

			case ShutdownableThread.DEAD:
				return "DEAD (State Code:" + ShutdownableThread.DEAD + ")";

			default:
				return "(State Code:" + ShutdownableThread.NOT_CREATED + ")";
		}

	}


	/**
	 * Starts this Thread, changing its state.
	 */
	public void start() {

		this.setState( ShutdownableThread.STARTED );
		super.start();
	}


	/**
	 * Makes this thread wait for an object.
	 * 
	 * @param o The object this thread should wait for.
	 */
	public void waitFor( Object o ) throws InterruptedException {

		synchronized ( o ) {
			this.println( "Will wait for object:" + o );
			this.setState( ShutdownableThread.NOT_RUNNABLE_WAIT );
			o.wait();
		}
	}


	/**
	 * Simply calls the super class run method, changing this thread state
	 * before and after this.
	 */
	public void run() {

		this.println( "Entered here..." );
		this.setState( ShutdownableThread.RUNNING );
		super.run();
		// For most threads there is not the run
		// method for the Thread when there is a runnable.
		this.setState( ShutdownableThread.DEAD );
		this.println( "Leaved here..." );

	}


	/**
	 * Gets the state of a this Thread.
	 * 
	 * @return Returns the state of a this Thread.
	 */
	public synchronized int getStateValue() {

		this.println( "::Returning state:" + state + "  of Thread:" + this );
		return state;
	}


	/**
	 * Sets the state of this Thread.
	 * 
	 * @param state The state to set.
	 */
	protected synchronized void setState( int state ) {

		this.println( "Thread " + this + " will change state:" + state );
		this.state = state;
		this.println( "Thread " + this + " changed state:" + this.state );
	}


	/**
	 * Gets the myRunnable object.
	 * 
	 * @return Returns the myRunnable object.
	 */
	public Runnable getMyRunnable() {

		return myRunnable;
	}


	/**
	 * General method that prints the string parameter on the standard output.
	 * 
	 * @param str String to be printed.
	 */
	public void println( String str ) {

		if ( DEBUG ) {
			System.out.println( str );
		}
	}
}
