/*
 * Copyright (c) 2002-2006 Universidade Federal de Campina Grande
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ourgrid.webstatus;

import java.io.Serializable;
import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Vector;

import org.ourgrid.common.config.LogConfiguration;
import org.ourgrid.common.id.ObjectID;
import org.ourgrid.common.spec.GumSpec;
import org.ourgrid.corepeer.CorePeer;
import org.ourgrid.peer.manager.status.StatusEntry;

/**
 * Description: This is a fake peer, created to join the OurGrid Community and
 * obtain the known peers list from the core peer and then talk to each known
 * peer to get it's status information.
 */
public class WebStatusImpl extends UnicastRemoteObject implements Runnable, WebStatus, Serializable {

	private static final long serialVersionUID = 3L;

	protected static final String WEBSTATUS_SELF_ID = "WebStatus";

	/**
	 * Default core peer url.
	 */
	protected static final String DEFAULT_CORE_PEER_URL = "lula.lsd.ufcg.edu.br";

	/**
	 * Default core peer port.
	 */
	protected static final String DEFAULT_CORE_PEER_PORT = "3085";

	/**
	 * Which port the webstatus will be bound property.
	 */
	protected static final String WEBSTATUS_PORT = "webstatus.port";

	/**
	 * WebStatus url property.
	 */
	protected static final String WEBSTATUS_URL = "webstatus.url";

	/**
	 * The time to update property.
	 */
	protected static final String WEBSTATUS_REFRESH_TIME = "webstatus.refresh.time";

	/**
	 * Core peer port property.
	 */
	protected static final String CORE_PEER_PORT = "corepeer.port";

	/**
	 * Core peer url property,
	 */
	protected static final String CORE_PEER_URL = "corepeer.url";

	/**
	 * Core peer url.
	 */
	private String corepeerURL;

	/**
	 * Known peer's list.
	 */
	private Collection<ObjectID> corePeerList;

	/**
	 * WebStatus peer ID. This is of no use since no other peer will contact the
	 * WebStatus, but it could be an identification to join the community.
	 */
	private ObjectID myPeerID;

	/**
	 * Refresh default time.
	 */
	protected static final int DEFAULT_REFRESH_TIME = 100;

	/**
	 * Default port.
	 */
	protected static final int DEFAULT_WEBSTATUS_PORT = 7676;

	private static final String BIND_OBJECT = "WebStatusImpl";

	/**
	 * The refresh time.
	 */
	private int refreshTime;

	/**
	 * The peer list of information.
	 */
	private Collection<PeerInfoPack> localPeerList;

	/**
	 * The date of snapshot
	 */
	private Date timeSnapshot;

	/**
	 * The status of thread (alive or dead).
	 */
	private boolean active = true;


	public WebStatusImpl() throws RemoteException {

	}


	/**
	 * Constructs a webstatus instance.
	 * 
	 * @param corepeerurl The core peer url.
	 * @param refreshTimeInteger The refresh time.
	 * @param port The port where it will be published.
	 */
	public WebStatusImpl( String corepeerurl, int refreshTimeInteger, int port ) throws MalformedURLException,
		RemoteException, NotBoundException {

		super( port );
		myPeerID = new ObjectID( WEBSTATUS_SELF_ID, new Long( Long.MAX_VALUE ), WEBSTATUS_SELF_ID );
		this.corepeerURL = corepeerurl;
		this.refreshTime = refreshTimeInteger * 100;
		this.localPeerList = new Vector<PeerInfoPack>();
		this.getPeerInfoListRemote();
	}


	/**
	 * Join the OurGrid Community.
	 * 
	 * @throws RemoteException If the corepeer could not be conctacted.
	 * @throws NotBoundException If the corepeer is not online.
	 * @throws MalformedURLException If the corepeer URL is invalid. This should
	 *         never occurs.
	 */
	private void joinCommunity() throws MalformedURLException, RemoteException, NotBoundException {

		try {
			CorePeer corePeer = null;

			corePeer = (CorePeer) Naming.lookup( corepeerURL );
			corePeerList = corePeer.getKnownPeers( myPeerID );

		} catch ( Exception e ) {
			System.out.println( "[INFO] Error when gets the CorePeer Known Peers" );
			e.printStackTrace();
			System.exit( 0 );
		}

	}


	/**
	 * Gets all peer status information from local peer list.
	 * 
	 * @return A collection of PeerInfoPack, one for each known peer. *
	 * @throws MalformedURLException If the corepeer URL is invalid. This should
	 *         never occurs.
	 * @throws RemoteException If the corepeer could not be conctacted.
	 * @throws NotBoundException If the corepeer is not online
	 */
	public Collection<PeerInfoPack> getLocalPeerInfoList() throws MalformedURLException, RemoteException,
		NotBoundException {

		return new LinkedList<PeerInfoPack>( this.localPeerList );
	}


	/**
	 * Gets all peer status information from all known (by the corepeer) peers
	 * remotely.
	 * 
	 * @throws MalformedURLException If the corepeer URL is invalid. This should
	 *         never occurs.
	 * @throws RemoteException If the corepeer could not be contacted.
	 * @throws NotBoundException If the corepeer is not online
	 */
	private void getPeerInfoListRemote() throws MalformedURLException, RemoteException, NotBoundException {

		joinCommunity();

		System.out.println( "Core Peer list with " + corePeerList.size() + " peers." );
		System.out.println( "--------------------------------------------------" );
		for ( ObjectID objID : corePeerList ) {
			System.out.println( "-> " + objID.getName() + " at " + objID.getObjectURL() );
		}
		System.out.println( "--------------------------------------------------" );
		System.out.println( "Local list with " + localPeerList.size() + " peers." );
		System.out.println( "--------------------------------------------------" );
		for ( PeerInfoPack ppack : localPeerList ) {
			System.out.println( "-> " + ppack.getPeerID().getName() + " at " + ppack.getPeerID().getObjectURL() );
		}
		System.out.println( "--------------------------------------------------" );

		Iterator<ObjectID> it = corePeerList.iterator();

		while ( it.hasNext() ) {

			ObjectID peerID = it.next();

			try {

				if ( !peerID.getName().equals( WEBSTATUS_SELF_ID ) ) {

					System.out.print( new Date() + " - Trying to contact " + peerID.getName() + " at "
							+ peerID.getObjectURL() + "..." );
					RemoteUIServices uiServices = new RemoteUIServices( peerID.getObjectURL() );
					PeerInfoPack pack = new PeerInfoPack( peerID, uiServices.getVersion(), uiServices.getEmail(),
							uiServices.getDescription(), uiServices.getUptime(), uiServices.getLatitude(), uiServices
									.getLongitude(), uiServices.getLocalGums(), uiServices.getRemoteGums(), uiServices
									.getDonatedGums(), uiServices.getOwnerGums(), uiServices.getLocalConsumers(),
							uiServices.getRemoteConsumers(), uiServices.getNetworkOfFavors() );
					System.out.println( "Successfully contacted!" );

					addPeer( pack );
				}

			} catch ( Exception e ) {
				System.out.println( "Could not get information from peer: " + peerID.getName() + " Cause: "
						+ e.getCause() );
				removePeer( peerID );
			}

		}

		updateLocalList();
	}


	/**
	 * Update local list, removing a peer that is not into Core Peer list.
	 */
	private void updateLocalList() {

		Iterator<PeerInfoPack> itLocal = this.localPeerList.iterator();

		while ( itLocal.hasNext() ) {
			PeerInfoPack pack = itLocal.next();

			Iterator<ObjectID> itCorePeer = this.corePeerList.iterator();
			boolean itFound = false;

			while ( itCorePeer.hasNext() && !itFound ) {
				ObjectID objId = itCorePeer.next();

				if ( pack.getPeerID().getIdValue().equals( objId.getIdValue() ) ) {
					itFound = true;
				}
			}

			if ( !itFound ) {
				synchronized ( this.localPeerList ) {
					itLocal.remove();
				}
			}

		}

	}


	/**
	 * Remove the peer from local list
	 * 
	 * @param peerID
	 */
	private void removePeer( ObjectID peerID ) {

		try {
			Iterator<PeerInfoPack> it = this.localPeerList.iterator();
			while ( it.hasNext() ) {
				PeerInfoPack ppack = it.next();
				if ( peerID.getIdValue().equals( ppack.getPeerID().getIdValue() ) ) {
					it.remove();
				}
			}
		} catch ( Exception e ) {
		}

	}


	/**
	 * Add a new peer into local list, replacing the old peer info if needed
	 * 
	 * @param pack Object with information about peer
	 */
	private void addPeer( PeerInfoPack pack ) {

		synchronized ( this.localPeerList ) {

			boolean find = false;
			Iterator<PeerInfoPack> it = localPeerList.iterator();

			while ( !find && it.hasNext() ) {

				PeerInfoPack p = it.next();
				if ( p.getPeerID().getIdValue().equals( pack.getPeerID().getIdValue() ) ) {
					it.remove();
				}

			}
			this.localPeerList.add( pack );
		}

	}


	/**
	 * Get the data/time value of last snapshot
	 * 
	 * @return actual time of snapshot
	 */
	public Date getLastTimeOfSnapshot() {

		return this.timeSnapshot;
	}


	/**
	 * Set the flag to define if thread will be run.
	 */
	public void shutdown() {

		this.active = false;
	}


	/**
	 * From time to time, obtains the peer information.
	 */
	public void run() {

		while ( active ) {

			try {
				this.timeSnapshot = new Date();
				getPeerInfoListRemote();
				Thread.sleep( this.refreshTime );
			} catch ( MalformedURLException e ) {
				System.err.println( "Problems when trying to get peer information list! Cause 1: " + e.getCause() );
			} catch ( RemoteException e ) {
				System.err.println( "Problems when trying to get peer information list! Cause 2: " + e.getCause() );
			} catch ( NotBoundException e ) {
				System.err.println( "Problems when trying to get peer information list! Cause 3: " + e.getCause() );
			} catch ( InterruptedException e ) {
				System.err.println( "Problems when sleeping! Cause: " + e.getCause() );
			}

		}

	}


	public static void main( String[ ] args ) {

		// Getting informations from properties parameters
		String corepeerURL = System.getProperty( CORE_PEER_URL );
		String corepeerPort = System.getProperty( CORE_PEER_PORT );
		String refreshTime = System.getProperty( WEBSTATUS_REFRESH_TIME );

		int webStatusImplPortInteger = DEFAULT_WEBSTATUS_PORT;

		System.out.println( "Initializing WebStatus Agent..." );

		int refreshTimeInteger = DEFAULT_REFRESH_TIME;
		if ( refreshTime != null ) {
			try {
				refreshTimeInteger = Integer.parseInt( refreshTime );
			} catch ( Exception e ) {
				System.err.println( "Invalid refresh time... using default value." );
			}
		}

		if ( corepeerPort == null ) {
			corepeerPort = DEFAULT_CORE_PEER_PORT;
		}

		if ( corepeerURL == null ) {
			corepeerURL = DEFAULT_CORE_PEER_URL;
		}

		corepeerURL = "rmi://" + corepeerURL + ":" + corepeerPort + "/CorePeer";

		System.out.println( "Monitoring " + corepeerURL );

		// creating registry
		Registry registry = null;
		try {
			/* Init the Registry using the port number specified by */
			registry = LocateRegistry.createRegistry( webStatusImplPortInteger );
		} catch ( Exception e ) {
			System.err.println( "Oops, problems when trying to create Registry! Cause: " + e.getCause() );
		}

		WebStatusImpl webStatusAgent = null;

		try {
			webStatusAgent = new WebStatusImpl( corepeerURL, refreshTimeInteger, webStatusImplPortInteger );
		} catch ( Exception e ) {
			e.printStackTrace();
			System.err.println( "Oops, problems when trying to create WebStatusImpl! Cause: " + e.getCause() );
		}

		Thread webStatusThread = new Thread( webStatusAgent );
		webStatusThread.start();

		try {
			registry.bind( BIND_OBJECT, webStatusAgent );
		} catch ( Exception e ) {
			System.err.println( "Oops, problems when trying to bind WebStatusImpl! Cause: " + e.getCause() );
		}

		// creating Status logger
		WebStatusLogger statusLogger = new WebStatusLogger( webStatusAgent );
		Thread statusLoggerThread = new Thread( statusLogger );
		statusLoggerThread.start();

		System.out.println( "End of WebStatus webStatusAgent initialization." );

	}


	public void disableWebStatusAgent() {

		this.active = false;
	}

}


/**
 * This class will be in charge of logging the Community status.
 */
class WebStatusLogger implements Runnable {

	private static transient final org.apache.log4j.Logger LOG = org.apache.log4j.Logger
			.getLogger( WebStatusLogger.class );

	/**
	 * The log file.
	 */
	private static final String LOG_FILE = "webstatus.log";

	/**
	 * The xml log file.
	 */
	private static final String LOG_XML_FILE = "webstatus.log.xml";

	/**
	 * The log properties file.
	 */
	private static final String LOG_PROPERTIES_FILE = "webstatus.log.properties";

	/**
	 * The WebStatus that will provide the Community status.
	 */
	private WebStatus webStatus;

	/**
	 * The property "update time".
	 */
	private static final String STATUS_LOGGER_UPDATE_TIME = "status.logger.update.time";

	/**
	 * A default update time in seconds.
	 */
	private static final int DEFAULT_STATUS_LOGGER_UPDATE_TIME = 1800;

	/**
	 * The update time in milliseconds.
	 */
	private int updateTimeInteger;

	/**
	 * A Map of peerIds that are online.
	 */
	private Map<Long,String> peerMap;


	/**
	 * Creates a WebStatusLogger with the WebStatus specified.
	 * 
	 * @param webStatus The WebStatus specified.
	 */
	public WebStatusLogger( WebStatus webStatus ) {

		/** **** Configuring Log4j ******* */
		LogConfiguration.configure( LOG_PROPERTIES_FILE, LOG_FILE, LOG_XML_FILE );

		this.webStatus = webStatus;

		String updateTime = System.getProperty( STATUS_LOGGER_UPDATE_TIME );
		this.updateTimeInteger = DEFAULT_STATUS_LOGGER_UPDATE_TIME;
		if ( updateTime != null ) {
			try {
				updateTimeInteger = Integer.parseInt( updateTime );
			} catch ( Exception e ) {
				System.err.println( "Invalid update time for the WebStatusLogger... using default value." );
			}
		}
		this.updateTimeInteger = updateTimeInteger * 1000;
	}


	/**
	 * Fetches the Community information and writes in the log file.
	 */
	public void run() {

		LOG.info( "WebStatusLogger is on-line." );
		while ( true ) {
			try {

				LOG.info( "Online peers:" );
				Collection<PeerInfoPack> localPeerInfoList = webStatus.getLocalPeerInfoList();

				this.peerMap = createPeerMap( localPeerInfoList );

				Iterator<PeerInfoPack> itLocalPeer = localPeerInfoList.iterator();
				while ( itLocalPeer.hasNext() ) {
					PeerInfoPack peerInfoPack = itLocalPeer.next();
					logPeerInformation( peerInfoPack );
				}

			} catch ( Exception e ) {
				e.printStackTrace();
				LOG.error( "Exception: " + e );
			} finally {
				try {
					Thread.sleep( updateTimeInteger );
				} catch ( InterruptedException e1 ) {
					LOG.error( "Interrupted Exception: " + e1.getMessage() );
					e1.printStackTrace();
				}
			}
		}
	}


	/**
	 * Creates a Map indexed by the PeerIds.
	 * 
	 * @param peerInfoList A Collection of PeerInfoPack
	 * @return A Map indexed by the PeerIds.
	 */
	private Map<Long,String> createPeerMap( Collection<PeerInfoPack> peerInfoList ) {

		Map<Long,String> peerMap = new TreeMap<Long,String>();
		Iterator<PeerInfoPack> peerInfoIt = peerInfoList.iterator();
		while ( peerInfoIt.hasNext() ) {
			PeerInfoPack peerInfoPack = peerInfoIt.next();
			peerMap.put( peerInfoPack.getPeerID().getIdValue(), peerInfoPack.getPeerID().getName() );
		}
		return peerMap;
	}


	/**
	 * Logs the peer information.
	 * 
	 * @param peerInfoPack The peer information.
	 */
	private void logPeerInformation( PeerInfoPack peerInfoPack ) {

		ObjectID peerID = peerInfoPack.getPeerID();
		LOG.info( "Peer: " + peerID.getName() );
		LOG.info( "Local GuMs: " + gumsNames( peerInfoPack.getLocalGums() ) );
		LOG.info( "Community GuMs: " + gumsNames( peerInfoPack.getRemoteGums() ) );
		LOG.info( "Accounting information: " + getAccountingInfo( peerInfoPack ) );

	}


	/**
	 * Returns the GuMs names given a Collection of GuMSpecs.
	 * 
	 * @param communityGuMs The collection of GuMSpecs.
	 * @return A String containing the name of all GuMs in the Collection.
	 */
	private String gumsNames( Collection<StatusEntry> gumsSpecs ) {

		String gumsNames = "[";

		try {
			Iterator<StatusEntry> gumsIterator = gumsSpecs.iterator();
			if ( gumsIterator.hasNext() ) {
				GumSpec guMSpec = (gumsIterator.next()).getGumSpec();
				gumsNames += guMSpec.getAttribute( GumSpec.ATT_NAME );
			}
			while ( gumsIterator.hasNext() ) {
				GumSpec guMSpec = (gumsIterator.next()).getGumSpec();
				gumsNames += "," + guMSpec.getAttribute( GumSpec.ATT_NAME );
			}
		} catch ( Exception e ) {
			// it occours if a name cannot be resolved
		}

		gumsNames += "]";
		return gumsNames;
	}


	private String getAccountingInfo( PeerInfoPack p ) {

		List<String> accountingInfoList = new LinkedList<String>();

		boolean exception = false;

		try {
			Map<ObjectID,Double> peerBalances = p.getNetworkOfFavors();
			System.out.println( "### Peer: " + p.getPeerID().getName() );
			System.out.println( "### Bal.: " + peerBalances );
			Iterator<ObjectID> itPeers = peerBalances.keySet().iterator();
			while ( itPeers.hasNext() ) {
				Object ob = itPeers.next();
				System.out.println( "Class: " + ob.getClass() );
				ObjectID objID = itPeers.next();
				String oneAccountingInfo = getOneAccountingInfo( objID.getIdValue(), peerBalances );
				if ( !oneAccountingInfo.equals( "" ) ) {
					accountingInfoList.add( oneAccountingInfo );
				}
			}
		} catch ( Exception e ) {
			// It occours if a peer balance cannot be resolved
			exception = true;
		}

		String accountingInfo = "[";
		Iterator<String> accountingInfoListIt = accountingInfoList.iterator();
		if ( accountingInfoListIt.hasNext() ) {
			accountingInfo += accountingInfoListIt.next();
			while ( accountingInfoListIt.hasNext() ) {
				accountingInfo += "," + accountingInfoListIt.next();
			}
			accountingInfo += "]";
		} else {
			accountingInfo = "This peer has no accounting information!";
		}

		if ( exception ) {
			accountingInfo = "Error when gets the peer balances";
		}

		return accountingInfo;

	}


	/**
	 * Returns one piece of information about the accounting according to the
	 * peerId.
	 * 
	 * @param peerId The peerId.
	 * @param peerBalances The peerBalances.
	 * @return One piece of information about the accounting according to the
	 *         peerId. If the peer is not online, an empty String will be
	 *         returned.
	 */
	private String getOneAccountingInfo( Long idValue, Map<ObjectID,Double> peerBalances ) {

		String oneAccountingInfo = "";
		Double peerBalance = peerBalances.get( idValue );
		String peerName = peerMap.get( idValue );
		if ( peerName != null ) {
			oneAccountingInfo += peerName + " = " + peerBalance;
		}
		return oneAccountingInfo;
	}
}
